/**
  *  \file arch/unix/cmdargs.cc
  *  \brief Platform-specific option parsing
  *
  *  These are in a separate file due to linking dependencies.
  *  Putting this in platform.cc would reference util/cmdline.cc
  *  each time it is used (backwards dependency in linking).
  *  This is not a problem when cmdargs.cc is separate: a file
  *  which calls Platform::initOptions most likely also calls
  *  the cmdline.cc module, hence the required symbols are available.
  */

#define _POSIX_SOURCE           // putenv
#define _BSD_SOURCE             // setenv
#define _XOPEN_SOURCE           // putenv. Hooray. It's POSIX, the manpage
                                // says it's POSIX, but <stdlib.h> declares it
                                // only for XOPEN/SVID. Need I say that they
                                // forgot the const, too?

#include <stdlib.h>
#include "arch/unix/platform.h"
#include "defs.h"
#include "util/nls.h"
#include "util/cmdline.h"
#include "cpluslib/misc.h"
#ifdef HAVE_SDL
#  include "SDL.h"
#endif

/** Handle -display option. Unfortunately, SDL has no better way to do
    this. */
static void
optSetDisplay(StandardCommandLineParser& p, const StandardCommandLineParser::Option&)
{
#if defined(HAVE_PUTENV)
    /* POSIX way. When platform doesn't have putenv, SDL emulates it */
    string_t value = string_t("DISPLAY=") + p.getArg();
    putenv(const_cast<char*>(value.c_str()));
#elif defined(HAVE_SETENV)
    /* BSD way */
    setenv("DISPLAY", p.getArg(), 1);
#endif
}

static const StandardCommandLineParser::Option options[] = {
    { " -display", optSetDisplay, 0, true, N_(" -display DPY       X display to use (if applicable)") }
};

void
Platform::initOptions(StandardCommandLineParser& par)
{
    par.addOptionSet(::options, countof(::options));
}
