/**
  *  \file arch/unix/dirs.cc
  *  \brief Unix Directory Reading
  */

// half-hearted attempt to get this compiled with -ansi
#define _POSIX_C_SOURCE 2
#define _XOPEN_SOURCE 600

#include <list>
#include <cerrno>
#include <cstring>
#include <unistd.h>
#include <sys/stat.h>
#include "fileio/dirs.h"
#include "cpluslib/assert.h"
#include "util/except.h"

namespace {
    // Unix directory implementation.
    class UnixDirectory : public Directory {
        class Entry;
        class Enum;
        string_t dir;
        friend class Enum;
     public:
        UnixDirectory(string_t aname);
        ~UnixDirectory();
        Ptr<DirectoryEntry> getEntryByName(string_t name);
        Ptr<Enumeration<Ptr<DirectoryEntry> > > getEntries();
        Ptr<Directory> getParent();
        string_t getName();
    };

    // Unix root directory enumerator.
    class UnixRootEnum : public Enumeration<Ptr<Directory> > {
        bool gotit;
     public:
        UnixRootEnum();
        ~UnixRootEnum();
        bool hasMoreElements();
        Ptr<Directory> getNextElement();
    };
}

/** Test whether file name is absolute.
    \returns true iff file name is absolute, that is, it has
    the same meaning on this computer no matter what the current
    working directory is. */
bool
Directory::isAbsoluteName(string_t name)
{
    return (name.length() && name[0] == PATH_SEPARATOR);
}

/** Test whether character is path separator.
    \returns true iff c is a path separator char (that is, something
    that can not occur in a file "basename") */
bool
Directory::isSeparatorChar(char c)
{
    return c == PATH_SEPARATOR;
}

/** Make file name from two components.
    \param dir     working directory
    \param name    file name
    \returns file name that names the file that would be accessed when
    we were in directory /dir/ and opened /name/. In particular, if
    /name/ is absolute, /dir/ is ignored. The returned file name is
    canonical. */
string_t
Directory::makeFileName(string_t dir, string_t name)
{
    if (isAbsoluteName(name))
        return getCanonicalName(name);
    if (dir.length() && !isSeparatorChar(dir[dir.length()-1]))
        dir += PATH_SEPARATOR;
    return getCanonicalName(dir + name);
}

/** Canonicalize directory name. Generates the canonical name for the
    specified file name/directory. It does so by removing redundant
    components (e.g. in Unix notation, `foo/./bar/../x' is the same
    as `foo/x').

    getCanonicalName("") returns the notation for the "current
    directory". */
string_t
Directory::getCanonicalName(string_t dir)
{
    if (!dir.length())
        return ".";

    bool is_absolute;
    int  up_count = 0;
    std::list<string_t> out;
    if (dir[0] == PATH_SEPARATOR) {
        is_absolute = true;
        dir.erase(0, 1);
    } else {
        is_absolute = false;
    }

    string_t::size_type pos = 0;
    while (pos < dir.length()) {
        string_t::size_type next = dir.find(PATH_SEPARATOR, pos);
        if (next == string_t::npos)
            next = dir.length();
        if (next == pos) {
            /* ignore consecutive slashes */
        } else if (next == pos+1 && dir[pos] == '.') {
            /* ignore "." components */
        } else if (next == pos+2 && dir.compare(pos, 2, "..", 2) == 0) {
            /* ".." component */
            if (!out.empty())
                out.pop_back();
            else
                ++up_count;
        } else {
            out.push_back(string_t(dir, pos, next-pos));
        }
        if (next >= dir.length())
            break;
        pos = next+1;
    }

    /* now generate output */
    string_t s;
    if (is_absolute) {
        s += PATH_SEPARATOR;
    } else {
        while (up_count--) {
            s += "..";
            s += PATH_SEPARATOR;
        }
    }
    for (std::list<string_t>::iterator i = out.begin(); i != out.end(); ++i) {
        s += *i;
        s += PATH_SEPARATOR;
    }
    if (s.length() > 1)
        s.erase(s.length() - 1);
    return s;
}

/** Get absolute name. Returns the canonical, absolute name for the
    specified directory.
    \post isAbsoluteName(result) */
string_t
Directory::getAbsoluteName(string_t dir)
{
    if (!isAbsoluteName(dir)) {
        string_t result = makeFileName(getWorkingDirectory(), dir);
        ASSERT(isAbsoluteName(result));
        return result;
    } else {
        return getCanonicalName(dir);
    }
}

/** Get file name part ("basename") of a name.
    This can be used to split a file name into pieces.
    \see getDirName
    \inv getAbsoluteName(dir) == getAbsoluteName(makeFileName(getDirName(dir), getFileName(dir))) */
string_t
Directory::getFileName(string_t dir)
{
    string_t::size_type d = dir.rfind(PATH_SEPARATOR);
    if (d != string_t::npos)
        return dir.substr(d+1);
    else
        return dir;
}

/** Get directory name part of a name. \see getFileName */
string_t
Directory::getDirName(string_t dir)
{
    string_t::size_type d = dir.rfind(PATH_SEPARATOR);
    if (d == 0)
        /* `/file' */
        return string_t(1, PATH_SEPARATOR);
    else if (d != string_t::npos)
        /* `dir/file' */
        return dir.substr(0, d);
    else
        /* `file' */
        return ".";
}

/** Get current working directory.
    \post isAbsoluteName(result)
    \throws FileIOException on error */
string_t
Directory::getWorkingDirectory()
{
    // FIXME: to handle symlink-awareness of some shells, we should
    // handle the "PWD" environment variable, too. Something like
    //    if ($PWD exists && $PWD and "." refer to same inode)
    //    then return $PWD
    //    else get physical directory name
    // Lusers expect `cd symlink; cat ../file' to work.
    // OTOH, only few programs actually *do* support that consistently
    // (if at all possible):
    //   $ cd cpluslib<RET>
    //   $ head ../acdefs<TAB>
    //   acdefs.h   acdefs.hin   acdefs.mak    acdefs.min
    //   $ head ../acdefs.hin<RET>
    //   head: ../acdefs.hin: No such file or directory
    //   $ cd ..<RET>
    //   $ head acdefs.hin<RET>
    //   .....
    std::size_t len = 0;
    char* buffer;
    
    while (1) {
        len += PATH_MAX;
        buffer = new char[len];
        if (getcwd(buffer, len))
            break;
        if (errno != ERANGE)
            throw FileIOException(perrorStr("getWorkingDirectory"));
        delete[] buffer;
    }

    string_t result(buffer);
    delete[] buffer;
    return result;
}

/** Create directory object for named directory. */
Ptr<Directory>
Directory::openDir(string_t dir)
{
    return new UnixDirectory(dir);
}

/** Get list of root directories ("drives"). For systems that support
    multiple roots, this returns an enumeration of those. For other
    systems, this may return a list of convenient start paths. */
Ptr<Enumeration<Ptr<Directory> > >
Directory::getRoots()
{
    return new UnixRootEnum();
}

//////////////////////////////////////////////////////////////////////

/**
  *  \class UnixDirectory
  *  \brief Directory on File System
  *
  *  This class represents an actual file system directory.
  */

/** File System Directory Entry. Implementation of DirectoryEntry
    for file system directories. */
class UnixDirectory::Entry : public DirectoryEntry {
    friend class UnixDirectory;
    friend class UnixDirectory::Enum;
    Ptr<UnixDirectory> parent;
    Ptr<Entry> link;
 protected:
    void getInfo();
 public:
    Entry(Ptr<UnixDirectory> fsd, string_t aname);
    Ptr<Directory> openDir();
    Ptr<Stream>    openFile(int mode);
    Ptr<Directory> openParent();
};

UnixDirectory::Entry::Entry(Ptr<UnixDirectory> fsd, string_t aname)
    : DirectoryEntry(aname), parent(fsd), link(0)
{ }

Ptr<Directory>
UnixDirectory::Entry::openDir()
{
    return new UnixDirectory(name);
}

Ptr<Stream>
UnixDirectory::Entry::openFile(int mode)
{
    return Stream::openFile(name, mode);
}

Ptr<Directory>
UnixDirectory::Entry::openParent()
{
    return parent;
}

void
UnixDirectory::Entry::getInfo()
{
    struct stat st;
    if (lstat(name.c_str(), &st) != 0)
        return;

    FileType ft = S_ISREG(st.st_mode) ? FT_FILE
        : S_ISDIR(st.st_mode) ? FT_DIRECTORY
        : FileType('X');
    setInfo(st.st_size, ft);
}

/** File System Directory Enumerator. Implementation of the directory
    enumeration for UnixDirectory. */
class UnixDirectory::Enum : public Enumeration<Ptr<DirectoryEntry> >
{
    Ptr<UnixDirectory> dir;           ///< The directory we're using
    Ptr<UnixDirectory::Entry> entry;  ///< Current entry
    DIR* dirp;                        ///< Current position. 0 when done iterating
    void next();
 public:
    Enum(Ptr<UnixDirectory> adir);
    ~Enum();
    bool hasMoreElements();
    Ptr<DirectoryEntry> getNextElement();
};

UnixDirectory::Enum::Enum(Ptr<UnixDirectory> adir)
    : dir (adir),
      entry (0)
{
    dirp = opendir(dir->dir.c_str());
    if (!dirp)
        throw FileOpenException(perrorStr(dir->dir));
    next();
}

UnixDirectory::Enum::~Enum()
{
    if (dirp)
        closedir(dirp);
}

/** Advance to next entry. Sets entry to the next directory entry. */
void
UnixDirectory::Enum::next()
{
    ASSERT(dirp);

    struct dirent* de;
    do
        de = readdir(dirp);
    while (de && (strcmp(de->d_name, ".") == 0 || strcmp(de->d_name, "..") == 0));
    if (!de) {
        closedir(dirp);
        entry = 0;
        dirp = 0;
    } else {
        entry = new UnixDirectory::Entry(dir, Directory::makeFileName(dir->dir, de->d_name));
    }
}

bool
UnixDirectory::Enum::hasMoreElements()
{
    return entry.ptr();
}

Ptr<DirectoryEntry>
UnixDirectory::Enum::getNextElement()
{
    ASSERT(entry.ptr());
    Ptr<DirectoryEntry> ret = entry;
    next();
    return ret;
}

/** Constructor. Construct a UnixDirectory for the
    directory specified by aname. */
UnixDirectory::UnixDirectory(string_t aname)
    : dir(getCanonicalName(aname))
{ }

UnixDirectory::~UnixDirectory()
{ }

Ptr<DirectoryEntry>
UnixDirectory::getEntryByName(string_t name)
{
    return new Entry(this, makeFileName(dir, name));
}

Ptr<Enumeration<Ptr<DirectoryEntry> > >
UnixDirectory::getEntries()
{
    return new Enum(this);
}

Ptr<Directory>
UnixDirectory::getParent()
{
    string_t my_abs = getAbsoluteName(dir);
    if (my_abs.length() <= 1)
        return 0;
    else
        return new UnixDirectory(makeFileName(my_abs, ".."));
}

string_t
UnixDirectory::getName()
{
    return dir;
}

/****************************** UnixRootEnum *****************************/

/** \class UnixRootEnum
    Enumerates Unix root directories.

    \todo right now, this returns only "/". It should return a list of
    mount points, plus the user's home */
UnixRootEnum::UnixRootEnum()
    : gotit(false)
{ }

UnixRootEnum::~UnixRootEnum()
{ }

bool
UnixRootEnum::hasMoreElements()
{
    return !gotit;
}

Ptr<Directory>
UnixRootEnum::getNextElement()
{
    gotit = true;
    return new UnixDirectory("/");
}
