/**
  *  \file arch/win32-bcc55/hacklink.cc
  *
  *  Borland tries to be clever when linking, and links in all modules
  *  with static initializers, even when they are not referenced
  *  otherwise. This program attempts to emulate the Unix linker
  *  semantics. Simply modify the normal command lines to start
  *  with "hacklink".
  *
  *  Step 1: "hacklink tlib foo.lib +a +b +c" resp.
  *  "hacklink mklib tlib foo.lib a b c" are modified so that they
  *  generate special-format library files (actually, just lists of
  *  filenames).
  *
  *  Step 2: "hacklink bcc32 prog.obj foo.lib" will generate a command
  *  line for bcc32 which contains the right things.
  *
  *  I know this code is ugly. In particular, we use C I/O all the time,
  *  the C++ library is just a PITA for binary I/O.
  */

#include <cstdio>
#include <map>
#include <string>
#include <vector>
#include <cctype>

using namespace std;

#define TMPFILE "hacklink.$$$"

#define VERBOSE if(1);else fprintf

typedef unsigned char uint8;

static bool
strcaseeq(const char* a, const char* b)
{
    while (*a && *b && toupper(*a) == toupper(*b))
        ++a, ++b;
    return *a==0 && *b==0;
}

map<string, bool> symbols;
vector<const char*> search_path;

class ObjReader {
 public:
    ObjReader() { }
    virtual bool process(uint8 type, const char* data, size_t size) = 0;
    virtual void pubdef(string name) = 0;
    virtual ~ObjReader() { }
    void read(const char* fn);
};

/* read an object file and mark symbols */
class Marker : public ObjReader {
 public:
    Marker() { }
    ~Marker() { }
    bool process(uint8 type, const char* data, size_t size);
    void pubdef(string name);
};

class Scanner : public ObjReader {
    string current;
 public:
    Scanner() { }
    ~Scanner() { }
    map<string, string> symtab;
    bool process(uint8 type, const char* data, size_t size);
    void pubdef(string name);
    void doRead(const char* fn);
};

int
getindex(const char*& p)
{
    uint8 x = *p++;
    if (x & 0x80)
        return ((x & 0x7F) << 8) + (uint8)*p++;
    else
        return x;
}

void
ObjReader::read(const char* fn)
{
    FILE* fp = fopen(fn, "rb");
    if (!fp) {
        perror(fn);
        return;
    }

    char* buffer = new char[128];
    size_t bufsize = 128;

    while (1) {
        uint8 tmp[3];
        if (fread(tmp, 1, 3, fp) != 3) {
            fprintf(stderr, "%s: too short (header)\n", fn);
            break;
        }
        size_t size = 256*tmp[2] + tmp[1];
        if (size >= bufsize) {
            delete[] buffer;
            while (size >= bufsize)
                bufsize *= 2;
            VERBOSE(stderr, "BUFSIZE: %u\n", bufsize);
            buffer = new char[bufsize];
        }
        if (fread(buffer, 1, size, fp) != size) {
            fprintf(stderr, "%s: too short (data)\n", fn);
            break;
        }
        if (tmp[0] == 0x91)
            printf("%s: PUBDEF32 record!\n", fn);
        if (tmp[0] == 0x90) {
            /* PUBDEF */
            const char* p = buffer, *end = buffer+size-1;
            int group = getindex(p);
            getindex(p);
            if (group == 0)
                getindex(p);
            while (p < end) {
                string name(p+1, (unsigned char) p[0]);
                p += name.length()+1+2;
                getindex(p);
                pubdef(name);
            }
            if (p > end)
                printf("%s: PUBDEF record overflown\n", fn);
        } else {
            buffer[size] = 0;
            if (process(tmp[0], buffer, size-1) || tmp[0] == 0x8A || tmp[0] == 0x8B)
                break;
        }
    }
    delete[] buffer;
    fclose(fp);
}

void
Scanner::doRead(const char* fn)
{
    VERBOSE(stderr, "SCANNING: %s\n", fn);
    current = fn;
    read(fn);
}

void
Scanner::pubdef(string name)
{
    symtab[name] = string(current);
}

bool
Scanner::process(uint8 type, const char* data, size_t /*size*/)
{
    if (type == 0xB0 || type == 0xB1)
        /* COMDEF */
        symtab[string(data+1, (unsigned char)data[0])] = string(current);

    if (type == 0xA0 || type == 0xA1)
        /* LEDATA. Assume all symbol definitions come before */
        return true;
    return false;
}

void
Marker::pubdef(string name)
{
    symbols[name] = true;
}

bool
Marker::process(uint8 type, const char* data, size_t size)
{
    if (type == 0xB0 || type == 0xB1)
        /* COMDEF */
        symbols[string(data+1, (unsigned char)data[0])] = true;
    
    if (type == 0x8C || type == 0x8D) {
        /* EXTDEF */
        const char* end = data + size;
        while (data < end) {
            string name(data+1, (unsigned char)data[0]);
            data += name.length() + 1;
            if ((unsigned char)*data & 0x80)
                data += 2;
            else
                ++data;
            if (symbols.find(name) == symbols.end())
                symbols[name] = false;
        }
    }

    if (type == 0xA0)
        /* LEDATA. Assume all symbol definitions come before */
        return true;
    return false;
}

FILE*
checklib(const char* fn, FILE* fp)
{
    char buf[7];
    (void) fn;
    VERBOSE(stderr, "CHECKLIB: %s\n", fn);
    if (fread(buf, 1, 7, fp) == 7 && memcmp(buf, "hacklib", 7) == 0) {
        int i;
        while ((i = fgetc(fp) != '\n') && i != EOF)
            ;
        VERBOSE(stderr, "IS-A-HACKLIB: %s\n", fn);
        return fp;
    }
    /*printf("%s: file was found but is not a hacklib\n", fn);*/
    VERBOSE(stderr, "NOT-A-HACKLIB: %s\n", fn);
    fclose(fp);
    return 0;
}

FILE*
libsearch(const char* lib)
{
    FILE* fp = fopen(lib, "rb");
    if (fp)
        return checklib(lib, fp);

    for (unsigned i = 0; i < search_path.size(); ++i) {
        string name = search_path[i];
        name += '\\';
        name += lib;
        fp = fopen(name.c_str(), "rb");
        if (fp)
            return checklib(name.c_str(), fp);
    }
    return 0;
}

void
readlib(FILE* lib, FILE* file)
{
    char name[1024];
    Scanner scan;
    map<string,bool> linked;
    
    while (fgets(name, sizeof name, lib)) {
        int i = strlen(name);
        if (i && name[i-1] == '\n')
            name[--i] = 0;
        linked[string(name)] = false;
        scan.doRead(name);
    }
    bool done = true;
    while (done) {
        done = false;
        for (map<string,bool>::const_iterator i = symbols.begin(); i != symbols.end(); ++i) {
            if (!i->second) {
                string name = i->first;
                map<string,string>::const_iterator j = scan.symtab.find(name);
                if (j != scan.symtab.end()) {
                    printf("Using %s\n", j->second.c_str());
                    linked[string(j->second)] = true;
                    fprintf(file, "\"%s\" ", j->second.c_str());
                    Marker().read(j->second.c_str());
                    done = true;
                    if (!symbols[name])
                        fprintf(stderr, "internal error: %s failed to define %s\n",
                                j->second.c_str(), name.c_str());
                    break;
                }
            }
        }
    }
    for (map<string,bool>::const_iterator i = linked.begin(); i != linked.end(); ++i)
        if (!i->second)
            printf("Not linking %s\n", i->first.c_str());
    VERBOSE(stderr, "DONE: LIB\n");
}

static void
doCompiler(char** argv)
{
    FILE* fp = fopen(TMPFILE, "w");
    if (!fp) {
        perror(TMPFILE);
        exit(1);
    }

    for (int i = 2; argv[i]; ++i) {
        if (argv[i][0] == '-') {
            VERBOSE(stderr, "OPT: %s\n", argv[i]);
            if (argv[i][1] == 'l')
                search_path.push_back(argv[i]+2);
            fprintf(fp, "\"%s\" ", argv[i]);
        } else {
            const char* p = argv[i], *q;
            while ((q = strpbrk(p+1, ":\\/.")))
                p = q;
            if (strcaseeq(p, ".obj")) {
                VERBOSE(stderr, "OBJ: %s\n", argv[i]);
                fprintf(fp, "\"%s\" ", argv[i]);
                Marker().read(argv[i]);
            } else if (strcaseeq(p, ".lib")) {
                VERBOSE(stderr, "LIB: %s\n", argv[i]);
                FILE* lib = libsearch(argv[i]);
                if (!lib) {
                    fprintf(fp, "\"%s\" ", argv[i]);
                } else {
                    readlib(lib, fp);
                    VERBOSE(stderr, "CLOSING %p\n", (void*)lib);
                    fclose(lib);
                }
            } else {
                VERBOSE(stderr, "OTHER: %s\n", argv[i]);
            }
        }
    }
    fclose(fp);

    string command = argv[1];
    command += " @" TMPFILE;
    printf("> %s\n", command.c_str());
    int i = system(command.c_str());
    if (i)
        printf("=> exit code is %d\n", i);

    remove(TMPFILE);
}

static void
doLibrarian(char** argv, bool is_tlib)
{
    /* tlib lib +foo +bar
       mklib tlib lib +foo +bar */
    int lib_index = (is_tlib ? 2 : 3);
    if (!argv[lib_index]) {
        fprintf(stderr, "no library name specified\n");
        exit(1);
    }
    FILE* fp = fopen(argv[lib_index], "wb");
    if (!fp) {
        perror(argv[lib_index]);
        exit(1);
    }
    fputs("hacklib\n", fp);
    for (int i = lib_index+1; argv[i]; ++i) {
        char* name = argv[i];
        if (is_tlib)
            ++name;
        fprintf(fp, "%s\n", name);
    }
    fclose(fp);
}

int
main(int argc, char** argv)
{
    if (argc < 2) {
        printf("usage: %s <compiler command line>\n", argv[0]);
        return 1;
    }

    /* check personality */
    char* prog = argv[1];
    if (strstr(prog, "tlib"))
        doLibrarian(argv, true);
    else if (strstr(prog, "mklib"))
        doLibrarian(argv, false);
    else
        doCompiler(argv);
}
