/**
  *  \file arch/win32-bcc55/output.cc
  *  \brief Output of Console Messages
  *
  *  It seems to be impossible under Win32 to write a program that
  *  remains attached to a console if run from one, and not opens
  *  a console when run from the GUI. So we can't just write the
  *  help messages etc to stdout. This displays the current console
  *  contents in a resizeable window.
  */

#include <iostream>
#include <cstdlib>
#include <windows.h>
#include "arch/win32-bcc55/platform.h"
#include "util/console.h"

#undef USE_LISTBOX

static const char class_name[] = "Console Window";

static HWND window, button, textfield;
static POINT mintracksize;
static bool done;

static void
complain(const char* title)
{
    LPVOID lpMsgBuf;

    FormatMessage(
                  FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                  NULL,
                  GetLastError(),
                  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
                  (LPTSTR) &lpMsgBuf,
                  0,
                  NULL
                  );

    // Display the string.
    MessageBox( NULL, (const char*) lpMsgBuf, title, MB_OK|MB_ICONINFORMATION );

    // Free the buffer.
    LocalFree( lpMsgBuf );
}

static void
updateGeometry(bool first)
{
    /* manual geometry management:
       ,-----------------------.
       | ,-------------------. |
       | |                   | |
       | |                   | |
       | |                   | |
       | |                   | |
       | |                   | |
       | `-------------------' |
       |                  [ok] |
       `-----------------------'
      The frame is 5 units, likewise the distance between the
      button and the listbox. The button is 50x15 units (standard
      size). */
    LONG u = GetDialogBaseUnits();
    int  x = LOWORD(u), y = HIWORD(u);
    int  xspace = 5 * x / 4;
    int  yspace = 5 * y / 8;
    int  button_wi = 50 * x / 4;
    int  button_he = 15 * y / 8;

    RECT r, rb, rl;

    mintracksize.x = 2*button_wi;
    mintracksize.y = button_he * 3;

    GetClientRect(window, &r);
    if (first && r.right - r.left < mintracksize.x || r.bottom - r.top < mintracksize.y) {
        SetWindowPos(window, 0, r.left, r.top, 2*mintracksize.x, 2*mintracksize.y,
                     SWP_NOZORDER);
        GetClientRect(window, &r);
    }

    rb.top    = r.bottom - yspace - button_he;
    rb.left   = r.right - xspace - button_wi;
    rb.bottom = r.bottom - yspace;
    rb.right  = r.right - xspace;

    rl.top    = r.top + yspace;
    rl.left   = r.left + xspace;
    rl.bottom = rb.top - yspace;
    rl.right  = r.right - xspace;

    SetWindowPos(button, 0, rb.left, rb.top, rb.right-rb.left,
                 rb.bottom-rb.top, SWP_NOACTIVATE | SWP_NOZORDER);
    SetWindowPos(textfield, 0, rl.left, rl.top, rl.right-rl.left,
                 rl.bottom-rl.top, SWP_NOACTIVATE | SWP_NOZORDER);
}

LRESULT CALLBACK
WindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg) {
     case WM_COMMAND:
        if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) {
            done = true;
            return 0;
        }
        break;
     case WM_GETMINMAXINFO:
        LPMINMAXINFO(lParam)->ptMinTrackSize = mintracksize;
        return 0;
     case WM_SIZE:
        updateGeometry(false);
        return 0;
     case WM_CLOSE:
        done = true;
        return 0;
//      case WM_KEYDOWN:
//         if (lParam == VK_ESC) {
//             done = true;
//             return 0;
//         }
//         break;
    }
    return DefDlgProc(hwnd, uMsg, wParam, lParam);
}

static bool window_registered = false;

void
Platform::showOutput()
{
    HINSTANCE inst = GetModuleHandle(0);
    if (!window_registered) {
        WNDCLASS wc;
        wc.style         = 0;
        wc.lpfnWndProc   = WindowProc;
        wc.cbClsExtra    = 0;
        wc.cbWndExtra    = DLGWINDOWEXTRA;
        wc.hInstance     = inst;
        wc.hIcon         = 0;
        wc.hCursor       = LoadCursor(0, IDC_ARROW);
        wc.hbrBackground = HBRUSH(COLOR_WINDOW);
        wc.lpszMenuName  = 0;
        wc.lpszClassName = class_name;
        if (!RegisterClass(&wc)) {
            complain ("Unable to register window class");
            return;
        }
        window_registered = true;
    }
    window = CreateWindow(class_name,
                          "Output",
                          WS_OVERLAPPEDWINDOW | DS_3DLOOK,
                          CW_USEDEFAULT, CW_USEDEFAULT,
                          CW_USEDEFAULT, CW_USEDEFAULT,
                          0,// parent
                          0,
                          inst,
                          0);
    if (!window) {
        complain("Unable to create window");
        return;
    }
    button = CreateWindow("BUTTON",
                          "&Close",
                          WS_CHILD | BS_DEFPUSHBUTTON | WS_TABSTOP,
                          0, 0, 1, 1,
                          window,
                          HMENU(IDOK),
                          inst,
                          0);
#ifdef USE_LISTBOX
    textfield = CreateWindow("LISTBOX",
                             "",
                             WS_CHILD | LBS_NOSEL | LBS_NOINTEGRALHEIGHT | WS_TABSTOP | WS_BORDER,
                             0, 0, 1, 1,
                             window,
                             HMENU(0),
                             inst,
                             0);
#else
    textfield = CreateWindow("EDIT",
                             "",
                             WS_CHILD | ES_AUTOHSCROLL | ES_AUTOVSCROLL | WS_TABSTOP
                             | ES_MULTILINE | WS_BORDER | ES_READONLY,
                             0, 0, 1, 1,
                             window,
                             HMENU(0),
                             inst,
                             0);
#endif
    if (!textfield || !button) {
        complain("Unable to create widgets");
        DestroyWindow(window);
        window = textfield = button = 0;
        return;
    }

    SendMessage(window, WM_INITDIALOG, (WPARAM) button, 0);

    /* fill listbox with text */
#ifdef USE_LISTBOX
    for (int i = console.getScrollbackFill(); i > 0; --i) {
        const Console::Message& m = console.getLine(i-1);
        SendMessage(textfield, LB_ADDSTRING, 0, (LPARAM) m.text.c_str());
    }
#else
    {
        string_t s;
        for (int i = console.getScrollbackFill(); i > 0; --i) {
            const Console::Message& m = console.getLine(i-1);
            s += m.text;
            s += "\r\n";
        }
        SendMessage(textfield, WM_SETTEXT, 0, (LPARAM) s.c_str());
        SendMessage(textfield, EM_SETSEL, 0, 0);
    }
#endif
    SendMessage(textfield, WM_SETFONT, (WPARAM) GetStockObject(ANSI_FIXED_FONT),
                MAKELPARAM(0, 0));
    updateGeometry(true);
    ShowWindow(button, SW_SHOW);
    ShowWindow(textfield, SW_SHOW);
    ShowWindow(window, SW_SHOW);

    MSG msg;
    done = false;
    while (!done && GetMessage(&msg, NULL, 0, 0)) {
        if (!IsDialogMessage(window, &msg)) {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
    }
    DestroyWindow(window);
}
