/**
  *  \file client/act-planet.cc
  *  \brief Planet Actions
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/actions.h"
#include "client/dlg-cargo.h"
#include "client/dlg-fcode.h"
#include "client/dlg-list.h"
#include "game/fcode.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "ui/commands.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "game/player.h"

/** Friendly code change, UI action. */
void
doPlanetFCodeChange(GPlanet& planet)
{
    GFCodeList list(GFCodeList::getInstance(), planet);
    WFCodeWindow dialog(_("Change Friendly Code"), list);
    dialog.setFCode(planet.getFCode());
    ui_root->centerWidget(dialog);
    ui_root->add(dialog);
    int rv = ui_root->run();
    ui_root->close(dialog);
    if (rv == cm_Enter) {
        planet.setFCode(dialog.getFCode());
        getDisplayedTurn().doUpdates();
    }
}

/** Cargo transfer to/from planet, UI action.
    \param planet    invoking planet
    \param unload    true if 'u' key, false if 'c' key. These have no
                     difference other than the heading of the target
                     selection window. */
void
doPlanetTransfer(GPlanet& planet, bool unload)
{
    GGameTurn& trn = getDisplayedTurn();

    GObjectList list;
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        GShip& ship = trn.getShip(i);
        if (ship.isExistant() && ship.getOwner() == planet.getOwner() && ship.getPos() == planet.getPos())
            list.addObject(trn.getShipType(), ship);
    }
    if (list.getSize() == 0) {
        messageBox(_("There's none of our ships orbiting this planet."),
                   _("Cargo Transfer"));
        return;
    }
    WObjectSelectionList wlist(0, 15, 320, list);
    GObjectId chosen = wlist.doStandardDialog(unload
                                              ? _("Unload ship...")
                                              : _("Transfer cargo to..."),
                                              0 /* FIXME */);
    if (chosen.isValid())
        doPlanetTransferFor(planet, chosen);
}

/** Planet cargo transfer to ship, UI action.
    \param planet  invoking planet
    \param chosen  target. Must be a ship. */
void
doPlanetTransferFor(GPlanet& planet, GObjectId chosen)
{
    Ptr<GCargoContainer> left  = new GPlanetTransfer(planet);
    Ptr<GCargoContainer> right = new GShipTransfer(*checked_cast<GShip*>(&chosen.getObject()));

    doCargoTransfer(left, right);
    getDisplayedTurn().doUpdates();
}

/** Starbase mission change, UI action. */
void
doBaseChangeMission(GPlanet& planet)
{
    UIStringList data;
    for (int i = 0; i <= 6; ++i)
        data.add(i, GPlanet::getBaseMissionName(i));

    UIStandardListbox lb(0, 7, 300, data);
    lb.scrollToKey(planet.getBaseMission());
    if (lb.doStandardListbox(_("Starbase Mission"), 0 /* FIXME: help id */)) {
        planet.setBaseMission(lb.getKey());
        getDisplayedTurn().doUpdates();
    }
}

/** Starbase shipyard order, UI action.
    \param planet the starbase we're talking about.
    \param pe     triggering event, used for binding the menu. */
void
doBaseFixRecycle(GPlanet& planet, const UIEvent& pe)
{
    GStarbaseFixRecycleAction action(planet);
    UIStringList d;
    if (action.canFix())
        d.add(sya_Fix, _("Fix (repair) a ship"));
    if (action.canRecycle())
        d.add(sya_Recycle, _("Recycle a ship"));
    if (planet.getBaseShipyardOrder() == sya_Recycle)
        d.add(sya_None, _("Cancel `recycle' order"));
    if (planet.getBaseShipyardOrder() == sya_Fix)
        d.add(sya_None, _("Cancel `fix' order"));

    if (d.getCount() == 0) {
        messageBox(_("There is no ship at this starbase which could be fixed or recycled."),
                   _("Shipyard"));
        return;
    }

    UIStandardListbox lb(0, 1, 200, d);
    if (!lb.doStandardMenu(pe))
        return;
    switch (int cmd = lb.getKey()) {
     case sya_Recycle:
     case sya_Fix:
        {
            GObjectList list;
            action.enumerateShipsFor(cmd, list);
            list.postprocessUser(true);

            const char* title = (cmd == sya_Recycle ? _("Recycle ship") : _("Fix ship"));
            GObjectId chosen = WObjectSelectionList(0, -1, 300, list).doStandardDialog(title, 0 /* FIXME */);
            if (chosen.isValid())
                action.setAction(cmd, chosen.getId());
        }
        break;
     case sya_None:
        action.setAction(sya_None, 0);
    }
    getDisplayedTurn().doUpdates();
}

/*********************** GStarbaseFixRecycleAction ***********************/

/** \class GStarbaseFixRecycleAction
    \brief Starbase shipyard order rule class.

    This class encapsulates all rules for a Fix/Recycle action. */

/** Initialize. */
GStarbaseFixRecycleAction::GStarbaseFixRecycleAction(GPlanet& the_planet)
    : num_own(0), num_foreign(0), the_planet(the_planet)
{
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        GShip& s = getDisplayedTurn().getShip(i);
        if (s.isExistant() && s.getPos() == the_planet.getPos()) {
            if (s.getOwner() == the_planet.getOwner())
                ++num_own;
            else
                ++num_foreign;
        }
    }
}

/** Check whether this starbase can do a "fix" command. */
bool
GStarbaseFixRecycleAction::canFix() const
{
    return num_own + num_foreign != 0;
}

/** Check whether this starbase can do a "recycle" command. */
bool
GStarbaseFixRecycleAction::canRecycle() const
{
    return num_own != 0;
}

/** Enumerate ships for a particular action.
    \param action desired action (sya_Fix, sya_Recycle).
    \param list   output list, should be empty. */
void
GStarbaseFixRecycleAction::enumerateShipsFor(int action, GObjectList& list)
{
    if (action == sya_None)
        return;
    ASSERT(action == sya_Recycle || action == sya_Fix);
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        GShip& s = getDisplayedTurn().getShip(i);
        if (s.isExistant() && s.getPos() == the_planet.getPos())
            if (s.getOwner() == the_planet.getOwner() || action == sya_Fix)
                list.addObject(getDisplayedTurn().getShipType(), s);
    }
}

bool
GStarbaseFixRecycleAction::setAction(int action, int ship_id)
{
    if (action == sya_None) {
        ship_id = 0;
    } else {
        if (ship_id <= 0 || ship_id >= NUM_SHIPS)
            return false;
        if (action != sya_Fix && action != sya_Recycle)
            return false;

        GShip& the_ship = getDisplayedTurn().getShip(ship_id);
        if (!the_ship.isExistant() || the_ship.getPos() != the_planet.getPos())
            return false;
        if (action != sya_Fix && the_ship.getOwner() != the_planet.getOwner())
            return false;
    }
    the_planet.setBaseShipyardOrder(action, ship_id);
    return true;
}

/*********************** GStarbaseTechUpgradeAction **********************/

/** \class GStarbaseTechUpgradeAction
    \brief Tech Upgrade rule class.

    This class encapsulates the rules for tech upgrades, and billing
    there-of. */

/** Cost for upgrading from Tech 1. Indexed by tech level, i.e.
    TECH_COSTS[4] = cost for upgrading from 1 to 4. */
static const int TECH_COSTS[] = {
    0, 0, 100, 300, 600, 1000, 1500, 2100, 2800, 3600, 4500
};

/** Constructor. Initializes the object and allocates resources.
    \pre !the_planet.getCargoHandler().isFacilityInUse(pf_TechLevels)
    \pre the_planet.hasBase()
    \param the_planet the planet we're talking about
    \param container  the cargo container through which we are billing
                      the transaction. This should be a planet; this
                      code assumes addCargo() never fails. */
GStarbaseTechUpgradeAction::GStarbaseTechUpgradeAction(GPlanet& the_planet,
                                                       GCargoContainer& container)
    : the_planet(the_planet), container(container),
      reserved_sup(0), reserved_mc(0), rest(0)
{
    for (int i = 0; i < 4; ++i)
        new_tech[i] = min_tech[i] = the_planet.getBaseTech(TTechLevel(i));
    the_planet.getCargoHandler().incUseCount();
    the_planet.getCargoHandler().acquireFacility(GPlanet::pf_TechLevels);
    ASSERT(getCost() == 0);
}

GStarbaseTechUpgradeAction::~GStarbaseTechUpgradeAction()
{
    GCargo& h = the_planet.getCargoHandler();

    h.releaseFacility(GPlanet::pf_TechLevels);
    h.decUseCount();
}

/** Check whether the user may raise his tech level to the specified
    value. Only checks rules, not resources. */
bool
GStarbaseTechUpgradeAction::isValid(int index, int new_val) const
{
    ASSERT(index >= 0 && index < 4);
    if (new_val < getMinTech(index) || new_val > getMaxTech(index))
        return false;
    return true;
}

/** Get maximum allowed tech level for this index. */
int
GStarbaseTechUpgradeAction::getMaxTech(int index) const
{
    if (allowRegistered(the_planet.getOwner()))
        return 10;
    else
        return 6;
}

/** Set tech level to specified value. This will not actually update
    the base, it will only register the request and update the cost
    accordingly. The actual update happens in commit().
    \return true iff request was honored (see isValid(int,int)). */
bool
GStarbaseTechUpgradeAction::setTech(int index, int new_val)
{
    if (isValid(index, new_val)) {
        new_tech[index] = new_val;
        changeAllocation(getCost());
        return true;
    } else {
        return false;
    }
}

/** Get accumulated cost so far. May be negative. */
int
GStarbaseTechUpgradeAction::getCost() const
{
    int cost = 0;
    for (int i = 0; i < 4; ++i)
        cost += TECH_COSTS[new_tech[i]] - TECH_COSTS[the_planet.getBaseTech(TTechLevel(i))];
    return cost;
}

/** True iff this transaction is valid and can be commited. */
bool
GStarbaseTechUpgradeAction::isValid() const
{
    return rest == 0;
}

/** Commit transaction. Caller must call commit() on the
    GCargoContainer to actually bill the costs. */
void
GStarbaseTechUpgradeAction::commit()
{
    ASSERT(isValid());
    for (int i = 0; i < 4; ++i)
        the_planet.setBaseTech(TTechLevel(i), new_tech[i]);
    container.mustCommit();
}

/** Cancel transaction. Caller should call cancel() on the
    GCargoContainer to actually cancel the cargo transaction. */
void
GStarbaseTechUpgradeAction::cancel()
{
    changeAllocation(0);
    container.mustCancel();
}

/** Change amount of mc allocated to this transaction. */
void
GStarbaseTechUpgradeAction::changeAllocation(int amount)
{
    /* give back everything we have. */
    container.addCargo(el_Supplies, reserved_sup, false);
    container.addCargo(el_Money,    reserved_mc,  false);

    /* (re-)acquire stuff. */
    reserved_mc = container.removeCargo(el_Money, amount, true);
    amount -= reserved_mc;
    reserved_sup = container.removeCargo(el_Supplies, amount, true);
    amount -= reserved_sup;
    rest = amount;
}
