/**
  *  \file client/chart.cc
  *  \brief Starchart Drawing
  *
  *  This module contains the routines to draw the starchart. It is a
  *  more-or-less straightforward port of the PCC 1.0.17 starchart
  *  code.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <climits>
#include "client/chart.h"
#include "client/marks.h"
#include "cpluslib/bitops.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/player.h"
#include "game/ship.h"
#include "gfx/clip.h"
#include "gfx/gfxcompl.h"
#include "gfx/palette.h"
#include "game/formula.h"

static const int
    SCRingRadius     = 3,       ///< Radius of planet ring, in ly.
    SCMaxRingRadius  = 6,       ///< Max.\ radius of scaled planet ring, in pixels.
    SCCrossRadius    = 6,       ///< Size of starbase cross, in ly.
    SCMaxCrossRadius = 12,      ///< Max.\ radius of scaled SB cross, in pixels.
    SCObjCrossRadius = 5,
    SCMaxObjCrossRadius = 10;

static const uint8 ionstorm_init[] = { 0x88, 0x00, 0x22, 0x00,
                                       0x88, 0x00, 0x22, 0x00 };
static const uint8 ionstorm_dense_init[] = { 0x88, 0x44, 0x22, 0x44,
                                             0x88, 0x44, 0x22, 0x44 };

const GfxFillPattern IONSTORM_FILL = ionstorm_init;
const GfxFillPattern IONSTORM_DENSE_FILL = ionstorm_dense_init;


/** Draw a, possibly filled, circle.
    \parm f    zero for outline circle, nonzero for filled. */
static void
drawObject(GfxCanvas* can, GfxParameters& parm, GfxPoint pt, int r, int32 f)
{
    if (f) {
        drawFilledCircle(can, parm, pt.x, pt.y, r);
    } else {
        drawCircle(can, parm, pt.x, pt.y, r);
    }
}

void
GChartViewport::drawMovingObject(GfxCanvas* can, GfxParameters& parm,
                                 GfxPoint pt, int radius,
                                 int speed, int heading, int32 f)
{
    drawObject(can, parm, pt, radius, f);
    if (speed > 0 && speed <= 20 && heading >= 0) {
        double h = heading * PI / 180;
        pt.x += round(radius * std::sin(h));
        pt.y -= round(radius * std::cos(h));

        int way = scale(speed * speed);
        int dx = round(way * std::sin(h));
        int dy = round(way * std::cos(h));

        drawArrow(can, parm, pt.x, pt.y, pt.x + dx, pt.y - dy,
                  2*mult < divi ? 3 : 5);
    }
}

/** Set fill style for drawing a minefield. Must frob the pattern a bit
    to make the minefield appear the same in all positions of the map. */
static void
setMineFillStyle(GfxParameters& parm, int x0, int y0, GGameTurn& /*trn*/,
                 GMinefield* pmf)
{
    if (pmf->getOwner() == ::getPlayerId())
        ++y0;
    if (pmf->isWeb()) {
        parm.fill_pattern = GRAY50_FILL;
        if ((x0 + y0) & 1)
            parm.fill_pattern.shiftUp(1);
    } else {
        parm.fill_pattern = GRAY25_FILL;
        parm.fill_pattern.shiftUp((15 + 2*(x0 & 1) - y0 & 3) & 3);
    }
}

/***************************** GChartViewport ****************************/

/*! \class GChartViewport
    This encapsulates the logic for drawing a starchart. Starcharts
    can be displayed in various zoom levels specified as fractions. */

/** Initialize. Call after setting the other parameters to compute
    derived information. */
void
GChartViewport::init()
{
    xm = view.x + view.w / 2;
    ym = view.y + view.h / 2;
    x0 = xm - scale(center.x);
    y0 = ym + scale(center.y);
}

/** Scale. Convert "world" distance into "screen" distance. The
    value n should be chosen so that the result is in range
    permissible for an int16 (because int may have 16 bits only, and
    drawing code might only be able to handle 16 bit coordinates). */
int
GChartViewport::scale(int n) const
{
    /* we want to round towards -INF instead of 0 */
    if ((-1)/2 == 0 && n < 0)
        return - (-long(n) * mult) / divi;
    return long(n) * mult / divi;
}

/** Scale, with Clipping. Compute scale(n)+add, but clip the result to
    be in the guaranteed range of int16. Unlike scale(), this is designed
    to handle out-of-range values of n. */
int
GChartViewport::scaleClip(int n, int add) const
{
    /* no need to fix rounding behaviour, this one's called to
       scale large values... */
    long l = long(n) * mult / divi + add;
    if (l < -0x7FFF)
        return -0x7FFF;
    else if (l > 0x7FFF)
        return 0x7FFF;
    else
        return l;
}

/** Unscale. Convert "screen" distance into "world" distance. */
int
GChartViewport::unscale(int n) const
{
    if ((-1)/2 == 0 && n < 0)
        return - (-long(n) * divi) / mult;
    return long(n) * divi / mult;
}

/** Draw Starchart, main entry point.
    \param can    Canvas to draw to
    \param opt    Options (what to draw)
    \param trn    Turn data to display
    \param bbox   Applicable bounding box */
void
GChartViewport::draw(Ptr<GfxCanvas> can, GChartOptions& opt, GGameTurn& trn,
                     const GChartBBox& bbox)
{
    Ptr<GfxCanvas> worker = new GfxClipRectFilter (view, can);

    init();
    // FIXME: set font....
    // FIXME: this order of starchart drawing has the drawback that in
    // wrapped mode, decoration *can* obscure ships (the "east" image
    // is drawn after the "center" image, hence an "east" planet ring
    // can obscure a "center" ship when they're close enough to the
    // border).
    if (chart_conf.wrapped) {
        int x0s = x0, y0s = y0;
        for (int i = -1; i <= 1; ++i) {
            for (int j = -1; j <= 1; ++j) {
                x0 = x0s + scale(i*chart_conf.xsize);
                y0 = y0s + scale(j*chart_conf.ysize);
                drawAux(worker, opt, trn, bbox);
            }
        }
    } else {
        drawAux(worker, opt, trn, bbox);
    }
}

/** Draw starchart, helper routine. This does the actual drawing of
    one starchart image. */
void
GChartViewport::drawAux(Ptr<GfxCanvas> can, GChartOptions& opt, GGameTurn& trn,
                        const GChartBBox& bbox)
{
    int range_x1 = scaleClip(bbox.min.x, x0);
    int range_x2 = scaleClip(bbox.max.x, x0);
    int range_y1 = scaleClip(-bbox.min.y, y0);
    int range_y2 = scaleClip(-bbox.max.y, y0);
    if (! can->isVisible(GfxRect(range_x1,
                                 range_y2,
                                 range_x2 - range_x1,
                                 range_y1 - range_y2)))
        return;

    drawSectors(can, opt.show);
    if (opt.show & GChartOptions::co_Mine)
        drawMines(can, trn, opt.fill);
// FIXME:
//     if (opt.show & co_Ufo)
//         drawUfos(can, trn, opt.fill & co_Ufo);
    if (opt.show & GChartOptions::co_Ion)
        drawIons(can, trn, opt.fill);
//     if (opt.show & co_Drawings)
//         drawDrawings(can, trn, opt.show);
    drawShipSelAndVectors(can, trn, opt.show);
    drawPlanets(can, trn, opt.show);
    drawShips(can, trn, opt.show);
}

/** Draw a cross, used for the center of Ufos and mine fields. */
void
GChartViewport::drawCross(Ptr<GfxCanvas> can, GfxParameters& parm, GfxPoint pt)
{
    int n = scale(SCObjCrossRadius);
    if (n <= 0)
        n = 1;
    else if (n > SCMaxObjCrossRadius)
        n = SCMaxObjCrossRadius;
    drawHLine(can, parm, pt.x - n, pt.y, pt.x + n);
    drawVLine(can, parm, pt.x, pt.y - n, pt.y + n);
}

void
GChartViewport::drawIons(Ptr<GfxCanvas> can, GGameTurn& trn, int fill)
{
    GfxParameters parm;
    parm.fill_pattern = IONSTORM_FILL;
    for (int sid = 1; sid <= NUM_STORMS; ++sid) {
        GIonStorm& i = trn.getIonStorm(sid);
        if (!i.isExistant())
            continue;

        if (i.getVoltage() < 50)
            parm.color = standard_colors[COLOR_BLUE];
        else if (i.getVoltage() < 100)
            parm.color = standard_colors[COLOR_GRAY];
        else if (i.getVoltage() < 150)
            parm.color = standard_colors[COLOR_WHITE];
        else
            parm.color = standard_colors[COLOR_RED];
        
        drawMovingObject(can, parm,
                         scaleXY(i.getPos()),
                         scale(i.getRadius()),
                         i.getSpeed(), i.getHeading(),
                         fill & GChartOptions::co_Ion);
    }
}

/** Draw Minefields. Call only if minefield drawing is actually enabled.
    \param fill  options.fill */
void
GChartViewport::drawMines(Ptr<GfxCanvas> can, GGameTurn& trn, int fill)
{
    GfxParameters parm;
    for (GGameTurn::MinefieldMap::iterator i = trn.getMinefieldBegin();
         i != trn.getMinefieldEnd(); ++i)
    {
        GMinefield* pmf = i->second;
        GfxPoint pt = scaleXY(pmf->getPos());
        setMineFillStyle(parm, x0, y0, trn, pmf);
        // FIXME: three colors
        parm.color = standard_colors[pmf->getOwner() != ::getPlayerId() ? COLOR_DARK : (COLOR_GREENSCALE + 4)];
        drawObject(can, parm, pt, scale(pmf->getRadius()), fill & GChartOptions::co_Mine);
        drawCross(can, parm, pt);
    }
}

/** Draw sector/starchart borders.
    \param show   options.show */
void
GChartViewport::drawSectors(Ptr<GfxCanvas> can, int show)
{
    GfxParameters parm;
    int  dx = std::min(chart_conf.xsize / 200, 10);
    int  dy = std::min(chart_conf.ysize / 200, 10);
    int  center_x = (chart_conf.xmin + chart_conf.xmax) / 2;
    int  center_y = (chart_conf.ymin + chart_conf.ymax) / 2;

    parm.color = standard_colors[COLOR_DARK];
    if (show & GChartOptions::co_Sectors) {
        parm.line_style = (y0 & 1) ? 0x55 : 0xAA;
        for (int i = -dx; i <= dx; ++i)
            drawVLine(can, parm,
                      scaleX(center_x + 100*i),
                      scaleY(center_y - 100*dy),
                      scaleY(center_y + 100*dy));

        parm.line_style = (x0 & 1) ? 0x55 : 0xAA;
        for (int i = -dy; i <= dy; ++i)
            drawHLine(can, parm,
                      scaleX(center_x - 100*dy),
                      scaleY(center_y + 100*i),
                      scaleX(center_x + 100*dy));
    }

    if (show & GChartOptions::co_Borders) {
        parm.line_style = ror8(0x27, y0);
        drawVLine(can, parm,
                  scaleX(chart_conf.xmin),
                  scaleY(chart_conf.ymin),
                  scaleY(chart_conf.ymax));
        drawVLine(can, parm,
                  scaleX(chart_conf.xmax),
                  scaleY(chart_conf.ymin),
                  scaleY(chart_conf.ymax));

        parm.line_style = ror8(0x27, x0);
        drawHLine(can, parm,
                  scaleX(chart_conf.xmin),
                  scaleY(chart_conf.ymin),
                  scaleX(chart_conf.xmax));
        drawHLine(can, parm,
                  scaleX(chart_conf.xmin),
                  scaleY(chart_conf.ymax),
                  scaleX(chart_conf.xmax));
    }
}

/** Draw Ships, part I. Draws ship selection and trails. These are
    drawn below the planets, to not obscure other stuff. */
void
GChartViewport::drawShipSelAndVectors(Ptr<GfxCanvas> can, GGameTurn& trn, int show)
{
    if (show & GChartOptions::co_Selection) {
        GfxParameters parm;
        parm.color = standard_colors[COLOR_YELLOW];
        for (int i = 1; i <= NUM_SHIPS; ++i) {
            GShip& s = trn.getShip(i);
            if (s.isExistant() && (s.getUIFlags() & GShip::ss_Marked)) {
                GfxPoint pt = scaleXY(s.getPos());
                drawSelection(can, parm, pt.x, pt.y, mult, divi);
            }
        }
    }
//   IF show AND co_Trails<>0 THEN BEGIN
//     FOR i:=1 TO NUM_SHIPS DO
//      IF (Ships^[i].HHist<>NIL) THEN
//       NShowVector(view, i);
//   END ELSE IF view.vec_id>0 THEN
//    NShowVector(view, view.vec_id);
}

/** Draw Ships, part II. These are drawn above the planets, so that planet
    rings do not obscure ships in warp wells, like in VPA, for example
    (hint hint! ;-) */
void
GChartViewport::drawShips(Ptr<GfxCanvas> can, GGameTurn& trn, int show)
{
    (void) show;
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        GShip& s = trn.getShip(i);
        if (! s.isExistant())
            continue;

        GfxPoint pt = scaleXY(s.getPos());
        int   color = fof_colors[isFriend(s.getOwner())];
        if (s.getOrbit() == 0)
            can->drawPixel(pt.x, pt.y, standard_colors[color], SDL_ALPHA_OPAQUE);
// FIXME:
//     IF (Fleets^[i].FleetID = i) AND (2*view.mult > view.divi) THEN BEGIN
//       PutPixel(x-1, y-1, Color);
//       PutPixel(x-1, y+1, Color);
//       PutPixel(x+1, y-1, Color);
//       PutPixel(x+1, y+1, Color);
//     END;
//     IF (Ships^[i].Text<>NIL) AND (show AND co_Labels<>0) THEN BEGIN
//       IF Ships^[i].Orbit<>0 THEN Inc(y, 4);
//       OutLTextXY(x, y+1, 600, Ships^[i].Text^);
//     END;
    }
}

/** Draw Planets. */
void
GChartViewport::drawPlanets(Ptr<GfxCanvas> can, GGameTurn& trn, int show)
{
    enum RingType { r_Enemy, r_Own, r_Unowned, r_Unknown };
    GfxParameters parm;
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        GPlanet& p = trn.getPlanet(i);
        if (!p.isExistant())
            continue;

        GfxPoint pt = scaleXY(p.getPos());

        /* Selection */
        if ((show & GChartOptions::co_Selection) && (p.getUIFlags() & GPlanet::ps_Marked)) {
            parm.color = standard_colors[COLOR_YELLOW];
            drawSelection(can, parm, pt.x, pt.y, mult, divi);
        }

        bool     base;          // true iff planet has base
        RingType ring;          // type of planet ring
        int      incy;          // Y increment for text
        bool     shipg;         // true iff ship ring is dotted (guessed)
        int      shipring;      // color of ship ring, 0 if none

        /* Compute decoration -- Planet */
        // FIXME: history & stuff
        incy = 1;
        if (p.getOwner() == getPlayerId()) {
            ring = r_Own;
            base = p.hasBase();
        } else {
            if (p.getOwner())
                ring = r_Enemy;
            else if (p.haveDat())
                ring = r_Unowned;
            else
                ring = r_Unknown;
            base = false;
        }
        /* Compute decoration -- Ship */
        shipg = false;
        shipring = 0;
        switch(p.getStatus() & (GPlanet::PLANET_OWNORBIT + GPlanet::PLANET_ENEMYORBIT + GPlanet::PLANET_ALLYORBIT)) {
         case 0:
            shipg = true;
            switch(p.getStatus() & (GPlanet::PLANET_GENEMYORBIT + GPlanet::PLANET_GALLYORBIT)) {
             case 0:
                shipring = 0;
                break;
             case GPlanet::PLANET_GALLYORBIT:
                shipring = fof_colors[FOF_FRIEND];
                break;
             case GPlanet::PLANET_GENEMYORBIT:
                shipring = COLOR_RED;
                break;
             default:
                shipring = COLOR_YELLOW;
                break;
            }
            break;
         case GPlanet::PLANET_OWNORBIT:
            shipring = COLOR_GREEN;
            break;
         case GPlanet::PLANET_ENEMYORBIT:
            shipring = COLOR_RED;
            break;
         case GPlanet::PLANET_ALLYORBIT:
         case GPlanet::PLANET_ALLYORBIT + GPlanet::PLANET_OWNORBIT:
            shipring = fof_colors[FOF_FRIEND];
            break;
         default:
            shipring = COLOR_YELLOW;
            break;
        }

        /* Draw the planet */
        if (ring != r_Own)
            parm.color = standard_colors[COLOR_BLUEBLACK];
        else
            parm.color = standard_colors[COLOR_BLUE];
        if (base) {
            int sbsize = std::min(scale(SCCrossRadius), SCMaxCrossRadius);
            drawHLine(can, parm, pt.x - sbsize, pt.y, pt.x + sbsize);
            drawVLine(can, parm, pt.x, pt.y - sbsize, pt.y + sbsize);
        }
        can->drawPixel(pt.x, pt.y, standard_colors[COLOR_WHITE], SDL_ALPHA_OPAQUE);

        int r = std::min(scale(SCRingRadius), SCMaxRingRadius);
        if (shipring == 0 && r == 0)
            r = 1;
        if (r) {
            if (ring == r_Enemy || ring == r_Own) {
                drawCircle(can, parm, pt.x, pt.y, r);
                incy += r;
            } else if (ring == r_Unowned) {
                drawDottedCircle(can, parm, pt.x, pt.y, r);
                incy += r;
            }
        }
        if (shipring != 0) {
            parm.color = standard_colors[shipring];
            ++r;
            if (shipg)
                drawDottedCircle(can, parm, pt.x, pt.y, r);
            else
                drawCircle(can, parm, pt.x, pt.y, r);
            incy = r+1;
        }
//     IF (Stars^[i].Text<>NIL) AND (show AND co_Labels<>0) THEN BEGIN
//       Color:=Gray;
//       OutLTextXY(x,y+IncY,600,Stars^[i].Text^);
//     END;
    }
}

/** Zoom in.
    \param bbox applicable bounding box (limits available zoom factors)
    \returns true iff zoom factor changed */
bool
GChartViewport::zoomIn(GChartBBox& bbox)
{
    if (divi > 1)
        --divi;
    else if (mult < bbox.max_zoom * divi)
        ++mult;
    else
        return false;           // can't zoom in further

    /* Zoom range may be exceeded. For example, old zoom 5:2, limit 4;
       new zoom 5:1 */
    if (mult > divi * bbox.max_zoom) {
        mult = bbox.max_zoom;
        divi = 1;
    }
    return true;
}

/** Zoom out.
    \param bbox applicable bounding box
    \returns true iff zoom factor changed */
bool
GChartViewport::zoomOut(GChartBBox& bbox)
{
    // FIXME: incomplete
    (void) bbox;
    if (mult > 1) {
        --mult;
        return true;
    } else if (divi < GChartBBox::MAX_ZOOM) {
        ++divi;
        return true;
    } else {
        return false;
    }
}

static GChartOptions chart_options[3] = {
    { GChartOptions::co_Default, GChartOptions::co_Default },
    { GChartOptions::co_Default, GChartOptions::co_Default },
    { GChartOptions::co_Default, GChartOptions::co_Default }
};

/** Get starchart options for a particular chart widget.
    \param is_scanner  true if scanner in a control screen, false if main chart
    \param mult,divi   zoom settings

    Currently implements the PCC 1.x policy. */
GChartOptions&
getChartOpts(bool is_scanner, int mult, int divi)
{
    if (is_scanner)
        return chart_options[2];
    else if (mult < divi)
        return chart_options[0];
    else
        return chart_options[1];
}
