/**
  *  \file client/chartusr.cc
  *  \brief Starchart -- User Interface
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/chartusr.h"
#include "client/chart.h"
#include "gfx/gfx.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "game/objlist.h"
#include "game/lock.h"
#include "game/ship.h"
#include "game/planet.h"
#include "client/cscreen.h"
#include "client/playerscr.h"
#include "client/dlg-list.h"
#include "game/player.h"
#include "ui/util.h"
#include "game/formula.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "ui/widgets.h"
#include "gfx/gfxalpha.h"

namespace {
    /** Current position in starchart. This structure is not
        necessarily POD. We still compare two of these via memcmp to
        check whether a redraw is needed. */
    struct ChartLocation {
        GObjectId     current;      ///< Currently-locked object.
        int           pid;          ///< Planet at current location.
        int           ship_count;   ///< Number of ships at this location.

        int           ol_top, ol_pos;///< Index into chart_objlist.

        void syncPositions();
        void setCurrentIndex(int y);
        void scroll(int delta);
    };
};

static const int OBJS_PER_PAGE = 15;

static ChartLocation  chart_location;
static GChartViewport chart_viewport;
static GObjectList    chart_objlist;

/*********************** Chart Location Management ***********************/

void
ChartLocation::syncPositions()
{
    if (ol_pos < ol_top)
        ol_top = ol_pos;
    if (ol_pos >= ol_top + OBJS_PER_PAGE)
        ol_top = ol_pos - OBJS_PER_PAGE + 1;
}

void
ChartLocation::setCurrentIndex(int y)
{
    ol_pos = y;
    syncPositions();
    current = chart_objlist[ol_pos];
}

void
ChartLocation::scroll(int delta)
{
    if (chart_objlist.getSize()) {
        int n = ol_pos + delta;
        if (n < 0)
            n = chart_objlist.getSize() - 1;
        if (n >= chart_objlist.getSize())
            n = 0;
        setCurrentIndex(n);
    }
}

static void
initChartLocation()
{
    // int sid = 0;

    chart_location.pid = getDisplayedTurn().getPlanetAt(chart_viewport.center);
    chart_location.ship_count = 0;

    chart_objlist.clear();
    chart_objlist.addObjectsAt(getDisplayedTurn(), chart_viewport.center,
                               ls_IncludeForeignShips + ls_IncludePlanet, 0);
    chart_objlist.postprocessUser(false);

    if (chart_objlist.getSize()) {
        chart_location.current = chart_objlist[0];
        chart_location.ol_pos = chart_location.ol_top = 0;
    } else {
        chart_location.current = GObjectId();
        chart_location.ol_pos = chart_location.ol_top = 0;
    }

//     for (int i = 1; i <= NUM_SHIPS; ++i) {
//         GShip& s = getDisplayedTurn().getShip(i);
//         if (s.isExistant() && s.getPos() == chart_viewport.center) {
//             ++chart_location.ship_count;
//             if (sid == 0)
//                 sid = i;
//         }
//     }

//     if (chart_location.pid)
//         chart_location.current = GObjectId(getDisplayedTurn().getPlanetType(),
//                                            getDisplayedTurn().getPlanet(chart_location.pid));
//     else if (sid)
//         chart_location.current = GObjectId(getDisplayedTurn().getShipType(),
//                                            getDisplayedTurn().getShip(sid));
//     else
//         chart_location.current = GObjectId();
}

/********************************** Main *********************************/

void
initStarchart()
{
    chart_viewport.center.x = chart_viewport.center.y = 2000;
}

void
setStarchartPosition(GPoint pt)
{
    chart_viewport.center = pt;
}

static void
doChartLock(GPoint& pt, TLockMode lm, bool marked)
{
    GLockData locker(pt, marked);
    findItem(getDisplayedTurn(), locker, lm);
    pt = locker.getPoint();
}

namespace {
    class StarchartWidget : public UIWidget {
     public:
        StarchartWidget();
        void draw(Ptr<GfxCanvas>, UpdateInfo&);
        bool handleEvent(const UIEvent& e, bool second_pass);

        void drawChart(Ptr<GfxCanvas> can);
        void getLayoutInfo(LayoutInfo&);
    };
}

StarchartWidget::StarchartWidget()
    : UIWidget(0)
{
    setState(st_Modal, true);
}

void
StarchartWidget::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    drawChart(can);
    ui.update(getExtent());
}

bool
StarchartWidget::handleEvent(const UIEvent& /*e*/, bool /*second_pass*/)
{
    /* FIXME? */
    return true;
}

void
StarchartWidget::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_Flexible;
    info.pref_size = info.min_size = GfxPoint(100,100); // FIXME?
}

static void
drawFancyBar(Ptr<GfxCanvas> can, GfxRect r)
{
    const int icolor = standard_colors[COLOR_SHIELD+3];
    const int fcolor = standard_colors[COLOR_SHIELD+5];
    const int z = standard_colors[COLOR_BLACK];

    int alpha;

    if (ui_screen_bits <= 8) {
        can->drawBar(r, icolor, z, GRAY50_FILL, SDL_ALPHA_OPAQUE);
        alpha = SDL_ALPHA_OPAQUE;
    } else {
        can->drawBar(r, icolor, z, SOLID_FILL, 128);
        alpha = 192;
    }

    can->drawBar(GfxRect(r.x-1,   r.y-2,   r.w+2, 2),   fcolor, z, SOLID_FILL, alpha);
    can->drawBar(GfxRect(r.x-1,   r.y+r.h, r.w+2, 2),   fcolor, z, SOLID_FILL, alpha);
    can->drawBar(GfxRect(r.x-2,   r.y,     2,     r.h), fcolor, z, SOLID_FILL, alpha);
    can->drawBar(GfxRect(r.x+r.w, r.y,     2,     r.h), fcolor, z, SOLID_FILL, alpha);
    can->drawPixel(r.x-2,     r.y-1,   fcolor, alpha);
    can->drawPixel(r.x+r.w+1, r.y-1,   fcolor, alpha);
    can->drawPixel(r.x-2,     r.y+r.h, fcolor, alpha);
    can->drawPixel(r.x+r.w+1, r.y+r.h, fcolor, alpha);
}

static void
drawChartButton(Ptr<GfxCanvas> can, int& x, int y, const string_t& text)
{
    int width = 8 + fonts[FONT_SMALL]->getTextWidth(text);
    UIButton::drawButton(can, GfxRect(x, y, width, 15), text, UIButton::bf_SmallSize);
    x += width + 1;
}

void
StarchartWidget::drawChart(Ptr<GfxCanvas> can)
{
    can->drawBar(chart_viewport.view, standard_colors[0],
                 standard_colors[0], SOLID_FILL,
                 SDL_ALPHA_OPAQUE);
#if 0
    /* looks quite funky, but isn't at all correct. Sure it's
       reasonable that when playing in a spherical universe, you see
       the "back side" of the sphere while scrolling through the
       "front" side. But just painting it isn't enough, we also have
       to turn it around 180. */
    {
        Ptr<GfxCanvas> ncan = new GfxAlphaFilter(64, can);
        GChartViewport vp2 = chart_viewport;
        vp2.divi *= 2;
        vp2.center.x = 4000 - chart_viewport.center.x;
        vp2.center.y = 4000 - chart_viewport.center.y;
        vp2.draw(ncan, getChartOpts(false, vp2.mult, vp2.divi),
                 getDisplayedTurn(), getDisplayedTurn().getChartBBox());
    }
#endif
    chart_viewport.draw(can,
                        getChartOpts(false, chart_viewport.mult, chart_viewport.divi),
                        getDisplayedTurn(),
                        getDisplayedTurn().getChartBBox());

    GfxParameters parm;
    GfxPoint sc = chart_viewport.getCenter();
    if (! chart_location.current.isValid()) {
        parm.color = standard_colors[COLOR_BLUE];
        drawHLine(can, parm, sc.x-30, sc.y, sc.x-6);
        drawHLine(can, parm, sc.x+30, sc.y, sc.x+6);
        drawVLine(can, parm, sc.x, sc.y-30, sc.y-6);
        drawVLine(can, parm, sc.x, sc.y+30, sc.y+6);
    }

    GfxRect screen_size(ui_screen->getSurface());
    const int h = font_heights[FONT_NORMAL] + 2;
    const int y = screen_size.h - h;
    drawFancyBar(can, GfxRect(1, y - 2, screen_size.w - 2, h));

    parm.color = standard_colors[COLOR_GRAY];
    outTextF(can, parm, fonts[FONT_NORMAL], 3, y, 1000, _("Starchart:"));
    int x = 8 + fonts[FONT_NORMAL]->getTextWidth(_("Starchart:"));
    drawChartButton(can, x, y, "ESC");
    drawChartButton(can, x, y, "F7");
    drawChartButton(can, x, y, "M");
    drawChartButton(can, x, y, "I");
    drawChartButton(can, x, y, "U");
    drawChartButton(can, x, y, "H");
}

void
drawObjectList(Ptr<GfxCanvas> can)
{
    if (!chart_objlist.getSize())
        return;

    int lines = chart_objlist.getSize();
    if (lines > OBJS_PER_PAGE)
        lines = OBJS_PER_PAGE;

    GfxPoint pt = chart_viewport.getCenter();
    int h = font_heights[FONT_NORMAL]*lines;
    int y = pt.y - h/2;

    drawFancyBar(can, GfxRect(pt.x - 300, y-2, 280, h+4));

    GfxParameters parm;
    for (int i = 0; i < lines; ++i) {
        GObjectId id = chart_objlist[i + chart_location.ol_top];
        int fof = id.isGameObject() ? isFriend(id.getObject().getOwner()) : FOF_TECHIE;
        if (i + chart_location.ol_top == chart_location.ol_pos) {
            if (ui_screen_bits <= 8)
                can->drawBar(GfxRect(pt.x-298, y+i*font_heights[FONT_NORMAL], 276, font_heights[FONT_NORMAL]),
                             standard_colors[COLOR_SHIELD+4], GFX_TRANSPARENT, SOLID_FILL, SDL_ALPHA_OPAQUE);
            else
                can->drawBar(GfxRect(pt.x-298, y+i*font_heights[FONT_NORMAL], 276, font_heights[FONT_NORMAL]),
                             standard_colors[COLOR_SHIELD+8], 0, SOLID_FILL, 128);
            parm.color = standard_colors[fof_colors[fof]];
            parm.bgcolor = GFX_TRANSPARENT;
            parm.alpha = SDL_ALPHA_OPAQUE;
        } else {
            parm.color = standard_colors[fof_colors[fof]];
            parm.bgcolor = GFX_TRANSPARENT;
            parm.alpha = SDL_ALPHA_OPAQUE;
        }
        outTextF(can, parm, fonts[FONT_NORMAL], pt.x-298, y+i*font_heights[FONT_NORMAL], 276,
                 chart_objlist[i+chart_location.ol_top].getName());
    }
}

static void
drawObject(Ptr<GfxCanvas> can)
{
    if (!chart_objlist.getSize())
        return;

    GObjectId id = chart_objlist[chart_location.ol_pos];
    if (GShip* s = dynamic_cast<GShip*>(&id.getObject())) {
        GfxPoint pt = chart_viewport.getCenter();
        GfxParameters parm;
        int h = 110;
        int y = pt.y - h/2;
        int hull = s->getHullId();
        int x;

        if (hull > 0) {
            drawFancyBar(can, GfxRect(pt.x + 20, y, 280, h));

            GfxRect imgrect(pt.x + 25, y + 5, 100, 100);
            Ptr<GfxPixmap> pix = getResourcePixmap(mkResId(rt_Ship, getHull(hull).getPicture() + 1000),
                                                   rc_Normal);
            if (pix)
                pix->blitSized(can, imgrect.x, imgrect.y, imgrect.w, imgrect.h);

            imgrect.grow(1, 1);
            parm.color = standard_colors[COLOR_SHIELD+5];
            parm.alpha = 192;
            drawRectangle(can, parm, imgrect);

            x = pt.x + 130;

            drawChartButton(can, x, y + 90, "F1");
            drawChartButton(can, x, y + 90, "F6");
            drawChartButton(can, x, y + 90, "Ins");

            x = pt.x + 130;
        } else {
            drawFancyBar(can, GfxRect(pt.x + 20, y, 180, h));
            x = pt.x + 30;
        }
        parm.color = standard_colors[COLOR_GRAY];
        parm.alpha = SDL_ALPHA_OPAQUE;
        parm.bgcolor = GFX_TRANSPARENT;
        outTextF(can, parm, fonts[FONT_NORMAL], x, y + 5, 150, s->getName());
        if (hull > 0)
            outTextF(can, parm, fonts[FONT_NORMAL], x, y + 5 + font_heights[FONT_NORMAL], 150,
                     format(_("a %s %s"), getAdjName(s->getOwner()), s->getHull().getShortName()));
        else
            outTextF(can, parm, fonts[FONT_NORMAL], x, y + 5 + font_heights[FONT_NORMAL], 150,
                     format(_("a %s starship"), getAdjName(s->getOwner())));
    }
}

static void
drawLocker(Ptr<GfxCanvas> can, int angle)
{
    static const int SIZE = 15;
    static const double FACTOR = 5 * PI / 180;

    angle = -angle;

    GfxPoint pt = chart_viewport.getCenter();
    GfxParameters parm;
    parm.color = standard_colors[COLOR_RED];

    /* it turns out that scaling the locker doesn't look that good. */
    const double scaled_size = SIZE;

    for (int i = 0; i < 2; ++i) {
        double d = angle * PI / 180;
        drawLine(can, parm,
                 pt.x + round(4*std::sin(d)), pt.y + round(4*std::cos(d)),
                 pt.x + round(scaled_size*std::sin(d)), pt.y + round(scaled_size*std::cos(d)));
        for (int j = -8; j < 8; ++j) {
            drawLine(can, parm,
                     pt.x + round(scaled_size*std::sin(d-FACTOR*j)),
                     pt.y + round(scaled_size*std::cos(d-FACTOR*j)),
                     pt.x + round(scaled_size*std::sin(d-FACTOR*j+FACTOR)),
                     pt.y + round(scaled_size*std::cos(d-FACTOR*j+FACTOR)));
        }
        angle += 180;
    }
}

int
doStarchart()
{
    Ptr<GfxPixmap> locker_img = new GfxPixmap(31, 31, ui_screen->getSurface()->format);
    int locker_angle = 0;

    enum RedrawMode {
        rNothing,               // need not redraw anything
        rRedrawMap,             // need to redraw map
        rRedrawPanel,           // need to redraw map and panel
        rUpdate                 // need to do complete update
    };

    RedrawMode redraw = rUpdate;

    chart_viewport.mult = chart_viewport.divi = 1;
    chart_viewport.view = GfxRect(ui_screen->getSurface());
    chart_viewport.vec_id = 0;
    chart_viewport.init();

    StarchartWidget chart_w;
    chart_w.setExtent(chart_viewport.view);

    GPoint        prev_center = chart_viewport.center;
    ChartLocation prev_loc;
    int pending_dx = 0, pending_dy = 0;
    int return_value = 0;

    ui_root->add(chart_w);

    while (1) {
        if (chart_viewport.center != prev_center)
            redraw = rUpdate;
        if (redraw < rRedrawPanel && std::memcmp(&prev_loc, &chart_location, sizeof chart_location))
            redraw = rRedrawPanel;

        if (redraw != rNothing) {
            if (redraw > rRedrawPanel)
                initChartLocation();

            ui_screen->lock();
            chart_w.drawChart(ui_screen);
            drawObjectList(ui_screen);
            drawObject(ui_screen);
            ui_screen->unlock();
            ui_screen->update(chart_viewport.view);
            redraw = rNothing;
            prev_center = chart_viewport.center;
            prev_loc  = chart_location;
            if (chart_location.current.isValid()) {
                GfxRect r(chart_viewport.getCenter().x - 15,
                          chart_viewport.getCenter().y - 15, 31, 31);
                locker_img->lock();
                locker_img->blitSurface(-r.x, -r.y, ui_screen->getSurface(), r);
                locker_img->unlock();
                locker_img->update(GfxRect(0, 0, 31, 31));

                ui_screen->lock();
                drawLocker(ui_screen, locker_angle);
                ui_screen->unlock();
                ui_screen->update(r);
            }
        }

        UIEvent e;
        getEvent(e, false, true);
        if (e.type == ev_Tick) {
            if (chart_location.current.isValid()) {
                GfxRect r(chart_viewport.getCenter().x - 15,
                          chart_viewport.getCenter().y - 15, 31, 31);

                locker_angle = (locker_angle + 3) % 360;
                locker_img->blit(ui_screen, r.x, r.y);
                ui_screen->lock();
                drawLocker(ui_screen, locker_angle);
                ui_screen->unlock();
                ui_screen->update(r);
            }
        } else if (e.type == ev_Mouse) {
            /* Mouse movement. When mouse movement does not immediately
               translate into screen movement, we accumulate mouse movement
               in pending_d[xy] instead of discarding it. This gives much
               smoother movement in large zoom levels. */
            pending_dx += e.mouse.x;
            pending_dy += e.mouse.y;
            GPoint new_pt(chart_viewport.center.x + chart_viewport.unscale(pending_dx),
                          chart_viewport.center.y - chart_viewport.unscale(pending_dy));

            if (e.mouse.b) {
                if (e.mouse.b & mb_Left)
                    doChartLock(new_pt, lm_Left, e.mouse.b & ss_Ctrl);
                else if (e.mouse.b & mb_Right)
                    doChartLock(new_pt, lm_Right, e.mouse.b & ss_Ctrl);
            }

            if (new_pt != chart_viewport.center) {
                chart_viewport.center = new_pt;
                redraw = rUpdate;
                pending_dx = pending_dy = 0;
            }
        } else if (e.type == ev_Key) {
            /* Keys */
            switch (e.key.code) {
             case SDLK_ESCAPE:
             case SDLK_F4:
                /* ESC, F4 -- exit */
                goto out;
             case SDLK_LEFT:
                chart_viewport.center.x -= e.key.arg ? e.key.arg : 10;
                break;
             case SDLK_RIGHT:
                chart_viewport.center.x += e.key.arg ? e.key.arg : 10;
                break;
             case SDLK_UP:
                chart_viewport.center.y += e.key.arg ? e.key.arg : 10;
                break;
             case SDLK_DOWN:
                chart_viewport.center.y -= e.key.arg ? e.key.arg : 10;
                break;
             case ss_Ctrl + SDLK_LEFT:
                chart_viewport.center.x -= 100;
                break;
             case ss_Ctrl + SDLK_RIGHT:
                chart_viewport.center.x += 100;
                break;
             case ss_Ctrl + SDLK_UP:
                chart_viewport.center.y += 100;
                break;
             case ss_Ctrl + SDLK_DOWN:
                chart_viewport.center.y -= 100;
                break;
             case ss_Shift + SDLK_LEFT:
                chart_viewport.center.x--;
                break;
             case ss_Shift + SDLK_RIGHT:
                chart_viewport.center.x++;
                break;
             case ss_Shift + SDLK_UP:
                chart_viewport.center.y++;
                break;
             case ss_Shift + SDLK_DOWN:
                chart_viewport.center.y--;
                break;
             case SDLK_RETURN:
                doChartLock(chart_viewport.center, lm_Right, false);
                break;
             case SDLK_SPACE:
                doChartLock(chart_viewport.center, lm_Left, false);
                break;
             case ss_Ctrl + SDLK_RETURN:
                doChartLock(chart_viewport.center, lm_Right, true);
                break;
             case ss_Ctrl + SDLK_SPACE:
                doChartLock(chart_viewport.center, lm_Left, true);
                break;
             case '-':
                if (chart_viewport.zoomOut(getDisplayedTurn().getChartBBox()))
                    redraw = rRedrawMap;
                break;
             case '+':
                if (chart_viewport.zoomIn(getDisplayedTurn().getChartBBox()))
                    redraw = rRedrawMap;
                break;
             case SDLK_F1:
             case SDLK_F1 + ss_Ctrl:
             case SDLK_F1 + ss_Shift:
                /* FIXME */
                break;
             case SDLK_F2:
             case SDLK_F2 + ss_Ctrl:
             case SDLK_F2 + ss_Shift:
                if (chart_location.pid)
                    putCommand(cm_CS_GotoCS, 2, chart_location.pid);
                break;
             case SDLK_F3:
             case SDLK_F3 + ss_Ctrl:
             case SDLK_F3 + ss_Shift:
             case SDLK_F8:
             case SDLK_F8 + ss_Ctrl:
             case SDLK_F8 + ss_Shift:
                if (chart_location.pid)
                    putCommand(cm_CS_GotoCS, 3, chart_location.pid);
                break;
             case SDLK_TAB:
                if (chart_objlist.getSize() > 1) {
                    chart_location.scroll(+1);
                    redraw = rRedrawPanel;
                }
                break;
             case ss_Shift + SDLK_TAB:
                if (chart_objlist.getSize() > 1) {
                    chart_location.scroll(-1);
                    redraw = rRedrawPanel;
                }
                break;
            }
        } else if (e.type == ev_Command) {
            if (e.cmd.code == cm_CS_GotoCS) {
                if (e.cmd.info == 0)
                    goto out;
                GObjectType& t = ::getObjectType(e.cmd.info);
                if (t.isExistant(e.cmd.info2)) {
                    t.setCurrent(e.cmd.info2);
                    return_value = e.cmd.info;
                    goto out;
                }
            }
        }
    }
 out:
    ui_root->close(chart_w);
    return return_value;
}
