/**
  *  \file client/cscr-ch.cc
  *  \brief Control Screens -- Starchart
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/cscreen.h"
#include "gfx/palette.h"
#include "gfx/clip.h"
#include "gfx/gfx.h"
#include "game/lock.h"

/****************************** WChartWidget *****************************/

/** \class WChartWidget
    \brief Widget displaying a static piece of the starmap

    This just displays a static piece of starchart, without any user
    interaction. */

/** Constructor. Call setCenterPosition() to set position to display. */
WChartWidget::WChartWidget()
{
    viewport.center = GPoint(2000, 2000);
    viewport.mult = viewport.divi = 1;
    viewport.vec_id = 0;
}

WChartWidget::~WChartWidget()
{ }

void
WChartWidget::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    viewport.view = getExtent();
    viewport.init();            // so that drawPre/drawPost can use it

    can->drawBar(getExtent(), standard_colors[COLOR_BLACK], standard_colors[COLOR_BLACK],
                 SOLID_FILL, SDL_ALPHA_OPAQUE);
    drawPre(can);
    viewport.draw(can.ptr(), getChartOpts(true, viewport.mult, viewport.divi),
                  getDisplayedTurn(),
                  getDisplayedTurn().getChartBBox());
    drawPost(can);
    ui.update(getExtent());
}

/** Set position of the center of this chart widget. Will update the
    display. */
void
WChartWidget::setCenterPosition(GPoint pt)
{
    if (pt == viewport.center)
        return;
    viewport.center = pt;
    drawWidget(false);
}

/** Pre-draw hook. Override in descendant classes to draw something
    below the map. */
void
WChartWidget::drawPre(Ptr<GfxCanvas> /*can*/)
{ }

/** Post-draw hook. Override in descendant classes to draw something
    atop the map. */
void
WChartWidget::drawPost(Ptr<GfxCanvas> /*can*/)
{ }

/** Zoom in. \return true iff zoom factor actually changed. */
bool
WChartWidget::zoomIn()
{
    if (viewport.zoomIn(getDisplayedTurn().getChartBBox())) {
        drawWidget(false);
        return true;
    }
    return false;
}

/** Zoom out. \return true iff zoom factor actually changed. */
bool
WChartWidget::zoomOut()
{
    if (viewport.zoomOut(getDisplayedTurn().getChartBBox())) {
        drawWidget(false);
        return true;
    }
    return false;
}

/************************** WScannerChartWidget **************************/

WScannerChartWidget::WScannerChartWidget()
    : scan(getViewport().center)
{ }

WScannerChartWidget::~WScannerChartWidget()
{ }

void
WScannerChartWidget::drawScannerXor(Ptr<GfxCanvas> can)
{
    Ptr<GfxCanvas> can1 = new GfxClipRectFilter(extent, new GfxXorFilter(can.ptr()));
    drawScanner(can1);
}

void
WScannerChartWidget::drawEndBig(GfxCanvas* can, GfxParameters& parm, int x, int y)
{
    drawVLine(can, parm, x-3, y-2, y+2);
    drawVLine(can, parm, x+3, y-2, y+2);
    drawHLine(can, parm, x-2, y-3, x+2);
    drawHLine(can, parm, x-2, y+3, x+2);
}

void
WScannerChartWidget::drawEndSmall(GfxCanvas* can, GfxParameters& parm, int x, int y)
{
    can->drawPixel(x+1, y, parm.color, SDL_ALPHA_OPAQUE);
    can->drawPixel(x-1, y, parm.color, SDL_ALPHA_OPAQUE);
    can->drawPixel(x, y+1, parm.color, SDL_ALPHA_OPAQUE);
    can->drawPixel(x, y-1, parm.color, SDL_ALPHA_OPAQUE);
}

void
WScannerChartWidget::drawScanner(Ptr<GfxCanvas> can)
{
    GfxParameters parm;
    parm.color = standard_colors[COLOR_YELLOW];
    const GChartViewport& vp = getViewport();
    void (*f)(GfxCanvas*,GfxParameters&,int,int) = vp.mult >= vp.divi ? drawEndBig : drawEndSmall;

    const int xm = vp.getCenterX();
    const int ym = vp.getCenterY();
    GfxPoint end = vp.scaleXY(scan);

    f(can, parm, xm, ym);
    if (end.x != xm || end.y != ym) {
        drawLine(can, parm, xm, ym, end.x, end.y);
        f(can, parm, end.x, end.y);
    }
}

/** Synchronize scanner beam with underlying data.
    \param dblclick true if reason of call is a double-click; false if
    normal movement only */
void
WScannerChartWidget::sync(bool /*dblclick*/)
{ }

void
WScannerChartWidget::drawPre(Ptr<GfxCanvas> can)
{
    const GfxPoint m = getViewport().getCenter();

    GfxParameters parm;
    parm.color = standard_colors[COLOR_BLUEBLACK];

    drawHLine(can, parm, m.x - 10, m.y,      m.x + 11);
    drawVLine(can, parm, m.x,      m.y - 10, m.y + 11);
}

void
WScannerChartWidget::drawPost(Ptr<GfxCanvas> can)
{
    drawScannerXor(can);
}

void
WScannerChartWidget::setScannerPosition(GPoint pt)
{
    const GChartViewport& vp = getViewport();
    const GPoint m           = vp.center;
    int xw = vp.unscale(extent.w) / 2;
    int yw = vp.unscale(extent.h) / 2;
    if (pt.x < m.x - xw)
        pt.x = m.x - xw;
    if (pt.x > m.x + xw)
        pt.x = m.x + xw;
    if (pt.y < m.y - yw)
        pt.y = m.y - yw;
    if (pt.y > m.y + yw)
        pt.y = m.y + yw;
    if (pt != scan) {
        if (hasState(st_Exposed)) {
            Ptr<GfxCanvas> can = owner->getCanvasFor(this);
            can->lock();
            drawScannerXor(can);
            scan = pt;
            sync(false);
            drawScannerXor(can);
            can->unlock();
            can->update(extent);
            sig_moved.raise();
        } else {
            scan = pt;
            sync(false);
            sig_moved.raise();
        }
    }
}

void
WScannerChartWidget::setCenterPosition(GPoint pt)
{
    const GChartViewport& vp = getViewport();
    if (pt != vp.center) {
        scan = pt;
        WChartWidget::setCenterPosition(pt);
        sig_moved.raise();
    } else {
        setScannerPosition(pt);
    }
}

void
WScannerChartWidget::moveKey(int dx, int dy)
{
    GPoint pt = getScannerPosition();
    setScannerPosition(GPoint(pt.x + dx, pt.y + dy));
}

bool
WScannerChartWidget::handleKey(UIEvent::Key k)
{
    switch (k.code) {
     case SDLK_LEFT:
        moveKey(k.arg ? -k.arg : -10, 0);
        return true;
     case SDLK_RIGHT:
        moveKey(k.arg ? k.arg : 10, 0);
        return true;
     case SDLK_UP:
        moveKey(0, k.arg ? k.arg : 10);
        return true;
     case SDLK_DOWN:
        moveKey(0, k.arg ? -k.arg : -10);
        return true;
     case ss_Shift + SDLK_LEFT:
        moveKey(-1, 0);
        return true;
     case ss_Shift + SDLK_RIGHT:
        moveKey(1, 0);
        return true;
     case ss_Shift + SDLK_UP:
        moveKey(0, 1);
        return true;
     case ss_Shift + SDLK_DOWN:
        moveKey(0, -1);
        return true;
     case ss_Ctrl + SDLK_LEFT:
        moveKey(-100, 0);
        return true;
     case ss_Ctrl + SDLK_RIGHT:
        moveKey(100, 0);
        return true;
     case ss_Ctrl + SDLK_UP:
        moveKey(0, 100);
        return true;
     case ss_Ctrl + SDLK_DOWN:
        moveKey(0, -100);
        return true;
     case '+':
        if (hasState(sc_Zoomable))
            zoomIn();
        return true;
     case '-':
        if (hasState(sc_Zoomable))
            zoomOut();
        return true;
     case ' ':
     case SDLK_RETURN:
     case ss_Ctrl + ' ':
     case ss_Ctrl + SDLK_RETURN:
        {
            const GChartViewport& vp = getViewport();
            GLockData locker(getScannerPosition(),
                             GPoint(vp.unscaleX(extent.x),
                                    vp.unscaleY(extent.y + extent.h - 1)),
                             GPoint(vp.unscaleX(extent.x + extent.w - 1),
                                    vp.unscaleY(extent.y)),
                             (k.code & ss_Ctrl) != 0);
            findItem(getDisplayedTurn(), locker,
                     (k.code & ~ss_Ctrl) == ' ' ? lm_Left : lm_Right);
            setScannerPosition(locker.getPoint());
        }
        return true;

        /* z
           m c-m */
    }
    return false;
}

bool
WScannerChartWidget::handleEvent(const UIEvent& e, bool second_pass)
{
    /* nothing selected -> no input accepted */
    /* FIXME: out-of-band signalling? */
    if (!isValid())
        return WChartWidget::handleEvent(e, second_pass);

    if (e.type == ev_Mouse && extent.contains(e.mouse.x, e.mouse.y)) {
        const GChartViewport& vp = getViewport();
        select();
        if (e.mouse.b || hasState(sc_KeyboardMode)) {
            if (e.mouse.b & mb_Double) {
                sync(true);
            } else {
                GLockData locker(GPoint(vp.unscaleX(e.mouse.x),
                                        vp.unscaleY(e.mouse.y)),
                                 GPoint(vp.unscaleX(extent.x),
                                        vp.unscaleY(extent.y + extent.h - 1)),
                                 GPoint(vp.unscaleX(extent.x + extent.w - 1),
                                        vp.unscaleY(extent.y)),
                                 0);
                if (e.mouse.b & mb_Left)
                    findItem(getDisplayedTurn(), locker, lm_Left);
                else if (e.mouse.b & mb_Right)
                    findItem(getDisplayedTurn(), locker, lm_Right);

                setScannerPosition(locker.getPoint());
            }
        }
        return true;
    } else if (e.type == ev_Key && hasState(sc_KeyboardMode | sc_Keyboard)) {
        return handleKey(e.key);
    } else {
        return WChartWidget::handleEvent(e, second_pass);
    }
}

void
WScannerChartWidget::getLayoutInfo(LayoutInfo& info)
{
    // FIXME? Probably preferred size should be different
    info.flags = li_Flexible;
    info.min_size = info.pref_size = GfxPoint(100,100);
}
