/**
  *  \file client/cscr-sel.cc
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/cscreen.h"
#include "ui/uiglobal.h"
#include "ui/widgets.h"
#include "gfx/palette.h"
#include "util/nls.h"
#include "ui/layout.h"
#include "cpluslib/holder.h"

/************************* WObjectSelectionTile *************************/

/*! \class WObjectSelectionTile
    \brief Interior of "Select-a-$THING" window

    This one provides the main widget of a select-a-thing window:
    browsing with up/down, selection.

    Derived classes must override its draw() method to do something
    useful.

    \see doSelectWindow */

WObjectSelectionTile::WObjectSelectionTile(GObjectType& atype)
    : WTile(GfxPoint(25 * font_heights[FONT_NORMAL],
                     10 * font_heights[FONT_NORMAL])),
      type(atype), current(atype.getCurrent())
{ }

void
WObjectSelectionTile::select(int id)
{
    if (id != current) {
        current = id;
        update();
    }
}

bool
WObjectSelectionTile::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Command) {
        switch (e.cmd.code) {
         case cm_Prev:
            UIWidget::select();
            select(type.getPrev(current, false));
            return true;
         case cm_Next:
            UIWidget::select();
            select(type.getNext(current, false));
            return true;
         case cm_PrevMarked:
            UIWidget::select();
            select(type.getPrev(current, true));
            return true;
         case cm_NextMarked:
            UIWidget::select();
            select(type.getNext(current, true));
            return true;
         case cm_CS_ToggleSel:
            UIWidget::select();
            type.toggleMark(current);
            update();
            return true;
         case cm_Enter:
            stop(current);
            return true;
         case cm_Escape:
            stop(0);
            return true;
        }
    }
    return WTile::handleEvent(e, second_pass);
}

/** Do "select-a-$THING" window. Constructs the window and runs it.
    \param inter  instance of WObjectSelectionTile specialised for displaying the desired object type
    \param title  for use in window title etc. */
int
doSelectWindow(WObjectSelectionTile& inter, string_t title)
{
    static const UIKeyTranslator::Init init[] = {
        { SDLK_UP,           cm_Prev },
        { SDLK_DOWN,         cm_Next },
        { ss_Ctrl+SDLK_UP,   cm_PrevMarked },
        { ss_Ctrl+SDLK_DOWN, cm_NextMarked },
        { '.',               cm_CS_ToggleSel }
    };
    inter.setTheme(WTile::tt_LightGray);

    ObjectHolder holder;
    // Hooray. Unlike g++, OpenC++ would at least have parsed this right.
    UIWindow win((0, GfxRect()), "Select " + title,
                 UIWindow::BLUE_WINDOW, COLOR_GRAY);
    UIGroup h(new UIHBoxLayout(5));
    win.setLayout(new UIVBoxLayout(5));

    win.add(inter);
    win.add(h);
    win.add(holder.add(new UIKeyTranslator(init, countof(init))));
    h.add(holder.add(new UIButton(cm_Prev,  "\036",            '-')));
    h.add(holder.add(new UIButton(cm_Next,  "\037",            '+')));
    h.add(holder.add(new UIButton(cm_Enter, _("ENTER - View"), SDLK_RETURN)));
    h.add(holder.add(new UIButton(cm_Print, _("P - Print"),    'P')));
    h.add(holder.add(new UISpacer()));
    h.add(holder.add(new UIButton(cm_Escape, _("ESC"),         SDLK_ESCAPE)));
    h.add(holder.add(new UIButton(cm_Help,  "H",               'H')));
    win.pack();
    win.setState(UIWindow::st_Modal, true);

    ui_root->centerWidget(win);
    ui_root->add(win);
    int rv = ui_root->run();
    ui_root->close(win);

    return rv;
}
