/**
  *  \file client/cscreen.cc
  *  \brief Control Screens, and Related Things
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/cscreen.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "gfx/palette.h"
#include "ui/widgets.h"
#include "ui/layout.h"
#include "gfx/clip.h"
#include "ui/util.h"
#include "resmgr/resmgr.h"
#include "game/planet.h"
#include "game/ship.h"
#include "client/playerscr.h"
#include "cpluslib/format.h"
#include "client/chartusr.h"
#include "client/dlg-list.h"

/********************************* WTile *********************************/

/** \class WTile
    \brief Widget displaying certain data (hooray.)

    A tile displays certain data, usually text. In particular, tiles
    can be moved, rearranged, and have their background changed
    (setTheme). This class provides an interface for that, and
    convenience functions for implementing actual display.

    Data is split into a "fixed" and a "variable" portion.

    (In PCC 1.x, the distinguishing property of a "Tile" was that it
    was freely moveable. In PCC 2, all widgets are moveable.)

    \todo maybe the look&feel should be moved into a class instead of
    a switch(). */

WTile::WTile(GfxPoint asize, int id, int when)
    : UIWidget(id), size(asize), update_when(when), theme(tt_DarkGray)
{ }

WTile::WTile(int wi, int he, int when)
    : UIWidget(0), size(wi, he), update_when(when), theme(tt_DarkGray)
{ }

static void
initParameters(GfxParameters& parm)
{
    parm.just_x = parm.just_y = 0;
}

void
WTile::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    parm.bgcolor = GFX_TRANSPARENT;
    initParameters(parm);
    drawFixed(can);
    initParameters(parm);
    drawData(can);
    ui.update(getExtent());
}

void
WTile::setTheme(Theme t)
{
    if (t != theme) {
        theme = t;
        drawWidget(false);
    }
}

void
WTile::update(int n)
{
    if (!n || !update_when || (n & update_when))
        update();
}

void
WTile::update()
{
    Ptr<GfxCanvas> can = getOwner()->getCanvasFor(this);
    if (theme == tt_Transparent) {
        /* this one is special. */
        can->lock();
        drawBackground(can, GfxRect(0, 0, extent.w, extent.h));
        parm.bgcolor = GFX_TRANSPARENT;
        initParameters(parm);
        drawFixed(can);
        initParameters(parm);
        drawData(can);
        can->unlock();
        can->update(extent);
        return;
    }

    switch(theme) {
     case tt_LightGray:
        parm.bgcolor = standard_colors[COLOR_GRAY];
        break;
     case tt_DarkGray:
        parm.bgcolor = standard_colors[COLOR_DARK];
        break;
     case tt_Transparent: /* hey compiler, don't warn me here */;
    }
    can->lock();
    initParameters(parm);
    drawData(can);
    can->unlock();
    can->update(extent);
}

/** Draw fixed data. Override this in derived classes to display data
    which does not change during the lifetime of this tile. */
void
WTile::drawFixed(Ptr<GfxCanvas>)
{ }

/** Draw variable data. Override this in derived classes to display
    data which does change during the lifetime of the tile. */
void
WTile::drawData(Ptr<GfxCanvas>)
{ }

/** Create controls belonging to this tile. Controls will be added to
    the group. The group will not have a layout manager, so use the
    two-argument add() version. */
void
WTile::addControls(UIGroup& /*g*/, ObjectHolder& /* holder */)
{ }

void
WTile::addControl(UIGroup& g, GfxRect r, UIWidget& w)
{
    r.moveBy(extent.x, extent.y);
    g.add(r, w);
}

void
WTile::setColor(Color c)
{
    static const uint8 color_table[3][5] = {
        { COLOR_GRAY,  COLOR_GREEN,      COLOR_YELLOW,       COLOR_RED, COLOR_WHITE },
        { COLOR_BLACK, COLOR_GREENBLACK, COLOR_BRIGHTORANGE, COLOR_RED, COLOR_WHITE },
        { COLOR_GRAY,  COLOR_GREEN,      COLOR_YELLOW,       COLOR_RED, COLOR_WHITE }
    };
    parm.color = standard_colors[color_table[theme][c]];
}

/** Draw text. Wrapper for ::outText, this one honors the theme settings.
    \param f   font number
    \param x,y position, relative coordinates */
void
WTile::outText(GfxCanvas* can, int f, int x, int y, string_t text)
{
    ::outText(can, parm, fonts[f], x + extent.x, y + extent.y, text);
}

/** Draw text. Wrapper for ::outTextF, this one honors the theme settings.
    \param f   font number
    \param x,y position, relative coordinates
    \param lim write to a box exactly this wide (truncate/fill as needed) */
void
WTile::outTextF(GfxCanvas* can, int f, int x, int y, int lim, string_t text)
{
    ::outTextF(can, parm, fonts[f], x + extent.x, y + extent.y, lim, text);
}

/** Draw background. Derived drawData()/drawFixed() use this to clear
    out regions of the background. \c r is in relative coordinates. */
void
WTile::drawBackground(GfxCanvas* can, GfxRect r)
{
    r.moveBy(extent.x, extent.y);
    r.intersect(extent);
    switch(theme) {
     case tt_LightGray:
        can->drawBar(r, standard_colors[COLOR_GRAY], GFX_TRANSPARENT,
                     SOLID_FILL, SDL_ALPHA_OPAQUE);
        break;
     case tt_DarkGray:
        can->drawBar(r, standard_colors[COLOR_DARK], GFX_TRANSPARENT,
                     SOLID_FILL, SDL_ALPHA_OPAQUE);
        break;
     case tt_Transparent:
        {
            Ptr<GfxCanvas> c1 = new GfxClipRectFilter(r, can);
            UpdateInfo ui(can->getSurface());
            getOwner()->drawBelow(this, c1, ui);
            // ui.update() -- no, we do that later
        }
        break;
    }
}

// GfxPoint
// WTile::getMinSize()
// {
//     return size;
// }

// GfxPoint
// WTile::getMaxSize()
// {
//     return GfxPoint(INT_MAX, INT_MAX);
// }

// GfxPoint
// WTile::getPreferredSize()
// {
//     return size;
// }

void
WTile::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.min_size = info.pref_size = size;
}

bool
WTile::handleEvent(const UIEvent& e, bool second_pass)
{
//     if (e.type == ev_Command && e.cmd.code == cm_Update)
//         update();
    return UIWidget::handleEvent(e, second_pass);
}

/***************************** WControlScreen ****************************/

/*
  Layout:

    ,-----------------------------------.
    |              ,------------------. |   <-- borders are 5 pixels
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              |                  | |
    |              `------------------' |  <-- tile_new_y  \ 3 lines for
    |              [                  ] |  ___ chart_y2    / scanner widget
    |                                   |
    |                                   |    8 lines free here for tiles
    |                                   |
    |                                   |
    `-----------------------------------'  ---

    |  320 pixels  |
                chart_x1
 */


WControlScreen::WControlScreen(GObjectType& atype, int screen_id, int update_when)
    : scanner(),
      screen_id(screen_id),
      update_when(update_when),
      pos_changed_cookie(atype.sig_pos_changed.add(this, &WControlScreen::onPosChanged)),
      type(atype), pic_id(0), pos(2000, 2000),
      chart(0),
      controls()
{
    setExtent(GfxRect(ui_screen->getSurface()), true);
    setGeometry();

    controls.setExtent(extent);
    add(controls);

    const int scanner_height = 2*font_heights[FONT_NORMAL];
    add(GfxRect(chart_x1+2,                chart_y2 + 5,
                extent.w - chart_x1 - 5-2, scanner_height),
        scanner);
}

WControlScreen::~WControlScreen()
{
    type.sig_pos_changed.remove(pos_changed_cookie);
}

void
WControlScreen::setGeometry()
{
    tile_width = 315;           // FIXME: should probably be related to font size somehow
    tile_new_y = extent.h - 8*font_heights[FONT_NORMAL];
    tile_y     = 5;
    tile_x     = 5;
    chart_y2   = tile_new_y - 3*font_heights[FONT_NORMAL];
    chart_x1   = tile_width + tile_x + 5;
}

void
WControlScreen::setData(ResID apic, GPoint apos,
                        const string_t& atit, const string_t& adesc)
{
    pic_id = apic;
    pos    = apos;
    title  = atit;
    desc   = adesc;
    chart->setCenterPosition(apos);
}

// void
// WControlScreen::addControl(Ptr<UIWidget> w)
// {
//     controls->add(w);
// }

void
WControlScreen::addTile(WTile& tile, const char* title)
{
    addTile(holder.add(new WHeadingTile(title)));
    addTile(tile);
}

void
WControlScreen::addTile(WTile& tile)
{
    const GfxPoint& tile_size = tile.getSize();
    if (tile_y + tile_size.y > extent.h - 5) {
        /* start new column */
        tile_y = tile_new_y;
        tile_x += tile_width;
    }
    tile.setExtent(GfxRect(tile_x, tile_y, tile_width, tile_size.y), true);
    tile.addControls(controls, holder);
    tile_y += tile_size.y;

    sig_changed.add(&tile, (void (WTile::*)(int)) &WTile::update);
    add(tile);
}

void
WControlScreen::addChart(WScannerChartWidget& chart)
{
    this->chart = &chart;
    add(GfxRect(chart_x1, 5, extent.w - chart_x1 - 5, chart_y2 - 5), chart);
    scanner.setChart(chart);
}

void
WControlScreen::onPosChanged(int /*id*/)
{
    update();
    sig_changed.raise(0);
}

void
WControlScreen::setCurrent(int id)
{
    if (id == type.getCurrent() || !type.isExistant(id))
        return;
    type.setCurrent(id);
}

void
WControlScreen::drawInterior(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    // FIXME: optimize!
    drawSolidBar(can, extent, standard_colors[COLOR_DARK]);
    drawFrameUp(can, extent);
    drawFrameDown(can, GfxRect(chart_x1-1, 5-1,        extent.w - chart_x1 - 3, chart_y2 - 5+2));
    drawFrameDown(can, GfxRect(chart_x1-1, chart_y2+4, extent.w - chart_x1 - 3, 2*font_heights[FONT_NORMAL]+4));
    ui.update(extent);
}

bool
WControlScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch(e.key.code) {
         case SDLK_PAGEUP:
         case '-':
            setCurrent(type.getPrev(type.getCurrent(), false));
            return true;
         case SDLK_PAGEDOWN:
         case '+':
            setCurrent(type.getNext(type.getCurrent(), false));
            return true;
         case SDLK_PAGEUP | ss_Ctrl:
         case '-' | ss_Ctrl:
            setCurrent(type.getPrev(type.getCurrent(), true));
            return true;
         case SDLK_PAGEDOWN | ss_Ctrl:
         case '+' | ss_Ctrl:
            setCurrent(type.getNext(type.getCurrent(), true));
            return true;
         case SDLK_ESCAPE:
            stop(0);
            return true;
         case '.':
            type.toggleMark(type.getCurrent());
            type.getTurn().doUpdates();
            return true;
         case SDLK_F1:
            doScreen(1);
            return true;
         case SDLK_F1 + ss_Shift:
         case SDLK_F1 + ss_Ctrl:
            if (chart->isValid()) {
                GObjectId id = doListShips(((e.key.code & ss_Ctrl)
                                            ? chart->getScannerPosition()
                                            : pos),
                                           ls_EarlyExit,
                                           (screen_id==1 || screen_id==11) ? type.getCurrent() : 0,
                                           _("Visual Scanner"),
                                           _("Go to"),
                                           0 /* FIXME */);
                if (id.isValid())
                    putCommand(cm_CS_GotoCS, 1, id.getId());
            }
            return true;
         case SDLK_F2:
            doScreen(2);
            return true;
         case SDLK_F2 + ss_Shift:
            putCommand(cm_CS_GotoCS, 2, type.getTurn().getPlanetAt(chart->getCenterPosition(), true));
            return true;
         case SDLK_F2 + ss_Ctrl:
            putCommand(cm_CS_GotoCS, 2, scanner.getScannedPlanetId(false));
            return true;
         case SDLK_F3:
            doScreen(3);
            return true;
         case SDLK_F3 + ss_Shift:
            putCommand(cm_CS_GotoCS, 3, type.getTurn().getPlanetAt(chart->getCenterPosition(), true));
            return true;
         case SDLK_F3 + ss_Ctrl:
         case SDLK_F8 + ss_Ctrl:
            putCommand(cm_CS_GotoCS, 3, scanner.getScannedPlanetId(false));
            return true;
         case SDLK_F4:
         case SDLK_F4 + ss_Shift:
            if (chart->isValid())
                setStarchartPosition(chart->getCenterPosition());
            stop(4);
            return true;
         case SDLK_F4 + ss_Ctrl:
            if (chart->isValid()) {
                setStarchartPosition(chart->getScannerPosition());
                stop(4);
            }
            return true;
         case 'l':
            if (chart->isValid()) {
                GObjectId id = doListShips(chart->getScannerPosition(),
                                           ls_IncludeForeignShips,
                                           0,
                                           _("Visual Scanner"),
                                           _("Go to"),
                                           0 /* FIXME */);
                if (id.isValid())
                    putCommand(cm_CS_GotoCS, 1, id.getId());
            }
            return true;
        }
    } else if (e.type == ev_Command) {
        switch(e.cmd.code) {
         case cm_CS_GotoCS:
            if (e.cmd.info == screen_id) {
                if (type.isExistant(e.cmd.info2))
                    setCurrent(e.cmd.info2);
            } else if (e.cmd.info == 0) {
                stop(0);
            } else {
                GObjectType& t = ::getObjectType(e.cmd.info);
                if (t.isExistant(e.cmd.info2)) {
                    t.setCurrent(e.cmd.info2);
                    stop(e.cmd.info);
                }
            }
            return true;
        }
    }
    return UIGroup::handleEvent(e, second_pass);
}

int
WControlScreen::run()
{
    update();
    setState(st_Modal, true);
    ui_root->add(*this);
    int rv = UIGroup::run();
    ui_root->close(*this);
    return rv;
}

void
WControlScreen::updateId(int id)
{
    if (id == type.getCurrent()) {
        int df = type.getDirtyFlags(id);
        if (df & update_when)
            update();
        sig_changed.raise(df);
    }
}

/************************ WControlScreenHeaderTile ***********************/

/*! \class WControlScreenHeaderTile
    \brief Header of a Control Screen

    Displays the top-most part of a control screen: name, description,
    picture, buttons. */

WControlScreenHeaderTile::WControlScreenHeaderTile(WControlScreen& acon, int buttons, int update)
      // FIXME: magic numbers
    : WTile(GfxPoint(315, 145), update), con(acon), buttons(buttons)
{ }

void
WControlScreenHeaderTile::drawData(Ptr<GfxCanvas> can)
{
    // FIXME: magic numbers
    int wi = (buttons & bName) ? extent.w-20 : extent.w;
    setColor(tc_White);
    outTextF(can, FONT_TITLE, 0, 0, wi, con.title);
    setColor(tc_Yellow);
    outTextF(can, FONT_NORMAL, 0, 25, extent.w, con.desc);

    drawFrameDown(can, GfxRect(extent.x + 108, extent.y + 45, 107, 95));

    Ptr<GfxPixmap> pix = getResourcePixmap(con.pic_id, rc_Normal);
    if (pix)
        pix->blitSized(can, extent.x + 109, extent.y + 46, 105, 93);

    drawColorFrame(can, GfxRect(extent.x + 108, extent.y + 45, 107, 95),
                   con.getType().isMarked(con.getType().getCurrent()) ? cf_Yellow : cf_None,
                   false);
}

void
WControlScreenHeaderTile::addControls(UIGroup& g, ObjectHolder& holder)
{
    addControl(g, GfxRect( 5,  45, 30, 25),
               holder.add(new UIButton("F1", SDLK_F1)));
    addControl(g, GfxRect(40,  45, 30, 25),
               holder.add(new UIButton("F2", SDLK_F2)));
    addControl(g, GfxRect( 5,  75, 30, 25),
               holder.add(new UIButton("F3", SDLK_F3)));
    addControl(g, GfxRect(40,  75, 30, 25),
               holder.add(new UIButton("F4", SDLK_F4)));
    addControl(g, GfxRect( 5, 105, 30, 25),
               holder.add(new UIButton("F6", SDLK_F5)));
    addControl(g, GfxRect(40, 105, 30, 25),
               holder.add(new UIButton("F7", SDLK_F7)));
    addControl(g, GfxRect(265, 114, 45, 25),
               holder.add(new UIButton("ESC", SDLK_ESCAPE)));
    addControl(g, GfxRect(230, 114, 25, 25),
               holder.add(new UIButton("H", 'h')));
    addControl(g, GfxRect(285, 45, 25, 25),
               holder.add(new UIButton(cm_CS_Xchg, "X", 'x', UIButton::aKey)));
    addControl(g, GfxRect(75, 45, 20, 25),
               holder.add(new UIButton("\036", '-')));
    addControl(g, GfxRect(75, 75, 20, 25),
               holder.add(new UIButton("\037", '+')));
    if (buttons & bSend)
        addControl(g, GfxRect(75, 105, 20, 25),
                   holder.add(new UIButton("I", 'i')));
    if (buttons & bAuto)
        addControl(g, GfxRect(230, 45, 50, 25),
                   holder.add(new UIButton("Auto", SDLK_RETURN)));
    if (buttons & bCycle)
        addControl(g, GfxRect(275, 75, 35, 25),
                   holder.add(new UIButton("\035", SDLK_TAB)));
    if (buttons & bAdd)
        addControl(g, GfxRect(230, 75, 40, 25),
                   holder.add(new UIButton("Add", SDLK_INSERT)));
    if (buttons & bName)
        addControl(g, GfxRect(extent.w-20, 0, 20, 20),
                   holder.add(new UIButton("N", 'n')));
}

/****************************** WHeadingTile *****************************/

/*! \class WHeadingTile
    \brief Section divider

    Displays a constant text string in big font, as a divider between
    several sections on a control screen. */

WHeadingTile::WHeadingTile(const string_t& s)
    : WTile(GfxPoint(315, font_heights[FONT_TITLE])), text(s)
{ }

WHeadingTile::WHeadingTile(const char* s)
    : WTile(GfxPoint(315, font_heights[FONT_TITLE])), text(s)
{ }

void
WHeadingTile::drawFixed(Ptr<GfxCanvas> can)
{
    setColor(tc_White);
    outTextF(can, FONT_TITLE, 0, 0, extent.w, text);
}

/******************************** WScanner *******************************/

WScanner::WScanner()
    : chart(0)
{ }

WScanner::~WScanner()
{ }

void
WScanner::setChart(WScannerChartWidget& achart)
{
    chart = &achart;
    chart->sig_moved.add(this, &WScanner::update);
    update();
}

void
WScanner::update()
{
    if (!chart || !chart->isValid()) {
        sid = xpid = gpid = 0;
        if (owner)
            owner->setStateOf(cm_CS_Xchg, st_Disabled, true);
        return;
    }

    sid  = getDisplayedTurn().getShipAt(chart->getScannerPosition());
    xpid = getDisplayedTurn().getPlanetAt(chart->getScannerPosition());
    if (xpid)
        gpid = xpid;
    else
        gpid = getDisplayedTurn().getGravityPlanetAt(chart->getScannerPosition());
    if (owner)
        owner->setStateOf(cm_CS_Xchg, st_Disabled, (xpid == 0 && gpid == 0 && sid == 0));

    drawWidget(false);
}

void
WScanner::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    const int x = extent.x;
    const int y = extent.y;
    const int n = font_heights[FONT_NORMAL];

    GfxParameters parm;
    parm.color = standard_colors[COLOR_GRAY];
    outText(can, parm, fonts[FONT_NORMAL], x, y,     _("Scan:"));
    outText(can, parm, fonts[FONT_NORMAL], x, y + n, _("Distance:"));

    if (chart) {
        string_t scan, dist;

        parm.color = standard_colors[COLOR_GREEN];
        parm.bgcolor = standard_colors[COLOR_DARK];

        if (!chart->isValid()) {
            scan = dist = "...";
        } else {
            if (xpid > 0)
                scan = getDisplayedTurn().getPlanet(xpid).getName();
            else if (sid > 0)
                scan = getDisplayedTurn().getShip(sid).getFullName();
            else
                scan = format(_("Deep Space (%d,%d)"),
                              chart->getScannerPosition().x,
                              chart->getScannerPosition().y);
            dist = format(_("%.2f ly"),
                          std::sqrt(double(distanceSquaredRaw(chart->getScannerPosition(),
                                                              chart->getCenterPosition()))));
        }
        outTextF(can, parm, fonts[FONT_NORMAL], x + 70, y,   extent.w - 70, scan);
        outTextF(can, parm, fonts[FONT_NORMAL], x + 70, y+n, extent.w - 70, dist);
    }
    ui.update(getExtent());
}

void
WScanner::getLayoutInfo(LayoutInfo& info)
{
    // FIXME?
    info.flags = 0;
    info.min_size = info.pref_size = GfxPoint(2*font_heights[FONT_NORMAL], 310);
}
