/**
  *  \file client/dlg-cargo.cc
  *  \brief Cargo transfer window.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-cargo.h"
#include "client/cscreen.h"
#include "ui/widgets.h"
#include "gfx/palette.h"
#include "util/nls.h"
#include "ui/window.h"
#include "ui/layout.h"
#include "ui/event.h"
#include "ui/msgbox.h"
#include "cpluslib/format.h"
#include "ui/util.h"
#include "cpluslib/holder.h"

namespace {
    class WCargoTransferWindow : public UIWindow {
        GCargoTransfer transfer;
     public:
        WCargoTransferWindow(Ptr<GCargoContainer> left, Ptr<GCargoContainer> right);
        ~WCargoTransferWindow();

        GCargoTransfer& getActor() { return transfer; }
    };
    
    class WCargoLine : public UIFocusGroup {
        GCargoType       type;
        GCargoTransfer&  actor;
        UIButton         le, ri;
        void  showText(GfxCanvas* c, GfxParameters& parm, int side, int x, int y);
     public:
        WCargoLine(int id, GCargoType type, GCargoTransfer& actor);
        void  drawInterior(Ptr<GfxCanvas> can, UpdateInfo& ui);
        void  doLayout();
        void  getLayoutInfo(LayoutInfo& info);
        bool  handleEvent(const UIEvent& e, bool second_pass);
    };

    class WCargoHeader : public UIWidget {
        void  showTitle(GfxCanvas* can, GCargoContainer& con, int x);
     public:
        WCargoHeader();
        void  draw(Ptr<GfxCanvas> can, UpdateInfo& ui);
        void  getLayoutInfo(LayoutInfo& info);
    };
};

/** \class WCargoHeader
    \brief Header line (two green boxes)

    A WCargoHeader displays the header with the names of the cargo
    transfer partners, plus the column headings

    \inv owner must be a WCargoTransferWindow */
WCargoHeader::WCargoHeader()
    : UIWidget()
{
    setState(st_Disabled, true);
}

void
WCargoHeader::showTitle(GfxCanvas* can, GCargoContainer& con, int x)
{
    GfxParameters parm;
    GfxRect toprect(x-10, extent.y, x+190, font_heights[FONT_NORMAL]*3-3);

    drawTiledArea(can, toprect, 0 /* pixmap, actually */,
                  standard_colors[COLOR_GREENSCALE+4], 0 /* alteration */);
    drawFrameDown(can, toprect);

    parm.color = standard_colors[COLOR_WHITE];
    outTextF(can, parm, fonts[FONT_NORMAL], x, extent.y + 5, 180, con.getName());
    drawHLine(can, parm, x, extent.y + font_heights[FONT_NORMAL] + 4, x + 180);

    parm.just_x = 1;
    const int y = extent.y + font_heights[FONT_NORMAL] + 6;
    outText(can, parm, fonts[FONT_NORMAL], x+45, y, _("has"));
    outText(can, parm, fonts[FONT_NORMAL], x+135, y, _("space left"));
}

void
WCargoHeader::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    WCargoTransferWindow& w = *checked_cast<WCargoTransferWindow*>(getOwner());
    showTitle(can, w.getActor().getContainer(0), extent.x + 10);
    showTitle(can, w.getActor().getContainer(1), extent.x + 430);
    ui.update(getExtent());
}

void
WCargoHeader::getLayoutInfo(LayoutInfo& info)
{
    // FIXME: width is same as WCargoLine
    info.flags = li_FlexibleH;
    info.min_size = info.pref_size = GfxPoint(430+190, 3*font_heights[FONT_NORMAL]);
}

/******************************* WCargoLine ******************************/

/** \class WCargoLine
    \brief One line in the cargo window

    Displays amounts for one cargo type, as well as two buttons for
    manipulating it. */

/** Constructor.
    \param id      widget id
    \param type    type of cargo
    \param actor   underlying transaction class */
WCargoLine::WCargoLine(int id, GCargoType type, GCargoTransfer& actor)
    : UIFocusGroup(id), type(type), actor(actor),
      le("\033", UIKey(SDLK_LEFT)),
      ri("\032", UIKey(SDLK_RIGHT))
{
    add(le);
    add(ri);
}

void
WCargoLine::drawInterior(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    GfxParameters parm;
    
    parm.color = standard_colors[COLOR_BLACK];
    
    string_t unit = getCargoUnit(type);
    if (unit.length())
        unit = " [" + unit + "]";
    outTextF(can, parm, fonts[FONT_NORMAL], extent.x + 230, extent.y + 2, 170,
             getCargoName(type) + unit);

    parm.bgcolor = standard_colors[COLOR_GRAY];

    showText(can, parm, 0, extent.x + 10,  extent.y + 2);
    showText(can, parm, 1, extent.x + 430, extent.y + 2);
    
    UIFocusGroup::drawInterior(can, ui); /* calls update */
}

void
WCargoLine::showText(GfxCanvas* can, GfxParameters& parm, int side,
                     int x, int y)
{
    int32 now = actor.getContainer(side).getCurrent(type);
    int32 max = actor.getContainer(side).getMax(type);

    parm.color  = standard_colors[COLOR_GREENBLACK];
    parm.just_x = 2;
    parm.just_y = 0;
    outTextF(can, parm, fonts[FONT_NORMAL], x + 90, y, 90, itoa(now));

    if (max > 20000)
        outTextF(can, parm, fonts[FONT_NORMAL], x + 180, y, 90, _("(unl)"));
    else
        outTextF(can, parm, fonts[FONT_NORMAL], x + 180, y, 90, itoa(max - now));
}

/* Display:

   Item              PCC1                     PCC2

   left              +19                      +10
                       +0  W90   amount        +0 W90
                       +90 W90   maximum       +90 W90
   button            +210  W19                +200 W20
   good name/unit    +240                     +230
   button            +410  W19                +400 W20
   right             +440                     +430          */
void
WCargoLine::doLayout()
{
    le.setExtent(GfxRect(extent.x + 200, extent.y + 2, 20, 16), true);
    ri.setExtent(GfxRect(extent.x + 400, extent.y + 2, 20, 16), true);
    drawWidget(false);
}

void
WCargoLine::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.min_size = info.pref_size = GfxPoint(430+190, 20);
}

bool
WCargoLine::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key && hasState(st_Focused)) {
        switch(e.key.code) {
         case SDLK_LEFT:
            actor.move(type, 10, 1, true, false);
            drawWidget(false);
            return true;
         case SDLK_RIGHT:
            actor.move(type, 10, 0, true, false);
            drawWidget(false);
            return true;
         case ss_Shift + SDLK_LEFT:
            actor.move(type, 1, 1, true, false);
            drawWidget(false);
            return true;
         case ss_Shift + SDLK_RIGHT:
            actor.move(type, 1, 0, true, false);
            drawWidget(false);
            return true;
         case ss_Ctrl + SDLK_LEFT:
            actor.move(type, 100, 1, true, false);
            drawWidget(false);
            return true;
         case ss_Ctrl + SDLK_RIGHT:
            actor.move(type, 100, 0, true, false);
            drawWidget(false);
            return true;
        }
    }
    return UIFocusGroup::handleEvent(e, second_pass);
}

/************************** WCargoTransferWindow *************************/

/** \class WCargoTransferWindow
    \brief Cargo Transfer Window (normal transfer) */

/** Constructor.
    \param left,right cargo transfer traits */
WCargoTransferWindow::WCargoTransferWindow(Ptr<GCargoContainer> left, Ptr<GCargoContainer> right)
    : UIWindow(_("Cargo Transfer"), BLUE_WINDOW, COLOR_GRAY),
      transfer(left, right)
{
    setLayout(new UIVBoxLayout(0, 5));
    setState(st_Modal, true);
}

WCargoTransferWindow::~WCargoTransferWindow()
{ }

/** The order in which cargo types appear on the window. This is the
    same order as used in PCC 1.x. Why we don't just use the order of
    the enumeration? Good question... */
static const GCargoType elements[] = {
    el_Neutronium,  el_Tritanium,  el_Duranium,  el_Molybdenum,
    el_Supplies,    el_Colonists,  el_Money,
    el_Fighters,
    el_Torps,
    GCargoType(el_Torps+1),
    GCargoType(el_Torps+2),
    GCargoType(el_Torps+3),
    GCargoType(el_Torps+4),
    GCargoType(el_Torps+5),
    GCargoType(el_Torps+6),
    GCargoType(el_Torps+7),
    GCargoType(el_Torps+8),
    GCargoType(el_Torps+9)
};

/** Normal cargo transfer, main entry point. Does all of the cargo
    transfer actions between left and right. By convention, left is
    the object the user called the cargo transfer from. */
void
doCargoTransfer(Ptr<GCargoContainer> left, Ptr<GCargoContainer> right)
{
    /* figure out what can be transferred */
    uint32 bits = 0;
    for (GCargoType i = el_FIRST; i < el_COUNT; ++i)
        if (left->canHave(i) && right->canHave(i)
            && (left->getCurrent(i) || right->getCurrent(i)))
            bits |= 1L << i;
    if ((bits & (1L << el_Supplies))
        && ((left->getFlags() | right->getFlags()) & GCargoContainer::can_Sup2mc))
        bits |= 1L << el_Money;

    if (!bits) {
        messageBox(format(_("There is nothing you could transfer to or from %s."), right->getName()),
                   _("Cargo Transfer"));
        return;
    }

    /* create window */
    ObjectHolder holder;
    WCargoTransferWindow win(left, right);
    win.add(holder.add(new WCargoHeader()));
    int line_id = 1000;
    for (unsigned i = 0; i < countof(elements); ++i)
        if (bits & (1L << elements[i]))
            win.add(holder.add(new WCargoLine(line_id++, elements[i], win.getActor())));
    win.add(holder.add(new UIFocusIterator(1000, line_id-1, UIFocusIterator::fi_Vert)));

    UIGroup button_list(new UIHBoxLayout(5));
    button_list.add(holder.add(new UISpacer()));
    button_list.add(holder.add(new UIButton(1, _("OK"), SDLK_RETURN, UIButton::aStop)));
    button_list.add(holder.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE, UIButton::aStop)));
    win.add(button_list);

    /* run it */
    win.pack();
    win.moveTo(0, 0);            // FIXME?
    ui_root->add(win);
    int rv = ui_root->run();
    ui_root->close(win);

    if (rv)
        win.getActor().commit();
    else
        win.getActor().cancel();
}
