/**
  *  \file client/dlg-fcode.cc
  *
  *  Friendly-code input dialog box.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-fcode.h"
#include "gfx/font.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "ui/uiglobal.h"
#include "util/nls.h"

WFCodeListbox::WFCodeListbox(int id, int lines, int width, const GFCodeList& list)
    : UIListbox(id, lines, width, list.getSize()), list(list)
{ }

WFCodeListbox::~WFCodeListbox()
{ }

void
WFCodeListbox::drawPart(Ptr<GfxCanvas> can, int from, int to)
{
    const int fcwidth = 3 * font_ems[FONT_NORMAL];
    GfxParameters parm;
    for (int line = from; line < to; ++line) {
        const int index = getTop() + line;
        const int y     = line * getItemHeight() + extent.y;

        if (index == getCurrent()) {
            parm.color = standard_colors[COLOR_WHITE];
            if (hasState(st_Focused))
                parm.bgcolor = standard_colors[COLOR_BLACK];
            else
                parm.bgcolor = standard_colors[COLOR_GRAY];
        } else {
            parm.color   = standard_colors[COLOR_BLACK];
            parm.bgcolor = standard_colors[COLOR_GRAY];
        }

        if (index >= list.getSize()) {
            drawSolidBar(can, GfxRect(extent.x, y, extent.w, getItemHeight()), parm.bgcolor);
        } else {
            outTextF(can, parm, fonts[FONT_NORMAL], extent.x, y, fcwidth,
                     " " + list[index].getFCode());
            outTextF(can, parm, fonts[FONT_NORMAL], extent.x + fcwidth, y, extent.w - fcwidth,
                     expandRaceNames(list[index].getDescription()));
        }
    }
}

void
WFCodeListbox::setFCode(const string_t& fc)
{
    GFCodeList::iterator i = list.getFCodeByName(fc);
    if (i != list.end())
        scrollTo(i - list.begin(), 1);
}

string_t
WFCodeListbox::getFCode() const
{
    if (getCurrent() < list.getSize())
        return list[getCurrent()].getFCode();
    else
        return string_t();
}

/****************************** WFCodeWindow *****************************/

WFCodeWindow::WFCodeWindow(const string_t& title, const GFCodeList& list)
    : UIWindow(title, BLUE_WINDOW, COLOR_GRAY),
      g(new UIHBoxLayout(5)),
      g2(new UIHBoxLayout(5)),
      hg(new UIHBoxLayout(0)),
      input(3, 10, ss_Alt + 'f', UIInputLine::ilf_NoHi, 1000),
      fclist(1001, 10, 380, list),
      s_text(_("FCode:"), UIStaticText::BlackTitle),
      b_ok(cm_Enter,  _("OK"),     SDLK_RETURN, UIButton::aStop),
      b_esc(cm_Escape, _("Cancel"), SDLK_ESCAPE, UIButton::aStop),
      b_help(cm_Help,   _("Help"),   ss_Alt + 'h', UIButton::aNothing),
      spacer(),
      scrollbar(fclist),
      fiter(1000, 1001, UIFocusIterator::fi_Tab)
{
    setLayout(new UIVBoxLayout(5));

    g.add(s_text);
    g.add(input);

    g2.add(b_ok);
    g2.add(b_esc);
    g2.add(spacer);
    g2.add(b_help);

    hg.add(fclist);
    hg.add(scrollbar);

    add(g);
    add(hg);
    add(g2);
    add(fiter);

    pack();

    fclist.sig_moved.add(this, &WFCodeWindow::scrolled);
    input.sig_changed.add(this, &WFCodeWindow::edited);

    setState(st_Modal, true);
}

WFCodeWindow::~WFCodeWindow()
{ }

void
WFCodeWindow::edited(int)
{
    fclist.setFCode(input.getText());
}

void
WFCodeWindow::scrolled(int)
{
    if (fclist.hasState(st_Focused))
        input.setText(fclist.getFCode());
}

void
WFCodeWindow::setFCode(const string_t& fc)
{
    input.setText(fc);         // triggers /edited/
}
