/**
  *  \file client/dlg-tech.cc
  *  \brief Tech Upgrade dialog
  */

#include "client/dlg-tech.h"
#include "util/nls.h"
#include "gfx/palette.h"
#include "ui/window.h"
#include "ui/util.h"
#include "cpluslib/holder.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "game/global.h"
#include "cpluslib/format.h"

namespace {
    class WTechCostDisplay : public UIWidget {
        GPlanetTransfer& planet;
        GStarbaseTechUpgradeAction& action;
     public:
        WTechCostDisplay(GPlanetTransfer& planet, GStarbaseTechUpgradeAction& action)
            : planet(planet), action(action) { }

        void draw(Ptr<GfxCanvas> can, UpdateInfo& ui);
        void getLayoutInfo(LayoutInfo& info);

        void updateTechBar(int);
        void maybeStop(int);

        // FIXME? unclean, but this is a local class
        UINumberSelector* tech[4];
    };
}

/******************************** WTechBar *******************************/

WTechBar::WTechBar(int id, int low, int high, string_t name)
    : UINumberSelector(id, low, high, 1), name(name)
{ }

WTechBar::~WTechBar()
{ }

GfxRect
WTechBar::getBarPosition() const
{
    GfxRect r = getExtent();
    r.grow(-2, -2);

    int reduce = r.w / 2;
    r.x += reduce;
    r.w -= reduce;
    return r;
}

void
WTechBar::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    static const int dif[7] = { -9, -6, -3, 0, +4, +7, +10 };

    GfxRect r = getBarPosition(), x = getExtent();
    GfxParameters parm;

    /* frame */
    parm.color = hasState(st_Focused) ? standard_colors[COLOR_BLACK] : standard_colors[COLOR_GRAY];
    drawRectangle(can, parm, x);

    /* text */
    parm.color = standard_colors[COLOR_BLACK];
    parm.just_y = 1;
    outTextF(can, parm, fonts[FONT_NORMAL],
             x.x + 3, x.y + x.h/2, r.x - x.x,  // not a hard limit
             name);
    parm.just_x = 2;
    parm.bgcolor = standard_colors[COLOR_GRAY];
    outTextF(can, parm, fonts[FONT_NORMAL],
             r.x - 5, x.y + x.h/2, 2 * font_ems[FONT_NORMAL],
             itoa(getValue()));

    /* the tech bar */
    drawFrameDown(can, r);
    r.grow(-1, -1);

    int xt = r.w * getMin() / 10;
    int xs = r.w * getValue() / 10;

    if (getMin() == getValue()) {
        drawSolidBar(can, GfxRect(r.x, r.y, xt, r.h), standard_colors[COLOR_GREENBLACK]);
    } else {
        drawSolidBar(can, GfxRect(r.x, r.y, xt-10, r.h), standard_colors[COLOR_GREENBLACK]);
        for (int i = 0; i < 6; ++i) {
            drawSolidBar(can, GfxRect(r.x + xt + dif[i], r.y,
                                      dif[i+1]-dif[i],   r.h),
                         standard_colors[COLOR_GREENSCALE6 + i]);
        }
        drawSolidBar(can, GfxRect(r.x + xt + 10, r.y,
                                  xs - xt - 10,  r.h),
                     standard_colors[COLOR_GREEN]);
    }
    if (xs < r.w)
        drawSolidBar(can, GfxRect(r.x + xs, r.y, r.w - xs, r.h), standard_colors[COLOR_BLACK]);

    int xlim = r.w * getMax() / 10;
    if (xlim < r.w)
        can->drawBar(GfxRect(r.x + xlim, r.y, r.w - xlim, r.h),
                     standard_colors[COLOR_GRAY],
                     GFX_TRANSPARENT,
                     LTSLASH_FILL,
                     SDL_ALPHA_OPAQUE);

    ui.update(getExtent());
}

bool
WTechBar::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key && hasState(st_Focused)) {
        if (e.key.code >= '1' && e.key.code <= '9') {
            select();
            setValue(e.key.code - '1');
            return true;
        } else if (e.key.code == '0') {
            select();
            setValue(10);
            return true;
        }
    } else if (e.type == ev_Mouse && e.mouse.b && isInside(e.mouse.x, e.mouse.y)) {
        select();
        setFocus();
        GfxRect r = getBarPosition();
        if (r.contains(e.mouse.x, e.mouse.y))
            setValue(1 + (e.mouse.x - r.x) * 10 / r.w);
        return true;
    }
    return UINumberSelector::handleEvent(e, second_pass);
}

void
WTechBar::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.pref_size = info.min_size = GfxPoint(2*20*10, 20);
}

void
WTechBar::stateChanged(int state, bool enabled)
{
    if (state & st_Focused)
        drawWidget(false);
    UINumberSelector::stateChanged(state, enabled);
}

/**************************** WTechCostDisplay ***************************/

void
WTechCostDisplay::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    GfxRect x = getExtent();
    GfxParameters parm;

    // FIXME: currently, we're using a fixed layout as in 1.x:
    // InitBill(105, 270, 535) => 430 wide
    //   You need___________  You have_____________________
    //   ^0             150^  ^165                     430^
    // ShowCash(105, 290, ...)
    //   Cash           NNNN  XXXX     NNN remaining
    //   Supplies       NNNN  XXXX
    //   ^0             150^  215^     ^230

    const int m = x.x + 165;

    /* Headings */
    parm.color = standard_colors[COLOR_BLACK];
    outText(can, parm, fonts[FONT_NORMAL], x.x, x.y, _("You need:"));
    outText(can, parm, fonts[FONT_NORMAL], m,   x.y, _("You have:"));
    drawHLine(can, parm, x.x, x.y + font_heights[FONT_NORMAL], m - 15);
    drawHLine(can, parm, m,   x.y + font_heights[FONT_NORMAL], x.x + x.w);

    /* Cash line */
    const int y1 = extent.y + font_heights[FONT_NORMAL] * 3/2;
    parm.bgcolor = standard_colors[COLOR_GRAY];
    outTextF(can, parm, fonts[FONT_NORMAL],  x.x, y1, 150, _("Cash:"));
    parm.just_x = 2;
    outTextF(can, parm, fonts[FONT_NORMAL],  m+50, y1, 50, itoa(planet.getCurrent(el_Money) + action.getReservedCash()));
    parm.bgcolor = GFX_TRANSPARENT;
    outTextF(can, parm, fonts[FONT_NORMAL], x.x+150, y1, 150, itoa(action.getCost()));
    parm.just_x = 0;

    /* Supplies line */
    const int y2 = y1 + font_heights[FONT_NORMAL];
    parm.bgcolor = standard_colors[COLOR_GRAY];
    outTextF(can, parm, fonts[FONT_NORMAL],  x.x, y2, 150, _("Supplies:"));
    parm.just_x = 2;
    outTextF(can, parm, fonts[FONT_NORMAL],  m+50, y2, 50, itoa(planet.getCurrent(el_Supplies) + action.getReservedSupplies()));
    parm.bgcolor = GFX_TRANSPARENT;
    outTextF(can, parm, fonts[FONT_NORMAL], x.x+150, y2, 150, "0");
    parm.just_x = 0;

    /* Conclusion */
    parm.bgcolor = standard_colors[COLOR_GRAY];
    const int x3 = extent.x + 230;
    if (action.getMissingCash()) {
        parm.color = standard_colors[COLOR_RED];
        outTextF(can, parm, fonts[FONT_NORMAL], x3, y1, 200,
                 format(_("You need %d mc"), action.getMissingCash()));
        outTextF(can, parm, fonts[FONT_NORMAL], x3, y2, 200,
                 _("and/or supplies more."));
    } else {
        parm.color = standard_colors[COLOR_GREENBLACK];
        outTextF(can, parm, fonts[FONT_NORMAL], x3, y1, 200,
                 format(_("%d mc remaining"), planet.getCurrent(el_Money)));
        outTextF(can, parm, fonts[FONT_NORMAL], x3, y2, 200,
                 format(_("%d kt remaining"), planet.getCurrent(el_Supplies)));
    }

    ui.update(getExtent());
}

void
WTechCostDisplay::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.min_size = info.pref_size = GfxPoint(430, 7 * font_heights[FONT_NORMAL] / 2);
}

void
WTechCostDisplay::updateTechBar(int id)
{
    id -= 100;
    ASSERT(id >= 0 && id < 4);
    ASSERT(action.isValid(id, tech[id]->getValue()));
    action.setTech(id, tech[id]->getValue());
    drawWidget(false);
}

void
WTechCostDisplay::maybeStop(int id)
{
    if (action.isValid())
        stop(cm_Enter);
}

void
doTechUpgradeDialog(GPlanet& planet)
{
    static const char*const tech_names[] = {
        N_("Engines:"),
        N_("Hulls:"),
        N_("Beam Weapons:"),
        N_("Torpedoes:")
    };

    GPlanetTransfer             transfer(planet);
    GStarbaseTechUpgradeAction  action(planet, transfer);

    ObjectHolder h;

    UIWindow win(_("Tech Upgrade"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
    WTechCostDisplay cost_dpy(transfer, action);
    win.setLayout(new UIVBoxLayout(5));
    win.add(h.add(new UIFocusIterator(100, 103, UIFocusIterator::fi_Vert | UIFocusIterator::fi_Tab)));

    for (int i = 0; i < 4; ++i) {
        WTechBar& bar = h.add(new WTechBar(100 + i, action.getMinTech(i), action.getMaxTech(i),
                                           tech_names[i]));
        bar.setValue(action.getTech(i));
        bar.sig_changed.add(&cost_dpy, &WTechCostDisplay::updateTechBar);
        win.add(bar);
        cost_dpy.tech[i] = &bar;
    }
    win.add(cost_dpy);

    UIButton ok_button(cm_Enter,  _("OK"),     SDLK_RETURN, UIButton::aNothing);
    ok_button.sig_fired.add(&cost_dpy, &WTechCostDisplay::maybeStop);

    UIGroup hg(new UIHBoxLayout(5));
    hg.add(h.add(new UISpacer()));
    hg.add(ok_button);
    hg.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE, UIButton::aStop)));
    win.add(hg);
    win.pack();

    win.setState(UIWindow::st_Modal, true);
    ui_root->centerWidget(win);
    ui_root->add(win);
    int rv = ui_root->run();
    ui_root->close(win);
    if (rv == cm_Enter) {
        action.commit();
        transfer.commit();
    } else {
        action.cancel();
        transfer.cancel();
    }
    getDisplayedTurn().doUpdates();
}
