/**
  *  \file client/gdchooser.cc
  *  \brief Game directory and race chooser
  *
  *  This provides two widgets for choosing game and race.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/gdchooser.h"
#include "cpluslib/assert.h"
#include "cpluslib/format.h"
#include "fileio/io.h"
#include "gfx/gfx.h"
#include "gfx/gfxcompl.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "util/console.h"
#include "util/nls.h"
#include "util/except.h"

WGamedirChooser::Entry::Entry(Action a, const string_t& name, Ptr<Directory> entry)
    : a(a), name(name), entry(entry)
{ }

WGamedirChooser::Entry::~Entry()
{ }

/**************************** WGamedirChooser ****************************/

/** \class WGamedirChooser
    \brief Game Directory Selection

    Widget that displays a directory list and lets the user choose from it. */

/** Initialize. Widget starts by displaying current working directory.
    \param id widget id. */
WGamedirChooser::WGamedirChooser(int id)
    : UIListbox(id, 14, 300, 0),
      current(Directory::openDir(Directory::getWorkingDirectory()))
{
    void (WGamedirChooser::*tmp)(int) = &WGamedirChooser::enter;
    sig_dblclick.add(this, tmp);

    scan(scanGamelist, current);
}

/** Initialize.
    \param current   initial directory
    \param oneup     if true, starts by displaying directory's parent with
                     cursor on directory; if false, starts by displaying
                     directory's content
    \param id        widget id */
WGamedirChooser::WGamedirChooser(Ptr<Directory> current, bool oneup, int id)
    : UIListbox(id, 14, 200, 0), current(current)
{
    void (WGamedirChooser::*tmp)(int) = &WGamedirChooser::enter;
    sig_dblclick.add(this, tmp);

    if (oneup) {
        Ptr<Directory> parent = current->getParent();
        if (parent) {
            scan(scanGamelist, parent);
            string_t fn = Directory::getFileName(current->getName());
            for (std::vector<Entry>::size_type i = 0; i < content.size(); ++i) {
                if (content[i].entry && Directory::getFileName(content[i].entry->getName()) == fn) {
                    scrollTo(i, 1);
                    break;
                }
            }
            return;
        }
    }
    scan(scanGamelist, current);
}

WGamedirChooser::~WGamedirChooser()
{ }

void
WGamedirChooser::drawPart(Ptr<GfxCanvas> can, int from, int to)
{
    GfxParameters parm;

    for (int line = from; line < to; ++line) {
        const int index = getTop() + line;
        const int y = extent.y + getItemHeight() * line;

        if (index >= content.size()) {
            drawSolidBar(can, GfxRect(extent.x, y, extent.w, getItemHeight()),
                         standard_colors[COLOR_GDC_BACKGROUND]);
        } else {
            bool r = false;
            if (index == getCurrent()) {
                if (hasState(st_Focused)) {
                    parm.color = standard_colors[COLOR_GDC_FRAME_TEXT];
                    parm.bgcolor = standard_colors[COLOR_GDC_FRAME];
                } else {
                    parm.color   = standard_colors[COLOR_GDC_SELECTED_TEXT];
                    parm.bgcolor = standard_colors[COLOR_GDC_BACKGROUND];
                    r            = true;
                }
            } else {
                parm.color = standard_colors[COLOR_GDC_NORMAL_TEXT];
                parm.bgcolor = standard_colors[COLOR_GDC_BACKGROUND];
            }
            outTextF(can, parm, fonts[FONT_NORMAL], extent.x, y, extent.w,
                     " " + content[index].name);
            if (r)
                drawRectangle(can, parm, GfxRect(extent.x, y, extent.w, getItemHeight()));
        }
    }
}

/** Scan directory content. */
void
WGamedirChooser::scanDirectory(Ptr<Directory> dir)
{
    std::vector<Entry>::size_type first_index = content.size();
    Ptr<Enumeration<Ptr<DirectoryEntry> > > list = dir->getEntries();
    while (list->hasMoreElements()) {
        Ptr<DirectoryEntry> entry = list->getNextElement();
        if (entry->getFileType() == DirectoryEntry::FT_DIRECTORY)
            content.push_back(Entry(aChdir, entry->getBaseName(),
                                    entry->openDir()));
    }
    std::sort(content.begin() + first_index, content.end(), isEntryLT);
}

/** Initialize directory. This will read the directory, and update display
    on success.
    \param with_gamelist try to read gamestat.dat if possible
    \param dir           directory to read */
void
WGamedirChooser::scan(ScanArg arg, Ptr<Directory> dir)
{
    Ptr<Stream> s;
    content.clear();

    if (arg == scanRoots) {
        /* scan roots (=devices) */
        Ptr<Enumeration<Ptr<Directory> > > roots = Directory::getRoots();
        while (roots->hasMoreElements()) {
            Ptr<Directory> dir = roots->getNextElement();
            content.push_back(Entry(aChdir, dir->getName(), dir));
        }
    } else {
        if (Ptr<Directory> p = dir->getParent())
            content.push_back(Entry(aParent, _("[Up one level]"), p));

        if (arg == scanGamelist && (s = dir->openFileNT("gamestat.dat", Stream::C_READ))) {
            /* read gamelist file */
            char buffer[8][206];
            if (s->read(buffer[0], sizeof buffer) == sizeof buffer) {
                /* game list is okay */
                for (int i = 0; i < 8; ++i) {
                    try {
                        // FIXME: I'm unsure whether we should recode here.
                        content.push_back(Entry(aChdir, " " + getBasicStringN(&buffer[i][184], 20),
                                                dir->openSubdir(format("vpwork%d", i+1))));
                    }
                    catch(...) { }
                }
                content.push_back(Entry(aRescanWithoutGamelist,
                                        _("[Directory Content]"), 0));
            } else
                scanDirectory(dir);
        } else
            scanDirectory(dir);

        content.push_back(Entry(aShowRoots, _("[Disks]"), 0));
    }

    current = dir;
    setCurrent(0);
    setSize(content.size());
    adjustPosition();
    drawWidget(false);
}

bool
WGamedirChooser::isEntryLT(const Entry& l, const Entry& r)
{
    return strCaseCmp(l.name, r.name) < 0;
}

/** Enter current directory. The parameter is ignored. \overload */
void
WGamedirChooser::enter(int)
{
    enter();
}

/** Enter current directory. When something goes wrong, tries to recover
    (= stay in current directory). */
void
WGamedirChooser::enter()
{
    if (getCurrent() < content.size()) {
        Ptr<Directory> c = current;
        try {
            switch (content[getCurrent()].a) {
             case aRescanWithoutGamelist:
                scan(scanNoGamelist, current);
                break;
             case aChdir:
             case aParent:
                scan(scanGamelist, content[getCurrent()].entry);
                break;
             case aShowRoots:
                scan(scanRoots, 0);
                break;
            }
        }
        catch(...) {
            /* no matter what happens, try to recover by returning
               to old directory. If that fails, bomb out. */
            scan(c ? scanGamelist : scanRoots, c);
        }
        sig_moved.raise(getId());
    }
}

/** Get current subdirectory (=the directory the cursor is on).
    \returns directory, or 0 if none (e.g. if cursor is on "[Parent]"
    entry */
Ptr<Directory>
WGamedirChooser::getSubdirectory()
{
    if (getCurrent() < content.size() && content[getCurrent()].a == aChdir)
        return content[getCurrent()].entry;
    return 0;
}

/**************************** WGameRaceChooser ***************************/

/** \class WGameRaceChooser
    \brief Race Chooser

    This displays a race list for the user to choose from.
    Derived classes must override scan() to provide actual
    initialisation. */

/** Initialize. Widget will be empty (no valid race)
    \param id widget Id */
WGameRaceChooser::WGameRaceChooser(int id)
    : UIListbox(id, NUM_PLAYERS, 270, 0), avail(0)
{
    setDirectory(0);
}

WGameRaceChooser::~WGameRaceChooser()
{ }

void
WGameRaceChooser::drawPart(Ptr<GfxCanvas> can, int from, int to)
{
    GfxParameters parm;
    for (int line = from; line < to; ++line) {
        const int index = getTop() + line;
        const int y = extent.y + getItemHeight() * line;

        if (index >= NUM_PLAYERS) {
            drawSolidBar(can, GfxRect(extent.x, y, extent.w, getItemHeight()),
                         standard_colors[COLOR_GDC_BACKGROUND]);
        } else {
            bool r = false;
            if (hasState(st_Disabled) || !isAccessible(index)) {
                parm.color = standard_colors[COLOR_GDC_INACTIVE_TEXT];
                parm.bgcolor = standard_colors[COLOR_GDC_BACKGROUND];
            } else if (index == getCurrent()) {
                if (hasState(st_Focused)) {
                    parm.color = standard_colors[COLOR_GDC_FRAME_TEXT];
                    parm.bgcolor = standard_colors[COLOR_GDC_FRAME];
                } else {
                    parm.color = standard_colors[COLOR_GDC_SELECTED_TEXT];
                    parm.bgcolor = standard_colors[COLOR_GDC_BACKGROUND];
                    r = true;
                }
            } else {
                parm.color = standard_colors[COLOR_GDC_NORMAL_TEXT];
                parm.bgcolor = standard_colors[COLOR_GDC_BACKGROUND];
            }

            string_t txt = " " + list.getShortName(index+1);
            if (infos[index].length()) {
                outTextF(can, parm, fonts[FONT_NORMAL], extent.x, y, extent.w - 100, txt);
                outTextF(can, parm, fonts[FONT_NORMAL], extent.x + extent.w - 100, y, 100, infos[index]);
            } else {
                outTextF(can, parm, fonts[FONT_NORMAL], extent.x, y, extent.w, txt);
            }
            if (r)
                drawRectangle(can, parm, GfxRect(extent.x, y, extent.w, getItemHeight()));
        }
    }
}

bool
WGameRaceChooser::isAccessible(int index)
{
    return (avail & (1 << (index+1))) != 0;
}

/** Change to specified directory. This will read the directory's race
    names, and call scan() to initialize the information part.
    \param dir directory. Can be 0 to set the widget into an inactive state */
void
WGameRaceChooser::setDirectory(Ptr<Directory> dir)
{
    /* first, read new race names */
    bool loaded = false;
    if (dir) {
        Ptr<Stream> s = dir->openFileNT("race.nm", Stream::C_READ);
        if (s) {
            try {
                list.load(s);
                loaded = true;
            }
            catch(FileOpenException& e) {
                console << "race.nm: " << e.what() << "\n";
                ui_root->doDeferredRedraws();
            }
            catch(...) { }
        }
    }
    if (!loaded)
        list = racename_data;
    for (int i = 0; i < NUM_PLAYERS; ++i)
        infos[i].clear();

    /* now, scan directory */
    avail = 0;
    if (dir)
        scan(dir);

    setTop(0);
    setSize(NUM_PLAYERS);

    if (avail) {
        setState(st_Disabled, false);
        scrollTo(0, 1);
        drawWidget(false);
    } else {
        setState(st_Disabled, true);
        drawWidget(false);
    }
}

/** Set information for a player. Should be called at most once per player
    by scan().
    \param player   1..NUM_PLAYERS
    \param avail    true iff this race is available. Default is "not available".
    \param text     info text */
void
WGameRaceChooser::setInfo(int player, bool avail, string_t text)
{
    ASSERT(player > 0 && player <= NUM_PLAYERS);
    if (avail)
        this->avail |= 1 << player;
    infos[player-1] = text;
}
