/**
  *  \file main.cc
  *  \brief Client Main Module
  *
  *  This file contains the main function for the planets client.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#include <cstdlib>
#include "util/cmdline.h"
#include "arch/platform.h"
#include "gfx/init.h"
#include "gfx/gfx.h"
#include "util/nls.h"
#include "game/global.h"
#include "util/debug.h"
#include "util/console.h"
#include "game/spec.h"
#include "game/player.h"
#include "game/config.h"
#include "ui/uiglobal.h"
#include "gfx/palette.h"
#include "game/load.h"
#include "client/chart.h"
#include "client/playerscr.h"
#include "ui/msgbox.h"
#include "cpluslib/format.h"
#include "client/chartusr.h"
#include "game/save.h"
#include "game/fcode.h"
#include "util/recode.h"
#include "version.h"

/** Option Parser for pcc-v2. */
class PCC2OptionParser : public StandardCommandLineParser {
    enum { MAX_NON_OPTIONS = 2 };
    int num_nonoptions;
    int player;
    string_t nonopts[MAX_NON_OPTIONS];
    static void help(StandardCommandLineParser&, const Option&);
    static void setCharset(StandardCommandLineParser&, const Option&);
 public:
    PCC2OptionParser();
    void handleNonoption(const string_t& str);
    void handleError(const string_t& msg);
    int getNumNonoptions() const { return num_nonoptions; }
    void parseFinished();

    string_t getGameDirName() const { return nonopts[0]; }
    string_t getRootDirName() const { return nonopts[1]; }
    int getPlayerId() const { return player; }
};

/** Initialize. */
PCC2OptionParser::PCC2OptionParser()
    : num_nonoptions(0), player(0)
{
    static const Option options[] = {
        { "help",    help,       0, false,
          N_(" -help              Show help screen") },
        { "charset", setCharset, 0, true,
          N_(" -charset EXT[:INT] Set character set translation") }
    };
    addOptionSet(options, countof(options));
}

/** Handle Non-Options. The first non-option is the game directory
    name, the second (optional) is the root directory. Player number
    can stand anywhere. */
void
PCC2OptionParser::handleNonoption(const string_t& str)
{
    if (!player) {
        int x = std::atoi(str.c_str());
        if (x > 0 && x <= NUM_PLAYERS) {
            player = x;
            return;
        }
    }
    if (num_nonoptions >= MAX_NON_OPTIONS)
        handleError(_("Too many parameters specified."));
    nonopts[num_nonoptions++] = str;
}

/** Show help screen. */
void
PCC2OptionParser::help(StandardCommandLineParser& par,
                       const StandardCommandLineParser::Option&)
{
    console << _("Usage: pcc-v2 [-options] gamedir [playerid] [rootdir]\n"
                 "\n"
                 "Valid options:\n");
    par.printHelp(console);
    console << _("\n(c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>\n");
    Platform::showOutput();
    std::exit(0);
}

void
PCC2OptionParser::setCharset(StandardCommandLineParser& par,
                             const StandardCommandLineParser::Option&)
{
    string_t cs = par.getArg();
    string_t intrn = "latin1";
    strSplit(cs, cs, intrn, ":");
    CharacterSet ics = Recoder::getCharsetByName(intrn);
    CharacterSet ecs = Recoder::getCharsetByName(cs);
    if (!ics.isValid())
        console.write(LOG_WARN, format(_("Character set `%s' is unknown"), intrn));
    if (!ecs.isValid())
        console.write(LOG_WARN, format(_("Character set `%s' is unknown"), cs));
    Recoder::setCharsets(ics, ecs);

}

/** Error handler. Instead of throwing an exception, show a message
    and bail out. */
void
PCC2OptionParser::handleError(const string_t& str)
{
    console.write(LOG_ERROR, str);
    console.write(LOG_ERROR, _("Try `pcc-v2 --help' for help."));
    Platform::showOutput();
    std::exit(1);
}

/** Parsing finished. Postprocess input. */
void
PCC2OptionParser::parseFinished()
{
    if (getNumNonoptions() == 0)
        handleError(_("No game directory name specified."));
    if (getNumNonoptions() == 1)
        nonopts[1] = Platform::getDataDirectory();
}

string_t
fmtflags(int fl)
{
    string_t rv;
    if (fl & GDirectoryOverview::have_RST)
        rv += " RST";
    if (fl & GDirectoryOverview::have_Gamedir)
        rv += " unpacked";
    if (fl & GDirectoryOverview::have_TRN)
        rv += " TRN";
    if (fl & GDirectoryOverview::have_NewRST)
        rv += " new-RST";
    if (fl & GDirectoryOverview::have_Conflict)
        rv += " *CONFLICT*";
    return rv;
}

int main(int /*argc*/, char** argv)
{
    Platform::initOutput();
    initNLS();
    PCC2OptionParser parser;
    Platform::initOptions(parser);
    initGraphicsOptions(parser);
    parser.parse(argv);

    SDL_Surface* sfc = initGraphics(0 /* FIXME: icon name */,
                                    "Planets Command Center II (v" PCC2_VERSION ")");
    if (!sfc) {
        Platform::showOutput();
        return 1;
    }

    try {
        initDebug();
        initGameGlobals(parser.getGameDirName(), parser.getRootDirName());
        initConfig();
        initSpecification();
        initFCodes();
        initRaceNames();
        initStarchart();
        initUIGlobals(new GfxSurface(sfc, false),
                      Directory::openDir(Platform::getResourceDirectory()));

        GDirectoryOverview ov;
        ov.scan(game_file_dir);

        console.write(LOG_INFO, _("=== Loading... ==="));
        if (int pid = parser.getPlayerId()) {
            if (! (ov.getPlayerFlags(pid) & (GDirectoryOverview::have_Gamedir | GDirectoryOverview::have_RST))) {
                console.write(LOG_ERROR, format(_("No data for player %d available"), pid));
            } else {
                Ptr<Stream> pxy = spec_file_dir->openFileT("xyplan.dat", Stream::C_READ);
                getCurrentTurn().init();
                getCurrentTurn().loadPlanetXY(pxy);
                if (ov.getPlayerFlags(pid) & GDirectoryOverview::have_RST) {
                    Ptr<Stream> rst =
                        game_file_dir->openFileT(format("player%d.rst", parser.getPlayerId()),
                                                 Stream::C_READ);
                    loadResult(getCurrentTurn(), rst, parser.getPlayerId());
                    if (ov.getPlayerFlags(pid) & GDirectoryOverview::have_TRN)
                        loadTurn(getCurrentTurn(),
                                 game_file_dir->openFileT(format("player%d.trn", parser.getPlayerId()),
                                                          Stream::C_READ),
                                 parser.getPlayerId());
                } else {
                    ASSERT(ov.getPlayerFlags(pid) & GDirectoryOverview::have_Gamedir);
                    loadDirectory(getCurrentTurn(), game_file_dir, pid);
                }
                setPlayerIds(pid, pid);
                getCurrentTurn().postprocess();
                doPlayerScreen();

                // saving the game
                console.write(LOG_INFO, _("=== Saving... ==="));
                if (! (ov.getPlayerFlags(pid) & GDirectoryOverview::have_RST))
                    saveDirectory(getCurrentTurn(), game_file_dir, 1 << pid);
                else
                    saveTurns(getCurrentTurn(), game_file_dir, 1 << pid);
            }
        }
    }
    catch(std::exception& e) {
        console.write(LOG_ERROR, string_t(_("Error during initialisation: ")) + e.what());
        Platform::showOutput();
    }
    return 0;
}
