/**
  *  \file client/marks.cc
  *  \brief Marker Drawing
  *
  *  This module provides functions to draw certain icons which are
  *  used in the map and otherwise.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/marks.h"
#include "gfx/gfxcompl.h"
#include "cpluslib/assert.h"

namespace {
/* Must put these into an anonymous namespace to allow forward-declarations. */
#define END -99
#define DEFINE_MARKER(name, height)              \
extern int8 name ## _data[];                     \
TMarkerImage name = { height, name ## _data };   \
int8 name ## _data[]

/* Normal "user" markers */

/* "+", normal size; type code 0 */
DEFINE_MARKER(um_plus, 4) = {
    -3, 0, -2, 0, -1, 0, 1, 0, 2, 0, 3, 0,
    0, -3, 0, -2, 0, -1, 0, 1, 0, 2, 0, 3, END
};

/* "!", normal size; type code 1 */
DEFINE_MARKER(um_exclam, 3) = {
    -1, -3, -1, -2, -1, -1, -1, 0, -1, 2, END
};

/* "x", normal size; type code 2 */
DEFINE_MARKER(um_times, 3) = {
    -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, -2, 2, -1, 1, 1, -1, 2,-2, END
};

/* diamond, normal size; type code 3 */
DEFINE_MARKER(um_diamond, 3) = {
    -2, 0, -1, 1, 2, 0, 1, 1, 0, 2, 1, -1, 0, -2, -1,-1, END
};

/* "p" (flag), normal size; type code 4 */
DEFINE_MARKER(um_flag, 1) = {
    0, 0, 0, -1, 0, -2, 0, -3, 0,-4, 0, -5,
    1, -5, 2, -5, 3, -4, 3, -3, 2, -2, 1, -2, END
};

/* up/down arrow, normal size; type code 5 */
DEFINE_MARKER(um_updown, 4) = {
    -2, 3, -1, 3, 0, 3, 1, 3, 2, 3, -2, -3,
    -1, -3, 0, -3, 1, -3, 2, -3,
    -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, -2, 2,
    -1, 1, 1, -1, 2, -2, END
};

/* left/right arrow, normal size; type code 6 */
DEFINE_MARKER(um_leftright, 3) = {
    -2, -2, -1, -1, 0, 0, 1, 1, 2, 2, -2, 2,
    -1, 1, 1, -1, 2, -2,
    -3, -2, -3, -1, -3, 0, -3, 1, -3, 2, 3, -2,
    3, -1, 3, 0, 3, 1, 3, 2, END
};

/* Small "user" markers */

DEFINE_MARKER(um_small_plus, 2) = {
    -1, 0, 1, 0, 0, -1, 0, 1, 0, 0, END
};

DEFINE_MARKER(um_small_exclam, 1) = {
    0, -1, 0, 0, 0, 2, END
};

DEFINE_MARKER(um_small_times, 1) = {
    0, 0, -1, -1, -1, 1, 1, 1, 1, -1, END
};

DEFINE_MARKER(um_small_diamond, 1) = {
    -1, 0, 1, 0, 0, -1, 0, 1, END
};

DEFINE_MARKER(um_small_flag, 1) = {
    0, 0, 0, -1, 0, -2, 1, -1, 1, -2, END
};

DEFINE_MARKER(um_small_updown, 2) = {
    0, -1, 0, 1, 0, 0, -1, -1, -1, 1, 1, 1, 1, -1, END
};

DEFINE_MARKER(um_small_leftright, 1) = {
    0, 0, -1, -1, -1, 1, 1, 1, 1, -1, -1, 0, 1, 0, END
};

/* Selection markers ("Andreas cross") */

DEFINE_MARKER(selection_1, 2) = {
    0, 0, -1, -1, -1, 1, 1, -1, 1, 1, -2, -1, -2, 1, 2, -1, 2, 1, END
};

DEFINE_MARKER(selection_2, 4) = {
    0, 0, -1, -1, -1, 1, 1, -1, 1, 1, -2, -1, -2, 1, 2, -1, 2, 1, -3, -2, 
    -3, 2, 3, -2, 3, 2, -4, -2, 4, -2, -4, 2, 4, 2, END
};

DEFINE_MARKER(selection_3, 6) = {
    0, 0, -1, -1, -1, 1, 1, -1, 1, 1, -2, -1, -2, 1, 2, -1, 2, 1, -3, -2, 
    -3, 2, 3, -2, 3, 2, -4, -2, 4, -2, -4, 2, 4, 2, -5, -3, 5, -3, -5, 3, 
    5, 3, -6, -3, 6, -3, -6, 3, 6, 3, END
};

/* Dotted circles (for starchart). These have been constructed
   so that two concentric circles with difference 1 in radius do
   look good. */

DEFINE_MARKER(dotted_1, 1) = {
    -1, -1, -1, 1, 1, 1, 1, -1, END
};

DEFINE_MARKER(dotted_2, 2) = {
    -2, -1, -1, -2, 1, -2, 2, -1, 2, 1, 1, 2, -1, 2, -2, 1, END
};

DEFINE_MARKER(dotted_3, 3) = {
    -3, -1, -3, 1, 3, -1, 3, 1, -1, -3, -1, 3, 1, -3, 1, 3, END
};

DEFINE_MARKER(dotted_4, 4) = {
    -4, -2, -4, 0, -4, 2, 4, -2, 4, 0, 4, 2, -2, -4, 0, -4, 2, -4, -2, 4,
    0, 4, 2, 4, END
};

DEFINE_MARKER(dotted_5, 5) = {
    -5, -1, -4, -3, -3, -4, -1, -5, 1, -5, 3, -4, 4, -3, 5, -1, 
    5, 1, 4, 3, 3, 4, 1, 5, -1, 5, -3, 4, -4, 3, -5, 1, END
};

DEFINE_MARKER(dotted_6, 6) = {
    -6, -1, -5, -3, -3, -5, -1, -6, 1, -6, 3, -5, 5, -3, 6, -1, 
    6, 1, 5, 3, 3, 5, 1, 6, -1, 6, -3, 5, -5, 3, -6, 1, END
};

DEFINE_MARKER(dotted_7, 7) = {
    -7, -1, -6, -3, -5, -5, -3, -6, -1, -7, 1, -7, 3, -6, 5, -5, 6, -3, 7, -1, 
    7, 1, 6, 3, 5, 5, 3, 6, 1, 7, -1, 7, -6, 3, -5, 5, -3, 6, -7, 1, END
};
} 

/** Draw a Marker.
    \param parm    parameters. We use color and alpha setting */
void
drawMarker(GfxCanvas* can, GfxParameters& parm, const TMarkerImage& marker,
           int x, int y)
{
    const int8* ptr = marker.data;
    while (*ptr != END) {
        can->drawPixel(x + ptr[0], y + ptr[1], parm.color, parm.alpha);
        ptr += 2;
    }
}

/** Draw a dotted circle.
    \param r   radius, in range [0, 7]. */
void
drawDottedCircle(GfxCanvas* can, GfxParameters& parm, int x, int y, int r)
{
    static const TMarkerImage*const markers[] = {
        &dotted_1, &dotted_2, &dotted_3, &dotted_4,
        &dotted_5, &dotted_6, &dotted_7
    };
    if (r <= 0)
        return;
    if (r <= int(countof(markers)))
        drawMarker(can, parm, *markers[r], x, y);
    // else what?
}

/** Draw Selection Marker.
    \param mult,divi  Zoom settings. 1:1 = standard size for big font,
                      1:2 = size for normal font, others = starchart zoom. */
void
drawSelection(GfxCanvas* can, GfxParameters& parm, int x, int y,
              int mult, int divi)
{
    if (mult > divi) {
        /* Zoom > 1 */
        int dx = std::min(6 * mult / divi, 12);
        int dy = dx/2;
        parm.line_style = ~0U;
        parm.line_thickness = 1;
        drawLine(can, parm, x - dx, y - dy, x + dx, y + dy);
        drawLine(can, parm, x - dx, y + dy, x + dx, y - dy);
    } else if (mult*3 > divi*2) {
        /* Zoom > 2/3 */
        drawMarker(can, parm, selection_3, x, y);
    } else if (divi > mult*2) {
        /* Zoom < 1/2 */
        drawMarker(can, parm, selection_1, x, y);
    } else {
        drawMarker(can, parm, selection_2, x, y);
    }
}

/** Get user marker. Returns marker data. Use as
    drawMarker(can, parm, getUserMarker(id, big), x, y). */
const TMarkerImage&
getUserMarker(int id, bool big)
{
    static const TMarkerImage*const user_markers[NUM_USER_MARKERS][2] = {
        { &um_plus,      &um_small_plus },
        { &um_exclam,    &um_small_exclam },
        { &um_times,     &um_small_times },
        { &um_diamond,   &um_small_diamond },
        { &um_flag,      &um_small_flag },
        { &um_updown,    &um_small_updown },
        { &um_leftright, &um_small_leftright }
    };
    ASSERT(id >= 0 && id < NUM_USER_MARKERS);
    return *user_markers[id][!big];
}
