/**
  *  \file client/playerscr.cc
  *  \brief Player Main Screen
  *
  *  Main screen, user interface and multiplexing to other control screens.
  *  PCC 1.x source: cc.pas, scrmux.pas.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <iostream>
#include <SDL.h>
#include "client/screens.h"
#include "client/chartusr.h"
#include "client/cscreen.h"
#include "client/dlg-list.h"
#include "client/playerscr.h"
#include "client/vcrplay.h"
#include "game/global.h"
#include "game/objlist.h"
#include "game/planet.h"
#include "game/ship.h"
#include "gfx/font.h"
#include "gfx/gfxcompl.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/ui.h"
#include "ui/util.h"
#include "ui/widgets.h"
#include "util/nls.h"
#include "cpluslib/holder.h"

class WPixmapButton : public UIButton {
    Ptr<GfxPixmap> pix;
 public:
    WPixmapButton(int aid, string_t aname, UIKey akey, Ptr<GfxPixmap> apix);
    void draw(Ptr<GfxCanvas> can, UpdateInfo& ui);
    void getLayoutInfo(LayoutInfo& info);
};

WPixmapButton::WPixmapButton(int aid, string_t aname,
                             UIKey akey, Ptr<GfxPixmap> apix)
    : UIButton(aid, aname, akey), pix(apix)
{ }

void
WPixmapButton::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    GfxParameters parm;
    int text_height = fonts[FONT_SMALL]->getTextHeight(getName());
    if (pix && extent.h > text_height-2)
        pix->blit(can, extent.x + 1, extent.y + 1,
                  GfxRect(0, 0, extent.w - 2, extent.h - text_height - 2));
    parm.color = standard_colors[COLOR_WHITE];
    parm.bgcolor = standard_colors[COLOR_BLACK];
    outTextF(can, parm, fonts[FONT_SMALL],
             extent.x + 1, extent.y + extent.h - text_height - 1,
             extent.w - 2, getName());
    drawFrameDown(can, extent);
    ui.update(extent);
}

void
WPixmapButton::getLayoutInfo(LayoutInfo& info)
{
    GfxPoint pt (fonts[FONT_SMALL]->getTextWidth(getName()) + 2,
                 fonts[FONT_SMALL]->getTextHeight(getName()) + 2);
    if (pix) {
        pt.x = pix->getWidth() + 2;
        pt.y += pix->getHeight();
    }
    info.flags = 0;
    info.min_size = info.pref_size = pt;
}

/***************************** WPlayerScreen *****************************/

class WPlayerScreen : public UIGroup {
    Ptr<GfxPixmap> pix;
    static UIKeyTranslator::Init keys[];
    ObjectHolder holder;
 public:
    WPlayerScreen();
    void drawInterior(Ptr<GfxCanvas> can, UpdateInfo& ui);
    bool handleEvent(const UIEvent& e, bool second_pass);
};

UIKeyTranslator::Init
WPlayerScreen::keys[] = {
    { SDLK_ESCAPE, cm_Escape }
};

WPlayerScreen::WPlayerScreen()
    : UIGroup(new UIHBoxLayout(5, 5))
{
    setExtent(GfxRect(ui_screen->getSurface()));
    pix = requestPixmap("bgtile");
    UIGroup& g = holder.add(new UIGroup(new UIVBoxLayout(5, 5)));
    g.add(holder.add(new WPixmapButton(1, _("F1 - Starships"), SDLK_F1,
                             requestPixmap("menu.ship"))));
    g.add(holder.add(new WPixmapButton(2, _("F2 - Planets"), SDLK_F2,
                             requestPixmap("menu.planet"))));
    g.add(holder.add(new WPixmapButton(3, _("F3 - Starbases"), SDLK_F3,
                             requestPixmap("menu.base"))));
    g.add(holder.add(new WPixmapButton(4, _("F4 - Starchart"), SDLK_F4,
                             requestPixmap("menu.chart"))));
    g.add(holder.add(new UISpacer()));
    add(g);
    g.pack();
    add(holder.add(new UISpacer()));
    doLayout();
    add(holder.add(new UIKeyTranslator(keys, countof(keys))));
//     add(GfxRect(200, 50, 300, 25),
//         new UIInputLine(100, ss_Alt + 'i', 0, 1234));
//    add(GfxRect(200, 100, 300, 300),
//         new WChartWidget());
//        new WScannerChartWidget());
//     add(GfxRect(200, 50, 300, 25),
//         new UICheckbox(100, "Click me", 'c', 3));

//     Ptr<UIRadioButton::Value> v = new UIRadioButton::Value(0);
//     add(GfxRect(200, 100, 300, 25),
//         new UIRadioButton(200, "Ras", 'r', v, 0));
//     add(GfxRect(200, 130, 300, 25),
//         new UIRadioButton(200, "Dwa", 'd', v, 1));
//     add(GfxRect(200, 160, 300, 25),
//         new UIRadioButton(200, "Tri", 't', v, 2));
}

void
WPlayerScreen::drawInterior(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    drawTiledArea(can, GfxRect(extent.x + 1, extent.y + 1,
                               extent.w - 2, extent.h - 2),
                  pix, standard_colors[COLOR_DARK], 16);
    drawFrameUp(can, extent);
    ui.update(extent);
}

bool
WPlayerScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Command) {
        switch(e.cmd.code) {
         case cm_Escape:
            stop(-1);
            return true;
         case 1:
         case 2:
         case 3:
         case 6:
         case 10:
            doMainLoop(-e.cmd.code);
            return true;
         case 4:
            doMainLoop(4);
            return true;
        }
    } else if (e.type == ev_Key) {
        if (e.key.code == 'v') {
//             playVcrs(getDisplayedTurn().getVcrs());
            return true;
//         } else if (e.key.code == 'm') {
//             Ptr<GOutbox> o = getDisplayedTurn().getOutbox();
//             for (int i = 0; i < o->getCount(); ++i)
//                 std::cout << "Message #" << i << ":\n"
//                           << o->getText(i) << "\n";
//             return true;
        } else if (e.key.code == 's') {
            GObjectList l;
//             l.addAllObjects(getDisplayedTurn().getPlanetType());
//             l.sort(sortByName);
            for (int i = 1; i <= NUM_SHIPS; ++i)
                if (getDisplayedTurn().getShip(i).getOwner())
                    l.addObject(getDisplayedTurn().getShipType(),
                                getDisplayedTurn().getShip(i));
            l.sort(sortByOwner);
            l.addDividers(diviOwner);
//             for (GObjectList::iterator i = l.begin(); i != l.end(); ++i)
//                 std::cout << "#" << i->getId() << ": " << i->getName() << "\n";
            Ptr<WObjectSelectionList> osl = new WObjectSelectionList(0, 15, 320, l);
            osl->doStandardDialog("Choose object", 0);
            return true;
        } else if (e.key.code == 'i') {
            static string_t s = "foo";
            UIInputLine::run(s, 30, "Title Bar", "Requester Text:", 20, 0);
            return true;
        } else if (e.key.code == 'l' || e.key.code == 'm') {
            UIStringList d;
            d.add(1, "One");
            d.add(2, "Two");
            d.add(3, "Three");
            d.add(4, "Four");
            d.add(5, "Five");
            d.add(6, "Six");
            d.add(7, "Seven");
            if (e.key.code == 'm') {
                UIStandardListbox(0, 1, 150, d).doStandardMenu(100, 100);
            } else {
                ObjectHolder h;
                UIWindow w("Window Title", UIWindow::BLUE_WINDOW, COLOR_GRAY, 10);
                UIGroup g;
                w.setLayout(new UIVBoxLayout());
                w.setState(UIWidget::st_Modal, true);
                w.add(h.add(new UIStandardListbox(1, 5, 200, d)));
                w.add(g);
                g.setLayout(new UIHBoxLayout());
                g.add(h.add(new UISpacer()));
                g.add(h.add(new UIButton(1, "OK", SDLK_RETURN, UIButton::aStop)));
                g.add(h.add(new UIButton(2, "Cancel", SDLK_ESCAPE, UIButton::aStop)));
                w.pack();
                ui_root->centerWidget(w);
                ui_root->add(w);
                ui_root->run();
                ui_root->close(w);
            }
            return true;
        }
    }
    return UIGroup::handleEvent(e, second_pass);
}

/***************************** Main Routines *****************************/

int
doPlayerScreen()
{
    WPlayerScreen ps;
    ui_root->add(ps);
    int rv = ui_root->run();
    ui_root->close(ps);
    return rv;
}

/** Main loop. Giant state machine. (cc.pas::RunScreen in PCC 1.x) */
int
doMainLoop(int state)
{
    while (1) {
        if (state < 0) {
            state = -state;
            if (int i = doSelectObject(state))
                getObjectType(state).setCurrent(i);
            else
                return 0;
        }
        
        switch(state) {
         case 1:
            state = doShipScreen();
            break;
         case 2:
            state = doPlanetScreen();
            break;
         case 3:
            state = doBaseScreen();
            break;
         case 4:
            state = doStarchart();
            break;
         default:
            return state;
        }
    }
}

/** Get type associated with a screen number. Screen numbers are function
    key numbers for the appropriate screens, without formal \#defines. */
GObjectType&
getObjectType(int screen, GGameTurn& trn)
{
    switch (screen) {
     case 1: case 11:
        return trn.getShipType();
     case 2: case 12:
        return trn.getPlanetType();
     case 3: case 13:
        return trn.getBaseType();
        // FIXME: fleets
     default:
        FAIL();
    }
}

/** Get type associated with a screen number. */
GObjectType&
getObjectType(int screen)
{
    return getObjectType(screen, getDisplayedTurn());
}

/** Do object selection window for a particular screen Id.
    \returns chosen Id or 0 */
int
doSelectObject(int screen)
{
    switch (screen) {
     case 1: case 11:
        return chooseShip();
     case 2: case 12:
        return choosePlanet();
     case 3: case 13:
        return chooseBase();
        // FIXME: ship history, fleets
     default:
        return 0;
    }
}

/** Do a control screen. Does selection window and posts a cm_CS_GotoCS
    if needed. */
void
doScreen(int screen_id)
{
    if (int i = doSelectObject(screen_id))
        putCommand(cm_CS_GotoCS, screen_id, i);
}
