/**
  *  \file client/scr-ship.cc
  *  \brief Starship Control Screen
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/cscreen.h"
#include "client/dlg-fcode.h"
#include "client/dlg-list.h"
#include "client/marks.h"
#include "client/screens.h"
#include "cpluslib/format.h"
#include "game/formula.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/player.h"
#include "game/ship.h"
#include "game/spec.h"
#include "gfx/palette.h"
#include "ui/listbox.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets.h"
#include "util/nls.h"
#include "game/fcode.h"
#include "client/dlg-cargo.h"
#include "ui/layout.h"

namespace {
    /// Starship selection window.
    class WShipSelectionTile : public WObjectSelectionTile {
     public:
        WShipSelectionTile();
        void drawData(Ptr<GfxCanvas> can);
    };

    /// Ship scanner (with waypoint etc).
    class WShipScannerChartWidget : public WScannerChartWidget {
     public:
        WShipScannerChartWidget();
        ~WShipScannerChartWidget();
        void drawScanner(Ptr<GfxCanvas> can);
        void update(int);
        void sync(bool);
    };

    /// Starship screen.
    class WShipScreen : public WControlScreen {
        WShipScannerChartWidget scanner_chart;
        void* sig_handle;
        void* sig_ch_handle;
     public:
        WShipScreen();
        ~WShipScreen();
        bool handleEvent(const UIEvent& e, bool second_pass);
        void update();
    };

    /// Ship equipment tile (crew/arms/damage).
    class WShipEquipmentTile : public WTile {
     public:
        WShipEquipmentTile();
        void drawFixed(Ptr<GfxCanvas> can);
        void drawData(Ptr<GfxCanvas> can);
    };

    /// Ship cargo tile (ore/clans/cash).
    class WShipCargoTile : public WTile {
        void drawCargoBar(GfxCanvas* can, int x, int y, int wi, int have, int max, int color, const char* fmt);
     public:
        WShipCargoTile();
        void drawFixed(Ptr<GfxCanvas> can);
        void drawData(Ptr<GfxCanvas> can);
    };

    /// Ship mission tile (mission/PE/FC).
    class WShipMissionTile : public WTile {
     public:
        WShipMissionTile();
        void drawFixed(Ptr<GfxCanvas> can);
        void drawData(Ptr<GfxCanvas> can);
    };

    /// Ship movement tile (waypoint/speed/mass/ETA).
    class WShipMovementTile : public WTile {
     public:
        WShipMovementTile();
        void drawFixed(Ptr<GfxCanvas> can);
        void drawData(Ptr<GfxCanvas> can);
    };
}

static int acquireWaypointPerm(const string_t& reason, GShip& ship);

/*************************** WShipSelectionTile **************************/

WShipSelectionTile::WShipSelectionTile()
    : WObjectSelectionTile(getDisplayedTurn().getShipType())
{ }

void
WShipSelectionTile::drawData(Ptr<GfxCanvas> can)
{
    GShip& ps = getDisplayedTurn().getShip(getCurrent());

    // FIXME: this is very incomplet
    int y = 0, rem = 0;
    parm.just_x = parm.just_y = 0;
    if (getType().isMarked(getCurrent())) {
        drawBackground(can, GfxRect(0, 0, 15, font_heights[FONT_NORMAL]));
        parm.color = standard_colors[COLOR_GREENBLACK];
        drawSelection(can, parm, extent.x + 5, extent.y + 7, 1, 2);
        rem = 15;
    }
    setColor(tc_Static);
    outTextF(can, FONT_NORMAL, rem, y, extent.w - rem,
             ps.getFullName() + ", " + ps.getHull().getName());
    y += font_heights[FONT_NORMAL] * 3/2;
    outTextF(can, FONT_NORMAL, 0, y, extent.w,
             _("Location: ") + getDisplayedTurn().getLocationName(ps.getPos(), locs_WW | locs_Orbit | locs_Verbose));
    y += font_heights[FONT_NORMAL];
    outTextF(can, FONT_NORMAL, 0, y, extent.w,
             _("Waypoint: ") + getDisplayedTurn().getLocationName(ps.getWaypoint(), locs_WW | locs_Verbose));
    y += font_heights[FONT_NORMAL];
    outTextF(can, FONT_NORMAL, 0, y, extent.w,
             format(_("Distance: %.2f ly"), distFromDX(ps.getWaypointDX(),
                                                       ps.getWaypointDY())));
}

int
chooseShip()
{
    GGameTurn& trn = getDisplayedTurn();
    int nships = trn.getShipType().getCount();
    if (nships == 0) {
        messageBox(_("You do not have any ships. "
                     "To build ships, use [B] on a starbase."),
                   _("Starships"));
        return 0;
    } else if (nships == 1) {
        return trn.getShipType().getCurrent();
    } else {
        WShipSelectionTile selector;
        return doSelectWindow(selector, _("Ship"));
    }
}

/*************************** WShipEquipmentTile **************************/

WShipEquipmentTile::WShipEquipmentTile()
    : WTile(315, 6*font_heights[FONT_NORMAL]+3)
{ }

void
WShipEquipmentTile::drawFixed(Ptr<GfxCanvas> can)
{
    const int y = font_heights[FONT_NORMAL];
    setColor(tc_Static);
    outText(can, FONT_NORMAL, 5,   0, _("Engines:"));
    outText(can, FONT_NORMAL, 5,   y, _("Primary:"));
    outText(can, FONT_NORMAL, 5, 2*y, _("Secondary:"));
    outText(can, FONT_NORMAL, 5, 4*y, _("Damage:"));
    outText(can, FONT_NORMAL, 5, 5*y, _("Crew:"));
}

void
WShipEquipmentTile::drawData(Ptr<GfxCanvas> can)
{
    const int y = font_heights[FONT_NORMAL];
    const int x = 85 + 5;
    const int w = extent.w - x;
    const GShip& ship = getDisplayedTurn().getShipType().getShip();

    setColor(tc_Green);

    GHull& hull = ship.getHull();

    /* engines */
    {
        string_t en = getEngine(ship.getEngine()).getName();
        if (hull.canDoSpecial(hf_Gravitonic, ship.getOwner()))
            en = "Grav. " + en;
        if (hull.getNumEngines() > 1)
            en = format("%d x %s", hull.getNumEngines(), en);
        outTextF(can, FONT_NORMAL, x, 0, w, en);
    }

    /* primary */
    if (ship.getBeamCount() && ship.getBeamType()) {
        string_t bm = getBeam(ship.getBeamType()).getName();
        if (ship.getBeamCount() != 1)
            bm = format("%d x %s", ship.getBeamCount(), bm);
        outTextF(can, FONT_NORMAL, x, y, w, bm);
    } else {
        outTextF(can, FONT_NORMAL, x, y, w, _("none"));
    }

    /* secondary */
    if (ship.getTorpType() && ship.getTorpLauncherCount()) {
        string_t tl = format(_("%d x %s launcher%!1{s%}"),
                             ship.getTorpLauncherCount(),
                             getTorp(ship.getTorpType()).getName());
        string_t tc = (ship.getAmmo()
                       ? string_t(format(_("%d torpedo%!1{es%} aboard"), ship.getAmmo()))
                       : string_t(_("no torpedoes aboard")));
        outTextF(can, FONT_NORMAL, x, 2*y, w, tl);
        outTextF(can, FONT_NORMAL, x, 3*y, w, tc);
    } else if (ship.getBayCount()) {
        string_t tl = format(_("%d fighter bay%!1{s%}"), ship.getBayCount());
        string_t tc = (ship.getAmmo()
                       ? string_t(format(_("%d fighter%!1{s%} ready"), ship.getAmmo()))
                       : string_t(_("no fighters aboard")));
        outTextF(can, FONT_NORMAL, x, 2*y, w, tl);
        outTextF(can, FONT_NORMAL, x, 3*y, w, tc);
    } else {
        outTextF(can, FONT_NORMAL, x, 2*y, w, _("none"));
        outTextF(can, FONT_NORMAL, x, 3*y, w, string_t());
    }

    /* damage */
    if (ship.getDamage()) {
        int i = ship.getSupplies() / 5;
        int j = ship.getDamage() - i;
        if (j < 0) {
            j = 0;
            setColor(tc_Yellow);
        } else {
            setColor(tc_Red);
        }
        if (i)
            outTextF(can, FONT_NORMAL, x, 4*y, w, format("%d%% (-%d: %d%%)", ship.getDamage(), i, j));
        else
            outTextF(can, FONT_NORMAL, x, 4*y, w, format("%d%%", ship.getDamage()));
        setColor(tc_Green);
    } else {
        outTextF(can, FONT_NORMAL, x, 4*y, w, _("none"));
    }

    /* crew */
    if (ship.getCrew() != hull.getCrew()) {
        if (ship.getCrew() > hull.getCrew())
            setColor(tc_Yellow);
        else
            setColor(tc_Red);
        outTextF(can, FONT_NORMAL, x, 5*y, w, format(_("%d of %d"), ship.getCrew(), hull.getCrew()));
    } else {
        outTextF(can, FONT_NORMAL, x, 5*y, w, itoa(ship.getCrew()));
    }
}

/***************************** WShipCargoTile ****************************/

WShipCargoTile::WShipCargoTile()
    // FIXME: magic numbers
    : WTile(315, 4 * font_heights[FONT_NORMAL] + 14 + 3)
{ }

void
WShipCargoTile::drawFixed(Ptr<GfxCanvas> can)
{
    const int y = font_heights[FONT_NORMAL];
    // FIXME: translate mineral names? I guess not.
    setColor(tc_Static);
    outText(can, FONT_NORMAL, 5,   0, "Neu:");
    outText(can, FONT_NORMAL, 5,   y, "Tri:");
    outText(can, FONT_NORMAL, 5, 2*y, "Dur:");
    outText(can, FONT_NORMAL, 5, 3*y, "Mol:");

    const int x2 = 105 + 5;
    outText(can, FONT_NORMAL, x2,   0, _("Colonists:"));
    outText(can, FONT_NORMAL, x2,   y, _("Supplies:"));
    outText(can, FONT_NORMAL, x2, 2*y, _("Money:"));
    outText(can, FONT_NORMAL, x2, 3*y, _("Total Mass:"));

    setColor(tc_Green);
    const int x3 = x2 + 141;
    outText(can, FONT_NORMAL, x3,   y, " kt");
    outText(can, FONT_NORMAL, x3, 2*y, " mc");
    outText(can, FONT_NORMAL, x3, 3*y, " kt");
}

void
WShipCargoTile::drawData(Ptr<GfxCanvas> can)
{
    GShip& s = getDisplayedTurn().getShipType().getShip();
    const int y = font_heights[FONT_NORMAL];

    parm.just_x = 2;
    setColor(tc_Green);

    const int x1 = 95 + 5;
    const int wi = fonts[FONT_NORMAL]->getTextWidth("10000 kt");
    outTextF(can, FONT_NORMAL, x1,   0, wi, itoa(s.getCargo(el_Neutronium)) + " kt");
    outTextF(can, FONT_NORMAL, x1,   y, wi, itoa(s.getCargo(el_Tritanium)) + " kt");
    outTextF(can, FONT_NORMAL, x1, 2*y, wi, itoa(s.getCargo(el_Duranium)) + " kt");
    outTextF(can, FONT_NORMAL, x1, 3*y, wi, itoa(s.getCargo(el_Molybdenum)) + " kt");

    const int x2 = x1 + 151;
    outTextF(can, FONT_NORMAL, x2,   0, 70, itoa(100L * s.getColonists()));
    outTextF(can, FONT_NORMAL, x2,   y, 70, itoa(s.getSupplies()));
    outTextF(can, FONT_NORMAL, x2, 2*y, 70, itoa(s.getMoney()));
    outTextF(can, FONT_NORMAL, x2, 3*y, 55, itoa(s.getMass()));

    drawCargoBar(can, extent.x + 5,       extent.y + 4*y + 4, 130, s.getCargo(el_Neutronium), s.getHull().getFuel(), COLOR_GREENBLACK, _("Fuel: %d free"));
    int cargo = s.getHull().getCargo();
    drawCargoBar(can, extent.x + 145, extent.y + 4*y + 4, 130, cargo - s.getFreeCargo(), cargo, COLOR_BLUE, _("Cargo: %d free"));
}

void
WShipCargoTile::drawCargoBar(GfxCanvas* can, int x, int y, int wi, int have, int max,
                             int color, const char* fmt)
{
    drawFrameDown(can, GfxRect(x, y, wi, 10));
    ++x;
    ++y;
    wi -= 2;

    int split = max ? wi * have / max : 0;
    if (have > 0 && split == 0)
        split = 1;
    if (have >= max)
        split = wi;
    else if (split >= wi)
        --split;

    if (split)
        drawSolidBar(can, GfxRect(x, y, split, 8), standard_colors[color]);
    if (split != wi)
        drawSolidBar(can, GfxRect(x + split, y, wi-split, 8), standard_colors[COLOR_BLACK]);

    parm.just_x = 1;
    parm.just_y = 0;
    if (have > max)
        parm.color = standard_colors[COLOR_RED];
    else
        parm.color = standard_colors[COLOR_WHITE];
    parm.bgcolor = GFX_TRANSPARENT;
    ::outTextF(can, parm, fonts[FONT_SMALL], x + wi/2, y-1, wi, format(fmt, max-have));
}

/**************************** WShipMissionTile ***************************/

WShipMissionTile::WShipMissionTile()
    : WTile(315, 9*font_heights[FONT_NORMAL]/2)
{ }

void
WShipMissionTile::drawFixed(Ptr<GfxCanvas> can)
{
    const int y = font_heights[FONT_NORMAL];
    setColor(tc_Static);
    outText(can, FONT_NORMAL, 5,   0, _("Mission:"));
    outText(can, FONT_NORMAL, 5,   y, _("Primary Enemy:"));
    outText(can, FONT_NORMAL, 5, 2*y, _("Friendly Code:"));
}

void
WShipMissionTile::drawData(Ptr<GfxCanvas> can)
{
    const int y = font_heights[FONT_NORMAL];
    const int x = 120;

    const GShip& ship = getDisplayedTurn().getShipType().getShip();

    setColor(tc_Green);
    outTextF(can, FONT_NORMAL, x-50, 0, 214, format("FIXME: M%d I%d T%d",
                                                    ship.getMission(),
                                                    ship.getIntId(),
                                                    ship.getTowId()));
    outTextF(can, FONT_NORMAL, x,    y, 164-25,
             (ship.getPE()
              ? getShortName(ship.getPE())
              : string_t(_("none"))));
    outTextF(can, FONT_NORMAL, x,  2*y, 100, ship.getFCode());

    // FIXME
    string_t comment = "FIXME";
    if (comment.empty()) {
        setColor(tc_Static);
        comment = _("(press [F9] to add a note)");
    } else {
        setColor(tc_Yellow);
    }
    outTextF(can, FONT_NORMAL, 5,  7*y/2, 265, comment);
}

/*************************** WShipMovementTile ***************************/

WShipMovementTile::WShipMovementTile()
    : WTile(GfxPoint(315, 7*font_heights[FONT_NORMAL]))
{ }

void
WShipMovementTile::drawFixed(Ptr<GfxCanvas> can)
{
    const int y = font_heights[FONT_NORMAL];

    setColor(tc_Static);
    outText(can, FONT_NORMAL, 5, 0,   _("Location:"));
    outText(can, FONT_NORMAL, 5, y,   _("Waypoint:"));
    outText(can, FONT_NORMAL, 5, 2*y, _("Distance:"));
    outText(can, FONT_NORMAL, 5, 3*y, _("Warp Factor:"));
    outText(can, FONT_NORMAL, 5, 4*y, _("E.T.A.:"));
    outText(can, FONT_NORMAL, 5, 5*y, _("Fuel Usage:"));
    outText(can, FONT_NORMAL, 5, 6*y, _("Engine Load:"));
}

void
WShipMovementTile::drawData(Ptr<GfxCanvas> can)
{
    const int y = font_heights[FONT_NORMAL];
    const int x = 110;
    const int w = extent.w - x;
    const GShip& s = getDisplayedTurn().getShipType().getShip();

    setColor(tc_Green);
    /* location */
    outTextF(can, FONT_NORMAL, x, 0, w, getDisplayedTurn().getLocationName(s.getPos(), locs_WW | locs_Verbose));

    /* waypoint */
    outTextF(can, FONT_NORMAL, x, y, w, "FIXME");

    /* warp */
    outTextF(can, FONT_NORMAL, x, 3*y, w, format(_("Warp %d")) << s.getWarp());

    // FIXME
}

/************************ WShipScannerChartWidget ************************/

WShipScannerChartWidget::WShipScannerChartWidget()
{ }

WShipScannerChartWidget::~WShipScannerChartWidget()
{ }

void
WShipScannerChartWidget::drawScanner(Ptr<GfxCanvas> can)
{
    const GChartViewport& vp = getViewport();

    GPoint                wp = getDisplayedTurn().getShipType().getShip().getWaypoint();
    GfxPoint       wp_screen = vp.scaleXY(wp);

    if (wp != getScannerPosition() && wp_screen != vp.getCenter()) {
        GfxParameters parm;
        parm.color = standard_colors[COLOR_RED];

        drawLine(can, parm, vp.getCenterX(), vp.getCenterY(), wp_screen.x, wp_screen.y);
        drawEndBig(can, parm, wp_screen.x, wp_screen.y);
    }

    WScannerChartWidget::drawScanner(can);
}

void
WShipScannerChartWidget::update(int)
{
    drawWidget(false);
}

void
WShipScannerChartWidget::sync(bool dblclick)
{
    if (dblclick) {
        GGameTurn& trn = getDisplayedTurn();
        if (int id = acquireWaypointPerm(_("waypoint"), trn.getShipType().getShip())) {
            trn.getShip(id).setWaypoint(getScannerPosition());
            trn.doUpdates();
        }
    }
}

/****************************** WShipScreen ******************************/

WShipScreen::WShipScreen()
    : WControlScreen(getDisplayedTurn().getShipType(), 1, GShip::ss_DirtyHead),
      scanner_chart()
{
    addTile(holder.add(new WControlScreenHeaderTile(*this, WControlScreenHeaderTile::bName + WControlScreenHeaderTile::bAuto + WControlScreenHeaderTile::bCycle + WControlScreenHeaderTile::bAdd,
                                                    GShip::ss_DirtyHead)));
    addTile(holder.add(new WShipEquipmentTile()),    _("Equipment & Crew:"));
    addTile(holder.add(new WShipCargoTile()),        _("Aboard:"));
    addTile(holder.add(new WShipMissionTile()),      _("Mission:"));
    addTile(holder.add(new WShipMovementTile()));
    addChart(scanner_chart);
    sig_handle = getDisplayedTurn().sig_ship_changed.add(static_cast<WControlScreen*>(this),
                                                         &WControlScreen::updateId);
    sig_ch_handle = sig_changed.add(static_cast<WShipScannerChartWidget*>(&scanner_chart),
                                    &WShipScannerChartWidget::update);
}

WShipScreen::~WShipScreen()
{
    getDisplayedTurn().sig_ship_changed.remove(sig_handle);
    sig_changed.remove(sig_ch_handle);
}

void
WShipScreen::update()
{
    int sid = getType().getCurrent();
    const GShip& ship = getDisplayedTurn().getShip(sid);
    setData(mkResId(rt_Ship, ship.getHull().getPicture() + 1000),
            ship.getPos(),
            ship.getName(),
            format(_("(Id #%d, %s)"), sid, ship.getHull().getName()));
}

static void
doShipXchg(WControlScreen* scr)
{
    WScanner* scan             = &scr->getScanner();
    WScannerChartWidget* chart = scan->getChart();
    int n;
    if (scan->getScannedShipId()) {
        n = getDisplayedTurn().countMyShipsAt(scr->getScannerPosition());
        if (chart->getCenterPosition() == chart->getScannerPosition())
            --n;                // discount current ship
    } else {
        n = 0;
    }

    if (n)
        putKey(ss_Ctrl + SDLK_F1);
    else
        putKey(ss_Ctrl + SDLK_F2);
}

/** Attempt to acquire permission to change waypoint. (this is used for fleets)
    \param reason reason, for user
    \param ship   ship we want to change
    \returns 0 if permission not granted, ship id to change waypoint on
    specified ship */
static int
acquireWaypointPerm(const string_t& reason, GShip& ship)
{
    if (!ship.getWaypointPermission()) {
        // FIXME: display dialog, leave fleet, ...
        (void) reason;
        return 0;
    } else {
        return ship.getId();
    }
}

void
doShipTransferFor(GShip& ship, GObjectId chosen)
{
    Ptr<GCargoContainer> left = new GShipTransfer(ship);
    Ptr<GCargoContainer> right;

    if (!chosen.isGameObject()) {
        /* jettison */
        right = new GShipTransporter(ship);
    } else if (GShip* s = dynamic_cast<GShip*>(&chosen.getObject())) {
        if (s->getOwner() == ship.getOwner()) {
            /* ship <-> ship */
            ASSERT(s->haveDat());
            right = new GShipTransfer(*s);
        } else {
            /* ship <-> foreign ship */
            if (ship.getCargoHandler().isFacilityInUse(GShip::sf_Transfer)) {
                messageBox(_("The cargo transporter is already in use."), _("Cargo Transfer"));
                return;
            }
            if (ship.getTransporter(true).id != 0
                && ship.getTransporter(true).id != s->getId()
                && ship.isTransporterActive(true))
            {
                messageBox(format(_("The cargo transporter is already transferring cargo to %s."),
                                  s->getFullName()), _("Cargo Transfer"));
                // FIXME: we could cancel the transfer here
                return;
            }
            // ship.cancelTranporter(true);
            right = new GShipTransporter(ship, true, *s);
        }
    } else if (GPlanet* p = dynamic_cast<GPlanet*>(&chosen.getObject())) {
        if (p->getOwner() == ship.getOwner() && p->haveDat()) {
            right = new GPlanetTransfer(*p);
        } else {
            right = new GShipTransporter(ship, false, *p);
        }
    } else {
        FAIL();
    }
    doCargoTransfer(left, right);
}

static void
doShipCargoTransfer(GGameTurn& trn, GShip& ship)
{
    GObjectList list;
    int pid = trn.getPlanetAt(ship.getPos());
    if (pid)
        list.addObject(trn.getPlanetType(), trn.getPlanet(pid));
    else
        list.addDivider(_("Jettison into space"), -1);

    for (int i = 1; i <= NUM_SHIPS; ++i)
        if (i != ship.getId() && trn.getShip(i).isExistant() && trn.getShip(i).getPos() == ship.getPos())
            list.addObject(trn.getShipType(), trn.getShip(i));

    Ptr<WObjectSelectionList> wlist = new WObjectSelectionList(0, 15, 320, list);
    GObjectId chosen = wlist->doStandardDialog(_("Transfer cargo to..."), 0 /* FIXME */);
    if (chosen.isValid())
        doShipTransferFor(ship, chosen);
}

static void
doShipNavigationChart(GGameTurn& trn, GShip& ship)
{
    int size = ui_root->getExtent().h;
    
    UIWindow the_window(_("Navigation Chart"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
    the_window.setLayout(new UIVBoxLayout(5, 0));
    the_window.setState(UIWindow::st_Modal, true);

    WScannerChartWidget chart;
    chart.setCenterPosition(ship.getPos());
    chart.setScannerPosition(ship.getWaypoint());
    chart.setState(chart.sc_StyleChange | chart.sc_Zoomable | chart.sc_Keyboard, true);
    the_window.add(chart);
    
    UIGroup buttons(new UIHBoxLayout(5));
    ObjectHolder h;
    buttons.add(h.add(new UISpacer()));
    buttons.add(h.add(new UIButton(cm_Enter, _("F10 - OK"), SDLK_F10, UIButton::aStop)));
    buttons.add(h.add(new UIButton(cm_Escape, "ESC",        SDLK_ESCAPE, UIButton::aStop)));
    the_window.add(buttons);
    the_window.setExtent(GfxRect(0, 0, size - 50, size - 10));
    the_window.doLayout();
    
    ui_root->centerWidget(the_window);
    ui_root->add(the_window);
    int rv = ui_root->run();
    ui_root->close(the_window);

    if (rv == cm_Enter)
        ship.setWaypoint(chart.getScannerPosition());        
}

bool
WShipScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        GGameTurn& trn = getDisplayedTurn();
        GShip&     ship = trn.getShipType().getShip();
        switch (e.key.code) {
         case 'n': {
             string_t s = ship.getName();
             if (UIInputLine::run(s, 20, format(_("Rename Starship #%d"),
                                                ship.getId()),
                             _("Enter new name:"), 20, 0))
                 ship.setName(s);
             trn.doUpdates();
             return true;
         }
         case 'e': {
             UIStringList d;
             d.add(0, _("0 - none"));
             for (int i = 1; i <= NUM_PLAYERS; ++i)
                 d.add(i, format("%X - %s", i, getShortName(i)));

             UIStandardListbox lb(0, -1, 240, d);
             lb.scrollToKey(ship.getPE());
             if (lb.doStandardListbox(_("Change Primary Enemy"), 0))
                 ship.setPE(lb.getKey());
             trn.doUpdates();
             return true;
         }
         case 'a':
            if (int id = acquireWaypointPerm(_("waypoint"), ship)) {
                doShipNavigationChart(trn, trn.getShip(id));
                trn.doUpdates();
            }
            return true;
         case 'c':
            doShipCargoTransfer(trn, ship);
            trn.doUpdates();
            return true;
         case 'u':
            if (ship.getOrbit()) {
                doShipTransferFor(ship, GObjectId(trn.getPlanetType(),
                                                  trn.getPlanet(ship.getOrbit())));
                trn.doUpdates();
            } else {
                messageBox(_("We are not orbiting a planet, sir. "
                             "If you want to jettison cargo, use the "
                             "'c' command instead."),
                           _("Cargo Transfer"));
            }
            return true;
         case 'x':
            doShipXchg(this);
            return true;
         case 'f': {
             GShip& ship = trn.getShipType().getShip();
             GFCodeList list(GFCodeList::getInstance(), ship);
             WFCodeWindow dialog(_("Change Friendly Code"), list);
             dialog.setFCode(ship.getFCode());
             ui_root->centerWidget(dialog);
             ui_root->add(dialog);
             int rv = ui_root->run();
             ui_root->close(dialog);
             if (rv == cm_Enter) {
                 ship.setFCode(dialog.getFCode());
                 trn.doUpdates();
             }
             return true;
         }
         case ss_Ctrl + 'w':
            if (int id = acquireWaypointPerm(_("waypoint"), trn.getShipType().getShip())) {
                trn.getShip(id).setWaypoint(getScannerPosition());
                trn.doUpdates();
            }
            return true;
         case ss_Ctrl + 'i':
            if (ship.getIntId())
                putCommand(cm_CS_GotoCS, 1, ship.getIntId());
            return true;
         case ss_Ctrl + 't':
            if (ship.getTowId())
                putCommand(cm_CS_GotoCS, 1, ship.getTowId());
            return true;
         case SDLK_F8:
         case ss_Shift + SDLK_F8:
            putCommand(cm_CS_GotoCS, 3, trn.getShipType().getShip().getOrbit());
            return true;
         case SDLK_TAB:
         case ss_Shift + SDLK_TAB:
         case ss_Ctrl + SDLK_TAB:
         case ss_Shift + ss_Ctrl + SDLK_TAB: {
             GObjectList l;
             l.addObjectsAt(getDisplayedTurn(), pos, 0, 0);
             l.postprocessUser(false);
             int n = l.getIndexFor(GObjectId(trn.getShipType(), trn.getShipType().getShip()));
             int overflow = 0;
             ASSERT(n >= 0);
             do {
                 if (e.key.code & ss_Shift) {
                     if (++n >= l.getCount())
                         n = 0, ++overflow;
                 } else {
                     if (--n < 0)
                         n = l.getCount()-1, ++overflow;
                 }
             } while (!l[n].getType().isMarked(l[n].getId()) && (e.key.code & ss_Ctrl) && overflow < 2);
             if (overflow < 2)
                 putCommand(cm_CS_GotoCS, getScreenId(), l[n].getId());
             break;
         }
            /* ship screen keys:

               c-p, c-j, c-s    // cargo transfer
               m-i, m-t         // search xceptor/tower
               c-n              // list next turn
               q                // toggle independantmove
               w                // warp
               a                // aim
               m                // mission
               b                // base commands
               ins              // sim
               s                // spec
               r                // remote
               g                // give
               d                // distribute cargo
               f9               // comment
               f10              // fleet
               s-f10            // join fleet/goto fleet
               m-r              // reset x/y
            */
        }
    }
    return WControlScreen::handleEvent(e, second_pass);
}

int
doShipScreen()
{
    return WShipScreen().run();
}
