/**
  *  \file vcranim.cc
  *  \brief VCR Animation
  *
  *  Animations/Sprites for the VCR.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <map>
#include <cstdlib>
#include <algorithm>
#include <cmath>
#include <iostream>
#include "client/vcranim.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "client/vcrvis.h"
#include "util/random.h"
#include "util/nls.h"
#include "cpluslib/format.h"

typedef Ptr<GfxPixmap> (*MakePixmapFunction)(const uint8*);
static Ptr<GfxAnimation> makeExplosion(int rx, int ry,
                                       int sx, int sy,
                                       int hotspots,
                                       int speed,
                                       MakePixmapFunction make_pix);
static Ptr<GfxPixmap> makeExplodePixmap(const uint8* buf);
static Ptr<GfxPixmap> makeShieldPixmap(const uint8* buf);

/* FIXME: memory leak */
typedef std::map<int, Ptr<GfxAnimation> > VcrExplAniCache;
static VcrExplAniCache* expl_cache;
static VcrExplAniCache* shield_cache;

/** \class VcrExplSprite
    \brief Explosion Sprite

    A sprite displaying an explosion, moving at a certain speed. */

/** Construct new explosion. This one does not move. \overload */
VcrExplSprite::VcrExplSprite(int x, int y, int layer, int size, int speed)
    : GfxAniSprite(x - EXPLODE_SIZE/2, y - EXPLODE_SIZE/2,
                   layer, getExplosion(size, speed)),
      dx(0), dy(0), sumx(0), sumy(0)
{ }

/** Construct new explosion.
    \param x,y    coordinates of center
    \param layer  layer in animation
    \param size   "size" parameter for explosion (number of hotspots -- not
                  linear in visual appearance!)
    \param speed  "speed" parameter for explosion (every Nth frame displayed)
    \param adx,ady Movement vector of explosion, in 1/16 pixels. */
VcrExplSprite::VcrExplSprite(int x, int y, int layer, int size, int speed,
                             int adx,  int ady)
    : GfxAniSprite(x - EXPLODE_SIZE/2, y - EXPLODE_SIZE/2,
                   layer, getExplosion(size, speed)),
      dx(adx), dy(ady), sumx(0), sumy(0)
{ }

void
VcrExplSprite::tick()
{
    GfxAniSprite::tick();
    int mx = sumx / 16;
    int my = sumy / 16;
    sumx += dx;
    sumy += dy;
    mx -= sumx / 16;
    my -= sumy / 16;
    if (mx || my)
        moveTo(pos.x - mx, pos.y - my);
}

/** Create an Explosion.
    \param size    size. Number of hotspots.
    \param speed   speed. Play only every Nth frame. */
Ptr<GfxAnimation>
VcrExplSprite::getExplosion(int size, int speed)
{
    if (!expl_cache)
        expl_cache = new VcrExplAniCache;

    int id = 10*size + speed;
    VcrExplAniCache::iterator i = expl_cache->find(id);
    if (i == expl_cache->end()) {
        Ptr<GfxAnimation> anim;
        if (size < 32)
            anim = makeExplosion(16384, 16384, 8192, 8192, size, speed, makeExplodePixmap);
        else
            anim = makeExplosion(32767, 32767, 16384, 16384, size, speed, makeExplodePixmap);
        (*expl_cache)[id] = anim;
        return anim;
    } else {
        return i->second;
    }
}

/**************************** VcrShieldSprite ****************************/

/** \class VcrShieldSprite
    \brief Shield Animation

    The shield animation is a variation of the explode effect, using
    different colors and parameters. Depending on the weapon, the
    shield discharge wave moves into a different direction, although the
    shield effect itself remains at the same position. */

VcrShieldSprite::VcrShieldSprite(int x, int y, int layer, int dir)
    : GfxAniSprite(x - VcrExplSprite::EXPLODE_SIZE / 2,
                   y - VcrExplSprite::EXPLODE_SIZE / 2,
                   layer, getShield(dir))
{ }

/** Get Shield Animation. Fetches it from cache or creates it.
    \param dir direction where shield moves to, in degrees (i.e. if
    something hits from the right, dir=270 for shield moving to the left). */
Ptr<GfxAnimation>
VcrShieldSprite::getShield(int dir)
{
    dir = (dir + 22) / 45;

    if (!shield_cache)
        shield_cache = new VcrExplAniCache;

    VcrExplAniCache::iterator i = shield_cache->find(dir);
    if (i != shield_cache->end())
        return i->second;

    /* need to create it */
    Ptr<GfxAnimation> anim;
    switch(dir) {
     case 0:
     case 8:
        /* NORTH */
        anim = makeExplosion(32768, 16384, 16384, 0, 16, 2, makeShieldPixmap);
        break;
     case 1:
        /* NORTHEAST */
        anim = makeExplosion(16384, 16384, 16384, 0, 16, 2, makeShieldPixmap);
        break;
     case 2:
        /* EAST */
        anim = makeExplosion(16384, 32768, 16384, 16384, 16, 2, makeShieldPixmap);
        break;
     case 3:
        /* SOUTHEAST */
        anim = makeExplosion(16384, 16384, 16384, 16384, 16, 2, makeShieldPixmap);
        break;
     case 4:
        /* SOUTH */
        anim = makeExplosion(32768, 16384, 16384, 16384, 16, 2, makeShieldPixmap);
        break;
     case 5:
        /* SOUTHWEST */
        anim = makeExplosion(16384, 16384, 0, 16384, 16, 2, makeShieldPixmap);
        break;
     case 6:
        /* WEST */
        anim = makeExplosion(16384, 32768, 0, 16384, 16, 2, makeShieldPixmap);
        break;
     case 7:
        /* NORTHEAST */
        anim = makeExplosion(16384, 16384, 0, 0, 16, 2, makeShieldPixmap);
        break;
     default:
        std::cerr << "huh? value is " << dir << "\n";
        ASSERT(0);
    }
    (*shield_cache)[dir] = anim;
    return anim;
}

/******************************* Explosions ******************************/

/** \name Explosions */

// The basic idea/algorithm was borrowed from the Allegro 3.0 demo.
// This may still be noticeable; although the algorithm inbetween was
// a Pascal program.
//@{

/** Make Explosion Pixmap from explosion buffer. /buf/ points to a
    EXPLODE_SIZE x EXPLODE_SIZE buffer containing explosion
    intensities. */
static Ptr<GfxPixmap>
makeExplodePixmap(const uint8* buf)
{
    if (ui_screen->getSurface()->format->BytesPerPixel == 1) {
        /* Palettized Image: use colorkey blit */

        Ptr<GfxPixmap> pix = new GfxPixmap(VcrExplSprite::EXPLODE_SIZE,
                                           VcrExplSprite::EXPLODE_SIZE,
                                           8, 0, 0, 0, 0);
        SDL_Surface* sfc = pix->getSurface();
        ASSERT(sfc->format->BytesPerPixel == 1);
        SDL_SetColorKey(sfc, SDL_SRCCOLORKEY, 255);
        SDL_SetPalette(sfc, SDL_LOGPAL, standard_palette, 0, NUM_STANDARD_COLORS);

        for (int y = 0; y < VcrExplSprite::EXPLODE_SIZE; ++y) {
            uint8* dest = static_cast<uint8*>(sfc->pixels) + y * sfc->pitch;
            for (int x = 0; x < VcrExplSprite::EXPLODE_SIZE; ++x) {
                register uint8 pix = *buf++;
                if (pix > 63)
                    *dest++ = 95-32;
                else if (pix < 3)
                    *dest++ = 255;
                else
                    *dest++ = pix/2 + 32;
            }
        }
        return pix;
    } else {
        /* RGB image: use alpha blit */
        // FIXME: use these only for explosions on ships, not
        // for fighters.
        Ptr<GfxPixmap> pix = new GfxPixmap(VcrExplSprite::EXPLODE_SIZE,
                                           VcrExplSprite::EXPLODE_SIZE,
                                           32,
                                           0x00FF0000UL,
                                           0x0000FF00UL,
                                           0x000000FFUL,
                                           0xFF000000UL);
        SDL_Surface* sfc = pix->getSurface();
        ASSERT(sfc->format->BytesPerPixel == 4);

        for (int y = 0; y < VcrExplSprite::EXPLODE_SIZE; ++y) {
            uint32* dest =
                static_cast<uint32*>(static_cast<void*>(static_cast<uint8*>(sfc->pixels) + y * sfc->pitch));
            for (int x = 0; x < VcrExplSprite::EXPLODE_SIZE; ++x) {
                register uint8 pix = *buf++;
                if (pix > 63) {
                    *dest++ = 0xFFFFFFFF;
                } else if (pix < 3) {
                    *dest++ = 0x00000000;
                } else {
                    int c = pix/2 + 32;
                    *dest++ = (pix + 64) << 25    // was: (pix << 26)
                        | standard_palette[c].r << 16
                        | standard_palette[c].g << 8
                        | standard_palette[c].b;
                }
            }
        }
        return pix;
    }
}

/** Make Shield Pixmap from Explosion buffer. /buf/ points to a
    EXPLODE_SIZE x EXPLODE_SIZE buffer containing explosion
    intensities. */
static Ptr<GfxPixmap>
makeShieldPixmap(const uint8* buf)
{
    if (ui_screen->getSurface()->format->BytesPerPixel == 1) {
        /* Palettized Image: use colorkey blit */
        Ptr<GfxPixmap> pix = new GfxPixmap(VcrExplSprite::EXPLODE_SIZE,
                                           VcrExplSprite::EXPLODE_SIZE,
                                           8, 0, 0, 0, 0);
        SDL_Surface* sfc = pix->getSurface();
        ASSERT(sfc->format->BytesPerPixel == 1);
        SDL_SetColorKey(sfc, SDL_SRCCOLORKEY, 255);
        SDL_SetPalette(sfc, SDL_LOGPAL, standard_palette, 0, NUM_STANDARD_COLORS);

        for (int y = 0; y < VcrExplSprite::EXPLODE_SIZE; ++y) {
            uint8* dest = static_cast<uint8*>(sfc->pixels) + y * sfc->pitch;
            for (int x = 0; x < VcrExplSprite::EXPLODE_SIZE; ++x) {
                register uint8 pix = *buf++;
                if (pix > 63)
                    *dest++ = 255; // 111-32;
                else if (pix < 3)
                    *dest++ = 255;
                else
                    *dest++ = pix/4 + 64;
            }
        }
        return pix;
    } else {
        /* RGB image: use alpha blit */
        // FIXME: use these only for explosions on ships, not
        // for fighters.
        Ptr<GfxPixmap> pix = new GfxPixmap(VcrExplSprite::EXPLODE_SIZE,
                                           VcrExplSprite::EXPLODE_SIZE,
                                           32,
                                           0x00FF0000UL,
                                           0x0000FF00UL,
                                           0x000000FFUL,
                                           0xFF000000UL);
        SDL_Surface* sfc = pix->getSurface();
        ASSERT(sfc->format->BytesPerPixel == 4);

        for (int y = 0; y < VcrExplSprite::EXPLODE_SIZE; ++y) {
            uint32* dest =
                static_cast<uint32*>(static_cast<void*>(static_cast<uint8*>(sfc->pixels) + y * sfc->pitch));
            for (int x = 0; x < VcrExplSprite::EXPLODE_SIZE; ++x) {
                register uint8 pix = *buf++;
                if (pix > 63) {
                    *dest++ = 0x00000000;  // was: 0x80FFFFFF
                } else if (pix < 3) {
                    *dest++ = 0x00000000;
                } else {
                    int c = pix/4 + 64;
                    *dest++ = (pix /*+ 64*/) << 25    // was: (pix << 26)
                        | standard_palette[c].r << 16
                        | standard_palette[c].g << 8
                        | standard_palette[c].b;
                }
            }
        }
        return pix;
    }
}

static const uint8 shift_tab[] = {
    4, 3, 3, 3, 2, 2, 2, 2, 2, 3, 3, 3, 4,
    3, 3, 3, 2, 2, 2, 1, 2, 2, 2, 3, 3, 3,
    3, 3, 2, 2, 2, 1, 1, 1, 2, 2, 2, 3, 3,
    3, 2, 2, 2, 1, 1, 1, 1, 1, 2, 2, 2, 3,
    2, 2, 2, 1, 1, 1, 0, 1, 1, 1, 2, 2, 2,
    2, 2, 1, 1, 1, 0, 0, 0, 1, 1, 1, 2, 2,
    2, 1, 1, 1, 0, 0, 0, 0, 0, 1, 1, 1, 2,
    2, 2, 1, 1, 1, 0, 0, 0, 1, 1, 1, 2, 2,
    2, 2, 2, 1, 1, 1, 0, 1, 1, 1, 2, 2, 2,
    3, 2, 2, 2, 1, 1, 1, 1, 1, 2, 2, 2, 3,
    3, 3, 2, 2, 2, 1, 1, 1, 2, 2, 2, 3, 3,
    3, 3, 3, 2, 2, 2, 1, 2, 2, 2, 3, 3, 3,
    4, 3, 3, 3, 2, 2, 2, 2, 2, 3, 3, 3, 4
};

/** Make Explosion.
    \param rx,ry    Movement x/y range
    \param sx,sy    Movement x/y offset
    \param hotspots Number of hotspots
    \param speed    Speed. Explosion will have EXPLODE_FRAMES/speed frames.

    Movement of hotspots is specified in 1/65536 pixels; there are
    no range checks, unlucky values for the movement values cause
    segfaults. */
static Ptr<GfxAnimation>
makeExplosion(int rx, int ry, int sx, int sy, int hotspots, int speed,
              MakePixmapFunction make_pix)
{
    struct Hotspot {
        long x, y;
        int xc, yc;
    };

    Hotspot hs[VcrExplSprite::MAX_HOTSPOTS];
    Ptr<GfxAnimation> anim = new GfxAnimation();

    for (int i = 0; i < hotspots; ++i) {
        hs[i].x = VcrExplSprite::EXPLODE_SIZE/2 * 65536L;
        hs[i].y = VcrExplSprite::EXPLODE_SIZE/2 * 65536L;
        hs[i].xc = myRandom(rx) - sx;
        hs[i].yc = myRandom(ry) - sy;
    }
    for (int i = 1; i <= VcrExplSprite::EXPLODE_FRAMES; ++i) {
        if (i % speed == 0) {
            int color = (i < 16) ? 2*i : (80-i) / 2;
            uint8 buf[VcrExplSprite::EXPLODE_SIZE2];
            std::fill(buf, buf + sizeof buf, 0);
            for (int j = 0; j < hotspots; ++j) {
                const uint8* sl = shift_tab;
                for (int x = -6; x <= 6; ++x)
                    for (int y = -6; y <= 6; ++y) {
                        uint16 xx = (hs[j].x >> 16) + x;
                        uint16 yy = (hs[j].y >> 16) + y;
                        buf[yy * VcrExplSprite::EXPLODE_SIZE + xx] += color >> *sl++;
                    }
                hs[j].x += hs[j].xc;
                hs[j].y += hs[j].yc;
            }
            anim->addFrame(make_pix(buf), anim->getNumFrames()+1, 1);
        } else {
            for (int j = 0; j < hotspots; ++j) {
                hs[j].x += hs[j].xc;
                hs[j].y += hs[j].yc;
            }
        }
    }

    anim->getFrame(anim->getNumFrames()-1).next = -1;
    return anim;
}
//@}

/**************************** VcrFighterSprite ***************************/

/** \class VcrFighterSprite
    \brief Fighter in VCR

    These will automatically update themselves from the VCR player, and
    move accordingly. */

/** New VCR Fighter Sprite.
    \param x,y         start x/y (center). Position of owning object.
    \param alayer      animation layer
    \param aside       side (0 or 1)
    \param atrack      fighter track (0..N)
    \param aplayer     owner of fighter
    \param avcr_player VCR player
    \param aatt,aretr  graphics of attacking/retracting fighter */
VcrFighterSprite::VcrFighterSprite(int x, int y,
                                   int alayer,
                                   int aside, int atrack, int aplayer,
                                   Ptr<VcrPlayer> avcr_player,
                                   Ptr<GfxPixmap> aatt, Ptr<GfxPixmap> aretr)
    : GfxSprite(GfxRect(x-WIDTH/2, y-HEIGHT/2, WIDTH, HEIGHT), alayer),
      side(aside), track(atrack), player(aplayer),
      vcr_player(avcr_player),
      attack(aatt), retract(aretr),
      target_x(x), target_y(getTrackY(atrack)),
      status(1), last_time(0)
{ }

VcrFighterSprite::VcrFighterSprite(const Status& st,
                                   int alayer,
                                   int aside, int atrack, int aplayer,
                                   Ptr<VcrPlayer> aplay,
                                   Ptr<GfxPixmap> aatt, Ptr<GfxPixmap> aretr)
    : GfxSprite(GfxRect(st.x, st.y, WIDTH, HEIGHT), alayer),
      side(aside), track(atrack), player(aplayer),
      vcr_player(aplay),
      attack(aatt), retract(aretr),
      target_x(st.target_x), target_y(st.target_y),
      status(st.status), last_time(0)
{ }

VcrFighterSprite::~VcrFighterSprite()
{ }

/** Compute final Y for Track. The fighter will steer to this Y. */
int
VcrFighterSprite::getTrackY(int tr)
{
    /* New way */
    int t19 = tr % 19;
    if (t19 & 1)
        t19 = 9 - t19/2;
    else
        t19 = t19/2 + 10;

    return 10*t19 + 2*(tr / 19) + 10;

//    /* "classic" way */
//    return 10*(tr % 19) + 2*(tr / 19) + 50;
}

static inline void
moveCoord(int& x, int target, int speed = 5)
{
    if (x < target) {
        x += speed;
        if (x > target)
            x = target;
    } else if (x > target) {
        x -= speed;
        if (x < target)
            x = target;
    }
}

void
VcrFighterSprite::tick()
{
    /* do nothing if time hasn't changed */
    if (last_time == vcr_player->getTime())
        return;
    last_time = vcr_player->getTime();

    /* Move sprite */
    int newx = pos.x, newy = pos.y;

    target_x = vcr_player->getFighterX(side, track);
    /* NEW */
//     if (vcr_player->getFighterStatus(side, track) == 1) {
//         int enemy_dist = target_x - vcr_player->getObjectX(1-side);
//         if (side == 1)
//             enemy_dist = -enemy_dist;
//         enemy_dist += 30;
//         if (enemy_dist > 0 && enemy_dist < 70) {
//             int ty = getTrackY(track);
//             target_y = ty + (VCRPLAY_OBJECT_Y - ty) * (enemy_dist*enemy_dist / 5400);
//         }
//     }                
    /* END */

    moveCoord(newy, target_y);
    int xmove = 5 - std::abs(newy - target_y);
    if (xmove < 3)
        xmove = 3;
    moveCoord(newx, target_x, xmove);
    moveTo(newx, newy);

    /* Change shape */
    bool oldst = status;
    status = vcr_player->getFighterStatus(side, track);
    if (oldst != status)
        update();
}

void
VcrFighterSprite::draw(Ptr<GfxCanvas> can)
{
    if (status == 1)
        attack->blit(can, pos.x, pos.y);
    else
        retract->blit(can, pos.x, pos.y);
}

/** Kill this fighter. Deletes it and starts an explosion. */
void
VcrFighterSprite::kill()
{
    int x = pos.x+WIDTH/2;
    int y = pos.y+HEIGHT/2;

    int xdir = vcr_player->getFighterStatus(side, track) == 1 ? +4 : -4;
    if (side)
        xdir = -xdir;
    int ydir =
        y < VCRPLAY_OBJECT_Y
        ? (y < VCRPLAY_OBJECT_Y / 2 ? -4 : -2)
        : (y < 3*VCRPLAY_OBJECT_Y / 2 ? -2 : -4);

    getManager()->addSprite(new VcrExplSprite(x, y, VCRPLAY_LAYER_BANG, 8, 2,
                                              xdir, ydir));
    deleteThis();
}

void
VcrFighterSprite::storeStatus(Status& st)
{
    st.status   = status;
    st.x        = pos.x;
    st.y        = pos.y;
    st.target_x = target_x;
    st.target_y = target_y;
}

/**************************** VcrWeaponSprite ****************************/

/** \class VcrWeaponSprite
    \brief Weapon Sprite

    Provides sprite counting (playback is stopped as long as there are
    active weapons) and logic for making bangs.

    \bug It happens from time to time that there remain active sprites
    when terminating the animation. There must be something fundamentally
    wrong. */

int VcrWeaponSprite::active_fx_sprites;

VcrWeaponSprite::VcrWeaponSprite(GfxRect apos, int alayer, int ax1, int ay1)
    : GfxSprite(apos, alayer), x1(ax1), y1(ay1)
{
    ++active_fx_sprites;
}

VcrWeaponSprite::~VcrWeaponSprite()
{
    --active_fx_sprites;
}

/** Create Bang. Returns a bang sprite for the current status in /ply/.
    \param ply       associated VCR player
    \param direction where the hit came from, in degrees: 0=N, 90=E, 180=S, 270=W
    \param side      object which was hit */
Ptr<GfxSprite>
VcrWeaponSprite::makeBang(VcrPlayer* ply, int direction, int side)
{
    while (direction < 0)
        direction += 360;
    if (ply->getShields(side) > 0) {
        return new VcrShieldSprite(x1, y1, VCRPLAY_LAYER_BANG,
                                   direction);
    } else {
        /* bang */
        return new VcrExplSprite(x1-5 + rand()%10,
                                 y1-5 + rand()%10,
                                 VCRPLAY_LAYER_BANG, 3+rand()%6, 4,
                                 0, 0);
    }
}

/***************************** VcrBeamSprite *****************************/

/** New Beam Sprite.
    \param ax0,ay0      position of cannon;
    \param ax1,ay1      position of target;
    \param alayer       layer;
    \param damage,kill  weapon data (used to determine color of weapon)

    \todo Long beams hurt performance, as they cause an update
    of almost the whole screen. */
VcrBeamSprite::VcrBeamSprite(int ax0, int ay0, int ax1, int ay1,
                             int alayer, int damage, int kill)
    : VcrWeaponSprite(GfxRect(std::min(ax0, ax1), std::min(ay0, ay1),
                              std::abs(ax0-ax1)+1, std::abs(ay0-ay1)+1),
                      alayer, ax1, ay1),
      scheme(getScheme(damage, kill)),
      status(0),
      x0(ax0), y0(ay0)
{ }

VcrBeamSprite::~VcrBeamSprite()
{ }

int
VcrBeamSprite::getScheme(int /*damage*/, int /*kill*/)
{
    return 0;
}

/* Status:
   ====--------
   ========----
   ============
   ....========
   ........====
   ............ */

void
VcrBeamSprite::tick()
{
    if (++status > 6)
        deleteThis();
    else
        update();
}

static int colors[6][3] = {
    { COLOR_WHITE, COLOR_DARKCYAN, COLOR_DARKCYAN },
    { COLOR_WHITE, COLOR_WHITE, COLOR_DARKCYAN },
    { COLOR_WHITE, COLOR_WHITE, COLOR_WHITE },
    { COLOR_BLUE, COLOR_WHITE, COLOR_WHITE },
    { COLOR_BLUE, COLOR_BLUE, COLOR_WHITE },
    { COLOR_BLUE, COLOR_BLUE, COLOR_BLUE },
};

void
VcrBeamSprite::draw(Ptr<GfxCanvas> can)
{
    int x2 = (2*x0+x1)/3;
    int y2 = (2*y0+y1)/3;
    int x3 = (2*x1+x0)/3;
    int y3 = (2*y1+y0)/3;

    ASSERT(status > 0 && status <= 6);
    GfxParameters parm;
    parm.color = standard_colors[colors[status-1][0]];
    drawLine(can, parm, x0, y0, x2, y2);
    parm.color = standard_colors[colors[status-1][1]];
    drawLine(can, parm, x2, y2, x3, y3);
    parm.color = standard_colors[colors[status-1][2]];
    drawLine(can, parm, x3, y3, x1, y1);
}

void
VcrBeamSprite::setBang(VcrPlayer* ply, int side)
{
    if (x0 == x1 && y0 == y1)
        return;                 // can't happen

    int angle = int(std::atan2(double(x1 - x0), double(y1 - y0)) * 180 / 3.14159265);
    Ptr<GfxSprite> s = makeBang(ply, angle, side);
    if (s)
        getManager()->addSprite(s);
}

/***************************** VcrTorpSprite *****************************/

/** \class VcrTorpSprite
    \brief Torpedo */

/** Construct new Torpedo Sprite.
    \param ax0,ay0    launcher coordinates
    \param ax1,ay1    target coordinates (point in space if misses)
    \param alayer     sprite layer
    \param apix       pixmap for torpedo
    \param ahits      object which is hit (0 or 1), or negative for miss */
VcrTorpSprite::VcrTorpSprite(int ax0, int ay0, int ax1, int ay1,
                             int alayer, Ptr<GfxPixmap> apix, int ahits,
                             SfxSample bang_sound)
    : VcrWeaponSprite(GfxRect(ax0, ay0, apix->getWidth(), apix->getWidth()),
                      alayer, ax1, ay1),
      x0(ax0), y0(ay0), pix(apix),
      hit_side(ahits), status(0),
      bang_sound(bang_sound)
{
    x0 -= apix->getWidth() / 2;
    y0 -= apix->getHeight() / 2;
    x1 -= apix->getWidth() / 2;
    y1 -= apix->getHeight() / 2;
    pos.x = x0;
    pos.y = y0;
}

VcrTorpSprite::~VcrTorpSprite()
{ }

void
VcrTorpSprite::tick()
{
    if (status < 8)
        ++status;

    moveCoord(x0, x1, 2*status);
    if (status > 5)
        moveCoord(y0, y1, 1);

    if (x0 == x1) {
        if (hit_side >= 0) {
            if (bang) {
                getManager()->addSprite(bang);
                bang_sound.play(100, panFromX(x0), VCR_SfxImportance);
            }
        }
        deleteThis();
    } else {
        moveTo(x0, y0);
    }
}

void
VcrTorpSprite::draw(Ptr<GfxCanvas> can)
{
    pix->blit(can, pos.x, pos.y);
}

void
VcrTorpSprite::setBang(VcrPlayer* ply, int side)
{
    bang = makeBang(ply, side ? 270 : 90, side);
}

/***************************** VcrTimeSprite *****************************/

VcrTimeSprite::VcrTimeSprite(int xr, int yb, int alayer,
                             Ptr<VcrPlayer> aply)
    : GfxSprite(GfxRect(0, 0, 0, 0), alayer),
      current(-1), vcr_player(aply)
{
    int wi = fonts[FONT_FIXED]->getTextWidth(_("Time: 000:00"));
    int he = fonts[FONT_FIXED]->getTextHeight(_("Time: 000:00"));
    pos = GfxRect(xr-wi, yb-he, wi, he);  // this sucks
}

void
VcrTimeSprite::tick()
{
    if (vcr_player->getTime() != current)
        update();
}

void
VcrTimeSprite::draw(Ptr<GfxCanvas> can)
{
    current = vcr_player->getTime();

    GfxParameters parm;
    parm.color = standard_colors[COLOR_WHITE];
    parm.just_x = 2;
    parm.just_y = 0;

    outText(can, parm, fonts[FONT_FIXED],
            pos.x + pos.w, pos.y,
            format(_("Time: %3d:%02d"), int(current/60), int(current%60)));
}

/*************************** VcrDistanceSprite ***************************/

VcrDistanceSprite::VcrDistanceSprite(int xl, int yb, int alayer,
                             Ptr<VcrPlayer> aply)
    : GfxSprite(GfxRect(0, 0, 0, 0), alayer),
      current(-1), vcr_player(aply)
{
    int wi = fonts[FONT_FIXED]->getTextWidth(_("Distance: 00000 m"));
    int he = fonts[FONT_FIXED]->getTextHeight(_("Distance: 00000 m"));
    pos = GfxRect(xl, yb-he, wi, he);  // this sucks
}

void
VcrDistanceSprite::tick()
{
    if (vcr_player->getDistance() != current)
        update();
}

void
VcrDistanceSprite::draw(Ptr<GfxCanvas> can)
{
    current = vcr_player->getDistance();

    GfxParameters parm;
    parm.color = standard_colors[COLOR_WHITE];
    parm.just_x = 2;
    parm.just_y = 0;

    outText(can, parm, fonts[FONT_FIXED],
            pos.x + pos.w, pos.y,
            format(_("Distance: %5d m"), current));
}

/**************************** VcrResultSprite ****************************/

VcrResultSprite::VcrResultSprite(GfxRect pos, int alayer, Ptr<VcrPlayer> ply,
                                 TVcr& vcr)
    : GfxSprite(pos, alayer), vcr_player(ply), vcr(vcr), res(0)
{ }

void
VcrResultSprite::tick()
{
    int ores = res;
    res = vcr_player->getStatusWord();
    if (res != ores)
        update();
}

void
VcrResultSprite::draw(Ptr<GfxCanvas> can)
{
    GfxParameters parm;
    parm.color = standard_colors[COLOR_WHITE];
    parm.just_x = 1;
    parm.just_y = 0;
    if (res > 0)
        outTextF(can, parm, fonts[FONT_TITLE], pos.x + pos.w/2, pos.y, pos.w,
                 VcrPlayer::getResultString(res, vcr));
}
