/**
  *  \file vcrplay.cc
  *  \brief VCR Player (GUI) Implementation
  *
  *  This implements the VCR player, except for the actual battle
  *  algorithms.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <cstdlib>
#include "client/vcranim.h"
#include "client/vcrplay.h"
#include "client/vcrvis.h"
#include "fileio/io.h"
#include "game/config.h"
#include "game/player.h"
#include "game/spec.h"
#include "game/vcrdb.h"
#include "gfx/anim.h"
#include "gfx/palette.h"
#include "gfx/sprite.h"
#include "ui/msgbox.h"
#include "ui/ui.h"
#include "ui/util.h"
#include "ui/widgets.h"
#include "ui/window.h"
#include "util/console.h"
#include "util/nls.h"
#include "ui/layout.h"
#include "ui/commands.h"
#include "cpluslib/format.h"
#include "game/pref.h"

/****************************** WVcrSelector *****************************/

/** Create a VCR Selector displaying the specified database. */
WVcrSelector::WVcrSelector(Ptr<VcrDatabase> adb)
    : UIWidget(),
      index(-1),
      db(adb),
      playing(false)
{
    player = db->getVcrPlayer();
    player->setVisualizer(new VcrNullVisualizer());

    selectVcr(0);
}

WVcrSelector::~WVcrSelector()
{ }

/** Select VCR to be displayed.
    \param aindex Index, 0-based. */
void
WVcrSelector::selectVcr(int aindex)
{
    if (aindex < 0)
        aindex = 0;
    if (aindex >= db->getVcrCount())
        aindex = db->getVcrCount()-1;
    if (index == aindex)
        return;

    index = aindex;
    entry = db->getEntry(index);
    vcr = entry->getVcr();
    disp_status = entry->getStatus();
    maybeRestartPlayer();
    drawWidget(false);
}

void
WVcrSelector::maybeRestartPlayer()
{
    if (! disp_status) {
        /* We do not know the result. Attempt to compute it. */
        if (player.ptr()) {
            /* we have a VCR player. */
            TVcr vcr_tmp = vcr;
            if (!player->setCapabilities(db->getCapabilities()) || player->checkVcr(vcr_tmp))
                disp_status = -1;
            else
                player->initVcr(vcr);
        } else {
            /* we don't have a VCR player. */
            disp_status = -1;
        }
    }
    setState(st_IdleNotify, disp_status == 0);
}

/** Draw information about warrior.
    \param can  canvas to draw onto
    \param x,y  coordinates
    \param w    width (height is constant 70)
    \param side side (0 or 1) */
void
WVcrSelector::drawWarrior(Ptr<GfxCanvas> can, int x, int y, int w, int side)
{
    // FIXME: this i18n approach is far from perfect
    TVcrObject& vcr_obj = vcr.objects[side];
    bool is_planet = (side == 1) && (vcr.battle_type != 0);
    int ybot = y + 70;

    /* Object title */
    string_t adj = (vcr.objects[side].owner == getPlayerId()
                    ? string_t(_("our"))
                    : string_t(format(_("a %s"), getAdjName(vcr.objects[side].owner))));
#if 0
    /* borland c++ doesn't grok this one */
    string_t type = (is_planet
                     ? string_t(_("planet"))
                     : entry->getGuessedHull(side)
                     ? getHull(entry->getGuessedHull(side)).getShortName()
                     : string_t(_("starship")));
#else
    string_t type;
    if (is_planet)
        type = _("planet");
    else if (entry->getGuessedHull(side))
        type = getHull(entry->getGuessedHull(side)).getShortName();
    else
        type = _("starship");
#endif
    string_t s = (format("%s (Id #%d, %s %s)")
                  << getBasicString(vcr.objects[side].name)
                  << vcr.objects[side].id
                  << adj
                  << type);

    parm.color = standard_colors[fof_colors_dark[isFriend(vcr.objects[side].owner)]];

    outTextF(can, parm, fonts[1], x, y, w, s);
    y += 15;

    /* Shield, Damage, Crew */
    int shield = vcr.shield[side] < 0 ? 0 : vcr.shield[side];
    s = format(_("%d%% shield (%d kt), %d %% damaged")) << shield
                                                        << vcr.mass[side]
                                                        << vcr.objects[side].damage;
    if (!is_planet)
        s += format(_(", %d %1{crewman%|crewmen%}")) << vcr.objects[side].crew;
    parm.color = standard_colors[COLOR_BLACK];
    outTextF(can, parm, fonts[1], x+10, y, w-10, s);
    y += 15;

    /* Beams */
    if (vcr.objects[side].beam_count) {
        outTextF(can, parm, fonts[1], x+10, y, w-10,
                 itoa(vcr.objects[side].beam_count) + " x "
                 + getBeam(vcr.objects[side].beam_type).getName());
        y += 15;
    }

    /* Torps/Fighters */
    if (vcr_obj.bay_count > 0) {
        if (vcr.torp_launcher_count[side] > 0) {
            int tcount = vcr.num_torps[side];
            s = format(_("%d %1{%s%|%ss%} and %d %1{fighter%|fighters%}"))
                << tcount
                << getTorp(vcr_obj.torp_type).getName()
                << vcr.num_fighters[side];
        } else {
            s = format(_("%d fighter bay%!1{s%} with %d fighter%!1{s%}"))
                << vcr_obj.bay_count
                << vcr.num_fighters[side];
        }
        outTextF(can, parm, fonts[1], x+10, y, w-10, s);
        y += 15;
    } else if (vcr.torp_launcher_count[side] > 0) {
        s = format(_("%d x %1{%s launcher%|%s launchers%} with %d torp%!1{s%}"))
                   << vcr.torp_launcher_count[side]
                   << getTorp(vcr_obj.torp_type).getName()
                   << vcr.num_torps[side];
        outTextF(can, parm, fonts[1], x+10, y, w-10, s);
        y += 15;
    } else {
        /* No auxiliary weapons. Can we still give more info? Yes, we
           can. When `NTP' is used, THost clears the "count" field,
           but keeps type and ammo count intact. Nice, isn't it?
           (thanks, Akseli, for that one!) We can still give the number
           of bays if we know the hull. PHost makes the ship appear
           with the correct weapon count, but no ammo, so this doesn't
           work here. */
        int ghull = entry->getGuessedHull(side);
        if (ghull) {
            const GHull& hull = getHull(ghull);
            if (hull.getNumBays()) {
                s = format(_("(%d fighter bay%!1{s%} %snot used)"))
                    << hull.getNumBays()
                    << (vcr.num_fighters[side]
                        ? string_t(format(_("with %d fighter%!1{s%} ")) << vcr.num_fighters[side])
                        : string_t());
                parm.color = standard_colors[COLOR_DARK];
                outTextF(can, parm, fonts[1], x+10, y, w-10, s);
                y += 15;
            } else if (hull.getMaxLaunchers()) {
                int torp_type = vcr.objects[side].torp_type;
                s = format(_("(up to %s %snot used)"))
                    << (torp_type
                        ? string_t(format(_("%d %s%!1{s%}"), hull.getMaxLaunchers(), getTorp(torp_type).getName()))
                        : string_t(format(_("%d torpedo launcher%!1{s%}"), hull.getMaxLaunchers())))
                    << (vcr.num_torps[side]
                        /* This string must end in a space */
                        ? string_t(format(_("with %d torp%!1{s%} ")) << vcr.num_torps[side])
                        : string_t());
                parm.color = standard_colors[COLOR_DARK];
                outTextF(can, parm, fonts[1], x+10, y, w-10, s);
                y += 15;
            }
        }
    }
    ASSERT(y <= ybot);
    can->drawBar(GfxRect(x, y, w, ybot-y), parm.bgcolor,
                 GFX_TRANSPARENT, SOLID_FILL, SDL_ALPHA_OPAQUE);
}

/** Draw battle status. This draws the line at the bottom explaining
    the status (disp_status). */
void
WVcrSelector::drawStatus(Ptr<GfxCanvas> can)
{
    if (user_pref.InstantBattleResult) {
        parm.color = standard_colors[COLOR_BLACK];
        outTextF(can.ptr(), parm, fonts[1].ptr(),
                 extent.x, extent.y + 195,
                 extent.w - 40,     // 40 = width of "Tab" button
                 VcrPlayer::getResultString(disp_status, vcr));
    }
}

void
WVcrSelector::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    parm.color = standard_colors[COLOR_BLACK];
    parm.bgcolor = standard_colors[COLOR_GRAY];
    outTextF(can, parm, fonts[0], extent.x, extent.y, extent.w - 80,
             format(_("Battle #%d of %d:"), index+1, db->getVcrCount()));
    parm.just_x = 2;
    outTextF(can, parm, fonts[1], extent.x + extent.w, extent.y, 80,
             player->getPlayerName());
    parm.just_x = 0;

    drawWarrior(can, extent.x, extent.y + 30, extent.w, 0);
    drawWarrior(can, extent.x, extent.y + 100, extent.w, 1);
    drawStatus(can);
    ui.update(getExtent());
}

void
WVcrSelector::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size = GfxPoint(420, 210);
    info.flags = 0;
}

void
WVcrSelector::playIt(int)
{
    if (disp_status < 0) {
        if (!player->setCapabilities(db->getCapabilities())) {
            messageBox(_("This battle can not be played. The recording uses a feature "
                         "not yet implemented in this VCR player."),
                       _("Error"));
        } else {
            messageBox(_("This battle can not be played. Most likely, you do not "
                         "have the correct configuration file and/or shiplist to "
                         "play it. This error could also be caused by transmission "
                         "errors or (heavens!) bugs in this program or in the Host."),
                       _("Error"));
        }
        return;
    }

    /* we must stop playing because WVcrPlayWindow will use the same
       instance of the player. In addition, it only hurts performance. */
    setState(st_IdleNotify, false);
    playing = true;

    if (VcrBeamSprite::active_fx_sprites) {
        /* FIXME: remove when stable. Should not happen any longer. */
        std::cout << "at start of play: " << VcrBeamSprite::active_fx_sprites
                  << " active FX sprites\n";
        messageBox("There were still " + itoa(VcrBeamSprite::active_fx_sprites)
                   + " active FX sprites from the previous run.", "Warning");
        VcrBeamSprite::active_fx_sprites = 0;
    }

    WVcrPlayWindow win(player, entry);
    ui_root->add(win);
    ui_root->run();
    ui_root->close(win);

    /* There must not be other references to the WVcrPlayWindow,
       so that the destructor runs at the end of this function. */
    // ASSERT(w.ref_count() == 1);

    maybeRestartPlayer();
    playing = false;
}

void
WVcrSelector::next(int)
{
    select();
    selectVcr(index+1);
}

void
WVcrSelector::prev(int)
{
    select();
    selectVcr(index - 1);
}

bool
WVcrSelector::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Command) {
        switch (e.cmd.code) {
         case cm_VCR_Next:
            next();
            return true;
         case cm_VCR_Prev:
            prev();
            return true;
         case cm_VCR_SetBattle:
            select();
            selectVcr(e.cmd.info);
            return true;
         case cm_Enter:
            playIt();
            ui_root->redraw();
            return true;
        }
    } else if (e.type == ev_Idle && !disp_status && !playing) {
        /* Idle Notify, and we have to do something. Play 10 cycles,
           so we waste less time looping through the event dispatcher. */
        ASSERT(player.ptr());
        for (int i = 0; i < 10; ++i) {
            if (! player->playCycle()) {
                /* battle finished */
                entry->setResultFromPlayer(*player);
                disp_status = entry->getStatus();
                select();
                drawWidget(false);
                ASSERT(disp_status != 0);
                setState(st_IdleNotify, false);
                break;
            }
        }
    }
    return UIWidget::handleEvent(e, second_pass);
}
