/**
  *  \file client/vcrvis.cc
  *  \brief VCR Visualisation Classes/Widgets
  *
  *  This module contains the visualisation classes and the
  *  main VCR window (WVcrPlayWindow).
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/vcrvis.h"
#include "game/spec.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "ui/util.h"
#include "ui/widgets.h"
#include "util/console.h"
#include "util/nls.h"
#include "resmgr/resmgr.h"
#include "cpluslib/format.h"
#include "game/player.h"

enum {
    cm_GotoStart, // = cm_FirstUser,
    cm_GotoEnd,
    cm_Backward,
    cm_Forward,
    cm_Play
};

/** Playback speed. Expressed as delay between two battle ticks
    (animation ticks are always at rate of TIMER_HZ). Can be
    zero ("as fast as possible, animation included"). */
int VcrSpriteVisualizer::speed = 2;

/**************************** WVcrShieldWidget ***************************/

/*! \class WVcrShieldWidget
    Displays shield gauge for one side.
    Parent should call drawWidget() to update this one. */

WVcrShieldWidget::WVcrShieldWidget(VcrPlayer* aplay, int aside)
    : UIFixedWidget(0),
      player(aplay), side(aside)
{ }

void
WVcrShieldWidget::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    int shield = player->getShields(side);
    int color = std::min(COLOR_STATUS + shield / 6, COLOR_STATUS + 15);
    int width = extent.w - 3;
    int x1 = extent.x + 1;
    int barw = shield * width / 100;
    can->drawBar(GfxRect(x1, extent.y + 1, barw, extent.h - 3),
                 standard_colors[color], GFX_TRANSPARENT,
                 SOLID_FILL, SDL_ALPHA_OPAQUE);
    can->drawBar(GfxRect(x1 + barw, extent.y + 1, width - barw, extent.h - 3),
                 standard_colors[0], GFX_TRANSPARENT,
                 SOLID_FILL, SDL_ALPHA_OPAQUE);

    GfxParameters parm;
    parm.color = standard_colors[shield > 50 ? COLOR_WHITE :
                                shield > 0 ? COLOR_GRAY : COLOR_DARK];
    parm.just_x = parm.just_y = 1;
    outText(can, parm, fonts[1],
            extent.x + extent.w / 2, extent.y + extent.h / 2,
            format(_("Shields: %d%%"), shield));

    parm.color = standard_colors[0];
    drawRectangle(can, parm, GfxRect(extent.x, extent.y, extent.w - 1, extent.h - 1));
    drawFrameDown(can, extent);

    ui.update(getExtent());
}

/**************************** WVcrDamageWidget ***************************/

/*! \class WVcrDamageWidget
    Displays damage/crew for one side of the VCR. Parent should call
    drawWidget() to update this one. */

WVcrDamageWidget::WVcrDamageWidget(VcrPlayer* aplay, int aside, bool aplanet)
    : UIFixedWidget(),
      player(aplay), side(aside), is_planet(aplanet)
{ }

void
WVcrDamageWidget::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    GfxParameters parm;

    drawFrameDown(can, extent);

    parm.color = standard_colors[0];
    drawBar(can, parm, GfxRect(extent.x, extent.y, extent.w - 1, extent.h - 1));

    parm.color = standard_colors[COLOR_WHITE];
    outText(can, parm, fonts[1], extent.x + 1, extent.y + 1,
            format(_("Damage: %d%%"), player->getDamage(side)));
    if (!is_planet)
        outText(can, parm, fonts[1], extent.x + 1, extent.y + 17,
                format(_("Crew: %d"), player->getCrew(side)));

    ui.update(extent);
}

/**************************** WVcrBeamDisplay ****************************/

/** Draw weapon system charge bar.
    \param can   Canvas
    \param x,y   Coordinates. Bar occupies GfxRect(x, y, 101, scale+1)
    \param len   Charge status, [0,100]
    \param scale Scaling factor = height of bar */
static void
drawWeaponBar(GfxCanvas* can, int x, int y, int len, int scale)
{
    if (len > 100)
        len = 100;

    int color = COLOR_STATUS + (len >> 2);
    if (color > COLOR_STATUS + 15)
        color = COLOR_STATUS + 15;

    can->drawBar(GfxRect(x, y, len, scale), standard_colors[color],
                 GFX_TRANSPARENT, SOLID_FILL, SDL_ALPHA_OPAQUE);
    can->drawBar(GfxRect(x + len, y, 101-len, scale), standard_colors[COLOR_BLACK],
                 GFX_TRANSPARENT, SOLID_FILL, SDL_ALPHA_OPAQUE);
    can->drawHLine(x, y + scale, x + 101, standard_colors[COLOR_BLACK],
                   0xFFU, SDL_ALPHA_OPAQUE);
}

WVcrBeamDisplay::WVcrBeamDisplay(int acount, int atype, int ascale)
    : count(acount), type(atype), scale(ascale)
{
    std::fill(status, status + countof(status), 0);
    pref_size = count * (scale + 1) + 4 + font_heights[FONT_SMALL];
}

void
WVcrBeamDisplay::drawOne(Ptr<GfxCanvas> can, int aindex)
{
    int y = getBeamY (aindex);
    int x = extent.x + 1;
    drawWeaponBar(can, x, y, status[aindex], scale);
}

void
WVcrBeamDisplay::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    GfxParameters parm;

    drawSolidBar(can, extent, standard_colors[COLOR_BLACK]);
    drawFrameDown(can, extent);

    parm.color = standard_colors[COLOR_WHITE];
    outText(can, parm, fonts[2], extent.x + 1, extent.y + 2,
            getBeam(type).getName());

    for (int i = 0; i < count; ++i)
        drawOne(can, i);

    ui.update(extent);
}

void
WVcrBeamDisplay::updateOne(int aindex, int aval)
{
    if (status[aindex] != aval) {
        status[aindex] = aval;
        if (hasState(st_Exposed)) {
            drawOne(ui_screen, aindex);
            ui_screen->update(GfxRect(extent.x + 1, getBeamY (aindex),
                                      101, scale));
        }
    }
}

/**************************** WVcrTorpDisplay ****************************/

WVcrTorpDisplay::WVcrTorpDisplay(int acount, int atype, int ascale)
    : count(acount), type(atype), scale(ascale), torp_count(-1)
{
    std::fill(status, status + countof(status), 0);
    pref_size = count * (scale + 1) + 4 + 2*font_heights[FONT_SMALL];
}

void
WVcrTorpDisplay::drawCount(Ptr<GfxCanvas> can)
{
    int y = getTorpY(count);
    int x = extent.x + 2;

    GfxParameters parm;
    parm.color = standard_colors[COLOR_WHITE];
    parm.bgcolor = standard_colors[COLOR_BLACK];
    outTextF(can, parm, fonts[FONT_SMALL], x, y, 100,
             format(_("Torpedoes: %d"), torp_count));
}

void
WVcrTorpDisplay::drawOne(Ptr<GfxCanvas> can, int aindex)
{
    int y = getTorpY (aindex);
    int x = extent.x + 1;
    drawWeaponBar(can, x, y, status[aindex], scale);
}

void
WVcrTorpDisplay::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    GfxParameters parm;

    drawSolidBar(can, extent, standard_colors[COLOR_BLACK]);
    drawFrameDown(can, extent);

    parm.color = standard_colors[COLOR_WHITE];
    outText(can, parm, fonts[2], extent.x + 2, extent.y + 2,
            getTorp(type).getName());
    drawCount(can);

    for (int i = 0; i < count; ++i)
        drawOne(can, i);

    ui.update(extent);
}

void
WVcrTorpDisplay::updateOne(int aindex, int aval)
{
    if (status[aindex] != aval) {
        status[aindex] = aval;
        if (hasState(st_Exposed)) {
            drawOne(ui_screen, aindex);
            ui_screen->update(GfxRect(extent.x + 1, getTorpY (aindex), 101, scale));
        }
    }
}

void
WVcrTorpDisplay::updateCount(int acount)
{
    if (acount == torp_count)
        return;

    torp_count = acount;
    if (hasState(st_Exposed)) {
        drawCount(ui_screen);
        ui_screen->update(GfxRect(extent.x + 2, getTorpY(count),
                                  101, font_heights[FONT_SMALL]));
    }
}

/*************************** WVcrFighterDisplay **************************/

WVcrFighterDisplay::WVcrFighterDisplay(int abays, int aftrs)
    : bays(abays), fighters(aftrs)
{ }

void
WVcrFighterDisplay::updateCount(int aftrs)
{
    if (aftrs != fighters) {
        fighters = aftrs;
        drawWidget(true);
    }
}

void
WVcrFighterDisplay::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    GfxParameters parm;

    drawSolidBar(can, extent, standard_colors[COLOR_BLACK]);
    drawFrameDown(can, extent);

    parm.color = standard_colors[COLOR_WHITE];
    outText(can, parm, fonts[FONT_SMALL],  extent.x + 2, extent.y + 2,
            format(_("%d fighter bay%!1{s%}"), bays));
    outTextF(can, parm, fonts[FONT_SMALL], extent.x + 2, extent.y + font_heights[FONT_SMALL] + 4,
             extent.w - 4, format(_("Fighters: %d"), fighters));

    ui.update(extent);
}

/************************** VcrSpriteVisualizer **************************/

/* must be here */
struct VcrSpriteVisualizer::Status {
    Ptr<VcrStatus> play_status;
    VcrFighterSprite::Status fighters[2][VcrPlayer::VCRMF_MAX_FTRS];

    Status(VcrSpriteVisualizer* vis);
};

/*! \class VcrSpriteVisualizer
    Displays battles, includes playback control (fast forward/rewind).
    Assumes to be displayed on a screen together with a WVcrPlayWindow.

    FIXME: this still uses Ptr for widget lifetime control.
 */

/** Initialize visualizer.
    \param vcr      The VCR we'll play;
    \param awindow  Window. Constructor will add widgets.
    \param player   Player that'll be used. */
VcrSpriteVisualizer::VcrSpriteVisualizer(const TVcr& vcr,
                                         WVcrPlayWindow* awindow,
                                         Ptr<VcrPlayer> player)
    : the_vcr(vcr), the_window(awindow),
      playing(true), done(false), disable_fx(false),
      ticks_per_status(VCRP_ORIG_TICKS)
{
    m.setBoundingBox(GfxRect(0, 0, 640, 240));
    objs[0] = ResManager::getInstance().getPixmap(mkResId(rt_Ship, the_vcr.objects[0].picture + 1000), rc_RightSmall);
    if (!objs[0])
        objs[0] = requestPixmap("vcr.ship");
    if (vcr.battle_type == 0) {
        objs[1] = ResManager::getInstance().getPixmap(mkResId(rt_Ship, the_vcr.objects[1].picture + 1000), rc_LeftSmall);
        if (!objs[1]) {
            objs[1] = requestPixmap("vcr.ship")->clone();
            objs[1]->flipHorizontal();
        }
    } else {
        objs[1] = requestPixmap("vcr.planet");
    }
    ftrs[0][0] = requestPixmap("vcr.ftr", vcr.objects[0].owner);
    ftrs[1][1] = requestPixmap("vcr.ftr", vcr.objects[1].owner);
    ftrs[0][1] = ftrs[0][0]->clone();
    ftrs[1][0] = ftrs[1][1]->clone();
    ftrs[0][1]->flipHorizontal();
    ftrs[1][0]->flipHorizontal();
    SDL_SetColorKey(ftrs[0][0]->getSurface(), SDL_SRCCOLORKEY, 255);
    SDL_SetColorKey(ftrs[0][1]->getSurface(), SDL_SRCCOLORKEY, 255);
    SDL_SetColorKey(ftrs[1][0]->getSurface(), SDL_SRCCOLORKEY, 255);
    SDL_SetColorKey(ftrs[1][1]->getSurface(), SDL_SRCCOLORKEY, 255);
    Ptr<GfxPixmap> pix = requestPixmap("vcr.torpedo");
    torps[0] = pix->clone();
    torps[1] = pix->clone();
    torps[1]->flipHorizontal();

    if (ui_screen->getSurface()->format->BytesPerPixel > 1) {
        /* only looks good in truecolor visuals */
        Ptr<GfxPixmap> pix = requestPixmap("vcr.bg");
        if (pix)
            m.addSprite(new GfxPixmapSprite(0, 0, VCRPLAY_LAYER_SCENE, pix));
    }
    m.addSprite(new VcrTimeSprite(640, 240, VCRPLAY_LAYER_TEXT, player));
    m.addSprite(new VcrDistanceSprite(0, 240, VCRPLAY_LAYER_TEXT, player));
    m.addSprite(new VcrResultSprite(GfxRect(0, 10, 640, font_heights[FONT_TITLE]),
                                    VCRPLAY_LAYER_TEXT, player, the_vcr));

    findMountpoints(0);
    findMountpoints(1);

    counter = speed;

    the_window->add(GfxRect(19,  400, 107, 20),
                    *(shieldw[0] = new WVcrShieldWidget(player, 0)));
    the_window->add(GfxRect(514, 400, 107, 20),
                    *(shieldw[1] = new WVcrShieldWidget(player, 1)));

    the_window->add(GfxRect(19,  422, 107, 34),
                    *(damagew[0] = new WVcrDamageWidget(player, 0, false)));
    if (the_vcr.battle_type == 0)
        the_window->add(GfxRect(514, 422, 107, 34),
                        *(damagew[1] = new WVcrDamageWidget(player, 1, false)));
    else
        the_window->add(GfxRect(514, 422, 107, 18),
                        *(damagew[1] = new WVcrDamageWidget(player, 1, true)));

    createWeaponWidgets(136, 0);
    createWeaponWidgets(399, 1);

    for (int i = 0; i < 2; ++i) {
        if (vcr.objects[i].beam_type > 0)
            snd_beam[i] = requestSound("vcr.beam", vcr.objects[i].beam_type);
        if (vcr.objects[i].beam_type > 0)
            snd_torp[i] = requestSound("vcr.tlaun", vcr.objects[i].torp_type);
    }
    snd_flaunch   = requestSound("vcr.flaun");
    snd_fbeam     = requestSound("vcr.beam", 0);
    snd_fexplode  = requestSound("vcr.bangf");
    snd_damaged   = requestSound("vcr.damaged");
    snd_destroyed = requestSound("vcr.bang");
    snd_torp_hit  = requestSound("vcr.thit");
}

/** Distribute weapon mount points.
    \param possible   (in) List of possible points;
    \param mount      (out) Used points;
    \param num        (in) Number of points we want. */
static void
distributeMountpoints(const std::vector<GfxPoint>& possible,
                      GfxPoint mount[],
                      int num)
{
    for (int x = 0; x < num; ++x) {
        int thispt = x * (possible.size()-1) / num;
        int nextpt = (x+1) * (possible.size()-1) / num;
        mount[x] = possible[(thispt + nextpt) / 2];
    }
}

/** Find weapon mount points (origins of shots). */
void
VcrSpriteVisualizer::findMountpoints(int side)
{
    Ptr<GfxPixmap> pix = objs[side];
    SDL_Surface* sfc = pix->getSurface();

    /* fallback version. */
    for (int i = 0; i < the_vcr.objects[side].beam_count; ++i)
        beam_mount[side][i] = GfxPoint(sfc->w/2, sfc->h/2);

    if (sfc->format->BytesPerPixel != 1) {
        /* we can only handle 8-bit pixmaps. */
        //console.write(LOG_WARN, _("VCR: object pixmap is not 8-bit."));
        static SDL_Palette pal = {
            NUM_STANDARD_COLORS,
            standard_palette
        };
        static SDL_PixelFormat fmt = {
            &pal,
            8, 1,               // bits/bytes per pixel
            0, 0, 0, 0,         // rgbamask
            0, 0, 0, 0,         // rgbashift
            0, 0, 0, 0,         // rgbaloss
            255,                // colorkey
            SDL_ALPHA_OPAQUE    // alpha
        };
        Ptr<GfxPixmap> p = pix->convertTo(&fmt);
        if (p) {
            pix = objs[side] = p;
            sfc = pix->getSurface();
            SDL_SetColorKey(sfc, SDL_SRCCOLORKEY, 0xFF);
        } else {
            console.write(LOG_WARN, _("VCR: object pixmap is not 8 bit, unable to convert"));
            return;
        }
    }

    /* Convert colors, somehow. Because we know nothing about
       transparency in the original image, make everything black
       transparent. */
    SDL_LockSurface(sfc);
    for (int y = 0; y < sfc->h; ++y) {
        uint8* line = static_cast<uint8*>(sfc->pixels) + sfc->pitch * y;
        for (int x = 0; x < sfc->w; ++x) {
            if (line[x] >= NUM_STANDARD_COLORS) {
                line[x] = 255;
            } else {
                const SDL_Color& c = standard_palette[line[x]];
                if ((c.r | c.g | c.b) == 0)
                    line[x] = 255;
            }
        }
    }
    SDL_UnlockSurface(sfc);
    SDL_UpdateRect(sfc, 0, 0, 0, 0);
    SDL_SetColorKey(sfc, SDL_SRCCOLORKEY, 0xFF);

    /* okay, find correct mount points. First, scan the image for
       possible points. */
    std::vector<GfxPoint> possible;
    for (int y = 0; y < sfc->h; ++y) {
        uint8* line = static_cast<uint8*>(sfc->pixels) + sfc->pitch * y;
        if (side == 0) {
            for (int x = sfc->w; x > 0; --x)
                if (line[x-1] != 255) {
                    possible.push_back(GfxPoint(x, y));
                    break;
                }
        } else {
            for (int x = 0; x < sfc->w; ++x)
                if (line[x] != 255) {
                    possible.push_back(GfxPoint(x, y));
                    break;
                }
        }
    }

    /* At this point, we have possible mount points. */
    if (possible.empty()) {
        console.write(LOG_WARN, _("VCR: object pixmap is empty?"));
        return;
    }

    distributeMountpoints(possible, beam_mount[side], the_vcr.objects[side].beam_count);
    distributeMountpoints(possible, torp_mount[side], the_vcr.torp_launcher_count[side]);
}

/** Every-tick update. Ensures fighters move correctly, and correct
    shields/damage are displayed. */
void
VcrSpriteVisualizer::update()
{
    if (objsprite[0])
        objsprite[0]->moveTo(ply->getObjectX(0) - objsprite[0]->getPos().w,
                             VCRPLAY_OBJECT_Y   - objsprite[0]->getPos().h/2);
    if (objsprite[1])
        objsprite[1]->moveTo(ply->getObjectX(1),
                             VCRPLAY_OBJECT_Y   - objsprite[1]->getPos().h/2);

    shieldw[0]->drawWidget(true);
    shieldw[1]->drawWidget(true);
    damagew[0]->drawWidget(true);
    damagew[1]->drawWidget(true);
}

/** Update after rewind/fast forward. */
void
VcrSpriteVisualizer::updateAfterWind()
{
    updatePanels();
    update();
}

/** Update weapon panels. Call at startup (to initialize them), and
    at end. */
void
VcrSpriteVisualizer::updatePanels()
{
    for (int i = 0; i < 2; ++i) {
        /* Beam panel (bars) */
        for (int b = 0; b < the_vcr.objects[i].beam_count; ++b)
            updateBeam(i, b);
        /* Torpedo panels (bars + count) */
        for (int l = 0; l < the_vcr.torp_launcher_count[i]; ++l)
            updateTorp(i, l);
        if (torpw[i])
            torpw[i]->updateCount(the_vcr.num_torps[i]);
        updateFighter(i);
    }
}

void
VcrSpriteVisualizer::redraw()
{ }

/** Start a fighter. The VcrFighterSprite automatically moves. */
void
VcrSpriteVisualizer::startFighter(int side, int track)
{
    VcrFighterSprite* bs =
        new VcrFighterSprite(objsprite[side]->getPosCX(),
                             objsprite[side]->getPosCY(),
                             VCRPLAY_LAYER_FTR,
                             side, track,
                             the_vcr.objects[side].owner,
                             ply,
                             ftrs[side][0], ftrs[side][1]);
    m.addSprite(fighters[side][track] = bs);
    updateFighter(side);
    if (!disable_fx)
        snd_flaunch.play(100, panFromX(objsprite[side]->getPosCX()), VCR_SfxImportance);
}

void
VcrSpriteVisualizer::landFighter(int side, int track)
{
    m.deleteSprite(fighters[side][track]);
    fighters[side][track] = 0;
    updateFighter(side);
}

/** Kill a fighter. Removes the fighter sprite and generates the bang. */
void
VcrSpriteVisualizer::killFighter(int side, int track)
{
    /* FIXME: the disable_fx distinction is probably not needed.
       Rewind looks pretty cool when we always do ...->kill(). */
    ASSERT(fighters[side][track]);
    if (!disable_fx) {
        fighters[side][track]->kill();
        snd_fexplode.play(30, panFromX(fighters[side][track]->getPosCX()), VCR_SfxImportance);
    } else {
        fighters[side][track]->deleteThis();
    }
    fighters[side][track] = 0;
    updateFighter(side);
}

/* hit >= 0: hit, hit < 0: miss */
void
VcrSpriteVisualizer::fireBeam(int side, int track, int target, int hit,
                              int damage, int kill)
{
    /* no effect when rewind/ffwd active */
    if (disable_fx)
        return;

    /* FIXME: missing beams should fire elsewhere */
    int x1 = track < 0 ? objsprite[side]->getPos().x + beam_mount[side][-track-1].x : fighters[side][track]->getPosCX();
    int y1 = track < 0 ? objsprite[side]->getPos().y + beam_mount[side][-track-1].y : fighters[side][track]->getPosCY();
    int x2 = target < 0 ? objsprite[!side]->getPosCX() : fighters[!side][target]->getPosCX();
    int y2 = target < 0 ? objsprite[!side]->getPosCY() : fighters[!side][target]->getPosCY();

    Ptr<VcrBeamSprite> beam = new VcrBeamSprite(x1, y1, x2, y2, VCRPLAY_LAYER_BEAM, damage, kill);
    m.addSprite(beam);
    if (target < 0 && hit >= 0)
        beam->setBang(ply, !side);

    if (track < 0)
        snd_beam[side].play(100, panFromX(x1), VCR_SfxImportance);
    else
        snd_fbeam.play(100, panFromX(x1), VCR_SfxImportance);
}

void
VcrSpriteVisualizer::fireTorp(int side, int hit, int launcher)
{
    /* no effect when rewind/ffwd active */
    if (disable_fx)
        return;

    int x1 = objsprite[side]->getPos().x + torp_mount[side][launcher].x;
    int y1 = objsprite[side]->getPos().y + torp_mount[side][launcher].y;
    int x2, y2;

    if (hit >= 0) {
        x2 = objsprite[!side]->getPosCX();
        // y2 = y1;                // torp moves horizontally
        y2 = objsprite[!side]->getPos().y +
            torp_mount[side][launcher].y * objsprite[!side]->getHeight() / objsprite[side]->getHeight();
        Ptr<VcrTorpSprite> torp = new VcrTorpSprite(x1, y1, x2, y2, VCRPLAY_LAYER_TORP,
                                                    torps[side], !side, snd_torp_hit);
        m.addSprite(torp);
        torp->setBang(ply, !side);
    } else {
        x2 = side ? 0 : 640;
        y2 = rand() % 240;
        m.addSprite(new VcrTorpSprite(x1, y1, x2, y2, VCRPLAY_LAYER_TORP,
                                      torps[side], -1, snd_torp_hit));
    }
    torpw[side]->updateCount(ply->getTorpCount(side));

    snd_torp[side].play(100, panFromX(x1), VCR_SfxImportance);
}

void
VcrSpriteVisualizer::killObject(int side)
{
    if (objsprite[side]) {
        m.addSprite(new VcrExplSprite(objsprite[side]->getPosCX(),
                                      objsprite[side]->getPosCY(),
                                      VCRPLAY_LAYER_BANG,
                                      48, 1, 0, 0));
        if (!disable_fx)
            snd_destroyed.play(100, panFromX(objsprite[side]->getPosCX()), VCR_SfxImportance);
    }
    if (!side || the_vcr.battle_type == 0) {
        objsprite[side]->deleteThis();
        objsprite[side] = 0;
    }

#if 0
    for (int i = 0; i < VcrPlayer::VCRMF_MAX_FTRS; ++i) {
        if (fighters[side][i]) {
            // killFighter(side, i);     // impressive but slow :)
            fighters[side][i]->deleteThis();
            fighters[side][i] = 0;
        }
    }
#endif
}

/** Create sprites for objects. Called at beginning, and after rewind. */
void
VcrSpriteVisualizer::createObjectSprites()
{
    objsprite[0] = new GfxPixmapSprite(ply->getObjectX(0) - objs[0]->getWidth(),
                                       VCRPLAY_OBJECT_Y - objs[0]->getHeight()/2,
                                       VCRPLAY_LAYER_OBJECTS, objs[0]);
    objsprite[1] = new GfxPixmapSprite(ply->getObjectX(1),
                                       VCRPLAY_OBJECT_Y - objs[1]->getHeight()/2,
                                       VCRPLAY_LAYER_OBJECTS, objs[1]);
    m.addSprite(objsprite[0]);
    m.addSprite(objsprite[1]);
}

void
VcrSpriteVisualizer::init()
{
    createObjectSprites();
    updatePanels();
}

void
VcrSpriteVisualizer::updateTorp(int side, int id)
{
    if (!disable_fx)
        torpw[side]->updateOne(id, ply->getTorpStatus(side, id));
}

void
VcrSpriteVisualizer::updateBeam(int side, int id)
{
    if (!disable_fx)
        beamw[side]->updateOne(id, ply->getBeamStatus(side, id));
}

void
VcrSpriteVisualizer::updateFighter(int side)
{
    if (!disable_fx && fighterw[side])
        fighterw[side]->updateCount(ply->getFighterCount(side));
}

/** Save current status. Call this after every ply->playCycle().
    Causes the status parameters to be updated accordingly. */
void
VcrSpriteVisualizer::saveStatus()
{
    if (ply->getTime() % ticks_per_status)
        return;

    /* when we run out of space (>5000 ticks default), increase interval
       between checkpoints */
    int slot;
    while ((slot = ply->getTime() / ticks_per_status) >= VCRP_MAX_STATUS) {
        ticks_per_status *= 2;
        for (unsigned i = 0; i < VCRP_MAX_STATUS/2; ++i)
            status[i] = status[2*i];
        for (unsigned i = VCRP_MAX_STATUS/2; i < VCRP_MAX_STATUS; ++i)
            status[i] = 0;
    }
    if (status[slot]) {
        /* FIXME: ASSERT(*status[i] == correct status) */
        return;
    }
    status[slot] = new Status(this);
}

/** Restore status from specified slot. Afterwards, everything is in
    consistent state *except* that the playground is not up-to-date. */
void
VcrSpriteVisualizer::restoreStatus(int slot)
{
    ASSERT(status[slot]);
    for (int side = 0; side < 2; ++side)
        for (int ftr = 0; ftr < VcrPlayer::VCRMF_MAX_FTRS; ++ftr)
            fighters[side][ftr] = 0;

    // keep TEXT layer
    m.deleteLayer(VCRPLAY_LAYER_TORP);
    m.deleteLayer(VCRPLAY_LAYER_BEAM);
    m.deleteLayer(VCRPLAY_LAYER_BANG);
    m.deleteLayer(VCRPLAY_LAYER_FTR);
    m.deleteLayer(VCRPLAY_LAYER_OBJECTS);
    // keep SCENE layers

    m.tickSkip(false);               // perform "delete" orders

    Status*const st = status[slot];
    ply->setStatus(st->play_status);
    ASSERT(ply->getTime() == slot * ticks_per_status);
    for (int side = 0; side < 2; ++side) {
        for (int ftr = 0; ftr < VcrPlayer::VCRMF_MAX_FTRS; ++ftr) {
            if (st->fighters[side][ftr].status)
                m.addSprite(fighters[side][ftr] =
                            new VcrFighterSprite(st->fighters[side][ftr],
                                                 VCRPLAY_LAYER_FTR,
                                                 side, ftr,
                                                 the_vcr.objects[side].owner,
                                                 ply,
                                                 ftrs[side][0],
                                                 ftrs[side][1]));
            else
                fighters[side][ftr] = 0;
        }
    }
    createObjectSprites();
}

/** Get latest known time. That's the latest time to which we can go by
    means of restoreStatus(). */
VcrPlayer::time_type
VcrSpriteVisualizer::getLatestStoredTime() const
{
    int i = VCRP_MAX_STATUS;
    while (--i >= 0)
        if (status[i])
            return i * ticks_per_status;
    return 0;
}

/** Wind to end. Right now, implemented as a windTo() call, but that's
    not guaranteed. */
void
VcrSpriteVisualizer::windToEnd()
{
    // FIXME: I'd like to use numeric_limits<VcrPlayer::time_type>, but
    // gcc-2.95.2 doesn't have it.
    windTo(0x7FFFFFFFL, false);
    if (!isDone())
        playing = true;
    the_window->updateStatus();
}

/** Wind playback to specified time, if possible. Does everything
    needed to do that, no postprocessing necessary. */
void
VcrSpriteVisualizer::windTo(VcrPlayer::time_type t, bool restart_play)
{
    /* only restart playing if we are currently playing */
    if (restart_play && !isPlaying())
        restart_play = false;
    
    VcrPlayer::time_type current = ply->getTime();

    /* preprocess */
    if (t < 0)
        t = 0;
    if (t == current)
        return;

    /* cases:
         1) t > current time
           1a) current time < lkt < t: go to lkt, play
           1b) current time < t < lkt: go to t
           1c) current time >= lkt: play
         2) t = current time
         3) t < current time: go to slot t / ticks_per_status, play */
    if (t > current) {
        /* case 1 */
        VcrPlayer::time_type lkt = getLatestStoredTime();
        if (current < lkt) {
            if (t < lkt)
                restoreStatus(t / ticks_per_status);
            else
                restoreStatus(lkt / ticks_per_status);
        } else {
            /* we're about to go forward, which is deadly when the battle
               is already at end. */
            if (done)
                return;
        }
    } else if (t < current) {
        /* case 3 */
        restoreStatus(t / ticks_per_status);
    }

    /* "play" part */
    done = playing = false;
    disable_fx = true;
    while (!done && ply->getTime() < t) {
        saveStatus();
        if (! ply->playCycle()) {
            TVcr tmp_vcr;
            ply->doneVcr(tmp_vcr);
            done = true;
        }
        /* FIXME: copied off update() */
        for (int side = 0; side < 2; ++side) {
            if (objsprite[side])
                objsprite[side]->moveToC(ply->getObjectX(side), VCRPLAY_OBJECT_Y);
        }
        m.tickSkip();
    }
    disable_fx = false;
    updateAfterWind();
    ui_screen->lock();
    m.draw(ui_screen);
    ui_screen->unlock();
    the_window->updateStatus();       // update buttons
    if (restart_play && !isDone())
        startPlaying();
}

/** Play one tick. Does all necessary processing.
    \param force Force advance of a tick, even when /playing/ is currently
                 false. */
void
VcrSpriteVisualizer::playTick(bool force)
{
    if (!playing && !force)
        return;

    if (!counter && !done && !VcrWeaponSprite::active_fx_sprites) {
        saveStatus();
        if (! ply->playCycle()) {
            TVcr tmp_vcr;
            ply->doneVcr(tmp_vcr);
            done = true;
            the_window->updateStatus();
        }
        update();
        counter = speed;
    }
    if (counter)
        --counter;

    /* FIXME: valid? Should probably be
       the_window->getOwner()->getCanvasFor(the_window),
       but windTo uses ui_screen, too. */
    m.tick(ui_screen);

    if (playing && isDone()) {
        playing = false;
        the_window->updateStatus();
    }
}

/** Create weapon widgets (torp/beam/fighter status gauges). */
void
VcrSpriteVisualizer::createWeaponWidgets(int x, int side)
{
    int scale, count, y;

    /* compute size of status bars */
    count = (the_vcr.objects[side].beam_count + the_vcr.torp_launcher_count[side]);
    if (the_vcr.objects[side].bay_count)
        count += 5;
    if (count >= 34)
        scale = 1;
    else if (count >= 26)
        scale = 2;
    else
        scale = 3;

    /* create widgets */
    y = 300;
    if (the_vcr.objects[side].beam_count) {
        beamw[side] = new WVcrBeamDisplay(the_vcr.objects[side].beam_count,
                                          the_vcr.objects[side].beam_type,
                                          scale);
        the_window->add(GfxRect(x, y, 103, beamw[side]->getPreferredHeight()),
                        *beamw[side]);
        y += beamw[side]->getPreferredHeight() + 2;
    } else {
        beamw[side] = 0;
    }

    if (the_vcr.torp_launcher_count[side]) {
        torpw[side] = new WVcrTorpDisplay(the_vcr.torp_launcher_count[side],
                                          the_vcr.objects[side].torp_type,
                                          scale);
        the_window->add(GfxRect(x, y, 103, torpw[side]->getPreferredHeight()),
                        *torpw[side]);
        y += torpw[side]->getPreferredHeight() + 2;
    } else {
        torpw[side] = 0;
    }

    if (the_vcr.objects[side].bay_count) {
        fighterw[side] = new WVcrFighterDisplay(the_vcr.objects[side].bay_count,
                                                the_vcr.num_fighters[side]);
        the_window->add(GfxRect(x, y, 103, 2*font_heights[FONT_SMALL] + 6),
                        *fighterw[side]);
        y += fighterw[side]->getExtent().h + 2;
    } else {
        fighterw[side] = 0;
    }
}

/** Start playing (">" button). */
void
VcrSpriteVisualizer::startPlaying()
{
    if (playing)
        return;

    if (isDone())
        windTo(0, false);

    playing = true;
    the_window->updateStatus();
}

/** Stop playing (">" or "[]" button). */
void
VcrSpriteVisualizer::stopPlaying()
{
    if (!playing)
        return;

    playing = false;
    the_window->updateStatus();
}

/** Change playback speed.
    \param dif relative change (positive: faster, negative: slower) */
void
VcrSpriteVisualizer::changeSpeed(int dif)
{
    int new_speed = speed - dif;
    if (new_speed < 0)
        new_speed = 0;
    if (new_speed > VCRP_MAX_SPEED)
        new_speed = VCRP_MAX_SPEED;

    if (new_speed != speed) {
        speed = new_speed;
        if (counter > speed)
            counter = speed;
        the_window->updateStatus();
    }
}

/** Save current status. */
VcrSpriteVisualizer::Status::Status(VcrSpriteVisualizer* vis)
    : play_status(vis->getPeer()->getStatusToken())
{
    for (int side = 0; side < 2; ++side) {
        for (int i = 0; i < VcrPlayer::VCRMF_MAX_FTRS; ++i) {
            VcrFighterSprite* fs = vis->fighters[side][i];
            if (fs) {
                fs->storeStatus(fighters[side][i]);
            } else {
                fighters[side][i].status = 0;
            }
        }
    }
}

/***************************** WVcrPlayWindow ****************************/

WVcrPlayWindow::WVcrPlayWindow(Ptr<VcrPlayer> aply, Ptr<VcrEntry> aen)
    : UIGroup(),
      player(aply),
      old_vis(aply->getVisualizer()),
      entry(aen),
      the_vcr(aen->getVcr())
{
    setExtent(GfxRect(ui_screen->getSurface()), true);
    setState(st_Modal, true);

    add(GfxRect( 19, 300, 107, 98),
        holder.add(new UIImageWidget(ResManager::getInstance().getPixmap(mkResId(rt_Ship, the_vcr.objects[0].picture + 1000), rc_Normal))));
    add(GfxRect(514, 300, 107, 98),
        holder.add(new UIImageWidget(ResManager::getInstance().getPixmap(mkResId(rt_Ship, the_vcr.objects[1].picture + 1000), rc_Normal))));
    
    add(GfxRect(210, 443, 24, 24),
        holder.add(new UIButton(cm_GotoStart, "\022\021", ss_Alt + SDLK_LEFT)));
    add(GfxRect(235, 443, 24, 24),
        holder.add(new UIButton(cm_Backward,  "\021\021", ss_Ctrl + SDLK_LEFT)));
    add(GfxRect(260, 443, 24, 24),
        holder.add(new UIButton(cm_Play,      "\020",  SDLK_RIGHT)));
    add(GfxRect(285, 443, 24, 24),
        holder.add(new UIButton(cm_Forward,   "\020\020", ss_Ctrl + SDLK_RIGHT)));
    add(GfxRect(310, 443, 24, 24),
        holder.add(new UIButton(cm_GotoEnd,   "\020\022", ss_Alt + SDLK_RIGHT)));
    add(GfxRect(340, 443, 45, 24),
        holder.add(new UIButton(0, "ESC", UIKey(SDLK_ESCAPE), UIButton::aStop)));
    add(GfxRect(390, 443, 24, 24),
        holder.add(new UIButton(0,             "H",  UIKey('H'))));

    sprite_vis = new VcrSpriteVisualizer(the_vcr, this, player);
    player->setVisualizer(sprite_vis);
    player->initVcr(the_vcr);
    updateStatus();
}

WVcrPlayWindow::~WVcrPlayWindow()
{
    player->setVisualizer(old_vis);
}

void
WVcrPlayWindow::updateStatus()
{
    setState(st_IdleNotify, VcrSpriteVisualizer::speed == 0 && sprite_vis->isPlaying());
    setStateOf(cm_Play, UIButton::bf_Highlight, sprite_vis->isPlaying());
}

void
WVcrPlayWindow::drawInterior(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    can->drawBar(getExtent(), standard_colors[COLOR_BLACK],
                 GFX_TRANSPARENT, SOLID_FILL, SDL_ALPHA_OPAQUE);
    sprite_vis->m.draw(can);

    UIWindow::draw(can, GfxRect(0, 240, 640, 240), COLOR_GRAY, // FIXME: should be gray texture
                   _("Visual Combat Recorder"));

    /* Names of ships */
    GfxParameters parm;
    parm.color  = standard_colors[COLOR_BLACK];
    parm.just_x = parm.just_y = 0;
    outText(can, parm, fonts[0], 19, 270, getBasicString(the_vcr.objects[0].name));
    parm.just_x = 2;
    outText(can, parm, fonts[0], 620, 270, getBasicString(the_vcr.objects[1].name));

    /* Race names */
    GfxRect r(245, 240+95, 150, 65);
    drawFrameDown(can, r);
    r.grow(-1,-1);
    drawSolidBar(can, r, standard_colors[COLOR_SHIELD+3]);
    parm.just_x = 1;
    parm.color  = standard_colors[COLOR_WHITE];
    outTextF(can, parm, fonts[FONT_NORMAL], 320, 240+100, 140, getShortName(the_vcr.objects[0].owner));
    outTextF(can, parm, fonts[FONT_NORMAL], 320, 240+120, 140, _("- versus -"));
    outTextF(can, parm, fonts[FONT_NORMAL], 320, 240+140, 140, getShortName(the_vcr.objects[1].owner));

    ui.update(getExtent());
}

bool
WVcrPlayWindow::handleEvent(const UIEvent& e, bool second_pass)
{
    if (VcrSpriteVisualizer::speed == 0 ? e.type == ev_Idle : e.type == ev_Tick) {
        if (!second_pass)
            sprite_vis->playTick();
    } else if (e.type == ev_Key) {
        if (e.key.code == SDLK_ESCAPE) {
            stop(0);
            return true;
        } else if (e.key.code == ss_Shift + SDLK_LEFT) {
            if (player->getTime())
                sprite_vis->windTo(player->getTime() - 1, false);
            return true;
        } else if (e.key.code == ss_Shift + SDLK_RIGHT) {
            if (! sprite_vis->isPlaying()) {
                VcrPlayer::time_type t = player->getTime();
                while (!sprite_vis->isDone() && player->getTime() == t)
                    sprite_vis->playTick(true);
            }
            return true;
        } else if (e.key.code == '+') {
            sprite_vis->changeSpeed(+1);
            return true;
        } else if (e.key.code == '-') {
            sprite_vis->changeSpeed(-1);
            return true;
        } else if (e.key.code == vk_Quit) {
            stop(0);
            putKey(vk_Quit);
            return true;
        }
    } else if (e.type == ev_Command) {
        if (e.cmd.code == cm_Backward) {
            sprite_vis->windTo(player->getTime() - 20, false);
            return true;
        } else if (e.cmd.code == cm_GotoStart) {
            sprite_vis->windTo(0, false);
            return true;
        } else if (e.cmd.code == cm_Forward) {
            sprite_vis->windTo(player->getTime() + 20, true);
            return true;
        } else if (e.cmd.code == cm_GotoEnd) {
            sprite_vis->windToEnd();
            return true;
        } else if (e.cmd.code == cm_Play) {
            if (sprite_vis->isPlaying())
                sprite_vis->stopPlaying();
            else
                sprite_vis->startPlaying();
            return true;
        }
    }
    return UIGroup::handleEvent(e, second_pass);
}
