/**
  *  \file client/vcrvis.h
  *  \brief VCR Visualisation Classes/Widgets
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_CLIENT_VCRVIS_H
#define PCC_V2_CLIENT_VCRVIS_H

#include "client/vcranim.h"
#include "cpluslib/smartptr.h"
#include "game/vcr.h"
#include "game/vcrdb.h"
#include "gfx/sprite.h"
#include "ui/ui.h"
#include "ui/uiglobal.h"
#include "sound/sound.h"
#include "cpluslib/holder.h"

/** VCR Player Constants. */
enum {
    /* Animation layers */
    VCRPLAY_LAYER_TEXT    = 0,
    VCRPLAY_LAYER_TORP    = 1,
    VCRPLAY_LAYER_BEAM    = 2,
    VCRPLAY_LAYER_BANG    = 3,
    VCRPLAY_LAYER_FTR     = 4,
    VCRPLAY_LAYER_OBJECTS = 5,
    VCRPLAY_LAYER_SCENE   = 6,

    /* Coordinates */
    VCRPLAY_OBJECT_Y = 110
    // NOTE: fix VcrFighterSprite::getTrackY() when this changes.
    // with new fighter movement pattern, this must be getTrackY(0)
};

class WVcrPlayWindow;

//! Shield display widget.
class WVcrShieldWidget : public UIFixedWidget {
    VcrPlayer* player;
    int side;
 public:
    WVcrShieldWidget(VcrPlayer* aplay, int aside);
    void draw(Ptr<GfxCanvas> can, UpdateInfo& ui);
};

//! Damage/crew level display widget.
class WVcrDamageWidget : public UIFixedWidget {
    VcrPlayer* player;
    int side;
    bool is_planet;
 public:
    WVcrDamageWidget(VcrPlayer* aplay, int aside, bool aplanet);
    void draw(Ptr<GfxCanvas> can, UpdateInfo& ui);
};

//! Beam status widget.
class WVcrBeamDisplay : public UIFixedWidget {
    int count, type, scale;
    int status[VcrPlayer::VCRMF_MAX_BEAMS];
    int pref_size;
    void drawOne(Ptr<GfxCanvas> can, int aindex);
    inline int getBeamY(int index) {
        return extent.y + index * (scale + 1) + 3 + font_heights[FONT_SMALL];
    }
 public:
    WVcrBeamDisplay(int acount, int atype, int scale);
    void draw(Ptr<GfxCanvas> can, UpdateInfo& ui);
    void updateOne(int aindex, int aval);
    int getPreferredHeight() const { return pref_size; }
};

//! Torpedo status display widget.
class WVcrTorpDisplay : public UIFixedWidget {
    int count, type, scale;
    int status[VcrPlayer::VCRMF_MAX_TORPS];
    int pref_size;
    int torp_count;
    void drawOne(Ptr<GfxCanvas> can, int aindex);
    void drawCount(Ptr<GfxCanvas> can);
    inline int getTorpY(int index) {
        return extent.y + index * (scale + 1) + 3 + font_heights[FONT_SMALL];
    }
 public:
    WVcrTorpDisplay(int acount, int atype, int scale);
    void draw(Ptr<GfxCanvas> can, UpdateInfo& ui);
    void updateOne(int aindex, int aval);
    void updateCount(int anewcount);
    int getPreferredHeight() const { return pref_size; }
};

//! Fighter status display widget.
class WVcrFighterDisplay : public UIFixedWidget {
    int bays, fighters;
 public:
    WVcrFighterDisplay(int abays, int aftrs);
    void updateCount(int aftrs);
    void draw(Ptr<GfxCanvas> can, UpdateInfo& ui);
};

//! VCR Visualizer with sprites.
class VcrSpriteVisualizer : public VcrVisualizer {
    friend class WVcrPlayWindow;
    class Status;
    friend class Status;

    Ptr<GfxPixmap> objs[2], ftrs[2][2], torps[2];
    Ptr<VcrFighterSprite> fighters[2][VcrPlayer::VCRMF_MAX_FTRS];
    Ptr<GfxPixmapSprite> objsprite[2];
    GfxPoint beam_mount[2][MAX_BEAM_LIMIT];
    GfxPoint torp_mount[2][MAX_TORP_LIMIT];

    Ptr<WVcrShieldWidget> shieldw[2];     // Shield display widgets
    Ptr<WVcrDamageWidget> damagew[2];     // Damage display widgets
    Ptr<WVcrBeamDisplay> beamw[2];        // Beam status widgets
    Ptr<WVcrTorpDisplay> torpw[2];        // Torpedo status widgets
    Ptr<WVcrFighterDisplay> fighterw[2];  // Fighter status widgets

    SfxSample snd_beam[2], snd_torp[2];
    SfxSample snd_flaunch, snd_fbeam, snd_fexplode;
    SfxSample snd_damaged, snd_destroyed;
    SfxSample snd_torp_hit;

    void findMountpoints(int side);

    TVcr the_vcr;
    WVcrPlayWindow* the_window;
    bool playing, done, disable_fx;
    int counter;

    void updatePanels();
    void createWeaponWidgets(int x, int side);
    void createObjectSprites();
    void saveStatus();
    void restoreStatus(int slot);
    void updateFighter(int side);
    VcrPlayer::time_type getLatestStoredTime() const;

    enum {
        VCRP_ORIG_TICKS = 50,   ///< Starting ticks_per_status value.
        VCRP_MAX_STATUS = 100,  ///< Maximum size of status.
        VCRP_MAX_SPEED  = 20
    };
    Ptr<Status> status[VCRP_MAX_STATUS];
    int ticks_per_status;

 public:
    GfxSpriteManager m;

    VcrSpriteVisualizer(const TVcr& vcr, WVcrPlayWindow* awindow,
                        Ptr<VcrPlayer> player);
    void startFighter(int side, int track);
    void landFighter(int side, int track);
    void killFighter(int side, int track);

    /* hit >= 0: hit, hit < 0: miss */
    void fireBeam(int side, int track, int target, int hit, int damage, int kill);
    void fireTorp(int side, int hit, int launcher);

    void updateBeam(int side, int id);
    void updateTorp(int side, int id);

    void killObject(int side);

    void redraw();
    void init();

    void update();
    void updateAfterWind();

//     Ptr<VcrFighterSprite> getFighterSprite(int side, int i) const
//         { return fighters[side][i]; }
//     void setFighterSprite(int side, int i, Ptr<VcrFighterSprite> fs);

    /* Playback Control */
    void playTick(bool force = false);
    void windTo(VcrPlayer::time_type t, bool restart_play);
    void windToEnd();

    void startPlaying();
    void stopPlaying();

    bool isPlaying() const { return playing; }
    bool isDone() const { return done && !VcrWeaponSprite::active_fx_sprites && m.isLayerEmpty(VCRPLAY_LAYER_BANG); }

    void changeSpeed(int dif);

    static int speed;
};

//! The whole VCR Display screen.
class WVcrPlayWindow : public UIGroup {
    friend class VcrSpriteVisualizer;

    ObjectHolder holder;

    Ptr<VcrPlayer> player;                // The VCR Player.
    Ptr<VcrVisualizer> old_vis;           // Previous visualizer.
    Ptr<VcrSpriteVisualizer> sprite_vis;  // Our visualizer.
    Ptr<VcrEntry> entry;                  // The VCR entry we play.
    TVcr the_vcr;                         // The VCR we play.
    void updateStatus();
 public:
    WVcrPlayWindow(Ptr<VcrPlayer> aply, Ptr<VcrEntry> aen);
    ~WVcrPlayWindow();

    void drawInterior(Ptr<GfxCanvas> can, UpdateInfo& ui);
    bool handleEvent(const UIEvent& e, bool second_pass);
};

#endif
