/**
  *  \file fileio/dirs.cc
  *  \brief Directory Handling
  *
  *  Directory reading and pathname arithmetic. Most of the functions
  *  are platform-dependant and implemented in arch/.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#include "arch/platform.h"
#include "cpluslib/assert.h"
#include "cpluslib/smartptr.h"
#include "fileio/dirs.h"
#include "util/except.h"
#include "util/nls.h"

/**
  *  \class DirectoryEntry
  *  \brief Abstract Directory Entry
  *
  *  This class provides an interface for directory entries.
  *  It can be opened to obtain the actual file stream or
  *  subdirectory.
  */

/** Constructor. Construct a directory entry for the specified
    name. The name should be the complete name of the file. */
DirectoryEntry::DirectoryEntry(string_t aname)
    : name(aname), statted(false)
{ }

DirectoryEntry::~DirectoryEntry()
{ }

/** Get size of file. Returns the size of the file in bytes,
    or -1 when an error occurs. Results are cached, that is,
    when you query the directory item the first time, the
    disk is actually read; later on the cached result is used. */
DirectoryEntry::fpos_t
DirectoryEntry::getSize()
{
    if (!statted)
        getInfo();
    if (statted)
        return file_size;
    else
        return -1;
}

/** Get file type. Returns the type of the file:
    - FT_FILE for regular files
    - FT_DIRECTORY for subdirectories
    - FT_UNKNOWN when the type can't be determined
    or some other implementation defined value. Results are cached
    like for getSize(). */
DirectoryEntry::FileType
DirectoryEntry::getFileType()
{
    if (!statted)
        getInfo();
    if (statted)
        return file_type;
    else
        return FT_UNKNOWN;
}

/** Set type/size information. Your implementation of getInfo()
    should call this method, filling in the appropriate values
    for size and type. */
void
DirectoryEntry::setInfo(fpos_t asize, FileType atype)
{
    file_size = asize;
    file_type = atype;
    statted   = true;
}

/** Get basename. Should return the file name with all path information
    removed. \default does standard pathname arithmetic. You should only
    override it if you can get the same information much more efficient. */
string_t
DirectoryEntry::getBaseName() const
{
    return Directory::getFileName(getName());
}

/************************** DummyDirectoryEntry **************************/

/**
  *  \class DummyDirectoryEntry
  *  \brief Dummy Entry
  *
  *  This is a directory entry usable for `dead ends'. It
  *  can not be opened in any way.
  */

/** Constructor.
    \param aname Name of item
    \param aparent Parent directory. Must not be 0. */
DummyDirectoryEntry::DummyDirectoryEntry(string_t aname,
                                         Ptr<Directory> aparent)
    : DirectoryEntry(aname), parent(aparent)
{
    ASSERT(aparent.ptr());
}

Ptr<Directory>
DummyDirectoryEntry::openDir()
{
    throw FileOpenException(_("No such directory."));
}

Ptr<Stream>
DummyDirectoryEntry::openFile(int)
{
    throw FileOpenException(_("No such file."));
}

Ptr<Directory>
DummyDirectoryEntry::openParent()
{
    return parent;
}

void
DummyDirectoryEntry::getInfo()
{ }

/******************************* Directory *******************************/

/**
  *  \class Directory
  *  \brief Abstract Directory
  *
  *  This class provides an abstract interface for directories.
  *  Directories can be queried for directory items (i.e. the
  *  files they contain); these are presented as instances of
  *  classes derived from DirectoryEntry.
  *
  *  In addition, there are functions for OS-dependant file name
  *  `arithmetic'.
  */

Directory::~Directory()
{ }

/** Open subdirectory. This is just a convenience function for
    DirectoryEntry::openDir(). */
Ptr<Directory>
Directory::openSubdir(string_t name)
{
    return getEntryByName(name)->openDir();
}

/** Open file. This is just a convenience function for
    DirectoryEntry::openFile(int). */
Ptr<Stream>
Directory::openFile(string_t name, int mode)
{
    return getEntryByName(name)->openFile(mode);
}

/** Open file, don't throw. When the file can't be opened, returns 0
    instead of throwing. \see openFile */
Ptr<Stream>
Directory::openFileNT(string_t name, int mode)
{
    try {
        return getEntryByName(name)->openFile(mode);
    }
    catch(...) {
        return 0;
    }
}

/** Open file, rethrow exception. FileOpenExceptions do not include the
    file name. This function rethrows the exception with the file name
    inserted when open fails. \see openFile */
Ptr<Stream>
Directory::openFileT(string_t name, int mode)
{
    try {
        return getEntryByName(name)->openFile(mode);
    }
    catch(FileOpenException& e) {
        throw FileOpenException(name + ": " + e.what());
    }
}
