/**
  *  \file fileio/dirs.h
  *  \brief Directory Handling
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_FILEIO_DIRS_H
#define PCC_V2_FILEIO_DIRS_H

#include <string>
#include <sys/types.h>
#include <dirent.h>
#include "fileio/streams.h"
#include "arch/platform.h"
#include "cpluslib/smartptr.h"
#include "cpluslib/string.h"
#include "cpluslib/misc.h"
#include "cpluslib/enum.h"

class Directory;

class DirectoryEntry : public Object {
 public:
    typedef Stream::fpos_t fpos_t;
    typedef char FileType;
    enum {
        FT_UNKNOWN    = '?',    ///< File type can't be detected
        FT_FILE       = '-',    ///< Regular file
        FT_DIRECTORY  = 'd'     ///< Directory
    };
 protected:
    string_t name;
    virtual void getInfo() = 0;
    void setInfo(fpos_t asize, FileType atype);
    DirectoryEntry(string_t aname);
 public:
    virtual ~DirectoryEntry();
    virtual Ptr<Directory> openDir() = 0;
    virtual Ptr<Stream>    openFile(int mode) = 0;
    virtual Ptr<Directory> openParent() = 0;

    string_t getName() const { return name; }
    virtual  string_t getBaseName() const;
    FileType getFileType();
    fpos_t   getSize();

 private:
    bool statted;               ///< true iff file_size/file_type are set
    fpos_t file_size;           ///< size of file
    FileType file_type;         ///< type of file (FT_xxx)
};

class DummyDirectoryEntry : public DirectoryEntry {
    Ptr<Directory> parent;
 protected:
    void getInfo();
 public:
    DummyDirectoryEntry(string_t aname, Ptr<Directory> aparent);
    Ptr<Directory> openDir();
    Ptr<Stream> openFile(int mode);
    Ptr<Directory> openParent();
};

class Directory : public Object {
 public:
    virtual ~Directory();
    virtual Ptr<DirectoryEntry> getEntryByName(string_t name) = 0;
    virtual Ptr<Enumeration<Ptr<DirectoryEntry> > > getEntries() = 0;
    virtual Ptr<Directory> getParent() = 0;
    virtual string_t getName() = 0;

    /* Convenience Functions */
    Ptr<Directory> openSubdir(string_t name);
    Ptr<Stream>    openFile(string_t name, int mode);
    Ptr<Stream>    openFileNT(string_t name, int mode);
    Ptr<Stream>    openFileT(string_t name, int mode);

    /* Static functions, platform specific (implemented in arch/) */
    static bool     isAbsoluteName(string_t name);
    static bool     isSeparatorChar(char c);
    static string_t makeFileName(string_t dir, string_t name);
    static string_t getCanonicalName(string_t dir);
    static string_t getAbsoluteName(string_t dir);
    static string_t getFileName(string_t dir);
    static string_t getDirName(string_t dir);
    static string_t getWorkingDirectory();
    static Ptr<Directory> openDir(string_t dir);
    static Ptr<Enumeration<Ptr<Directory> > > getRoots();

    static const char PATH_SEPARATOR = Platform::PATH_SEPARATOR;
};

#endif
