/**
  *  \file io.cc
  *  \brief Binary File Parsing
  *
  *  Routines to read and write DOS-format structures. This provides
  *  the actual marshalling primitives, as well as routines to read
  *  and write structures using a definition table.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "fileio/io.h"
#include "util/recode.h"

/** Compute structure size.
    Returns the packed (DOS) size of the structure defined by parser.
    If parser is FooType::parser, this should equal FooType::size. */
std::size_t
getStructureSize(const TStructureParser* parser)
{
    std::size_t size = 0;
    while(parser->type != T_END) {
        switch(parser->type) {
         case T_INT16:
         case T_UINT16:
         case T_BOOL:
         case T_TRISTATE:
            size += 2 * parser->amount;
            break;
         case T_INT32:
         case T_UINT32:
            size += 4 * parser->amount;
            break;
         case T_STRING:
         case T_RSTRING:
            size += parser->amount;
            break;
         case T_TYPEREF:
            size += parser->amount * getStructureSize(parser->ref);
            break;
         case T_END:;
        }
        ++parser;
    }
    return size;
}

/** Parse structure.
   
    Unpacks a DOS structure into a C++ structure. Returns a pointer
    to just after the parsed DOS structure.
  
    \param src    pointer to DOS structure
    \param data   pointer to C++ structure
    \param parser conversion information
    
    \pre  src points to at least getStructureSize(parser) bytes
    \post getStructure(src, data, parser) == src + getStructureSize(parser) */
const char*
getStructure(const char* src, void* data,
             const TStructureParser* parser)
{
    while(parser->type != T_END) {
        char* target = static_cast<char*>(data) + parser->offset;
        switch(parser->type) {
         case T_INT16:
            for (int i = 0; i < parser->amount; ++i) {
                ((int16*) target)[i] = getInt16(src);
                src += 2;
            }
            break;
         case T_UINT16:
            for (int i = 0; i < parser->amount; ++i) {
                ((uint16*) target)[i] = getUint16(src);
                src += 2;
            }
            break;
         case T_BOOL:
            for (int i = 0; i < parser->amount; ++i) {
                ((bool*) target)[i] = getInt16(src);
                src += 2;
            }
            break;
         case T_TRISTATE:
            for (int i = 0; i < parser->amount; ++i) {
                ((TTristate*) target)[i] = TTristate(getInt16(src));
                src += 2;
            }
            break;
         case T_INT32:
            for (int i = 0; i < parser->amount; ++i) {
                ((int32*) target)[i] = getInt32(src);
                src += 4;
            }
            break;
         case T_UINT32:
            for (int i = 0; i < parser->amount; ++i) {
                ((uint32*) target)[i] = getUint32(src);
                src += 4;
            }
            break;
         case T_STRING:
         case T_RSTRING:
            std::copy(src, src + parser->amount, target);
            if (parser->type == T_RSTRING)
                Recoder::getE2I().recodeArray(target, parser->amount);
            src += parser->amount;
            break;
         case T_TYPEREF:
            for(int i = 0; i < parser->amount; ++i) {
                src = getStructure(src, target, parser->ref);
                target += parser->size;
            }
            break;
         case T_END:;
        }
        ++parser;
    }
    return src;
}

namespace {
    /** Used for implementation of getPartialStructure. \internal */
    class StructureReader {
        const char* src;
        char*       data;
        std::size_t offset;
        std::size_t size;
        bool        advance(std::size_t size);
     public:
        StructureReader(const char* src, void* data, std::size_t offset, std::size_t size)
            : src(src), data(static_cast<char*>(data)), offset(offset), size(size) { }
        void parse(const TStructureParser* parser, std::size_t add);
        const char* getSourcePtr() const { return src; }
    };
}

/** Advance input by /now/ bytes.
    \returns true if caller should convert the appropriate item and advance
    input pointer (src).
    \returns false if caller should not convert the item. In this case,
    this function will advance the input pointer itself (needed for
    partial fields) */
bool
StructureReader::advance(std::size_t now)
{
    if (offset) {
        /* we're not yet reading */
        if (offset > now) {
            offset -= now;
        } else {
            /* partial read, maybe */
            now -= offset;
            offset = 0;
            if (now > size)
                size = 0;       // we have less than one field (!)
            else
                size -= now;    // skip partial field
            src += now;
        }
        return false;
    } else if (size >= now) {
        size -= now;
        return true;
    } else {
        src += size;
        size = 0;
        return false;
    }
}

void
StructureReader::parse(const TStructureParser* parser, std::size_t add)
{
    while(parser->type != T_END) {
        char* target = data + parser->offset + add;
        switch(parser->type) {
         case T_INT16:
            for (int i = 0; i < parser->amount; ++i) {
                if (advance(2)) {
                    ((int16*) target)[i] = getInt16(src);
                    src += 2;
                }
            }
            break;
         case T_UINT16:
            for (int i = 0; i < parser->amount; ++i) {
                if (advance(2)) {
                    ((uint16*) target)[i] = getUint16(src);
                    src += 2;
                }
            }
            break;
         case T_BOOL:
            for (int i = 0; i < parser->amount; ++i) {
                if (advance(2)) {
                    ((bool*) target)[i] = getInt16(src);
                    src += 2;
                }
            }
            break;
         case T_TRISTATE:
            for (int i = 0; i < parser->amount; ++i) {
                if (advance(2))
                    ((TTristate*) target)[i] = TTristate(getInt16(src));
                src += 2;
            }
            break;
         case T_INT32:
            for (int i = 0; i < parser->amount; ++i) {
                if (advance(4)) {
                    ((int32*) target)[i] = getInt32(src);
                    src += 4;
                }
            }
            break;
         case T_UINT32:
            for (int i = 0; i < parser->amount; ++i) {
                if (advance(4)) {
                    ((uint32*) target)[i] = getUint32(src);
                    src += 4;
                }
            }
            break;
         case T_STRING:
         case T_RSTRING:
            if (advance(parser->amount)) {
                std::copy(src, src + parser->amount, target);
                if (parser->type == T_RSTRING)
                    Recoder::getE2I().recodeArray(target, parser->amount);
                src += parser->amount;
            }
            break;
         case T_TYPEREF:
            for(int i = 0; i < parser->amount; ++i) {
                parse(parser->ref, add + parser->offset + i*parser->size);
                target += parser->size;
            }
            break;
         case T_END:;
        }
        ++parser;
    }
}

/** Get partial structure.
    \param src     input data (partial structure in DOS format)
    \param data    pointer to C++ structure
    \param parser  conversion information
    \param offset  relative position of input data in DOS structure
    \param size    number of available bytes in input data

    For example, when /src/ points at eight bytes (hence /size/ = 8)
    and /offset/ is 4, this will determine what fields of the
    structure are at offset 4 in DOS and overwrite them. When the structure
    is defined as
    \verbatim
      struct Foo {
        int16 a, b;
        int16 c, d;
        int32 x;
      };
    \endverbatim
    this will overwrite items c, d and x.

    Partial fields will not be touched. If, in the above example, /size/ is 7,
    x will not be changed.

    \returns pointer to after consumed input data.
    Iff getStructureSize(parser) >= offset + size, then
    return value = src + size */
const char*
getPartialStructure(const char* src, void* data,
                    const TStructureParser* parser,
                    std::size_t offset, std::size_t size)
{
    StructureReader r(src, data, offset, size);
    r.parse(parser, 0);
    return r.getSourcePtr();
}

/** Store Structure.

    Reads the C++ structure and packs it into the DOS structure.

    \param dest    pointer to structure in DOS format
    \param data    pointer to C++ structure
    \param parser  conversion information

    \return pointer behind DOS structure
    \post storeStructure(dest, data, parser) == dest + getStructureSize(parser) */
char* storeStructure(char* dest, const void* data,
                     const TStructureParser* parser)
{
    while(parser->type != T_END) {
        const char* source = static_cast<const char*>(data) + parser->offset;
        switch(parser->type) {
         case T_INT16:
            for (int i = 0; i < parser->amount; ++i) {
                storeInt16(dest, ((const int16*) source)[i]);
                dest += 2;
            }
            break;
         case T_UINT16:
            for (int i = 0; i < parser->amount; ++i) {
                storeUint16(dest, ((const uint16*) source)[i]);
                dest += 2;
            }
            break;
         case T_BOOL:
            for (int i = 0; i < parser->amount; ++i) {
                storeInt16(dest, ((const bool*) source)[i]);
                dest += 2;
            }
            break;
         case T_TRISTATE:
            for (int i = 0; i < parser->amount; ++i) {
                storeInt16(dest, ((const TTristate*) source)[i]);
                dest += 2;
            }
            break;
         case T_INT32:
            for (int i = 0; i < parser->amount; ++i) {
                storeInt32(dest, ((const int32*) source)[i]);
                dest += 4;
            }
            break;
         case T_UINT32:
            for (int i = 0; i < parser->amount; ++i) {
                storeUint32(dest, ((const uint32*) source)[i]);
                dest += 4;
            }
            break;
         case T_STRING:
         case T_RSTRING:
            std::copy(source, source + parser->amount, dest);
            if (parser->type == T_RSTRING)
                Recoder::getI2E().recodeArray(dest, parser->amount);
            dest += parser->amount;
            break;
         case T_TYPEREF:
            for(int i = 0; i < parser->amount; ++i) {
                dest = storeStructure(dest, source, parser->ref);
                source += parser->size;
            }
            break;
         case T_END:;
        }
        ++parser;
    }
    return dest;
}

/** Append DOS structure to a buffer. Do not call this function directly,
    use the storeStructure(Buffer<char>&, T) template instead.
    \param buf      buffer
    \param data     data (object of type T)
    \param size     size of T in DOS
    \param parser   conversion information */
void
storeAppendStructure(Buffer<char>& buf, const void* data,
                     std::size_t size, const TStructureParser* parser)
{
    std::size_t bufpos = buf.getSize();
    buf.ensureSize(bufpos + size);
    storeStructure(buf.ptr() + bufpos, data, parser);
}

/** Parse BASIC string. Parses a string of maximum length n into a
    standard C++ string. Note: you should use getBasicString if you
    can, because that gets the size right automatically. */
string_t
getBasicStringN(const char* name, std::size_t n)
{
    const char* end = name;
    while(n && *end)
        ++end, --n;
    while(end > name && end[-1] == ' ')
        --end;
    return string_t(name, end-name);
}

/** Store BASIC string. Stores string s in BASIC format, using
    n bytes. Note: you should use storeBasicString if you can,
    because that gets the size right automatically. */
void
storeBasicStringN(char* name, std::size_t n, const string_t s)
{
    string_t::size_type rlen = s.copy(name, n);
    if(rlen < n)
        std::fill_n(name + rlen, n - rlen, ' ');
}
