/**
  *  \file io.h
  *  \brief Binary File Parsing
  *
  *  Functions to parse binary files (in particular, those written by
  *  certain BASIC programs). 
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_FILEIO_IO_H
#define PCC_V2_FILEIO_IO_H

#include <string>
#include <cstddef>
#include "cpluslib/types.h"
#include "cpluslib/buffer.h"
#include "fileio/streams.h"

enum TStructureMember {
    T_END,                      ///< end of definition
    T_BOOL,                     ///< boolean type (word)
    T_INT16,                    ///< signed 16 bit ("WORD")
    T_INT32,                    ///< signed 32 bit ("DWORD")
    T_UINT16,                   ///< unsigned 16 bit
    T_UINT32,                   ///< unsigned 32 bit
    T_STRING,                   ///< string
    T_RSTRING,                  ///< recoded string
    T_TRISTATE,                 ///< Tristate. Treated as 16-bit word.
    T_TYPEREF                   ///< member structure
};

/** Tristate options. */
enum TTristate {
    tri_No,
    tri_Yes,
    tri_Ally
};

struct TStructureParser {
    TStructureMember type;       ///< member type
    short int amount;            ///< number of members, if array
    std::size_t offset;          ///< offset in C++ structure
    std::size_t size;            ///< T_TYPEREF: size of member type (C++)
    const TStructureParser* ref; ///< T_TYPEREF: definition of member type
};

std::size_t getStructureSize(const TStructureParser* parser);
const char* getStructure(const char* src, void* data,
                         const TStructureParser* parser);
const char* getPartialStructure(const char* src, void* data,
                                const TStructureParser* parser,
                                std::size_t offset, std::size_t size);
char* storeStructure(char* dest, const void* data,
                     const TStructureParser* parser);
void storeAppendStructure(Buffer<char>& buf, const void* data,
                          std::size_t size, const TStructureParser* parser);

template<class T>
const char*
getStructure(const char* src, T& data) {
    return getStructure(src, &data, T::parser);
}

template<class T>
char*
storeStructure(char* dest, const T& data) {
    return storeStructure(dest, &data, T::parser);
}

template<class T>
void
storeAppendStructure(Buffer<char>& buf, const T& data)
{
    storeAppendStructure(buf, &data, T::size, T::parser);
}

template<class T>
inline bool
storeStructure(Stream* s, const T& data)
{
    char buffer[T::size];
    storeStructure(buffer, &data, T::parser);
    return s->write(buffer, T::size) == T::size;
}

template<class T>
inline bool
getStructure(Stream* s, T& data)
{
    char buffer[T::size];
    if (s->read(buffer, T::size) != T::size)
        return false;
    getStructure(buffer, &data, T::parser);
    return true;
}

/******************** Word I/O ********************/

/** Parse unsigned 16-bit quantity.
    Parses two bytes, returns unsigned value (0..64k) */
template<class T>
inline uint16
getUint16(const T& src, int dif=0)
{
    return (unsigned char) src[dif]
        + 0x100U * (unsigned char) src[dif+1];
}

/** Parse unsigned 32-bit quantity.
    Parses four bytes, returns unsigned value (0..4G) */
template<class T>
inline uint32
getUint32(const T& src, int dif=0)
{
    return (unsigned char) src[dif]
        + 0x100U * (unsigned char) src[dif+1]
        + 0x10000U * (unsigned char) src[dif+2]
        + 0x1000000U * (unsigned char) src[dif+3];
}

/** Parse signed 16-bit quantity.
    Parses two bytes, returns signed ("WORD") value (-32k..+32k) */
template<class T>
inline int16
getInt16(const T& src, int dif=0)
{
    /* ultra-portable way to parse two's complement -- gcc optimizes
       this away if possible. This may not even be necessary according
       to 4.7p2 [conv.integral]. */
    uint16 u = getUint16(src, dif);
    if(u & 0x8000U)
        return -1-~u;
    else
        return u;
}

/** Parse unsigned 32-bit quantity.
    Parses four bytes, returns signed ("DWORD") value (-2G..+2G) */
template<class T>
inline int32
getInt32(const T& src, int dif=0)
{
    uint32 u = getUint32(src, dif);
    if(u & 0x80000000U)
        return -1-~u;
    else
        return u;
}

inline void
storeUint16(char* dest, uint16 value)
{
    unsigned char* d = (unsigned char*) dest;
    *d++ = (value & 255);
    *d   = (value >> 8) & 255;
}

inline void
storeUint32(char* dest, uint32 value)
{
    unsigned char* d = (unsigned char*) dest;
    *d++ = (value & 255);
    *d++ = (value >>= 8) & 255;
    *d++ = (value >>= 8) & 255;
    *d   = (value >> 8) & 255;
}

inline void
storeInt16(char* dest, int16 value)
{
    if(value < 0)
        storeUint16(dest, ~(-1-value));
    else
        storeUint16(dest, value);
}

inline void
storeInt32(char* dest, int32 value)
{
    if(value < 0)
        storeUint32(dest, ~(-1-value));
    else
        storeUint32(dest, value);
}

inline void
storeInt16(char* dest, const int16* src, int count)
{
    while (--count >= 0) {
        storeInt16(dest, *src++);
        dest += 2;
    }
}

inline void
storeUint16(char* dest, const uint16* src, int count)
{
    while (--count >= 0) {
        storeUint16(dest, *src++);
        dest += 2;
    }
}

inline void
storeInt32(char* dest, const int32* src, int count)
{
    while (--count >= 0) {
        storeInt32(dest, *src++);
        dest += 4;
    }
}

inline void
storeUint32(char* dest, const uint32* src, int count)
{
    while (--count >= 0) {
        storeUint32(dest, *src++);
        dest += 4;
    }
}

inline void
getInt16(const char* src, int16* dest, int count)
{
    while (--count >= 0) {
        *dest++ = getInt16(src);
        src += 2;
    }
}

inline void
getUint16(const char* src, uint16* dest, int count)
{
    while (--count >= 0) {
        *dest++ = getUint16(src);
        src += 2;
    }
}

inline void
getInt32(const char* src, int32* dest, int count)
{
    while (--count >= 0) {
        *dest++ = getInt32(src);
        src += 4;
    }
}

inline void
getUint32(const char* src, uint32* dest, int count)
{
    while (--count >= 0) {
        *dest++ = getUint32(src);
        src += 4;
    }
}

string_t getBasicStringN(const char* name, std::size_t n);
void storeBasicStringN(char* name, std::size_t n, const string_t s);

#ifdef __BORLANDC__
/* Work around a compiler bug */
#define getBasicString(str) (getBasicStringN((str), sizeof (str)))
#define storeBasicString(str,v) (storeBasicStringN((str), sizeof (str), (v)))
#else
template<int N>
string_t
getBasicString(const char (&name)[N]) {
    return getBasicStringN(name, N);
}

/* For some reason, g++ can't derive this one automatically from the
   one above. */
template<int N>
string_t
getBasicString(char (&name)[N]) {
    return getBasicStringN(name, N);
}

template<int N>
void
storeBasicString(char (&name)[N], const string_t s) {
    storeBasicStringN(name, N, s);
}
#endif

#endif
