/**
  *  \file fileio/iostr.cc
  *
  *  #Stream / iostream bridge.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "fileio/iostr.h"
#include "util/compat.h"
#include "cpluslib/assert.h"

// BEGIN HACK
// This is so that we can refer to traits::XX, even though the GNU
// libio implementation does not have them
// struct traits {
//     static int eof() { return EOF; }
//     static int to_int_type(char c) { return (unsigned char) c; }
// };
//typedef ios ios_base;
// END HACK

/** Streambuf for our streams. This provides a bridge between our
    stream implementation and standard iostreams.

    It implements reading, writing and seeking, so that all
    #Stream operations are supported.

    The following functions are not implemented:
    - xsgetn (an optimisation for reading)
    - showmanyc
    - uflow (used for unbuffered I/O)
    - pbackfail (anyone needs this?) */
StandardStreambuf::StandardStreambuf()
    : stream(0)
{
    create_buffer();
}

StandardStreambuf::StandardStreambuf(Ptr<Stream> astream)
    : stream(astream)
{
    create_buffer();
}

StandardStreambuf::~StandardStreambuf()
{
    delete[] buffer;
}

void
StandardStreambuf::create_buffer()
{
    buf_len = 1024;
    buffer = new char[buf_len];
}

bool
StandardStreambuf::is_open()
{
    return stream.ptr();
}

StandardStreambuf*
StandardStreambuf::open(Ptr<Stream> astream)
{
    close();
    stream = astream;
    if (stream.ptr())
        return this;
    else
        return 0;
}

StandardStreambuf*
StandardStreambuf::close()
{
    sync();
    stream = 0;
    return this;
}

int
StandardStreambuf::underflow()
{
    /* warranties according to CD2: 27.5.2.4.3 */
    ASSERT(!gptr() || gptr() >= egptr());

    if (pptr())                 // make put buffer empty
        overflow(EOF);
    setp(0, 0);

    if (!stream.ptr() || !stream->canDo(Stream::C_READ))
        return STREAM_CHAR_TRAITS::eof();

    std::size_t count = stream->read(buffer, buf_len);
    if (count <= 0) {
        setg(0, 0, 0);          // error -- set empty buffer
        return STREAM_CHAR_TRAITS::eof();
    } else {
        setg(buffer, buffer, buffer + count);
        return STREAM_CHAR_TRAITS::to_int_type(*buffer);
    }
}

int
StandardStreambuf::overflow(int c)
{
    if (!stream.ptr() || !stream->canDo(Stream::C_WRITE))
        return STREAM_CHAR_TRAITS::eof();

    setg(0, 0, 0);
    if (!pptr())
        setp(buffer, buffer);

    std::size_t w = pptr() - pbase();
    if (c != STREAM_CHAR_TRAITS::eof()) {
        *pptr() = c;
        ++w;
    }

    if (stream->write(pbase(), w)) {
        setp(buffer, buffer + buf_len - 1);  // one less to allow stuffing one additional character on next call
        return 0;
    }
    setp(0, 0);
    return STREAM_CHAR_TRAITS::eof();    
}

int
StandardStreambuf::sync()
{
    if (pptr() && pptr() > pbase())
        return overflow(STREAM_CHAR_TRAITS::eof());
    
    return 0;
}

StandardStreambuf::pos_type
StandardStreambuf::seekoff(off_type off, ios_base::seekdir way,
                           ios_base::openmode which)
{
    if (!stream.ptr())
        return -1;              // FIXME: "invalid stream position"
    
    if (which) {
        /* Note that this function is only provided for completeness.
           Especially the ios_base::end case is hopelessly inefficient
           if the underlying stream is a file (it translates into no
           less then four(!) lseek system calls). */
        switch(way) {
         case ios_base::beg:
            /* easy case */
            stream->seek(off);
            break;
         case ios_base::end:
            /* must flush output to get the size right */
            sync();
            stream->seek(stream->getSize() + off);
            break;
         case ios_base::cur:
            /* complicated case. When we're reading, the pointer
               got too far, and we must rewind accordingly. */
            if (gptr())
                off -= egptr() - gptr();
            else if (pptr())
                sync();
            stream->seek(stream->getPos() + off);
            break;
        }
    }
    
    return stream->getPos();    
}

StandardStreambuf::pos_type
StandardStreambuf::seekpos(pos_type sp, ios_base::openmode which)
{
    return seekoff(sp, ios_base::beg, which);
}

/** \class StandardStream
    \brief Standard I/O Stream Wrapper

    This class wraps a #Stream for use with standard I/O, mainly
    text I/O.

    It has the following constraints:
    - I/O is always buffered. Function calls to change the buffering
      method have no effect.
    - if you switch between reading and writing, you must have
      a seek operation inbetween (like in C stdio).

    This design was carefully chosen for clarity and ease of use and
    implementation, although it might seem a little is inefficient.    
    - making a #Stream implement the std::streambuf interface sounds
      tempting, but has the problem that most streams are managed
      via a Ptr<>; this does not go well together with iostreams.
      Also, this solution would not be that different from the current
      one; this one has just one indirection more (and the other one
      would probably have some tests more to avoid having the
      actual stream buffer around all the time);
    - making a #Stream a iostream would mean carrying around a
      streambuf all the time, and having a very huge interface.
      In addition, specifying the interaction between the iostream
      part and the #Stream part would be hard. */

/** Create blank stream. The stream must be opened with
    open(Ptr<Stream>) until you can do something useful with it. */
StandardStream::StandardStream()
    : std::iostream(0)
{
    init (&buf);
}

/** Create open stream. The StandardStream is opened for whatever
    operation the Stream was opened. */
StandardStream::StandardStream(Ptr<Stream> astream)
    : std::iostream(0), buf()
{
    init (&buf);
    buf.open(astream);
}

StandardStream::~StandardStream()
{
    if (buf.is_open())
        buf.close();
}

/** Open specified stream. Opens the iostream for whatever
    operation the Stream was opened. If this StandardStream was
    already open, it is closed before. */
void
StandardStream::open(Ptr<Stream> astream)
{
    clear();
    if (!buf.open(astream));
//        clear(rdstate() & ios::badbit);
//        setstate(ios::failbit);
}

/** Close stream. This will also release the reference to the
    original Stream. */
void
StandardStream::close()
{
    if (buf.is_open()) {
        if (buf.close())
            clear();
        else
            setstate(std::ios::failbit);
    }
}
