/**
  *  \file fileio/multidir.cc
  *  \brief Multiple directories in one ("bind mount" :-)
  *
  *  A class for accessing multiple directories (from a search path)
  *  as if it were a single directory.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <set>
#include "fileio/multidir.h"

/**
  *  \class MultiDirectory
  *  \brief Directory Entry Multiplexer
  *
  *  This provides a directory that behaves as the ``union'' of
  *  multiple other directories. Files are looked up in all directories,
  *  taking the first matching one. This can be used to implement
  *  search paths. You can not create new entries in a MultiDirectory.
  */
class MultiDirectory::Enum : public Enumeration<Ptr<DirectoryEntry> > {
    typedef std::set<string_t> entry_set;

    Ptr<MultiDirectory> dir;
    Ptr<DirectoryEntry> cur_entry;
    Ptr<Enumeration<Ptr<DirectoryEntry> > > cur_iter;
    
    entry_set entries;
    dir_vec::size_type index;

    void next();
 public:
    Enum(Ptr<MultiDirectory> adir);
    bool hasMoreElements();
    Ptr<DirectoryEntry> getNextElement();
};

MultiDirectory::Enum::Enum(Ptr<MultiDirectory> adir)
    : dir(adir), index(0)
{
    next();
}

bool
MultiDirectory::Enum::hasMoreElements()
{
    return cur_entry.ptr();
}

Ptr<DirectoryEntry>
MultiDirectory::Enum::getNextElement()
{
    Ptr<DirectoryEntry> ret = cur_entry;
    next();
    return ret;
}

void
MultiDirectory::Enum::next()
{
    while (1) {
        while (index == 0 || !cur_iter->hasMoreElements()) {
            if (index >= dir->dirs.size()) {
                cur_entry = 0;
                return;
            }
            cur_iter = dir->dirs[index++]->getEntries();
        }
        cur_entry = cur_iter->getNextElement();

        string_t basename = cur_entry->getBaseName();
        if (entries.find(basename) == entries.end()) {
            entries.insert(basename);
            return;
        }
    }
}

/** Create empty MultiDirectory. All requests will yield a dummy entry
    until you add new directories. */
MultiDirectory::MultiDirectory()
{ }

/** Create MultiDirectory. All requests will go to the specified
    directory until you add more. */
MultiDirectory::MultiDirectory(Ptr<Directory> dir)
{
    addDirectory(dir);
}

/** Add directory. Adds another directory to the end of the search
    path. It will be searched last for new entries. */
void
MultiDirectory::addDirectory(Ptr<Directory> dir)
{
    dirs.push_back(dir);
}

Ptr<DirectoryEntry>
MultiDirectory::getEntryByName(string_t name)
{
    for(dir_vec::iterator i = dirs.begin(); i != dirs.end(); ++i) {
        Ptr<DirectoryEntry> e = (*i)->getEntryByName(name);
        if (e->getFileType() != DirectoryEntry::FT_UNKNOWN)
            return e;
    }
    return new DummyDirectoryEntry(name, this);
}

Ptr<Enumeration<Ptr<DirectoryEntry> > >
MultiDirectory::getEntries()
{
    return new Enum(this);
}

Ptr<Directory>
MultiDirectory::getParent()
{
    return 0;
}

string_t
MultiDirectory::getName()
{
    string_t sum;
    for (dir_vec::iterator i = dirs.begin(); i != dirs.end(); ++i) {
        if (sum.length())
            sum += " + ";
        sum += (*i)->getName();
    }
    return sum;
}
