/**
  *  \file fileio/rle.cc
  *  \brief Run-Length Encoded Data
  *
  *  We encode bitmap data and other stuff using a simple RLE variant.
  *
  *  Each file has the following format:
  *  - one dword total size (uncompressed)
  *  - sequence of compressed chunks
  *  - zero-length chunk (one word of value zero)
  *
  *  Each chunk has the following format:
  *  - word with chunk size (uncompressed)
  *  - byte with prefix code for this chunk (chosen dynamically
  *    for each chunk)
  *  - compressed data. Either a byte to be copied verbatim, or
  *    a (prefix,counter,value) triples
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "fileio/rle.h"
#include "fileio/io.h"
#include "util/except.h"
#include "util/nls.h"

/** \class RleReaderStream
    \brief Reading RLE data

    Supports only reading. Input is read into memory completely. */

/** Read RLE data from stream. */
RleReaderStream::RleReaderStream(Ptr<Stream> base)
    : name(base->getName())
{
    base->readBuffer(buffer);

    if(buffer.getSize() < 4)
        throw FileIOException(_("Damaged compression"));
    
    size = getUint32(buffer, 0);
    counter = chr = code = 0;
    index = 4;
    chunk_len = 0;
    pos = 0;
    eof_reached = false;
}

RleReaderStream::~RleReaderStream()
{ }

/** Fetch one byte. Decodes the data on-the fly.
    \returns character read. Sets eof_reached when we reached end-of-file. */
char
RleReaderStream::get()
{
    if (counter) {
        --counter;
        --chunk_len;
        return chr;
    }
    if (!chunk_len) {
        if (index+2 >= buffer.getSize()) {
            eof_reached = true;
            return 0;
        }
        chunk_len = getUint16(buffer, index);
        if (!chunk_len) {
            eof_reached = true;
            return 0;
        }
        code = buffer[index+2];
        index += 3;
    }
    if (index >= buffer.getSize()) {
        throw FileIOException(_("Damaged compression"));
    }

    char c = buffer[index++];
    if (c == code) {
        if (index+1 >= buffer.getSize() || !(counter = buffer[index++]) || counter > chunk_len) {
            throw FileIOException(_("Damaged compression"));
        }
        chr = buffer[index++];
        --counter;
        --chunk_len;
        return chr;
    } else {
        --chunk_len;
        return c;
    }
}

std::size_t
RleReaderStream::read(char* buffer, std::size_t max)
{
    std::size_t i = 0;
    if (!eof_reached) {
        while (i < max && !eof_reached) {
            *buffer++ = get();
            ++pos;
            ++i;
        }
        if (eof_reached) {
            --i;
            --pos;
        }
    }
    return i;
}

std::size_t
RleReaderStream::write(const char*, std::size_t)
{
    return 0;
}

int
RleReaderStream::getCapabilities() const
{
    return C_READ;
}

bool
RleReaderStream::seek(fpos_t)
{
    return false;
}

RleReaderStream::fpos_t
RleReaderStream::getPos()
{
    return pos;
}

RleReaderStream::fpos_t
RleReaderStream::getSize()
{
    return size;
}

string_t
RleReaderStream::getName()
{
    return name;
}
