/**
  *  \file fileio/streams.cc
  *  \brief Stream Implementation
  *
  *  The stream base class, and the simple standard stream classes.
  *  Implementation of the basic stream primitives and the SDL stream
  *  bridge.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cerrno>
#include "fileio/streams.h"
#include "util/except.h"

#ifdef CONF_FILEIO_SDL_INTERFACE
extern "C" {
#include <SDL_error.h>
/** \ingroup SDLInterface
    SDL Interface. These functions provide access to a Stream from SDL. */
//@{
static int
sdl_seek(SDL_RWops* context, int offset, int whence) {
    Stream* s = static_cast<Stream*>(context);
    if (!s->canDo(Stream::C_SEEK)) {
        SDL_SetError("Seek error");
        return -1;
    } else {
        bool b = false;
        switch(whence) {
         case SEEK_SET:
            b = s->seek(offset);
            break;
         case SEEK_CUR:
            b = s->seek(s->getPos() + offset);
            break;
         case SEEK_END:
            b = s->seek(s->getSize() + offset);
            break;
        }
        if (!b) {
            SDL_SetError("Seek error");
            return -1;
        }
        return s->getPos();
    }
}

static int
sdl_read(SDL_RWops* context, void* ptr, int size, int maxnum)
{
    Stream* s = static_cast<Stream*>(context);
    try {
        std::size_t r = s->read(static_cast<char*>(ptr), unsigned(size) * maxnum);
        if (!r) {
            /* according to SDL_rwops.c, it is an error if only
               0 items could be read */
            SDL_SetError("Read error (end of file)");
            return 0;
        }
        return r/size;
    }
    catch(std::exception& e) {
        SDL_SetError(e.what());
        return 0;
    }
}

static int
sdl_write(SDL_RWops* context, const void* ptr, int size, int num)
{
    Stream* s = static_cast<Stream*>(context);
    try {
        std::size_t w = s->write(static_cast<const char*>(ptr), unsigned(size) * num);
        if (!w) {
            /* according to SDL_rwops.c, it is an error if only
               0 items could be written */
            SDL_SetError("Write error (disk full)");
            return 0;
        }
        return w/size;
    }
    catch(std::exception& e) {
        SDL_SetError(e.what());
        return 0;
    }
}

static int
sdl_close(SDL_RWops*)
{
    /* No delete code. Object is deleted automatically. */
    return 0;
}
}
//@}
#endif

/*! \class Stream
    
    This class provides an abstract interface to streams, mostly
    intended for binary I/O.

    When the config define CONF_FILEIO_SDL_INTERFACE is set, every stream
    automatically is an SDL_RWops object and can be used as such. Note
    that SDL can not delete a Stream object; lifetime control is still
    up to you (resp. Ptr<>) even if the SDL code calls `close'. */

Stream::Stream()
{
#ifdef CONF_FILEIO_SDL_INTERFACE
    SDL_RWops::read  = &sdl_read;
    SDL_RWops::write = &sdl_write;
    SDL_RWops::seek  = &sdl_seek;
    SDL_RWops::close = &sdl_close;
#endif
}

/** \fn Stream::read(char*,std::size_t)
    \brief Read data
    
    Should read up to max bytes into the buffer; returns the number of
    bytes read. Returns zero if and only if EOF is reached. Need only
    be supported if capabilities include C_READ.

    \throws FileIOException on error. If appropriate, the stream may
    also go into an error state (where future reads return zero). */

/** \fn Stream::write(const char*,std::size_t)
    \brief Write data

    Should write up to max bytes from the buffer to the stream;
    returns the number of bytes written. Should return less than specified
    in max only if output device is full. Need only be supported if
    capabilities include C_WRITE.

    \throws FileIOException on error. If appropriate, the stream may
    also go into an error state (where future writes return zero). */

/** \fn Stream::seek(fpos_t)
    \brief Set file pointer

    Should position the file pointer at the specified position. The first
    byte in the stream is position 0. Need only be supported if
    capabilities include C_SEEK.

    \returns true if successful, false if not. */

/** \fn Stream::getPos()
    \brief Get file pointer

    Should return the current position of the file pointer. Need only
    be supported if capabilities include C_SEEK.

    \returns negative value on error. */

/** \fn Stream::getSize()
    \brief Get data size

    Should return the total number of bytes that can be read from the
    stream. Need only be supported if capabilities include C_SEEK.

    \returns negative value on error. */

/** \fn Stream::getCapabilities() const
    \brief Get capabilities of this stream

    Should return the capabilities of this stream, as a bitfield of the
    C_xxx constants. */

Stream::~Stream()
{ }

/** Check stream capabilities.
    \return true iff the stream can do all the things in what. */
bool
Stream::canDo(int what)
{
    return (getCapabilities() & what) == what;
}

/** Read stream into buffer. Reads the complete stream (starting at
    the current file position) into the specified buffer. */
void
Stream::readBuffer(Buffer<char>& buffer)
{
    char b[4096];

    while(std::size_t n = read(b, sizeof b))
        buffer.append(b, n);
}

/** Stream factory. \overload
    \returns stream reading file \c name opened with mode \c mode
    \throw FileOpenException when open fails */
Ptr<Stream>
Stream::openFile(const char* name, int mode)
{
    return openFile(string_t(name), mode);
}

/** Stream factory. Like openFile(), but if the open fails, this
    one rethrows the exception with the file name inserted. */
Ptr<Stream>
Stream::openFileT(const string_t& name, int mode)
{
    try {
        return openFile(name, mode);
    }
    catch(FileOpenException& e) {
        throw FileOpenException(name + ": " + e.what());
    }
}

/** Stream factory. Like openFile(), but instead of throwing an
    exception, this one returns 0 if the file can't be opened. */
Ptr<Stream>
Stream::openFileNT(const string_t& name, int mode)
{
    try {
        return openFile(name, mode);
    }
    catch(...) {
        return 0;
    }
}

/***************************** SubStreamBase *****************************/

/*! \class SubStreamBase
    This class provides a "window" into a stream, by restricting certain
    access.

    setRange() sets origin and length:
    - origin >= 0: users can only access positions after that one.
    - origin < 0 disables seeking, so users can only read forward.
    - length >= 0: users can only access this many bytes after origin.
    - length < 0: length is not limited.

    setDisallowedOperations() sets a bitfield of disallowed access modes
    (C_READ to disable reading etc.).

    The SubStreamBase class is intended as a base class for things that
    need this functionality. Users usually instanciate the SubStream class
    instead. */
SubStreamBase::SubStreamBase(Ptr<Stream> awork)
    : origin(-1), length(-1), position(0), worker(awork), op_mask(0)
{ }

/** Set ranges. See class description for explanation. */
void
SubStreamBase::setRange(fpos_t aorigin, fpos_t alength)
{
    origin   = aorigin;
    length   = alength;
    position = 0;
    seek(0);
}

int
SubStreamBase::getCapabilities() const
{
    int op = worker->getCapabilities() & ~op_mask;
    if (origin < 0)
        op &= C_SEEK;
    return op;
}

std::size_t
SubStreamBase::read(char* buffer, std::size_t max)
{
    if (op_mask & C_READ)
        return 0;
    
    if(length >= 0 && max > length - position)
        max = length - position;

    std::size_t r = worker->read(buffer, max);
    position += r;
    return r;
}

std::size_t
SubStreamBase::write(const char* buffer, std::size_t max)
{
    if(op_mask & C_WRITE)
        return 0;
    if(length >= 0 && max > length - position)
        max = length - position;

    std::size_t w = worker->write(buffer, max);
    position += w;
    return w;
}

bool
SubStreamBase::seek(fpos_t pos)
{
    if ((op_mask & C_SEEK) || origin < 0)
        return false;
    if (pos < 0 || (length >= 0 && pos > length))
        return false;
    if (! worker->seek(pos + origin))
        return false;
    position = pos;
    return true;
}

SubStreamBase::fpos_t
SubStreamBase::getPos()
{
    return position;
}

SubStreamBase::fpos_t
SubStreamBase::getSize()
{
    fpos_t o = worker->getSize();
    if (o < 0)
        return length;
    else if (origin >= 0 && origin <= o)
        return std::min(o - origin, length);
    else
        return -1;
}

string_t
SubStreamBase::getName()
{
    return worker ? worker->getName() : string_t("?");
}

/******************************* SubStream *******************************/

/** \class SubStream
    This is the user-callable version of SubStreamBase (see there for
    more information). */

SubStream::SubStream(Ptr<Stream> awork)
    : SubStreamBase(awork)
{ }
