/**
  *  \file game/cargo-tr.cc
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/cargo.h"
#include "cpluslib/assert.h"

/***************************** GCargoTransfer ****************************/

GCargoTransfer::GCargoTransfer(Ptr<GCargoContainer> left, Ptr<GCargoContainer> right)
{
    objs[0] = left;
    objs[1] = right;
}

GCargoTransfer::~GCargoTransfer()
{
    if (!isFinished())
        cancel();
}

/** Move stuff.
    \param type    type to move
    \param amount  amount to move. May be negative; semantics are
                   preserved.
    \param from    source object (0=left, 1=right)
    \param partial true to allow partial transfers, false to reject
    \param s2m     sell supplies on-the-fly
    \return amount moved.
    \inv !partial => return == 0 || return == amount
    \inv partial  => return >= 0 && return <= amount */
int32
GCargoTransfer::move(GCargo::CargoType type, int32 amount, int from, bool partial,
                     bool s2m)
{
    /* trivial case? */
    if (!amount)
        return 0;

    /* reverse transfer? */
    if (amount < 0)
        return -move(type, -amount, 1-from, partial, false);

    /* figure out src/dest objects and type */
    GCargoContainer* src  = objs[from];
    GCargoContainer* dest = objs[1 - from];

    GCargo::CargoType dtype = (type == el_Supplies && s2m
                               ? el_Money
                               : type);

    /* can we actually move it? */
    if (!src->canHave(type) || !dest->canHave(dtype))
        return 0;

    int32 space = dest->getMax(dtype) - dest->getCurrent(dtype);
    if (amount > space) {
        if (!partial)
            return 0;
        amount = space;
    }

    /* do the transfer */
    int32 got = src->removeCargo(type, amount, partial);
    int32 put = dest->addCargo(dtype, got, partial);
    ASSERT(got == put);
    ASSERT(got <= amount);

    return got;
}

void
GCargoTransfer::cancel()
{
    GTransaction::cancel();
    objs[0]->cancel();
    objs[1]->cancel();
}

void
GCargoTransfer::commit()
{
    GTransaction::commit();
    objs[0]->commit();
    objs[1]->commit();
}
