/**
  *  \file game/cargo.h
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_CARGO_H
#define PCC_V2_GAME_CARGO_H

#include "cpluslib/types.h"
#include "cpluslib/smartptr.h"
#include "game/struct.h"
#include "game/transact.h"

/** Cargo type. */
enum GCargoType {
    el_Neutronium,
    el_Tritanium,
    el_Duranium,
    el_Molybdenum,
    el_Fighters,
    el_Colonists,
    el_Supplies,
    el_Money,
    el_Torps,
    el_COUNT = el_Torps + NUM_TORPS,
    el_FIRST = el_Neutronium
};

/** Cargo Arbiter. This class arbitrates interleaved / overlapping
    cargo transaction.
    
    - when you manipulate cargo, obtain a cargo reservation
      (foo.getCargoHandler().incUseCount())
    - when you wish to remove cargo from somewhere, first reserve
      the cargo on the cargo handler. It will disappear from display.
    - finally, you can commit the transaction by removing the
      reservation and pending-add and writing back.
    - drop the reservation (foo.getCargoHandler().decUseCount()))

    In addition, every object can have several "facilities" (up to
    32, numbered 0 to 31) which, for simplicity, don't have their
    own cargo arbiter. Only one entity can use such a facility at a
    time. Facilities include:
    - a ship's cargo transporters
    - a starbase's hull/engine/beam/torp launcher store */
class GCargo {
 public:
    typedef GCargoType CargoType;
 private:
    struct Impl {
        int ref_count;

        /** Reserved cargo which is not available. */
        int32 reserved_cargo[el_COUNT];

        Impl();
    };
    Impl*   data;
    uint32  used_facilities;
 public:
    GCargo()
        : data(0), used_facilities(0) { }
    ~GCargo();

    void  incUseCount();
    void  decUseCount() throw();

    /** Return reserved amount, for user interface purposes. */
    int32 getReserved(CargoType t) const { return data ? data->reserved_cargo[t] : 0; }

    /** Change reservation. */
    int32 changeReservation(CargoType t, int32 diff) throw();

    /** Query whether facility is in use. */
    bool isFacilityInUse(int n) { return (used_facilities & (1L << n)) != 0; }
    
    void acquireFacility(int n) throw();
    void releaseFacility(int n) throw();
};

/* bummer. Why can't C++ have these built-in? */
inline GCargo::CargoType operator++(GCargo::CargoType& t) { return t = GCargo::CargoType(t+1); }
inline GCargo::CargoType operator--(GCargo::CargoType& t) { return t = GCargo::CargoType(t-1); }
inline GCargo::CargoType operator++(GCargo::CargoType& t,int) { GCargo::CargoType rv = t; ++t; return rv; }
inline GCargo::CargoType operator--(GCargo::CargoType& t,int) { GCargo::CargoType rv = t; --t; return rv; }

//! Abstract cargo container.
class GCargoContainer : public GTransaction {
 public:
    typedef GCargo::CargoType CargoType;
    enum {
        can_Unload     = 1,     ///< this object can provide stuff for unloading
        can_GetUnload  = 2,     ///< this object can receive unloaded stuff
        can_Sup2mc     = 4,     ///< this object can sell supplies

        can_Overload   = 8      ///< do overload. Set by driver code.
    };
 private:
    int32   diffs[el_COUNT];    ///< Target differences.
    int32   reserved[el_COUNT]; ///< Amounts reserved for us. Probably useless.
    int     flags;
 protected:
    // Interface to underlying object
    /** return amount this object has (reserved stuff has been subtracted). */
    virtual int32  getAmount(CargoType type) = 0;
    /** add some amount to object (may be negative). */
    virtual void   changeAmount(CargoType type, int32 amount) = 0;
    /** reserve some amount on object (negative = cancel reservation). */
    virtual void   reserveAmount(CargoType type, int32 amount) = 0;
 public:
    GCargoContainer();
    virtual ~GCargoContainer();

    // Public interface to underlying object
    /** true iff this object can carry specified cargo type. */
    virtual bool   canHave(CargoType cargo_id) = 0;
    /** return maximum amount of cargo this object can carry. */
    virtual int32  getMax(CargoType cargo_id) = 0;

    virtual string_t getName() const = 0;

    // Public manipulation interface
    /** return current "virtual" amount. */
    int32 getCurrent(CargoType cargo_id);
    /** remove some amount from this object. */
    int32 removeCargo(CargoType type, int32 amount, bool partial);
    int32 addCargo(CargoType type, int32 amount, bool partial);

    int   getFlags() const { return flags; }
    void  setFlags(int fl) { flags = fl; }
    bool  hasFlag(int fl) const { return flags & fl; }

    // Transaction interface
    void cancel();
    void commit();
};

//! Cargo transfer transaction.
class GCargoTransfer : public GTransaction {
    Ptr<GCargoContainer> objs[2];
 public:
    GCargoTransfer(Ptr<GCargoContainer> left, Ptr<GCargoContainer> right);
    ~GCargoTransfer();

    GCargoContainer& getContainer(int side) { return *objs[side]; }

    int32 move(GCargo::CargoType type, int32 amount, int from, bool partial, bool s2m);

    void cancel();
    void commit();
};

string_t getCargoName(GCargoType type);
string_t getCargoUnit(GCargoType type);

#endif
