/**
  *  \file game/checksum-crc.cc
  *  \brief CRC/Digest Computations
  *
  *  This module provides functions to compute the PHost digest values.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/checksum.h"

namespace {
    struct DigestTable {
        uint32 numbers[256];
        DigestTable();
        uint32 operator[](uint8 index) { return numbers[index]; }
    };
};

static DigestTable digest_table;

/* bcc dumps an obscure "poly(double,int,double[])" into our namespace */
static const uint32 crc_poly = 0x10811;

DigestTable::DigestTable()
{
    for (int b = 0; b < 256; ++b) {
        uint32 x = b;
        numbers[b] = 0;
        for (int i = 0; i < 8; ++i) {
            if (x & 1)
                x ^= crc_poly;
            x >>= 1;
            numbers[b] += (x+1) * crc_poly;
        }
    }
}

/** Compute digest.
    \param data   data block
    \param length number of bytes
    \param value  starting value.
    \invariant getDigest(p+n, m, getDigest(p, n, x)) == getDigest(p, n+m, x)
    \invariant in particular, getDigest(p, 0, x) = x */
uint32
getDigest(const char* data, std::size_t length, uint32 value)
{
    while (length--)
        value += digest_table[(value ^ *data++) & 255] + value >> 16;
    return value;
}
