/**
  *  \file game/checksum-ctl.cc
  *  \brief Checksums -- Control File Management
  *
  *  In Dosplan, a file `control.dat' stores checksums over ship,
  *  pdata and bdata records for each directory. In Winplan, such a
  *  file is created for each player. This module manages these
  *  checksums. We maintain the checksums only to please Tim's
  *  maketurns, we don't check them ourselves.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "fileio/io.h"
#include "fileio/streams.h"
#include "game/checksum.h"
#include "game/gamedirs.h"
#include "util/console.h"
#include "util/nls.h"
#include "cpluslib/assert.h"
#include "cpluslib/format.h"

static unsigned const CONTROL_MIN = 1501, CONTROL_MAX = 2499;

static int32* control_data;

/** Get index into control file. */
static inline int
getControlIndex(int otyp, int oid)
{
    switch(otyp) {
     case 1:
        if (oid <= 500)
            return oid-1;
        else
            return oid+1499;
     case 2:
        return oid + 499;
     case 3:
        return oid + 999;
     default:
        FAIL();
        return 0;
    }
}

/** Get control entry for an object. */
int32
getControlEntry(int otyp, int oid)
{
    return control_data[getControlIndex(otyp, oid)];
}

/** Store control entry (checksum) for an object. */
void
setControlEntry(int otyp, int oid, int32 checksum)
{
    control_data[getControlIndex(otyp, oid)] = checksum;
}

/** Load control file for player.
    \returns value to pass to saveControl(). Zero in Dosplan games, player
    number in Winplan. */
int
loadControl(int player)
{
    initControl();

    int type;
    Ptr<Stream> f = game_file_dir->openFileNT("control.dat", Stream::C_READ);
    if (f) {
        type = 0;
    } else if ((f = game_file_dir->openFileNT(format("contrl%d.dat", player), Stream::C_READ))) {
        type = player;
    } else {
        console.write(LOG_INFO, _("No control file (checksums) loaded"));
        // FIXME
        return -1;
    }

    /* load control file */
    char buffer[4 * CONTROL_MAX];
    std::size_t count = f->read(buffer, sizeof buffer);
    if (count > 0) {
        count /= 4;
        if (count > CONTROL_MAX)
            count = CONTROL_MAX;
        getInt32(buffer, control_data, count);
    }
    return type;
}

/** Save control file. */
void
saveControl(int type)
{
    Ptr<Stream> f =
        game_file_dir->openFileNT((type == 0
                                   ? string_t("control.dat")
                                   : string_t(format("contrl%d.dat", type))),
                                  Stream::C_WRITE | Stream::C_CREATE);
    if (!f) {
        console.write(LOG_ERROR, _("Control file (checksums) can't be created"));
    } else {
        /* figure out size of file */
        int size = 6002;
        char buffer[CONTROL_MAX * 4];
        buffer[6000] = buffer[6001] = 0;

        for (unsigned i = CONTROL_MIN; i < CONTROL_MAX; ++i) {
            if (control_data[i]) {
                size = CONTROL_MAX*4;
                break;
            }
        }

        storeInt32(buffer, control_data, size / 4);
        f->write(buffer, size);
    }
}

/** Initialize this module. This call is optional when you use
    loadControl(). */
void
initControl()
{
    if (!control_data)
        control_data = new int32[CONTROL_MAX];
    std::fill_n(control_data, CONTROL_MAX, 0);
}

/** Shut down this module. Frees memory allocated by initControl()
    or loadControl(). */
void
doneControl()
{
    if (control_data) {
        delete[] control_data;
        control_data = 0;
    }
}
