/**
  *  \file game/cmds.cc
  *  \brief Auxiliary Command Management
  *
  *  This module provides functions to manage auxiliary commands (PHost
  *  command processor, THost allies) -- basically all this information
  *  that is not represented in a .dat/.dis file pair but must still be
  *  sent to the host.
  *
  *  We parse PHost commands, and send them out in "canonical" format. 
  *  Basically, when player says "re c 3" using some message-writing
  *  program and we see it, the spelled-out form "remote control 3"
  *  will be sent to the host.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/cmds.h"
#include "cpluslib/format.h"
#include "cpluslib/assert.h"
#include "util/misc.h"

/******************************** GCommand *******************************/

/** \class GCommand
    \brief Auxiliary command.

    A command is uniquely identified by its Type/cmd pair. A second
    command with the same pair "overwrites" the first one. Some
    commands require the id to be zero, see the comment in the Type
    enum definition.

    The phc_Other command is an exception to all these rules: it
    contains a command that is not parsed by PCC and hence not
    identified in any way. Any number of phc_Other commands can
    coexist.

    The actual lookup/replace policy is implemented in
    GCommandContainer. */

GCommand::~GCommand()
{ }

/** Return complete text of command. This text can be written
    cmdX.txt. Not all commands are suitable for command messages. */
string_t
GCommand::getCommandText() const
{
    /* PCC 1.x: phost.pas::PHostCommandString */

    switch (cmd) {
     case phc_Other:
        return arg;
     case phc_Language:
        return "language " + arg;
     case phc_SendConfig:
        return "send config";
     case phc_SendRacenames:
        return "send racenames";
     case phc_SendFcodes:
        return "send fcodes";
     case phc_SetRaceName:
        {
            ASSERT(id >= 0 && id <= 2);
            static const char*const rn_id[] = { "long ", "short ", "adj " };
            return string_t("race ") + rn_id[id] + arg;
        }
     case phc_Filter:
        return "filter " + arg;
     case phc_ConfigAlly:
        return format("allies config %d %s", id, arg);
     case phc_AddDropAlly:
        return format("allies %s %d", arg, id);
     case phc_RemoteControl:
        return format("remote %s %d", arg, id);
     case phc_RemoteDefault:
        return "remote " + arg + " default";
     case phc_Beamup:
        return format("beamup %d %s", id, arg);
     case phc_GiveShip:
        return format("give ship %d to %s", id, arg);
     case phc_GivePlanet:
        return format("give planet %d to %s", id, arg);
     case phc_TAlliance:
        return "$thost-allies " + arg;
    }
    return string_t();
}

// FIXME: these functions should probably be elsewhere
/* i=0 = first word */
static string_t::size_type
findNthWordStart(const string_t& s, int i)
{
    string_t::size_type k = 0;
    while (k < s.length() && s[k] == ' ')
        ++k;
    while (i > 0) {
        while (k < s.length() && s[k] != ' ')
            ++k;
        while (k < s.length() && s[k] == ' ')
            ++k;
    --i;
    }
    return k;
}

static string_t
findNthWord(const string_t& s, int i)
{
    string_t::size_type k = findNthWordStart(s, i), j = k;
    while (j < s.length() && s[j] != ' ')
        ++j;
    if (k >= s.length())
        return string_t();
    else
        return string_t(s, k, j-k);
}

static string_t
findNthWordRest(const string_t& s, int i)
{
    string_t::size_type k = findNthWordStart(s, i);
    if (k >= s.length())
        return string_t();
    else
        return string_t(s, k);
}

/** Parse a command. \return command, 0 if not recognized.
    \param fromfile true iff this command comes from a cmdX.txt file, false
    if from a message. */
GCommand*
GCommand::parseCommand(string_t text, bool fromfile)
{
    text = strLTrim(text);

    string_t verb = findNthWord(text, 0);
    int wordoffs  = 0;

    /* Private Commands */
    if (fromfile && strCaseCmp(verb, "$THOST-ALLIES") == 0)
        return new GCommand(phc_TAlliance, 0, findNthWordRest(text, 1));
    
    /* PHost 4 Command Syntax */
    if (strCaseCmp(verb, "phost") == 0 || strCaseCmp(verb, "phost:") == 0) {
        ++wordoffs;
        verb = findNthWord(text, wordoffs);
    }

    string_t arg = findNthWord(text, wordoffs + 1);

    if (stringMatch("Send", verb)) {
        /* send config|racename|fcodes */
        if (stringMatch("Config", arg))
            return new GCommand(phc_SendConfig, 0, string_t());
        else if (stringMatch("Racename", arg))
            return new GCommand(phc_SendRacenames, 0, string_t());
        else if (stringMatch("Fcodes", arg))
            return new GCommand(phc_SendFcodes, 0, string_t());
    } else if (stringMatch("Language", verb) && !arg.empty()) {
        /* language <language> */
        return new GCommand(phc_Language, 0, arg);
    } else if (stringMatch("Filter", verb) && !arg.empty()) {
        /* filter yes|no */
        return new GCommand(phc_Filter, 0, arg);
    } else if (stringMatch("Give", verb)) {
        /* give ship|planet <id> [to] <race> */
        int id;
        if (stringValue(findNthWord(text, wordoffs + 2), id)) {
            int k = stringMatch("To", findNthWord(text, wordoffs + 3)) ? 4 : 3;
            if (stringMatch("Ship", arg))
                return new GCommand(phc_GiveShip, id, findNthWord(text, wordoffs + k));
            else if (stringMatch("Planet", arg))
                return new GCommand(phc_GivePlanet, id, findNthWord(text, wordoffs + k));
        }
    } else if (stringMatch("Allies", verb)) {
        int player;
        if (stringValue(findNthWord(text, wordoffs + 2), player)) {
            if (stringMatch("Config", arg)) {
                /* allies config <race> <flags> */
                string_t rest = findNthWordRest(text, wordoffs + 3);
                if (!rest.empty())
                    return new GCommand(phc_ConfigAlly, player, rest);
            } else if (stringMatch("Add", arg) || stringMatch("Drop", arg)) {
                /* allies add|drop <race> */
                return new GCommand(phc_AddDropAlly, player, arg);
            }
        }
    } else if (stringMatch("REmote", verb)) {
        /* remote <verb> <id>|default */
        int sid;
        string_t id = findNthWord(text, wordoffs + 2);
        if (stringMatch("Default", id)) {
            return new GCommand(phc_RemoteDefault, 0, arg);
        } else if (stringValue(id, sid)) {
            return new GCommand(phc_RemoteControl, sid, arg);
        }
    } else if (stringMatch("BEamup", verb)) {
        /* beamup <pid> <stuff> */
        int pid;
        if (stringValue(findNthWord(text, wordoffs + 1), pid))
            return new GCommand(phc_Beamup, pid, findNthWordRest(text, wordoffs + 2));
    } else if (stringMatch("RAcename", verb)) {
        /* racename long|short|adj <name> */
        int what =
            stringMatch("Long", arg) ? phcrn_Long :
            stringMatch("Short", arg) ? phcrn_Short :
            stringMatch("Adj", arg) ? phcrn_Adjective : -1;
        if (what >= 0)
            return new GCommand(phc_SetRaceName, what, findNthWordRest(text, wordoffs + 2));
    }
    return 0;
}

/** Check for message introducer commands. True iff text is a command that
    starts a message (i.e. the rest of the message *is* the message, not
    further commands). */
bool
GCommand::isMessageIntroducer(string_t text)
{
    string_t cmd = findNthWord(text, 0);
    return (stringMatch("RUmor", cmd) || stringMatch("RUmour", cmd) || stringMatch("Message", cmd));
}

/*************************** GCommandContainer ***************************/

/** \class GCommandContainer
    \brief Command Container

    This class manages a set of commands (GCommand). It implements
    storing / querying the command list, and the replace policy for
    the commands.

    \todo currently, this is a simple list, like it was in PCC v1. 
    Should performance be a problem, change it to std::map or
    somesuch. */

GCommandContainer::GCommandContainer()
{ }

GCommandContainer::~GCommandContainer()
{
    clear();
}

/** Clear this container. Discards all the commands. */
void
GCommandContainer::clear()
{
    while (!cmds.empty()) {
        delete cmds.front();
        cmds.pop_front();
    }
}

GCommandContainer::command_list::iterator
GCommandContainer::findCommand(GCommand::Type typ, int id)
{
    /* any number of phc_Other can coexist, so we can not find a
       particular one by just giving a typ/id pair */
    if (typ == GCommand::phc_Other)
        return cmds.end();
    
    command_list::iterator i = cmds.begin();
    while (i != cmds.end() && (*i)->getId() != id && (*i)->getCommand() != typ)
        ++i;
    return i;
}

/** Fetch command typ/id. \returns command, or 0 if no such command
    exists. */
const GCommand*
GCommandContainer::getCommand(GCommand::Type typ, int id)
{
    command_list::iterator i = findCommand(typ, id);
    if (i != cmds.end())
        return *i;
    else
        return 0;
}

/** Submit a command. When a command with the same typ/id pair exists,
    it is overwritten; otherwise a new command is created.
    \returns the GCommand created/reused by this operation. */
const GCommand*
GCommandContainer::setCommand(GCommand::Type typ, int id, string_t arg)
{
    command_list::iterator i = findCommand(typ, id);
    if (i != cmds.end()) {
        (*i)->setArg(arg);
        return *i;
    } else {
        GCommand* c = new GCommand(typ, id, arg);
        cmds.push_back(c);
        return c;
    }
}

/** Submit a command. Like setCommand(GCommand::Type,int,string_t),
    but this one uses an existing GCommand object. The GCommand*
    pointer gets owned by this GCommandContainer.
    \returns cmd */
const GCommand*
GCommandContainer::setCommand(GCommand* cmd)
{
    command_list::iterator i = findCommand(cmd->getCommand(), cmd->getId());
    if (i != cmds.end()) {
        delete *i;
        *i = cmd;
    } else {
        cmds.push_back(cmd);
    }
    return cmd;
}

/** Remove a command. \returns true iff command found, false if not. */
bool
GCommandContainer::removeCommand(GCommand::Type typ, int id)
{
    command_list::iterator i = findCommand(typ, id);
    if (i != cmds.end()) {
        cmds.erase(i);
        return true;
    } else
        return false;
}
