/**
  *  \file game/config.cc
  *  \brief PConfig Management
  *
  *  This file contains functions to manage and parse PConfig.
  *  HConfig is a strict subset of PConfig, so it is handled
  *  here, too.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include <cstring>
#include <cstdio>
#include <vector>
#include "game/config.h"
#include "game/struct.h"
#include "fileio/io.h"
#include "util/bsearch.h"
#include "fileio/iostr.h"
#include "util/except.h"
#include "util/confpars.h"
#include "util/misc.h"
#include "defs.h"
#include "util/console.h"
#include "util/nls.h"
#include "game/gamedirs.h"
#include "cpluslib/stringcv.h"
#include "cpluslib/assert.h"
#include "cpluslib/format.h"

/* Configuration structure */
#include "game/config-s.inc"

template<class T, class O>
T& member_ref(O& orig, std::size_t offset) {
    return *(T*)(((char*)&(orig)) + offset);
}

template<class T, class O>
const T& const_member_ref(const O& orig, std::size_t offset) {
    return *(const T*)(((const char*)&(orig)) + offset);
}

static void
doThrow(const string_t& str)
{
    throw std::domain_error(toStandardString(str));
}

/** Game configuration. This structure contains the configuration
    of the game (HCONFIG/PCONFIG). */
TGameConfig config;

/** Host version. */
THostType host_type = HT_UNKNOWN;

/** Host version code. Encoded in hex, e.g. THost 3.22.006 is
    0x031606. Zero if not determined. */
uint32 host_version_code = 0;

/** PCONFIG parser. This class maps the ConfigParser interface
    into calls to assignConfig. \internal */
class PConfigParser : public ConfigParser {
    string_t name;
    std::vector<bool> set_options;
 public:
    PConfigParser(string_t aname);
    void assign(int line_id, string_t name, string_t value,
                string_t line);
    void error(int line_id, string_t name);
    void postprocess();
    void setName(const string_t& name) { this->name = name; }
};

PConfigParser::PConfigParser(string_t aname)
    : name(aname), set_options(PCI_MAX)
{ }

void
PConfigParser::assign(int line_id, string_t name, string_t value,
                      string_t)
{
    try {
        int index = lookupConfigOption(name);
        if (index < 0) {
            error(line_id, format(_("unknown parameter `%s'"), name));
        } else {
            assignConfig(index, value);
            set_options[index] = true;
        }
    }
    catch(std::domain_error& e) {
        error(line_id, e.what());
    }
}

void
PConfigParser::error(int line_id, string_t msg)
{
    console.write(LOG_WARN, format(_("%s:%d: %s -- line has been ignored"), name, line_id, msg));
}

void
PConfigParser::postprocess()
{
    for (unsigned int i = 0; i < countof(config_info); ++i) {
        if (!set_options[i]) {
            assignConfig(i, config_info[i].deflt);
            set_options[i] = true;
        }
    }
}

/** Load PCONFIG.SRC.
    \param pconf pconfig.src file
    \param shiplist shiplist.txt file, may be null. */
void
loadPConfig(Ptr<Stream> pconf, Ptr<Stream> shiplist)
{
    StandardStream in(pconf);
    PConfigParser parser(pconf->getName());
    parser.parse(in, "phost", true);
    if (shiplist) {
        parser.setName(shiplist->getName());
        parser.parse(in, "phost", false);
    }
    parser.postprocess();
    if (!host_type)
        host_type = HT_PHOST;
}

static string_t
buildString(const TGameConfig& config, int offset, int count, TStructureMember type)
{
    string_t rv;
    while (count-- > 0) {
        if (rv.length())
            rv.append(1, ',');
        switch (type) {
         case T_BOOL:
            if (const_member_ref<bool>(config, offset))
                rv.append("yes");
            else
                rv.append("no");
            offset += sizeof(bool);
            break;
         case T_TRISTATE:
            switch (const_member_ref<TTristate>(config, offset)) {
             case tri_Yes:
                rv.append("yes");
                break;
             case tri_No:
                rv.append("no");
                break;
             default:
                rv.append("allies");
                break;
            }
            offset += sizeof(TTristate);
            break;
         case T_INT16:
            rv.append(itoa(const_member_ref<int16>(config, offset)));
            offset += 2;
            break;
         case T_INT32:
            rv.append(itoa(const_member_ref<int32>(config, offset)));
            offset += 4;
            break;
         case T_STRING:
            rv.append(const_member_ref<string_t>(config, offset));
            offset += sizeof(string_t);
            break;
         default:
            ASSERT(0);
        }
    }
    return rv;
}

/** Copy HCONFIG Item.
    \param buffer HCONFIG.HST image
    \param size   its size
    \param pci    relevant config item

    Does nothing if HCONFIG is too short. */
static void
maybeParseHConfigItem(char* buffer, int size,
                      const TPConfigConvert& pci)
{
    using namespace std;        // BCC chokes on `std::strcpy' when optimizing
    
    /* use the dull version? For first time, yes.
       Actually, field sizes are the same in THost and PHost,
       so we could optimize this a lot. FIXME */
    char value[200];
    char* ptr = value;

    buffer += pci.t_ofs;
    size -= pci.t_ofs;

    /* construct a string. Yeah, C, I know... */
    for (unsigned int i = 0; i < pci.t_count; ++i) {
        if (i)
            *ptr++ = ',';
        switch(pci.t_type) {
         case T_BOOL:
            if ((size -= 2) < 0)
                return;
            strcpy(ptr, getUint16(buffer) ? "Yes" : "No");
            buffer += 2;
            break;
         case T_INT16:
            if ((size -= 2) < 0)
                return;
            sprintf(ptr, "%d", getInt16(buffer));
            buffer += 2;
            break;
         case T_INT32:
            if ((size -= 4) < 0)
                return;
            sprintf(ptr, "%ld", (long) getInt32(buffer));
            buffer += 4;
            break;
         default:
            FAIL();
        }
        ptr += strlen(ptr);
    }

    // FIXME FIXME FIXME
    assignConfig(lookupConfigOption(pci.name), value);
}

/** Load HCONFIG.HST.
    \todo do host version guessing here? */
void
loadHConfig(Ptr<Stream> hconf)
{
    Stream::fpos_t size = hconf->getSize();
    if (size < 0 || size > CONF_MAX_HCONFIG_SIZE)
        throw FileFormatException(format(_("File `%s' has invalid size"),
                                         hconf->getName()));

    ArrayPtr<char> data (size);
    if (hconf->read(data.ptr(), size) != unsigned(size))
        throw FileIOException(format(_("Error reading `%s'"), hconf->getName()));

    /* File read correctly. Now parse it. */
    for (unsigned i = 0; i < countof(config_info); ++i)
        if (config_info[i].t_count)
            maybeParseHConfigItem(data.ptr(), size, config_info[i]);

    /* Guess version. */
    if (!host_type)
        host_type = HT_THOST;
    
    if (!host_version_code) {
        if (size > 338)
            host_version_code = 0x031624; // 3.22.036
        else if (size > 336)
            host_version_code = 0x031616; // 3.22.022
        else if (size > 302)
            host_version_code = 0x031400; // 3.20
        else if (size > 298)
            host_version_code = 0x030D02; // 3.13b
        else if (size > 288)
            host_version_code = 0x030C03; // 3.12c
        else if (size > 186)
            host_version_code = 0x030B01; // 3.11a
        else if (size > 10)
            host_version_code = 0x030A00; // 3.10 (?)
        else
            host_version_code = 0x030000; // 3.00
    }
}

/** Load default config. Initializes the config variable with all
    default values. The defaults are the default values from PHost,
    which should also be the THost defaults. */
void
loadDefaultConfig()
{
    for (unsigned i = 0; i < countof(config_info); ++i)
        assignConfig(i, config_info[i].deflt);
    config.GameName = "Unnamed Game";   // be neutral ;)
}

/** \defgroup pconfigparse PConfig Parsing Functions
    These functions are used by the config parser to do the individual
    assignments. */
//@{

/** Parse Integer. Parses value into an integer and verifies its range
    using the information in pci. \throws std::domain_error if there
    is a syntax or range error. */
static int32
parseInt(const TPConfigConvert* pci, string_t value)
{
    char* endptr;
    int32 l = std::strtol(value.c_str(), &endptr, 0);
    if (!value.empty() && *endptr == 0)
        if (l >= pci->min && l <= pci->max)
            return l;
        else
            doThrow(format(_("Value out of range for option `%s'"), pci->name));
    else
        doThrow(format(_("Invalid integer value for option `%s'"), pci->name));
}

/** Assign integer (T_INT16). Assign to the int16 quantity at \c offset. */    
static void
assignConfigInt(const TPConfigConvert* pci, std::size_t offset, string_t value)
{
    member_ref<int16>(config, offset) = parseInt(pci, value);
}

/** Assign integer (T_INT32). Assign to the int32 quantity at \c offset. */
static void
assignConfigLong(const TPConfigConvert* pci, std::size_t offset, string_t value)
{
    member_ref<int32>(config, offset) = parseInt(pci, value);    
}

/** Assign boolean (T_BOOL). Assign to the bool at \c offset. */
static void
assignConfigBool(const TPConfigConvert* pci, std::size_t offset, string_t value)
{
    if (stringMatch("Yes", value) || stringMatch("True", value) || stringMatch("External", value))
        member_ref<bool>(config, offset) = true;
    else if (stringMatch("No", value) || stringMatch("False", value))
        member_ref<bool>(config, offset) = false;
    else
        doThrow(format(_("Invalid boolean value `%s' for `%s'"), value, pci->name));
}

/** Assign tristate (T_TRISTATE). Assign to the tristate at \c offset. */
static void
assignConfigTristate(const TPConfigConvert* pci, std::size_t offset, string_t value)
{
    if (stringMatch("Yes", value) || stringMatch("True", value))
        member_ref<TTristate>(config, offset) = tri_Yes;
    else if (stringMatch("No", value) || stringMatch("False", value))
        member_ref<TTristate>(config, offset) = tri_Yes;
    else if (stringMatch("Allies", value))
        member_ref<TTristate>(config, offset) = tri_Ally;
    else
        doThrow(format(_("Invalid value `%s' for `%s'"), value, pci->name));
}

/** Assign config string (T_STRING). Assign to the string at \c offset */
static void
assignConfigString(const TPConfigConvert* /*pci*/, std::size_t offset, string_t value)
{
    member_ref<string_t>(config, offset) = value;
}

/** Parse one config item.
    This splits array options, expands arrayized options, and calls the
    assign function for each of them.
    \param pci    Definition of config item
    \param assign Assignment function.
    \param size   Size of one item. Used for array/arrayized options.
    \param value  The value from the config file. */
static void
parseConfigItem(const TPConfigConvert* pci,
                void (*assign)(const TPConfigConvert*, std::size_t, string_t),
                std::size_t size,
                string_t value)
{
    std::size_t offset = pci->offset;
    unsigned remain    = pci->count;

    while (remain > 1) {
        string_t::size_type n = value.find (",");
        if (n == string_t::npos)
            break;
        assign(pci, offset, strTrim(value.substr(0, n)));
        value.erase(0, n+1);
        offset += size;
        --remain;
    }

    if (remain > 1 && !pci->arrayized)
        doThrow(string_t("Too few values in assignment to `") + pci->name + "'");

    value = strTrim(value);
    while (remain > 0) {
        assign(pci, offset, value);
        offset += size;
        --remain;
    }
}
//@}

/** Assign one config item.
    \param config_name   Name of the option (e.g. "AllowHiss")
    \param config_value  Value (e.g. "Yes")
    \throws std::domain_error if an error occurs. */
void
assignConfig(int index, string_t config_value)
{
    config_value = strTrim(config_value);
    if (config_value.length() && config_value[0] == '$') {
        /* it's a reference */
        int refindex = lookupConfigOption(strLTrim(config_value.substr(1)));
        if (refindex < 0)
            doThrow("invalid reference `" + config_value + "'");
        // FIXME: we'd probably like to know whether this option is already set
        config_value = buildString(config,
                                   config_info[refindex].offset,
                                   std::min(config_info[refindex].count,
                                            config_info[index].count),
                                   config_info[refindex].type);
    }
    const TPConfigConvert* pci = &config_info[index];
    switch (pci->type) {
     case T_INT16:
        parseConfigItem(pci, assignConfigInt, sizeof(int16), config_value);
        break;
     case T_INT32:
        parseConfigItem(pci, assignConfigLong, sizeof(int32), config_value);
        break;
     case T_BOOL:
        parseConfigItem(pci, assignConfigBool, sizeof(bool), config_value);
        break;
     case T_TRISTATE:
        parseConfigItem(pci, assignConfigTristate, sizeof(TTristate), config_value);
        break;
     case T_STRING:
        if (pci->count == 1)
            /* shortcut, to permit commas in values (GameName) */
            member_ref<string_t>(config, pci->offset) = config_value;
        else
            parseConfigItem(pci, assignConfigString, sizeof(string_t), config_value);
        break;
     default:
        FAIL();
    }
}

/** Assign one config item. \overload */
void
assignConfig(string_t config_name, string_t config_value)
{
    assignConfig(config_name.c_str(), config_value.c_str());
}

/** Compare Config Options. Predicate object for use with bin_search
    to look up config options by name. */
struct CompareConfigOption {
    const char* name;
    CompareConfigOption(const char* aname) : name(aname) { }

    bool isEqual(const TPConfigConvert& rhs) const {
        return strCaseCmp(name, rhs.name) == 0;
    }
    bool isLess(const TPConfigConvert& rhs) const {
        return strCaseCmp(name, rhs.name) < 0;
    }
    bool isEqual(const TPConfigAliasTable& rhs) const {
        return strCaseCmp(name, rhs.name) == 0;
    }
    bool isLess(const TPConfigAliasTable& rhs) const {
        return strCaseCmp(name, rhs.name) < 0;
    }
};

/** Look up config option. Returns index into config_info,
    or -1 on error. */
int
lookupConfigOption(const char* name)
{
    /* bcc55 needs `&config_info[0]' instead of `config_info'. Argh. */
    int i = bin_search(&config_info[0], countof(config_info),
                       CompareConfigOption(name));
    if (i < 0) {
        i = bin_search(&config_alias[0], countof(config_alias),
                       CompareConfigOption(name));
        if (i > 0)
            i = config_alias[i].index;
    }
    return i;
}

/** Look up config option. \overload */
int
lookupConfigOption(string_t name)
{
    return lookupConfigOption(name.c_str());
}

void
initConfig()
{
    Ptr<Stream> ps;

    loadDefaultConfig();

    ps = spec_file_dir->openFileNT("pconfig.src", Stream::C_READ | Stream::C_TEXT);
    if (ps.ptr()) {
        loadPConfig(ps, spec_file_dir->openFileNT("shiplist.txt", Stream::C_READ | Stream::C_TEXT));
        return;
    }
    ps = spec_file_dir->openFileNT("hconfig.hst", Stream::C_READ);
    if (ps.ptr()) {
        loadHConfig(ps);
        return;
    }
    console.write(LOG_WARN, _("No host configuration file found -- using defaults"));
}
