/**
  *  \file game/fcode-load.cc
  *  \brief Friendly Code Loading
  *
  *  This module provides functions for loading friendly code definitions
  *  from definition files (fcodes.cc).
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <iostream>
#include "game/fcode.h"
#include "fileio/iostr.h"
#include "game/global.h"
#include "util/console.h"
#include "cpluslib/string.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "util/except.h"

/** Load friendly code list from a file. Syntax errors are logged on
    the console.
    \param inp   the file
    \param name  name of the file, for error messages. */
void
GFCodeList::loadFromFile(std::istream& inp, const string_t& name)
{
    int line_nr = 0;
    string_t line;
    while (getline(inp, line)) {
        ++line_nr;
        line = strTrim(line);
        if (line.empty() || line[0] == ';')
            continue;

        string_t::size_type p = line.find_first_of("=,");
        if (p == string_t::npos)
            console.write(LOG_ERROR, format(_("%s:%d: missing delimiter"), name, line_nr));
        if (line[p] == ',') {
            string_t fc = strTrim(line.substr(0, p));
            line.erase(0, p+1);
            if (fc.length() > 3)
                console.write(LOG_ERROR, format(_("%s:%d: friendly code too long; truncated"), name, line_nr));
            try {
                addFCode(GFCode(fc, line));
            }
            catch(UserException& e) {
                console.write(LOG_ERROR, format("%s:%d: %s", name, line_nr, e.what()));
            }
        } else {
            /* FIXME: additional assignments. These still need to be
               implemented. */
        }
    }
    sort();
}

/** Load friendly codes. This initializes all the friendly-code
    related data structures. */
void
initFCodes()
{
    GFCodeList::getInstance().clear();
    Ptr<Stream> s = spec_file_dir->openFileNT("fcodes.cc", Stream::C_READ);
    if (s) {
        StandardStream inp(s);
        GFCodeList::getInstance().loadFromFile(inp, s->getName());
    }
    s = spec_file_dir->openFileNT("fcodes.usr", Stream::C_READ);
    if (s) {
        StandardStream inp(s);
        GFCodeList::getInstance().loadFromFile(inp, s->getName());
    }
    // FIXME: xtrfcode.txt
}
