/**
  *  \file game/fcode.cc
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include "game/fcode.h"
#include "game/ship.h"
#include "game/planet.h"
#include "game/spec.h"
#include "util/misc.h"
#include "fileio/io.h"
#include "util/except.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "util/random.h"
#include "fileio/streams.h"

GFCode::GFCode()
{ }

/** Construct from fcodes.cc file line. This one handles a line with
    that originally contained `c + "," + desc_line'.
    \param c          friendly code
    \param desc_line  description line */
GFCode::GFCode(string_t c, string_t desc_line)
{
    storeBasicString(code, c);

    string_t flag_str;
    if (!strSplit(desc_line, flag_str, description, ","))
        throw UserException(format(_("Friendly code `%s' lacking description"), c));

    players = ~0;
    if (!parseFlags(flag_str, "SPBCARU", &flags, &players, true))
        throw UserException(format(_("Malformed flags for friendly code `%s'"), c));
}

GFCode::~GFCode()
{ }

/** Check whether code works on specified object.
    \param o a ship or planet */
bool
GFCode::worksOn(const GObject& o) const
{
    if (const GShip* s = dynamic_cast<const GShip*>(&o))
        return worksOn(*s);
    else if (const GPlanet* p = dynamic_cast<const GPlanet*>(&o))
        return worksOn(*p);
    else
        return false;
}

/** Check whether code works on ship. \overload */
bool
GFCode::worksOn(const GShip& s) const
{
    if ((players & (1 << raceId(s.getOwner()))) == 0)
        return false;
    if ((flags & fc_Ship) == 0)
        return false;
    if ((flags & fc_Capital) && (s.getBeamCount() == 0)
        && (s.getTorpLauncherCount() == 0) && (s.getBayCount() == 0))
        return false;
    if ((flags & fc_Alchemy) && !s.canDoFunction(hf_MerlinAlchemy)
        && !s.canDoFunction(hf_NeutronicRefinery)
        && !s.canDoFunction(hf_AriesRefinery))
        return false;
    return true;
}

/** Check whether code works on planet. \overload */
bool
GFCode::worksOn(const GPlanet& p) const
{
    if ((players & (1 << raceId(p.getOwner()))) == 0)
        return false;
    if ((flags & fc_Planet) || ((flags & fc_Base && p.hasBase())))
        return true;
    return false;
}

bool
GFCode::isNumeric(const string_t& fc)
{
    if (fc.length() != 3)
        return false;
    return ((fc[0] >= '0' && fc[0] <= '9') || (fc[0] == '-' && isPHost()))
        && (fc[1] >= '0' && fc[1] <= '9')
        && (fc[2] >= '0' && fc[2] <= '9');
}

/** Get a friendly code's numeric value. */
int
GFCode::getNumericValue(const string_t& fc)
{
    if (isNumeric(fc)) {
        int n;
        stringValue(fc, n);
        return n;
    } else
        return 1000;
}

/** Check whether fc is permitted as a "random" code. Random codes must be
    - not special
    - not numeric
    - "random enough" */
bool
GFCode::isAllowedRandom(const string_t& fc)
{
    return fc.length() == 3
        && fc.compare(0, 2, "mf", 2) == 0
        && fc[0] != 'X' && fc[0] != 'x'
        && fc[1] != fc[0] && fc[2] != fc[0] && fc[1] != fc[2]
        && !isSpecial(fc)
        && !isExtraFC(fc)
        && !isNumeric(fc);
}

/** Generate a random friendly code. */
string_t
GFCode::getRandomFCode()
{
    /* to guarantee termination, bail out if we didn't find a good enough
       code after a while. I never saw that happening, but lusers can
       easily misconfigure their system to trigger it (by defining every
       nonblank character special via xtrfcode.txt) */
    int paranoia_counter = 200;
    string_t rv(3, ' ');
    do {
        rv[0] = 33 + myRandom(90);
        rv[1] = 33 + myRandom(90);
        rv[2] = 33 + myRandom(90);
    } while (!isAllowedRandom(rv) && --paranoia_counter);
    return rv;
}

static Buffer<char> extra_fcodes;

/** Check whether fc is a special friendly code (according to xtrafcode.txt). */
bool
GFCode::isExtraFC(const string_t& fc)
{
    using namespace std;
    char* p = extra_fcodes.begin();
    char* end = extra_fcodes.end();
    while (p < end) {
        while (p < end && isspace(static_cast<unsigned char>(*p)))
            ++p;
        
        char* q = p;
        while (p < end && !isspace(static_cast<unsigned char>(*p)))
            ++p;
        if (p != q && p-q <= fc.length() && fc.compare(0, fc.length(), q, p-q)==0)
            return true;
    }
    return false;
}

/** Clear extra fcodes list.
    \post isExtraFC(x) == false for all codes */
void
GFCode::clearExtraFC()
{
    extra_fcodes.clear();
}

/** Load extra fcodes list. This will append the specified file to the
    current list. To replace the list, clear it first
    (clearExtraFC()). When s is null, does nothing. */
void
GFCode::loadExtraFC(Stream* s)
{
    if (s) {
        s->readBuffer(extra_fcodes);
        extra_fcodes.append(' ');
    }
}
