/**
  *  \file game/gen.cc
  *  \brief "GEN" file management
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include "game/gen.h"

const char GGen::NOPASSWORD[] = "NOPASSWORD";

/** Initialize with unpacked GEN record. */
GGen::GGen(const TGen& agen)
    : gen(agen)
{
    genSig();
}

/** Initialize with packed GEN record (from RST). */
GGen::GGen(const TResultGen& agen)
{
    std::memcpy(gen.timestamp, agen.timestamp, sizeof(gen.timestamp));
    std::memcpy(gen.scores,    agen.scores,    sizeof(gen.scores));
    gen.player_id = agen.player_id;
    std::memcpy(gen.password,  agen.password,  sizeof(agen.password));
    gen.unused = 0;
    gen.ship_checksum = gen.pdata_checksum = gen.bdata_checksum = 0;
    gen.new_password_flag = 0;
    std::memset(gen.new_password, 0, sizeof(gen.new_password));
    gen.turn = agen.turn;
    gen.timestamp_checksum = agen.timestamp_checksum;
    genSig();
}

/** Fetch password, and store it in p. Returns length of password.
    p must point to 10 bytes. This routine is consistent with PCC 1.x
    in handling "bogus" passwords. */
unsigned
GGen::getPassword(char* p) const
{
    if (gen.new_password_flag == 13) {
        /* New password has been set */
        for (int i = 0; i < 10; ++i)
            p[i] = gen.new_password[i] - 50;
    } else {
        /* No new password */
        for (int i = 0; i < 10; ++i) {
            int k = gen.password[i] - gen.password[19-i] + 32;
            if (k < 32 || k > 127)
                p[i] = 32;
            else
                p[i] = k;
        }
    }

    /* compute length of password */
    unsigned n = 10;
    while (n > 0 && p[n-1] <= 32)
        --n;
    return n;
}

/** Test password. Returns true iff apass is the valid password. */
bool
GGen::isPassword(string_t apass) const
{
    char buffer[10];
    unsigned len = getPassword(buffer);
    return apass.length() == len && apass.compare(0, len, buffer, len) == 0;
}

/** Test whether password active. Returns true iff this turn has a
    password active. */
bool
GGen::hasPassword() const
{
    char buffer[10];
    return !(getPassword(buffer) == 10 && std::memcmp(buffer, NOPASSWORD, 10) == 0);
}

/** Change password. */
void
GGen::setPassword(string_t apass)
{
    for (unsigned i = 0; i < 10; ++i)
        if (i >= apass.length())
            gen.new_password[i] = 50;
        else
            gen.new_password[i] = apass[i] + 50;
    gen.new_password_flag = 13;
}

void
GGen::setPasswordFromTurn(const char* data)
{
    std::copy(data, data + 10, gen.new_password);
    gen.new_password_flag = 13;
}

void
GGen::genSig()
{
    for (int i = 0; i < 10; ++i)
        sig2[i] = gen.password[i+10] + i+1;
}
