/**
  *  \file game/global-loc.cc
  *  \brief Global Data -- Location Names
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/global.h"
#include "game/config.h"
#include "game/coord.h"
#include "game/formula.h"
#include "util/nls.h"
#include "game/planet.h"
#include "game/ship.h"
#include "cpluslib/format.h"

/** Get planet at specified position, or 0.
    \post return!=0 => getPlanet(return).getPos() == pt */
int
GGameTurn::getPlanetAt(GPoint pt)
{
    pt = pt.getNormalized();
    for (int i = 1; i <= NUM_PLANETS; ++i)
        if (getPlanet(i).isExistant() && getPlanet(i).getPos() == pt)
            return i;
    return 0;
}

/** Get planet at (x,y).
    \param gravity if true, also use gravity wells. If false, this function
    is exactly like getPlanetAt(GPoint).
    \returns Id of planet, 0 if none. */
int
GGameTurn::getPlanetAt(GPoint pt, bool gravity)
{
    int rv = getPlanetAt(pt);
    if (!rv && gravity)
        rv = getGravityPlanetAt(pt);
    return rv;
}

/** Get planet whose warp wells are in effect at pt, or 0.
    \pre getPlanetAt(pt) == 0 */
int
GGameTurn::getGravityPlanetAt(GPoint pt)
{
    /* easy case */
    if (!config.AllowGravityWells)
        return 0;

    if (isPHost()) {
        /* PHost gravity wells */
        int sqs = square(config.GravityWellRange);
        for (int i = NUM_PLANETS; i > 0; --i) {
            GPlanet& p = getPlanet(i);
            if (p.isExistant()) {
                if (config.RoundGravityWells) {
                    if (distanceSquared(p.getPos(), pt) <= sqs)
                        return i;
                } else {
                    GPoint p2 = p.getPos().getNearestTo(pt);
                    if (square(p2.x - pt.x) <= sqs
                        || square(p2.x - pt.x) <= sqs)
                        return i;
                }
            }
        }
        return 0;
    } else {
        /* THost gravity wells: round, 3 ly, not wrapped, "cumulative" */
        int pid = 0;
        for (int i = 1; i <= NUM_PLANETS; ++i) {
            GPlanet& p = getPlanet(i);
            if (p.isExistant() && distanceSquaredRaw(p.getPos(), pt) <= 9) {
                pt = p.getPos();  // (!)
                pid = i;
            }
        }
        return pid;
    }
}

static string_t
formatPlanetName(GGameTurn& trn, int pid, int flags)
{
    return format(((flags & locs_Orbit)
                   ? ((flags & locs_Verbose)
                      ? _("Orbit of %s (Planet #%d)")
                      : _("Orbit of %s (#%d)"))
                   : ((flags & locs_Verbose)
                      ? _("%s (Planet #%d)")
                      : _("%s (#%d)"))),
                  trn.getPlanet(pid).getName(), pid);
}

/** Get name of a location in human-readable form.
    \param pt     location
    \param flags  details of requested string
    - locs_Ships: also use ship names as location names, not just planets
    - locs_WW: return "near PLANETNAME" if applicable
    - locs_Orbit: say "Orbit of PLANET" if applicable
    - locs_Verbose: be more verbose */
string_t
GGameTurn::getLocationName(GPoint pt, int flags)
{
    /* is it a bird? */
    int pid = getPlanetAt(pt);
    if (pid)
        return formatPlanetName(*this, pid, flags);

    /* is it a plane? */
    if (flags & locs_Ships) {
        int sid = getShipAt(pt);
        if (sid)
            return getShip(sid).getFullName();
    }
    if (flags & locs_WW) {
        pid = getGravityPlanetAt(pt);
        if (pid)
            return "near " + formatPlanetName(*this, pid, flags & ~locs_Orbit);
    }
    return format((flags & locs_Verbose) ? _("Deep Space (%d,%d)") : "(%d,%d)",
                  pt.x, pt.y);
}

/** Get ship at position. Any race does. This is mainly used for naming
    locations, and for things like to tell whether a "L" command would succeed.
    \returns Id number of a ship at position \c pt, or zero if none. */
int
GGameTurn::getShipAt(GPoint pt)
{
    pt = pt.getNormalized();
    for (int i = 1; i <= NUM_SHIPS; ++i)
        if (getShip(i).isExistant() && getShip(i).getPos() == pt)
            return i;
    return 0;
}

/** Count own ships at position.
    \returns number of ships played by the player which are at the
    specified position \c pt. */
int
GGameTurn::countMyShipsAt(GPoint pt)
{
    int n = 0;
    pt = pt.getNormalized();
    for (int i = 1; i <= NUM_SHIPS; ++i)
        if (getShip(i).isExistant() && getShip(i).getPos() == pt && isPlaying(getShip(i).getOwner()))
            ++n;
    return n;
}
