/**
  *  \file game/global.cc
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/global.h"
#include "game/ship.h"
#include "util/nls.h"
#include "game/player.h"
#include "util/console.h"
#include "util/except.h"
#include "game/planet.h"
#include "cpluslib/format.h"
#include "util/recode.h"

/** Global turn instance. */
Ptr<GGameTurn> current_turn_internal;
Ptr<GGameTurn> display_turn_internal;

/******************************* GGameTurn *******************************/

/*! \class GGameTurn
    This class contains everything that makes up a game turn: things
    that change each turn. There's one global instance of this class
    which is used by most code; some special-purpose code will work
    on different instances.

    This contains many functions that otherwise would be global,
    hence the huuuuuuuge interface.

    During a run, we expect all planet/starbase objects to be present,
    but possibly empty. To not have to do this during static
    initialisation (which isn't possible with Ptr<> anyway), one
    must call init() before loading the turn. Basically, do
    \code
        try {
            GGameTurn& t = getCurrentTurn();
            t.init();
            // load game...
            t.postprocess();
        }
        catch(std::exception& e) {
            // ...
        }
    \endcode
*/

GGameTurn::GGameTurn()
    : ship_type(0), planet_type(0), base_type(0),
      storm_type(0), playing_mask(0), data_mask(0),
      turn_nr(0)
{
    std::fill_n(ships,   countof(ships),   (GShip*) 0);
    std::fill_n(planets, countof(planets), (GPlanet*) 0);
    std::fill_n(player_data, countof(player_data), (Player*) 0);
    std::fill_n(storms,  countof(storms),  (GIonStorm*) 0);
    ship_type   = new GShipType(*this);
    planet_type = new GPlanetType(*this);
    base_type   = new GBaseType(*this);
    storm_type  = new GIonStormType(*this);
}

GGameTurn::~GGameTurn()
{
    done();
}

void
GGameTurn::init()
{
    for (int i = 1; i <= NUM_SHIPS; ++i)
        ships[i-1] = new GShip(i);
    for (int i = 1; i <= NUM_PLANETS; ++i)
        planets[i-1] = new GPlanet(i);
    for (int i = 1; i <= NUM_STORMS; ++i)
        storms[i-1] = new GIonStorm(i);
    inbox = new GInbox();
    vcrs  = new VcrDatabase();
    clearAllMines();
}

void
GGameTurn::done()
{
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        delete ships[i-1];
        ships[i-1] = 0;
    }
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        delete planets[i-1];
        planets[i-1] = 0;
    }
    for (int i = 1; i <= NUM_STORMS; ++i) {
        delete storms[i-1];
        storms[i-1] = 0;
    }
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        delete player_data[i-1];
        player_data[i-1] = 0;
    }
    inbox = 0;
    vcrs  = 0;
    clearAllMines();
}

void
GGameTurn::clearAllMines()
{
    for (MinefieldMap::iterator i = mines.begin(); i != mines.end(); ++i)
        delete i->second;
    MinefieldMap().swap(mines);
}

GGameTurn::Player&
GGameTurn::createPlayerData(int player)
{
    ASSERT(player > 0 && player <= NUM_PLAYERS);
    if (!player_data[player-1])
        player_data[player-1] = new Player(player);
    return *player_data[player-1];
}

void
GGameTurn::setGen(int player, const GGen& agen)
{
    getPlayerData(player).gen = agen;
    if (!turn_nr)
        turn_nr = agen.getTurnNr();
}

void
GGameTurn::setHaveData(int player)
{
    createPlayerData(player);
    data_mask |= (1 << player);
}

/** Load Ion Storm Names. */
void
GGameTurn::loadStormNames()
{
    Ptr<Stream> p = spec_file_dir->openFileNT("storm.nm", Stream::C_READ);
    if (!p) {
        /* dummy version */
        for (int i = 1; i <= NUM_STORMS; ++i)
            getIonStorm(i).setName(_("Noname Storm"));
    } else {
        char buffer[NUM_STORMS * 20];
        if (p->read(buffer, sizeof buffer) != sizeof buffer)
            throw FileTooShortException("storm.nm");
        Recoder::getE2I().recodeArray(buffer, sizeof buffer);
        for (int i = 1; i <= NUM_STORMS; ++i)
            getIonStorm(i).setName(getBasicStringN(&buffer[20 * (i-1)], 20));
    }
}

/** Load Planet Names. */
void
GGameTurn::loadPlanetNames()
{
    Ptr<Stream> p = spec_file_dir->openFileT("planet.nm", Stream::C_READ);
    char buffer[NUM_PLANETS * 20];
    if (p->read(buffer, sizeof buffer) != sizeof buffer)
        throw FileTooShortException("planet.nm");
    Recoder::getE2I().recodeArray(buffer, sizeof buffer);
    for (int i = 1; i <= NUM_PLANETS; ++i)
        getPlanet(i).setName(getBasicStringN(&buffer[20 * (i-1)], 20));
}

/** Load Planet Coordinates. */
void
GGameTurn::loadPlanetXY(Stream* s)
{
    char buffer[NUM_PLANETS * 6];
    if (s->read(buffer, sizeof buffer) != sizeof buffer)
        throw FileTooShortException(s->getName());
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        int x = getInt16(&buffer[6*i-6]);
        int y = getInt16(&buffer[6*i-4]);
        if (! GPlanet::isClippedCoord(GPoint(x, y)))
            getPlanet(i).setXY(x, y);
    }
}

/** Postprocess after loading. */
void
GGameTurn::postprocess()
{
    // FIXME: use real values
    playing_mask = data_mask;

    loadPlanetNames();
    loadStormNames();

    console.write(LOG_INFO, _("Compiling starchart..."));

    for (int i = 1; i < NUM_PLANETS; ++i)
        getPlanet(i).internalCheck();
    for (int i = 1; i < NUM_SHIPS; ++i)
        getShip(i).internalCheck();
    for (int i = 1; i < NUM_SHIPS; ++i)
        getShip(i).combinedCheck1(*this);
    for (int i = 1; i < NUM_PLANETS; ++i)
        getPlanet(i).combinedCheck2(*this);
    for (MinefieldMap::iterator i = mines.begin(); i != mines.end(); /* empty */) {
        i->second->internalCheck(*this);
        if (i->second->isExistant()) {
            ++i;
        } else {
            delete i->second;
            mines.erase(i++);
        }
    }

    ship_type->init();
    planet_type->init();
    base_type->init();
    storm_type->init();

    chart_bbox.initFromTurn(*this);

    console.write(LOG_INFO, _("=== Loading complete ==="));
    for (int i = 1; i <= NUM_PLAYERS; ++i)
        if (isPlaying(i))
            console.write(LOG_INFO, format(_("Playing player %d: %s"), i, getFullName(i)));
}

/** Get Minefield.
    \param id     minefield you want
    \param cause  cause for this request. MFA_xxx or -1.
    \post cause >= 0 ==> return value != NULL */
GMinefield*
GGameTurn::getMinefield(int id, int cause)
{
    MinefieldMap::iterator i = mines.find(id);
    if (i == mines.end()) {
        if (cause >= 0) {
            GMinefield* mf = new GMinefield(id, cause);
            mines.insert(std::make_pair(id, mf));
            return mf;
        } else {
            return 0;
        }
    } else {
        if (cause > i->second->getAction())
            i->second->setAction(cause);
        return i->second;
    }
}

void
GGameTurn::doUpdates()
{
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        GPlanet& p = getPlanet(i);
        if (p.getUIFlags() & p.ps_Dirty) {
            /* call sig_planet_changed in a loop, in case it causes the
               planet to be modified further. */
            int flags = 0;
            do {
                flags |= p.getUIFlags();
                sig_planet_changed.raise(i);
            } while (p.getUIFlags() & ~flags);
            p.setUIFlags(p.getUIFlags() & ~p.ps_Dirty);
        }
    }

    for (int i = 1; i <= NUM_SHIPS; ++i) {
        GShip& s = getShip(i);
        if (s.getUIFlags() & s.ss_Dirty) {
            int flags = 0;
            do {
                flags |= s.getUIFlags();
                sig_ship_changed.raise(i);
            } while (s.getUIFlags() & ~flags);
            s.setUIFlags(s.getUIFlags() & ~s.ss_Dirty);
        }
    }
}

GGameTurn::Player::Player(int pid)
    : outbox(new GOutbox(pid))
{ }

GGameTurn::Player::~Player()
{ }

/**************************** Global Functions ***************************/

Signal0 sig_turn_changed;

void
setDisplayedTurn(Ptr<GGameTurn> disp)
{
    GGameTurn* orig = display_turn_internal;
    if (disp)
        display_turn_internal = disp;
    else
        display_turn_internal = current_turn_internal;
    if (orig != display_turn_internal)
        sig_turn_changed.raise();
}

void
initTurn()
{
    current_turn_internal = new GGameTurn();
    setDisplayedTurn(0);
}

void
initGameGlobals(string_t gamedir, string_t rootdir)
{
    initDirectories(gamedir, rootdir);
    initTurn();
}
