/**
  *  \file global.h
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GLOBAL_H
#define PCC_V2_GLOBAL_H

#include <map>
#include "fileio/streams.h"
#include "fileio/dirs.h"
#include "game/vcrdb.h"
#include "game/msg.h"
#include "game/gen.h"
#include "game/mine.h"
#include "game/cmds.h"
#include "game/gamedirs.h"      // re-exported
#include "cpluslib/sigslot.h"
#include "cpluslib/assert.h"
#include "game/ion.h"

/**************************** Data Structures ****************************/

class GGameTurn;
class GShip;
class GPlanet;
class GShipType;
class GPlanetType;
class GBaseType;

//! Chart Bounding Box.
struct GChartBBox {
    enum {
        MAX_ZOOM = 10
    };

    GPoint min, max;
    int    max_zoom;
    int    min_zoom;

    void initFromTurn(GGameTurn& trn);
    void initMinZoom();
    void initMaxZoom();
};

/** Flags for getLocationName */
enum {
    locs_Orbit   = 1,
    locs_Verbose = 2,
    locs_Ships   = 4,
    locs_WW      = 8
};

/** Everything that represents a game turn. */
class GGameTurn {
    /* I'm still unsure what to do with inbox and VCRs */
    struct Player {
//         Ptr<GInbox>      inbox;
//         Ptr<VcrDatabase> vcrs;
        Ptr<GOutbox>      outbox;
        GGen              gen;
        GCommandContainer commands;
        Player(int pid);
        ~Player();
    };
 public:
    typedef std::map<int, GMinefield*> MinefieldMap;
 private:
    GShip*            ships[NUM_SHIPS];
    GPlanet*          planets[NUM_PLANETS];
    GIonStorm*        storms[NUM_STORMS];
    Ptr<GInbox>       inbox;
//     Ptr<GOutbox>      outbox;
//     GGen              gen;
    Ptr<VcrDatabase>  vcrs;
    MinefieldMap      mines;
    GShipType*        ship_type;
    GPlanetType*      planet_type;
    GBaseType*        base_type;
    GIonStormType*    storm_type;
    GChartBBox        chart_bbox;
    Player*           player_data[NUM_PLAYERS];
    int               playing_mask, data_mask;
    int               turn_nr;

    void clearAllMines();

    Player&    createPlayerData(int player);
    Player&    getPlayerDataNC(int player) const { return *player_data[player-1]; }
    Player&    getPlayerData(int player)         { return player_data[player-1] ? *player_data[player-1] : createPlayerData(player); }

    void loadPlanetNames();
    void loadStormNames();
public:
    GGameTurn();
    ~GGameTurn();

    void init();
    void done();

    /* "GEN" accessors */
    void         setGen(int player, const GGen& agen);
    GGen&        getGen(int player)       { return getPlayerData(player).gen; }
    const GGen&  getGen(int player) const { return getPlayerDataNC(player).gen; }

    int          getTurnNr() const { return turn_nr; }
    
    bool         isPlaying(int player_nr) const { return playing_mask & (1 << player_nr); }
    bool         isData(int player_nr)    const { return data_mask & (1 << player_nr); }
    void         setHaveData(int player_nr);

    /* Accessors */
    GShip&          getShip(int sid)           { ASSERT(sid > 0 && sid <= NUM_SHIPS); return *ships[sid-1]; }
    GPlanet&        getPlanet(int pid)         { ASSERT(pid > 0 && pid <= NUM_PLANETS); return *planets[pid-1]; }
    GIonStorm&      getIonStorm(int sid)       { ASSERT(sid > 0 && sid <= NUM_STORMS); return *storms[sid-1]; }
    GMinefield*     getMinefield(int id, int cause);
    GMinefield*     getMinefield(int id)       { return getMinefield(id, -1); }
    MinefieldMap::iterator getMinefieldBegin() { return mines.begin(); }
    MinefieldMap::iterator getMinefieldEnd()   { return mines.end(); }

    int              getPlanetAt(GPoint pt);
    int              getPlanetAt(GPoint pt, bool gravity);
    int              getShipAt(GPoint pt);
    int              getGravityPlanetAt(GPoint pt);
    int              countMyShipsAt(GPoint pt);
    string_t         getLocationName(GPoint pt, int flags);

    Ptr<GInbox>      getInbox() const { return inbox; }
    Ptr<GOutbox>     getOutbox(int player) const { return getPlayerDataNC(player).outbox; }
    Ptr<VcrDatabase> getVcrs() const  { return vcrs; }

    GShipType&       getShipType()    { return *ship_type; }
    GPlanetType&     getPlanetType()  { return *planet_type; }
    GBaseType&       getBaseType()    { return *base_type; }
    GIonStormType&   getIonStormType() { return *storm_type; }

    GChartBBox&      getChartBBox()   { return chart_bbox; }

    GCommandContainer& getCommands(int player) { return getPlayerDataNC(player).commands; }

    /* Loading */
    void loadPlanetXY(Stream*);

    void postprocess();

    /* Change Notification */
    void doUpdates();

    Signal1<int> sig_planet_changed, sig_ship_changed;
};

/** Everything that's static during turns. */
struct GGameStatic {
    // ship location information (trails) goes here
    // rationale: when chart displays future/past (not the global
    // GGameTurn object), it can still get the trails from here
};

/******************************* Accessors *******************************/

// These are exposed here only for the following inline functions
extern Ptr<GGameTurn> current_turn_internal;
extern Ptr<GGameTurn> display_turn_internal;

/** Get global turn instance. */
inline GGameTurn&  getCurrentTurn()   { return *current_turn_internal; }
inline GGameTurn&  getDisplayedTurn() { return *display_turn_internal; }

void setDisplayedTurn(Ptr<GGameTurn> disp);

extern Signal0 sig_turn_changed;

/** Global Data. */
void initTurn();
void initGameGlobals(string_t gamedir, string_t rootdir);

#endif
