/**
  *  \file game/load-dir.cc
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "fileio/io.h"
#include "game/global.h"
#include "game/load.h"
#include "game/player.h"
#include "game/struct.h"
#include "util/console.h"
#include "util/nls.h"
#include "util/except.h"
#include "cpluslib/format.h"

/* arbitrary limit */
#define MAX_MSG_SIZE 16000

/** Load Version-3.5 Message Outbox File (mess35x.dat).
    \param trn   Game turn record
    \param s     Stream, read pointer at position 0 */
static void
loadOutbox35(GGameTurn& trn, Stream* s, int player)
{
    /* file format is
       - word      number of messages
       - 17 bytes  garbage
       - N records: 1 byte    garbage
                    1 byte    "valid" flag
                    12 bytes  receivers
                    word      text length
                    N bytes   text (encrypted) */
    TInt16 n;
    char header[20];

    trn.getOutbox(player)->setFormat(true);

    /* read count */
    if (!getStructure(s, n) || n.value == 0)
        return;

    /* skip garbage */
    s->read(header, 17);
    while (n.value-- && s->read(header, 16) == 16) {
        uint16 msgsize = getUint16(header, 14);
        if (msgsize > MAX_MSG_SIZE)
            throw FileFormatException(format(_("File `%s' is invalid: message too big"),
                                             s->getName()));
        if (msgsize == 0)
            continue;

        Array<char> msgtext(msgsize);
        if (s->read(msgtext.ptr(), msgsize) != msgsize)
            throw FileFormatException(format(_("File `%s' is invalid: unexpected end of file"),
                                             s->getName()));

        if (header[1] == '1') {
            /* message is valid. Gather receivers */
            uint16 recv = 0;
            for (int i = 1; i <= 12; ++i)
                if (header[1+i] == '1')
                    recv |= 1 << i;
            addMessageFromFile(trn, getMessageFromArray(msgtext.ptr(), msgsize, false), player, recv);
        }
    }
}

/** Load version 3.0 (Dosplan) outbox file (messX.dat).
    \param s      stream, pointer at start of directory. */
static void
loadOutbox(GGameTurn& trn, Stream* s, int player)
{
    TInt16 n;

    trn.getOutbox(player)->setFormat(false);

    getStructure(s, n);
    if (n.value <= 0)
        return;

    std::size_t header_size = TOutgoingMessageHeader::size * n.value;
    Array<char> header(header_size);
    if (s->read(header.ptr(), header_size) != header_size)
        throw FileFormatException(format(_("File `%s' is invalid: directory truncated"), s->getName()));

    for (int i = 0; i < n.value; ++i) {
        TOutgoingMessageHeader o;
        getStructure(header.ptr() + TOutgoingMessageHeader::size * i, o);
        if (o.address <= 0)
            throw FileFormatException(format(_("File `%s' is invalid: invalid directory"), s->getName()));
        s->seek(o.address - 1);

        if (o.length < 0 || o.length > MAX_MSG_SIZE)
            throw FileFormatException(format(_("File `%s' is invalid: message too big"),
                                             s->getName()));
        if (o.length == 0)
            continue;

        Array<char> data(o.length);
        if (s->read(data.ptr(), o.length) != o.length)
            throw FileFormatException(format(_("File `%s' is invalid: unexpected end of file"),
                                             s->getName()));

        addMessageFromFile(trn, getMessageFromArray(data.ptr(), o.length, false),
                           player, 1 << o.to);
    }
}

static void
loadKore(GGameTurn& trn, Stream* s, int /*player*/)
{
    char buffer[50 * 12];

    /* Header */
    if (s->read(buffer, 102) != 102)
        throw FileFormatException(format(_("File `%s' is invalid: unexpected end of file"),
                                         s->getName()));
    if (getInt16(buffer, 0) != trn.getTurnNr()) {
        console.write(LOG_INFO, format(_("File `%s' is stale and will be ignored."),
                                       s->getName()));
        return;
    }

    /* Minefields */
    loadKoreMinefields(trn, s, 500);

    /* Ion storms */
    loadKoreIonStorms(trn, s, 50);

    /* FIXME: bangs */
    s->read(buffer, 50 * 4);

    /* FIXME: ufos */
    for (int i = 1; i <= 100; ++i)
        s->read(buffer, 78);

    /* visual contacts */
    if (s->read(buffer, 24) == 24 && std::memcmp(buffer, "1120", 4) == 0) {
        int32 count = getInt32(buffer, 20);
        if (count < 0 || count > NUM_SHIPS)
            throw FileFormatException(format(_("File `%s' is invalid: unbelievable number of visual contacts"),
                                      s->getName()));
        loadTargets(trn, s, count, true);
    }
}

/** Load unpacked game directory. */
void
loadDirectory(GGameTurn& trn, Ptr<Directory> d, int player)
{
    TInt16 n;
    TGen   gen;

    console.write(LOG_INFO, format(_("Loading %s data..."), getAdjName(player)));

    /* gen.dat */
    Ptr<Stream> s = d->openFileT(format("gen%d.dat", player), Stream::C_READ);
    if (!getStructure(s, gen) || gen.player_id != player)
        throw FileFormatException(format(_("File `%s' is invalid"), s->getName()));
    trn.setHaveData(player);
    trn.setGen(player, GGen(gen));

    /* ships */
    s = d->openFileT(format("ship%d.dat", player), Stream::C_READ);
    getStructure(s, n);
    loadShips(trn, s, n.value, WHERE_DAT, 1 << player);
    s = d->openFileT(format("ship%d.dis", player), Stream::C_READ);
    getStructure(s, n);
    loadShips(trn, s, n.value, WHERE_DIS, 1 << player);

    /* targets */
    s = d->openFileT(format("target%d.dat", player), Stream::C_READ);
    getStructure(s, n);
    loadTargets(trn, s, n.value, false);

    /* planets */
    s = d->openFileT(format("pdata%d.dat", player), Stream::C_READ);
    getStructure(s, n);
    loadPlanets(trn, s, n.value, WHERE_DAT, 1 << player);
    s = d->openFileT(format("pdata%d.dis", player), Stream::C_READ);
    getStructure(s, n);
    loadPlanets(trn, s, n.value, WHERE_DIS, 1 << player);

    /* starbases */
    s = d->openFileT(format("bdata%d.dat", player), Stream::C_READ);
    getStructure(s, n);
    loadBases(trn, s, n.value, WHERE_DAT, 1 << player);
    s = d->openFileT(format("bdata%d.dis", player), Stream::C_READ);
    getStructure(s, n);
    loadBases(trn, s, n.value, WHERE_DIS, 1 << player);

    /* messages */
    s = d->openFileT(format("mdata%d.dat", player), Stream::C_READ);
    trn.getInbox()->loadInbox(s);

    /* shipxy */
    s = d->openFileT(format("shipxy%d.dat", player), Stream::C_READ);
    loadShipXY(trn, s, 0, 0);

    /* VCRs */
    s = d->openFileT(format("vcr%d.dat", player), Stream::C_READ);
    trn.getVcrs()->loadFromFile(s);

    /* Outbox */
    s = d->openFileNT(format("mess35%d.dat", player), Stream::C_READ);
    if (s) {
        loadOutbox35(trn, s, player);
    } else {
        s = d->openFileNT(format("mess%d.dat", player), Stream::C_READ);
        if (s)
            loadOutbox(trn, s, player);
    }

    /* Kore */
    s = d->openFileNT(format("kore%d.dat", player), Stream::C_READ);
    if (s)
        loadKore(trn, s, player);

    // FIXME: kore, skore, ...
}
