/**
  *  \file game/load-rst.cc
  *  \brief RST Loading
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/load.h"
#include "game/rst.h"
#include "game/player.h"
#include "game/global.h"
#include "util/console.h"
#include "util/nls.h"
#include "util/except.h"
#include "cpluslib/format.h"

/** Load a RST file. */
void
loadResult(GGameTurn& trn, Ptr<Stream> s, int player)
{
    TInt16 n;
    TResultGen gen;

    console.write(LOG_INFO, format(_("Loading %s RST file..."),
                                   getAdjName(player)));
    
    GResultFile rst(s);

    s->seek(rst.getSectionOffset(GResultFile::SEC_GEN));
    getStructure(s, gen);
    if (gen.player_id != player)
        throw FileFormatException(format(_("File %s is owned by player %d, should be %d"),
                                         s->getName(), gen.player_id, player));
    trn.setHaveData(player);
    trn.setGen(player, GGen(gen));

    /* Ships */
    s->seek(rst.getSectionOffset(GResultFile::SEC_SHIPS));
    getStructure(s, n);
    loadShips(trn, s, n.value, WHERE_DAT | WHERE_DIS, 1 << gen.player_id);

    /* Targets */
    s->seek(rst.getSectionOffset(GResultFile::SEC_TARGETS));
    getStructure(s, n);
    loadTargets(trn, s, n.value, false);

    /* Planets */
    s->seek(rst.getSectionOffset(GResultFile::SEC_PLANETS));
    getStructure(s, n);
    loadPlanets(trn, s, n.value, WHERE_DAT | WHERE_DIS, 1 << gen.player_id);

    /* Starbases */
    s->seek(rst.getSectionOffset(GResultFile::SEC_BASES));
    getStructure(s, n);
    loadBases(trn, s, n.value, WHERE_DAT | WHERE_DIS, 1 << gen.player_id);

    /* Messages */
    s->seek(rst.getSectionOffset(GResultFile::SEC_MESSAGES));
    trn.getInbox()->loadInbox(s);

    /* SHIPXY */
    s->seek(rst.getSectionOffset(GResultFile::SEC_SHIPXY));
    loadShipXY(trn, s, (rst.getSectionOffset(GResultFile::SEC_SHIPXY+1)
                        - rst.getSectionOffset(GResultFile::SEC_SHIPXY)), 0);

    /* VCRs */
    s->seek(rst.getSectionOffset(GResultFile::SEC_VCRS));
    trn.getVcrs()->loadFromFile(s);

    /* Windows part of RST */
    if (rst.getVersion() >= 0) {
        /* KORE */
        // FIXME: we load only the minefields; as of 17/Feb/2002 we can't
        // handle anything else
        s->seek(rst.getSectionOffset(GResultFile::SEC_KORE));
        loadKoreMinefields(trn, s, 500);
        loadKoreIonStorms(trn, s, 50);

        s->seek(s->getPos() + 50*4 + 682 + 7800);

        char buffer[8];
        if (s->read(buffer, 8) == 8 && std::memcmp(buffer, "1120", 4) == 0) {
            int32 count = getInt32(buffer, 4);
            if (count < 0 || count > NUM_SHIPS)
                throw FileFormatException(format(_("File `%s' is invalid: unbelievable number of visual contacts"),
                                                 s->getName()));
            loadTargets(trn, s, count, true);
        }
    }
//         SEC_KORE,               ///< "KORE" (minefields, race names, storms, contacts, UFOs).
//         SEC_LEECH,              ///< "LEECH.DAT".
//         SEC_SKORE               ///< "SKORE" (more UFOs).
}
