/**
  *  \file game/load-trn.cc
  *  \brief Loading a TRN file
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/cmds.h"
#include "game/global.h"
#include "game/load.h"
#include "game/player.h"
#include "game/trn.h"
#include "util/console.h"
#include "util/nls.h"
#include "game/ship.h"
#include "game/planet.h"
#include "util/except.h"
#include "cpluslib/format.h"

/** Load turn file.
    \param trn    Game turn data
    \param s      Stream to load (see GTurnfile constructor)
    \param player Player number */
void
loadTurn(GGameTurn& trn, Ptr<Stream> s, int player)
{
    ASSERT(trn.isData(player));
    
    console.write(LOG_INFO, format(_("Loading %s turn file..."),
                                   getAdjName(player)));

    GTurnfile f(s);

    if (f.getFeatures() & GTurnfile::TRNF_TACCOM)
        console.write(LOG_INFO, format(_("Turn file contains %d attachment%!1{s%}"), f.getFileCount()));
    f.sortCommands();

    /* Pass 1: verify commands */
    for (int i = 0; i < f.getNumCommands(); ++i) {
        int cmd_class = f.getCommandType(i);
        int cmd_id    = f.getCommandId(i);
        switch (cmd_class) {
         case GTurnfile::UNDEF_COMMAND:
            throw FileFormatException(format(_("Turn file `%s' contains invalid command code %d"),
                                             s->getName(), f.getCommand(i)));
         case GTurnfile::SHIP_COMMAND:
            if (cmd_id <= 0 && cmd_id > NUM_SHIPS)
                throw FileFormatException(format(_("Turn file `%s' refers to non-existant ship %d"),
                                                 s->getName(), cmd_id));
            if ((trn.getShip(cmd_id).getSrcFlags() & (1 << player)) == 0)
                throw FileFormatException(format(_("Turn file `%s' refers to ship %d which is not ours"),
                                                 s->getName(), cmd_id));
            break;
         case GTurnfile::PLANET_COMMAND:
            if (cmd_id <= 0 && cmd_id > NUM_PLANETS)
                throw FileFormatException(format(_("Turn file `%s' refers to non-existant planet %d"),
                                                 s->getName(), cmd_id));
            if ((trn.getPlanet(cmd_id).getPlanetSrcFlags() & (1 << player)) == 0)
                throw FileFormatException(format(_("Turn file `%s' refers to planet %d which is not ours"),
                                                 s->getName(), cmd_id));
            break;
         case GTurnfile::BASE_COMMAND:
            if (cmd_id <= 0 && cmd_id > NUM_PLANETS)
                throw FileFormatException(format(_("Turn file `%s' refers to non-existant starbase %d"),
                                                 s->getName(), cmd_id));
            if ((trn.getPlanet(cmd_id).getBaseSrcFlags() & (1 << player)) == 0)
                throw FileFormatException(format(_("Turn file `%s' refers to starbase %d which is not ours"),
                                                 s->getName(), cmd_id));
            break;
         case GTurnfile::OTHER_COMMAND:
            ;
        }
    }

    /* Pass 2: process commands */
    string_t tim_allies;
    for (int i = 0; i < f.getNumCommands(); ++i) {
        int cmd_class = f.getCommandType(i);
        int cmd_id    = f.getCommandId(i);
        switch (cmd_class) {
         case GTurnfile::SHIP_COMMAND:
            trn.getShip(cmd_id).applyTurnCommand(f, i);
            if (f.getCommand(i) == tcm_ShipChangeFc && i+1 < f.getNumCommands()
                && f.getCommand(i+1) == tcm_ShipChangeFc && f.getCommandId(i+1) == cmd_id)
            {
                /* might be THost alliance command */
                const char* data = f.getCommandData(i);
                uint16 hack = getUint16(data);
                if (hack == 0x6666 /* ff */ || hack == 0x4646 /* FF */ || hack == 0x6565 /* ee */)
                    tim_allies.append(data, 3);
            }
            break;
         case GTurnfile::PLANET_COMMAND:
         case GTurnfile::BASE_COMMAND:
            trn.getPlanet(cmd_id).applyTurnCommand(f, i);
            break;
         case GTurnfile::OTHER_COMMAND:
            if (f.getCommand(i) == tcm_SendMessage) {
                const char* data = f.getCommandData(i);
                /* format of message is:
                     id     = length
                     data+0 = sender
                     data+2 = receiver
                     data+4 = text */
                int recv = getUint16(data+2);
                if (recv > 0 && recv <= 12)
                    addMessageFromFile(trn, getMessageFromArray(data+4, cmd_id, false),
                                       player, 1 << recv);
            } else if (f.getCommand(i) == tcm_ChangePassword) {
                trn.getGen(player).setPasswordFromTurn(f.getCommandData(i));
            }
        }
    }
    if (tim_allies.length())
        trn.getCommands(player).setCommand(GCommand::phc_TAlliance, 0, tim_allies);
}
