/**
  *  \file game/load.cc
  *  \brief Game Loading -- Common Code
  *
  *  This module contains routines for loading miscellaneous parts
  *  of the game data.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/load.h"
#include "util/except.h"
#include "util/nls.h"
#include "game/global.h"
#include "util/console.h"
#include "game/formula.h"
#include "game/cmds.h"
#include "game/ship.h"
#include "game/planet.h"
#include "cpluslib/format.h"

/** Load Ships. Loads SHIPx.DAT or the appropriate section from a RST.
    \param trn   Game Turn object
    \param s     Stream to read from
    \param count Number of ships to load
    \param where where to store data (WHERE_DAT, WHERE_DIS)
    \param asrc  Source flags for ship (1<<pid) */
void
loadShips(GGameTurn& trn, Ptr<Stream> s, int count, int where, int asrc)
{
    while (count > 0) {
        TShip ship;
        if (!getStructure(s, ship))
            throw FileTooShortException(s->getName());
        if (ship.ship_id <= 0 || ship.ship_id > NUM_SHIPS)
            throw FileFormatException(format(_("Invalid ship Id #%d"), ship.ship_id));
        if (where & WHERE_DIS)
            trn.getShip(ship.ship_id).addDisEntry(ship);
        if (where & WHERE_DAT)
            trn.getShip(ship.ship_id).addDatEntry(ship, asrc);
        --count;
    }
}

/** Load Planets. Loads PDATAx.DAT or the appropriate section from a RST.
    \param trn   Game Turn object
    \param s     Stream to read from
    \param count Number of planets to load
    \param where where to store data (WHERE_DAT, WHERE_DIS)
    \param asrc  Source flags for planet (1<<pid) */
void
loadPlanets(GGameTurn& trn, Ptr<Stream> s, int count, int where, int asrc)
{
    while (count > 0) {
        TPlanet planet;
        if (!getStructure(s, planet))
            throw FileTooShortException(s->getName());
        if (planet.planet_id <= 0 || planet.planet_id > NUM_PLANETS)
            throw FileFormatException(format(_("Invalid planet Id #%d"), planet.planet_id));
        if (where & WHERE_DIS)
            trn.getPlanet(planet.planet_id).addDisEntry(planet);
        if (where & WHERE_DAT)
            trn.getPlanet(planet.planet_id).addDatEntry(planet, asrc);
        --count;
    }
}

/** Load Starbases. Loads BDATAx.DAT or the appropriate section from a RST.
    \param trn   Game Turn object
    \param s     Stream to read from
    \param count Number of starbases to load
    \param where where to store data (WHERE_DAT, WHERE_DIS)
    \param asrc  Source flags for starbase (1<<pid) */
void
loadBases(GGameTurn& trn, Ptr<Stream> s, int count, int where, int asrc)
{
    while (count > 0) {
        TStarbase base;
        if (!getStructure(s, base))
            throw FileTooShortException(s->getName());
        if (base.base_id <= 0 || base.base_id > NUM_PLANETS)
            throw FileFormatException(format(_("Invalid starbase Id #%d"), base.base_id));
        if (where & WHERE_DIS)
            trn.getPlanet(base.base_id).addBaseDisEntry(base);
        if (where & WHERE_DAT)
            trn.getPlanet(base.base_id).addBaseDatEntry(base, asrc);
        --count;
    }
}

/** Load targets. Load TARGETx.DAT, TARGETx.EXT, or appropriate section
    from KOREx.DAT, UTILx.DAT or RST.
    \param trn   Game Turn object
    \param s     Stream to read from
    \param count Number of targets to read
    \param decode true iff ship names are scrambled. This is needed for
                 KOREx.DAT (and the appropriate RST section). */
void
loadTargets(GGameTurn& trn, Ptr<Stream> s, int count, bool decode)
{
    while (count > 0) {
        TShipTarget target;

        /* do not use getStructure(stream) here. We have to decrypt
           before recoding. */
        char buffer[TShipTarget::size];
        if (s->read(buffer, sizeof buffer) != sizeof buffer)
            throw FileTooShortException(s->getName());
        if (decode)
            for (int i = 0; i < 20; ++i)
                buffer[i+14] ^= 154-i;
        getStructure(buffer, target);

        if (target.ship_id <= 0 || target.ship_id > NUM_SHIPS) {
            console.write(LOG_ERROR,
                          format(_("Invalid ship Id #%d for visual contact -- ignored"), target.ship_id));
        } else {
            trn.getShip(target.ship_id).addTargetEntry(target);
        }
        --count;
    }
}

/** Load SHIPXY file.
    \param avail  Number of bytes known to be available.
    \param reject Reject entries of these players, bitfield (1<<pid).
                  This is used for allied file loading; to not wreak
                  havoc if player accidentially mixed up his files. */
void
loadShipXY(GGameTurn& trn, Ptr<Stream> s, int avail, int reject)
{
    static const int CHUNK_SIZE = 100;

    CTASSERT(NUM_SHIPS == 999);
    int nships = avail && avail >= 999*TShipXY::size ? 999 : 500;
    int id = 0;
    char buffer[TShipXY::size * CHUNK_SIZE];

    while (nships > 0) {
        /* Read file in chunks of CHUNK_SIZE */
        unsigned now = std::min(nships, CHUNK_SIZE);
        if (s->read(buffer, now * TShipXY::size) != now * TShipXY::size)
            throw FileTooShortException(s->getName());
        for (unsigned i = 0; i < now; ++i) {
            TShipXY entry;
            getStructure(&buffer[i * TShipXY::size], entry);
            ++id;
            /* Detect bogus files made by Winplan999/Unpack999 when used
               with Host500. The SHIPXY file continues with a (mangled)
               copy of GENx.DAT which results in unlikely high coordinates.
               Only test for ship #501, to keep the risk of false positives
               low (if someone actually goes that far -- it's not forbidden
               after all). Stupid "solution" for stupid problem. */
            if (id == 501 && (entry.x < 0 || entry.x >= 0x3030 || entry.owner >= 0x2020))
                return;

            if (entry.owner > 0 && entry.owner <= NUM_OWNERS && (reject & (1 << entry.owner)) == 0)
                trn.getShip(id).addShipXYEntry(entry);
        }
        nships -= now;
    }
}

/** Load Minefields from KORE-style file. */
void loadKoreMinefields(GGameTurn& trn, Ptr<Stream> s, int count)
{
    /* FIXME: probably should load all at once instead of
       one-at-a-time. */
    for (int i = 1; i <= count; ++i) {
        TKoreMine mf;
        if (!getStructure(s, mf))
            throw FileTooShortException(s->getName());
        if (mf.owner != 0) {
            /* Minefield was seen. If it is zero-size, do not let it
               come to existance if it is not already there. */
            int      action = mf.radius > 0 ? GMinefield::MFA_SCANNED : -1;
            GMinefield* pmf = trn.getMinefield(i, action);
            int32 new_units = square(mf.radius);
            int32 max_units = square(mf.radius + 1);
            int   new_owner = mf.owner == 12 ? 7 : mf.owner;
            if (pmf) {
                if (pmf->getOwner() == new_owner
                    && pmf->getPos() == GPoint(mf.x, mf.y)
                    && pmf->getTurn() == trn.getTurnNr()
                    && pmf->getUnits() >= new_units
                    && pmf->getUnits() < max_units)
                {
                    /* The minefield was already seen this turn, with
                       better information (exact unit count). */
                } else {
                    /* Update minefield */
                    if (new_owner == 7)
                        pmf->setData(mf.x, mf.y, new_owner, new_units, trn.getTurnNr(), mf.owner == 12);
                    else
                        pmf->setData(mf.x, mf.y, new_owner, new_units, trn.getTurnNr());
                }
            }
        }
    }
}

void
loadKoreIonStorms(GGameTurn& trn, Ptr<Stream> s, int count)
{
    for (int i = 1; i <= count; ++i) {
        TKoreStorm st;
        if (!getStructure(s, st))
            throw FileTooShortException(s->getName());
        if (st.volt > 0 && st.radius > 0)
            trn.getIonStorm(i).addKoreData(st);
    }
}

static string_t
extractCommands(GGameTurn& trn, string_t text, int player)
{
    string_t::size_type n = text.find_first_not_of(" ");
    if (n != text.npos && text[n] == '<')
        /* it's a genuine message. Don't parse. */
        return text;

    string_t s;
    do {
        string_t now = strFirst(text, "\n");
        if (GCommand::isMessageIntroducer(now))
            /* the rest is a message. */
            return s + text;
        if (GCommand* cmd = GCommand::parseCommand(now, false)) {
            trn.getCommands(player).setCommand(cmd);
        } else {
            s += now;
            s += '\n';
        }
    } while (strRemove(text, "\n"));

    string_t::size_type x = s.length();
    while (x > 0 && s[x-1] == '\n')
        --x;
    if (x < s.length())
        s.erase(x);
    return s;
}

void
addMessageFromFile(GGameTurn& trn, string_t text, int sender, uint16 recv)
{
    if (recv == (1 << sender)) {
        /* It's a message to us. Is it a command message? */
        text = extractCommands(trn, text, sender);
        if (text.empty())
            return;
    }

    trn.getOutbox(sender)->addMessageFromFile(text, recv);
}
