/**
  *  \file game/lock.cc
  *  \brief Locking onto Objects
  *
  *  This module provides functions for locking at objects (mouse click
  *  into starchart). In addition, it provides a means for configuring
  *  what mouse button locks onto what.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <climits>
#include "game/lock.h"
#include "game/global.h"
#include "game/ship.h"
#include "game/planet.h"

static int lock_config[lm_MAX] = {
    li_Planet | li_Minefield | li_Ufo,
    li_Ship | li_Marker
};

/** \class GLockData
    \brief Picking the nearest point of a set

    This class encapsulates the logic for picking the nearest point of
    a set. Instanciate it with the right parameters, invoke addPoint() for
    all points, and grab the result using getPoint(). */

/** Unbounded area.
    \param clicked point the user clicked on
    \param marked  true to only process marked points */
GLockData::GLockData(GPoint clicked, bool marked)
    : clicked(clicked), min(INT_MIN, INT_MIN), max(INT_MAX, INT_MAX),
      found(clicked), marked(marked), min_dist(LONG_MAX)
{ }

/** Bounded area.
    \param clicked point the user clicked on
    \param min     minimum coordinates
    \param max     maximum coordinates
    \param marked  true to only process marked points */
GLockData::GLockData(GPoint clicked, GPoint min, GPoint max, bool marked)
    : clicked(clicked), min(min), max(max), found(clicked), marked(marked), min_dist(LONG_MAX)
{ }

/** Add a point to the set. */
void
GLockData::addPoint(GPoint pt, bool marked)
{
    if (this->marked && !marked)
        return;
    if (pt.x < min.x || pt.y < min.y || pt.x > max.x || pt.y > max.y)
        return;

    long dist = distanceSquared(pt, clicked);
    if (dist < min_dist) {
        found    = pt.getNearestTo(clicked);
        min_dist = dist;
    }
}

void
findPlanet(GGameTurn& trn, GLockData& data)
{
    for (int pid = 1; pid <= NUM_PLANETS; ++pid) {
        GPlanet& p = trn.getPlanet(pid);
        if (p.isExistant())
            data.addPoint(p.getPos(), p.getUIFlags() & GPlanet::ps_Marked);
    }
}

void
findShip(GGameTurn& trn, GLockData& data)
{
    for (int sid = 1; sid <= NUM_SHIPS; ++sid) {
        GShip& s = trn.getShip(sid);
        if (s.isExistant())
            data.addPoint(s.getPos(), s.getUIFlags() & GShip::ss_Marked);
    }
}
    
void
findUfo(GGameTurn& /*trn*/, GLockData& /*data*/)
{
    /* FIXME */
}

void
findMarker(GGameTurn& /*trn*/, GLockData& /*data*/)
{
    /* FIXME */
}

void
findMinefield(GGameTurn& trn, GLockData& data)
{
    GGameTurn::MinefieldMap::iterator e = trn.getMinefieldEnd();
    for (GGameTurn::MinefieldMap::iterator i = trn.getMinefieldBegin(); i != e; ++i)
    {
        data.addPoint(i->second->getPos(), false);
    }
}

/** Configure locking. */
void
setLockMode(TLockMode mode, int items)
{
    lock_config[mode] = items;
}

int
getLockMode(TLockMode mode)
{
    return lock_config[mode];
}

/** Main entry point: do locking according to mode.
    \param trn    turn data to process
    \param data   GLockData object that does the work
    \param mode   corresponds to the mouse button the user pressed */
void
findItem(GGameTurn& trn, GLockData& data, TLockMode mode)
{
    int items = lock_config[mode];
    if (items & li_Planet)
        findPlanet(trn, data);
    if (items & li_Ship)
        findShip(trn, data);
    if (items & li_Ufo)
        findUfo(trn, data);
    if (items & li_Marker)
        findMarker(trn, data);
    if (items & li_Minefield)
        findMinefield(trn, data);
}
