/**
  *  \file game/mine.cc
  *  \brief Minefield Implementation
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "game/mine.h"
#include "game/global.h"
#include "util/nls.h"
#include "game/config.h"
#include "game/formula.h"
#include "cpluslib/format.h"

/** Create empty minefield with specified Id. */
GMinefield::GMinefield(int aid, int ac)
    : id(aid), x(0), y(0), owner(0), radius_now(0), units_now(0),
      web(false), units(0), turn(0), units_prev(0), last_action(ac)
{ }

/** Postprocessing after load. */
void
GMinefield::internalCheck(GGameTurn& trn)
{
    int32 u = units;
    for (int i = turn; i < trn.getTurnNr(); ++i)
        u = getUnitsAfterDecay(u);
    units_now = u;
    radius_now = int(std::sqrt(double(units_now)));
}

/** Compute mine decay for one turn. */
int32
GMinefield::getUnitsAfterDecay(int32 orig_units)
{
    int decay_rate = web ? config.WebMineDecayRate : config.MineDecayRate[playerIndex(owner)];
    if (isPHost())
        /* PHost formula */
        return orig_units * (100 - decay_rate) / 100;
    else
        /* THost formula (3.22.040). Actual formula is
             ERND(orig_units - orig_units*decay_rate/100) */
        return rdivaddl(orig_units * (100-decay_rate), 100, 0);
}

/** Set data. Use this to "assimilate" a minefield report into the
    database. */
void
GMinefield::setData(int ax, int ay, int aown, int32 aunits, int aturn)
{
    if (aturn != turn) {
        if (x == ax && y == ay) {
            turn_prev = turn;
            units_prev = units;
        } else {
            turn_prev = 0;
            units_prev = 0;
        }
    }
    x     = ax;
    y     = ay;
    owner = aown;
    units = aunits;
    turn  = aturn;
}

/** Set data. Use this to "assimilate" a minefield report into the
    database. This also sets the minefield type. */
void
GMinefield::setData(int ax, int ay, int aown, int32 aunits, int aturn, bool aweb)
{
    setData(ax, ay, aown, aunits, aturn);
    web = aweb;
}

string_t
GMinefield::getName() const
{
    if (web)
        return format(_("Web Mine Field #%d"), id);
    else
        return format(_("Mine Field #%d"), id);
}

string_t
GMinefield::getFullName() const
{
    return getName();
}

int
GMinefield::getId() const
{
    return id;
}

int
GMinefield::getOwner() const
{
    return owner;
}

bool
GMinefield::isExistant() const
{
    return units_now != 0;
}

GPoint
GMinefield::getPos() const
{
    return GPoint(x, y);
}

int
GMinefield::getRadius() const
{
    return radius_now;
}
