/**
  *  \file game/msg-out.cc
  *  \brief Message Outbox Handling
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "cpluslib/assert.h"
#include "cpluslib/bitops.h"
#include "cpluslib/types.h"
#include "game/msg.h"
#include "game/player.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "util/recode.h"

struct GOutbox::Message {
    string_t text;
    uint16   receivers;

    Message(const string_t text, uint16 rec)
        : text(text), receivers(rec) { }
};

/** Header line for a universal message. */
const char GOutbox::UNIVERSAL_TEXT[] = "  <<< Universal Message >>>";

/** Get "TO:" line for a receiver bitfield. */
static string_t
getReceiverText(uint16 bits)
{
    /* Note: do not translate "Host" here, because this function is
       also used to generate title lines for sent messages */
    bits &= GOutbox::ALL_RECEIVERS_MASK;
    if (bits == 0) {
        return _("Nobody");
    } else if (isPowerOfTwo(bits)) {
        /* one receiver */
        for (int i = 1; i <= 12; ++i)
            if (bits & (1 << i))
                return i == 12 ? string_t("Host") : getFullName(i);
        return "Huh?";
    } else {
        /* many receivers */
        string_t rv;
        for (int i = 1; i <= 12; ++i) {
            if (bits & (1 << i)) {
                if (rv.length())
                    rv += " ";
                if (i == 12)
                    rv += "Host";
                else
                    rv += itoa(i);
            }
        }
        return rv;
    }
}

/******************************** GOutbox ********************************/

/** Construct outbox.
    \param sender owner of this outbox: the player
                  whom all messages contained in this
                  messagebox belong to. */
GOutbox::GOutbox(int sender)
    : sender(sender), is_winplan(true)
{ }

GOutbox::~GOutbox()
{
    clear();
}

/** Discard all messages. */
void
GOutbox::clear()
{
    for (std::vector<Message*>::iterator i = messages.begin(); i != messages.end(); ++i)
        delete *i;
    messages.clear();
}

int
GOutbox::getCount()
{
    return messages.size();
}

/** Get text of message /index/ (0-based). The return value will
    include headers. */
string_t
GOutbox::getText(int index)
{
    ASSERT(index >= 0 && index < messages.size());

    uint16   receivers = messages[index]->receivers & ALL_RECEIVERS_MASK;
    string_t recv_text = getReceiverText(receivers);
    string_t text = format(_("FROM: %s\nTO: %s\n"),
                           getFullName(sender), recv_text);
    if ((receivers & UNIVERSAL_MASK) == UNIVERSAL_MASK) {
        text += UNIVERSAL_TEXT;
        text += '\n';
    } else if (!isPowerOfTwo(receivers)) {
        text += "CC: ";
        text += recv_text;
        text += '\n';
    }
    text += messages[index]->text;
    return text;
}

string_t
GOutbox::getHeading(int index)
{
    ASSERT(index >= 0 && index < messages.size());
    uint16 receivers = messages[index]->receivers & ALL_RECEIVERS_MASK;
    if ((receivers & UNIVERSAL_MASK) == UNIVERSAL_MASK)
        return "Universal Message";
    else
        return format(_("To: %s"), getReceiverText(receivers));
}

string_t
GOutbox::getRawText(int index)
{
    ASSERT(index >= 0 && index < messages.size());
    return messages[index]->text;
}

/** Get prefix for message when physically sent. The prefix can be
    carelessly concatenated to the message text.
    \param index    message number
    \param to       addressee of this incarnation of the message. */
string_t
GOutbox::getSendPrefix(int index, int to)
{
    ASSERT(index >= 0 && index < messages.size());
    uint16 receivers = messages[index]->receivers & ALL_RECEIVERS_MASK;

    /* to 11 players, or 11 players + host */
    if ((receivers & UNIVERSAL_MASK) == UNIVERSAL_MASK)
        return string_t(UNIVERSAL_TEXT) + '\n';

    receivers &= ~(1 << to);
    if (receivers) {
        const char* pfx = (sender == to ? "<CC: " : "CC: ");
        return pfx + getReceiverText(receivers) + '\n';
    }
    return string_t();
}

/** Get bitmask of receivers. 1<<1 means player 1, 1<<2 means player
    2, and so on up to 1<<12 for the host. */
uint16
GOutbox::getReceiverMask(int index)
{
    ASSERT(index >= 0 && index < messages.size());
    return messages[index]->receivers & ALL_RECEIVERS_MASK;
}

int
GOutbox::getStatus(int /*index*/)
{
    return 0;                   // FIXME?
}

void
GOutbox::setStatus(int /*index*/, int /*newstat*/)
{ }

/** Add message (send).
    \param text       message body without headers
    \param receivers  receiver bitmap (1 << player) */
void
GOutbox::addMessage(string_t text, uint16 receivers)
{
    messages.push_back(new Message(strRTrim(text), receivers));
}

static string_t
maybeStripHeaders(string_t msg, uint16 rec)
{
    /* strip headers only from unicast messages */
    if (!isPowerOfTwo(rec))
        return msg;

    string_t first_line = strFirst(msg, "\n");
    if (first_line == GOutbox::UNIVERSAL_TEXT || first_line.compare(0, 4, "CC: ", 4) == 0
        || first_line.compare(0, 5, "<CC: ", 5) == 0)
    {
        if (strRemove(msg, "\n"))
            return msg;
        else
            return string_t();  // silly people sending empty messages
    } else
        return msg;
}

/** Add message (from file). Unlike addMessage, this one attempts to
    merge multicast messages. */
void
GOutbox::addMessageFromFile(string_t text, uint16 receivers)
{
    /* attempt to merge messages. Preconditions:
       - message box contains at least one message
       - receivers don't overlap
       - message bodies are identical, sans headers */
    text = strRTrim(text);
    string_t raw_text = maybeStripHeaders(text, receivers);

    if (!messages.empty()
        && (messages.back()->receivers & receivers) == 0
        && (maybeStripHeaders(messages.back()->text, messages.back()->receivers) == raw_text))
    {
        /* merge */
        messages.back()->receivers |= receivers;
        messages.back()->text = raw_text;
    } else {
        /* don't merge */
        addMessage(text, receivers);
    }
}

void
GOutbox::deleteMessagesAfter(int index)
{
    if (index >= getCount())
        return;

    for (std::vector<Message*>::size_type m = index; m < messages.size(); ++m)
        delete messages[m];
    messages.erase(messages.begin() + index, messages.end());
}

void
GOutbox::deleteMessage(int index)
{
    ASSERT(index >= 0 && index < messages.size());

    delete messages[index];
    for (std::vector<Message*>::size_type n = index+1; n < messages.size(); ++n)
        messages[n-1] = messages[n];
    messages.pop_back();
}

/** Encode message, in-place. Recodes the message for output, and
    encrypts it. */
void
encodeMessage(string_t& text)
{
    for (string_t::size_type i = 0; i < text.length(); ++i)
        if (text[i] == '\n')
            text[i] = 26;
        else
            text[i] = Recoder::getI2E().recodeCharacter(text[i]) + 13;
    
    /* terminate with a carriage return to avoid that
       people make fun of Akseli :-) */
    if (text.length() && text[text.length()-1] != 26)
        text += char(26);
}
