/**
  *  \file game/objlist-sort.cc
  *  \brief Sorting of Object Lists
  *
  *  This module provides the sort-related parts of object lists.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/objlist.h"
#include "game/player.h"

namespace {
    struct SortHelper {
        int (*func)(GObject&,GObject&);
        bool operator()(GObjectId l, GObjectId r) {
            int res = func(l.getObject(), r.getObject());
            if (res == 0)
                return l.getId() < r.getId();
            else
                return res < 0;
        };
        SortHelper(int (*func)(GObject&,GObject&))
            : func(func) { }
    };
    struct SortIntHelper {
        int32 (*func)(GObject&);
        bool operator()(GObjectId l, GObjectId r) {
            int32 lv = func(l.getObject());
            int32 rv = func(r.getObject());
            return (lv < rv) || (lv == rv && l.getId() < r.getId());
        };
        SortIntHelper(int32 (*func)(GObject&))
            : func(func) { }
    };
}

/** Sort using a binary comparison function.
    \param func function returning a strcmp-like result (-1 if first comes
    before second, 0 if equal, 1 if comes after. Objects that compare equal
    will be sorted by Id. */
void
GObjectList::sort(int (*func)(GObject&,GObject&))
{
    removeDividers();
    std::sort(list.begin(), list.end(), SortHelper(func));
}

/** Sort using an unary weight function.
    \param func function returning a weight. Large weights are sorted
    towards the end, small weights come first. Equal weights are sorted
    by Id. */
void
GObjectList::sort(int32 (*func)(GObject&))
{
    removeDividers();
    std::sort(list.begin(), list.end(), SortIntHelper(func));
}

/***************************** Sort Functions ****************************/

/** Function for GObjectList::sort: sort by ascending Id number. */
int32
sortById(GObject& o)
{
    return o.getId();
}

/** Function for GObjectList::sort: sort by owner. */
int32
sortByOwner(GObject& o)
{
    return o.getOwner();
}

/** Function for GObjectList::sort: sort by name. */
int
sortByName(GObject& l, GObject& r)
{
    string_t left  = strUCase(l.getName());
    string_t right = strUCase(r.getName());
    return left.compare(right);
}

/*************************** Divider Functions ***************************/

/** Function for GObjectList::addDividers: show owners. Input should
    already be sorted by owner. */
string_t
diviOwner(GObject& o)
{
    return ::getShortName(o.getOwner());
}
