/**
  *  \file game/objlist.cc
  *  \brief Object Lists
  *
  *  This module contains the simple parts of object list management.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/objlist.h"
#include "cpluslib/assert.h"
#include "game/global.h"
#include "game/ship.h"
#include "game/planet.h"

/** \class GObjectId
    Handle to an object. This provides a handle to a game object, to be
    put in a list. It does not do any memory management / ownership
    management (therefore, copying these creatures is dirt cheap). Objects
    this refers to must be owned by someone else:
    - a GGameTurn instance (for most; users can only create references
      to these)
    - a GObjectList (for GDividerObject) */

/***************************** GDividerObject ****************************/

/** \class GDividerObject
    Divider in an Object List. This class should generally not be used
    by user code.

    A GDividerObject can only be owned by one GObjectList at a time;
    GObjectList attempts to make sure this holds.

    Divider objects have an Id number. Dividers with a nonzero Id can
    be selected in a selection list. */

GDividerObject::~GDividerObject()
{ }

string_t
GDividerObject::getName() const
{
    return name;
}

string_t
GDividerObject::getFullName() const
{
    return name;
}

int
GDividerObject::getId() const
{
    return id;
}

int
GDividerObject::getOwner() const
{
    return 0;
}

bool
GDividerObject::isExistant() const
{
    return true;
}

/****************************** GObjectList ******************************/

/** \class GObjectList
    List of Game Objects. You can add objects and dividers; we additionally
    provide functions to sort it. This class provides O(1) access to
    individual members. */

/** Make new, empty list. */
GObjectList::GObjectList()
{ }

GObjectList::~GObjectList()
{
    freeDividers();
}

/** Add all objects from a particular type. Type t should be sorted by
    Id; objects will be added in Id order. If type t is not sorted,
    all objects will be added but in an unpredictable order. */
void
GObjectList::addAllObjects(GObjectType& t)
{
    /* trivial case */
    if (!t.getCount())
        return;

    /* figure out lowest Id */
    int id = t.getCurrent();
    int prev;
    while ((prev = t.getPrev(id, false)) < id)
        id = prev;

    /* enumerate objects */
    int first = id;
    do {
        addObject(t, t.getObject(id));
        id = t.getNext(id, false);
    } while (id != first);
}

/** Add specified object.
    \param id  object; must be a game object */
void
GObjectList::addObject(GObjectId id)
{
    ASSERT(id.isGameObject());
    list.push_back(id);
}

/** Add specified object. */
void
GObjectList::addObject(GObjectType& type, GObject& obj)
{
    list.push_back(GObjectId(type, obj));
}

/** Add a divider line.
    \param s  divider text
    \param id divider identifier. Dividers with nonzero identifier
              can be selected in a selection list. Dividers with a
              zero Id are just that -- dividers. */
void
GObjectList::addDivider(const string_t& s, int id)
{
    GDividerObject* p = new GDividerObject(s, id);
    p->setParent(this);
    try {
        list.push_back(p);
    }
    catch(...) {
        delete p;
        throw;
    }
}

/** Add objects at a particular location. This is used for
    certain ship lists.
    \param trn   turn data
    \param pt    location
    \param flags ls_IncludePlanet: include planet in list;
                 ls_IncludeForeignShips: also list foreign ships
                 (default: only list ships we're playing)
    \param exclude_ship do not include this ship */
void
GObjectList::addObjectsAt(GGameTurn& trn, GPoint pt, int flags, int exclude_ship)
{
    if (flags & ls_IncludePlanet) {
        if (int pid = trn.getPlanetAt(pt))
            addObject(trn.getPlanetType(), trn.getPlanet(pid));
    }
    for (int sid = 1; sid <= NUM_SHIPS; ++sid) {
        GShip& s = trn.getShip(sid);
        if (sid != exclude_ship
            && s.isExistant()
            && s.getPos() == pt
            && ((flags & ls_IncludeForeignShips)
                || trn.isPlaying(s.getOwner())))
            addObject(trn.getShipType(), s);
    }
}

/** Remove dividers. */
void
GObjectList::freeDividers()
{
    for (container_t::iterator i = list.begin(); i != list.end(); ++i) {
        if (!i->isGameObject()) {
            GDividerObject* pdo = checked_cast<GDividerObject*>(&i->getObject());
            ASSERT(pdo->getParent() == this);
            delete pdo;
        }
    }
}

/** Gather dividers. This will set all "parent" fields of the dividers. */
void
GObjectList::gatherDividers()
{
    for (container_t::iterator i = list.begin(); i != list.end(); ++i) {
        if (!i->isGameObject()) {
            GDividerObject* pdo = checked_cast<GDividerObject*>(&i->getObject());
            ASSERT (!pdo->getParent() || pdo->getParent() == this);
            pdo->setParent(this);
        }
    }
}

/** Add dividers according a function. The passed function will return
    a string for every object; a heading (divider) will be added whenever
    that string changes. For example, when the list contains
    \example
       ...object...    ...func returns...
       Ship #1         Crystalline
       Ship #9         Crystalline
       Ship #3         Empire
       Ship #17        Crystalline
    \endexample
    the list will look like this afterwards ("==foo==" are dividers):
    \example
       == Crystalline ==
       Ship #1
       Ship #9
       == Empire ==
       Ship #3
       == Crystalline ==
       Ship #17
    \endexample */
void
GObjectList::addDividers(string_t (*func)(GObject&))
{
    container_t tmp;
    string_t    current;
    try {
        for (container_t::iterator i = list.begin(); i != list.end(); ++i) {
            if (i->isGameObject()) {
                string_t next = func(i->getObject());
                if (next.length() && current != next) {
                    current = next;
                    tmp.push_back(GObjectId(new GDividerObject(current)));
                }
                tmp.push_back(*i);
            }
        }
        freeDividers();
        tmp.swap(list);
        gatherDividers();
    }
    catch(...) {
        for (container_t::iterator i = list.begin(); i != list.end(); ++i)
            if (!i->isGameObject())
                delete checked_cast<GDividerObject*>(&i->getObject());
    }
}

/** Remove all dividers. */
void
GObjectList::removeDividers()
{
    /* in-place deletion */
    container_t::iterator o = list.begin();
    for (container_t::iterator i = o; i != list.end(); ++i) {
        if (i->isGameObject())
            *o = *i, ++o;
        else
            delete checked_cast<GDividerObject*>(&i->getObject());
    }
    list.erase(o, list.end());
}

void
GObjectList::clear()
{
    freeDividers();
    list.clear();
}

/** Return index for object id. Returns -1 if not found. */
int
GObjectList::getIndexFor(GObjectId id) const
{
    container_t::const_iterator o = list.begin();
    int i = 0;
    while (o != list.end()) {
        if (*o == id)
            return i;
        ++o, ++i;
    }
    return -1;
}

GObjectList::size_type
GObjectList::getCount() const
{
    container_t::const_iterator o = list.begin();
    size_type i = 0;
    while (o != list.end()) {
        if (o->isGameObject())
            ++i;
        ++o;
    }
    return i;
}
