/**
  *  \file game/objlist.h
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_OBJLIST_H
#define PCC_V2_GAME_OBJLIST_H

#include <vector>
#include "game/objtypes.h"
#include "game/obj.h"
#include "cpluslib/types.h"

/** Flags for ship list and GObjectList::addObjectsAt(). */
enum {
    /* addObjectsAt */
    ls_IncludeForeignShips = 1,
    ls_IncludePlanet       = 2,

    /* ship list */
    ls_AllowForeignShips   = 8,
    ls_AllowPlanet         = 16,
    ls_EarlyExit           = 32
};

class GObjectList;

class GDividerObject : public GObject {
    string_t      name;
    GObjectList*  parent;
    int           id;
 public:
    GDividerObject(const string_t& name)
        : name(name), parent(0), id(0) { }
    GDividerObject(const string_t& name, int id)
        : name(name), parent(0), id(id) { }
    ~GDividerObject();

    string_t getName() const;
    string_t getFullName() const;
    int      getId() const;
    int      getOwner() const;
    bool     isExistant() const;

    GObjectList* getParent() const { return parent; }
    void         setParent(GObjectList* par) { parent = par; }
};

/// Handle to an object.
class GObjectId {
    GObjectType* type;
    GObject*     object;
 private:
    friend class GObjectList;
    GObjectId(GDividerObject* pdo)
        : type(0), object(pdo) { }
 public:
    GObjectId(GObjectType& type, GObject& object)
        : type(&type), object(&object) { }
    GObjectId()
        : type(0), object(0) { }

    int      getId() const        { return object->getId(); }
    string_t getName() const      { return object->getFullName(); }
    string_t getClassName() const { return type->getClassName(); }

    /** True iff object is a game object (i.e. not a divider line or somesuch) */
    bool     isGameObject() const { return type != 0; }
    /** True iff this GObjectId has been initialized */
    bool     isValid() const      { return object != 0; }

    GObject& getObject() const    { return *object; }
    GObjectType& getType() const  { return *type; }

    bool operator==(const GObjectId& o) const { return type==o.type && object==o.object; }
    bool operator!=(const GObjectId& o) const { return type!=o.type || object!=o.object; }
};

class GObjectList {
 public:
    typedef std::vector<GObjectId> container_t, container_type;
    typedef container_t::size_type size_type;
    typedef container_t::const_iterator iterator;
 private:
    GObjectList(const GObjectList&);
    GObjectList& operator=(const GObjectList&);

    container_t list;
    void freeDividers();
    void gatherDividers();
 public:
    GObjectList();
    ~GObjectList();

    /* Adding things */
    void addAllObjects(GObjectType& t);
    void addObject(GObjectId id);
    void addObject(GObjectType& type, GObject& obj);
    void addDivider(const string_t& s, int id = 0);
    void addObjectsAt(GGameTurn& trn, GPoint pt, int flags, int exclude_ship = 0);

    /* User interface */
    // void addObjectsUser(GGameTurn& trn, GPoint pt, int flags, int exclude_ship = 0);
    void postprocessUser(bool with_dividers);

    /* Container interface */
    GObjectId operator[](container_t::size_type i) const { return list[i]; }
    size_type getSize() const { return list.size(); }
    size_type getCount() const;
    iterator  begin() const { return list.begin(); }
    iterator  end() const { return list.end(); }

    /* Manipulation */
    void      addDividers(string_t (*func)(GObject&));
    void      removeDividers();
    void      sort(int (*func)(GObject&,GObject&));
    void      sort(int32 (*func)(GObject&));
    void      clear();

    /* Accessors */
    int       getIndexFor(GObjectId id) const;
};

int32 sortById(GObject& o);
int32 sortByOwner(GObject& o);
int   sortByName(GObject& l, GObject& r);

string_t diviOwner(GObject& o);

#endif
