/**
  *  \file game/objtypes.cc
  *  \brief Object types
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/objtypes.h"
#include "cpluslib/assert.h"

/****************************** GObjectType ******************************/

/** \class GObjectType
    \brief Object Type

    A class derived from GObjectType defines a set of objects, and
    iteration through them. In addition, it contains a general
    interface for object selection. Control screens and various other
    features use these as a back-end.

    A GObjectType object also manages a "current position". */

/** Initialize.
    \param amax  Maximum object Id.
    \param atrn  Related game turn object. */
GObjectType::GObjectType(int amax, GGameTurn& atrn)
    : trn(atrn), current(0), count(0), max(amax)
{ }

GObjectType::~GObjectType()
{ }

/** Initialize. Call this after any change to the object set. */
void
GObjectType::init()
{
    count = 0;
    if (!isExistant(current))
        current = 0;
    for (int i = 1; i <= max; ++i) {
        if (isExistant(i)) {
            ++count;
            if (!current)
                current = i;
        }
    }
}

/** Get next object after /index/.
    \param mark iff true, only access marked objects
    \returns 0 if no object (only if count=0!)
    \default goes through objects in Id order. */
int
GObjectType::getNext(int index, bool mark)
{
    bool wrapped = false;
    while (1) {
        if (++index > max) {
            index = 0;
            if (wrapped)
                return 0;
            wrapped = 1;
        }
        if (isExistant(index) && (!mark || isMarked(index)))
            return index;
    }
}

/** Get previous object after /index/.
    \param mark iff true, only access marked objects
    \returns 0 if no object (only if count=0!)
    \default goes through objects in Id order. */
int
GObjectType::getPrev(int index, bool mark)
{
    bool wrapped = false;
    while (1) {
        if (--index <= 0) {
            index = max;
            if (wrapped)
                return 0;
            wrapped = 1;
        }
        if (isExistant(index) && (!mark || isMarked(index)))
            return index;
    }
}

/** Set position. */
void
GObjectType::setCurrent(int index)
{
    ASSERT(isExistant(index));
    if (current != index) {
        current = index;
        sig_pos_changed.raise(index);
    }
}
