/**
  *  \file game/planet-form.cc
  *  \brief Planet-related Formulas
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "game/planet.h"
#include "util/nls.h"
#include "cpluslib/misc.h"
#include "game/config.h"
#include "game/formula.h"

/** Get name for a temperature. */
const char*
GPlanet::getTempStr(int temp)
{
    if (temp <= 14)
        return _("arctic");
    else if (temp <= 39)
        return _("cool");
    else if (temp <= 64)
        return _("warm");
    else if (temp <= 84)
        return _("tropical");
    else
        return _("desert");
}

const char*
GPlanet::getNativeRaceName(int nr)
{
    static const char*const names[] = {
        N_("Humanoid"),
        N_("Bovinoid"),
        N_("Reptilian"),
        N_("Avian"),
        N_("Amorphous"),
        N_("Insectoid"),
        N_("Amphibian"),
        N_("Ghipsoldal"),
        N_("Siliconoid")
    };
    if (nr <= 0 || nr > int(countof(names)))
        return "???";
    else
        return _(names[nr-1]);
}

const char*
GPlanet::getNativeGovernmentName(int nr)
{
    static const char*const names[] = {
        N_("Anarchy"),
        N_("Pre-Tribal"),
        N_("Early-Tribal"),
        N_("Tribal"),
        N_("Feudal"),
        N_("Monarchy"),
        N_("Representative"),
        N_("Participatory"),
        N_("Unity")
    };
    if (nr <= 0 || nr > int(countof(names)))
        return "???";
    else
        return _(names[nr-1]);
}

const char*
GPlanet::getHappyStr(int happy)
{
    if (happy >= 90)
        return _("happy");
    else if (happy >= 70)
        return _("calm");
    else if (happy >= 50)
        return _("unhappy");
    else if (happy >= 40)
        return _("very angry");
    else if (happy >= 20)
        return _("rioting");
    else
        return _("fighting");
}

const char*
GPlanet::getChangeStr(int change)
{
    if (change < -5)
        return _("They HATE you!");
    else if (change < 0)
        return _("They are angry about you!");
    else if (change == 0)
        return _("They are undecided about you.");
    else if (change > 4)
        return _("They LOVE you.");
    else
        return _("They like your leadership.");
}

const char*
GPlanet::getBaseMissionName(int mission)
{
    switch (mission) {
     case 0:
        return _("none");
     case 1:
        return _("Refuel ships");
     case 2:
        return _("Maximize defense");
     case 3:
        return _("Load torpedoes onto ships");
     case 4:
        return _("Unload incoming ships");
     case 5:
        return _("Repair base");
     case 6:
        return _("Force surrender");
     default:
        return "???";
    }
}

static double
pdiv(int a, int b)
{
    if (isPHost())
        return double(a) / b;
    else
        return a / b;
}

/** Native happiness change.
    \param tax   desired tax rate
    \param mifa  mines plus factories on planet */
int
GPlanet::getNativeChange(int tax, int mifa) const
{
    /* formula is correct for PHost and THost */
    int change =
        int32(500 + 50 * getNativeGovernment()
              - 85 * tax
              - pdiv(mifa, 2)
              - std::sqrt(double(getNatives())))
        / 100;
    if (getNativeRace() == nr_Avian)
        change += 10;
    return change;
}

/** Native happiness change.
    \returns happiness change for current situation. */
int
GPlanet::getNativeChange() const
{
    return getNativeChange(getNativeTax(), getMines() + getFactories());
}

/** Colonists change.
    \param tax   desired tax rate
    \param mifa  mines plus factories on planet */
int
GPlanet::getColonistChange(int tax, int mifa) const
{
    double common = 1000 - 80*tax - std::sqrt(double(getColonists()));
    bool   crystal = (raceId(getOwner()) == 7) && config.CrystalsPreferDeserts;
    if (isPHost()) {
        if (crystal)
            return int32(common - mifa/3.0 - (100-getTemp())/0.66) / 100;
        else
            return int32(common - mifa/3.0 - std::abs(getTemp()-50)/0.33) / 100;
    } else {
        if (crystal)
            return int32(common - (mifa/3 + 3*(100-getTemp()))) / 100;
        else
            return int32(common - (mifa/3 + 3*std::abs(getTemp()-50))) / 100;
    }
}

/** Colonists happiness change on this planet.
    \returns change on this planet in current situation */
int
GPlanet::getColonistChange() const
{
    return getColonistChange(getColonistTax(), getMines() + getFactories());
}

/** Maximum number of structures.
    \param clans      clans on planet
    \param threshold  below threshold, growth is linear; above it's rootic :-) */
int
GPlanet::getMaxStructures(int32 clans, int threshold)
{
    if (clans <= threshold)
        return clans;
    else
        /* the rounding function is actually ERnd, but that never
           makes a difference; sqrt(INT) never ends in .5 exactly */
        return threshold + int(std::sqrt(double(clans - threshold)) + 0.5);
}

/** Maximum population on planet, for a race.
    \returns maximum population in clans. */
int32
GPlanet::getMaxSupportedColonists(int race) const
{
    bool crystal = (raceId(race) == 7 && config.CrystalsPreferDeserts);
    int32 limit;
    if (isPHost()) {
        if (!config.ClimateLimitsPopulation)
            return 250000;
        if (crystal) {
            if (config.CrystalSinTempBehavior)
                if (getTemp() >= 15)
                    return int32(100000 * std::sin(getTemp() * PI / 100));
                else
                    return 3 + getTemp() * config.MaxColTempSlope / 100;
            else
                return 1000 * getTemp();
        }

        if (getTemp() >= 85)
            limit = config.MaxColTempSlope * (100-getTemp()) / 100 + 1;
        else if (getTemp() <= 14)
            limit = config.MaxColTempSlope * getTemp() / 100 + 3;
        else
            limit = int32(100000 * std::sin(getTemp() * PI / 100));
    } else {
        if (!config.ClimateLimitsPopulation)
            return 100000;
        if (crystal)
            return 1000 * getTemp();

        if (getTemp() >= 85)
            limit = 2 * (100 - getTemp()) + 1;
        else if (getTemp() <= 14)
            limit = 2 * getTemp() + 3;
        else
            limit = int32(100000 * std::sin((100 - getTemp()) * 0.0314) + 0.5);
    }

    if (raceId(race) == 10 && getTemp() <= 19 && limit < 90000)
        limit = 90000;
    if ((raceId(race) == 4 || raceId(race) >= 9) && getTemp() >= 84 && limit < 60)
        limit = 60;

    return limit;
}
