/**
  *  \file game/planet.cc
  *  \brief Planet / Starbase Management
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/planet.h"
#include "util/console.h"
#include "util/nls.h"
#include "game/player.h"
#include "game/ship.h"
#include "game/global.h"
#include "cpluslib/format.h"
#include "game/config.h"

/******************************** GPlanet ********************************/

GPlanet::GPlanet(int aid)
    : id(aid), dat(0), dis(0), base_dat(0), base_dis(0), x(0), y(0),
      name("?"), flags(0), psrc_flags(0), bsrc_flags(0), ui_flags(0)
{ }

GPlanet::~GPlanet()
{
    delete dat;
    delete dis;
    delete base_dat;
    delete base_dis;
}

/** \name Loading */
//@{

/** Add .dat file entry. This assumes that when we see a planet
    through several .dat files, we get the same one each time.
    \param adat   parsed .dat file entry
    \param asrcf  source flag value to use for this entry */
void
GPlanet::addDatEntry(const TPlanet& adat, int asrcf)
{
    if (!dat)
        dat = new TPlanet(adat);
    psrc_flags |= asrcf;
}

/** Add .dis file entry. */
void
GPlanet::addDisEntry(const TPlanet& adis)
{
    if (!dis)
        dis = new TPlanet(adis);
}

/** Add starbase .dat file entry. See addDatEntry() for more
    information. */
void
GPlanet::addBaseDatEntry(const TStarbase& abdat, int asrcf)
{
    if (!base_dat)
        base_dat = new TStarbase(abdat);
    bsrc_flags |= asrcf;
}

/** Add starbase .dis file entry. */
void
GPlanet::addBaseDisEntry(const TStarbase& abdis)
{
    if (!base_dis)
        base_dis = new TStarbase(abdis);
}

/** Set planet position. */
void
GPlanet::setXY(int ax, int ay)
{
    x = ax;
    y = ay;
}

/** Set planet name. */
void
GPlanet::setName(const string_t& aname)
{
    name = aname;
}

//@}

//! \name Checking
//@{

/** Do internal checks for this planet. Internal checks do not require
    a partner to interact with. This will fix the problems, and
    display appropriate messages. */
void
GPlanet::internalCheck()
{
    flags = 0;
    /* set existance flag according to x/y */
    if (isClippedCoord(GPoint(x, y)))
        flags &= ~PLANET_EXISTS;
    else
        flags |= PLANET_EXISTS;
    
    if (dat) {
        /* we have a DAT entry. This planet MUST exist. */
        if (!(flags & PLANET_EXISTS)) {
            console.write(LOG_WARN,
                          format(_("Planet #%d exists for Host, but is outside valid range"), id));
            flags |= PLANET_EXISTS;
        }
    }

    if (base_dat) {
        /* we have a BDATA entry. There must also be a PDATA entry */
        if (!dat) {
            console.write(LOG_WARN,
                          format(_("Starbase #%d does not have a planet -- deleting it"), id));
            delete base_dat;
            delete base_dis;
            base_dat = base_dis = 0;
        }
    }
}

/** Combined checks, phase 2. This will do all post-processing which
    needs a partner to interact with. */
void
GPlanet::combinedCheck2(GGameTurn& trn)
{
    (void) trn;
}
//@}

//! @name Accessors
//@{

/** True iff this planet has a starbase. */
bool
GPlanet::hasBase() const
{
    return base_dat != 0;
}

void
GPlanet::setShipInOrbit(const GShip& s)
{
    int f = isFriend(s.getOwner());
    if (f == FOF_ME)
        flags |= PLANET_OWNORBIT;
    else if (f == FOF_FRIEND)
        flags |= PLANET_ALLYORBIT;
    else
        flags |= PLANET_ENEMYORBIT;
}

/** Check whether specified coordinates are clipped (i.e.\ nonexistant
    planet). There is no way for clients to tell that a planet is
    non-existant; the ad-hoc solution is moving planets far away. 
    Consistent with PCC 1.x handling of this. */
bool
GPlanet::isClippedCoord(GPoint pt)
{
    if (pt.x <= 0 || pt.y <= 0 || pt.x > 9000 || pt.y > 9000 || !pt.isOnMap())
        return true;
    else
        return false;
}

string_t
GPlanet::getName() const
{
    return name;
}

string_t
GPlanet::getFullName() const
{
    return format(_("%s (Planet #%d)"), name, id);
}

int
GPlanet::getId() const
{
    return id;
}

int
GPlanet::getOwner() const
{
    if (dat)
        return dat->owner;
    else
        return 0;
}

bool
GPlanet::isExistant() const
{
    return flags & PLANET_EXISTS;
}

GPoint
GPlanet::getPos() const
{
    return GPoint(x, y);
}

string_t
GPlanet::getFCode() const
{
    return getBasicString(dat->fcode);
}

void
GPlanet::setFCode(const string_t& str)
{
    storeBasicString(dat->fcode, str);
    ui_flags |= ps_DirtyData;
}

/** Get cargo of a particular type, as to be used by most code.
    This will remove reservations from the amount reported. */
int32
GPlanet::getCargo(GCargoType type) const
{
    return getCargoRaw(type) - cargo.getReserved(type);
}

/** Get cargo, as it is in underlying data structure. */
int32
GPlanet::getCargoRaw(GCargoType type) const
{
    switch(type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        return dat->ore_mined[type];
     case el_Supplies:
        return dat->supplies;
     case el_Money:
        return dat->money;
     case el_Fighters:
        if (base_dat)
            return base_dat->ammo_store[10];
        return 0;
     case el_Colonists:
        return dat->colonists;
     default:
        ASSERT(type >= el_Torps && type < el_Torps + NUM_TORPS);
        if (base_dat)
            return base_dat->ammo_store[type];
        return 0;
    }
}

void
GPlanet::changeCargoRaw(GCargoType type, int32 amount)
{
    if (!amount)
        return;
    ui_flags |= ps_DirtyData;
    
    switch(type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        dat->ore_mined[type] += amount;
        return;
     case el_Supplies:
        dat->supplies += amount;
        return;
     case el_Money:
        dat->money += amount;
        return;
     case el_Fighters:
        ASSERT(base_dat);
        base_dat->ammo_store[10] += amount;
        return;
     case el_Colonists:
        dat->colonists += amount;
        return;
     default:
        ASSERT(base_dat);
        ASSERT(type >= el_Torps && type < el_Torps + NUM_TORPS);
        base_dat->ammo_store[type] += amount;
        return;
    }
}

void
GPlanet::setBaseMission(int mission)
{
    if (mission != getBaseMission()) {
        base_dat->mission = mission;
        ui_flags |= ps_DirtyData;
    }
}

void
GPlanet::setBaseShipyardOrder(int action, int id)
{
    if (action != getBaseShipyardOrder() || id != getBaseShipyardId()) {
        base_dat->shipyard_action = action;
        base_dat->shipyard_id     = id;
        ui_flags |= ps_DirtyData;
    }
}

void
GPlanet::setBaseTech(TTechLevel level, int value)
{
    ASSERT(value > 0 && value <= 10);
    if (value != base_dat->tech_levels[level]) {
        base_dat->tech_levels[level] = value;
        ui_flags |= ps_DirtyData;
    }
}

int
GPlanet::getBaseBuildHull() const
{
    if (base_dat->build.hull)
        return getTruehull(base_dat->owner, base_dat->build.hull);
    else
        return 0;
}

/** Get ship which is trying to clone at this planet. This runs
    through the ship list, so use with care.
    \pre this planet is on the displayed turn */
int
GPlanet::getShipCloningHere() const
{
    // FIXME: check whether 'cln' works
    if (!config.AllowShipCloning)
        return 0;
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        GShip& s = getDisplayedTurn().getShip(i);
        if (s.isExistant() && s.isCloningAt(*this))
            return i;
    }
    return 0;
}

/****************************** GPlanetType ******************************/

GPlanetType::GPlanetType(GGameTurn& atrn)
    : GObjectType(NUM_PLANETS, atrn)
{ }

bool
GPlanetType::isExistant(int index)
{
    if (index > 0 && index <= NUM_PLANETS) {
        GPlanet& p = trn.getPlanet(index);
        if (p.isExistant() && trn.isPlaying(p.getOwner()) && p.haveDat())
            return true;
    }
    return false;
}

bool
GPlanetType::isMarked(int index)
{
    return trn.getPlanet(index).getUIFlags() & GPlanet::ps_Marked;
}

void
GPlanetType::toggleMark(int index)
{
    GPlanet& p = trn.getPlanet(index);
    p.setUIFlags((p.getUIFlags() ^ GPlanet::ps_Marked) | GPlanet::ps_DirtyHead);
}

GPlanet&
GPlanetType::getObject(int index)
{
    return trn.getPlanet(index);
}

int
GPlanetType::getDirtyFlags(int index)
{
    return trn.getPlanet(index).getUIFlags() & GPlanet::ps_Dirty;
}

string_t
GPlanetType::getName(int index)
{
    return trn.getPlanet(index).getFullName();
}

string_t
GPlanetType::getClassName()
{
    return string_t(_("Planet"));
}

/******************************* GBaseType *******************************/

GBaseType::GBaseType(GGameTurn& atrn)
    : GObjectType(NUM_PLANETS, atrn)
{ }

bool
GBaseType::isExistant(int index)
{
    if (index > 0 && index <= NUM_PLANETS) {
        GPlanet& p = trn.getPlanet(index);
        if (p.isExistant() && trn.isPlaying(p.getOwner()) && p.hasBase())
            return true;
    }
    return false;
}

bool
GBaseType::isMarked(int index)
{
    return trn.getPlanet(index).getUIFlags() & GPlanet::ps_Marked;
}

void
GBaseType::toggleMark(int index)
{
    GPlanet& p = trn.getPlanet(index);
    p.setUIFlags((p.getUIFlags() ^ GPlanet::ps_Marked) | GPlanet::ps_DirtyHead);
}

GPlanet&
GBaseType::getObject(int index)
{
    return trn.getPlanet(index);
}

int
GBaseType::getDirtyFlags(int index)
{
    return trn.getPlanet(index).getUIFlags() & GPlanet::ps_Dirty;
}

string_t
GBaseType::getName(int index)
{
    return trn.getPlanet(index).getFullName();
}

string_t
GBaseType::getClassName()
{
    return string_t(_("Starbase"));
}

/**************************** GPlanetTransfer ****************************/

GPlanetTransfer::GPlanetTransfer(GPlanet& planet)
    : planet(planet)
{
    planet.getCargoHandler().incUseCount();
}

GPlanetTransfer::~GPlanetTransfer()
{
    planet.getCargoHandler().decUseCount();
}

int32
GPlanetTransfer::getAmount(CargoType type)
{
    return planet.getCargo(type);
}

void
GPlanetTransfer::changeAmount(CargoType type, int32 amount)
{
    planet.changeCargoRaw(type, amount);
}

void
GPlanetTransfer::reserveAmount(CargoType type, int32 amount)
{
    planet.getCargoHandler().changeReservation(type, amount);
}

bool
GPlanetTransfer::canHave(CargoType cargo_id)
{
    if (cargo_id == el_Fighters
        || (cargo_id >= el_Torps && cargo_id < el_Torps + NUM_TORPS))
    {
        if (!planet.hasBase())
            return false;
    }
    return true;
}

int32
GPlanetTransfer::getMax(CargoType cargo_id)
{
    /* the torp/ore limits are arbitrary, but should match Tim's TRN check. */
    if (cargo_id == el_Fighters)
        return 60;
    else if (cargo_id >= el_Torps && cargo_id < el_Torps + NUM_TORPS)
        return 10000;
    else
        return 1000000000;
}

string_t
GPlanetTransfer::getName() const
{
    // FIXME: PCC 1.x says "$NAME starbase" or "planet $NAME" here.
    // should we use same wording?
    return planet.getName();
}
