/**
  *  \file game/planet.h
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_PLANET_H
#define PCC_V2_GAME_PLANET_H

#include "game/struct.h"
#include "game/coord.h"
#include "game/obj.h"
#include "game/objtypes.h"
#include "game/cargo.h"

class GShip;
class GGameTurn;
class GTurnfile;

enum {
    nr_None,
    nr_Humanoid,
    nr_Bovinoid,
    nr_Reptilian,
    nr_Avian,
    nr_Amorphous,
    nr_Insectoid,
    nr_Amphibian,
    nr_Ghipsoldal,
    nr_Siliconoid
};

/// Shipyard actions. Value of the shipyard_action field.
enum {
    sya_None,                   ///< No operation.
    sya_Fix,                    ///< Fix a ship.
    sya_Recycle                 ///< Recycle a ship.
};

class GPlanet : public GMapObject {
 private:
    GPlanet();
    GPlanet(const GPlanet&);
    GPlanet& operator=(const GPlanet&);

    static int getMaxStructures(int32 clans, int threshold);
 public:
    enum {
        PLANET_DAT         = 1,   ///< We have a DAT entry.
        PLANET_DIS         = 2,   ///< We have a DIS entry.
        PLANET_EXISTS      = 4,   ///< This planet exists.
        PLANET_OWNORBIT    = 8,   ///< Own ship(s) in orbit.
        PLANET_ALLYORBIT   = 16,  ///< Allied ship(s) in orbit.
        PLANET_ENEMYORBIT  = 32,  ///< Enemy ship(s) in orbit.
        PLANET_GALLYORBIT  = 64,  ///< Guessed allied ship(s) in orbit.
        PLANET_GENEMYORBIT = 128  ///< Guessed enemy ship(s) in orbit.
    };
 private:
    int id;
    TPlanet* dat;
    TPlanet* dis;
    TStarbase* base_dat;
    TStarbase* base_dis;
    int x, y;
    string_t name;
    int flags;
    /** Source flags. Bitfields (1<<player) of who's .dat files the
        information is from, separate for planets and bases. Used for
        saving. */
    int psrc_flags, bsrc_flags;
    int ui_flags;
    GCargo cargo;               ///< Reservation management.
 public:
    GPlanet(int id);
    ~GPlanet();

    void addDatEntry(const TPlanet& adat, int asrcf);
    void addDisEntry(const TPlanet& adis);
    void addBaseDatEntry(const TStarbase& abdat, int asrcf);
    void addBaseDisEntry(const TStarbase& abdis);
    void setXY(int x, int y);
    void setName(const string_t& name);

    const TPlanet& getRawData() const { return *dat; }
    const TStarbase& getRawBaseData() const { return *base_dat; }

    bool haveDat() const { return dat; }
    bool haveBaseDat() const { return base_dat; }
    int  getPlanetSrcFlags() const { return psrc_flags; }
    int  getBaseSrcFlags() const { return bsrc_flags; }
    void applyTurnCommand(GTurnfile& trn, int index);
    void makePlanetTurnCommands(GTurnfile& trn);
    void makeBaseTurnCommands(GTurnfile& trn);

    /* GMapObject interface */
    string_t getName() const;
    string_t getFullName() const;
    int      getId() const;
    int      getOwner() const;
    bool     isExistant() const;
    GPoint   getPos() const;

    /* Accessors */
    bool hasBase() const;
    int getStatus() const { return flags; }

    // FIXME: these should be made smarter
    int   getColonistChange(int tax, int mifa) const;
    int   getColonistChange() const;
    int   getColonistHappiness() const   { return dat->colonist_happy; }
    int   getColonistTax() const         { return dat->colonist_tax; }
    int32 getColonists() const           { return dat->colonists - cargo.getReserved(el_Colonists); }
    int   getDefense() const             { return dat->defense; }
    int   getFactories() const           { return dat->factories; }
    int   getMaxFactories() const        { return getMaxFactories(getColonists()); }
    int   getMaxMines() const            { return getMaxMines(getColonists()); }
    int   getMaxDefense() const          { return getMaxDefense(getColonists()); }
    int32 getMaxSupportedColonists(int race) const;
    int32 getMaxSupportedColonists() const { return getMaxSupportedColonists(getOwner()); }
    int   getMines() const               { return dat->mines; }
    int32 getMoney() const               { return dat->money - cargo.getReserved(el_Money); }
    int   getNativeChange() const;
    int   getNativeChange(int tax, int mifa) const;
    int   getNativeGovernment() const    { return dat->native_gov; }
    int   getNativeHappiness() const     { return dat->native_happy; }
    int   getNativeRace() const          { return dat->native_race; }
    int   getNativeTax() const           { return dat->native_tax; }
    int32 getNatives() const             { return dat->natives; }
    // int32 getOre(int index) const        { return dat->ore_mined[index] - cargo.getReserved(GCargo::CargoType(index)); }
    int   getOreDensity(int index) const { return dat->ore_density[index]; }
    int32 getOreGround(int index) const  { return dat->ore_ground[index]; }
    int32 getSupplies() const            { return dat->supplies - cargo.getReserved(el_Supplies); }
    int   getTemp() const                { return 100 - dat->temp_code; }
    bool  isBuildingBase() const         { return dat->build_base; }

    string_t getFCode() const;
    void     setFCode(const string_t& str);

    int32 getCargo(GCargoType type) const;
    int32 getCargoRaw(GCargoType type) const;

    GCargo& getCargoHandler()            { return cargo; }

    const char* getTempStr() const       { return getTempStr(getTemp()); }

    int   getBaseDefense() const         { return base_dat->defense; }
    int   getBaseDamage()  const         { return base_dat->damage; }
    int   getBaseMission() const         { return base_dat->mission; }
    const char* getBaseMissionName() const { return getBaseMissionName(getBaseMission()); }
    int   getBaseShipyardOrder() const   { return base_dat->shipyard_action; }
    int   getBaseShipyardId() const      { return base_dat->shipyard_id; }

    void  setBaseMission(int mission);
    void  setBaseShipyardOrder(int action, int id);

    int   getBaseTech(TTechLevel level) const { return base_dat->tech_levels[level]; }
    void  setBaseTech(TTechLevel level, int value);

    int   getBaseBuildHull() const;
    int   getShipCloningHere() const;
    // int16 engine_tech;
    // int16 hull_tech;
    // int16 beam_tech;
    // int16 torp_tech;
    // int16 engine_store[9];
    // int16 hull_store[20];
    // int16 beam_store[10];
    // int16 launcher_store[10];
    // int   getBaseTorps(int index) const  { return base_dat->ammo_store[index-1] ** RESERVATION; }
    // int   getBaseFighters() const        { return base_dat->ammo_store[10] ** RESERVATION; }
    // TBuildOrder build;

    /* Setters */
    void changeCargoRaw(GCargoType type, int32 amount);

    void internalCheck();
    void combinedCheck2(GGameTurn& trn);

    void setShipInOrbit(const GShip& s);

    int  getUIFlags() const { return ui_flags; }
    void setUIFlags(int fl) { ui_flags = fl; }

    static bool isClippedCoord(GPoint p);

    enum {
        ps_Marked    = 1,
        ps_DirtyHead = 2,
        ps_DirtyData = 4,

        ps_Dirty     = ps_DirtyHead | ps_DirtyData
    };

    static const char* getTempStr(int temp);
    static const char* getNativeRaceName(int race);
    static const char* getNativeGovernmentName(int gov);
    static const char* getHappyStr(int happy);
    static const char* getChangeStr(int change);
    static const char* getBaseMissionName(int mission);

    static int getMaxFactories(int32 clans) { return getMaxStructures(clans, 100); }
    static int getMaxMines(int32 clans)     { return getMaxStructures(clans, 200); }
    static int getMaxDefense(int32 clans)   { return getMaxStructures(clans, 50); }

    enum Facility {
        pf_TechLevels,
        pf_Structures
    };
};

class GPlanetType : public GObjectType {
 public:
    GPlanetType(GGameTurn& atrn);

    bool isExistant(int index);
    bool isMarked(int index);
    void toggleMark(int index);
    GPlanet& getObject(int index);

    int getDirtyFlags(int index);
    string_t getName(int index);
    string_t getClassName();

    GPlanet& getPlanet() { return getObject(getCurrent()); }
};

class GBaseType : public GObjectType {
 public:
    GBaseType(GGameTurn& atrn);

    bool isExistant(int index);
    bool isMarked(int index);
    void toggleMark(int index);
    GPlanet& getObject(int index);

    int getDirtyFlags(int index);
    string_t getName(int index);
    string_t getClassName();

    GPlanet& getPlanet() { return getObject(getCurrent()); }
};

class GPlanetTransfer : public GCargoContainer {
    GPlanet& planet;
 protected:
    int32  getAmount(CargoType type);
    void   changeAmount(CargoType type, int32 amount);
    void   reserveAmount(CargoType type, int32 amount);
 public:
    GPlanetTransfer(GPlanet& planet);
    ~GPlanetTransfer();

    bool     canHave(CargoType cargo_id);
    int32    getMax(CargoType cargo_id);
    string_t getName() const;
};

#endif
