/**
  *  \file game/player.cc
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <string>
#include "game/player.h"
#include "game/global.h"
#include "util/except.h"
#include "fileio/io.h"
#include "gfx/palette.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "util/recode.h"

char player_chars[] = "0123456789ABX";

/** Definition of the race.nm file. */
enum {
    RN_LONG_ORIG  = 0,
    RN_SHORT_ORIG = RN_LONG_ORIG + 30*11,
    RN_ADJ_ORIG   = RN_SHORT_ORIG + 20*11,
    RN_FILE_SIZE  = RN_ADJ_ORIG + 12*11
};

GRaceNameList racename_data;

/***************************** GRaceNameList *****************************/

GRaceNameList::GRaceNameList(Stream* s)
{
    if (s)
        load(s);
}

GRaceNameList::~GRaceNameList()
{ }

void
GRaceNameList::load(Stream* s)
{
    char buffer[RN_FILE_SIZE];
    if (s->read(buffer, sizeof(buffer)) != sizeof(buffer))
        throw FileFormatException(format(_("File `%s' is invalid"), s->getName()));
    Recoder::getE2I().recodeArray(buffer, sizeof buffer);

    for(int i = 1; i <= 11; ++i) {
        long_names[i]  = getBasicStringN(&buffer[RN_LONG_ORIG  + 30*(i-1)], 30);
        short_names[i] = getBasicStringN(&buffer[RN_SHORT_ORIG + 20*(i-1)], 20);
        adjectives[i]  = getBasicStringN(&buffer[RN_ADJ_ORIG   + 12*(i-1)], 12);
    }

    long_names[0]  = _("Nobody");
    short_names[0] = _("Nobody");
    adjectives[0]  = _("unowned");

    long_names[12]  = _("The Alien Marauder Alliance");
    short_names[12] = _("Alien Marauders");
    adjectives[12]  = _("Alien");
}

/** Get short race name. Returns the short (20 character) name of
    the specified player. Also handles unowned (0) and alien (12). */
const string_t&
GRaceNameList::getShortName(int player)
{
    ASSERT(player >= 0 && player <= 12);
    return short_names[player];
}

/** Get full race name. Returns the full (30 character) name of
    the specified player. Also handles unowned (0) and alien (12). */
const string_t&
GRaceNameList::getFullName(int player)
{
    ASSERT(player >= 0 && player <= 12);
    return long_names[player];
}

/** Get race adjective. Returns the adjective (12 characters) of the
    specified player. Also handles unowned (0) and alien (12). */
const string_t&
GRaceNameList::getAdjName(int player)
{
    ASSERT(player >= 0 && player <= 12);
    return adjectives[player];
}

/** Expand race names.
    "%N" expands to race N's short name.
    "%-N" expands to race N's adjective. */
string_t
GRaceNameList::expandNames(string_t text)
{
    string_t rv;

    while (text.size()) {
        string_t::size_type n = text.find('%');
        if (n == text.npos || n+1 >= text.length()) {
            rv += text;
            break;
        }
        rv.append(text, 0, n);
        text.erase(0, n+1);

        bool adj = false;
        if (text.size() > 0 && text[0] == '-') {
            adj = true;
            text.erase(0, 1);
        }
        if (text.size() > 0) {
            if (text[0] == '%') {
                rv += '%';
            } else {
                int race = getPlayerIdFromChar(text[0]);
                if (race > 0 && race <= NUM_OWNERS) {
                    if (adj)
                        rv += getAdjName(race);
                    else
                        rv += getShortName(race);
                }
            }
            text.erase(0, 1);
        }
    }
    return rv;
}

/************************************  ***********************************/

static int player_id, real_player_id;

int fof_colors[] = {
    COLOR_GREEN,
    COLOR_DARKYELLOWSCALE + 6,  // 0x96
    COLOR_RED,
    COLOR_WHITE,
    COLOR_WHITE
};
int fof_colors_dark[] = {
    COLOR_GREENBLACK,
    COLOR_BRIGHTORANGE,         // 0x72
    COLOR_RED,
    COLOR_BLACK,
    COLOR_BLACK
};

/** Load Race Names. Loads the RACE.NM file. */
void
initRaceNames()
{
    Ptr<Stream> file = spec_file_dir->openFile("race.nm", Stream::C_READ);
    racename_data.load(file);
}

/** Get current player. That's the one whose data we're looking at,
    0 if none. */
int
getPlayerId()
{
    return player_id;
}

/** Get current authenticated player. That's the one whose password
    we have. May be different from current player if he used a password
    cookie to get somebody else. */
int
getRealPlayerId()
{
    return real_player_id;
}

/** Friend-or-Foe test. Classifies the specified player, and returns
    one of the FOF_xxx constants. */
int
isFriend(int player)
{
    return (player == 0
            ? FOF_UNOWNED
            : (player_id == player ? FOF_ME : FOF_ENEMY));
}

/** Parse player character into player number. This is used for
    subspace messages, fcodes.cc, etc.
    \returns player id [0,11], or -1. */
int
getPlayerIdFromChar(char c)
{
    if (c >= '0' && c <= '9')
        return c - '0';
    else if (c == 'A' || c == 'a')
        return 10;
    else if (c == 'B' || c == 'b')
        return 11;
    else
        return -1;
}

/** Set player Ids.
    \param eff   "effective" id. The one in whose name we're giving commands.
    \param real  "real" id. The one which was authenticated. */
void
setPlayerIds(int eff, int real)
{
    real_player_id = real;
    player_id = eff;
}

/** FIXME. */
bool
allowRegistered(int player)
{
    return true;
}
