/**
  *  \file game/pvcr.cc
  *  \brief PVCR Implementation
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/pvcr.h"
#include "game/config.h"
#include "game/spec.h"
#include "game/formula.h"

/** \class VcrPlayerPHost
    \brief PHost VCR Player

    This is the PHost VCR player, started as a port from the PCC
    v1.0.15 player, and went through serious tweaking.

    This player supports all PHost versions from 2.4 upwards, up to at
    least version 4.0.

    Things remaining to do:
    - some things should be done in subroutines (killFighter...)
    - cache weapon specs

    Believe me or not, I have not copied code off the PDK or PHost.
    The PDK is still (or, "again") faster... */

/*
  This seems to be pretty optimisation-resistant.

  All tests run on a vanilla Pentium 200. Each test is the mean of at
  least two runs.

  + original:                   AC: 15.74,  Non-AC: 3.87
  + inlining:                   AC: 15.27,  Non-AC: 3.72
  + PVCR_INTEGER:               AC: 15.25,  Non-AC: 3.74(!)
  + f_getDamage():              AC: 16.11(!) [not used]
  + caching config:             AC: 13.58,  Non-AC: 3.59
  + fighterMove:                AC: 11.91,  Non-AC: 3.41
  + beamFindNearestFighter:     AC:  6.70,  Non-AC: 2.04
  + template<int side>:         AC:  5.36,  Non-AC: 1.64  (currently not used)

  For comparison, PDK 4.2:      AC: 9.50,   Non-AC: 2.44
  PDK 4.3 (beamFindNearest.)    AC: 5.53,   Non-AC: 1.72

  The one reason why the PDK is faster is probably that the PDK
  contains for loops (for(side=0; side<2; ++side)), where this code
  contains two function calls for each phase. gcc can optimise that
  better. The speedup gained by templatizing seems to support that
  supposition.

  Some things which do not get anything:
  + Status& st = status[side]
  + side^1 instead of !side (although it should give better code;
    in theory, that is) */

/** Bitmask of VCR capabilities supported by this player. */
static const unsigned SUPPORTED_CAP = vcrcap_DeathRays | vcrcap_Experience;

#ifdef CONF_STATIC_VCR
TVcr VcrPlayerPHost::vcr;
VcrPlayerPHost::Status VcrPlayerPHost::status[2];
uint16 VcrPlayerPHost::seed;
bool VcrPlayerPHost::done;
int VcrPlayerPHost::one_f, VcrPlayerPHost::right_probab;
unsigned VcrPlayerPHost::capabilities;
#endif

/** Add with saturation at limits.
    \returns a+b, limited to range [min,max]. */
static inline int
addsl(int a, int b, int min, int max)
{
    long sum = long(a) + b;
    if (sum < min)
        return min;
    else if (sum > max)
        return max;
    else
        return sum;
}

/** Get config /KEY/ for object on /SIDE/. Use this during
    initialisation. */
#define CONFIG1(KEY, SIDE) (config.KEY[vcr.objects[SIDE].owner-1])

/** Get config /KEY/ for object on /SIDE/. Use this during
    the battle. This uses the cached values. */
#define CONFIG2(KEY, SIDE) (status[SIDE].f.KEY)

/** Distance. Shortcut for later optimisatitons. */
#define DISTANCE() (long(status[1].r.obj_x) - status[0].r.obj_x)

/** Get experience-modified value for an option. */
#define CONFIG1EX(KEY, SIDE, LEVEL, MIN, MAX) \
      (addsl(CONFIG1(KEY, SIDE), (LEVEL) ? config.EMod##KEY[LEVEL-1] : 0, MIN, MAX))

/********************************** RNG **********************************/

/** Random Number Generator. Basic linear congruence. */
uint16
VcrPlayerPHost::random64k()
{
    seed = 13489 - seed * 259;
    return seed;
}

/** Random Number Generator. Returns a random number uniformly
    in [0, max) */
int
VcrPlayerPHost::randomRange(int max)
{
    int divisor = 65535U / max;
    int i;
    do
        i = random64k() / divisor;
    while (i >= max);
    return i;
}

/** Check for planet. Returns true iff /side/ is the planet. */
inline bool
VcrPlayerPHost::isPlanet(int side)
{
    return side && vcr.battle_type;
}

/************************** Standard "Hit" Stuff *************************/

/** \name Standard "Hit" Stuff */
//@{

#ifdef PVCR_INTEGER
/** Compute shield damage. Documented formula.
    Integer-only version: returns actual value (non-AC) or value scaled
    with mass+1. */
int32
VcrPlayerPHost::computeShieldDamageS(int expl, int kill, int mass, int side)
{
    if (config.AllowAlternativeCombat) {
        int32 damage = (CONFIG2(ShieldDamageScaling, side) * int32(expl)
                        + CONFIG2(ShieldKillScaling, side) * int32(kill));
        return std::min(damage, status[side].f.max_scaled);
    } else {
        int32 damage = radiv(CONFIG2(ShieldDamageScaling, side) * int32(expl)
                             + CONFIG2(ShieldKillScaling, side) * int32(kill),
                             mass + 1) + 1;
        if (damage > 10000)
            return 10000;
        else
            return damage;
    }
}

/** Compute hull damage. Documented formula.
    Integer-only version: returns actual value (non-AC) or value scaled
    with mass+1. */
int32
VcrPlayerPHost::computeHullDamageS(int expl, int kill, int mass, int side)
{
    if (config.AllowAlternativeCombat) {
        int32 d = int32(expl) * CONFIG2(HullDamageScaling, side);
        return std::min(d, status[side].f.max_scaled);
    } else {
        int32 d = radiv(computeShieldDamageS(expl, kill, mass, side) * CONFIG2(HullDamageScaling, side),
                       mass + 1) + 1;
        if (d > 10000)
            return 10000;
        else
            return d;
    }
}

/** Compute killed crew. Documented formula.
    Integer-only version: returns actual value (non-AC) or value scaled
    with mass+1. */
int32
VcrPlayerPHost::computeCrewKilledS(int kill, int mass, bool death, int side)
{
    if (config.AllowAlternativeCombat) {
        return CONFIG2(CrewKillScaling, side) * int32(kill);
    } else {
        int32 rv = radiv(CONFIG2(CrewKillScaling, side) * int32(kill), mass + 1);
        if (death && rv == 0)
            return 1;
        else
            return rv;
    }
}
#else
/** Compute shield damage. Documented formula. Returns actual value. */
VcrPlayerPHost::damage_t
VcrPlayerPHost::computeShieldDamage(int expl, int kill, int mass, int side)
{
    double damage = (CONFIG2(ShieldDamageScaling, side) * double(expl)
                     + CONFIG2(ShieldKillScaling, side) * double(kill))
        / double(mass + 1);
    if (damage > 10000)
        damage = 10000;
    if (! config.AllowAlternativeCombat)
        return (int) (damage + 1.5);
    else
        return damage;
}

/** Compute hull damage. Documented formula. Returns actual value. */
VcrPlayerPHost::damage_t
VcrPlayerPHost::computeHullDamage(int expl, int kill, int mass, int side)
{
    if (config.AllowAlternativeCombat) {
        double d = double(expl) * CONFIG2(HullDamageScaling, side) / double(mass + 1);
        if (d > 10000)
            return 10000;
        else
            return d;
    } else {
        double d = computeShieldDamage(expl, kill, mass, side) * CONFIG2(HullDamageScaling, side) / double(mass + 1);
        if (d > 10000)
            return 10000;
        else
            return (int) (d + 1.5);
    }
}

/** Compute shield damage. Documented formula. Returns actual value. */
VcrPlayerPHost::damage_t
VcrPlayerPHost::computeCrewKilled(int kill, int mass, bool death, int side)
{
    double d = CONFIG2(CrewKillScaling, side) * double(kill) / double(mass + 1);
    if (config.AllowAlternativeCombat) {
        return d;
    } else {
        int i = int(d + 0.5);
        if (death && i == 0)
            return 1;
        else
            return i;
    }
}
#endif

/** Hit object.
    \param side  object being hit
    \param kill  anti-life (x-ray) power of weapon
    \param expl  explosive power of weapon
    \param is_death_ray true iff this weapon emits death rays. "hit"
    can't itself check for this, because a beam may end up here with
    expl=0 when fired while partially charged */
bool
VcrPlayerPHost::hit(int side, int kill, int expl, bool is_death_ray)
{
#ifdef PVCR_INTEGER
    int damage_rate = 100;
    bool over = false;          // FIXME: this variable use is too complicated
    Status& st = status[side];

    if (kill <= 0)
        kill = 1;
    if (expl <= 0)
        expl = 1;

    if (is_death_ray && (capabilities & vcrcap_DeathRays))
        goto death_ray;
    is_death_ray = false;

    /* Hit shields */
    if (st.r.shield_scaled > 0) {
        int32 damage_s = computeShieldDamageS(expl, kill, vcr.mass[side], side);
        if (st.r.shield_scaled <= damage_s) {
            // Shields are completely worn down by this hit
            damage_rate = (damage_s - st.r.shield_scaled) * 100 / damage_s;
            st.r.shield_scaled = 0;
        } else {
            // Shields still hold
            damage_rate = 0;
            st.r.shield_scaled -= damage_s;
        }
    }

    /* Shields are down -- do damage */
    if (damage_rate <= 0)
        return over;

    st.r.damage_scaled2 += computeHullDamageS(expl, kill, vcr.mass[side], side) * damage_rate;

    if (st.r.damage_scaled2 >= st.f.damage_limit_scaled) {
        over = true;
    } else {
     death_ray:
        if (!isPlanet(side)) {
            int32 killed_s = computeCrewKilledS(kill, vcr.mass[side], is_death_ray, side) * damage_rate;
            st.r.crew_scaled2 -= killed_s;
            if (st.r.crew_scaled2 < 50*st.f.scale) {
                st.r.crew_scaled2 = 0;
                over = true;
            }
        }
    }
    return over;
#else
    int damage_rate = 100;
    bool over = false;

    if (kill <= 0)
        kill = 1;
    if (expl <= 0)
        expl = 1;

    if (is_death_ray && (capabilities & vcrcap_DeathRays))
        goto death_ray;
    is_death_ray = false;

    /* Hit shields */
    if (status[side].r.shield > 0) {
        double damage = computeShieldDamage(expl, kill, vcr.mass[side], side);
        if (status[side].r.shield <= damage) {
            // Shields are completely worn down by this hit
            damage_rate = int((damage - status[side].r.shield) * 100 / damage);
            status[side].r.shield = 0;
        } else {
            // Shields still hold
            damage_rate = 0;
            status[side].r.shield -= damage;
        }
    }

    /* Shields are down -- do damage */
    if (damage_rate <= 0)
        return over;

    status[side].r.damage += computeHullDamage(expl, kill, vcr.mass[side], side) * damage_rate / 100.0;

    if (int(status[side].r.damage + 0.5) >= status[side].f.damage_limit) {
        over = true;
    } else {
     death_ray:
        if (!isPlanet(side)) {
            double killed = computeCrewKilled(kill, vcr.mass[side], is_death_ray, side) * damage_rate / 100.0;
            status[side].r.crew -= killed;
            if (status[side].r.crew < 0.5) {
                status[side].r.crew = 0;
                over = true;
            }
        }
    }
    return over;
#endif
}
//@}

/******************************** Fighters *******************************/

/** \name Fighters */
//@{

/** Compute bay recharge rate. Documented formula. Used for initialisation. */
int
VcrPlayerPHost::computeBayRechargeRate(int num, int side, int exp)
{
//     int i = CONFIG1(BayRechargeBonus, side) * num
//         + CONFIG1(BayRechargeRate, side);
    int i = CONFIG1EX(BayRechargeBonus, side, exp, -500, 500) * num
        + CONFIG1EX(BayRechargeRate, side, exp, 0, 16384);
    return i > 1 ? i : 1;
}

/** Recharge Fighter Bays. */
void
VcrPlayerPHost::fighterRecharge(int side)
{
    register const int mx = vcr.objects[side].bay_count;
    register Status& st = status[side];
    for (int i = 0; i < mx; ++i)
        if (st.r.bay_status[i] < 1000)
            st.r.bay_status[i] += randomRange(st.f.bay_recharge);
}

/** Launch Fighters. */
void
VcrPlayerPHost::fighterLaunch(int side)
{
    register Status& st = status[side];

    /* can we launch a fighter? */
    if (vcr.num_fighters[side] == 0
        || st.r.launched_fighters >= CONFIG2(MaxFightersLaunched, side)
        || st.r.launch_countdown > 0)
        return;

    /* yes, we can! */
    int bay_mx = vcr.objects[side].bay_count;
    for (int bay = 0; bay < bay_mx; ++bay)
        if (st.r.bay_status[bay] >= 1000)
            for (int track = 0; track < CONFIG2(MaxFightersLaunched, side); ++track)
                if (st.r.fighter_status[track] == 0) {
                    /* okay, we have a bay which is ready, and an empty track. */
                    st.r.fighter_status[track]     = 1;
                    st.r.fighter_x[track]          = st.r.obj_x;
                    st.r.fighter_shots_left[track] = CONFIG2(StrikesPerFighter, side);
                    st.r.bay_status[bay]           = 0;
                    st.r.launched_fighters++;
                    vcr.num_fighters[side]--;
                    st.r.launch_countdown = CONFIG2(BayLaunchInterval, side);
                    vis->startFighter(side, track);
                    return;
                }
}

/** Move Fighters. Takes back returned fighters. */
void
VcrPlayerPHost::fighterMove(int side)
{
    if (status[side].r.launched_fighters == 0)
        return;

    int dir = (side == 0) ? CONFIG2(FighterMovementSpeed, 0) : -CONFIG2(FighterMovementSpeed, 1);

    for (int track = 0; track < CONFIG2(MaxFightersLaunched, side); ++track)
        if (status[side].r.fighter_status[track] == 1) {
            status[side].r.fighter_x[track] += dir;
        } else if (status[side].r.fighter_status[track] == 2) {
            status[side].r.fighter_x[track] -= dir;
            if ((side == 0)
                ? status[side].r.fighter_x[track] < status[side].r.obj_x
                : status[side].r.fighter_x[track] > status[side].r.obj_x)
            {
                /* fighter comes back to baseship */
                status[side].r.launched_fighters--;
                vcr.num_fighters[side]++;
                vis->landFighter(side, track);
                status[side].r.fighter_status[track] = 0;
            }
        }
}

/** Do fighter intercepts. */
void
VcrPlayerPHost::fighterIntercept()
{
    static const int NEVER = -0x4000;

    /* fighter intercept only happens if both have fighters */
    if (status[0].r.launched_fighters == 0 || status[1].r.launched_fighters == 0)
        return;

    /* compute screen positions */
    int match[2][VCR_MAX_FTRS];
    for (int w = 0; w < 2; ++w)
        for (int i = 0; i < CONFIG2(MaxFightersLaunched, w); ++i)
            if (status[w].r.fighter_status[i] == 0)
                match[w][i] = NEVER;
            else
                match[w][i] = status[w].r.fighter_x[i] >> 7;

    for (int lf = 0; lf < CONFIG2(MaxFightersLaunched, 0); ++lf)
        if (match[0][lf] != NEVER)
            if (randomRange(100) < one_f)
                for (int rf = 0; rf < CONFIG2(MaxFightersLaunched, 1); ++rf)
                    if (match[1][rf] == match[0][lf]) {
                        /* two fighters at the same place, intercepting */
                        /* theoretically, here is a slight imbalance. If
                           the "right" fighter survives, it can fire again
                           while a "left" one can not. Whether this is
                           relevant in practice is unknown. */
                        if (randomRange(100) < right_probab) {
                            vis->fireBeam(1, rf, lf, 1, CONFIG2(FighterBeamExplosive, 1), CONFIG2(FighterBeamKill, 1));
                            status[0].r.launched_fighters--;
                            vis->killFighter(0, lf);
                            status[0].r.fighter_status[lf] = 0;
                        } else {
                            vis->fireBeam(0, lf, rf, 1, CONFIG2(FighterBeamExplosive, 0), CONFIG2(FighterBeamKill, 0));
                            status[1].r.launched_fighters--;
                            vis->killFighter(1, rf);
                            status[1].r.fighter_status[rf] = 0;
                            match[1][rf] = NEVER;
                        }
                        break;
                    }
}

/** Fighters attack enemy. */
bool
VcrPlayerPHost::fighterAttack(int side)
{
    /* only if we have fighters */
    if (status[side].r.launched_fighters == 0)
        return false;

    int enemy_x = status[!side].r.obj_x;
    for (int i = 0; i < CONFIG2(MaxFightersLaunched, side); ++i)
        if (status[side].r.fighter_status[i] == 1
            && std::abs(int32(status[side].r.fighter_x[i]) - enemy_x) <= CONFIG2(FighterFiringRange, side))
        {
            status[side].r.fighter_shots_left[i]--;
            if (!status[side].r.fighter_shots_left[i])
                status[side].r.fighter_status[i] = 2;

            bool hitres = hit(!side, CONFIG2(FighterBeamKill, side), CONFIG2(FighterBeamExplosive, side), false);
            vis->fireBeam(side, i, -1, 1,
                          CONFIG2(FighterBeamExplosive, side),
                          CONFIG2(FighterBeamKill, side));
            if (hitres)
                return true;
        }
    return false;
}
//@}

/********************************* Beams *********************************/

/** \name Beams */
//@{

/** Compute beam hit odds. Documented formula. Used during initialisation. */
int
VcrPlayerPHost::computeBeamHitOdds(int beam, int side, int exp)
{
    GBeam& b = getBeam(beam);
//     int i = CONFIG1(BeamHitBonus, side) * (b.getKill() + b.getDamage()) / 100
//         + CONFIG1(BeamHitOdds, side);
    int i = CONFIG1EX(BeamHitBonus, side, exp, -4095, 4095) * (b.getKill() + b.getDamage()) / 100
        + CONFIG1EX(BeamHitOdds, side, exp, 0, 100);
    return i < 0 ? 0 : i;
}

/** Compute beam recharge rate. Documented formula. Used during
    initialisation. */
int
VcrPlayerPHost::computeBeamRechargeRate(int beam, int side, int exp)
{
    GBeam& b = getBeam(beam);
//     int i = ((b.getKill() + b.getDamage()) * CONFIG1(BeamRechargeBonus, side)) / 100
//         + CONFIG1(BeamRechargeRate, side);
    int i = ((b.getKill() + b.getDamage()) * CONFIG1EX(BeamRechargeBonus, side, exp, -4095, 4095)) / 100
        + CONFIG1EX(BeamRechargeRate, side, exp, 0, 16384);
    return i < 1 ? 1 : i;
}

/** Recharge beams. */
void
VcrPlayerPHost::beamRecharge(int side)
{
    register const int mx = vcr.objects[side].beam_count;
    for (int i = 0; i < mx; ++i) {
        if (status[side].r.beam_status[i] < 1000) {
            status[side].r.beam_status[i] += randomRange(status[side].f.beam_recharge);
            vis->updateBeam(side, i);
        }
    }
}

/** Find nearest-possible fighter. Returns fighter index [0,VCR_MAX_FTRS),
    or -1 if none. */
int
VcrPlayerPHost::beamFindNearestFighter(int side)
{
    int opp = !side;
    long mindist = 100000;
    int fighter = -1;

    /* When enemy does not have a fighter, don't fire at one... */
    if (status[opp].r.launched_fighters == 0)
        return -1;

    long my_x = status[side].r.obj_x;

    /* Look for a matching fighters. When we're asked to fire on
       attacking fighters only, accept only those first. */
    for (int i = 0; i < CONFIG2(MaxFightersLaunched, opp); ++i)
        if (status[opp].r.fighter_status[i] == 1
            || (!config.FireOnAttackFighters && status[opp].r.fighter_status[i] != 0))
        {
            long d = std::abs(my_x - status[opp].r.fighter_x[i]);
            if (d < mindist) {
                mindist = d;
                fighter = i;
            }
        }

    /* No fighter found. Check retreating fighters. Any of them does. */
    if (fighter < 0 && config.FireOnAttackFighters) {
        for (int i = 0; i < CONFIG2(MaxFightersLaunched, opp); ++i)
            if (status[opp].r.fighter_status[i] != 0) {
                fighter = i;
                break;
            }
        return fighter;
    } else {
        /* Found a fighter on first pass. Must be in range. */
        if (mindist > CONFIG2(BeamHitFighterRange, side))
            return -1;
        else
            return fighter;
    }
}

/** Fire beams on /side/. \returns true iff battle ends. */
bool
VcrPlayerPHost::beamFire(int side)
{
    int opp = !side;
    for (int beam = 0; beam < vcr.objects[side].beam_count; ++beam) {
        const GBeam& gb = getBeam(vcr.objects[side].beam_type);
        /* Can we fire at a fighter? */
        if (status[side].r.beam_status[beam] >= CONFIG2(BeamHitFighterCharge, side)) {
            /* PVCR tests `&& distance <= 100000', but that's not needed */
            bool missing = randomRange(100) >= status[side].f.beam_hit_odds;
            int fighter = beamFindNearestFighter(side);
            if (fighter >= 0) {
                /* We fire at a fighter. */
                vis->fireBeam(side, -1-beam, fighter, missing ? -1 : +1,
                              gb.getDamage(), gb.getKill());
                status[side].r.beam_status[beam] = 0;
                vis->updateBeam(side, beam);
                if (!missing) {
                    vis->killFighter(opp, fighter);
                    status[opp].r.fighter_status[fighter] = 0;
                    status[opp].r.launched_fighters--;
                }
                return false;
            }
        }

        /* Can we fire at the enemy? */
        if (status[opp].r.launched_fighters <= 0
            && status[side].r.beam_status[beam] >= CONFIG2(BeamHitShipCharge, side)
            && DISTANCE() <= CONFIG2(BeamFiringRange, side))
        {
            bool missing = randomRange(100) >= status[side].f.beam_hit_odds;
            int kill = gb.getKill()   * (status[side].r.beam_status[beam]/10) / 100;
            int dest = gb.getDamage() * (status[side].r.beam_status[beam]/10) / 100;
            if (raceId(vcr.objects[side].owner) == 5)
                kill *= 3;

            status[side].r.beam_status[beam] = 0;
            vis->updateBeam(side, beam);

            if (!missing) {
                bool hitr = hit(opp, kill, dest, gb.getDamage() == 0);
                vis->fireBeam(side, -1-beam, -1, 1, dest, kill);
                if (hitr)
                    return true;
            } else {
                vis->fireBeam(side, -1-beam, -1, -1, dest, kill);
            }
            return false;
        }
    }
    return false;
}
//@}

/******************************* Torpedoes *******************************/

/** \name Torpedoes */
//@{

/** Compute torpedo hit odds. Documented formula. Used in initialisation. */
int
VcrPlayerPHost::computeTorpHitOdds(int torp, int side, int exp)
{
    const GTorpLauncher& tl = getTorp(torp);
//     int i = ((CONFIG1(TorpHitBonus, side) * (tl.getKill() + tl.getDamage())) / 100
//              + CONFIG1(TorpHitOdds, side));
    int i = ((CONFIG1EX(TorpHitBonus, side, exp, -4095, 4095) * (tl.getKill() + tl.getDamage())) / 100
             + CONFIG1EX(TorpHitOdds, side, exp, 0, 100));
    return i < 0 ? 0 : i;
}

/** Compute torpedo recharge rate. Documented formula. Used in
    initialisation. */
int
VcrPlayerPHost::computeTubeRechargeRate(int torp, int side, int exp)
{
    const GTorpLauncher& tl = getTorp(torp);
//     int i = ((CONFIG1(TubeRechargeBonus, side) * (tl.getKill() + tl.getDamage())) / 100
//              + CONFIG1(TubeRechargeRate, side));
    int i = ((CONFIG1EX(TubeRechargeBonus, side, exp, -4095, 4095) * (tl.getKill() + tl.getDamage())) / 100
             + CONFIG1EX(TubeRechargeRate, side, exp, 0, 16384));
    return i < 1 ? 1 : i;
}

/** Recharge torpedo launchers. */
void
VcrPlayerPHost::torpsRecharge(int side)
{
    register const int mx = vcr.torp_launcher_count[side];
    for (int i = 0; i < mx; ++i)
        if (status[side].r.torp_status[i] < 1000) {
            status[side].r.torp_status[i] += randomRange(status[side].f.torp_recharge);
            vis->updateTorp(side, i);
        }
}

/** Fire torpedoes. */
bool
VcrPlayerPHost::torpsFire(int side)
{
    if (DISTANCE() > CONFIG2(TorpFiringRange, side) || vcr.num_torps[side] == 0)
        return false;

    /* Scaling factor for torpedo effect. Tim scales with 2 for some reason. */
    int factor = config.AllowAlternativeCombat ? 1 : 2;

    for (int launcher = 0; launcher < vcr.torp_launcher_count[side]; ++launcher)
        if (status[side].r.torp_status[launcher] >= 1000) {
            /* we're firing a torpedo */
            int rr = randomRange(100);

            vcr.num_torps[side]--;
            status[side].r.torp_status[launcher] = 0;
            vis->updateTorp(side, launcher);
            if (rr <= status[side].f.torp_hit_odds) {
                /* we hit the enemy */
                GTorpLauncher& tl = getTorp(vcr.objects[side].torp_type);
                bool hitr = hit(!side, tl.getKill() * factor, tl.getDamage() * factor, tl.getDamage() == 0);
                vis->fireTorp(side, rr, launcher);
                return hitr;
            } else {
                /* miss */
                vis->fireTorp(side, -1-rr, launcher);
                return false;
            }
        }
    return false;
}
//@}

/******************************** Control ********************************/

/** \name Main Stuff */
//@{

/** Move objects towards each other. */
void
VcrPlayerPHost::moveObjects()
{
    long remain = DISTANCE() - config.StandoffDistance;
    if (remain <= 0)
        return;

    /* move objects. Ensure StandoffDistance is not violated */
    long move_left = std::min (remain, long(CONFIG2(ShipMovementSpeed, 0)));
    status[0].r.obj_x += move_left;
    remain -= move_left;

    if (!isPlanet(1)) {
        long move_right = std::min (remain, long(CONFIG2(ShipMovementSpeed, 1)));
        status[1].r.obj_x -= move_right;
    }
}

/** Check whether object still has offensive capabilities. */
bool
VcrPlayerPHost::canStillFight(int side)
{
    const bool drcheck = !(capabilities & vcrcap_DeathRays) || !isPlanet(!side);
    return (vcr.objects[side].beam_count > 0 && (drcheck || getBeam(vcr.objects[side].beam_type).getDamage()))
        || (vcr.num_fighters[side] > 0 && vcr.objects[side].bay_count > 0)
        || (status[side].r.launched_fighters > 0)
        || (vcr.num_torps[side] > 0 && (drcheck || getTorp(vcr.objects[side].torp_type).getDamage()));
}

/** Play one cycle. Slightly different than in PCC 1.x due to
    more-precisely specified semantics. */
bool
VcrPlayerPHost::playCycle()
{
    if (!canStillFight(0) && !canStillFight(1))
        done = true;

    if (done)
        return false;

    ++time;
    if (status[0].r.launch_countdown > 0)
        --status[0].r.launch_countdown;
    if (status[1].r.launch_countdown > 0)
        --status[1].r.launch_countdown;

    fighterRecharge(0);
    fighterRecharge(1);
    beamRecharge(0);
    beamRecharge(1);
    torpsRecharge(0);
    torpsRecharge(1);
    fighterLaunch(0);
    fighterLaunch(1);

    if (fighterAttack(0) || fighterAttack(1)
        || torpsFire(0) || torpsFire(1)
        || beamFire(0) || beamFire(1)) {
        done = true;
    } else {
        fighterIntercept();
        fighterMove(0);
        fighterMove(1);
        moveObjects();
    }
    return true;
}

/** Initialize player.
    \param av3   true iff we're playing PHost 3, false for PHost 2. */
VcrPlayerPHost::VcrPlayerPHost(bool av3)
    : phost3(av3)
{ }

/** Verify one side of VCR. */
bool
VcrPlayerPHost::checkVcrSide(TVcr& vcr, int side)
{
    bool err = false;
    TVcrObject& o = vcr.objects[side];

    /* owner and types must be in range */
    if (o.owner <= 0 || o.owner > 12)
        o.owner = 12;

    if (o.beam_type < 0 || o.beam_type > NUM_BEAMS) {
        o.beam_type = o.beam_count = 0;
        err = true;
    }

    if (o.torp_type < 0 || o.torp_type > NUM_TORPS) {
        o.torp_type = vcr.torp_launcher_count[side] = 0;
        err = true;
    }

    /* validate weapon counts */
    if (o.beam_count > VCR_MAX_BEAMS) {
        o.beam_count = VCR_MAX_BEAMS;
        err = true;
    }

    if (vcr.torp_launcher_count[side] > VCR_MAX_TORPS) {
        vcr.torp_launcher_count[side] = VCR_MAX_TORPS;
        err = true;
    }

    if (o.bay_count > VCR_MAX_BAYS) {
        o.bay_count = VCR_MAX_BAYS;
        err = true;
    }

    /* ensure experience level is consistent with configuration */
    if (o.experience_level) {
        if (!(capabilities & vcrcap_Experience) || o.experience_level > config.NumExperienceLevels)
            err = true;
    }

    return err;
}

/** Check and correct VCR. Returns true iff an error was corrected,
    false if the VCR was ok. */
bool
VcrPlayerPHost::checkVcr(TVcr& vcr)
{
    bool l = checkVcrSide(vcr, 0);
    bool r = checkVcrSide(vcr, 1);
    return l || r;
}

/** Initialize VCR Player. Creates shadow copies of configuration
    parameters and pre-charges weapons. */
void
VcrPlayerPHost::initVcr(TVcr& avcr)
{
    vcr = avcr;
    if (checkVcr(vcr)) {
        status_word = VCRS_INVALID;
        return;
    }

    /* Initialize Playback */
    status_word = 0;
    time = 0;
    seed = vcr.random_seed;
    done = false;

    status[0].r.obj_x = -29000;
    status[1].r.obj_x = +29000;
    for (int side = 0; side < 2; ++side) {
#ifdef PVCR_INTEGER
        if (config.AllowAlternativeCombat)
            status[side].f.scale = vcr.mass[side] + 1;
        else
            status[side].f.scale = 1;
        status[side].r.shield_scaled  = vcr.shield[side] * int32(status[side].f.scale);
        status[side].r.crew_scaled2   = vcr.objects[side].crew * int32(status[side].f.scale) * 100;
        status[side].r.damage_scaled2 = vcr.objects[side].damage * int32(status[side].f.scale) * 100;
        status[side].f.max_scaled = int32(10000) * status[side].f.scale;
#else
        status[side].r.shield  = vcr.shield[side];
        status[side].r.crew    = vcr.objects[side].crew;
        status[side].r.damage  = vcr.objects[side].damage;
#endif

        if (vcr.objects[side].beam_count > 0) {
            status[side].f.beam_hit_odds = computeBeamHitOdds     (vcr.objects[side].beam_type, side,
                                                                   vcr.objects[side].experience_level);
            status[side].f.beam_recharge = computeBeamRechargeRate(vcr.objects[side].beam_type, side,
                                                                   vcr.objects[side].experience_level);
        } else {
            status[side].f.beam_hit_odds = status[side].f.beam_recharge = 0;
        }
        if (vcr.torp_launcher_count[side] > 0) {
            status[side].f.torp_hit_odds = computeTorpHitOdds(vcr.objects[side].torp_type, side,
                                                              vcr.objects[side].experience_level);
            status[side].f.torp_recharge = computeTubeRechargeRate(vcr.objects[side].torp_type, side,
                                                                   vcr.objects[side].experience_level);
        } else {
            status[side].f.torp_hit_odds = status[side].f.torp_recharge = 0;
        }
        if (vcr.objects[side].bay_count > 0) {
            status[side].f.bay_recharge = computeBayRechargeRate(vcr.objects[side].bay_count, side,
                                                                 vcr.objects[side].experience_level);
        } else {
            status[side].f.bay_recharge = 0;
        }

        // FIXME: this is still missing, but since the original
        // line is bogus (should be p.Shield, not VCR.Shield) and it works,
        // it's probably not needed.
        //   IF NOT IsPlanet(Who) AND IsFreighter(Who) THEN VCR.Shield[Who]:=0;
        // Change line below if needed.

        int charge = (vcr.shield[side] == 100) ? 1000 : 0;
        std::fill_n(status[side].r.beam_status, int(VCR_MAX_BEAMS), charge);
        std::fill_n(status[side].r.torp_status, int(VCR_MAX_TORPS), charge);
        std::fill_n(status[side].r.bay_status,  int(VCR_MAX_BAYS),  0); // !!!
        std::fill_n(status[side].r.fighter_status, int(VCR_MAX_FTRS), 0);
        std::fill_n(status[side].r.fighter_x, int(VCR_MAX_FTRS), 0);  // not needed?
        std::fill_n(status[side].r.fighter_shots_left, int(VCR_MAX_FTRS), 0);  // not needed?
        status[side].r.launch_countdown = status[side].r.launched_fighters = 0;

#ifdef PVCR_INTEGER
        /* Original comparison was:
            if (st.r.damage_scaled2 + 50 * st.f.scale >=
                st.f.damage_limit * 100 * st.f.scale)          */
        status[side].f.damage_limit_scaled =
            raceId(vcr.objects[side].owner) == 2
            ? (150 * 2 - 1) * 50 * status[side].f.scale
            : (100 * 2 - 1) * 50 * status[side].f.scale;
#else
        status[side].f.damage_limit = (raceId(vcr.objects[side].owner) == 2 ? 150 : 100);
#endif

        CONFIG2(ShieldDamageScaling, side)  = CONFIG1(ShieldDamageScaling, side);
        CONFIG2(ShieldKillScaling, side)    = CONFIG1(ShieldKillScaling, side);
        CONFIG2(HullDamageScaling, side)    = CONFIG1(HullDamageScaling, side);
        CONFIG2(CrewKillScaling, side)      = CONFIG1(CrewKillScaling, side);
        CONFIG2(MaxFightersLaunched, side)  = CONFIG1(MaxFightersLaunched, side);
        CONFIG2(StrikesPerFighter, side)    = CONFIG1(StrikesPerFighter, side);
        CONFIG2(BayLaunchInterval, side)    = CONFIG1(BayLaunchInterval, side);
        CONFIG2(FighterMovementSpeed, side) = CONFIG1(FighterMovementSpeed, side);
        CONFIG2(FighterBeamExplosive, side) = CONFIG1(FighterBeamExplosive, side);
        CONFIG2(FighterBeamKill, side)      = CONFIG1(FighterBeamKill, side);
        CONFIG2(FighterFiringRange, side)   = CONFIG1(FighterFiringRange, side);
        CONFIG2(BeamHitFighterRange, side)  = CONFIG1(BeamHitFighterRange, side);
        CONFIG2(BeamHitFighterCharge, side) = CONFIG1(BeamHitFighterCharge, side);
        CONFIG2(BeamFiringRange, side)      = CONFIG1(BeamFiringRange, side);
        CONFIG2(BeamHitShipCharge, side)    = CONFIG1(BeamHitShipCharge, side);
        CONFIG2(TorpFiringRange, side)      = CONFIG1(TorpFiringRange, side);
        CONFIG2(ShipMovementSpeed, side)    = CONFIG1(ShipMovementSpeed, side);

        if (vcr.objects[side].experience_level) {
            /* must be -1 for array indexing. Therefore, "<"
               comparison instead of "<=" */
            const int level = vcr.objects[side].experience_level - 1;
            ASSERT(level < config.NumExperienceLevels);
            ASSERT(level < 10);
            ASSERT(capabilities & vcrcap_Experience);

#define ADD_BONUS(x, min, max) \
            CONFIG2(x, side) = addsl(CONFIG2(x, side), config.EMod##x[level], min, max)

            ADD_BONUS(ShieldDamageScaling,   0, 32767);
            ADD_BONUS(ShieldKillScaling,     0, 32767);
            ADD_BONUS(HullDamageScaling,     0, 32767);
            ADD_BONUS(CrewKillScaling,       0, 32767);
            ADD_BONUS(MaxFightersLaunched,   0, VCR_MAX_FTRS);
            ADD_BONUS(StrikesPerFighter,     1, 100);
            // BayLaunchInterval
            ADD_BONUS(FighterMovementSpeed,  1, 10000);
            ADD_BONUS(FighterBeamExplosive,  1, 1000);
            ADD_BONUS(FighterBeamKill,       1, 1000);
            // FighterFiringRange, BeamHitFighterRange
            ADD_BONUS(BeamHitFighterCharge,  0, 1000);
            // BeamFiringRange, BeamHitShipCharge, TorpFiringRange, ShipMovementSpeed
        }
    }

    /* pre-compute fighter intercept probabilities */
    if (phost3) {
        int left_f   = (100 - CONFIG1(FighterKillOdds, 0)) * CONFIG1(FighterKillOdds, 1);
        int right_f  = (100 - CONFIG1(FighterKillOdds, 1)) * CONFIG1(FighterKillOdds, 0);
        one_f        = (left_f + right_f) / 100;
        right_probab = right_f / one_f;
        // FIXME: battle.c seems to do `right_probab = left_f / one_f'
    } else {
        /* In PHost 2, combat options were not arrayized. Hence, for
           a valid pconfig, all FighterKillOdds values are the same
           and we can pick any one */
        one_f        = config.FighterKillOdds[0];
        right_probab = 50;
    }

    vis->init();
}

/** Finish up VCR. */
void
VcrPlayerPHost::doneVcr(TVcr& avcr)
{
    if (status_word == VCRS_INVALID) {
        avcr = vcr;
        return;
    }

    ASSERT(done);

    for (int side = 0; side < 2; ++side) {
#ifdef PVCR_INTEGER
        if (config.AllowAlternativeCombat) {
            vcr.objects[side].damage = radiv(status[side].r.damage_scaled2, status[side].f.scale * 100);
            vcr.objects[side].crew   = radiv(status[side].r.crew_scaled2,   status[side].f.scale * 100);
            vcr.shield[side]         = radiv(status[side].r.shield_scaled,  status[side].f.scale);
        } else {
            vcr.objects[side].damage = radiv(status[side].r.damage_scaled2, 100);
            vcr.objects[side].crew   = radiv(status[side].r.crew_scaled2, 100);
            vcr.shield[side]         = status[side].r.shield_scaled;
        }
#else
        vcr.objects[side].damage = int(status[side].r.damage + 0.5);
        vcr.objects[side].crew   = int(status[side].r.crew + 0.5);
        vcr.shield[side]         = int(status[side].r.shield + 0.5);
#endif
        if (vcr.objects[side].damage > 100)   // FIXME: take out?
            vcr.objects[side].damage = 100;

#ifdef PVCR_INTEGER
        if (canStillFight(side) && vcr.objects[side].damage < 100)
#else
        if (canStillFight(side) && status[side].r.damage < 99.5)
#endif
        {
            for (int i = 0; i < CONFIG2(MaxFightersLaunched, side); ++i) {
                if (status[side].r.fighter_status[i]) {
                    ++vcr.num_fighters[side];
                    --status[side].r.launched_fighters;
                    vis->landFighter(side, i);
                    status[side].r.fighter_status[i] = 0;
                }
            }
            ASSERT(status[side].r.launched_fighters == 0);
        }
        status[side].r.launched_fighters = 0;
    }

    status_word = 0;

    /* Lizards fight up to 150, but explode afterwards. */
    if (vcr.objects[0].damage >= 100)
        status_word |= VCRS_LEFT_DESTROYED;
    else if (vcr.objects[0].crew <= 0)
        status_word |= VCRS_LEFT_CAPTURED;

    if (vcr.objects[1].damage >= 100)
        status_word |= VCRS_RIGHT_DESTROYED;
    else if (vcr.battle_type == 0 && vcr.objects[1].crew <= 0)
        status_word |= VCRS_RIGHT_CAPTURED;

    if (status_word & VCRS_LEFT_DESTROYED)
        vis->killObject(0);
    if (status_word & VCRS_RIGHT_DESTROYED)
        vis->killObject(1);

    if (status_word == 0)
        /* FIXME: can we guarantee that every status not caught
           by the above is a stalemate? */
        status_word = VCRS_STALEMATE;

    avcr = vcr;
}

bool
VcrPlayerPHost::setCapabilities(uint16 cap)
{
    if (cap & ~SUPPORTED_CAP)
        return false;
    capabilities = cap;
    return true;
}

/** Fast forward. Unfortunately we don't have a real fast-path for
    PHost, and it's unlikely we ever find one due to the ugly random
    number generator. */
void
VcrPlayerPHost::playFastForward()
{ }
//@}

/** \name Accessor Interface */
//@{

int
VcrPlayerPHost::getBeamStatus(int side, int id)
{
    return status[side].r.beam_status[id] / 10;
}

int
VcrPlayerPHost::getTorpStatus(int side, int id)
{
    return status[side].r.torp_status[id] / 10;
}

int
VcrPlayerPHost::getTorpCount(int side)
{
    return vcr.num_torps[side];
}

int
VcrPlayerPHost::getFighterCount(int side)
{
    return vcr.num_fighters[side];
}

int
VcrPlayerPHost::getShields(int side)
{
    /* Round, because driver code also rounds. */
#ifdef PVCR_INTEGER
    return radiv(status[side].r.shield_scaled, status[side].f.scale);
#else
    return int(status[side].r.shield + 0.5);
#endif
}

int
VcrPlayerPHost::getDamage(int side)
{
#ifdef PVCR_INTEGER
    return radiv(status[side].r.damage_scaled2, status[side].f.scale * 100);
#else
    return int(status[side].r.damage + 0.5);
#endif
}

int
VcrPlayerPHost::getCrew(int side)
{
#ifdef PVCR_INTEGER
    return radiv(status[side].r.crew_scaled2, status[side].f.scale * 100);
#else
    return int(status[side].r.crew + 0.5);
#endif
}

int
VcrPlayerPHost::getFighterX(int side, int id)
{
    return (status[side].r.fighter_x[id]) / 128 + 320;
}

int
VcrPlayerPHost::getFighterStatus(int side, int id)
{
    return status[side].r.fighter_status[id];
}

int
VcrPlayerPHost::getObjectX(int side)
{
    return (status[side].r.obj_x) / 128 + 320;
}

long
VcrPlayerPHost::getDistance()
{
    return status[1].r.obj_x - status[0].r.obj_x;
}

string_t
VcrPlayerPHost::getPlayerName() const
{
    return string_t(phost3 ? "PHost 3" : "PHost 2");
}

/** Save status. Easy mindless way, just save everything. Actually, we
    could re-compute status[].f and one_f/right_probab from the VCR
    record, but I'm too lazy to do that now. */
Ptr<VcrStatus>
VcrPlayerPHost::getStatusToken()
{
    Ptr<StatusToken> s = new StatusToken(time);

    s->status[0]    = status[0];
    s->status[1]    = status[1];
    s->seed         = seed;
    s->status_word  = status_word;
    s->done         = done;
    s->one_f        = one_f;
    s->right_probab = right_probab;
    s->vcr          = vcr;

    return s;
}

void
VcrPlayerPHost::setStatus(Ptr<VcrStatus> s)
{
    StatusToken* t = checked_cast<StatusToken*>(s.ptr());
    status[0]    = t->status[0];
    status[1]    = t->status[1];
    seed         = t->seed;
    status_word  = t->status_word;
    done         = t->done;
    one_f        = t->one_f;
    right_probab = t->right_probab;
    vcr          = t->vcr;
    time         = t->time;
}

//@}

VcrPlayerPHost::StatusToken::StatusToken(time_type t)
    : VcrStatus(t)
{ }
