/**
  *  \file game/reg.cc
  *  \brief Functions for dealing with Registrations
  *
  *  This contains only code to *read* registrations, none to
  *  create them. Do not add more modificators.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/reg.h"
#include "fileio/io.h"
#include "version.h"
#include "util/except.h"
#include "game/gamedirs.h"
#include "util/console.h"
#include "util/nls.h"
#include "cpluslib/format.h"

/** Global registration info. Use getRegInfo() to access it. */
GRegInfo reginfo;

/* Magic numbers */
uint32 GRegInfo::unreg_fizz[51] = {
  1118,  1846,  2535,  1664,  5200,  8424,  8827, 11440, 11817, 15080,
 16445,  4992, 19435, 18928, 18915, 23712, 22321, 27846, 23959, 29640,
 27573,  9152,  9568,  9984, 10400,  1040,  1742,  2613,  1664,  4745,
  5694,  2912,  3328,  3744,  4160,  4576,  4992,  5408,  5824,  6240,
  6656,  7072,  7488,  7904,  8320,  8736,  9152,  9568,  9984, 10400,
 485451
};

/******************************** GRegInfo *******************************/

/** \class GRegInfo
    Complete registration information. The actual reg key (player name
    and town in DOS, serial and reg date in Winplan) is read from
    FIZZ.BIN, the additional information (name/address in Winplan)
    is read from REG.KEY.

    This does not include more knowledge than absolutely needed to
    fetch the above values from files. In particular, we don't check
    checksums -- invalid keys are accepted and sent to host (which
    will then reject them). */
GRegInfo::GRegInfo()
{
    initUnregistered();
    initUnowned();
}

/** Initialize registration strings from data of a FIZZ.BIN file.
    \param data 51 dwords (offset 136 of a FIZZ.BIN file). */
void
GRegInfo::initFromFizz(const uint32* data)
{
    std::copy(data, data + 51, fizz);
}

/** Create unregistered key. */
void
GRegInfo::initUnregistered()
{
    initFromFizz(unreg_fizz);
}

/** Make this key "unowned" (name/address not set). */
void
GRegInfo::initUnowned()
{
    setRegStr3("Client: Planets Command Center II (v" PCC2_VERSION ")");
    setRegStr4(PCC2_URL);
}

string_t
GRegInfo::decode(int n) const
{
    /* hehe. */
    char buffer[25];
    for (int i = 0; i < 25; ++i, ++n)
        buffer[i] = fizz[n] * 5042 / (i+1) >> 16;
    return getBasicString(buffer);
}

/** Returns registration string 1 (name of player/serial). */
string_t
GRegInfo::getRegStr1() const
{
    return decode(0);
}

/** Returns registration string 2 (town of player/reg date). */
string_t
GRegInfo::getRegStr2() const
{
    return decode(25);
}

/** Returns registration string 3 (name as entered by player). */
string_t
GRegInfo::getRegStr3() const
{
    return getBasicString(wpstrings[0]);
}

/** Returns registration string 4 (address as entered by player). */
string_t
GRegInfo::getRegStr4() const
{
    return getBasicString(wpstrings[1]);
}

/** Check if this key is registered. */
bool
GRegInfo::isRegistered() const
{
    for (int i = 0; i < 25; ++i)
        if (fizz[i] != unreg_fizz[i])
            return true;
    return false;
}

/** Modify registration string 3. This is allowed, players can do it
    themselves by reinstalling Winplan. */
void
GRegInfo::setRegStr3(const string_t& str)
{
    storeBasicString(wpstrings[0], str);
}

/** Modify registration string 4. This is allowed, players can do it
    themselves by reinstalling Winplan. */
void
GRegInfo::setRegStr4(const string_t& str)
{
    storeBasicString(wpstrings[1], str);
}

/** Initialize from FIZZ.BIN.
    \param s FIZZ.BIN file, file pointer at beginning. */
void
GRegInfo::parseFizz(Stream* s)
{
    char buffer[340];
    uint32 cooked[51];
        
    if (s->read(buffer, sizeof buffer) != sizeof buffer)
        throw FileTooShortException(s->getName());
    for (int i = 0; i < 51; ++i)
        cooked[i] = getUint32(buffer, 136 + 4*i);

    initFromFizz(cooked);
}

/** Initialize from REG.KEY.
    \param s REG.KEY file, file pointer at beginning. */
void
GRegInfo::parseKey(Stream* s)
{
    char buffer[154];
    uint32 flag;

    if (s->read(buffer, sizeof buffer) != sizeof buffer)
        throw FileTooShortException(s->getName());
    flag = getUint32(buffer, 150);
    if (flag == 13) {
        /* valid key */
        for (int i = 0; i < 100; ++i)
            buffer[i] -= 13;
        setRegStr3(getBasicStringN(buffer,    50));
        setRegStr4(getBasicStringN(buffer+50, 50));
    } else if (flag == 666771) {
        /* valid, but not unlocked */
        initUnowned();
    } else {
        /* reject */
        throw FileFormatException(format(_("File `%s' is invalid"), "reg.key"));
    }
}

/** Load Registration information.
    \pre initGameGlobals has been called.
    \throws FileOpenException if FIZZ.BIN can't be found */
void
initRegInfo()
{
    getRegInfo().initUnregistered();
    getRegInfo().initUnowned();

    Ptr<Stream> s = game_file_dir->openFile("fizz.bin", Stream::C_READ);
    getRegInfo().parseFizz(s);

    if (winplan_dir) {
        s = winplan_dir->openFileNT("reg.key", Stream::C_READ);
        if (s)
            getRegInfo().parseKey(s);
    }
}
