/**
  *  \file game/rst.cc
  *  \brief Result File Parsing
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include "game/rst.h"
#include "fileio/io.h"
#include "util/except.h"
#include "util/nls.h"
#include "cpluslib/assert.h"

/*! \class GResultFile
    This class parses result files. */

GResultFile::GResultFile(Ptr<Stream> afile)
    : file(afile)
{
    file->seek(0);
    if (! getStructure(file, header))
        throw FileFormatException(_("Result file too short"));

    /* Figure out RST version */
    version = -1;
    parseWinplanVersion();
}

GResultFile::~GResultFile()
{ }

void
GResultFile::parseWinplanVersion()
{
    /* RST must be seekable */
    Stream::fpos_t size = file->getSize();
    if (size < 0)
        throw FileFormatException(_("Result file is not a regular file"));

    /* Size of Winplan header */
    static const int WINSIZE = 500*8 + 50*12 + 50*4 + 682 + 7800;
    
    if (std::memcmp(header.signature, "VER3.5", 6) == 0
        && header.signature[6] >= '0' && header.signature[6] <= '9'
        && header.signature[7] >= '0' && header.signature[7] <= '9')
    {
        /* may be Winplan RST. Host occasionally sends out RSTs
           bearing the 3.5 header, which are not actually Winplan-style. */
        int32 kore_ofs = header.offset_win - 1;
        if (kore_ofs < 0 || kore_ofs > size - WINSIZE)
            return;

        char buf[4];
        file->seek(kore_ofs + WINSIZE);
        if (file->read(buf, 4) != 4 ||
            (std::memcmp(buf, "1211", 4) != 0 && std::memcmp(buf, "1120", 4) != 0))
            return;

        /* It is a Winplan file. */
        version = 10*(header.signature[6]-'0') + header.signature[7]-'0';
    }
}

/** Get offset of a RST section.
    \param section Section number (SEC_xxxx). Only those section present
    in the current type of RST can be queried.
    \returns section offset, suitable for seek. */
int32
GResultFile::getSectionOffset(int section) const
{
    if (section >= 0 && section < 8) {
        return header.offsets[section] - 1;
    } else if (section == SEC_KORE) {
        ASSERT(version >= 0);
        return header.offset_win - 1;
    } else if (section == SEC_LEECH) {
        ASSERT(version >= 0);
        return header.offset_leech - 1;
    } else if (section == SEC_SKORE) {
        ASSERT(version >= 1);
        return header.offset_skore - 1;
    } else {
        FAIL();
        return 0;
    }
}
