/**
  *  \file game/save-dir.cc
  *  \brief Saving Game Data -- Normal unpacked directories
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/save.h"
#include "game/checksum.h"
#include "cpluslib/format.h"
#include "fileio/streams.h"
#include "game/spec.h"
#include "util/console.h"
#include "util/nls.h"
#include "game/player.h"
#include "game/global.h"
#include "game/struct.h"
#include "util/misc.h"

/** Save message outbox, file format 3.5.
    \param outbox  Message box to save
    \param dir     Target directory
    \param player  Player owning the file (should be same as
                   outbox.getSender()) */
static void
saveOutbox35(GOutbox& outbox, Ptr<Directory> dir, int player)
{
    Ptr<Stream> f = dir->openFileT(format("mess35%d.dat", player), Stream::C_CREATE);
    char header_buffer[19];
    zeroFill(header_buffer);
    storeUint16(header_buffer, outbox.getCount());
    if (!outbox.getCount()) {
        f->write(header_buffer, 2);
    } else {
        f->write(header_buffer, 19);
        for (int i = 0; i < outbox.getCount(); ++i) {
            /* FIXME: right now, we enforce the 600-limit */
            string_t text = outbox.getRawText(i);
            uint16   recv = outbox.getReceiverMask(i);
            encodeMessage(text);
            for (string_t::size_type n = text.length(); n > 0; --n)
                if (text[n-1] == 26)
                    text.insert(n, 1, 23);
            
            char message_buffer[616];
            message_buffer[0] = 0;     // filler
            message_buffer[1] = '1';   // valid flag
            for (int p = 1; p <= 12; ++p)
                message_buffer[1+p] = (recv & (1 << p)) ? '1' : '0';
            storeUint16(message_buffer + 14, 600);

            std::fill_n(message_buffer + 16, 600, char(32+13));
            copy_n(text.begin(), std::min(text.length(), 600U), message_buffer + 16);
            f->write(message_buffer, 616);
        }
    }
}

/** Save message outbox, file format 3.0.
    \param outbox  Message box to save
    \param dir     Target directory
    \param player  Player owning the file (should be same as
                   outbox.getSender()) */
void
saveOutbox(GOutbox& outbox, Ptr<Directory> dir, int player)
{
    /* count messages */
    int nmessages = 0;
    for (int i = 0; i < outbox.getCount(); ++i) {
        uint16 recv = outbox.getReceiverMask(i);
        for (int p = 1; p <= NUM_PLAYERS+1; ++p)
            if (recv & (1 << p))
                ++nmessages;
    }

    // FIXME: some kind of overflow check here?
    
    /* allocate buffer for message headers */
    int nslots = nmessages + 50;
    if (nslots < 150)
        nslots = 150;

    int msgh_size = nslots * TOutgoingMessageHeader::size + 2;
    Array<char> buffer(msgh_size);
    buffer[0U] = buffer[1U] = 0;

    /* write file */
    Ptr<Stream> f = dir->openFileT(format("mess%d.dat", player), Stream::C_CREATE);
    f->write(buffer.ptr(), msgh_size);

    uint32 output_index = msgh_size + 1;  // +1 for BASIC
    int    header_index = 2;
    for (int i = 1; i <= outbox.getCount(); ++i) {
        uint16 recv = outbox.getReceiverMask(i);
        for (int p = 1; p <= NUM_PLAYERS+1; ++p) {
            if (recv & (1 << p)) {
                /* send one message */
                string_t text_to_send = outbox.getSendPrefix(i, p) + outbox.getRawText(i);
                encodeMessage(text_to_send);
                
                TOutgoingMessageHeader h;
                h.address = output_index;
                h.length  = text_to_send.length();
                h.from    = outbox.getSender();
                h.to      = p;
                storeStructure(buffer.ptr() + header_index, h);
                f->write(text_to_send.data(), text_to_send.size());
                header_index += TOutgoingMessageHeader::size;
                output_index += text_to_send.length();
            }
        }
    }

    ASSERT(header_index == msgh_size);
    storeUint16(buffer.ptr(), nmessages);
    f->seek(0);
    f->write(buffer, msgh_size);
}

// FIXME: maybe this should be in cmds.cc
void
sendCommands(GOutbox& outbox, GCommandContainer& cms)
{
    string_t accum;
    for (GCommandContainer::iterator i = cms.begin(); i != cms.end(); ++i) {
        string_t text = (*i)->getCommandText();
        if (text.length() && text[0] != '$') {
            if (accum.length() + text.length() > 500) {
                outbox.addMessage(accum, 1 << outbox.getSender());
                accum.clear();
            }
            if (accum.length())
                accum += '\n';
            accum += text;
        }
    }
    if (accum.length())
        outbox.addMessage(accum, 1 << outbox.getSender());
}

// FIXME: this function is very incomplete.
void
saveDirectory(GGameTurn& trn, Ptr<Directory> dir, int player_mask)
{
    bool have_control = false;
    int  control_id   = 0;

    for (int pid = 1; pid <= NUM_PLAYERS; ++pid) {
        if (! (player_mask & (1 << pid)))
            continue;
        
        if (!have_control) {
            // FIXME: this assumes dir == game directory
            control_id = loadControl(pid);
            have_control = true;
        }

        console.write(LOG_INFO, format(_("Writing %d data..."), ::getAdjName(pid)));

        {
            Ptr<Stream> s;

            s = dir->openFileT(format("ship%d.dat", pid), Stream::C_CREATE | Stream::C_WRITE);
            saveShips(trn, s, 1 << pid, 0, save_ChecksumCount);
            s->write(trn.getGen(pid).getSig2(), 10);
            
            s = dir->openFileT(format("pdata%d.dat", pid), Stream::C_CREATE | Stream::C_WRITE);
            savePlanets(trn, s, 1 << pid, 0, save_ChecksumCount);
            s->write(trn.getGen(pid).getSig2(), 10);

            s = dir->openFileT(format("bdata%d.dat", pid), Stream::C_CREATE | Stream::C_WRITE);
            saveBases(trn, s, 1 << pid, 0, save_ChecksumCount);
            s->write(trn.getGen(pid).getSig2(), 10);

            /* messages and commands. We add the commands to the message
               box, save that, and remove them again. */
            Ptr<GOutbox> out = trn.getOutbox(pid);
            int marker = out->getCount();
            sendCommands(*out, trn.getCommands(pid));
            if (out->getFormat())
                saveOutbox35(*out, dir, pid);
            else
                saveOutbox(*out, dir, pid);
            out->deleteMessagesAfter(marker);

            /* FIXME: save cmdX.txt with remaining commands */
        }

        if (control_id) {
            // we are playing DOS planets
            saveControl(control_id);
            have_control = false;
        }
    }

    if (have_control)
        saveControl(control_id);
}
