/**
  *  \file game/save-trn.cc
  *  \brief Saving Turn Files
  *
  *  This module provides functions to save a game turn into a set of
  *  TRN files.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "cpluslib/format.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/player.h"
#include "game/reg.h"
#include "game/save.h"
#include "game/ship.h"
#include "game/trn.h"
#include "util/console.h"
#include "util/nls.h"
#include "game/msg.h"
#include "util/recode.h"

/** Send a single message. */
static void
sendMessage(GTurnfile& trn, int from, int to, string_t text)
{
    encodeMessage(text);

    char buffer[4];
    storeUint16(buffer + 0, from);
    storeUint16(buffer + 2, to);

    trn.addCommand(tcm_SendMessage, text.length(), buffer, sizeof buffer);
    trn.addData(text.data(), text.length());
}

/** Send a message outbox. */
static void
sendOutbox(GTurnfile& trn, GOutbox& outbox)
{
    for (int i = 0; i < outbox.getCount(); ++i) {
        uint16 rec = outbox.getReceiverMask(i);
        for (int p = 1; p <= NUM_PLAYERS+1; ++p) {
            if (rec & (1 << p))
                sendMessage(trn, outbox.getSender(), p,
                            outbox.getSendPrefix(i, p) + outbox.getRawText(i));
        }
    }
}

/** Send THost alliance commands. For THost, the mere presence of a
    friendly code change command triggers an alliance action. When we
    change the FCode back to what it should be, we can set alliances
    without sacrificing a ship for each action. */
static void
sendTHostAllies(GTurnfile& trn, GCommand* cmd, GShip& ship)
{
    char buffer[3];
    
    string_t s = cmd->getArg();
    while (s.length() > 0) {
        storeBasicString(buffer, s);
        trn.addCommand(tcm_ShipChangeFc, ship.getId(), buffer, 3);
        if (s.length() <= 3)
            break;
        s.erase(0, 3);
    }
    storeBasicString(buffer, ship.getFCode());
    trn.addCommand(tcm_ShipChangeFc, ship.getId(), buffer, 3);
}

/** Save turn files (TRNs). This will generate all turns at once.
    \param trn          data base
    \param dir          target directory
    \param player_mask  whose turns to generate (1 << pid) */
void
saveTurns(GGameTurn& trn, Ptr<Directory> dir, int player_mask)
{
    Ptr<GTurnfile> trns[NUM_PLAYERS];

    if (! player_mask)
        return;

    console.write(LOG_INFO, _("Generating turn commands..."));
    for (int pid = 1; pid <= NUM_PLAYERS; ++pid) {
        if (! (player_mask & (1 << pid)))
            continue;

        /* red tape */
        trns[pid-1] = new GTurnfile(pid, trn.getGen(pid).getTimestamp());
        trns[pid-1]->setFeatures(GTurnfile::TRNF_WINPLAN);
        trns[pid-1]->setRegInfo(::getRegInfo(), trn.getTurnNr());

        /* make commands */
        int src = 1 << pid;
        int tha_ship_id = 0;
        for (int i = 1; i <= NUM_SHIPS; ++i) {
            if (trn.getShip(i).isExistant() && (trn.getShip(i).getSrcFlags() & src)) {
                if (tha_ship_id == 0)
                    tha_ship_id = i;
                trn.getShip(i).makeTurnCommands(*trns[pid-1]);
            }
        }

        for (int i = 1; i <= NUM_PLANETS; ++i)
            if (trn.getPlanet(i).isExistant() && (trn.getPlanet(i).getPlanetSrcFlags() & src))
                trn.getPlanet(i).makePlanetTurnCommands(*trns[pid-1]);

        for (int i = 1; i <= NUM_PLANETS; ++i)
            if (trn.getPlanet(i).isExistant() && trn.getPlanet(i).hasBase() && (trn.getPlanet(i).getBaseSrcFlags() & src))
                trn.getPlanet(i).makeBaseTurnCommands(*trns[pid-1]);

        /* messages */
        if (Ptr<GOutbox> out = trn.getOutbox(pid))
            sendOutbox(*trns[pid-1], *out);

        /* command messages */
        {
            GCommandContainer& cms = trn.getCommands(pid);
            string_t accum;
            for (GCommandContainer::iterator i = cms.begin(); i != cms.end(); ++i) {
                if ((*i)->getCommand() == GCommand::phc_TAlliance) {
                    if (!tha_ship_id)
                        console.write(LOG_WARN, _("Player has no ship; alliance changes not transmitted"));
                    else
                        sendTHostAllies(*trns[pid-1], *i, trn.getShip(tha_ship_id));
                } else {
                    string_t text = (*i)->getCommandText();
                    if (text.length() && text[0] != '$') {
                        if (accum.length() + text.length() > 500) {
                            sendMessage(*trns[pid-1], pid, pid, accum);
                            accum.clear();
                        }
                        if (accum.length())
                            accum += '\n';
                        accum += text;
                    }
                }
            }
            if (accum.length())
                sendMessage(*trns[pid-1], pid, pid, accum);
        }

        /* generate turn */
        trns[pid-1]->update();
    }

    /* update trailers */
    uint32 cksums[NUM_PLAYERS];
    for (int i = 1; i <= NUM_PLAYERS; ++i)
        if (trns[i-1])
            cksums[i-1] = trns[i-1]->getDosTrailer().checksum;
        else
            cksums[i-1] = 0;

    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (trns[i-1]) {
            trns[i-1]->setTemplock(cksums);
            trns[i-1]->updateTrailer();
        }
    }

    /* Write 'em out */
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (trns[i-1]) {
            Ptr<Stream> out = dir->openFileT(format("player%d.trn", i), Stream::C_CREATE + Stream::C_WRITE);
            console.write(LOG_INFO,
                          format(_("Writing %s turn file (%d command%!1{s%})..."),
                                 ::getAdjName(i),
                                 trns[i-1]->getNumCommands()));
            trns[i-1]->write(out);
        }
    }
}
