/**
  *  \file game/save.cc
  *  \brief Saving Game Data -- General Utilities
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/save.h"
#include "game/global.h"
#include "game/ship.h"
#include "fileio/io.h"
#include "game/checksum.h"
#include "game/planet.h"

/** Save ships to stream. Writes a "count" word plus ship records.
    \param trn       turn to save
    \param str       stream to save in
    \param src       source flag mask to use (generally 1<<pid)
    \param checksum  pointer to checksum word to update, or null */
void
saveShips(GGameTurn& trn, Ptr<Stream> str, int src, uint32* checksum, int flags)
{
    /* pass 1: count */
    TInt16 v = { 0 };
    for (int i = 1; i <= NUM_SHIPS; ++i)
        if (trn.getShip(i).isExistant() && (trn.getShip(i).getSrcFlags() & src))
            ++v.value;
    storeStructure(str, v);
    if ((flags & save_ChecksumCount) && checksum)
        *checksum += getChecksum(v.value);

    /* pass 2: data */
    char buffer[TShip::size];
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        if (trn.getShip(i).isExistant() && (trn.getShip(i).getSrcFlags() & src)) {
            storeStructure(buffer, trn.getShip(i).getRawData());
            str->write(buffer, sizeof buffer);

            uint32 csum = getChecksum(buffer, sizeof buffer);
            setControlEntry(1, i, csum);
            if (checksum)
                *checksum += csum;
        }
    }
}

/** Save planets to stream. Writes a "count" word plus planet records.
    \param trn       turn to save
    \param str       stream to save in
    \param src       source flag mask to use (generally 1<<pid)
    \param checksum  pointer to checksum word to update, or null */
void
savePlanets(GGameTurn& trn, Ptr<Stream> str, int src, uint32* checksum, int flags)
{
    /* pass 1: count */
    TInt16 v = { 0 };
    for (int i = 1; i <= NUM_PLANETS; ++i)
        if (trn.getPlanet(i).isExistant() && (trn.getPlanet(i).getPlanetSrcFlags() & src))
            ++v.value;
    storeStructure(str, v);
    if ((flags & save_ChecksumCount) && checksum)
        *checksum += getChecksum(v.value);

    /* pass 2: data */
    char buffer[TPlanet::size];
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        if (trn.getPlanet(i).isExistant() && (trn.getPlanet(i).getPlanetSrcFlags() & src)) {
            storeStructure(buffer, trn.getPlanet(i).getRawData());
            str->write(buffer, sizeof buffer);

            uint32 csum = getChecksum(buffer, sizeof buffer);
            setControlEntry(2, i, csum);
            if (checksum)
                *checksum += csum;
        }
    }
}

/** Save bases to stream. Writes a "count" word plus base records.
    \param trn       turn to save
    \param str       stream to save in
    \param src       source flag mask to use (generally 1<<pid)
    \param checksum  pointer to checksum word to update, or null */
void
saveBases(GGameTurn& trn, Ptr<Stream> str, int src, uint32* checksum, int flags)
{
    /* pass 1: count */
    TInt16 v = { 0 };
    for (int i = 1; i <= NUM_PLANETS; ++i)
        if (trn.getPlanet(i).isExistant()
            && trn.getPlanet(i).hasBase()
            && (trn.getPlanet(i).getBaseSrcFlags() & src))
            ++v.value;
    storeStructure(str, v);
    if ((flags & save_ChecksumCount) && checksum)
        *checksum += getChecksum(v.value);

    /* pass 2: data */
    char buffer[TStarbase::size];
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        if (trn.getPlanet(i).isExistant()
            && trn.getPlanet(i).hasBase()
            && (trn.getPlanet(i).getBaseSrcFlags() & src))
        {
            storeStructure(buffer, trn.getPlanet(i).getRawBaseData());
            str->write(buffer, sizeof buffer);

            uint32 csum = getChecksum(buffer, sizeof buffer);
            setControlEntry(3, i, csum);
            if (checksum)
                *checksum += csum;
        }
    }
}
