/**
  *  \file game/ship-xfer.cc
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/ship.h"
#include "game/spec.h"
#include "game/cargo.h"
#include "util/nls.h"
#include "game/planet.h"

/***************************** GShipTransfer *****************************/

/** \class GShipTransfer
    \brief Ship Cargo Transfer Traits */

GShipTransfer::GShipTransfer(GShip& ship)
    : ship(ship)
{
    ship.getCargoHandler().incUseCount();
}

GShipTransfer::~GShipTransfer()
{
    ship.getCargoHandler().decUseCount();
}

int32
GShipTransfer::getAmount(CargoType type)
{
    return ship.getCargo(type);
}

void
GShipTransfer::changeAmount(CargoType type, int32 amount)
{
    ship.changeCargoRaw(type, amount);
}

void
GShipTransfer::reserveAmount(CargoType type, int32 amount)
{
    ship.getCargoHandler().changeReservation(type, amount);
}

bool
GShipTransfer::canHave(CargoType cargo_id)
{
    switch(cargo_id) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
     case el_Colonists:
     case el_Supplies:
     case el_Money:
        return true;
     case el_Fighters:
        return (ship.getBayCount() > 0);
     default:
        return (ship.getTorpLauncherCount() > 0
                && ship.getTorpType() == (cargo_id - el_Torps + 1));
    }
}

int32
GShipTransfer::getMax(CargoType cargo_id)
{
    switch(cargo_id) {
     case el_Neutronium:
        return ship.getHull().getFuel();
     case el_Money:
        return 10000;
     default:
        /* this is pretty hairy, but probably the best we can do */
        int32 rest = ship.getHull().getCargo()
            - getCurrent(el_Tritanium) - getCurrent(el_Duranium)
            - getCurrent(el_Molybdenum) - getCurrent(el_Supplies)
            - getCurrent(el_Colonists);
        if (ship.getTorpLauncherCount() > 0)
            rest -= getCurrent(GCargoType(el_Torps + ship.getTorpType() - 1));
        if (ship.getBayCount() > 0)
            rest -= getCurrent(el_Fighters);
        return rest + getCurrent(cargo_id);
    }
}

string_t
GShipTransfer::getName() const
{
    return ship.getName();
}

/**************************** Ship Transporter ***************************/

/* FIXME: For "undo jettison ammo", we must even lock this location in
   space, not just "this transporter". */

/** Construct a transporter which transports to /other/.
    \param ship       the "source" ship
    \param ship_ship  true if "target" object is a ship, false if planet
    \param other      "target" object
    \pre type of other matches ship_ship
    \pre cargo transporter is empty or already transferring to /other/
    \pre respective facility of ship not already in use */
GShipTransporter::GShipTransporter(GShip& ship, bool ship_ship, GObject& other)
    : ship(ship), transfer(ship.getTransporter(ship_ship)), other(&other),
      ship_ship(ship_ship), reserve()
{
    ASSERT(ship_ship ? dynamic_cast<GShip*>(this->other) != 0 : dynamic_cast<GPlanet*>(this->other) != 0);
    ASSERT(transfer.id == other.getId() || transfer.id == 0);

    ship.getCargoHandler().incUseCount();
    ship.getCargoHandler().acquireFacility(ship_ship ? GShip::sf_Transfer : GShip::sf_Unload);

    transfer.id = other.getId();
}

/** Construct a jettison transporter for /ship/.
    \param ship      ship
    \pre sf_Unload facility not already in use */
GShipTransporter::GShipTransporter(GShip& ship)
    : ship(ship), transfer(ship.getTransporter(false)), other(0),
      ship_ship(false), reserve()
{
    ASSERT(transfer.id == 0);
    
    ship.getCargoHandler().incUseCount();
    ship.getCargoHandler().acquireFacility(GShip::sf_Unload);
}

GShipTransporter::~GShipTransporter()
{
    if (!ship.isTransporterActive(ship_ship))
        ship.cancelTranporter(ship_ship);
    
    ship.getCargoHandler().releaseFacility(ship_ship ? GShip::sf_Transfer : GShip::sf_Unload);
    ship.getCargoHandler().decUseCount();
}

int32
GShipTransporter::getAmount(CargoType type)
{
    if (type <= 3)
        return transfer.ore[type] - reserve.ore[type];
    else if (type == el_Colonists)
        return transfer.colonists - reserve.colonists;
    else if (type == el_Supplies)
        return transfer.supplies - reserve.supplies;
    else
        return 0;
}

void
GShipTransporter::changeAmount(CargoType type, int32 amount)
{
    switch(type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        transfer.ore[type] += amount;
        return;
     case el_Supplies:
        transfer.supplies += amount;
        return;
     case el_Colonists:
        transfer.colonists += amount;
        return;
     default:;
    }
}

void
GShipTransporter::reserveAmount(CargoType type, int32 amount)
{
    switch(type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        reserve.ore[type] += amount;
        return;
     case el_Supplies:
        reserve.supplies += amount;
        return;
     case el_Colonists:
        reserve.colonists += amount;
        return;
     default:;
    }
}

bool
GShipTransporter::canHave(CargoType cargo_id)
{
    switch(cargo_id) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
     case el_Supplies:
     case el_Colonists:
        return true;
     default:
        return false;
    }
}

int32
GShipTransporter::getMax(CargoType /*cargo_id*/)
{
    return 10000;               // Tim's turn check
}

string_t
GShipTransporter::getName() const
{
    if (other)
        return other->getName();
    else
        return string_t(_("Jettison"));
}
