/**
  *  \file game/ship.h
  *  \brief Ships
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_SHIP_H
#define PCC_V2_GAME_SHIP_H

#include "game/struct.h"
#include "game/obj.h"
#include "game/objtypes.h"
#include "game/spec.h"
#include "game/cargo.h"

class GGameTurn;
class GTurnfile;
class GPlanet;

class GShip : public GMapObject {
 private:
    GShip();
    GShip(const GShip&);
    GShip& operator=(const GShip&);
 public:
    enum {
        SHIP_DAT    = 1,
        SHIP_DIS    = 2,
        SHIP_TARGET = 4,
        SHIP_XY     = 8
    };
 private:
    int id;
    TShip* dat;
    TShip* dis;
    TShipTarget* target;
    TShipXY shipxy;
    int orbit;
    /** Source flags. Bitfield (1<<player) of who's .dat file the
        information is from. User for saving. */
    int src_flags;
    /** Remote control. 0 if not RCd, -1 if forbidden, otherwise real owner */
    int rc_flag;
    /** UI Flags. (select, dirty, ...) */
    int ui_flags;
    /** Cargo arbiter. */
    GCargo cargo;
 public:
    GShip(int aid);
    ~GShip();

    /* Initialisation */
    void addDatEntry(const TShip& adat, int src_flags);
    void addDisEntry(const TShip& adis);
    void addTargetEntry(const TShipTarget& atgt);
    void addShipXYEntry(TShipXY axy);
    void addRCEntry(int flag);

    const TShip& getRawData() const { return *dat; }

    /* Enquiry */
    bool haveDat() const { return dat; }
    bool haveTarget() const { return target; }
    int getSrcFlags() const { return src_flags; }
    
    void applyTurnCommand(GTurnfile& trn, int index);
    void makeTurnCommands(GTurnfile& trn);

    /* Checking */
    void internalCheck();
    void combinedCheck1(GGameTurn& trn);

    /* Accessors */
    int    getMass() const;
    int    getHullId() const;
    GHull& getHull() const { return ::getHull(getHullId()); }
    int    getOrbit() const { return orbit; }
    int    getRealOwner() const;

    GPoint getWaypoint() const;
    int    getWaypointDX() const { return dat->waypoint_dx; }
    int    getWaypointDY() const { return dat->waypoint_dy; }

    bool   canDoFunction(HullFunction hf) const;
    int    getWarp() const;
    bool   isHyperdriving() const;

    int    getEngine() const { return dat->engine; }
    int    getBeamType() const { return dat->beam_type; }
    int    getBeamCount() const { return dat->beam_count; }
    int    getBayCount() const { return dat->bay_count; }
    int    getTorpType() const { return dat->torp_type; }
    int    getTorpLauncherCount() const { return dat->torp_launcher_count; }
    int    getAmmo() const; // { return dat->ammo; }

    int    getMission() const { return dat->mission; }
    int    getTowId() const { return dat->mission_tow; }
    int    getIntId() const { return dat->mission_intercept; }
    int    getPE() const { return dat->primary_enemy; }
    int    getDamage() const { return dat->damage; }
    int    getCrew() const { return dat->crew; }
    // int    getOre(int index) const { return dat->ore_cargo[index]; }
    int    getColonists() const { return getCargo(el_Colonists); }
    int    getSupplies() const { return getCargo(el_Supplies); }
    int    getMoney() const { return getCargo(el_Money); }
    int    getCargoRaw(GCargoType type) const;
    int    getCargo(GCargoType type) const;

    int    getFreeCargo() const;

    int    getUIFlags() const { return ui_flags; }
    void   setUIFlags(int fl) { ui_flags = fl; }

    string_t getFCode() const { return getBasicString(dat->fcode); }

    GCargo& getCargoHandler()            { return cargo; }

    TShipTransfer& getTransporter(bool ship_ship) { return ship_ship ? dat->transfer : dat->unload; }
    const TShipTransfer& getTransporter(bool ship_ship) const { return ship_ship ? dat->transfer : dat->unload; }
    bool           isTransporterActive(bool ship_ship) const throw();
    bool           isTransporterCancellable(bool ship_ship) const;
    void           cancelTranporter(bool ship_ship) throw();

    bool   getWaypointPermission() const;

    bool   isCloningAt(const GPlanet& planet) const;
    
    /* Setters */
    void   setName(const string_t& str);
    void   setFCode(const string_t& str);
    void   setPE(int pe);
    void   setWaypoint(GPoint pt);
    void   changeCargoRaw(GCargoType type, int amount);

    void   leaveFleet();

    /* GMapObject interface */
    GPoint   getPos() const;
    string_t getName() const;
    string_t getFullName() const;
    int      getId() const;
    int      getOwner() const;
    bool     isExistant() const;

    /** Ship flags. */
    enum {
        ss_Marked    = 1,       /**< Ship is marked. */
        ss_DirtyData = 2,       /**< Ship was modified. */
        ss_DirtyHead = 4,       /**< Ship name/owner/selection/type...\ was modified.  */
        ss_DirtyXY   = 8,       /**< Ship position/waypoint was modified. */

        ss_Dirty     = ss_DirtyData | ss_DirtyHead | ss_DirtyXY
    };

    /** Ship facilities. */
    enum Facility {
        sf_Unload,              ///< Unload transporter.
        sf_Transfer             ///< Ship/Ship transporter.
    };
};

class GShipType : public GObjectType {
 public:
    GShipType(GGameTurn& atrn);
    
    bool isExistant(int index);
    bool isMarked(int index);
    void toggleMark(int index);

    int  getDirtyFlags(int index);
    string_t getName(int index);
    string_t getClassName();

    GShip& getObject(int index);
    GShip& getShip() { return getObject(getCurrent()); }
};

class GShipTransfer : public GCargoContainer {
    GShip& ship;
 protected:
    int32  getAmount(CargoType type);
    void   changeAmount(CargoType type, int32 amount);
    void   reserveAmount(CargoType type, int32 amount);
 public:
    GShipTransfer(GShip& ship);
    ~GShipTransfer();
    
    bool     canHave(CargoType cargo_id);
    int32    getMax(CargoType cargo_id);
    string_t getName() const;
};

class GShipTransporter : public GCargoContainer {
    GShip&             ship;     //!< The ship whose transporter we're using.
    TShipTransfer&     transfer; //!< The transporter we're using
    GObject*           other;    //!< The other object (null if jettison)
    bool               ship_ship;
    TShipTransfer      reserve;    
 protected:
    int32  getAmount(CargoType type);
    void   changeAmount(CargoType type, int32 amount);
    void   reserveAmount(CargoType type, int32 amount);
 public:
    GShipTransporter(GShip& ship, bool ship_ship, GObject& other);
    GShipTransporter(GShip& ship);
    ~GShipTransporter();

    bool     canHave(CargoType cargo_id);
    int32    getMax(CargoType cargo_id);
    string_t getName() const;
};

#endif
