/**
  *  \file game/spec.cc
  *
  *  Specification files. This file contains the classes which
  *  correspond to static game information (hullspec etc.), and
  *  functions to load/access the static specification files.
  *
  *  \todo Management of short names
  *  \todo Free-form field for starship hulls?
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "game/spec.h"
#include "fileio/streams.h"
#include "util/except.h"
#include "util/nls.h"
#include "game/gamedirs.h"
#include "fileio/iostr.h"
#include "util/console.h"
#include "util/confpars.h"
#include "util/misc.h"
#include "game/config.h"
#include "cpluslib/format.h"

/** Beam specifications. Indexed 0 to NUM_BEAMS-1 */
std::vector<GBeam> beam_spec;

/** Torpedo specifications. Indexed 0 to NUM_TORPS-1 */
std::vector<GTorpLauncher> torp_spec;

/** Engine specifications. Indexed 0 to NUM_ENGINES-1 */
std::vector<GEngine> eng_spec;

/** Hull specification. Indexed 0 to NNN */
std::vector<GHull> hull_spec;

/** Truehull. For each player the hulls he can build. */
TTruehull truehull_spec;

/** \class GComponent
    \brief Component of a game item

    This class represents an abstract component of something
    (weapon, structure, ...) */
GComponent::~GComponent()
{ }

/** \class GBeam
    \brief Beam Weapon

    This class represents a beam weapon. */
GBeam::GBeam(int aid, const TBeam& adata)
    : GWeapon(aid), data(adata)
{ }

/** \class GTorpLauncher
    \brief Torpedo Launcher

    This class represents a torpedo launcher. */
GTorpLauncher::GTorpLauncher(int aid, const TTorpedo& adata)
    : GWeapon(aid), data(adata)
{ }

/** \class GEngine
    \brief Engine

    This class represents a starship engine. */
GEngine::GEngine(int aid, const TEngine& adata)
    : GComponent(aid), data(adata)
{ }

/** Get fuel factor. With the standard fuel consumption formula, this
    is the amount of fuel used by a 100000 kt ship travelling one
    month at the specified speed. The value commonly communicated
    to the user is `getFuelFactor(warp)/(warp*warp)' (fuel per lightyear); */
int
GEngine::getFuelFactor(int warp)
{
    if (warp == 0)
        return 0;

    ASSERT(warp > 0 && warp < 10);
    return data.fuel[warp-1];
}

/** \class GHull
    \brief Starship Hull

    Represents a starship hull. This is information gathered
    together from multiple files: HULLSPEC, HULLFUNC. */
GHull::GHull(int aid, const THull& adata)
    : GComponent(aid), data(adata)
{
    std::fill(specials, specials + hf_MAX, 0);
}

TCost
GHull::getCost() const
{
    TCost c = { data.mc, data.t, data.d, data.m };
    return c;
}

/** Set special functions.
    \param hf      Function (hf_xxx)
    \param players Bitmask. 1<<1 == Player 1 can build it, etc. */
void
GHull::setSpecialFunction(HullFunction hf, int16 players)
{
    specials[hf] |= players;
}

void
GHull::clearSpecialFunctions()
{
    std::fill_n(specials, countof(specials), 0);
}

/** Check for special function. Returns true iff the ship can do
    function hf if owned by the specified player. */
bool
GHull::canDoSpecial(HullFunction hf, int player) const
{
    return specials[hf] & (1 << player);
}

/** Check for cloaking device. Returns true iff the ship can
    cloak if owned by the specified player. */
bool
GHull::canCloak(int player) const
{
    return (specials[hf_Cloak] | specials[hf_AdvancedCloak]) & (1 << player);
}

/** Check for bioscanner. Returns true iff the ship has a bioscanner
    when owned by the specified player. */
bool
GHull::canBioscan(int player) const
{
    return (specials[hf_Bioscan] | specials[hf_FullBioscan]) & (1 << player);
}

/** Check for alchemy. Returns true iff the ship can do alchemy
    when owned by the specified player. */
bool
GHull::canAlchemy(int player) const
{
    return (specials[hf_MerlinAlchemy] | specials[hf_NeutronicRefinery]
            | specials[hf_AriesRefinery]) & (1 << player);
}

/******************************** Loading ********************************/

/** Load beamspec file. Throws on error. */
static void
loadBeams()
{
    Ptr<Stream> file = spec_file_dir->openFile("beamspec.dat", Stream::C_READ);

    beam_spec.erase(beam_spec.begin(), beam_spec.end());
    for (int i = 1; i <= NUM_BEAMS; ++i) {
        char buffer[TBeam::size];
        TBeam data;
        
        if (file->read(buffer, sizeof(buffer)) != sizeof(buffer))
            throw FileFormatException(format(_("Error reading `%s'"), "beamspec.dat"));
        getStructure(buffer, data);
        beam_spec.push_back(GBeam(i, data));
    }
}

/** Load torpspec file. Throws on error. */
static void
loadTorps()
{
    Ptr<Stream> file = spec_file_dir->openFile("torpspec.dat", Stream::C_READ);

    torp_spec.erase(torp_spec.begin(), torp_spec.end());
    for (int i = 1; i <= NUM_TORPS; ++i) {
        char buffer[TTorpedo::size];
        TTorpedo data;
        
        if (file->read(buffer, sizeof(buffer)) != sizeof(buffer))
            throw FileFormatException(format(_("Error reading `%s'"), "torpspec.dat"));
        getStructure(buffer, data);
        torp_spec.push_back(GTorpLauncher(i, data));
    }
}

/** Load engspec file. Throws on error. */
static void
loadEngines()
{
    Ptr<Stream> file = spec_file_dir->openFile("engspec.dat", Stream::C_READ);

    eng_spec.erase(eng_spec.begin(), eng_spec.end());
    for (int i = 1; i <= NUM_ENGINES; ++i) {
        char buffer[TEngine::size];
        TEngine data;
        
        if (file->read(buffer, sizeof(buffer)) != sizeof(buffer))
            throw FileFormatException(format(_("Error reading `%s'"), "engspec.dat"));
        getStructure(buffer, data);
        eng_spec.push_back(GEngine(i, data));
    }
}

/** Load default special functions. Loads the default special functions
    as hardcoded in THost and PHost 2. */
static void
setDefaultSpecials()
{
    struct SpecialInit {
        int hull;
        HullFunction func;
    };
    static const SpecialInit init[] = {
        { 3,   hf_HeatsTo50 },             // Bohemian
        { 3,   hf_ScansAllWormholes },     // Bohemian
        { 7,   hf_LokiAnticloak },         // Loki
        { 8,   hf_CoolsTo50 },             // Eros
        { 9,   hf_Bioscan },               // Brynhild
        { 21,  hf_Cloak },                 // Reptile
        { 22,  hf_Cloak },                 // Lizard Class Cruiser
        { 25,  hf_Cloak },                 // Saurian
        { 26,  hf_Cloak },                 // White Falcon
        { 27,  hf_Cloak },                 // Swift Heart
        { 28,  hf_Cloak },                 // Fearless Wing
        { 29,  hf_AdvancedCloak },         // Dark Wing
        { 31,  hf_AdvancedCloak },         // Resolute
        { 32,  hf_Cloak },                 // Bright Heart
        { 33,  hf_Cloak },                 // Deth Specula
        { 36,  hf_Cloak },                 // D7 Coldpain
        { 38,  hf_Cloak },                 // D3 Thorn
        { 39,  hf_D19bGlory },             // D19b Nefarious
        { 41,  hf_SaberGlory },            // Saber
        { 42,  hf_LadyRoyale },            // Lady Royale
        { 43,  hf_Cloak },                 // Dwarfstar
        { 44,  hf_Gravitonic },            // BR4 
        { 44,  hf_Cloak },                 // BR4 
        { 45,  hf_Gravitonic },            // BR5 
        { 45,  hf_Cloak },                 // BR5 
        { 46,  hf_Gravitonic },            // MBR 
        { 46,  hf_Cloak },                 // MBR 
        { 47,  hf_Cloak },                 // Red Wind 
        { 51,  hf_Hyperdrive },            // B200 
        { 56,  hf_FirecloudChunnel },      // Firecloud 
        { 64,  hf_HeatsTo100 },            // Onyx 
        { 69,  hf_ImperialAssault },       // SSD 
        { 77,  hf_Hyperdrive },            // PL21 
        { 84,  hf_FullBioscan },           // Pawn 
        { 87,  hf_Hyperdrive },            // Falcon 
        { 96,  hf_Ramscoop },              // Cobol 
        { 96,  hf_Bioscan },               // Cobol 
        { 97,  hf_AriesRefinery },         // Aries 
        { 104, hf_NeutronicRefinery },     // NR ship 
        { 105, hf_MerlinAlchemy }          // Merlin 
    };

    for (unsigned i = 0; i < countof(init); ++i)
        if (init[i].hull <= getNumHulls())
            getHull(init[i].hull).setSpecialFunction(init[i].func, ~0U);
}

static void
clearSpecials()
{
    for (int i = 1; i < getNumHulls(); ++i)
        getHull(i).clearSpecialFunctions();
}

namespace {
    /** HULLFUNC.TXT parser. */
    class HullfuncParser : public ConfigParser {
        int chosen_hull;
        int chosen_function;
        bool assign_to_hull;
     public:
        HullfuncParser() : chosen_hull(-1), chosen_function(-1), assign_to_hull(true) { }
        void assignSpecial(int line_id, string_t value, bool byrace);
        void assign(int line_id, string_t name, string_t value, string_t line);
        void error(int line_id, string_t msg);
    };
}

void
HullfuncParser::assign(int line_id, string_t name, string_t value, string_t /*line*/)
{
    static const char*const special_names[] = {
        "ALCHEMY",
        "REFINERY",
        "ADVANCEDREFINERY",
        "HEATSTO50",
        "COOLSTO50",
        "HEATSTO100",
        "HYPERDRIVE",
        "GRAVITONIC",
        "SCANSALLWORMHOLES",
        "GAMBLING",
        "ANTICLOAK",
        "IMPERIALASSAULT",
        "CHUNNELING",
        "RAMSCOOP",
        "FULLBIOSCAN",
        "ADVANCEDCLOAK",
        "CLOAK",
        "BIOSCAN",
        "GLORYDEVICELOWDAMAGE",
        "GLORYDEVICEHIGHDAMAGE"
    };
    CTASSERT (countof(special_names) == hf_MAX);
    if (stringMatch("Initialize", name)) {
        /* "Initialize = Clear" or "Initialize = Default" */
        if (stringMatch("Clear", value)) {
            clearSpecials();
        } else if (stringMatch("Default", value)) {
            clearSpecials();
            setDefaultSpecials();
        } else {
            error(line_id, _("Invalid argument to `Initialize'"));
        }
    } else if (stringMatch("Hull", name)) {
        /* "Hull = number (Garbage)" or "Hull = Start of name" */
        int h;
        string_t::size_type pos;
        if ((stringValue(value, h, pos)
             || (pos > 0 && stringValue(value.substr(0, pos), h)))
            && (h > 0 && h <= getNumHulls()))
        {
            chosen_hull = h;
        } else {
            value = strUCase(value);
            chosen_hull = -1;
            for (int i = 1; i <= getNumHulls(); ++i) {
                if (strUCase(getHull(i).getName()).compare(0, value.length(), value) == 0) {
                    chosen_hull = h;
                    break;
                }
            }
            if (chosen_hull < 0)
                error(line_id, _("Invalid hull name"));
        }
    } else if (stringMatch("Function", name)) {
        /* "Function = Number (garbage)" or "Function = Name" */
        int h;
        string_t::size_type pos;
        if ((stringValue(value, h, pos)
             || (pos > 0 && stringValue(value.substr(0, pos), h)))
            && (h >= 0 && h < hf_MAX))
        {
            chosen_function = h;
        } else {
            value = strUCase(value);
            chosen_function = -1;
            for (int i = 0; i < hf_MAX; ++i) {
                using namespace std;    // hooray.
                if (value.length() <= strlen(special_names[i])
                    && value.compare(0, value.length(), special_names[i], value.length()) == 0)
                {
                    chosen_function = i;
                    break;
                }
            }
            if (chosen_function < 0)
                error(line_id, _("Invalid hull function"));
        }
    } else if (stringMatch("Racesallowed", name)) {
        if (assign_to_hull)
            assignSpecial(line_id, value, true);
    } else if (stringMatch("Playersallowed", name)) {
        if (assign_to_hull)
            assignSpecial(line_id, value, false);
    } else if (stringMatch("Assignto", name)) {
        assign_to_hull = stringMatch("Hull", value);
    } else {
        error(line_id, _("Invalid keyword"));
    }
}
    
void
HullfuncParser::assignSpecial(int line_id, string_t value, bool byrace)
{
    if (chosen_function < 0)
        error(line_id, _("No function selected for assignment"));
    if (chosen_hull < 0)
        error(line_id, _("No hull selected for assignment"));

    unsigned players = 0;
    while (value.length()) {
        bool neg = false;
        if (value[0] == '-') {
            value.erase(0, 1);
            if (value.length() && value[0] >= '0' && value[0] <= '9') {
                neg = true;
                goto digit;
            }
            players = 0;
        } else if (value[0] == '+') {
            value.erase(0, 1);
            if (value.length() && value[0] >= '0' && value[0] <= '9')
                goto digit;
            players = ~0U;
        } else if (value[0] >= '0' && value[0] <= '9') {
         digit:
            string_t::size_type pos;
            int num;
            if (stringValue(value, num, pos)) {
                value.clear();
            } else {
                if (pos && stringValue(value.substr(0, pos), num))
                    /* okay */;
                else
                    error(line_id, _("Invalid number"));
            }
            if (byrace) {
                for (int i = 1; i <= NUM_PLAYERS; ++i) {
                    if (config.PlayerRace[i-1] == num) {
                        if (neg)
                            players &= ~(1U << i);
                        else
                            players |= 1U << i;
                    }
                }
            } else {
                if (num > 0 && num <= NUM_PLAYERS)
                    if (neg)
                        players &= ~(1U << num);
                    else
                        players |= 1U << num;
                else
                    error(line_id, _("Invalid player number"));
            }
        } else {
            value.erase(0, 1);
        }
    }
    getHull(chosen_hull).setSpecialFunction(HullFunction(chosen_function), players);
}

void
HullfuncParser::error(int line_id, string_t msg)
{
    console.write(LOG_WARN, format(_("Line %d: %s"), line_id, msg));
}

/** Load Hull function file. */
static void
loadHullfunc(Ptr<Stream> file, bool in_section)
{
    StandardStream in(file);
    HullfuncParser parser;
    parser.parse(in, "hullfunc", in_section);
}

/** Load hullspec and truehull file. Throws on error.
    \todo Load HULLSPEC.CC(?)
    \pre  configuration has been loaded. */
static void
loadHulls()
{
    Ptr<Stream> file = spec_file_dir->openFile("hullspec.dat", Stream::C_READ);

    hull_spec.erase(hull_spec.begin(), hull_spec.end());
    int nhulls = file->getSize() / THull::size;
    if (nhulls < 10)
        throw FileTooShortException("hullspec.dat");
    if (nhulls > MAX_HULLS)
        nhulls = MAX_HULLS;
    
    for (int i = 1; i <= nhulls; ++i) {
        char buffer[THull::size];
        THull data;
        
        if (file->read(buffer, sizeof(buffer)) != sizeof(buffer))
            throw FileFormatException(format(_("Error reading `%s'"), "hullspec.dat"));
        getStructure(buffer, data);
        hull_spec.push_back(GHull(i, data));
    }

    file = spec_file_dir->openFile("truehull.dat", Stream::C_READ);

    char buffer[TTruehull::size];
    if (file->read(buffer, sizeof(buffer)) != sizeof(buffer))
        throw FileFormatException(format(_("Error reading `%s'"), "truehull.dat"));
    getStructure(buffer, truehull_spec);

    for (int p = 0; p < NUM_PLAYERS; ++p)
        for (int i = 0; i < NUM_HULLS_PER_PLAYER; ++i)
            if (truehull_spec.hulls[p][i] < 0 || unsigned(truehull_spec.hulls[p][i]) > hull_spec.size())
                throw FileFormatException(format(_("File `%s' is invalid"), "truehull.dat"));

    setDefaultSpecials();
    if ((file = spec_file_dir->openFileNT("shiplist.txt", Stream::C_READ | Stream::C_TEXT)))
        loadHullfunc(file, false);
    else if ((file = spec_file_dir->openFileNT("hullfunc.txt", Stream::C_READ | Stream::C_TEXT)))
        loadHullfunc(file, true);
}

/**************************** Public Interface ***************************/

/** Load Specification Files.
    \pre initGameGlobals() has been called.
    \pre Configuration was loaded. */
void
initSpecification()
{
    loadTorps();
    loadBeams();
    loadEngines();
    loadHulls();
}

/** Get Truehull slot. Returns the position of hull hid in player pid's
    Truehull list. Returns 0 if he can not build that hull. */
int
getTruehullSlot(int pid, int hid)
{
    for (int i = 1; i <= NUM_HULLS_PER_PLAYER; ++i)
        if (getTruehull(pid, i) == hid)
            return i;
    return 0;
}
