/**
  *  \file spec.h
  *
  *  Specification files. This file contains the classes which
  *  correspond to static game information (hullspec etc.)
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_SPEC_H
#define PCC_V2_SPEC_H

#include <string>
#include <vector>
#include "game/struct.h"
#include "fileio/io.h"
#include "cpluslib/assert.h"

/* Starship Component */
class GComponent {
    int id;
 public:
    GComponent(int aid) : id(aid) { }
    virtual ~GComponent();
    
    virtual int getMass() const = 0;
    virtual int getTech() const = 0;
    virtual TCost getCost() const = 0;
    virtual string_t getName() const = 0;

    int getId() const { return id; }
};

/* Weapon */
class GWeapon : public GComponent {
 public:
    GWeapon(int aid) : GComponent(aid) { }
    virtual int getKill() const = 0;
    virtual int getDamage() const = 0;
};

/* Beam */
class GBeam : public GWeapon {
    TBeam data;
 public:
    GBeam(int aid, const TBeam& adata);

    string_t getName() const { return getBasicString(data.name); }
    int getKill() const { return data.kill; }
    int getDamage() const { return data.damage; }
    int getMass() const { return data.mass; }
    int getTech() const { return data.tech; }
    TCost getCost() const { return data.cost; }

    int f_getDamage() const { return data.damage; }
    int f_getKill() const { return data.kill; }
};

/* Torpedo Launcher */
class GTorpLauncher : public GWeapon {
    TTorpedo data;
 public:
    GTorpLauncher(int aid, const TTorpedo& adata);
    
    string_t getName() const { return getBasicString(data.name); }
    int getKill() const { return data.kill; }
    int getDamage() const { return data.damage; }
    int getMass() const { return data.laun_mass; }
    int getTech() const { return data.tech; }
    TCost getCost() const { return data.laun_cost; }

    int f_getDamage() const { return data.damage; }
    int f_getKill() const { return data.kill; }
};

/* Engine */
class GEngine : public GComponent {
    TEngine data;
 public:
    GEngine(int aid, const TEngine& adata);
    
    int getMass() const { return 0; }
    int getTech() const { return data.tech; }
    TCost getCost() const { return data.cost; };
    string_t getName() const { return getBasicString(data.name); }
    
    int getFuelFactor(int warp);
};

enum HullFunction {
    hf_MerlinAlchemy,             ///< 0 = Merlin Alchemy: 9 Sup -> 3 Min
    hf_NeutronicRefinery,         ///< 1 = Neutronic Refinery: Min + Sup -> Fuel
    hf_AriesRefinery,             ///< 2 = Aries Refinery: Min -> Fuel
    hf_HeatsTo50,                 ///< 3 = Bohemian Terraformer
    hf_CoolsTo50,                 ///< 4 = Eros Terraformer
    hf_HeatsTo100,                ///< 5 = Onyx Terraformer
    hf_Hyperdrive,                ///< 6
    hf_Gravitonic,                ///< 7
    hf_ScansAllWormholes,         ///< 8 = Bohemian
    hf_LadyRoyale,                ///< 9 = Lady Royale
    hf_LokiAnticloak,             ///< 10 = Loki
    hf_ImperialAssault,           ///< 11 = SSD
    hf_FirecloudChunnel,          ///< 12 = Firecloud
    hf_Ramscoop,                  ///< 13 = Cobol
    hf_FullBioscan,               ///< 14 = Pawn
    hf_AdvancedCloak,             ///< 15 = Dark Wing
    hf_Cloak,                     ///< 16
    hf_Bioscan,                   ///< 17
    hf_SaberGlory,                ///< 18 = Saber (10% damage to own ships)
    hf_D19bGlory,                 ///< 19 = D19b (20% damage to own ships)
    
    hf_MAX                        
};

/* Starship Hull */
class GHull : public GComponent {
    THull data;
    int16 specials[hf_MAX];     ///< Special Functions. 1<<player = player can use it
 public:
    GHull(int aid, const THull& adata);

    int getMass() const { return data.mass; }
    int getTech() const { return data.tech; }
    TCost getCost() const;
    string_t getName() const { return getBasicString(data.name); }
    string_t getShortName() const { return getName(); /* FIXME!!! */ }

    int getPicture() const { return data.picture; }
    int getFuel() const { return data.fuel; }
    int getCrew() const { return data.crew; }
    int getNumEngines() const { return data.engines; }
    int getCargo() const { return data.cargo; }
    int getNumBays() const { return data.bays; }
    int getMaxLaunchers() const { return data.launchers; }
    int getMaxBeams() const { return data.beams; }

    void setSpecialFunction(HullFunction hf, int16 players);
    void clearSpecialFunctions();
    bool canDoSpecial(HullFunction hf, int player) const;
    bool canCloak(int player) const;
    bool canAlchemy(int player) const;
    bool canBioscan(int player) const;
    int  getPlayersThatCan(HullFunction hf) const { return specials[hf]; }
};

extern std::vector<GBeam> beam_spec;
extern std::vector<GTorpLauncher> torp_spec;
extern std::vector<GEngine> eng_spec;
extern std::vector<GHull> hull_spec;
extern TTruehull truehull_spec;

void initSpecification();

/** Get beam. bid = beam number (1..NUM_BEAMS) */
inline GBeam&
getBeam(int bid) {
    ASSERT(bid > 0 && bid <= NUM_BEAMS);
    return beam_spec[bid-1];
}

/** Get torpedo launcher. tid = torpedo number (1..NUM_TORPS) */
inline GTorpLauncher&
getTorp(int tid) {
    ASSERT(tid > 0 && tid <= NUM_TORPS);
    return torp_spec[tid-1];
}

/** Get engine. eid = engine number (1..NUM_ENGINES) */
inline GEngine&
getEngine(int eid) {
    ASSERT(eid > 0 && eid <= NUM_ENGINES);
    return eng_spec[eid-1];
}

/** Get hull. hid = hull number (1..getNumHulls()) */
inline GHull&
getHull(int hid) {
    ASSERT(hid > 0 && unsigned(hid) <= hull_spec.size());
    return hull_spec[hid-1];
}

/** Get truehull slot. pid = player number, slot = 1..20 */
inline int
getTruehull(int pid, int slot) {
    ASSERT(pid > 0 && pid <= NUM_PLAYERS);
    ASSERT(slot > 0 && slot <= NUM_HULLS_PER_PLAYER);
    return truehull_spec.hulls[pid-1][slot-1];
}

/** Get number of hulls. */
inline int
getNumHulls() {
    return hull_spec.size();
}

int getTruehullSlot(int pid, int hid);

#endif
