/**
  *  \file game/transact.cc
  *  \brief Transactions
  *
  *  This provides the basic transaction class, on which most cargo
  *  manipulation stuff is based.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/transact.h"
#include "cpluslib/assert.h"

/** \class GTransaction

    This class supports user-controlled transactions. The transaction
    model is as follows:
    - we start the transaction by constructing a GTransaction descendant
    - the user prepares the transaction by calling some of the
      descentants methods
    - when the user hits "OK", the transaction is committed, otherwise
      it is cancelled.

    Note that a transaction must always be cleanly committed or
    cancelled, it is an error to destroy a transaction via a
    destructor without having called one of commit() or cancel().

    GTransaction descendants must supply the cancel/commit methods.
    Those must eventually call GTransaction::cancel() resp. commit()
    to tell the core that the transaction is finished. If you wish,
    you can call cancel/commit from your destructor.

    Our transactions can be nested, overlapped and all other nasty
    things. We do not support pre-emptive concurrency, though.

    In a multi-part transaction, things may depend on each other. For
    example, a sub-transaction may state "if this transaction is
    committed, another sub-transaction must be committed as well". To
    ensure such things, you can use mustCommit() and mustCancel();
    these will trigger an assert fault when the transaction is not
    committed. */
GTransaction::~GTransaction()
{
    ASSERT(isFinished());
}

void
GTransaction::finish(Result r)
{
    ASSERT(must_result == Nothing || must_result == r);
    result = r;
}

/** Make sure this transaction is cancelled. Triggers an assert
    failure when the transaction is committed instead. */
void
GTransaction::mustCancel()
{
    ASSERT(must_result != Committed && result != Committed);
    must_result = Cancelled;
}

/** Make sure this transaction is committed. Triggers an assert
    failure when the transaction is cancelled instead. */
void
GTransaction::mustCommit()
{
    ASSERT(must_result != Cancelled && result != Cancelled);
    must_result = Committed;
}
