/**
  *  \file game/trn-cmd.cc
  *  \brief Applying turn commands to ship/planet
  *
  *  These functions provide the link between the game objects and
  *  GTurnfile. They belong somewhere between save.cc and trn.cc.
  *
  *  General notes: because this code is based on the marshalling
  *  primitives to dissect structures, there's no need to worry about
  *  character sets. getPartialStructure() and storeStructure() will
  *  translate just fine.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/ship.h"
#include "game/planet.h"
#include "game/trn.h"
#include "cpluslib/assert.h"
#include "fileio/io.h"

/* Ships */

/** Apply turn command for a ship. \pre haveDat() */
void
GShip::applyTurnCommand(GTurnfile& trn, int index)
{
    ASSERT(dat);
    getPartialStructure(trn.getCommandData(index),
                        dat, dat->parser,
                        trn.getRecordIndexForCode(trn.getCommand(index)),
                        trn.getCommandLength(index));
}

/** Generate turn commands for a ship. \pre haveDat() */
void
GShip::makeTurnCommands(GTurnfile& trn)
{
    ASSERT(dat && dis);
    char dat_buffer[TShip::size], dis_buffer[TShip::size];

    storeStructure(dat_buffer, *dat);
    storeStructure(dis_buffer, *dis);
    trn.makeShipCommands(getId(), dis_buffer, dat_buffer);
}

/* Planets */

/** Apply turn commands for a planet.
    \pre haveDat() or haveBaseDat(), depending on command. */
void
GPlanet::applyTurnCommand(GTurnfile& trn, int index)
{
    int cmd = trn.getCommand(index);
    if (cmd == tcm_PlanetBuildBase) {
        ASSERT(dat);
        dat->build_base = 1;
    } else if (cmd >= tcm_BaseFIRST) {
        ASSERT(base_dat);
        getPartialStructure(trn.getCommandData(index),
                            base_dat, base_dat->parser,
                            trn.getRecordIndexForCode(trn.getCommand(index)),
                            trn.getCommandLength(index));
    } else {
        ASSERT(dat);
        getPartialStructure(trn.getCommandData(index),
                            dat, dat->parser,
                            trn.getRecordIndexForCode(trn.getCommand(index)),
                            trn.getCommandLength(index));
    }
}

/** Generate turn commands for a planet. \pre haveDat() */
void
GPlanet::makePlanetTurnCommands(GTurnfile& trn)
{
    ASSERT(dat && dis);
    char dat_buffer[TPlanet::size], dis_buffer[TPlanet::size];

    storeStructure(dat_buffer, *dat);
    storeStructure(dis_buffer, *dis);
    trn.makePlanetCommands(getId(), dis_buffer, dat_buffer);
}

/** Generate turn commands for a starbase. \pre haveBaseDat() */
void
GPlanet::makeBaseTurnCommands(GTurnfile& trn)
{
    ASSERT(base_dat && base_dis);
    char dat_buffer[TStarbase::size], dis_buffer[TStarbase::size];

    storeStructure(dat_buffer, *base_dat);
    storeStructure(dis_buffer, *base_dis);
    trn.makeBaseCommands(getId(), dis_buffer, dat_buffer);
}
