/**
  *  \file game/trn.h
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_TRN_H
#define PCC_V2_GAME_TRN_H

#include "game/struct.h"
#include "fileio/streams.h"
#include "game/reg.h"

/** TRN Command codes. The names are the same as in UN-TRN, the file
    format list, and some utilities inspired by the above.
    Those marked "<-" have been renamed, which does not mean the list
    would be consistent now.*/
enum {
    /* Ship commands */
    tcm_ShipFIRST              = 1,
    tcm_ShipChangeFc           = 1,        // sid, 3 bytes FC
    tcm_ShipChangeSpeed        = 2,        // sid, 1 word
    tcm_ShipChangeWaypoint     = 3,        // sid, 2 words
    tcm_ShipChangeMission      = 4,        // sid, 1 word
    tcm_ShipChangePrimaryEnemy = 5,        // sid, 1 word
    tcm_ShipTowShip            = 6,        // sid, 1 word
    tcm_ShipChangeName         = 7,        // sid, 20 bytes
    tcm_ShipBeamDownCargo      = 8,        // sid, 7 words NTDMCS+id
    tcm_ShipTransferCargo      = 9,        // sid, 7 words NTDMCS+id
    tcm_ShipIntercept          = 10,       // sid, 1 word
    tcm_ShipChangeNeutronium   = 11,       // sid, 1 word
    tcm_ShipChangeTritanium    = 12,       // sid, 1 word
    tcm_ShipChangeDuranium     = 13,       // sid, 1 word
    tcm_ShipChangeMolybdenum   = 14,       // sid, 1 word
    tcm_ShipChangeSupplies     = 15,       // sid, 1 word
    tcm_ShipChangeColonists    = 16,       // sid, 1 word
    tcm_ShipChangeTorpedoes    = 17,       // sid, 1 word
    tcm_ShipChangeMoney        = 18,       // sid, 1 word
    tcm_ShipLAST               = 18,

    /* Planet commands */
    tcm_PlanetFIRST            = 21,
    tcm_PlanetChangeFc         = 21,       // pid, 3 bytes
    tcm_PlanetChangeMines      = 22,       // pid, 1 word <-
    tcm_PlanetChangeFactories  = 23,       // pid, 1 word
    tcm_PlanetChangeDefense    = 24,       // pid, 1 word
    tcm_PlanetChangeNeutronium = 25,       // pid, 1 dword
    tcm_PlanetChangeTritanium  = 26,       // pid, 1 dword
    tcm_PlanetChangeDuranium   = 27,       // pid, 1 dword
    tcm_PlanetChangeMolybdenum = 28,       // pid, 1 dword
    tcm_PlanetChangeColonists  = 29,       // pid, 1 dword
    tcm_PlanetChangeSupplies   = 30,       // pid, 1 dword
    tcm_PlanetChangeMoney      = 31,       // pid, 1 dword
    tcm_PlanetColonistTax      = 32,       // pid, 1 word
    tcm_PlanetNativeTax        = 33,       // pid, 1 word
    tcm_PlanetLAST             = 33,       // BuildBase is special
    tcm_PlanetBuildBase        = 34,       // pid, NO DATA

    /* Starbase commands */
    tcm_BaseFIRST              = 40,
    tcm_BaseChangeDefense      = 40,       // bid, 1 word
    tcm_BaseUpgradeEngineTech  = 41,       // bid, 1 word
    tcm_BaseUpgradeHullTech    = 42,       // bid, 1 word <-
    tcm_BaseUpgradeWeaponTech  = 43,       // bid, 1 word <-
    tcm_BaseBuildEngines       = 44,       // bid, 9 words
    tcm_BaseBuildHulls         = 45,       // bid, 20 words
    tcm_BaseBuildWeapons       = 46,       // bid, 10 words
    tcm_BaseBuildLaunchers     = 47,       // bid, 10 words
    tcm_BaseBuildTorpedoes     = 48,       // bid, 10 words
    tcm_BaseBuildFighters      = 49,       // bid, 1 word
    tcm_BaseFixRecycleShipId   = 50,       // bid, 1 word <-
    tcm_BaseFixRecycleShip     = 51,       // bid, 1 word action
    tcm_BaseChangeMission      = 52,       // bid, 1 word
    tcm_BaseBuildShip          = 53,       // bid, 7 words
    tcm_BaseUpgradeTorpTech    = 54,       // bid, 1 word
    tcm_BaseLAST               = 54,

    /* Rest */
    tcm_SendMessage            = 60,       // len, from, to, text
    tcm_ChangePassword         = 61,       // zero, 10 bytes
    tcm_SendBack               = 62        // recv, type, size, data
};

//! Turn File Handling.
class GTurnfile {
    struct CommandDef;
    static const CommandDef command_defs[];
    class CommandCompare;
    friend class CommandCompare;

    /* Turn file structure */
    TTurnHeader         trn_header;     ///< TRN header.
    TTaccomTurnHeader   taccom_header;  ///< Taccom header (TRN directory). Verbatim from turn file (offsets 1-based).
    TTurnDosTrailer     dos_trailer;    ///< DOS trailer.
    TTurnWindowsTrailer win_trailer;    ///< Windows trailer.
    Buffer<char>        data;           ///< Miscellaneous data. The TRN, usually ;)
    Buffer<int32>       offsets;        ///< Offsets of commands, pointing into data. zero-based. NOT the pointer array from the turn file!
    int                 version;        ///< TRN file sub-version (Winplan only).
    int                 features;       ///< TRN file features (bitfield TRNF_xxx).
    int                 trn_placement;  ///< Taccom: place TRN before Nth attachment.
    int                 turn_nr;

    /* Internal stuff */
    bool dirty;                 ///< True if data is dirty. If false, data is a valid turn file.

    void updateTurnfile(Buffer<char>& data, Buffer<int32>& offsets);
    void parseTurnfile(int32 offset, int32 length);
    void checkRange(int32 offset, int32 length);
    void makeCommands(int id, int low, int up, char* old, char* neu);
 public:
    /** Feature Flags. */
    enum {
        TRNF_WINPLAN = 1,
        TRNF_TACCOM  = 2
    };
    /** Manifest Constants. */
    enum {
        CURRENT_VERSION = 1,    ///< We know TRN sub-version 1 (Host 3.22).
        MAX_ATTACH      = 10    ///< Maximum number of attachments (Taccom).
    };
    /** Command Types. */
    enum {
        UNDEF_COMMAND,
        SHIP_COMMAND,
        PLANET_COMMAND,
        BASE_COMMAND,
        OTHER_COMMAND
    };

    GTurnfile(int player, const string_t& atime);
    GTurnfile(Ptr<Stream> stream);
    ~GTurnfile();

    /* Accessors */
    int      getPlayer() const { return trn_header.player; }
    string_t getTimestamp() const { return getBasicString(trn_header.timestamp); }
    int      getNumCommands() const { return offsets.getSize(); }

    int getCommand(int index) const;
    int getCommandLength(int index) const;
    int getCommandId(int index) const;
    int getCommandType(int index) const;
    int32 getCommandPosition(int index) const;
    const char* getCommandName(int index) const;
    const char* getCommandData(int index) const;

    static int         getTypeForCode(int code); // FIXME: these names suck
    static const char* getNameForCode(int code);
    static int         getRecordIndexForCode(int code);

    int getFeatures() const { return features; }
    int getVersion() const { return version; }

    void setFeatures(int f);

    /* Taccom Accessors. Since the only one to use that is currently
       un-trn, don't bloat it with stuff. */
    const TTaccomTurnHeader& getTaccomHeader() const { return taccom_header; }
    int getTaccomTurnPlace() const { return trn_placement; }

    /* Header Accessors. */
    const TTurnWindowsTrailer& getWindowsTrailer() const { return win_trailer; }
    const TTurnDosTrailer& getDosTrailer() const { return dos_trailer; }
    const TTurnHeader& getTurnHeader() const { return trn_header; }

    /* Modificators */
    void addCommand(int cmd, int id);
    void addCommand(int cmd, int id, const char* data, std::size_t len);
    void addData(const char* data, std::size_t len);

    void deleteCommand(int index);

    void update();
    void updateTrailer();
    int32 computeTurnChecksum() const;
    void setTemplock(const uint32* data);

    void write(Ptr<Stream> stream);

    void setRegInfo(const GRegInfo& reg, int turn_nr);

    /* Iteration */
    void sortCommands();

    /* Maketurn */
    // FIXME: this is probably too much an indirection
    void makeShipCommands(int id, char* old, char* neu);
    void makePlanetCommands(int id, char* old, char* neu);
    void makeBaseCommands(int id, char* old, char* neu);

    /* Taccom */
    int addFile(const char* file_data, std::size_t len, const string_t& name);
    int addFile(const Buffer<char>& data, const string_t& name);
    void deleteFile(int index);
    int getFileCount() const;
};

#endif
