/**
  *  \file tvcr.cc
  *  \brief THost VCR Implementation
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "game/tvcr.h"
#include "game/spec.h"
#include "game/config.h"
#include "game/formula.h"

/** \class VcrPlayerTHost
    \brief THost VCR Player

    This is the THost VCR player, started as a port from the PCC
    v1.0.13 player.

    It includes the speed optimisations (fast forward, psrandom
    tables), but is not yet completely tuned. Things that remain are
    - maybe widen the internally used data values (TVcr) or shorten
      those in #Status
    - maybe add some shortcuts for visualisation vs. simulation */

/* Benchmark results: `testvcr tests/vcr/vcr2.dat . 100', vcr2.dat from
   Manos-II / 35:

   + obfuscated seed advance:  2.350 user
   + 1-side instead of !side:  2.480 user
   + rand_table_type=uchar:    2.520 user
   + original:                 2.540 user
   + #undef CONF_STATIC_VCR:   2.950 user

   Fast access paths (f_getDamage/f_getKill) seems not to get much. */

#ifdef CONF_STATIC_VCR
TVcr VcrPlayerTHost::vcr;
VcrPlayerTHost::Status VcrPlayerTHost::status[2];
int VcrPlayerTHost::seed;
#endif

// /** Compute a/b+plus using IEEE rounding. The result is rounded to the
//     nearest integer. If the fractional part is exactly 0.5, rounds to
//     the nearest even integer. Uses only integer math. This is hard to
//     implement efficiently in C++ with floating point math, anyway. */
// static long
// rdivaddl(long a, int b, int plus)
// {
//     long x = a / b + plus;
//     int r = a % b;
//     if (r*2 + (x&1) > b)
//         ++x;
//     return x;
// }

// /** Compute a/b+plus using IEEE rounding. \see rdivaddl */
// static int
// rdivaddw(int a, int b, int plus)
// {
//     int x = a / b + plus;
//     int r = a % b;
//     if (r*2 + (x&1) > b)
//         ++x;
//     return x;
// }

template<class T>
static inline void
limitMax(T& v, int max)
{
    if (v > max)
        v = max;
}

static const VcrPlayerTHost::rand_table_type
rand_table_1_20[119] = {
   	9,	8,	11,	8,	5,	5,
	9,	10,	15,	2,	10,	4,	14,
	18,	1,	14,	15,	17,	2,	4,
	10,	13,	16,	17,	11,	10,	14,
	7,	2,	8,	13,	13,	18,	6,
	13,	12,	6,	12,	6,	14,	4,
	1,	20,	16,	16,	2,	8,	10,
	18,	4,	20,	16,	17,	15,	6,
	19,	16,	14,	2,	15,	11,	6,
	9,	17,	15,	4,	3,	12,	16,
	19,	12,	18,	11,	13,	13,	8,
	3,	2,	15,	5,	12,	6,	10,
	6,	9,	16,	20,	19,	18,	17,
	11,	1,	4,	12,	7,	13,	15,
	5,	7,	12,	3,	3,	7,	14,
	10,	18,	13,	3,	16,	14,	4,
	13,	9,	14,	2,	9,	7,	4,	15
};

static const VcrPlayerTHost::rand_table_type
rand_table_1_100[119] = {
	42,	36,	54,	39,	23,	21,
	41,	45,	73,	5,	47,	14,	71,
	89,	2,	70,	76,	83,	5,	16,
	50,	64,	78,	87,	53,	47,	66,
	33,	5,	37,	63,	61,	88,	29,
	62,	58,	26,	61,	30,	67,	16,
	2,	98,	78,	81,	7,	37,	46,
	88,	15,	99,	77,	82,	75,	25,
	96,	79,	69,	5,	71,	54,	25,
	43,	87,	75,	17,	13,	58,	78,
	96,	57,	87,	52,	63,	64,	36,
	14,	5,	73,	23,	58,	29,	48,
	27,	43,	77,	99,	95,	88,	84,
	55,	2,	15,	57,	33,	61,	76,
	22,	31,	61,	11,	13,	31,	70,
	45,	92,	61,	11,	80,	71,	14,
	62,	44,	70,	4,	40,	32,	18,	74
};

static const VcrPlayerTHost::rand_table_type
rand_table_1_17[119] = {
	8,	7,	10,	7,	5,	4,
	7,	8,	13,	2,	8,	3,	12,
	15,	1,	12,	13,	14,	2,	3,
	9,	11,	13,	15,	9,	8,	12,
	6,	2,	7,	11,	11,	15,	5,
	11,	10,	5,	11,	6,	12,	3,
	1,	17,	14,	14,	2,	7,	8,
	15,	3,	17,	13,	14,	13,	5,
	16,	14,	12,	2,	12,	10,	5,
	8,	15,	13,	4,	3,	10,	13,
	16,	10,	15,	9,	11,	11,	7,
	3,	2,	13,	5,	10,	5,	9,
	5,	8,	13,	17,	16,	15,	14,
	10,	1,	3,	10,	6,	11,	13,
	4,	6,	11,	3,	3,	6,	12,
	8,	16,	11,	3,	14,	12,	3,
	11,	8,	12,	2,	7,	6,	4,	13
};

void
VcrPlayerTHost::Status::clear()
{
    for (int i = 0; i < VCR_MAX_BEAMS; ++i)
        beam_status[i] = 0;
    for (int i = 0; i < VCR_MAX_TORPS; ++i)
        torp_status[i] = 0;
    for (int i = 0; i < VCR_MAX_FTRS; ++i)
        ftr_x[i] = ftr_active[i] = 0;
    cur_x = 0;
    damage_limit = 100;
}

/** Random number between 1 and 20. The infamous VCR random number
    generator. Fetches the value from a table. */
VcrPlayerTHost::rand_type
VcrPlayerTHost::psrandom_1_20()
{
    /* The obfuscated form makes gcc generate better code at -O2
       (only two references to `seed', no in-memory increment) */
//     if (seed >= 119)
//         seed = 0;
//     return rand_table_1_20[seed++];
    return rand_table_1_20[(seed = (seed >= 119 ? 1 : seed+1))-1];
}

/** Random number between 1 and 100. The infamous VCR random number
    generator. Fetches the value from a table. */
VcrPlayerTHost::rand_type
VcrPlayerTHost::psrandom_1_100()
{
//     if (seed >= 119)
//         seed = 0;
//     return rand_table_1_100[seed++];
    return rand_table_1_100[(seed = (seed >= 119 ? 1 : seed+1))-1];
}

/** Random number between 1 and 17. The infamous VCR random number
    generator. Fetches the value from a table. */
VcrPlayerTHost::rand_type
VcrPlayerTHost::psrandom_1_17()
{
//     if (seed >= 119)
//         seed = 0;
//     return rand_table_1_17[seed++];
    return rand_table_1_17[(seed = (seed >= 119 ? 1 : seed+1))-1];
}

/** Check if planet. \return true iff side (0 or 1) is the planet. */
bool
VcrPlayerTHost::isPlanet(int side)
{
    return (side != 0) && (vcr.battle_type != 0);
}

/** Check if planet. \return true iff side (0 or 1) is a freighter. */
bool
VcrPlayerTHost::isFreighter(int side)
{
    TVcrObject& o = vcr.objects[side];
    return o.beam_count == 0
        && vcr.torp_launcher_count[side] == 0
        && o.bay_count == 0;
}

/** Hit object.
    \param side     who is hit
    \param damage   `damage' (explosive) power of weapon
    \param kill     `kill' (x-ray) power of weapon */
void
VcrPlayerTHost::hit(int side, int damage, int kill)
{
    int shld = -rdivaddw (80*damage, vcr.mass[side] + 1, 1 - vcr.shield[side]);
    if (shld < 0) {
        vcr.shield[side] = 0;
        long l = rdivaddl (-80L*shld, vcr.mass[side] + 1, vcr.objects[side].damage + 1);
        if (l > 9999)
            l = 9999;
        vcr.objects[side].damage = l;
        shld = 0;
        if (isPlanet(side) && damage > 1) {
            int beam = 10 - vcr.objects[1].damage / 10;
            // bug emulation
            if (beam <= 0) {
                if (vcr.objects[0].beam_type > 0)
                    vcr.objects[1].beam_type = vcr.objects[0].beam_type;
            } else {
                if (vcr.objects[1].beam_type > beam)
                    vcr.objects[1].beam_type = beam;
            }
        }
    }

    if (vcr.shield[side] == 0) {
        int crew = vcr.objects[side].crew;
        if (!isPlanet(side)) {
            crew = - rdivaddw (80*kill, vcr.mass[side] + 1, -crew);
            if (crew < 0)
                vcr.objects[side].crew = 0;
            else
                vcr.objects[side].crew = crew;
        }
    }
    vcr.shield[side] = shld;
}

/** \name Fighters */
//@{

/** Start a fighter. Attempts to launch a fighter.
    \pre Object has fighters. */
void
VcrPlayerTHost::launchFighter(int side)
{
    for (int i = 0; i < VCR_MAX_FTRS; ++i)
        if (! status[side].ftr_active[i]) {
            vcr.num_fighters[side]--;
            status[side].ftr_active[i] = 1;
            status[side].ftr_x[i]      = status[side].cur_x;
            vis->startFighter(side, i);
            return;
        }
}

/** Start fighter. Start fighter from specified side, if
    preconditions fulfilled. */
void
VcrPlayerTHost::launchFighters(int side)
{
    if (vcr.objects[side].bay_count > 0) {
        int n = psrandom_1_20();
        if (n <= vcr.objects[side].bay_count && vcr.num_fighters[side] > 0)
            launchFighter(side);
    }
}

/** Move fighters. */
void
VcrPlayerTHost::moveFighters()
{
    for (int i = 0; i < VCR_MAX_FTRS; ++i) {
        /* left side */
        if (status[0].ftr_active[i] == 1 && status[0].ftr_x[i] > status[1].cur_x + 10)
            status[0].ftr_active[i] = 2;
        if (status[0].ftr_active[i] == 2 && status[0].ftr_x[i] < status[0].cur_x) {
            vcr.num_fighters[0]++;
            vis->landFighter(0, i);
            status[0].ftr_active[i] = 0;
        }
        if (status[0].ftr_active[i] == 1)
            status[0].ftr_x[i] += 4;
        else if (status[0].ftr_active[i] == 2)
            status[0].ftr_x[i] -= 4;

        /* right side */
        if (status[1].ftr_active[i] == 1 && status[1].ftr_x[i] < status[0].cur_x - 10)
            status[1].ftr_active[i] = 2;
        if (status[1].ftr_active[i] == 2 && status[1].ftr_x[i] > status[1].cur_x) {
            vcr.num_fighters[1]++;
            vis->landFighter(1, i);
            status[1].ftr_active[i] = 0;
        }
        if (status[1].ftr_active[i] == 1)
            status[1].ftr_x[i] -= 4;
        else if (status[1].ftr_active[i] == 2)
            status[1].ftr_x[i] += 4;
    }
}

/** Left fighter fires. The left fighter on track i fires at its
    opponent, if it's close enough. */
void
VcrPlayerTHost::fighterShootLeft(int i)
{
    if (status[0].ftr_active[i] == 1
        && std::abs(status[0].ftr_x[i] - status[1].cur_x) < 20)
    {
        hit(1, 2, 2);
        vis->fireBeam(0, i, -1, 1, 2, 2);
    }
}

/** Right fighter fires. The right fighter on track i fires at its
    opponent, if it's close enough. */
void
VcrPlayerTHost::fighterShootRight(int i)
{
    if (status[1].ftr_active[i] == 1
        && std::abs(status[1].ftr_x[i] - status[0].cur_x) < 20)
    {
        hit(0, 2, 2);
        vis->fireBeam(1, i, -1, 1, 2, 2);
    }
}

/** Kill fighter. The fighter on track i, specified side, is killed
    and removed from the game. */
void
VcrPlayerTHost::killFighter(int i, int side)
{
    vis->killFighter(side, i);
    status[side].ftr_active[i] = 0;
}

/** Fighter stuff.

    - Fighter movement
    - Fighters fire at enemy
    - Fighter intercept */
void
VcrPlayerTHost::fighterStuff()
{
    moveFighters();
    for (int i = 0; i < VCR_MAX_FTRS; ++i) {
        fighterShootLeft(i);
        fighterShootRight(i);
    }

    if (vcr.objects[0].bay_count > 0 && vcr.objects[1].bay_count > 0) {
        for (int i = 0; i < VCR_MAX_FTRS; ++i)
            if (status[0].ftr_active[i])
                for (int j = 0; j < VCR_MAX_FTRS; ++j)
                    if (status[1].ftr_active[j])
                        if (status[0].ftr_x[i] == status[1].ftr_x[j]) {
                            /* fighter intercept */
                            int n = psrandom_1_100();
                            if (status[0].ftr_active[i] == 0) {
                                /* Tim is the king! A dead fighter can
                                   still fire! Our visualisation can't
                                   handle that, so use the short way.
                                   See tests/vcr/deadfire.vcr for an
                                   example fight. */
                                if (n >= 50)
                                    killFighter(j, 1);
                            } else {
                                /* regular fighter intercept code */
                                if (n < 50) {
                                    vis->fireBeam(1, j, i, 1, 2, 2);
                                    killFighter(i, 0);
                                } else {
                                    vis->fireBeam(0, i, j, 1, 2, 2);
                                    killFighter(j, 1);
                                }
                            }
                        }
    }
}
//@}

/** \name Beam Weapons. */
//@{

/** Recharge beams of one side. Beams recharge randomly up to
    level 100. */
void
VcrPlayerTHost::rechargeBeams(int side)
{
    const int mx = vcr.objects[side].beam_count;
    for (int i = 0; i < mx; ++i) {
        register int j = status[side].beam_status[i];
        if (psrandom_1_100() > 50 && j < 100) {
            status[side].beam_status[i] = j+1;
            vis->updateBeam(side, i);
        }
    }
}

/** Fire beam. Hits the enemy with a beam of object from, beam number
    which. */
void
VcrPlayerTHost::fireBeam(int from, int which)
{
    int charge = status[from].beam_status[which];
    GBeam& b = getBeam(vcr.objects[from].beam_type);
    int da = rdivaddw (charge * b.f_getDamage(), 100, 0);
    int ki = rdivaddw (charge * b.f_getKill(),   100, 0);

    if (raceId(vcr.objects[from].owner) == 5)
        ki *= 3;

    hit(1-from, da, ki);
    vis->fireBeam(from, -1 - which, -1, 1, da, ki);
    status[from].beam_status[which] = 0;
    vis->updateBeam(from, which);
}

/** Fire beams from specified object. Fires all beams that can
    fire. */
void
VcrPlayerTHost::fireBeams(int from)
{
    for (int i = 0; i < vcr.objects[from].beam_count; ++i) {
        int n = psrandom_1_20();
        if (n < 7 && status[from].beam_status[i] > 50)
            fireBeam(from, i);
    }
}

/** Fire at fighters. Fires the specified beam at an enemy fighter,
    if possible. */
void
VcrPlayerTHost::fireAtFighter(int from, int beam)
{
    int min_dist = 600;
    int ftr_id   = -1;
    int at       = 1-from;

    if (vcr.objects[at].bay_count == 0)
        return;

    for (int i = 0; i < VCR_MAX_FTRS; ++i) {
        if (status[at].ftr_active[i]) {
            int j = std::abs(status[at].ftr_x[i] - status[from].cur_x);
            if (j < min_dist) {
                min_dist = j;
                ftr_id   = i;
            }
        }
    }

    if (ftr_id >= 0) {
        vis->fireBeam(from, -1 - beam, ftr_id, 1, 2, 2 /* FIXME */);
        killFighter(ftr_id, at);
        status[from].beam_status[beam] = 0;
        vis->updateBeam(from, beam);
    }
}

/** Fire beams at fighter. Fire all beams that can fire at enemy
    fighters. */
void
VcrPlayerTHost::fireBeamsAtFighter(int fm)
{
    for (int i = 0; i < vcr.objects[fm].beam_count; ++i) {
        int n = psrandom_1_20();
        if (status[fm].beam_status[i] > 40 && n < 5)
            fireAtFighter(fm, i);
    }
}
//@}

/** \name Torpedoes. */
//@{

/** Fire torpedo. Launches one torpedo from the specified
    object. */
void
VcrPlayerTHost::fireTorp(int from, int launcher)
{
    register int n = psrandom_1_100();
    if (n >= 35) {
        GTorpLauncher& t = getTorp(vcr.objects[from].torp_type);
        hit(1-from, 2*t.f_getDamage(), 2*t.f_getKill());
        vis->fireTorp(from, n, launcher);
    } else {
        vis->fireTorp(from, -n, launcher);
    }
}

/** Fire torpedo. Fires all ready torpedoes from the specified
    object at the enemy. */
void
VcrPlayerTHost::fireTorpedoes(int from)
{
    for (int i = 0; i < vcr.torp_launcher_count[from]; ++i) {
        if (vcr.num_torps[from] > 0) {
            int n = psrandom_1_17();
            if (status[from].torp_status[i] > 40
                || (status[from].torp_status[i] > 30 && n < vcr.objects[from].torp_type))
            {
                vcr.num_torps[from]--;
                status[from].torp_status[i] = 0;
                vis->updateTorp(from, i);
                fireTorp(from, i);
            }
        }
        status[from].torp_status[i]++;
        vis->updateTorp(from, i);
    }
}
//@}

/** \name Red Tape. */
//@{

/** Charge weapons. If the specified object has full shields,
    it also starts with fully-charged weapons. */
void
VcrPlayerTHost::preloadWeapons(int side)
{
    if (vcr.shield[side] == 100) {
        for (int i = 0; i < VCR_MAX_BEAMS; ++i) {
            status[side].torp_status[i] = 30;
            status[side].beam_status[i] = 100;
        }
    }
}

/** Init. Set up for playing. */
void
VcrPlayerTHost::initVcr(TVcr& avcr)
{
    vcr = avcr;
    if (checkVcr(vcr)) {
        status_word = VCRS_INVALID;
        return;
    }

    /* okay, we'll be able to play it. */
    status_word = 0;

    seed = vcr.random_seed % 119;

    /* clear status */
    for (int i = 0; i < 2; ++i)
        status[i].clear();

    status[0].cur_x = 30;       // ship
    if (vcr.battle_type == 0)
        status[1].cur_x = 610;
    else
        status[1].cur_x = 570;

    limitMax(vcr.shield[0], 100 - vcr.objects[0].damage);
    limitMax(vcr.shield[1], 100 - vcr.objects[1].damage);

    if (vcr.shield[0] < 0 || isFreighter(0))
        vcr.shield[0] = 0;
    if (vcr.battle_type == 0) {
        if (isFreighter(1))
            vcr.shield[1] = 0;
    } else {
        if (vcr.objects[1].crew <= 0)
            vcr.shield[1] = 0;
    }
    if (vcr.shield[1] < 0)
        vcr.shield[1] = 0;

    preloadWeapons(0);
    preloadWeapons(1);

    status[0].damage_limit = (raceId(vcr.objects[0].owner) == 2) ? 151 : 100;
    status[1].damage_limit = (raceId(vcr.objects[1].owner) == 2) ? 151 : 100;

    time = 0;
    vis->init();
}

/** Check one side of VCR. Fixes out-of-range values, and
    returns true iff there was such a fix needed. */
bool
VcrPlayerTHost::checkVcrSide(TVcr& vcr, int side)
{
    bool err = false;
    TVcrObject& o = vcr.objects[side];

    if (o.owner <= 0 || o.owner > 12)
        o.owner = 12;

    if (o.beam_type < 0 || o.beam_type > NUM_BEAMS) {
        o.beam_type = o.beam_count = 0;
        err = true;
    }

    if (o.torp_type < 0 || o.torp_type > NUM_TORPS) {
        o.torp_type = vcr.torp_launcher_count[side] = 0;
        err = true;
    }

    if (o.beam_count > VCR_MAX_BEAMS) {
        o.beam_count = VCR_MAX_BEAMS;
        err = true;
    }

    if (vcr.torp_launcher_count[side] > VCR_MAX_TORPS) {
        vcr.torp_launcher_count[side] = VCR_MAX_TORPS;
        err = true;
    }

    return err;
}

/** Check VCR. */
bool
VcrPlayerTHost::checkVcr(TVcr& vcr)
{
    bool l = checkVcrSide(vcr, 0);
    bool r = checkVcrSide(vcr, 1);
    return l || r;
}

/** Finish up VCR. Take back fighters, explode, compute status. */
void
VcrPlayerTHost::doneVcr(TVcr& avcr)
{
    if (status_word == VCRS_INVALID) {
        avcr = vcr;
        return;
    }

    // FIXME: explode fighters when mothership explodes -- NOT! too
    // expensive in visualisation
    for (int i = 0; i < VCR_MAX_FTRS; ++i) {
        if (status[0].ftr_active[i]) {
            vcr.num_fighters[0]++;
            vis->landFighter(0, i);
            status[0].ftr_active[i] = 0;
        }
        if (status[1].ftr_active[i]) {
            vcr.num_fighters[1]++;
            vis->landFighter(1, i);
            status[1].ftr_active[i] = 0;
        }
    }

    status_word = 0;
    if (vcr.battle_type != 0) {
        if (vcr.objects[0].damage >= 100 || vcr.objects[0].crew <= 0)
            status_word |= VCRS_LEFT_DESTROYED;
        if (vcr.objects[1].damage >= 100)
            status_word |= VCRS_RIGHT_DESTROYED;
    } else {
        if (vcr.objects[0].damage >= status[0].damage_limit) {
            status_word |= VCRS_LEFT_DESTROYED;
        } else if (vcr.objects[0].crew <= 0) {
            if (vcr.objects[0].damage < status[1].damage_limit)
                status_word |= VCRS_LEFT_CAPTURED;
            else
                status_word |= VCRS_LEFT_DESTROYED;
        }
        if (vcr.objects[1].damage >= status[1].damage_limit) {
            status_word |= VCRS_RIGHT_DESTROYED;
        } else if (vcr.objects[1].crew <= 0) {
            if (vcr.objects[1].damage < status[0].damage_limit)
                status_word |= VCRS_RIGHT_CAPTURED;
            else
                status_word |= VCRS_RIGHT_DESTROYED;
        }
    }

    if (status_word == 0)
        status_word = VCRS_TIMEOUT;

    if (status_word & VCRS_LEFT_DESTROYED)
        vis->killObject(0);
    if (status_word & VCRS_RIGHT_DESTROYED)
        vis->killObject(1);

    avcr = vcr;
}

bool
VcrPlayerTHost::setCapabilities(uint16 cap)
{
    return (cap == 0);
}

/** Play one cycle. */
bool
VcrPlayerTHost::playCycle()
{
    if (status_word != 0)
        return false;

    if (vcr.objects[0].damage >= status[0].damage_limit
        || vcr.objects[1].damage >= status[1].damage_limit)
        return false;
    if (vcr.objects[0].crew <= 0 || vcr.objects[1].crew <= 0)
        if (vcr.battle_type == 0)
            return false;
    if (time >= 2000)
        return false;

    ++time;

    /* Movement */
    int distance = status[1].cur_x - status[0].cur_x;
    if (distance > 30) {
        ++status[0].cur_x;
        --distance;
        if (vcr.battle_type == 0) {
            --status[1].cur_x;
            --distance;
        }
    }

    /* Beams */
    if (distance < 200)
        fireBeams(0);
    fireBeamsAtFighter(0);
    fireBeamsAtFighter(1);
    if (distance < 200)
        fireBeams(1);

    /* Torpedoes */
    if (distance < 300) {
        fireTorpedoes(0);
        fireTorpedoes(1);
    }

    /* Fighters */
    launchFighters(0);
    launchFighters(1);
    if (vcr.objects[0].bay_count > 0 || vcr.objects[1].bay_count > 0)
        fighterStuff();

    /* Recharge beams */
    rechargeBeams(0);
    rechargeBeams(1);

    return true;
}

/** Fast forward. In torp/torp battles, nothing happens until ships are
    in torpedo range. So start there instead.
    \todo the same applies to freighters */
void
VcrPlayerTHost::playFastForward()
{
    if (status_word || time || vcr.battle_type)
        /* not applicable if we already have a result, have played
           a tick, or fighting a planet. */
        return;

    if (vcr.objects[0].bay_count || vcr.objects[1].bay_count)
        return;

    if (vcr.shield[0] != 100 || vcr.shield[1] != 100)
        return;

    /* Start at distance 304 (to compensate off-by-one errors,
       300 should suffice ;-) */
    status[0].cur_x = 168;
    status[1].cur_x = 472;
    time = 138;

    seed = (seed + 138 * 2 * (vcr.objects[0].beam_count + vcr.objects[1].beam_count)) % 119;
}
//@}

/** \name Accessors */
//@{
int
VcrPlayerTHost::getBeamStatus(int side, int id)
{
    return status[side].beam_status[id];
}

int
VcrPlayerTHost::getTorpStatus(int side, int id)
{
    /* scale [0,40] to [0,100] */
    return status[side].torp_status[id] * 5 / 2;
}

int
VcrPlayerTHost::getTorpCount(int side)
{
    if (vcr.torp_launcher_count[side])
        return vcr.num_torps[side];
    return 0;
}

int
VcrPlayerTHost::getFighterCount(int side)
{
    if (vcr.objects[side].bay_count)
        return vcr.num_fighters[side];
    return 0;
}

int
VcrPlayerTHost::getShields(int side)
{
    return vcr.shield[side];
}

int
VcrPlayerTHost::getCrew(int side)
{
    return vcr.objects[side].crew;
}

int
VcrPlayerTHost::getDamage(int side)
{
    return vcr.objects[side].damage;
}

int
VcrPlayerTHost::getFighterX(int side, int id)
{
    return status[side].ftr_x[id];
}

int
VcrPlayerTHost::getFighterStatus(int side, int id)
{
    return status[side].ftr_active[id];
}

int
VcrPlayerTHost::getObjectX(int side)
{
    return status[side].cur_x;
}

long
VcrPlayerTHost::getDistance()
{
    return (status[1].cur_x - status[0].cur_x) * 100L;
}
//@}

VcrPlayerTHost::StatusToken::StatusToken(VcrPlayerTHost::time_type t)
    : VcrStatus(t)
{ }

/** Save status. The easy mindless way -- just save everything. */
Ptr<VcrStatus>
VcrPlayerTHost::getStatusToken()
{
    Ptr<StatusToken> t = new StatusToken(time);
    t->status[0]       = status[0];
    t->status[1]       = status[1];
    t->seed            = seed;
    t->status_word     = status_word;
    t->vcr             = vcr;
    return t;
}

/** Restore status. */
void
VcrPlayerTHost::setStatus(Ptr<VcrStatus> st)
{
    StatusToken* t = checked_cast<StatusToken*>(st.ptr());
    status[0]      = t->status[0];
    status[1]      = t->status[1];
    seed           = t->seed;
    status_word    = t->status_word;
    vcr            = t->vcr;
    time           = t->time;
}

/** Get Player Name. */
string_t
VcrPlayerTHost::getPlayerName() const
{
    return "Host";
}
