/**
  *  \file game/un-trn.cc
  *  \brief Portable Un-Trn
  *
  *  Main code for UN-TRN.
  *
  *  i18n note: do not translate the output of this program; this is
  *  intended to be in English to be (possibly) postprocessed by
  *  someone (e.g. Serge's Turnback script).
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <iostream>
#include <cstdlib>
#include <cstdio>
#include "game/trn.h"
#include "fileio/streams.h"
#include "fileio/io.h"
#include "util/except.h"
#include "game/checksum.h"
#include "game/un-trnflt.h"
#include "util/nls.h"
#include "cpluslib/format.h"

#define AS_INDENT 20
#define COM_INDENT 50

#define VERSION "1.99.2"

static char* progname;
static bool  opt_comments = true;

/** Integer to Hex String conversion. */
static string_t
itox(long l)
{
    char buf[100];
    std::sprintf(buf, "%08lX", (unsigned long) l);
    return string_t(buf);
}

/** Exit with error. */
static void
die(string_t str)
{
    std::cerr << progname << ": " << str << std::endl;
    std::exit(1);
}

/** Quote a string, C-like. */
static string_t
quoteString(const string_t& value)
{
    string_t result = "\"";
    for (string_t::size_type i = 0; i < value.size(); ++i)
        switch(value[i]) {
         case '\n':
            result += "\\n";
            break;
         case '\r':
            result += "\\r";
            break;
         case '\\':
         case '\"':
            result += '\\';
            result += value[i];
            break;
         default:
            if (value[i] >= 0 && value[i] < 32) {
                result += '\\';
                result += '0' + value[i] / 64;
                result += '0' + value[i] / 8 % 8;
                result += '0' + value[i] % 8;
            } else {
                result += value[i];
            }
        }
    return result + "\"";
}

/** Generate /len/ spaces. Also handles negative lengths. */
static string_t
spaces(int len)
{
    if (len > 0)
        return string_t(len, ' ');
    else
        return string_t();
}

/** Show a line in output, most general version.
    \param name    left side
    \param value   rhs of "=", may be null (no assignment then)
    \param comment comment, may be null */
static void
showLine(const char* name, const char* value, const char* comment)
{
    string_t output = name;
    if (value)
        output += spaces(AS_INDENT - output.length()) + " = " + value;
    if (comment && *comment && opt_comments) {
        if (output.length() < COM_INDENT - 2)
            output += spaces(COM_INDENT - output.length()) + "; " + comment;
        else
            output += "\n" + spaces(COM_INDENT) + "; " + comment;
    }
    std::cout << output << "\n";
}

/** Show `name=int' pair. */
static void
var(string_t name, int32 value)
{
    showLine(name.c_str(), itoa(value).c_str(), 0);
}

/** Show `name=int' pair with comment. */
static void
var(string_t name, int32 value, string_t comment)
{
    showLine(name.c_str(), itoa(value).c_str(), comment.c_str());
}

/** Show `name=string' pair. */
static void
var(string_t name, string_t value)
{
    showLine(name.c_str(), quoteString(value).c_str(), 0);
}

/** Show `name=string' pair with comment. */
static void
var(string_t name, string_t value, string_t comment)
{
    showLine(name.c_str(), quoteString(value).c_str(), comment.c_str());
}

/** Show Taccom header.
    \param taccom   The Taccom header itself
    \param place    Conceptual position of TRN between attachments */
static void
showTaccom(const TTaccomTurnHeader& taccom, int place)
{
    bool shown = false;
    std::cout << ";\n; Taccom-format Turn File Directory:\n";
    for (int i = 0; i < int(countof(taccom.attach)); ++i) {
        if (place == i) {
            std::cout << ";   turn data, " << taccom.turn_size << " bytes, position " << taccom.turn_pos
                      << "\n";
            shown = true;
        }
        string_t name = getBasicString(taccom.attach[i].name);
        if (name.length()) {
            std::cout << ";   file `" << name << "', " << taccom.attach[i].file_size
                      << " bytes, position " << taccom.attach[i].file_pos << "\n";
        }
    }
    if (!shown)
        std::cout << ";   turn data, " << taccom.turn_size << " bytes, position " << taccom.turn_pos
                  << "\n";
}

/** Show parameters of a `word' command.
    \param data  command data
    \param n     number of parameters
    \param text  names, N consecutive strings ("first\0second\0third") */
static void
showWord(const char* data, int n, const char* text)
{
    while (n--) {
        string_t title = text;
        int16 value = getInt16(data);
        var("  " + title, value);
        text += title.length() + 1;
        data += 2;
    }
}

/** Show parameters of a `word' command with "indexed" names.
    Indexed commands are used for starbase storage.
    \param data  command data
    \param n     number of parameters
    \param text  names. Used as `text1', `text2', ... */
static void
showWordIndexed(const char* data, int n, const char* text)
{
    int i = 0;
    while (i < n) {
        string_t title = text + itoa(++i);
        int16 value = getInt16(data);
        var("  " + title, value);
        data += 2;
    }
}

/** Show parameters of a `dword' command.
    \param data  command data
    \param n     number of parameters
    \param text  names, N consecutive strings ("first\0second\0third") */
static void
showLong(const char* data, int n, const char* text)
{
    while (n--) {
        string_t title = text;
        int32 value = getInt32(data);
        var("  " + title, value);
        text += title.length() + 1;
        data += 4;
    }
}

/** Show a message.
    \param length message
    \param text   encrypted message */
static void
showMessage(int length, const char* text)
{
    string_t msg(text, length);
    for (string_t::size_type i = 0; i < msg.length(); ++i)
        msg[i] -= 13;

    while (msg.length()) {
        string_t::size_type i = msg.find('\r');
        string_t sub = msg.substr(0, i);
        msg.erase(0, i);
        if (msg.length())
            msg.erase(0, 1);
        std::cout << "    " << quoteString(sub) << "\n";
    }
}

static void
showHex(const char* data, std::size_t size)
{
    using namespace std;
    bool trunc = false;
    if (size > 256) {
        size = 256;
        trunc = true;
    }

    size_t i = 0;
    while (i < size) {
        std::cout << "  " << itox(i) << ": ";
        for (size_t c = 0; c < 16; ++c) {
            char buf[10];
            if (i+c > size) {
                cout << "  ";
            } else {
                sprintf(buf, "%02X", (unsigned char) data[i+c]);
                cout << buf;
            }
            cout << " ";
            if (c == 7)
                cout << " ";
        }
        for (std::size_t c = 0; c < 16 && i+c < size; ++c)
            if (isprint(data[i+c]))
                cout << data[i+c];
            else
                cout << ".";
        cout << "\n";
        i += 16;
    }

    if (trunc)
        std::cout << "; ...rest omitted\n";
}

static void
showUtilData(const char* data)
{
    uint16 size = getUint16(data, 2);
    uint16 type = getUint16(data, 0);
    if (type == 34) {
        var("    File Name", getBasicStringN(data+4, 12));
        var("    File Size", size-13);
        var("    Flags", data[16]);
        showHex(data+17, size-13);
    } else {
        showHex(data+4, size);
    }
}

/** Show one turn command. */
static void
showCommand(GTurnfile& trn, int index)
{
    int code = trn.getCommand(index);
    int type = trn.getCommandType(index);
    const char* name = trn.getCommandName(index);
    int32 thispos = trn.getCommandPosition(index);

    if (!name) {
        var("Command", code, "index " + itoa(index+1) + ", position " + itox(thispos));
        std::cout << "; Unknown command\n";
        return;
    }

    showLine(name, 0, ("index " + itoa(index+1) + ", position " + itox(thispos)).c_str());
    switch(type) {
     case GTurnfile::SHIP_COMMAND:
        var("  Ship Id", trn.getCommandId(index));
        break;
     case GTurnfile::PLANET_COMMAND:
        var("  Planet Id", trn.getCommandId(index));
        break;
     case GTurnfile::BASE_COMMAND:
        var("  Base Id", trn.getCommandId(index));
        break;
    }

    const char* data = trn.getCommandData(index);
    switch(code) {
     case tcm_ShipChangeFc:
     case tcm_PlanetChangeFc:       // pid, 3 bytes
        var("  FCode", getBasicStringN(data, 3));
        break;
     case tcm_ShipChangeSpeed: // sid, 1 word
        showWord(data, 1, "Speed");
        break;
     case tcm_ShipChangeWaypoint: // sid, 2 words
        showWord(data, 2, "WaypointDX\0WaypointDY");
        break;
     case tcm_ShipChangeMission: // sid, 1 word
        showWord(data, 1, "Mission");
        break;
     case tcm_ShipChangePrimaryEnemy: // sid, 1 word
        showWord(data, 1, "Player");
        break;
     case tcm_ShipTowShip: // sid, 1 word
        showWord(data, 1, "Towee Id");
        break;
     case tcm_ShipChangeName: // sid, 20 bytes
        var("  Name", getBasicStringN(data, 20));
        break;
     case tcm_ShipBeamDownCargo:        // sid, 7 words NTDMCS+id
     case tcm_ShipTransferCargo:        // sid, 7 words NTDMCS+id
        showWord(data, 7, "Neutronium\0Tritanium\0Duranium\0"
                 "Molybdenum\0Clans\0Supplies\0Target Id");
        break;
     case tcm_ShipIntercept:       // sid, 1 word
        showWord(data, 1, "Target Id");
        break;
     case tcm_ShipChangeNeutronium:       // sid, 1 word
        showWord(data, 1, "Neutronium");
        break;
     case tcm_ShipChangeTritanium:       // sid, 1 word
        showWord(data, 1, "Tritanium");
        break;
     case tcm_ShipChangeDuranium:       // sid, 1 word
        showWord(data, 1, "Duranium");
        break;
     case tcm_ShipChangeMolybdenum:       // sid, 1 word
        showWord(data, 1, "Molybdenum");
        break;
     case tcm_ShipChangeSupplies:       // sid, 1 word
        showWord(data, 1, "Supplies");
        break;
     case tcm_ShipChangeColonists:       // sid, 1 word
        showWord(data, 1, "Clans");
        break;
     case tcm_ShipChangeTorpedoes:       // sid, 1 word
        showWord(data, 1, "Ammo");
        break;
     case tcm_ShipChangeMoney:       // sid, 1 word
        showWord(data, 1, "Money");
        break;

    /* Planet commands */
     case tcm_PlanetChangeMines:       // pid, 1 word <-
        showWord(data, 1, "Mines");
        break;
     case tcm_PlanetChangeFactories:       // pid, 1 word
        showWord(data, 1, "Factories");
        break;
     case tcm_PlanetChangeDefense:       // pid, 1 word
        showWord(data, 1, "Defense");
        break;
     case tcm_PlanetChangeNeutronium:       // pid, 1 dword
        showLong(data, 1, "Neutronium");
        break;
     case tcm_PlanetChangeTritanium:       // pid, 1 dword
        showLong(data, 1, "Tritanium");
        break;
     case tcm_PlanetChangeDuranium:       // pid, 1 dword
        showLong(data, 1, "Duranium");
        break;
     case tcm_PlanetChangeMolybdenum:       // pid, 1 dword
        showLong(data, 1, "Molybdenum");
        break;
     case tcm_PlanetChangeColonists:       // pid, 1 dword
        showLong(data, 1, "Clans");
        break;
     case tcm_PlanetChangeSupplies:       // pid, 1 dword
        showLong(data, 1, "Supplies");
        break;
     case tcm_PlanetChangeMoney:       // pid, 1 dword
        showLong(data, 1, "Money");
        break;
     case tcm_PlanetColonistTax:       // pid, 1 word
     case tcm_PlanetNativeTax:       // pid, 1 word
        showWord(data, 1, "Tax Rate");
        break;
     case tcm_PlanetBuildBase:       // pid, NO DATA
        std::cout << "; no data for this command\n";
        break;

    /* Starbase commands */
     case tcm_BaseChangeDefense:       // bid, 1 word
        showWord(data, 1, "Defense");
        break;
     case tcm_BaseUpgradeEngineTech:       // bid, 1 word
     case tcm_BaseUpgradeHullTech:       // bid, 1 word <-
     case tcm_BaseUpgradeWeaponTech:       // bid, 1 word <-
     case tcm_BaseUpgradeTorpTech:       // bid, 1 word
        showWord(data, 1, "Tech");
        break;
     case tcm_BaseBuildEngines:       // bid, 9 words
        showWordIndexed(data, 9, "Engine");
        break;
     case tcm_BaseBuildHulls:       // bid, 20 words
        showWordIndexed(data, 20, "Hull");
        break;
     case tcm_BaseBuildWeapons:       // bid, 10 words
        showWordIndexed(data, 10, "Beam");
        break;
     case tcm_BaseBuildLaunchers:       // bid, 10 words
        showWordIndexed(data, 10, "Launcher");
        break;
     case tcm_BaseBuildTorpedoes:       // bid, 10 words
        showWordIndexed(data, 10, "Torp");
        break;
     case tcm_BaseBuildFighters:       // bid, 1 word
        showWord(data, 1, "Fighters");
        break;
     case tcm_BaseFixRecycleShipId:       // bid, 1 word <-
        showWord(data, 1, "Ship Id");
        break;
     case tcm_BaseFixRecycleShip:       // bid, 1 word action
        switch (int16 v = getInt16(data)) {
         case 0:
            var("  Action", 0, "none");
            break;
         case 1:
            var("  Action", 1, "Fix");
            break;
         case 2:
            var("  Action", 2, "Recycle");
            break;
         default:
            var("  Action", v, "INVALID");
        }
        break;
     case tcm_BaseChangeMission:       // bid, 1 word
        showWord(data, 1, "Mission");
        break;
     case tcm_BaseBuildShip:       // bid, 7 words
        showWord(data, 7, "Hull Type\0Engine Type\0Beam Type\0Beam Count\0"
                 "Torp Type\0Torp Count\0Unused");
        break;

     /* Rest */
     case tcm_SendMessage:       // len, from, to, text
        showWord(data, 2, "From\0To");
        std::cout << "  Text =\n";
        showMessage(trn.getCommandId(index), data + 4);
        break;
     case tcm_ChangePassword:       // zero, 10 bytes
        std::cout << "; Intentionally not decoded.\n";
        break;
     case tcm_SendBack:
        var("  Receiver", trn.getCommandId(index));
        showWord(data, 2, "Type\0Size");
        showUtilData(data);
        break;
    }

    /* Diagnose positioning errors */
    if (index < trn.getNumCommands()-1) {
        int32 nextpos = trn.getCommandPosition(index+1);
        int32 thislen = trn.getCommandLength(index) + 4;
        int32 endpos = thispos + thislen;
        if (endpos != nextpos) {
            std::cout << "; WARNING: next command not at expected position "
                      << itox(endpos) << "\n";
            if (endpos < nextpos) {
                std::cout << "; there's a " << (nextpos-endpos) << " bytes gap\n";
            } else {
                int32 overlap = endpos - nextpos;
                if (overlap < thislen) {
                    std::cout << "; there's a " << overlap << " bytes overlap between commands\n";
                } else {
                    /* next command is at/before the start of this one.
                       How do we name that? */
                    std::cout << "; this TRN is screwed.\n";
                }
            }
        }
    }
}

/** Show turn file header. */
static void
showHeader (const GTurnfile& trn, const char* filename)
{
    std::cout << ";==================================================\n"
                 "; Listing of `" << filename << "'\n"
                 ";==================================================\n"
                 ";\n";
    if (trn.getFeatures() & trn.TRNF_TACCOM)
        std::cout << "; Taccom file format\n";
    if (trn.getFeatures() & trn.TRNF_WINPLAN)
        std::cout << "; Winplan trailer present, sub-version " << trn.getVersion() << "\n";
    if (trn.getFeatures() & trn.TRNF_TACCOM)
        showTaccom(trn.getTaccomHeader(), trn.getTaccomTurnPlace());
    std::cout << "\n";
    var("Player", trn.getPlayer());
    var("Commands", trn.getNumCommands());
    var("Timestamp", trn.getTimestamp());

    int32 real_sum = getChecksum(trn.getTurnHeader().timestamp, 18);
    string_t result;
    /* I'd like to use the ternary operator here, but Borland doesn't
       let me :-| */
    if (trn.getTurnHeader().time_checksum == real_sum)
        result = string_t("okay");
    else
        result = "WRONG, should be " + itoa(real_sum);
    var("Time checksum", trn.getTurnHeader().time_checksum, result);
}

/** Decode reg string, Winplan trailer. */
static string_t
decode(const char t[2][25])
{
    char res[25];
    for (int i = 0; i < 25; ++i)
        res[i] = t[0][i] ^ t[1][i];
    return getBasicString(res);
}

/** Decode reg string, standard trailer. */
static string_t
decode13(const int32 t[], bool& flag)
{
    char res[25];
    for (int i = 0; i < 25; ++i) {
        if (t[i] % (13+13*i)) {
            flag = true;
            res[i] = '?';
        } else {
            res[i] = t[i] / (13+13*i);
        }
    }
    return getBasicString(res);
}

/** Show turn trailer(s). */
static void
showTrailer(const GTurnfile& trn)
{
    if (trn.getFeatures() & trn.TRNF_WINPLAN) {
        /* Windows trailer */
        std::cout << "; Version 3.5 file format (Winplan)\n"
                  << "; Sub-version " << trn.getVersion() << "\n";
        const TTurnWindowsTrailer& wt = trn.getWindowsTrailer();
        showLine("VPH A", itox(wt.vph[0]).c_str(), ("-> VPH = " + itox(wt.vph[0] ^ wt.vph[1])).c_str());
        showLine("VPH B", itox(wt.vph[1]).c_str(), 0);
        var("RegStr1", decode(wt.regstr1));
        var("RegStr2", decode(wt.regstr2));
        var("RegStr3", getBasicString(wt.regstr3), "Player Name");
        var("RegStr4", getBasicString(wt.regstr4), "Player Address");
        std::cout << "\n; DOS Trailer follows:\n";
    } else {
        std::cout << "; Version 3.0 file format (DOS)\n";
    }

    /* DOS Trailer */
    int32 csum = trn.computeTurnChecksum();
    const TTurnDosTrailer& dt = trn.getDosTrailer();
    if (csum == dt.checksum)
        showLine("Checksum", itox(dt.checksum).c_str(), "okay");
    else
        showLine("Checksum", itox(dt.checksum).c_str(),
                 ("ERROR: should be " + itox(csum)).c_str());

    const char* mtname =
        dt.sig == 0x32434350 || dt.sig == 0x49494343 ? "PCC II"
        : dt.sig == 0x21434350 ? "PCC"
        : dt.sig == 0x2E522E53 ? "Stefan's Portable Maketurn"
        : dt.sig == 0x6F72656B || dt.sig == 0x6F72654B ? "k-Maketurn"
        : dt.sig == 0 ? "Tim's Maketurn or VPmaketurn"
        : "Tim's Maketurn";
    showLine("Unused", itox(dt.sig).c_str(), mtname);

    bool res = false;
    var("RegStr1", decode13(dt.reg, res));
    var("RegStr2", decode13(dt.reg + 25, res));
    int32 rsum = 668;
    for (int i = 0; i < 50; ++i)
        rsum += dt.reg[i];
    if (rsum == dt.reg[50])
        showLine("RegSum", itox(dt.reg[50]).c_str(), "okay");
    else
        showLine("RegSum", itox(dt.reg[50]).c_str(),
                 ("ERROR: should be " + itox(rsum)).c_str());
    if (res)
        std::cout << "; WARNING: Encoding error (indicated with `?')\n";

    std::cout << "\nPlayerLog\n";
    for (int i = 0; i < 11; ++i)
        showLine(("  Player" + itoa(i+1)).c_str(), itox(dt.players[i]).c_str(), 0);
}

/** Friendly message (-h). */
static void
help()
{
    std::cout << format(_("Turn File Decompiler v%s - (c) 2001,2002,2003 Stefan Reuther\n"), VERSION)
              << format(_("\n"
                          "Usage:\n"
                          "  %s [-h|-v]\n"
                          "  %0$s [-pPtTsckr] [-d] [-f expr] playerX.trn\n\n"
                          "Filter expressions:\n"
                          "  a&b, a|b, !a, (a), command, com*, id, id-id, true, false, #num, #num-num\n\n"
                          "About:\n"
                          " -h        this help summary\n"
                          " -v        show version number\n"
                          "\n"
                          "Options:\n"
                          " -p/-P     show/don't show turn header\n"
                          " -t/-T     show/don't show turn trailer\n"
                          " -f expr   show only matching commands\n"
                          " -s        sort turn before doing anything\n"
                          " -c        do not show most comments\n"
                          " -a        ignore Taccom attachments (delete them if -r used)\n"
                          "\n"
                          "Actions:\n"
                          " -r        re-write turn\n"
                          " -d        delete matching commands (ignores -pPtT, implies -r)\n"
                          " -n        no output (ignores -pPtT)\n"
                          "\n"
                          "Report bugs to <Streu@gmx.de>\n"),
                        progname);
    std::exit(0);
}

/** Friendly message (-v). */
static void
version()
{
    std::cout << format(_("Turn File Decompiler v%s - (c) 2001,2002 Stefan Reuther\n"), VERSION);
    std::exit(0);
}

static void
saveTurn(GTurnfile& trn, const char* fn)
{
    trn.update();
    Ptr<Stream> file = Stream::openFile(fn, Stream::C_CREATE);
    trn.write(file);
    std::cout << format(_("Wrote turn file with %d command%!1{s%}\n"), trn.getNumCommands());
}

/** Main program. Parameter parsing, loading, control. */
int
main(int /*argc*/, char** argv)
{
    enum { List, Delete, Rewrite, Nothing } opt_action = List;
    const char*           opt_filename = 0;
    Ptr<Filter>           opt_filter = 0;
    bool                  opt_kill_taccom = false;
    //bool                  opt_rewrite = false;
    int                   opt_header = 0;             // 1=do, 2=don't
    int                   opt_trailer = 0;            // 1=do, 2=don't
    bool                  opt_sort = false;

    progname = *argv;

    initNLS();
    while (char* p = *++argv) {
        if (*p == '-')
            while (*++p) {
                switch (*p) {
                 case 'h':
                    help();
                 case 'v':
                    version();
                    break;
                 case 'p':
                    opt_header = 1;
                    break;
                 case 'P':
                    opt_header = 2;
                    break;
                 case 't':
                    opt_trailer = 1;
                    break;
                 case 'T':
                    opt_trailer = 2;
                    break;
                 case 's':
                    opt_sort = true;
                    break;
                 case 'a':
                    opt_kill_taccom = true;
                    break;
                 case 'c':
                    opt_comments = false;
                    break;
                 case 'f':
                    if (! *++argv)
                        die(_("option `-f' needs an argument (the filter expression)"));
                    try {
                        Ptr<Filter> f = parseFilter(*argv);
                        if (opt_filter)
                            opt_filter = new FilterOr(opt_filter, f);
                        else
                            opt_filter = f;
                    }
                    catch (ParseException& ex) {
                        die(_("syntax error in filter expression: ") + string_t(ex.what()));
                    }
                    break;
                 case 'd':
                    opt_action = Delete;
                    break;
                 case 'n':
                    opt_action = Nothing;
                    break;
                 case 'r':
                    opt_action = Rewrite;
                    break;
                 default:
                    die(format(_("invalid option specified. Use `%s -h' for help"), progname));
                    break;
                }
            }
        else if (!opt_filename)
            opt_filename = p;
        else
            die(_("too many arguments"));
    }
    if (!opt_filename)
        die(format(_("no file name specified. Use `%s -h' for help"), progname));
    if (opt_action == Delete && !opt_filter)
        die(_("no filter specified. Deleting (`-d') needs a filter"));

    if (!opt_header)
        opt_header = opt_filter ? 2 : 1;
    if (!opt_trailer)
        opt_trailer = opt_filter ? 2 : 1;

    int status = opt_filter ? 2 : 0;

    try {
        Ptr<Stream> file = Stream::openFile(opt_filename, Stream::C_READ);
        GTurnfile trn(file);
        file = 0;

        if (opt_sort)
            trn.sortCommands();
        if (opt_kill_taccom) {
            trn.setFeatures(trn.getFeatures() & ~GTurnfile::TRNF_TACCOM);
            trn.update();       // needed by computeTurnChecksum()
        }

        switch (opt_action) {
         case List:
            {
                bool nl = false, match = false;
                if (opt_header == 1) {
                    showHeader(trn, opt_filename);
                    nl = true;
                }
                
                /* Now process commands */
                for (int i = 0; i < trn.getNumCommands(); ++i) {
                    if (!opt_filter || opt_filter->accept(trn, i)) {
                        if (nl)
                            std::cout << "\n";
                        nl = true;
                        showCommand(trn, i);
                        status = 0;
                        match = true;
                    }
                }
                if (!match && opt_filter)
                    std::cout << _("No command matched the filter.\n");

                if (opt_trailer == 1) {
                    if (nl)
                        std::cout << "\n";
                    showTrailer(trn);
                }
            }
            break;
         case Delete:
            {
                bool match = false;
                for (int i = 0; i < trn.getNumCommands(); ++i) {
                    if (opt_filter->accept(trn, i)) {
                        trn.deleteCommand(i);
                        status = 0;
                        match = true;
                    }
                }
                if (!match)
                    std::cout << _("No command matched the filter.\n");
            }
            /* FALLTHROUGH */
         case Rewrite:
            saveTurn(trn, opt_filename);
            break;
         case Nothing:
            break;
        }
    }
    catch(std::exception& e) {
        die (_("Error: ") + string_t(e.what()));
    }
    return status;
}
