/**
  *  \file game/un-trnflt.cc
  *  \brief Implementation of Filter Classes for Un-TRN
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <cctype>
#include "game/un-trnflt.h"
#include "util/nls.h"

/* ISO C++ mandates these are not macros, but libc5 does not care. */
#undef isspace
#undef isalpha
#undef isdigit

bool
FilterConst::accept(GTurnfile&, int)
{
    return res;
}

bool
FilterName::accept(GTurnfile& trn, int index)
{
    const char* cmname = trn.getCommandName(index);
    if (!cmname)
        return false;

    string_t s(cmname);
    if (wild && s.length() > str.length())
        s.erase(str.length());
    
    if (strCaseCmp(s, str) == 0)
        return true;
    else
        return false;
}

bool
FilterString::accept(GTurnfile& trn, int index)
{
    string_t text;
    switch(trn.getCommand(index)) {
     case tcm_ShipChangeFc:
     case tcm_PlanetChangeFc:
        text = getBasicStringN(trn.getCommandData(index), 3);
        break;
     case tcm_ShipChangeName:
        text = getBasicStringN(trn.getCommandData(index), 20);
        break;
     case tcm_SendMessage:
        text.assign(trn.getCommandData(index), trn.getCommandId(index));
        for (string_t::size_type i = 0; i < text.length(); ++i)
            text[i] -= 13;
        break;
     default:
        return false;
    }
    return strUCase(text).find(str) != string_t::npos;
}

bool
FilterId::accept(GTurnfile& trn, int index)
{
    int type = trn.getCommandType(index);
    if (type == GTurnfile::SHIP_COMMAND
        || type == GTurnfile::PLANET_COMMAND
        || type == GTurnfile::BASE_COMMAND)
    {
        int id = trn.getCommandId(index);
        return (id >= lower) && (id <= upper);
    } else {
        return false;
    }
}

bool
FilterAnd::accept(GTurnfile& trn, int index)
{
    return lhs->accept(trn, index) && rhs->accept(trn, index);
}

bool
FilterOr::accept(GTurnfile& trn, int index)
{
    return lhs->accept(trn, index) || rhs->accept(trn, index);
}

bool
FilterNeg::accept(GTurnfile& trn, int index)
{
    return !other->accept(trn, index);
}

bool
FilterIndex::accept(GTurnfile& /*trn*/, int index)
{
    ++index;
    return (index >= from && index <= to);
}

/*
   filter ::= or
            | or "|" filter
   or     ::= and
            | and "&" or
   and    ::= '!' and
            | ex
   ex     ::= single
            | single ex
   single ::= "(" filter ")"
            | identifier
            | identifier "*"
            | number
            | number "-" number
            | "#" number
            | "#" number "-" number
            | "'" character* "'"
            | """ character* """
  */

Ptr<Filter> parseFilter1(const char*& text);

static void
parseWs(const char*& text)
{
    while (*text && std::isspace(*text))
        ++text;
}

int
parseNumber(const char*& text)
{
    int result = 0;
    while (std::isdigit(*text))
        result = 10*result + (*text++ - '0');
    return result;
}

bool
isSingle(const char* text)
{
    return *text == '('
        || *text == '\''
        || *text == '"'
        || std::isalpha(*text)
        || std::isdigit(*text);
}

Ptr<Filter>
parseSingle(const char*& text)
{
    parseWs(text);
    if (*text == '(') {
        ++text;
        Ptr<Filter> result = parseFilter1(text);
        parseWs(text);
        if (*text != ')')
            throw ParseException(_("`)' expected"));
        ++text;
        return result;
    } else if (*text == '\'' || *text == '"') {
        char delim = *text++;
        string_t value;
        while (*text && *text != delim)
            value += *text++;
        if (*text != delim)
            throw ParseException(_("unterminated string constant"));
        ++text;
        return new FilterString(value);
    } else if (std::isdigit(*text) || *text == '#') {
        bool index = false;
        if (*text == '#') {
            index = true;
            ++text;
            parseWs(text);
            if (!std::isdigit(*text))
                throw ParseException(_("Number expected after `#'"));
        }
        int low = parseNumber(text);
        int high = low;
        parseWs(text);
        if (*text == '-') {
            ++text;
            parseWs(text);
            if (!std::isdigit(*text))
                throw ParseException(_("Number expected after `-'"));
            high = parseNumber(text);
        }
        if (index)
            return new FilterIndex(low, high);
        else
            return new FilterId(low, high);
    } else if (std::isalpha(*text)) {
        int i = 0;
        while (std::isalpha(text[i]))
            ++i;
        string_t value (text, i);
        text += i;
        parseWs(text);
        if (strCaseCmp(value, "true") == 0) {
            return new FilterConst(true);
        } else if (strCaseCmp(value, "false") == 0) {
            return new FilterConst(false);
        } else {
            if (*text == '*') {
                ++text;
                return new FilterName(value, true);
            } else {
                return new FilterName(value);
            }
        }
    } else
        throw ParseException(_("Invalid expression"));
}

Ptr<Filter>
parseEx(const char*& text)
{
    Ptr<Filter> result = parseSingle(text);
    while (1) {
        parseWs(text);
        if (! isSingle(text))
            break;
        result = new FilterAnd(result, parseSingle(text));
    }
    return result;
}

Ptr<Filter>
parseAnd(const char*& text)
{
    parseWs(text);
    if (*text == '!') {
        ++text;
        return new FilterNeg(parseAnd(text));
    } else
        return parseEx(text);
}

Ptr<Filter>
parseOr(const char*& text)
{
    Ptr<Filter> result = parseAnd(text);
    while (1) {
        parseWs(text);
        if (*text != '&')
            break;
        ++text;
        result = new FilterAnd(result, parseAnd(text));
    }
    return result;
}

Ptr<Filter>
parseFilter1(const char*& text)
{
    Ptr<Filter> result = parseOr(text);
    while (1) {
        parseWs(text);
        if (*text != '|')
            break;
        ++text;
        result = new FilterOr(result, parseOr(text));
    }
    return result;
}

/** Parse filter expression.
    \returns filter expression
    \throws ParseException on error; exception object contains
    error message. */
Ptr<Filter>
parseFilter(const char* text)
{
    Ptr<Filter> result = parseFilter1(text);
    parseWs(text);
    if (*text)
        throw ParseException(_("Expression incorrectly terminated"));
    return result;
}
