/**
  *  \file game/un-trnflt.h
  *  \brief Filter Classes for Un-TRN
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_UN_TRNFLT_H
#define PCC_V2_GAME_UN_TRNFLT_H

#include "game/trn.h"
#include "util/except.h"
#include "cpluslib/stringcv.h"
#include "cpluslib/misc.h"

//! Syntax error while parsing a filter expression.
class ParseException : public std::runtime_error {
 public:
    ParseException(const string_t& awhat)
        : std::runtime_error(toStandardString(awhat)) { }
};

//! Abstract Filter Class.
class Filter : public Object {
 public:
    virtual bool accept(GTurnfile& trn, int index) = 0;
};

class FilterConst : public Filter {
    bool res;
 public:
    FilterConst(bool res) : res(res) { }
    virtual bool accept(GTurnfile& trn, int index);
};

//! Filtering for command names.
class FilterName : public Filter {
    bool wild;
    string_t str;
 public:
    FilterName(string_t astr, bool awild = false)
        : wild(awild), str(astr) { }
    bool accept(GTurnfile& trn, int index);
};

class FilterString : public Filter {
    string_t str;
 public:
    FilterString(string_t astr)
        : str(strUCase(astr)) { }
    bool accept(GTurnfile& trn, int index);
};

//! Filtering for object Ids.
class FilterId : public Filter {
    int lower, upper;
 public:
    FilterId(int alow, int aup)
        : lower(alow), upper(aup) { }
    bool accept(GTurnfile& trn, int index);
};

//! Conjunction of two filter expressions.
class FilterAnd : public Filter {
    Ptr<Filter> lhs, rhs;
 public:
    FilterAnd(Ptr<Filter> alhs, Ptr<Filter> arhs)
        : lhs(alhs), rhs(arhs) { }
    bool accept(GTurnfile& trn, int index);
};

//! Disjunction of two filter expressions.
class FilterOr : public Filter {
    Ptr<Filter> lhs, rhs;
 public:
    FilterOr(Ptr<Filter> alhs, Ptr<Filter> arhs)
        : lhs(alhs), rhs(arhs) { }
    bool accept(GTurnfile& trn, int index);
};

//! Negation of a filter expression.
class FilterNeg : public Filter {
    Ptr<Filter> other;
 public:
    FilterNeg(Ptr<Filter> ao)
        : other(ao) { }
    bool accept(GTurnfile& trn, int index);
};

class FilterIndex : public Filter {
    int from, to;
 public:
    FilterIndex(int from, int to)
        : from(from), to(to) { }
    bool accept(GTurnfile& trn, int index);
};

Ptr<Filter> parseFilter(const char* text);

#endif
