/**
  *  \file game/utilread.cc
  *  \brief UTILx.DAT Reading
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/utilread.h"
#include "util/except.h"
#include "fileio/io.h"
#include "game/utildata.h"

/****************************** GUtilReader ******************************/

/*! \class GUtilReader
    This encapsulates the logic for reading UTILx.DAT. Derived classes
    can override process() to actually process the records. */

/** UTILx.DAT reader for reading the specified stream. */
GUtilReader::GUtilReader(Ptr<Stream> as)
    : s(as)
{ }

GUtilReader::~GUtilReader()
{ }

/** Read UTILx.DAT. Reads the file, and calls process() for each record,
    in order. */
void
GUtilReader::read()
{
    char header[4];
    while ((s->read(header, sizeof(header))) == sizeof(header)) {
        std::size_t size = getUint16(header, 2);
        int         type = getInt16(header, 0);
        if (size) {
            Array<char> data (size);
            if (s->read(data.ptr(), size) != size)
                throw FileTooShortException(s->getName());
            if (!process(type, size, data.ptr()))
                return;
        } else {
            if (!process(type, size, 0))
                return;
        }
    }
}

/** Check whether this is a valid UTILx.DAT. Valid files start with a
    valid control record. Optionally, return the timestamp in *ts.
    \returns true iff file is valid.
    \pre stream is seekable.
    \throws never, except when the stream throws. */
bool
GUtilReader::check(string_t* ts)
{
    Stream::fpos_t pos = s->getPos();
    char header[4];
    TUtil13Control data;
    if (s->read(header, sizeof header) != sizeof header
        || getInt16(header, 0) != 13
        || getUint16(header, 2) < data.size
        || !getStructure(s, data))
    {
        s->seek(pos);
        return false;
    }
    s->seek(pos);
    if (ts)
        ts->assign(data.timestamp, sizeof(data.timestamp));
    return true;
}

/** Process one record. Can throw an exception to abort parsing.
    \param rec_id  record id
    \param size    size of record, in bytes
    \param data    "raw" data
    \pre size != 0 => data != 0
    \return true to continue parsing, false to stop. */
bool
GUtilReader::process(int rec_id, std::size_t size, const char* data)
{
    return true;
}
