/**
  *  \file vcr.cc
  *  \brief VCR Player Interface
  *
  *  The abstract interface for playing battles. This defines the
  *  interface used by players and visualizers.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/vcr.h"
#include "game/config.h"
#include "game/player.h"
#include "util/nls.h"
#include "cpluslib/format.h"

/********************************** TVcr *********************************/

/*! \class TVcr

    This is a complete VCR record, unpacked to support torps and
    fighters simultaneously. The brain-dead PVCR format is just too
    stupid to handle.

    \see VcrEntry */
TVcr::TVcr(const TRawVcr& raw)
    : TRawVcr(raw)
{
    unpackSide(0);
    unpackSide(1);
}

/** Pack VCR. Packs this VCR into a TRawVcr.
    \returns true if ok, false on error. */
bool
TVcr::pack(TRawVcr& raw)
{
    raw = *this;
    return packSide(raw, 0) & packSide(raw, 1);  // this must be "&"
}

/** Unpack VCR side. */
void
TVcr::unpackSide(int side)
{
    /* This is pretty much the same code as originally was in PVCR. */
    TVcrObject& o = objects[side];
    
    if (o._torp_launcher_count) {
        if (side && battle_type && config.PlanetsHaveTubes) {
            torp_launcher_count[side] = o._torp_launcher_count & 0xFF;
            num_torps[side]           = (o._torp_launcher_count >> 8) & 0xFF;
        } else {
            torp_launcher_count[side] = o._torp_launcher_count;
            num_torps[side]           = o._ammo;
            o._ammo                   = 0;
        }
    } else {
        torp_launcher_count[side] = num_torps[side] = 0;
    }

    if (o.bay_count) {
        num_fighters[side] = o._ammo;
    } else {
        num_fighters[side] = 0;
    }
}

/** Pack VCR side. Assumes that `raw=*this' already performed. */
bool
TVcr::packSide(TRawVcr& raw, int side)
{
    TVcrObject& o = raw.objects[side];
    
    if (o._torp_launcher_count) {
        if (side && battle_type && config.PlanetsHaveTubes) {
            if (torp_launcher_count[side] > 255 || num_torps[side] > 255)
                return false;
            o._torp_launcher_count = (num_torps[side] << 8) | torp_launcher_count[side];
        } else {
            if (o.bay_count)
                return false;
            o._torp_launcher_count = torp_launcher_count[side];
            o._ammo                = num_torps[side];
        }
    } else {
        o.torp_type = 0;
        o._torp_launcher_count = 0;
    }

    if (o.bay_count) {
        o._ammo = num_fighters[side];
    } else {
        o._ammo = 0;
    }
    return true;
}

/******************************* VcrPlayer *******************************/

#ifdef CONF_STATIC_VCR
Ptr<VcrVisualizer> VcrPlayer::vis;
VcrPlayer::time_type VcrPlayer::time;
int VcrPlayer::status_word;
#endif

/*! \class VcrPlayer

    This class is the base for all VCR player algorithms.
    Its interface includes saving and restoring state, and,
    of course, playing the VCR.

    Playing interface:
    - initVcr(vcr) initializes the player for the given VCR.
      This includes a call to the visualizer's init();
    - optionally call playFastForward();
    - call playCycle() repeatedly. If playCycle() returns true,
      the battle continues. If playCycle() returns false,
      the battle has ended, and playCycle() did nothing.
      The first thing playCycle() does is incrementing the
      timer.
    - doneVcr(res) finishes playing and determines the result.
      This does final exploding and such.

    playVcr(vcr) does all but the last step at once.

    Before calling the playing interface, you must associate
    a visualizer with the VcrPlayer.

    Depending on the config macro CONF_STATIC_VCR, the VCR player may
    be static and thus not re-entrant (i.e. thread-safe). It still can
    use nested calls when done properly (get a status token, do the
    "other" computation, continue using the status token). */

/** Constructor. Creates an empty VCR player not associated with
    a viewer. */
VcrPlayer::VcrPlayer()
{
    vis = 0;
}

VcrPlayer::~VcrPlayer()
{
    if (vis.ptr())
        vis->setPeer(0);
    vis = 0;
}

/** Play VCR. Shortcut for playing a whole VCR at once.
    Does not necessarily keep display up-to-date. */
void
VcrPlayer::playVcr(TVcr& vcr)
{
    initVcr(vcr);
    playFastForward();
    while (playCycle())
        ;
}

/** Set Visualizer. Before anything goes here, there must be
    a visualizer connected to the player. */
void
VcrPlayer::setVisualizer(Ptr<VcrVisualizer> avis)
{
    if (vis.ptr())
        vis->setPeer(0);
    vis = avis;
    if (vis.ptr())
        vis->setPeer(this);
}

/** Fast forward. If possible, jump forward in time.
    Does not need to keep the display up-to-date.
    If the VCR algorithm supports that, override this
    method. \default does nothing. */
void
VcrPlayer::playFastForward()
{
}

/** Format as string. \returns human-readable description of result status
    for battle vcr. */
string_t
VcrPlayer::getResultString(int status, const TVcr& vcr)
{
    int me;
    if (vcr.objects[0].owner == getPlayerId())
        me = 0;
    else if (vcr.objects[1].owner == getPlayerId())
        me = 1;
    else
        me = 2;

    string_t s;
    switch (status) {
     case 0:
        s = _("unknown. Wait while computing...");
        break;
     case VcrPlayer::VCRS_INVALID:
        s = _("Battle cannot be played!");
        break;
     case VcrPlayer::VCRS_TIMEOUT:
        s = _("Battle timed out (too long).");
        break;
     case VcrPlayer::VCRS_STALEMATE:
        s = _("Stalemate.");
        break;
     case VcrPlayer::VCRS_LEFT_DESTROYED:
        s = me == 0 ? string_t(_("We were destroyed.")) :
            me == 1 ? string_t(_("We won.")) : string_t(format(_("%s won."), getBasicString(vcr.objects[1].name)));
        break;
     case VcrPlayer::VCRS_RIGHT_DESTROYED:
        s = me == 1 ? string_t(_("We were destroyed.")) :
            me == 0 ? string_t(_("We won.")) : string_t(format(_("%s won."), getBasicString(vcr.objects[0].name)));
        break;
     case VcrPlayer::VCRS_LEFT_CAPTURED:
        s = me == 0 ? string_t(_("They have captured our ship.")) :
            me == 1 ? string_t(_("We captured their ship.")) : string_t(format(_("%s was captured."), getBasicString(vcr.objects[0].name)));
        break;
     case VcrPlayer::VCRS_RIGHT_CAPTURED:
        s = me == 1 ? string_t(_("They have captured our ship.")) :
            me == 0 ? string_t(_("We captured their ship.")) : string_t(format(_("%s was captured."), getBasicString(vcr.objects[1].name)));
        break;
     case VcrPlayer::VCRS_RIGHT_DESTROYED + VcrPlayer::VCRS_LEFT_DESTROYED:
        s = _("Both were destroyed.");
        break;
     default:
        s = _("Both are disabled.");
    }
    return s;
}

/*! \class VcrStatus
    VCR Status token. VcrPlayer descendants derive a class from this
    and store information they need in here. */
VcrStatus::VcrStatus(VcrPlayer::time_type atime)
    : time (atime)
{ }

VcrStatus::~VcrStatus()
{ }

/*! \class VcrVisualizer

    The visualizer's methods are called by a VCR player to do actual
    screen display. Not all events generate visualizer callbacks, hence
    the visualisation code must also poll for changes (fighter positions,
    damage/kill values, ...). It can use the accessor interface of
    VcrPlayer for that. */
VcrVisualizer::VcrVisualizer()
    : ply(0)
{ }

VcrVisualizer::~VcrVisualizer()
{ }

/** Set Player. This method is called by the VcrPlayer to inform
    this visualizer whom it works for. */
void
VcrVisualizer::setPeer(VcrPlayer* aply)
{
    ply = aply;
}

/** \class VcrNullVisualizer
    \brief Null Visualizer

    This VCR visualizer does nothing. It is to be used for
    simulations, etc. */
void
VcrNullVisualizer::startFighter(int /*side*/, int /*track*/)
{ }

void
VcrNullVisualizer::landFighter(int /*side*/, int /*track*/)
{ }

void
VcrNullVisualizer::killFighter(int /*side*/, int /*track*/)
{ }

void
VcrNullVisualizer::fireBeam(int /*side*/, int /*track*/, int /*target*/, int /*hit*/, int, int)
{ }

void
VcrNullVisualizer::fireTorp(int /*side*/, int /*hit*/, int)
{ }

void
VcrNullVisualizer::updateBeam(int /*side*/, int /*id*/)
{ }

void
VcrNullVisualizer::updateTorp(int /*side*/, int /*id*/)
{ }

void
VcrNullVisualizer::killObject(int /*side*/)
{ }

void
VcrNullVisualizer::redraw()
{ }

void
VcrNullVisualizer::init()
{ }
