/**
  *  \file game/vcrdb.cc
  *  \brief VCR Database Implementation
  *
  *  Management of VCR Entries ("battles") and databases containing
  *  them.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/vcrdb.h"
#include "util/except.h"
#include "game/tvcr.h"
#include "spec.h"
#include "game/pvcr.h"
#include "util/nls.h"

/******************************** VcrEntry *******************************/

/*! \class VcrEntry

    A VcrEntry is to a TVcr what GBeam is to TBeam: the high-level
    representation.

    For VCRs, we have a three-stage representation:
    - TRawVcr is a VCR as it appears in VCR.DAT files;
    - Since the VCR representation is ambiguous, mostly for
      PlanetsHaveTubes things, it is unpacked into a TVcr.
      Much code uses these;
    - VcrEntry is the high-level representation for "one battle".

    Right now, there is no accessor interface. */
VcrEntry::VcrEntry(const TVcr& avcr)
    : status(0),
      vcr(avcr)
{
    hulls[0] = guessHull(0);
    hulls[1] = guessHull(1);
}

/** Guess hull for specified side.
    \returns hull if unique, 0 if none possible or ambiguous. */
int
VcrEntry::guessHull(int side) const
{
    /* planets don't have hulls. */
    if (isPlanet(side))
        return 0;

    /* see if PHost sent us the hull type. */
    if (vcr.objects[side].hull_or_zero)
        return vcr.objects[side].hull_or_zero;

    /* otherwise, try all hulls. */
    int type = 0;
    for (int h = 1; h <= getNumHulls(); ++h)
        if (canBeHull(side, h)) {
            if (type == 0)
                type = h;
            else
                return 0;       // ambiguous, can't guess
        }
    return type;
}

/** Check if hull matches.
    \param side     0 or 1 (left or right side of VCR)
    \param hull_id  hull id to check [1, getNumHulls()].
    \pre we're actually talking about a ship. */
bool
VcrEntry::canBeHull(int side, int hull_id) const
{
    const GHull& hull = getHull(hull_id);
    const TVcrObject& vcro = vcr.objects[side];

    if (vcro.hull_or_zero) {
        /* hull known - case clear */
        return (vcro.hull_or_zero == hull_id);
    } else {
        /* picture must match */
        if (vcro.picture != hull.getPicture())
            return false;
        /* must not have more beams/torps than hull allows */
        if (vcro.beam_count > hull.getMaxBeams()
            || vcr.torp_launcher_count[side] > hull.getMaxLaunchers())
            return false;
        /* for fighter bays, the only k.o. criterion is that ship has
           fighters but hull has not. The number of bays can be
           smaller (NTP, damage) or larger (scotty bonus). */
        if (vcro.bay_count && !hull.getNumBays())
            return false;
        return true;
    }
}

/** Store result. Fetches the result from the specified
    VCR Player. The player must have called playVcr(), but
    not yet doneVcr(). */
void
VcrEntry::setResultFromPlayer(VcrPlayer& player)
{
    player.doneVcr(result);
    status = player.getStatusWord();
}

/** Set Result. Sets the result status from the specified
    paramters. */
void
VcrEntry::setResult(int astatus, const TVcr& avcr)
{
    status = astatus;
    result = avcr;
}

/****************************** VcrDatabase ******************************/

/*! \class VcrDatabase

    A set of VCRs that can be played; like a mailbox contains messages
    that can be dealt with, this one contains VCRs to deal with.

    It is indexed 0-based.

    A VCR database knows which player must be used to play it. It also
    knows how to parse a file into a set of VCRs and determine the
    player required for it. */

VcrDatabase::VcrDatabase()
    : capabilities(0), vcr_type(vcr_None)
{ }

static inline bool
hasPHostMagic(const TRawVcr& vcr)
{
    return ((vcr.random_seed + vcr.signature) & 0xFFFFU) == 48879;
}

/** Load VCRs from a file. Assumes that \c file is a stream, pointing
    at the beginning of a VCR file. This also figures out the type
    of the VCRs. */
void
VcrDatabase::loadFromFile(Ptr<Stream> file)
{
    char buffer[TRawVcr::size];
    bool maybe_phost;
    uint16 count;

    ASSERT(file.ptr());

    if (file->read(buffer, 2) != 2)
        throw FileTooShortException(file->getName());

    count = getUint16(buffer);
    maybe_phost = false;
    vcr_type = vcr_Host;
    capabilities = 0;
    while (count--) {
        TRawVcr raw_vcr;

        if (file->read(buffer, sizeof(buffer)) != sizeof buffer)
            throw FileTooShortException(file->getName());
        getStructure(buffer, raw_vcr);
        if (contents.empty()) {
            /* this is the first VCR */
            maybe_phost = hasPHostMagic(raw_vcr);
        } else if (maybe_phost && !count) {
            /* this is the last VCR. Some things to note:
               - it may be PHost 2's config battle. Discard it but
                 memorize type.
               - it may be the bogus battle from CORR (all but name
                 zeroed) */
            if (hasPHostMagic(raw_vcr) && raw_vcr.battle_type >= 2) {
                /* FIXME: right now, we assume that the pconfig is
                   right. For a standalone PVCR, we should parse the
                   config structure. */
                if ((raw_vcr._planet_temp_code & 255) == 2)
                    vcr_type = vcr_PHost2;
                break;
            }
            if (raw_vcr.objects[0].owner == 0 || raw_vcr.objects[1].owner == 0)
                /* CORR */
                break;
        }
        contents.push_back(new VcrEntry(TVcr(raw_vcr)));
    }
    if (maybe_phost && vcr_type == vcr_Host) {
        vcr_type = vcr_PHost3;

        uint16 cw = contents.front()->getVcr()._planet_temp_code;
        if (cw & vcrcap_Valid)
            capabilities = cw & ~vcrcap_Valid;
        if (capabilities)
            vcr_type = vcr_PHost3;
    }
}

/** Add a VCR record. This is to be used by simulators etc. */
void
VcrDatabase::addVcr(const TVcr& avcr)
{
    contents.push_back(new VcrEntry(avcr));
}

/** Set VCR player to use for this database.
    \post getVcrPlayer() == avp */
void
VcrDatabase::setVcrPlayer(Ptr<VcrPlayer> avp)
{
    vcr_player = avp;
}

/** Get VCR player to use for this VCR database. If you have not yet
    set one with setVcrPlayer(), but the VcrDatabase was initialized
    with loadFromFile(), it can guess a player automatically. */
Ptr<VcrPlayer>
VcrDatabase::getVcrPlayer()
{
    ASSERT(vcr_player.ptr() || vcr_type != vcr_None);

    if (!vcr_player.ptr()) {
        switch(vcr_type) {
         case vcr_Host:
            vcr_player = new VcrPlayerTHost();
            break;
         case vcr_PHost4:
         case vcr_PHost3:
         case vcr_PHost2:
            vcr_player = new VcrPlayerPHost(vcr_type != vcr_PHost2);
            break;
         case vcr_None:
            ASSERT(0);
        }
    }
    return vcr_player;
}
