/**
  *  \file gfx/clip.cc
  *  \brief Simple Clipping Implementation
  *
  *  Classes for clipping output at a single rectangle.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "gfx/clip.h"

/** \class GfxClipRect
    \brief Clip at a Rectangle

    This class provides the back-end for drawing with clipping
    at a rectangle.

    It does not contain a public constructor or methods to modify the
    clip rectangle; derived classes must implement these when needed. */

GfxClipRect::GfxClipRect(const GfxRect& arect, Ptr<GfxCanvas> apar)
    : GfxFilter(apar), rect(arect)
{ }

void
GfxClipRect::drawHLine(int x1, int y1, int x2, uint32 color, uint8 pat, alpha_type alpha)
{
    GfxRect r (x1, y1, x2-x1, 1);
    r.clipRect(rect);
    if (r.exists ())
        parent->drawHLine(r.x, r.y, r.x + r.w, color, pat, alpha);
}

void
GfxClipRect::drawVLine(int x1, int y1, int y2, uint32 color, uint8 pat, alpha_type alpha)
{
    GfxRect r (x1, y1, 1, y2-y1);
    r.clipRect(rect);
    if (r.exists())
        parent->drawVLine(r.x, r.y, r.y + r.h, color, pat, alpha);
}

void
GfxClipRect::drawPixel(int x, int y, uint32 color, alpha_type alpha)
{
    if (rect.contains(x, y))
        parent->drawPixel(x, y, color, alpha);
}

void
GfxClipRect::drawBar(GfxRect r, uint32 color, uint32 bg,
                     const GfxFillPattern& pat, alpha_type alpha)
{
    r.clipRect(rect);
    if (r.exists())
        parent->drawBar(r, color, bg, pat, alpha);
}

void
GfxClipRect::blitSurface(int x, int y, SDL_Surface* src, GfxRect r)
{
    GfxRect rclip = rect;
    rclip.moveBy(-x, -y);
    r.clipRect(rclip);
    if (r.exists())
        parent->blitSurface(x, y, src, r);
}

void
GfxClipRect::blitPattern(GfxRect r, int x, int y, int bytes_per_line,
                         const uint8* data, uint32 color,
                         uint32 bg, alpha_type alpha)
{
    r.clipRect(rect);
    if (r.exists())
        parent->blitPattern(r, x, y, bytes_per_line, data, color, bg, alpha);
}

GfxRect
GfxClipRect::computeClipRect(GfxRect r)
{
    r.clipRect(rect);
    return parent->computeClipRect(r);
}

/*************************** GfxClipRectFilter ***************************/

/** \class GfxClipRectFilter
    \brief Clipping at a Rectangle

    This class provides clipping at a rectangle. This can be used by
    normal users to clip output. */

GfxClipRectFilter::GfxClipRectFilter(const GfxRect& arect, Ptr<GfxCanvas> apar)
    : GfxClipRect(arect, apar)
{ }

/******************************* GfxOffset *******************************/

GfxOffset::GfxOffset(int adx, int ady, Ptr<GfxCanvas> apar)
    : GfxFilter(apar),
      dx(adx), dy(ady)
{ }
    
void
GfxOffset::drawHLine(int x1, int y1, int x2, uint32 color, uint8 pat, alpha_type alpha)
{
    parent->drawHLine(x1 + dx, y1 + dy, x2 + dx, color, pat, alpha);
}

void
GfxOffset::drawVLine(int x1, int y1, int y2, uint32 color, uint8 pat, alpha_type alpha)
{
    parent->drawVLine(x1 + dx, y1 + dy, y2 + dy, color, pat, alpha);
}

void
GfxOffset::drawPixel(int x, int y, uint32 color, alpha_type alpha)
{
    parent->drawPixel(x + dx, y + dy, color, alpha);
}

void
GfxOffset::drawBar(GfxRect rect, uint32 color, uint32 bg,
                   const GfxFillPattern& pat, alpha_type alpha)
{
    rect.moveBy(dx, dy);
    parent->drawBar(rect, color, bg, pat, alpha);
}

void
GfxOffset::blitSurface(int x, int y, SDL_Surface* src, GfxRect rect)
{
    parent->blitSurface(x + dx, y + dy, src, rect);
}

void
GfxOffset::blitPattern(GfxRect rect, int x, int y, int bytes_per_line,
                       const uint8* data, uint32 color, uint32 bg,
                       alpha_type alpha)
{
    rect.moveBy(dx, dy);
    parent->blitPattern(rect, x + dx, y + dy, bytes_per_line,
                        data, color, bg, alpha);
}

GfxRect
GfxOffset::computeClipRect(GfxRect r)
{
    r.moveBy(dx, dy);
    GfxRect q = parent->computeClipRect(r);
    q.moveBy(-dx, -dy);
    return q;
}

void
GfxOffset::update(GfxRect r)
{
    r.moveBy(dx, dy);
    parent->update(r);
}

uint32
GfxOffset::getPixel(int x, int y)
{
    return parent->getPixel(x + dx, y + dy);
}
