/**
  *  \file gfx/font.cc
  *  \brief Font Drivers
  *
  *  This module contains the basic font driver interface, and the
  *  implementation of a standard driver.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "gfx/font.h"
#include "fileio/io.h"
#include "util/except.h"
#include "util/nls.h"
#include "cpluslib/assert.h"

/** \class GfxStdFont
    \brief Standard Font Driver

    This class implements the font format that happened to get
    the default in my graphics programming stuff, and that was
    used in PCC v1.x.

    This is a bitmap font system with an optional `dumbfire'
    antialiasing function.

    Restrictions: no charset translation (yet). Sharing of bitmap data
    between characters and/or fonts is possible in the file format,
    but not supported here. The font editor does not create files with
    shared bitmap data. */

/** Construct font from stream.
    Loads the font from a stream.
    \param aindex Font files can contain multiple fonts. This is the
                  index of font to use. It is usually omitted because
                  in practice, font files contain only one font.
    \throws FileFormatException on error. */
GfxStdFont::GfxStdFont(Ptr<Stream> stream, int aindex)
{
    Buffer<char> buf;
    stream->readBuffer(buf);
    image = buf.getArray();

    initialize(aindex);
}

/** Construct font from array.
    Constructs the font from an already-loaded array.
    \param aindex Index of font to use.
    \overload */
GfxStdFont::GfxStdFont(ArrayPtr<char> aimage, int aindex)
    : image(aimage)
{
    initialize(aindex);
}

/** Initialisation. Common constructor code. Checks the file format,
    and fixes up the bitmap data. */
void
GfxStdFont::initialize(int aindex)
{
    if (image[0] != 'F' || image[1] != 'N')
        throw FileFormatException(_("Invalid font file"));
    if (aindex < 0 || aindex >= getInt16(image, 2))
        throw FileFormatException(_("Invalid font number"));

    base_ptr  = getInt32(image, 4 + 8*aindex);
    antialias = (image[base_ptr] == 3);

    if (antialias) {
        /* We assume in outText() that the AA pixels are not touched
           by the blitPattern() function. So clear them in the bitmap
           data. */
        int height = getUint16 (image, base_ptr + 2);
        for (int i = 0; i < 256; ++i) {
            const char* c = getChar(i);
            int width     = getUint16(c, 6);
            int bpl       = (width + 7) / 8;
            char* cdata   = &image[getUint32 (c, 0)];
            const char* adata = cdata + bpl*height;
            int count = getInt16(adata);
            while (count-- > 0) {
                int dx = getInt16(adata, 2);
                int dy = getInt16(adata, 4);
                adata += 4;
                if (dx >= 0 && dx < width && dy >= 0 && dy < height)
                    cdata[dy * bpl + (dx / 8)] &= ~(0x80 >> (dx & 7));
            }
        }
    }
}

/** Get character.
    \returns pointer to the font data for character c */
inline const char*
GfxStdFont::getChar(char c)
{
    unsigned char uc = c;
    return &image[base_ptr + 6 + 8*uc];
}

/** Output text. Implementation of the text output primitive. */
void
GfxStdFont::outText(GfxCanvas* out, GfxParameters& parm, int x, int y,
                    const string_t text)
{
    string_t::size_type len = text.length();
    int height = getUint16 (image, base_ptr + 2);

    /* arithmetic rounding, so alpha=255 gets to aa_alpha=128 */
    alpha_type aa_alpha = (parm.alpha+1)/2;

    for (string_t::size_type i = 0; i < len; ++i) {
        const char* c = getChar(text[i]);
        int width     = getUint16(c, 6);
        int bpl       = (width + 7) / 8;
        if (width <= 0)
            continue;
        const char* cdata = &image[getUint32 (c, 0)];

        out->blitPattern(GfxRect(x, y, width, height),
                         x, y,
                         bpl,
                         (const uint8*) cdata,
                         parm.color,
                         parm.bgcolor,
                         parm.alpha);

        if (antialias) {
            /* FIXME: this can be optimized. When
               parm.bgcolor != GFX_TRANSPARENT, we know the color of
               that pixel exactly. */
            cdata += bpl*height;
            int count = getInt16(cdata);
            while (count-- > 0) {
                int dx = getInt16(cdata, 2);
                int dy = getInt16(cdata, 4);
                cdata += 4;
                out->drawPixel(x + dx, y + dy, parm.color, aa_alpha);
            }
        }

        x += width;
    }
}

/** Compute text width. */
int
GfxStdFont::getTextWidth(const string_t text)
{
    string_t::size_type len = text.length();
    int width = 0;
    for (string_t::size_type i = 0; i < len; ++i)
        width += getUint16(getChar(text[i]), 6);
    return width;
}

/** Compute text height. */
int
GfxStdFont::getTextHeight(string_t /*text*/)
{
    return getUint16 (image, base_ptr + 2);
}

/** Set scaling factor. This driver only supports 1:1 scaling. */
bool
GfxStdFont::setScale(int mul, int div)
{
    return (mul == div);
}

/** Return scaling factor. We only support 1:1 scaling. */
GfxPoint
GfxStdFont::getScale()
{
    return GfxPoint(1, 1);
}

/** Get capabilities. This font is linear and supports alpha blending. */
unsigned
GfxStdFont::getCapabilities()
{
    return FN_LINEAR | FN_ALPHA;
}

/************************ "Complicated" Functions ************************/

/** Output Text, using Alignment Parameters.
    \param out    Canvas
    \param parm   Graphics parameters
    \param fnt    Font to use
    \param x,y    Anchor point
    \param text   Text to output */
void
outText(GfxCanvas* out, GfxParameters& parm, GfxFont* fnt, int x, int y, string_t text)
{
    if (parm.just_x)
        x -= fnt->getTextWidth(text) * parm.just_x / 2;
    if (parm.just_y)
        y -= fnt->getTextHeight(text) * parm.just_y / 2;
    fnt->outText(out, parm, x, y, text);
}

/** Output Text with fixed maximum width. Outputs the text so that it
    occupies at most maxw pixels on the screen. When non-transparent output
    is used (parm.bgcolor != GFX_TRANSPARENT), draws appropriate boxes so that
    exactly the specified area is covered. Otherwise like outText().

    This function moves the graphics cursor, depending on the text
    justification:
    - parm.y is set to y;
    - when parm.just_x=0 (left justify), parm.x is set to just after
      the output text;
    - when parm.just_x=2 (right justify), parm.x is set to just before
      the output text;
    - when parm.just_x=1 (centered), parm.x is set to x.

    The rationale is that (when not using centered output) one can do
    several outText()s in a row and always know the correct place without
    explicitly having to compute the width. */
void
outTextF(GfxCanvas* out, GfxParameters& parm, GfxFont* fnt, int x, int y, int maxw, string_t text)
{
    int width, height;
    while (1) {
        width = fnt->getTextWidth(text);
        if (width <= maxw || text.empty())
            break;
        text.erase(text.length() - 1);
    }

    height = fnt->getTextHeight(text);
    parm.y = y;
    y -= height * parm.just_y / 2;

    switch (parm.just_x) {
     case 0:
        /* Left */
        fnt->outText(out, parm, x, y, text);
        if (parm.bgcolor != GFX_TRANSPARENT && maxw > width)
            out->drawBar(GfxRect(x + width, y, maxw - width, height),
                         parm.bgcolor, GFX_TRANSPARENT, SOLID_FILL, parm.alpha);
        parm.x = x + width;
        break;
     case 1:
        /* Centered */
        {
            int xl = x - maxw/2;
            int xtl = x - width/2;
            int xr = xl + maxw;
            int xtr = xtl + width;
            fnt->outText(out, parm, xtl, y, text);
            if (parm.bgcolor != GFX_TRANSPARENT) {
                if (xl != xtl)
                    out->drawBar(GfxRect(xl, y, xtl - xl, height),
                                 parm.bgcolor, GFX_TRANSPARENT, SOLID_FILL, parm.alpha);
                if (xr != xtr)
                    out->drawBar(GfxRect(xtr, y, xr - xtr, height),
                                 parm.bgcolor, GFX_TRANSPARENT, SOLID_FILL, parm.alpha);
            }
            parm.x = x;
            break;
        }
     case 2:
        /* Right */
        fnt->outText(out, parm, x - width, y, text);
        if (parm.bgcolor != GFX_TRANSPARENT && maxw > width)
            out->drawBar(GfxRect(x - maxw, y, maxw - width, height),
                         parm.bgcolor, GFX_TRANSPARENT, SOLID_FILL, parm.alpha);
        parm.x = x - width;
        break;
     default:
        ASSERT(0);
    }
}
