/**
  *  \file gfx/gfx-modetr.h
  *  \brief Mode Traits
  *
  *  Generalized access to SDL_Surface pixel data.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GFX_GFX_MODETR_H
#define PCC_V2_GFX_GFX_MODETR_H

/** \name Mode Traits.
    Each type supports the following expressions:
    
    - T::pix_type      = type which can hold a pixel
    - T::peek(ptr)     = read video memory at specified address
    - T::poke(ptr,val) = modify video memory at specified address
    - T::add(ptr,diff) = advance ptr by diff pixels
    - T::mix(fmt,val,val,alpha) = alpha mixing */
//@{
struct Gfx8bitModeTraits {
    typedef uint8 pix_type;
    static inline uint8  peek(uint8* ptr) { return *ptr; }
    static inline void   poke(uint8* ptr, uint8 val) { *ptr = val; }
    static inline uint8* add(uint8* ptr, int n) { return ptr + n; }
    static inline uint8  mix(SDL_PixelFormat* fmt, uint8 a, uint8 b, alpha_type balpha);
};

struct Gfx16bitModeTraits {
    typedef uint16 pix_type;
    static inline uint16 peek(uint8* ptr) { return *(uint16*)ptr; }
    static inline void   poke(uint8* ptr, uint16 val) { *(uint16*)ptr = val; }
    static inline uint8* add(uint8* ptr, int n) { return ptr + 2*n; }
    static inline uint16 mix(SDL_PixelFormat* fmt, uint16 a, uint16 b, alpha_type balpha);
};

struct Gfx32bitModeTraits {
    typedef uint32 pix_type;
    static inline uint32 peek(uint8* ptr) { return *(uint32*)ptr; }
    static inline void   poke(uint8* ptr, uint32 val) { *(uint32*)ptr = val; }
    static inline uint8* add(uint8* ptr, int n) { return ptr + 4*n; }
    static inline uint32 mix(SDL_PixelFormat* fmt, uint32 a, uint32 b, alpha_type balpha);
};
//@}

/** Alpha mixing.
    \param a      background color
    \param b      color to write
    \param alpha  alpha value of color to write. */
template<class T>
T
alphaMix(T a, T b, alpha_type alpha)
{
    return a + ((b-a) * alpha >> 8);
}

uint8
Gfx8bitModeTraits::mix(SDL_PixelFormat* fmt, uint8 a, uint8 b, alpha_type balpha)
{
    /* FIXME: caching? */
    register SDL_Color* pal = fmt->palette->colors;
    return SDL_MapRGB (fmt,
                       alphaMix (pal[a].r, pal[b].r, balpha),
                       alphaMix (pal[a].g, pal[b].g, balpha),
                       alphaMix (pal[a].b, pal[b].b, balpha));
}

uint16
Gfx16bitModeTraits::mix(SDL_PixelFormat* fmt, uint16 a, uint16 b, alpha_type balpha)
{
    uint16 re = alphaMix (a & fmt->Rmask, b & fmt->Rmask, balpha) & fmt->Rmask;
    uint16 gr = alphaMix (a & fmt->Gmask, b & fmt->Gmask, balpha) & fmt->Gmask;
    uint16 bl = alphaMix (a & fmt->Bmask, b & fmt->Bmask, balpha) & fmt->Bmask;
    return re | gr | bl;
}

uint32
Gfx32bitModeTraits::mix(SDL_PixelFormat* fmt, uint32 a, uint32 b, alpha_type balpha)
{
    /* FIXME: Alpha channel (Amask) ? */
    /* FIXME2: can this overflow (if Rmask=0xFF000000)? */
    uint32 re = alphaMix (a & fmt->Rmask, b & fmt->Rmask, balpha) & fmt->Rmask;
    uint32 gr = alphaMix (a & fmt->Gmask, b & fmt->Gmask, balpha) & fmt->Gmask;
    uint32 bl = alphaMix (a & fmt->Bmask, b & fmt->Bmask, balpha) & fmt->Bmask;
    return re | gr | bl;
}

/** Pixel Format Switch.
    \param function name of a function defined as
                    `template<class T> void func(args)'
    \param args     arguments to use for call

    Calls the correct function with the appropriate mode traits filled
    in. The surface being examined must be in scope under the name
    `sfc'. */
/* FIXME: 24 bpp not supported yet */
#define GFX_MODE_SWITCH(function, args)                        \
    do switch(sfc->format->BytesPerPixel) {                    \
     case 1:                                                   \
        function<Gfx8bitModeTraits>args;                       \
        break;                                                 \
     case 2:                                                   \
        function<Gfx16bitModeTraits>args;                      \
        break;                                                 \
     case 3:                                                   \
        FAIL();                                                \
        break;                                                 \
     case 4:                                                   \
        function<Gfx32bitModeTraits>args;                      \
        break;                                                 \
    } while(0)

#endif
