/**
  *  \file gfx/gfx.cc
  *  \brief Graphics Kernel
  *
  *  Implementation or graphics kernel (higher-level API wrapped
  *  around SDL).
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <SDL_error.h>
#include "gfx/gfx.h"
#include "util/except.h"
#include "cpluslib/assert.h"
#include "cpluslib/bitops.h"

/*! \class GfxParameters
    This class contains graphics parameters used for the complex
    graphics primitives. Instead of passing around all parameters
    (like we do for the simple graphics primitives), we pass around
    GfxParameters objects. */

/// Construct an empty GfxParameters object, with all values default.
GfxParameters::GfxParameters()
    : color (~0),
      bgcolor (GFX_TRANSPARENT),
      line_thickness (1),
      line_style (0xFF),
      fill_pattern (0xFF),
      alpha (SDL_ALPHA_OPAQUE),
      x (0), y (0),
      just_x (0), just_y (0)
{ }

//////////////////////////////////////////////////////////////////////

/** \class GfxCanvas
    \brief Drawing GfxSurface

    Offers an interface for simple graphics primitives. Descendants
    have to actually implement them. Based on the simple graphics
    primitives, we offer a few more complex ones in gfxcompl.h. The
    complex primitives use a GfxParameters object to store the common
    parameters. This permits `stateful' programming. The simple
    graphics primitives are stateless.
    
    The general procedure for drawing something on a canvas is
    \code
       c->lock();
       // draw something
       c->unlock();
       c->update(...);     // update relevant parts
    \endcode
    We'll be going to hide that somewhere deep in the GUI class
    redraw management. */

GfxCanvas::~GfxCanvas()
{ }

/** Check if rectangle is clipped. If it is not clipped, it is completely
    visible. */
bool
GfxCanvas::isClipped(GfxRect r)
{
    return r != computeClipRect(r);
}

/** Check if rectangle is visible. A rectangle is visible if at least
    one pixel of it can be seen. */
bool
GfxCanvas::isVisible(GfxRect r)
{
    return computeClipRect(r).exists();
}

//////////////////////////////////////////////////////////////////////

#include "gfx/gfx-modetr.h"

/** Pixel. Implementation of the `put pixel' primitive for all modes. */
template<class T>
inline void
doPixel(SDL_Surface* sfc, int x, int y, typename T::pix_type color)
{
    T::poke (T::add ((uint8*)sfc->pixels + y * sfc->pitch, x), color);
}

/** Pixel. Implementation of the `put pixel with alpha' primitive for
    all modes. */
template<class T>
inline void
doPixelAlpha(SDL_Surface* sfc, int x, int y, typename T::pix_type color,
             alpha_type alpha)
{
    uint8* p = T::add ((uint8*)sfc->pixels + y * sfc->pitch, x);
    T::poke (p, T::mix (sfc->format, T::peek(p), color, alpha));
}

/** Horizontal Line. Implementation of the `horizontal line' primitive
    for all modes. */
template<class T>
inline void
doHLine(SDL_Surface* sfc, int x1, int y1, int x2, typename T::pix_type color,
        uint8 pat)
{
    if (pat == 0) {
        /* nothing */
    } else if (pat == 255) {
        /* solid line */
        uint8* p = T::add ((uint8*)sfc->pixels + y1 * sfc->pitch, x1);
        while (x1 < x2) {
            T::poke(p, color);
            p = T::add(p, 1);
            ++x1;
        }
    } else {
        /* pattern line */
        uint8 mask = ror8 (0x80, x1);
        uint8* p = T::add ((uint8*)sfc->pixels + y1 * sfc->pitch, x1);
        while (x1 < x2) {
            if (mask & pat)
                T::poke(p, color);
            p = T::add(p, 1);
            ++x1;
            mask = ror8 (mask, 1);
        }
    }
}

/** Horizontal Line. Implementation of the `horizontal line with alpha
    channel' primitive for all modes. */
template<class T>
inline void
doHLineAlpha(SDL_Surface* sfc, int x1, int y1, int x2, typename T::pix_type color,
             uint8 pat, alpha_type alpha)
{
    if (pat == 0 || alpha == 0) {
        /* nothing */
    } else if (pat == 255) {
        /* solid line */
        uint8* p = T::add ((uint8*)sfc->pixels + y1 * sfc->pitch, x1);
        while (x1 < x2) {
            T::poke(p, T::mix(sfc->format, T::peek(p), color, alpha));
            p = T::add(p, 1);
            ++x1;
        }
    } else {
        /* pattern line */
        uint8 mask = ror8 (0x80, x1);
        uint8* p = T::add ((uint8*)sfc->pixels + y1 * sfc->pitch, x1);
        while (x1 < x2) {
            if (mask & pat)
                T::poke(p, T::mix(sfc->format, T::peek(p), color, alpha));
            p = T::add(p, 1);
            ++x1;
            mask = ror8 (mask, 1);
        }
    }
}

/** Vertical Line. Implementation of the `vertical line' primitive
    for all modes. */
template<class T>
inline void
doVLine(SDL_Surface* sfc, int x1, int y1, int y2, typename T::pix_type color,
        uint8 pat)
{
    if (pat == 0) {
        /* nothing */
    } else if (pat == 255) {
        uint8* p = T::add ((uint8*)sfc->pixels + y1 * sfc->pitch, x1);
        while (y1 < y2) {
            T::poke(p, color);
            p += sfc->pitch;
            ++y1;
        }
    } else {
        uint8 mask = ror8 (0x80, y1);
        uint8* p = T::add ((uint8*)sfc->pixels + y1 * sfc->pitch, x1);
        while (y1 < y2) {
            if (mask & pat)
                T::poke(p, color);
            p += sfc->pitch;
            ++y1;
            mask = ror8 (mask, 1);
        }
    }
}

/** Vertical Line. Implementation of the `vertical line with Alpha channel'
    primitive for all modes. */
template<class T>
inline void
doVLineAlpha(SDL_Surface* sfc, int x1, int y1, int y2, typename T::pix_type color,
             uint8 pat, alpha_type alpha)
{
    if (pat == 0 || alpha == 0) {
        /* nothing */
    } else if (pat == 255) {
        uint8* p = T::add ((uint8*)sfc->pixels + y1 * sfc->pitch, x1);
        while (y1 < y2) {
            T::poke(p, T::mix(sfc->format, T::peek(p), color, alpha));
            p += sfc->pitch;
            ++y1;
        }
    } else {
        uint8 mask = ror8 (0x80, y1);
        uint8* p = T::add ((uint8*)sfc->pixels + y1 * sfc->pitch, x1);
        while (y1 < y2) {
            if (mask & pat)
                T::poke(p, T::mix(sfc->format, T::peek(p), color, alpha));
            p += sfc->pitch;
            ++y1;
            mask = ror8 (mask, 1);
        }
    }
}

/** Implementation of the `blit pattern' primitive.
    \param x,y             position of MSB of data[0]
    \param bytes_per_line  size of one line (no restrictions!)
    \param data            bitmap data
    \param color           color value in target format */
template<class T>
inline void
doBlitPatternTransp(SDL_Surface* sfc, GfxRect rect, int x, int y,
                    int bytes_per_line, const uint8* data, typename T::pix_type color)
{
    x = rect.x - x;                // columns to skip in data image
    y = rect.y - y;                // lines to skip in data image
    data += bytes_per_line * y + x/8;

    uint8  zmask = 0x80 >> (x & 7);
    uint8* zmem = T::add ((uint8*)sfc->pixels + rect.y * sfc->pitch, rect.x);
    while (rect.h > 0) {
        uint8 mask = zmask;
        uint8* mem = zmem;
        const uint8* pat = data;
        for (unsigned k = 0; k < rect.w; ++k) {
            if (*pat & mask)
                T::poke (mem, color);
            mem = T::add(mem, 1);
            mask >>= 1;
            if (!mask) {
                mask = 0x80;
                ++pat;
            }
        }
        --rect.h;
        data += bytes_per_line;
        zmem += sfc->pitch;
    }
}

template<class T>
inline void
doBlitPattern(SDL_Surface* sfc, GfxRect rect, int x, int y,
              int bytes_per_line, const uint8* data, typename T::pix_type color,
              typename T::pix_type bg)
{
    x = rect.x - x;                // columns to skip in data image
    y = rect.y - y;                // lines to skip in data image
    data += bytes_per_line * y + x/8;

    uint8  zmask = 0x80 >> (x & 7);
    uint8* zmem = T::add ((uint8*)sfc->pixels + rect.y * sfc->pitch, rect.x);
    while (rect.h > 0) {
        uint8 mask = zmask;
        uint8* mem = zmem;
        const uint8* pat = data;
        for (unsigned k = 0; k < rect.w; ++k) {
            if (*pat & mask)
                T::poke (mem, color);
            else
                T::poke (mem, bg);
            mem = T::add(mem, 1);
            mask >>= 1;
            if (!mask) {
                mask = 0x80;
                ++pat;
            }
        }
        --rect.h;
        data += bytes_per_line;
        zmem += sfc->pitch;
    }
}

template<class T>
inline void
doBlitPatternAlpha(SDL_Surface* sfc, GfxRect rect, int x, int y,
                   int bytes_per_line, const uint8* data, typename T::pix_type color,
                   uint32 bg, alpha_type alpha)
{
    x = rect.x - x;                // columns to skip in data image
    y = rect.y - y;                // lines to skip in data image
    data += bytes_per_line * y + x/8;

    uint8  zmask = 0x80 >> (x & 7);
    uint8* zmem = T::add ((uint8*)sfc->pixels + rect.y * sfc->pitch, rect.x);
    while (rect.h > 0) {
        uint8 mask = zmask;
        uint8* mem = zmem;
        const uint8* pat = data;
        for (unsigned k = 0; k < rect.w; ++k) {
            if (*pat & mask)
                T::poke (mem, T::mix (sfc->format, T::peek(mem), color, alpha));
            else if (bg != GFX_TRANSPARENT)
                T::poke (mem, T::mix (sfc->format, T::peek(mem), bg, alpha));
            mem = T::add(mem, 1);
            mask >>= 1;
            if (!mask) {
                mask = 0x80;
                ++pat;
            }
        }
        --rect.h;
        data += bytes_per_line;
        zmem += sfc->pitch;
    }
}

template<class T>
inline typename T::pix_type
doPeek(SDL_Surface* sfc, int x, int y)
{
    return T::peek(T::add((uint8*) sfc->pixels + y * sfc->pitch, x));
}

//////////////////////////////////////////////////////////////////////

/** \class GfxSurface
    \brief SDL GfxSurface.
    This is a direct interface to an SDL \c GfxSurface. */

/** Use existing surface. We'll draw on the specified surface.
    If aowner is true, the surface will be deleted when this
    object is destroyed. The surface must be unlocked; noone
    else must play with the lock state.

    This is used to get a drawing surface for the screen after
    SDL_SetVideoMode(). */
GfxSurface::GfxSurface(SDL_Surface* asfc, bool aowner)
    : sfc(asfc), owner(aowner), locked(0)
{ }

/** Memory GfxSurface. Creates a surface with the specified pixel
    format and the specified dimensions. Generally, one does
    \code
      Ptr<GfxSurface> bmp = new GfxSurface(screen->getSurface()->format,
                                     width, height);
    \endcode
    to get a blittable surface. */
GfxSurface::GfxSurface(int wi, int he, SDL_PixelFormat* fmt)
    : sfc(SDL_CreateRGBSurface(SDL_SWSURFACE, wi, he,
                               fmt->BitsPerPixel,
                               fmt->Rmask, fmt->Gmask,
                               fmt->Bmask, fmt->Amask)),
      owner(true),
      locked(0)
{
    if (!sfc)
        throw GraphicsIOException(SDL_GetError());
}

GfxSurface::~GfxSurface()
{
    if (owner && sfc)
        SDL_FreeSurface(sfc);
}

void
GfxSurface::setSurface(SDL_Surface* asfc, bool aown)
{
    /* FIXME: lock count? for now, assume it's not locked. */
    ASSERT(!locked);

    if (owner)
        SDL_FreeSurface(sfc);
    sfc = asfc;
    owner = aown;
}

void
GfxSurface::drawHLine(int x1, int y1, int x2, uint32 color, uint8 pat, alpha_type alpha)
{
    /* clipping */
    if (y1 < 0 || y1 >= sfc->h)
        return;
    if (x1 < 0)
        x1 = 0;
    if (x2 > sfc->w)
        x2 = sfc->w;
    if (x1 >= x2)
        return;

    if (alpha == 255)
        GFX_MODE_SWITCH(doHLine, (sfc, x1, y1, x2, color, pat));
    else
        GFX_MODE_SWITCH(doHLineAlpha, (sfc, x1, y1, x2, color, pat, alpha));
    return;
}

void
GfxSurface::drawVLine(int x1, int y1, int y2, uint32 color, uint8 pat, alpha_type alpha)
{
    /* clipping */
    if (x1 < 0 || x1 >= sfc->w)
        return;
    if (y1 < 0)
        y1 = 0;
    if (y2 > sfc->h)
        y2 = sfc->h;
    if (y1 >= y2)
        return;

    if (alpha == 255)
        GFX_MODE_SWITCH(doVLine, (sfc, x1, y1, y2, color, pat));
    else
        GFX_MODE_SWITCH(doVLineAlpha, (sfc, x1, y1, y2, color, pat, alpha));
    return;
}

void
GfxSurface::drawPixel(int x, int y, uint32 color, alpha_type alpha)
{
    if (x < 0 || x >= sfc->w || y < 0 || y >= sfc->h)
        return;

    if (alpha == 255)
        GFX_MODE_SWITCH(doPixel, (sfc, x, y, color));
    else
        GFX_MODE_SWITCH(doPixelAlpha, (sfc, x, y, color, alpha));
}

void
GfxSurface::drawBar(GfxRect rect, uint32 color, uint32 bg,
                    const GfxFillPattern& pat, alpha_type alpha)
{
    rect.clipSurface(sfc);
    if (!rect.exists())
        return;

    if (alpha == 255 && pat.isBlack()) {
        SDL_FillRect(sfc, &rect, color);
    } else {
        int x1 = rect.x, y1 = rect.y, x2 = rect.x + rect.w, h = rect.h;
        if (alpha == 255) {
            if (bg == GFX_TRANSPARENT)
                while (h > 0) {
                    GFX_MODE_SWITCH (doHLine, (sfc, x1, y1, x2, color, pat[y1]));
                    ++y1; --h;
                }
            else
                while (h > 0) {
                    GFX_MODE_SWITCH (doHLine, (sfc, x1, y1, x2, color, pat[y1]));
                    GFX_MODE_SWITCH (doHLine, (sfc, x1, y1, x2, bg, ~pat[y1]));
                    ++y1; --h;
                }
        } else {
            while (h > 0) {
                GFX_MODE_SWITCH(doHLineAlpha, (sfc, x1, y1, x2, color, pat[y1], alpha));
                if (bg != GFX_TRANSPARENT)
                    GFX_MODE_SWITCH(doHLineAlpha, (sfc, x1, y1, x2, bg, ~pat[y1], alpha));
                ++y1;
                --h;
            }
        }
    }
}

/** Lock surface. Locks the surface if needed. */
bool
GfxSurface::lock()
{
    if (SDL_MUSTLOCK (sfc)) {
        if (SDL_LockSurface(sfc) == 0) {
            ++locked;
            return true;
        } else {
            return false;
        }
    } else {
        return true;
    }
}

/** Unlock surface. Unlocks the surface if needed. */
void
GfxSurface::unlock()
{
    if (SDL_MUSTLOCK (sfc)) {
        SDL_UnlockSurface(sfc);
        --locked;
    }
}

/** Copy other surface (pixmap).
    \param x,y   anchor point. src's (0,0) is here;
    \param src   source surface;
    \param rect  clip rectangle, in src coordinates.

    `blitSurface(x, y, src, GfxRect(src))' will display the whole
    surface. `blitSurface(x, y, src, GfxRect(2, 3, 4, 5))' will
    display a 4x5 area from surface src, at (x+2, y+3) up to but not
    including (x+2+4, y+3+5). */
void
GfxSurface::blitSurface(int x, int y, SDL_Surface* src, GfxRect rect)
{
    SDL_Rect dstrect = { x + rect.x, y + rect.y, 0, 0 };
    int lockc = locked;
    while (locked)
        unlock();
    SDL_BlitSurface(src, &rect, sfc, &dstrect);
    while (locked < lockc)
        lock();
}

/** Display pattern (monochrome bitmap).
    \param rect    position, screen coordinates;
    \param x,y     position of bit 7, data[0], on the screen;
    \param bytes_per_line  bytes per line in data;
    \param data    pixel data. bit 7 = left, bit 0 = right (VGA bit order);
    \param color   color of "on" bits;
    \param bg      color of "off" bits, or GFX_TRANSPARENT;
    \param alpha   alpha channel.

    \pre x <= rect.x, y <= rect.y

    Note that this function has no constraints on the bytes_per_line
    parameter. bytes_per_line can also be zero or negative for
    interesting effects. */
void
GfxSurface::blitPattern(GfxRect rect, int x, int y, int bytes_per_line,
                        const uint8* data, uint32 color, uint32 bg,
                        alpha_type alpha)
{
    rect.clipSurface(sfc);
    if (rect.exists()) {
        if (alpha == 255) {
            if (bg == GFX_TRANSPARENT)
                GFX_MODE_SWITCH(doBlitPatternTransp, (sfc, rect, x, y, bytes_per_line, data, color));
            else
                GFX_MODE_SWITCH(doBlitPattern, (sfc, rect, x, y, bytes_per_line, data, color, bg));
        } else {
            GFX_MODE_SWITCH(doBlitPatternAlpha, (sfc, rect, x, y, bytes_per_line, data, color, bg, alpha));
        }
    }
}

/** Update. Causes the specified region be redisplayed. */
void
GfxSurface::update(GfxRect r)
{
    r.clipSurface(sfc);
    if (r.w && r.h)
        SDL_UpdateRect(sfc, r.x, r.y, r.w, r.h);
}

SDL_Surface*
GfxSurface::getSurface()
{
    return sfc;
}

GfxRect
GfxSurface::computeClipRect(GfxRect r)
{
    r.clipSurface(sfc);
    return r;
}

uint32
GfxSurface::getPixel(int x, int y)
{
    if (x >= 0 && y >= 0 && x < sfc->w && y < sfc->h)
        GFX_MODE_SWITCH(return doPeek, (sfc, x, y));
    return 0;
}

//////////////////////////////////////////////////////////////////////

GfxFilter::GfxFilter(Ptr<GfxCanvas> aparent)
    : parent(aparent)
{ }

GfxFilter::GfxFilter()
{ }

bool
GfxFilter::lock()
{
    return parent->lock();
}

void
GfxFilter::unlock()
{
    parent->unlock();
}

void
GfxFilter::update(GfxRect r)
{
    parent->update(r);
}

SDL_Surface*
GfxFilter::getSurface()
{
    return parent->getSurface();
}

uint32
GfxFilter::getPixel(int x, int y)
{
    return parent->getPixel(x, y);
}
