/**
  *  \file gfx/gfx.h
  *  \brief Graphics Kernel Interface
  *
  *  Prototypes and declarations for graphics kernel (higher-level
  *  API wrapped around SDL).
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GFX_GFX_H
#define PCC_V2_GFX_GFX_H

#include <SDL_video.h>
#include "gfx/gfxutil.h"
#include "cpluslib/misc.h"
#include "cpluslib/smartptr.h"

/** Transparent color. When a background color is requested, specify
    this value to draw transparent. This value has been chosen to be
    a very unlikely if not impossible return value of SDL_MapRGB(). */
static const uint32 GFX_TRANSPARENT = 0xFFFFFFFEU;

/** Type for alpha channel. In theory, any type does (and int should
    be the fastest of them). Right now, only uint8 seems to work, though. */
typedef uint8 alpha_type;

/// Graphics Parameter Collection.
struct GfxParameters {
    uint32 color;                  ///< Color in target format
    uint32 bgcolor;                ///< Background color, if used
    unsigned line_thickness;       ///< Line thickness
    uint8 line_style;              ///< Line pattern
    GfxFillPattern fill_pattern;   ///< Bar fill pattern
    alpha_type alpha;              ///< Alpha channel
    int x, y;                      ///< Graphics Cursor
    int just_x, just_y;            ///< Alignment for ::outText()

    GfxParameters();
    void moveTo(int ax, int ay) { x = ax, y = ay; }
    void moveRel(int ax, int ay) { x += ax, y += ay; }
};

class GfxCanvas : public Object {
 public:
    virtual ~GfxCanvas();

    virtual void drawHLine(int x1, int y1, int x2, uint32 color, uint8 pat, alpha_type alpha) = 0;
    virtual void drawVLine(int x1, int y1, int y2, uint32 color, uint8 pat, alpha_type alpha) = 0;
    virtual void drawPixel(int x, int y, uint32 color, alpha_type alpha) = 0;
    virtual void drawBar(GfxRect rect, uint32 color, uint32 bg, const GfxFillPattern& pat, alpha_type alpha) = 0;

    virtual void blitSurface(int x, int y, SDL_Surface* src, GfxRect rect) = 0;
    virtual void blitPattern(GfxRect rect, int x, int y, int bytes_per_line, const uint8* data, uint32 color, uint32 bg, alpha_type alpha) = 0;

    virtual bool lock() = 0;
    virtual void unlock() = 0;
    virtual void update(GfxRect r) = 0;

    virtual GfxRect computeClipRect(GfxRect r) = 0;
    virtual bool isVisible(GfxRect r);
    virtual bool isClipped(GfxRect r);

    virtual SDL_Surface* getSurface() = 0;

    virtual uint32 getPixel(int x1, int y1) = 0;
};

class GfxSurface : public GfxCanvas {
 protected:
    SDL_Surface* sfc;
    bool owner;
    int locked;

    void setSurface(SDL_Surface* asfc, bool aown = true);

 public:
    GfxSurface(SDL_Surface* asfc, bool aowner);
    GfxSurface(int wi, int he, SDL_PixelFormat* fmt);
    ~GfxSurface();

    void drawHLine(int x1, int y1, int x2, uint32 color, uint8 pat, alpha_type alpha);
    void drawVLine(int x1, int y1, int y2, uint32 color, uint8 pat, alpha_type alpha);
    void drawPixel(int x, int y, uint32 color, alpha_type alpha);
    void drawBar(GfxRect rect, uint32 color, uint32 bg, const GfxFillPattern& pat, alpha_type alpha);

    void blitSurface(int x, int y, SDL_Surface* src, GfxRect rect);
    void blitPattern(GfxRect rect, int x, int y, int bytes_per_line, const uint8* data, uint32 color, uint32 bg, alpha_type alpha);

    bool lock();
    void unlock();
    void update(GfxRect r);

    GfxRect computeClipRect(GfxRect r);

    SDL_Surface* getSurface();

    uint32 getPixel(int x, int y);
};

class GfxFilter : public GfxCanvas {
 protected:
    Ptr<GfxCanvas> parent;
    GfxFilter(Ptr<GfxCanvas> aparent);
    GfxFilter();
 public:
    void setParent(Ptr<GfxCanvas> apar) { parent = apar; }
    Ptr<GfxCanvas> getParent() const { return parent; }

    bool lock();
    void unlock();
    void update(GfxRect r);

    SDL_Surface* getSurface();

    uint32 getPixel(int x, int y);
};

class GfxXorFilter : public GfxCanvas {
    Ptr<GfxCanvas> parent;
 public:
    GfxXorFilter(Ptr<GfxCanvas> apar);
    ~GfxXorFilter();

    void drawHLine(int x1, int y1, int x2, uint32 color, uint8 pat, alpha_type alpha);
    void drawVLine(int x1, int y1, int y2, uint32 color, uint8 pat, alpha_type alpha);
    void drawPixel(int x, int y, uint32 color, alpha_type alpha);
    void drawBar(GfxRect rect, uint32 color, uint32 bg, const GfxFillPattern& pat, alpha_type alpha);

    void blitSurface(int x, int y, SDL_Surface* src, GfxRect rect);
    void blitPattern(GfxRect rect, int x, int y, int bytes_per_line, const uint8* data, uint32 color, uint32 bg, alpha_type alpha);

    bool lock();
    void unlock();
    void update(GfxRect r);

    GfxRect computeClipRect(GfxRect r);

    SDL_Surface* getSurface();
    
    uint32 getPixel(int x, int y);
};

#include "gfx/gfxcompl.h"

#endif
