/**
  *  \file gfx/gfxcompl.cc
  *  \brief Complex Graphics Primitives
  *
  *  Graphics primitives built atop the others. Hence clipping need not
  *  deal with clipping a circle or a general line.
  *
  *  In addition, graphics primitives that are just wrappers with a more
  *  convenient interface.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <cmath>
#include "gfx/gfx.h"
#include "gfx/gfxcompl.h"

#define LINE_PATTERN(parm, t) ((parm).line_style & (0x80 >> (t & 7)))

static void
line_pset(GfxCanvas* c, int x1, int y1, GfxParameters& parm)
{
    c->drawPixel (x1, y1, parm.color, parm.alpha);
}

static void
line_hline(GfxCanvas* c, int x1, int y1, GfxParameters& parm)
{
    c->drawHLine (x1, y1, x1+parm.line_thickness, parm.color, 0xFF, parm.alpha);
}

static void
line_vline(GfxCanvas* c, int x1, int y1, GfxParameters& parm)
{
    c->drawVLine (x1, y1, y1+parm.line_thickness, parm.color, 0xFF, parm.alpha);
}

/** \name Complex Graphics Primitives

    These graphics primitives are built atop the other primitives.
    They need not, and can not, be overridden in GfxCanvas' descendants.

    They take their parameters from a GfxParameters object. */
//@{

/** Draw Horizontal Line. Draw a horizontal line from (x1,y1) to (x2,y1),
    inclusive. Uses the parameters from \c parm. */
void
drawHLine(GfxCanvas* can, GfxParameters& parm, int x1, int y1, int x2)
{
    if (x1 > x2)
        std::swap(x1, x2);
    x2++;

    if (parm.line_thickness == 1) {
        can->drawHLine(x1, y1, x2, parm.color, parm.line_style, parm.alpha);
    } else {
        GfxFillPattern pat (parm.line_style);
        GfxRect r (x1,    y1 - parm.line_thickness/2,
                   x2-x1, parm.line_thickness);
        can->drawBar(r, parm.color, GFX_TRANSPARENT, pat, parm.alpha);
    }
}

/** Draw Vertical Line. Draw a vertical line from (x1,y1) to (x1,y2),
    inclusive. Uses the parameters from \c parm. */
void
drawVLine(GfxCanvas* can, GfxParameters& parm,
          int x1, int y1, int y2)
{
    if (y2 < y1)
        std::swap(y1, y2);
    y2++;

    if (parm.line_thickness == 1) {
        can->drawVLine(x1, y1, y2, parm.color, parm.line_style, parm.alpha);
    } else {
        GfxFillPattern pat;
        for (int i = 0; i < 8; ++i)
            if ((parm.line_style << i) & 0x80)
                pat[i] = 0xFF;
        GfxRect r (x1 - parm.line_thickness/2, y1,
                   parm.line_thickness,        y2-y1);
        can->drawBar(r, parm.color, GFX_TRANSPARENT, pat, parm.alpha);
    }
}

/** Draw Line. Draws a general line from (x1,y1) to (x2,y2).
    This is a standard Bresenham algorithm, with shortcuts for
    horizontal and vertical lines. Supports all graphics parameters. */
void
drawLine(GfxCanvas* can, GfxParameters& parm,
         int x1, int y1, int x2, int y2)
{
    /* Handle easy cases first */
    if (x1 == x2) {
        drawVLine(can, parm, x1, y1, y2);
        return;
    }
    if (y1 == y2) {
        drawHLine(can, parm, x1, y1, x2);
        return;
    }

    /* Vanilla Bresenham */
    int dx = x2 - x1;
    int dy = y2 - y1;
    if(dx < 0)
        dx = -dx;
    if(dy < 0)
        dy = -dy;

    if(dx < dy) {                // Y is the major axis
	if(y1 > y2) {
	    std::swap(y1, y2);
	    std::swap(x1, x2);
	}

        void (*plot) (GfxCanvas*,int,int,GfxParameters&);
	int error = dy >> 2;
	int addx  = (x2 < x1) ? -1 : 1;

	if(parm.line_thickness == 1) {
            plot = line_pset;
	} else {
            /* FIXME: correct error term for even line thicknesses */
	    x1 -= parm.line_thickness / 2;
	    plot = line_hline;
	}

	while(y1 <= y2) {
            if (LINE_PATTERN (parm, y1))
                plot (can, x1, y1, parm);
	    y1++;
	    error += dx;
	    if(error > dy) {
		error -= dy;
		x1 += addx;
	    }
	}
    } else {                // X is the major axis
	if(x1 > x2) {
	    std::swap(y1, y2);
	    std::swap(x1, x2);
	}

        void (*plot) (GfxCanvas*,int,int,GfxParameters&);
	int error = dx >> 2;
	int addy  = (y2<y1) ? -1 : 1;

	if(parm.line_thickness==1) {
            plot = line_pset;
	} else {
            /* FIXME: correct error term for even line thicknesses */
	    y1 -= parm.line_thickness/2;
	    plot = line_vline;
	}

	while(x1 <= x2) {
            if (LINE_PATTERN (parm, x1))
                plot (can, x1, y1, parm);
	    x1++;
	    error += dy;
	    if(error > dx) {
		error -= dx;
		y1 += addy;
	    }
	}
    }
}

/** Draw Line to. Draws a line between the current graphics cursor in
    \c parm to the specified (x,y) coordinates. Otherwise like line(). */
void
drawLineTo (GfxCanvas* can, GfxParameters& parm, int x, int y)
{
    drawLine (can, parm, parm.x, parm.y, x, y);
    parm.moveTo (x, y);
}

/** Draw Relative Line. Draws a line from the current graphics cursor in
    \c parm to the point offset by (dx,dy) from it. Otherwise like
    line(). */
void
drawLineRel (GfxCanvas* can, GfxParameters& parm, int dx, int dy)
{
    drawLineTo (can, parm, parm.x + dx, parm.y + dy);
}

/** Circle. Draws a circle around (x0,y0), with the specified radius.
    All line-related parameters are taken into account.

    \todo some pixels in the diagonals are painted twice which is a
    Bad Thing(tm) for alpha circles */
void
drawCircle(GfxCanvas* can, GfxParameters& parm, int x0, int y0, int r)
{
    /*
     *  A home-brew circle algorithm (but probably not at all new):
     *  . it uses the equation sum(1,3,5,...,2n-1) = n^2 to compute
     *    pythagoras' theorem.
     *  . first implemented in Z80 assembly language
     *  If you have a better algorithm, tell me.
     */
#if 1
    /* FIXME: This is the circle routine used in PCC I, which does not
       support line-patterned circles and is quite slow. The other one
       below has different (=wrong) rounding characteristics. We want
       a circle of radius 3 to appear as a "rounded rectangle" with
       diameter 7, the other draws squares with diameter 5 instead. */
    register int w = 0;
    register int k = 1;
    register int z = 0;
    register int y2 = 0;
    register int xk = 2*r-1;
    register int x = r;
    while (x >= 0) {
        while (z < y2) {
            z += k;
            ++w;
            k += 2;
        }
        if (w >= x) {
            can->drawPixel(x0 - x, y0 - w, parm.color, parm.alpha);
            can->drawPixel(x0 + x, y0 - w, parm.color, parm.alpha);
            can->drawPixel(x0 - w, y0 - x, parm.color, parm.alpha);
            can->drawPixel(x0 + w, y0 - x, parm.color, parm.alpha);
            can->drawPixel(x0 - w, y0 + x, parm.color, parm.alpha);
            can->drawPixel(x0 + w, y0 + x, parm.color, parm.alpha);
            can->drawPixel(x0 - x, y0 + w, parm.color, parm.alpha);
            can->drawPixel(x0 + x, y0 + w, parm.color, parm.alpha);
        }
        y2 += xk;
        xk -= 2;
        --x;
    }
#else
    register int w = 0;
    register int k = 1;
    register int z = 0;
    register int y2 = 0;
    register int xk = 2*r-1;
    register int x = r;

    register int w0 = w;
    while(x >= w) {
        while(z < y2) {
            z += k;
            drawVLine(can, parm, x0+x, y0+w0, y0+w);
            drawVLine(can, parm, x0-x, y0+w0, y0+w);
            drawHLine(can, parm, x0+w0, y0+x, x0+w);
            drawHLine(can, parm, x0+w0, y0-x, x0+w);
            if (w) {
                drawVLine(can, parm, x0+x, y0-w0, y0-w);
                drawVLine(can, parm, x0-x, y0-w0, y0-w);
                drawHLine(can, parm, x0-w0, y0+x, x0-w);
                drawHLine(can, parm, x0-w0, y0-x, x0-w);
            }
            w++;
            w0 = w;
            k += 2;
        }
        y2 += xk;
        xk -= 2;
        x--;
    }
#endif
}

void
drawFilledCircle(GfxCanvas* can, GfxParameters& parm, int x0, int y0, int r)
{
    /* Roughly the same algorithm as /Circle/ */
    register int w = 0;
    register int k = 1;
    register int z = 0;
    register int y2 = 0;
    register int xk = 2*r-1;
    register int x = r;

    register int w0 = w;
    while(x >= 0) {
        while(z < y2) {
            z += k;
            drawBar(can, parm, x0 - x, y0 + w0, x0 + x, y0 + w);
            if (w)
                drawBar(can, parm, x0 - x, y0 - w0, x0 + x, y0 - w);
            w++;
            w0 = w;
            k += 2;
        }
        y2 += xk;
        xk -= 2;
        x--;
    }
}

/** Draw Filled Bar. Draws a filled bar between the corners (x1,y1)
    and (x2,y2), inclusive. Uses color and fill pattern from \c parm. */
void
drawBar (GfxCanvas* can, GfxParameters& parm, int x1, int y1, int x2, int y2)
{
    if (x1 > x2)
        std::swap (x1, x2);
    if (y1 > y2)
        std::swap (y1, y2);

    GfxRect r(x1, y1, x2-x1+1, y2-y1+1);
    can->drawBar (r, parm.color, GFX_TRANSPARENT, parm.fill_pattern, parm.alpha);
}

/** Draw Filled Bar. Fills rectangle r with color and fill pattern
    from \c parm. Note that when you know alpha and color, you can
    also call drawBar() directly. */
void
drawBar (GfxCanvas* can, GfxParameters& parm, const GfxRect& r)
{
    can->drawBar (r, parm.color, GFX_TRANSPARENT, parm.fill_pattern, parm.alpha);
}

void
drawSolidBar(GfxCanvas* can, const GfxRect& r, uint32 color)
{
    can->drawBar(r, color, GFX_TRANSPARENT, SOLID_FILL, SDL_ALPHA_OPAQUE);
}

void
drawRectangle(GfxCanvas* can, GfxParameters& parm, const GfxRect& r)
{
    if (r.w == 0 || r.h == 0) {
        return;
    } else if (r.w == 1) {
        drawVLine(can, parm, r.x, r.y, r.y + r.h - 1);
    } else if (r.h == 1) {
        drawHLine(can, parm, r.x, r.y, r.x + r.w - 1);
    } else {
        int x2 = r.x + r.w - 1;
        int y2 = r.y + r.h - 1;
        drawHLine(can, parm, r.x, r.y, x2);
        drawHLine(can, parm, r.x,  y2, x2);
        drawVLine(can, parm, r.x, r.y+1, y2-1);
        drawVLine(can, parm, x2,  r.y+1, y2-1);
    }
}

static inline int32 square(int32 x) { return x*x; }

/** Draw an arrow.
    \param can,parm  drawing parameters
    \param x1,y1     starting coordinates
    \param x2,y2     ending coordinates (the pointer is added here)
    \param ptsize    length of pointer in pixels, approximately */
void
drawArrow(GfxCanvas* can, GfxParameters& parm, int x1, int y1, int x2, int y2, int ptsize)
{
    drawLine(can, parm, x1, y1, x2, y2);

    int l = int(std::sqrt(double(square(x2-x1) + square(y2-y1))));
    if (l) {
        int dx = (x2 - x1) * ptsize / l;
        int dy = (y2 - y1) * ptsize / l;
        drawLine(can, parm, x2, y2, x2-dy-dx, y2+dx-dy);
        drawLine(can, parm, x2, y2, x2+dy-dx, y2-dx-dy);
    }
}

//@}
