/**
  *  \file gfx/gfxutil.cc
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <iostream>
#include "gfx/gfxutil.h"
#include "cpluslib/bitops.h"

static const uint8 solid_init[]   = { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };
static const uint8 gray50_init[]  = { 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA };
static const uint8 gray25_init[]  = { 0xAA, 0x00, 0x55, 0x00, 0xAA, 0x00, 0x55, 0x00 };
static const uint8 ltslash_init[] = { 0x88, 0x44, 0x22, 0x11, 0x88, 0x44, 0x22, 0x11 };

const GfxFillPattern SOLID_FILL(solid_init),
    GRAY50_FILL(gray50_init),
    GRAY25_FILL(gray25_init),
    GRAY50_FILL2(gray50_init+1),
    LTSLASH_FILL(ltslash_init);

/** Construct blank pattern. Constructs a pattern that is all-zero. */
GfxFillPattern::GfxFillPattern()
{
    std::fill(pat, pat+8, uint8(0));
}

/** Construct pattern from byte array.
    \param init pointer to 8 bytes */
GfxFillPattern::GfxFillPattern(const uint8* init)
{
    std::copy(init, init+8, pat);
}

/** Construct pattern with specified value. Constructs a pattern that
    contains the specified value in each line. */
GfxFillPattern::GfxFillPattern(uint8 init)
{
    std::fill(pat, pat+8, init);
}

/** Check for blank pattern. \returns true iff the pattern is all-zero,
    i.e. when it is used for filling something, nothing would happen. */
bool
GfxFillPattern::isBlank() const
{
    for (int i = 0; i < 8; ++i)
        if (pat[i])
            return false;
    return true;
}

/** Check for black pattern. \returns true iff the pattern is all-one,
    i.e. filling something with it would result in a completely filled
    surface. */
bool
GfxFillPattern::isBlack() const
{
    for (int i = 0; i < 8; ++i)
        if (pat[i] != 255)
            return false;
    return true;
}

/** Shift pattern to the left. \returns reference to this pattern
    to allow chaining. */
GfxFillPattern&
GfxFillPattern::shiftLeft(int amount)
{
    for (int i = 0; i < 8; ++i)
        pat[i] = rol8(i, amount);
    return *this;
}

/** Shift pattern to the right. \returns reference to this pattern
    to allow chaining. */
GfxFillPattern&
GfxFillPattern::shiftRight(int amount)
{
    for (int i = 0; i < 8; ++i)
        pat[i] = ror8(i, amount);
    return *this;
}

/** Shift pattern up. \returns reference to this pattern
    to allow chaining. */
GfxFillPattern&
GfxFillPattern::shiftUp(int amount)
{
    GfxFillPattern tmp = *this;
    for (int i = 0; i < 8; ++i)
        pat[i] = tmp.pat[unsigned(i+amount) % 8];
    return *this;
}

/** Shift pattern down. \returns reference to this pattern
    to allow chaining. */
GfxFillPattern&
GfxFillPattern::shiftDown(int amount)
{
    GfxFillPattern tmp = *this;
    for (int i = 0; i < 8; ++i)
        pat[i] = tmp.pat[unsigned(i-amount) % 8];
    return *this;
}

/** `Or' two patterns. The result has 1-bits where either of the two
    patterns had 1-bits. \returns reference to this pattern to allow
    chaining. */
GfxFillPattern&
GfxFillPattern::operator |= (const GfxFillPattern& rhs)
{
    for (int i = 0; i < 8; ++i)
        pat[i] |= rhs.pat[i];
    return *this;
}

/** `Or' pattern with value. \overload */
GfxFillPattern&
GfxFillPattern::operator |= (uint8 value)
{
    for (int i = 0; i < 8; ++i)
        pat[i] |= value;
    return *this;
}

/** `And' two patterns. The result has 1-bits where both of the two
    patterns had 1-bits. \returns reference to this pattern to allow
    chaining. */
GfxFillPattern&
GfxFillPattern::operator &= (const GfxFillPattern& rhs)
{
    for (int i = 0; i < 8; ++i)
        pat[i] &= rhs.pat[i];
    return *this;
}

/** `And' pattern and value. \overload */
GfxFillPattern&
GfxFillPattern::operator &= (uint8 value)
{
    for (int i = 0; i < 8; ++i)
        pat[i] &= value;
    return *this;
}

/** `Xor' two patterns. The result has 1-bits where either of the two
    patterns (but not both) had 1-bits. \returns reference to this
    pattern to allow chaining. */
GfxFillPattern&
GfxFillPattern::operator ^= (const GfxFillPattern& rhs)
{
    for (int i = 0; i < 8; ++i)
        pat[i] ^= rhs.pat[i];
    return *this;
}

/** `Xor' pattern and value. \overload */
GfxFillPattern&
GfxFillPattern::operator ^= (uint8 value)
{
    for (int i = 0; i < 8; ++i)
        pat[i] ^= value;
    return *this;
}

/** Invert pattern. \returns reference to this pattern to allow
    chaining. */
GfxFillPattern&
GfxFillPattern::invert()
{
    for (int i = 0; i < 8; ++i)
        pat[i] = ~ pat[i];
    return *this;
}

//////////////////////////////////////////////////////////////////////

/** Construct empty GfxRect.
    \post contains(p) == false for every p */
GfxRect::GfxRect()
{
    x = y = w = h = 0;
}

/** Construct GfxRect from SDL_Rect. */
GfxRect::GfxRect(const SDL_Rect& r)
    : SDL_Rect(r)
{ }

/** Construct GfxRect from explicit parameters.
    \param x left
    \param y top
    \param w width
    \param h height */
GfxRect::GfxRect(int x, int y, int w, int h)
{
    this->x = x;
    this->y = y;
    this->w = w;
    this->h = h;
}

/** Construct bounding rectangle for GfxSurface.
    Constructs a rectangle that covers the whole surface \c sfc. */
GfxRect::GfxRect(const SDL_Surface* sfc)
{
    x = y = 0;
    w = sfc->w;
    h = sfc->h;
}

/** Clip at surface. Modify this rectangle so that it lies entirely
    within \c sfc. */
void
GfxRect::clipSurface(const SDL_Surface* sfc)
{
    register long sw = w, sh = h;
    if (x < 0)
        sw += x, x = 0;
    if (y < 0)
        sh += y, y = 0;
    if (x + sw > sfc->w)
        sw = sfc->w - x;
    if (y + h > sfc->h)
        sh = sfc->h - y;
    w = (sw < 0) ? 0 : sw;
    h = (sh < 0) ? 0 : sh;
}

/** Clip at rectangle. Modify this rectangle so that it lies entirely
    within the other rectangle, \c r.

    This function is also available under the alternate name intersect(). */
void
GfxRect::clipRect(const SDL_Rect& r)
{
    register long sw = w, sh = h;
    if (x < r.x)
        sw += x - r.x, x = r.x;
    if (y < r.y)
        sh += y - r.y, y = r.y;
    if (x + sw > r.w + r.x)
        sw = r.w - x + r.x;
    if (y + sh > r.h + r.y)
        sh = r.h - y + r.y;
    w = (sw < 0) ? 0 : sw;
    h = (sh < 0) ? 0 : sh;
}

/** Include rectangle. Modifies the rectangle so that it includes
    the specified rectangle, r. */
void
GfxRect::include(const SDL_Rect& r)
{
    if (r.w <= 0 || r.h <= 0)
        return;
    if (!exists()) {
        *this = r;
        return;
    }

    if (r.x < x) {
        w += x - r.x;
        x = r.x;
    }
    if (r.y < y) {
        h += y - r.y;
        y = r.y;
    }
    if (r.x + r.w - x > w)
        w = r.x + r.w - x;
    if (r.y + r.h - y > h)
        h = r.y + r.h - y;
}

bool
GfxRect::operator==(const SDL_Rect& rhs) const
{
    return x == rhs.x
        && y == rhs.y
        && w == rhs.w
        && h == rhs.h;
}

bool
GfxRect::operator!=(const SDL_Rect& rhs) const
{
    return x != rhs.x
        || y != rhs.y
        || w != rhs.w
        || h != rhs.h;
}

GfxPoint
GfxRect::moveTo(GfxPoint where)
{
    GfxPoint moved_by (where.x - x, where.y - y);
    moveBy(moved_by);
    return moved_by;
}

bool
GfxRect::contains(const GfxRect& r) const
{
    if (!exists())
        return false;
    if (!r.exists())
        return true;

    /* check whether r AND *this == r */
    GfxRect rcopy = r;
    rcopy.intersect(*this);
    return rcopy == r;
}

//////////////////////////////////////////////////////////////////////

/** \class GfxRectSet
    \brief Rectangle Set

    */

/** Construct empty set.
    \post contains(x,y) == false for any x,y */
GfxRectSet::GfxRectSet()
{
}

/** Construct rectangular set. The set will contain the one rectangle r. */
GfxRectSet::GfxRectSet(const GfxRect& r)
{
    the_list.push_back(r);
}

/** Intersect this rectangle set with r.
    \post r.contains(x,y)==false => contains(x,y)==false
    \todo maybe this is more efficient to do in-place? */
void
GfxRectSet::intersect(const GfxRect& r)
{
    std::list<GfxRect> l2;
    for(std::list<GfxRect>::iterator i = the_list.begin(); i != the_list.end(); ++i) {
        GfxRect q(*i);
        q.intersect(r);
        if (q.exists ())
            l2.push_back(q);
    }
    the_list.swap(l2);
}

/** Rectangle Set Subtraction. Constructs in \c l2 the set of rectangles
    in \c other which are outside \c r. Rectangles are split when needed. */
void
GfxRectSet::doSubtract(const GfxRect& r, std::list<GfxRect>& l2,
                       std::list<GfxRect>& other)
{
    for(std::list<GfxRect>::iterator i = other.begin(); i != other.end(); ++i) {
        GfxRect q(*i);
        q.intersect(r);
        if (!q.exists()) {
            l2.push_back(*i);
        } else {
            int qby = q.y + q.h, qbx = q.x + q.w;
            int iby = i->y + i->h, ibx = i->x + i->w;
            if(q.y > i->y)
                l2.push_back(GfxRect(i->x, i->y, i->w, q.y - i->y));
            if(q.x > i->x)
                l2.push_back(GfxRect(i->x, q.y, q.x - i->x, q.h));
            if(qbx < ibx)
                l2.push_back(GfxRect(qbx, q.y, ibx - qbx, q.h));
            if(qby < iby)
                l2.push_back(GfxRect(i->x, qby, i->w, iby - qby));
        }
    }
}

/** Add rectangle to set.
    \post r.contains(x,y) => contains(x,y)
    \post contains(r) */
void
GfxRectSet::add(const GfxRect& r)
{
    std::list<GfxRect> l2;
    l2.push_back(r);
    for(iterator i = the_list.begin(); i != the_list.end(); ++i) {
        std::list<GfxRect> result;
        doSubtract(*i, result, l2);
        l2.swap(result);
    }

    the_list.splice(the_list.end(), l2);
}

/** Remove rectangle from set.
    \post r.contains(x,y) => !contains(x,y) */
void
GfxRectSet::remove(const GfxRect& r)
{
    std::list<GfxRect> l2;
    doSubtract(r, l2, the_list);
    the_list.swap(l2);
}

/** Check if point is in set.
    \returns true iff (x,y) is in this set. */
bool
GfxRectSet::contains(int x, int y) const
{
    for(iterator i = begin(); i != end(); ++i)
        if(i->contains(x, y))
            return true;
    return false;
}

/** Check if rectangle in set. */
bool
GfxRectSet::contains(const GfxRect& r) const
{
    std::list<GfxRect> l2;
    l2.push_back(r);
    for(iterator i = the_list.begin(); i != the_list.end(); ++i) {
        std::list<GfxRect> result;
        doSubtract(*i, result, l2);
        l2.swap(result);
    }

    return !l2.empty();
}

GfxRect
GfxRectSet::getBoundingRectangle() const
{
    GfxRect br;
    for (iterator i = begin(); i != end(); ++i)
        br.include(*i);
    return br;
}

//////////////////////////////////////////////////////////////////////

std::ostream&
operator<<(std::ostream& os, const GfxRect& r)
{
    return os << "(" << r.x << ", " << r.y
              << ") +(" << r.w << ", " << r.h << ")";
}

std::ostream&
operator<<(std::ostream& os, GfxPoint p)
{
    return os << "(" << p.x << ", " << p.y << ")";
}
