/**
  *  \file gfx/gfxutil.h
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GFX_GFXUTIL_H
#define PCC_V2_GFX_GFXUTIL_H

#include <iosfwd>
#include <list>
#include <SDL_video.h>
#include "cpluslib/types.h"

/** Fill Pattern. An 8x8 pattern usable to fill a rectangle.
    Consists of 8 bytes (one per line, first byte is topmost),
    the most-significant bit of each byte is leftmost. */
class GfxFillPattern {
    uint8 pat[8];
 public:
    bool isBlank() const;
    bool isBlack() const;

    uint8& operator[](unsigned i) { return pat[i & 7]; }
    uint8 operator[](unsigned i) const { return pat[i & 7]; }

    GfxFillPattern();
    GfxFillPattern(const uint8* init);
    GfxFillPattern(uint8 val);

    GfxFillPattern& shiftLeft(int amount);
    GfxFillPattern& shiftRight(int amount);
    GfxFillPattern& shiftUp(int amount);
    GfxFillPattern& shiftDown(int amount);

    GfxFillPattern& operator |= (const GfxFillPattern& rhs);
    GfxFillPattern& operator |= (uint8 value);
    GfxFillPattern& operator &= (const GfxFillPattern& rhs);
    GfxFillPattern& operator &= (uint8 value);
    GfxFillPattern& operator ^= (const GfxFillPattern& rhs);
    GfxFillPattern& operator ^= (uint8 value);

    GfxFillPattern& invert();
};

struct GfxPoint {
    int x, y;

    GfxPoint() { }
    GfxPoint(int ax, int ay) : x(ax), y(ay) { }

    bool operator==(GfxPoint rhs) const { return x == rhs.x && y == rhs.y; }
    bool operator!=(GfxPoint rhs) const { return x != rhs.x || y != rhs.y; }
};

/** GfxRect. An SDL_Rect or a faster implementation.
    Represents rectangles as position plus extent. */
class GfxRect : public SDL_Rect {
 public:
    GfxRect();
    GfxRect(const SDL_Rect& r);
    GfxRect(int x, int y, int w, int h);
    explicit GfxRect(const SDL_Surface* sfc);

    void clipSurface(const SDL_Surface* sfc);
    void clipRect(const SDL_Rect& r);
    void intersect(const SDL_Rect& r) { clipRect(r); }

    void include(const SDL_Rect& r);

    bool exists() const { return w > 0 && h > 0; }

    bool operator==(const SDL_Rect& rhs) const;
    bool operator!=(const SDL_Rect& rhs) const;

    bool contains(int px, int py) const {
        px -= x;
        py -= y;
        return px >= 0 && py >= 0 && unsigned(px) < w && unsigned(py) < h;
    }
    bool contains(GfxPoint pt) const { return contains(pt.x, pt.y); }
    bool contains(const GfxRect& r) const;

    GfxPoint moveTo(GfxPoint where);
    GfxPoint moveTo(int px, int py) { return moveTo(GfxPoint(px, py)); }

    void moveBy(GfxPoint dist) { x += dist.x; y += dist.y; }
    void moveBy(int dx, int dy) { x += dx; y += dy; }
    void moveByRev(GfxPoint dist) { x -= dist.x; y -= dist.y; }
    void moveByRev(int dx, int dy) { x -= dx; y -= dy; }

    void grow(int dx, int dy) { x -= dx; y -= dy; w += 2*dx; h += 2*dy; }
};

class GfxRectSet {
    std::list<GfxRect> the_list;

    static void doSubtract(const GfxRect& r, std::list<GfxRect>& l2, std::list<GfxRect>& other);
 public:
    typedef std::list<GfxRect>::const_iterator iterator;

    GfxRectSet();
    GfxRectSet(const GfxRect& r);

    void intersect(const GfxRect& r);
    void add(const GfxRect& r);
    void remove(const GfxRect& r);
    void clear() { the_list.clear(); }

    bool contains(int x, int y) const;
    bool contains(const GfxPoint& r) const { return contains(r.x, r.y); }
    bool contains(const GfxRect& r) const;

    GfxRect getBoundingRectangle() const;

    iterator begin() const { return the_list.begin(); }
    iterator end() const { return the_list.end(); }

    bool empty() const { return the_list.empty(); }
};

extern const GfxFillPattern SOLID_FILL, GRAY50_FILL, GRAY25_FILL,
    GRAY50_FILL2, LTSLASH_FILL;

std::ostream& operator<<(std::ostream&, const GfxRect&);
std::ostream& operator<<(std::ostream&, GfxPoint);

#endif
