/**
  *  \file gfx/gfxxor.cc
  *  \brief "XOR" mode drawing.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "gfx/gfx.h"
#include "cpluslib/assert.h"

/** \class GfxXorFilter
    \brief A canvas that supports "XOR" drawing.

    Instead of putting this into the "core" code and passing around
    the "XOR" status explicitly all the time, we use a separate class. 
    This one "peeks" into the underlying canvas, and uses explicit
    draw commands to do actual drawing.

    Pros: simple. Zero overhead for non-XOR. Output clipping is natural.

    Cons: slow. XOR drawing is intended for small markers only, anyway.

    The semantics of transparency for XOR drawing is a bit weird
    currently; the XOR results are drawn in that alpha level (losing
    us the "draw again to restore original" property). blitSurface and
    blitPattern are not supported. */
    
GfxXorFilter::GfxXorFilter(Ptr<GfxCanvas> apar)
    : parent(apar)
{ }

GfxXorFilter::~GfxXorFilter()
{ }

void
GfxXorFilter::drawHLine(int x1, int y1, int x2, uint32 color, uint8 pat, alpha_type alpha)
{
    SDL_Surface* sfc = parent->getSurface();
    if (x1 < 0 || x1 >= sfc->w || y1 < 0 || y1 >= sfc->h || pat == 0)
        return;

    while (x1 < x2) {
        if (pat & (0x80 >> (x1 & 7)))
            parent->drawPixel(x1, y1, parent->getPixel(x1, y1) ^ color, alpha);
        ++x1;
    }
}

void
GfxXorFilter::drawVLine(int x1, int y1, int y2, uint32 color, uint8 pat, alpha_type alpha)
{
    SDL_Surface* sfc = parent->getSurface();
    if (x1 < 0 || x1 >= sfc->w || y1 < 0 || y1 >= sfc->h || pat == 0)
        return;
    
    while (y1 < y2) {
        if (pat & (0x80 >> (y1 & 7)))
            parent->drawPixel(x1, y1, parent->getPixel(x1, y1) ^ color, alpha);
        ++y1;
    }
}

void
GfxXorFilter::drawPixel(int x, int y, uint32 color, alpha_type alpha)
{
    parent->drawPixel(x, y, parent->getPixel(x, y) ^ color, alpha);
}

void
GfxXorFilter::drawBar(GfxRect rect, uint32 color, uint32 bg, const GfxFillPattern& pat, alpha_type alpha)
{
    if (rect.w > 0) {
        while (rect.h > 0) {
            drawHLine(rect.x, rect.y, rect.x + rect.w, color, pat[rect.y & 7], alpha);
            if (bg != GFX_TRANSPARENT)
                drawHLine(rect.x, rect.y, rect.x + rect.w, bg, ~pat[rect.y & 7], alpha);
            --rect.h;
            ++rect.y;
        }
    }
}

void
GfxXorFilter::blitSurface(int /*x*/, int /*y*/, SDL_Surface* /*src*/, GfxRect /*rect*/)
{
    FAIL();
}

void
GfxXorFilter::blitPattern(GfxRect /*rect*/, int /*x*/, int /*y*/, int /*bytes_per_line*/, const uint8* /*data*/, uint32 /*color*/, uint32 /*bg*/, alpha_type /*alpha*/)
{
    // FIXME: this one should probably be implemented
    FAIL();
}

bool
GfxXorFilter::lock()
{
    return parent->lock();
}

void
GfxXorFilter::unlock()
{
    parent->unlock();
}

void
GfxXorFilter::update(GfxRect r)
{
    parent->update(r);
}

GfxRect
GfxXorFilter::computeClipRect(GfxRect r)
{
    return parent->computeClipRect(r);
}

SDL_Surface*
GfxXorFilter::getSurface()
{
    return parent->getSurface();
}

uint32
GfxXorFilter::getPixel(int x, int y)
{
    return parent->getPixel(x, y);
} 
