/**
  *  \file gfx/init.cc
  *  \brief Common Graphics Initialisation code
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include <iostream>
#include "gfx/init.h"
#include "util/cmdline.h"
#include "arch/platform.h"
#include "util/nls.h"
#include "cpluslib/misc.h"
#include "util/console.h"
#include "fileio/dirs.h"

/** The graphics subsystem options. */
//@{
static bool opt_fullscreen = false;
static bool opt_hwsurface  = false;
static int  opt_bpp        = 0;
static int  opt_wi = 0, opt_he = 0;
static void optSetBpp(StandardCommandLineParser&, const StandardCommandLineParser::Option&);
static void optSetSize(StandardCommandLineParser&, const StandardCommandLineParser::Option&);

static StandardCommandLineParser::Option options[] = {
    { "fullscreen", &optSetValue<bool,true>,  &opt_fullscreen, false, N_(" -fullscreen        Run fullscreen") },
    { "windowed",   &optSetValue<bool,false>, &opt_fullscreen, false, N_(" -windowed          Run in a window") },
    { "bpp",        &optSetBpp,               0,               true,  N_(" -bpp=N             Use N bits per pixel") },
    { "hw",         &optSetValue<bool,true>,  &opt_hwsurface,  false, N_(" -hw                Use hardware display") },
    { "size",       &optSetSize,              0,               true,  N_(" -size=WI[xHE]      Use specified screen size") }
};
//@}

void
optSetBpp(StandardCommandLineParser& p, const StandardCommandLineParser::Option&)
{
    // FIXME: we don't support drawing on 24bpp surfaces yet */
    int v = std::atoi(p.getArg());
    if (v != 8 && v != 16 /* && v != 24 */ && v != 32) {
        p.handleError(_("Argument to `-bpp' must be 8, 16 or 32"));
    } else {
        opt_bpp = v;
    }
}

void
optSetSize(StandardCommandLineParser& p, const StandardCommandLineParser::Option&)
{
    string_t arg = p.getArg();
    string_t::size_type n = arg.find_first_of("xX*");
    if (n == string_t::npos) {
        int wi = std::atoi(arg.c_str());
        if (wi < GFX_MIN_WIDTH || wi > 10000) {
            p.handleError(_("Invalid argument to `-size'"));
        } else {
            opt_wi = wi;
            opt_he = 0;
        }
    } else {
        int wi = std::atoi(arg.substr(0, n).c_str());
        int he = std::atoi(arg.c_str() + n+1);
        if (wi < GFX_MIN_WIDTH || he < GFX_MIN_HEIGHT || wi > 10000 || he > 10000) {
            p.handleError(_("Invalid argument to `-size'"));
        } else {
            opt_wi = wi;
            opt_he = he;
        }
    }
}

static void
loadIcon(const char* name)
{
    string_t fname = Directory::makeFileName(Platform::getResourceDirectory(), name);
    SDL_Surface* icon = SDL_LoadBMP(fname.c_str());
    if (icon) {
        SDL_WM_SetIcon(icon, 0);
        SDL_FreeSurface(icon);
    }
}

/**************************** Public Interface ***************************/

/** Make command line options known to command line parser.
    Call this before parsing the command line in a graphics mode
    program. */
void
initGraphicsOptions(StandardCommandLineParser& par)
{
    par.addOptionSet(options, countof(options));
}

/** Init graphics. Does all the necessary red-tape, and sets up
    graphics mode. Returns screen surface, or null on error. */
SDL_Surface*
initGraphics(const char* icon_file, const char* window_title)
{
    /* preprocess parameters */
    int flags = SDL_HWPALETTE;
    if (opt_fullscreen)
        flags |= SDL_FULLSCREEN;
    if (opt_hwsurface)
        flags |= SDL_HWSURFACE;
    if (! opt_bpp) {
        flags |= SDL_ANYFORMAT;
        opt_bpp = 16;
    }

    /* Get resolution */
    if (! opt_wi)
        opt_wi = GFX_MIN_WIDTH;
    if (! opt_he) {
        opt_he = (opt_wi == 640 ? 480
                  : opt_wi == 800 ? 600
                  : opt_wi == 1024 ? 768
                  : opt_wi == 1200 ? 1024
                  : opt_wi == 1600 ? 1200
                  : (opt_wi * 3 / 4));
    }

    if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) != 0) {
        console << _("Error initializing SDL: ") << SDL_GetError() << std::endl;
        return 0;
    }
    std::atexit(SDL_Quit);
    Platform::init(true);

    if (icon_file)
        loadIcon(icon_file);
    if (window_title)
        SDL_WM_SetCaption(window_title, window_title);
    
    SDL_Surface* sfc = SDL_SetVideoMode(opt_wi, opt_he, opt_bpp, flags);
    if (!sfc) {
        console << _("Can't set video mode: ") << SDL_GetError() << std::endl;
        return 0;
    }
    if (sfc->format->BitsPerPixel == 3) {
        /* FIXME: do a SDL_FreeSurface(sfc); here? */
        console << _("24-bit modes are not yet supported.") << std::endl;
        return 0;
    }
    return sfc;
}
