/**
  *  \file gfx/multiclip.cc
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "gfx/multiclip.h"

/** \class GfxMultiClip
    \brief Clipping at a Rectangle Set

    This class provides clipping at a set of rectangles.
    The rectangle set can be added to and removed from. */

/** Construct blank GfxMultiClip. The filter is originally
    `opaque', i.e. doesn't let anything through. */
GfxMultiClip::GfxMultiClip(Ptr<GfxCanvas> aparent)
    : GfxFilter(aparent)
{ }

void
GfxMultiClip::drawHLine(int x1, int y1, int x2, uint32 color, uint8 pat,
                        alpha_type alpha)
{
    for (GfxRectSet::iterator i = rect.begin(); i != rect.end(); ++i) {
        GfxRect r (x1, y1, x2-x1, 1);
        r.clipRect(*i);
        if (r.exists ())
            parent->drawHLine(r.x, r.y, r.x + r.w, color, pat, alpha);
    }
}

void
GfxMultiClip::drawVLine(int x1, int y1, int y2, uint32 color, uint8 pat,
                        alpha_type alpha)
{
    for (GfxRectSet::iterator i = rect.begin(); i != rect.end(); ++i) {
        GfxRect r (x1, y1, 1, y2-y1);
        r.clipRect(*i);
        if (r.exists())
            parent->drawVLine(r.x, r.y, r.y + r.h, color, pat, alpha);
    }
}

void
GfxMultiClip::drawPixel(int x, int y, uint32 color, alpha_type alpha)
{
    if (rect.contains(x, y))
        parent->drawPixel(x, y, color, alpha);
}

void
GfxMultiClip::drawBar(GfxRect rb, uint32 color, uint32 bg,
                      const GfxFillPattern& pat, alpha_type alpha)
{
    for (GfxRectSet::iterator i = rect.begin(); i != rect.end(); ++i) {
        GfxRect r = rb;
        r.clipRect(*i);
        if (r.exists())
            parent->drawBar(r, color, bg, pat, alpha);
    }
}

void
GfxMultiClip::blitSurface(int x, int y, SDL_Surface* src, GfxRect rb)
{
    for (GfxRectSet::iterator i = rect.begin(); i != rect.end(); ++i) {
        GfxRect rclip = *i, r = rb;
        rclip.moveBy(-x, -y);
        r.clipRect(rclip);
        if (r.exists())
            parent->blitSurface(x, y, src, r);
    }
}

void
GfxMultiClip::blitPattern(GfxRect rb, int x, int y, int bytes_per_line,
                          const uint8* data, uint32 color, uint32 bg, alpha_type alpha)
{
    for (GfxRectSet::iterator i = rect.begin(); i != rect.end(); ++i) {
        GfxRect r = rb;
        r.clipRect(*i);
        if (r.exists())
            parent->blitPattern(r, x, y, bytes_per_line, data, color, bg, alpha);
    }
}

GfxRect
GfxMultiClip::computeClipRect(GfxRect r)
{
    /* clip rect = smallest rectangle c such that bar(c) and bar(r)
       have the same effect. */
    GfxRect u;
    for (GfxRectSet::iterator i = rect.begin(); i != rect.end(); ++i) {
        GfxRect pi = *i;
        pi.intersect(r);
        u.include(pi);
    }
    return u;
}

bool
GfxMultiClip::isClipped(GfxRect r)
{
    /* true iff r is only partially visible */
    /* i.e. we want to check that
              r AND rect == r
       i.e.   r - rect == 0 */
    GfxRectSet set (r);
    for (GfxRectSet::iterator i = rect.begin(); i != rect.end(); ++i)
        set.remove(*i);

    return !set.empty();
}

bool
GfxMultiClip::isVisible(GfxRect r)
{
    for (GfxRectSet::iterator i = rect.begin(); i != rect.end(); ++i) {
        GfxRect x = *i;
        x.intersect(r);
        if (x.exists())
            return true;
    }
    return false;
}

/** Add rectangle. Points within \c r will then be visible. */
void
GfxMultiClip::add(const GfxRect& r)
{
    rect.add(r);
}

/** Remove rectangle. Points within \c r will then not be accessible any
    longer. */
void
GfxMultiClip::remove(const GfxRect& r)
{
    rect.remove(r);
}

/** Clip region at rectangle. Points outside \c r will no longer be
    accessible. */
void
GfxMultiClip::clipRegionAtRectangle(const GfxRect& r)
{
    rect.intersect(r);
}
