/**
  *  \file gfx/pixmap.cc
  *  \brief Pixmap Implementation
  *
  *  GfxPixmap class, and image loading routines.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include <SDL_error.h>
#include "gfx/pixmap.h"
#include "util/except.h"
#include "fileio/io.h"
#include "gfx/palette.h"
#include "fileio/rle.h"

#ifdef HAVE_LIBSDL_IMAGE
#  include <SDL_image.h>
#endif

#include "gfx/gfx-modetr.h"
#include "util/nls.h"
#include "gfx/clip.h"
#include "cpluslib/assert.h"

/** \class GfxPixmap
    \brief Pixmap

    This represents a pixmap. A pixmap is a normal SDL_Surface, but
    with additional routines for managing it (format conversion,
    blitting it, ...). */

/** Create new Pixmap. Creates a new pixmap with the specified
    parameters. */
GfxPixmap::GfxPixmap(int w, int h,
                     int depth,
                     uint32 rmask, uint32 gmask, uint32 bmask, uint32 amask)
    : GfxSurface(SDL_CreateRGBSurface(SDL_SWSURFACE,
                                      w, h,
                                      depth,
                                      rmask, gmask, bmask, amask), true)
{
    if (!sfc)
        throw GraphicsIOException(SDL_GetError());
}

/** Create new Pixmap. Creates a new pixmap with the specified size
    and pixel format. */
GfxPixmap::GfxPixmap(int w, int h, SDL_PixelFormat* format)
    : GfxSurface(w, h, format)
{
    if (format->BitsPerPixel <= 8)
        SDL_SetPalette(sfc, SDL_LOGPAL | SDL_PHYSPAL,
                       format->palette->colors, 0, format->palette->ncolors);
}

/** Use existing Surface. This wraps an existing surface in this
    GfxPixmap object. */
GfxPixmap::GfxPixmap(SDL_Surface* sfc, bool aown)
    : GfxSurface(sfc, aown)
{ }

/** Modify this surface's format. This will destroy the current contents.
    \overload */
void
GfxPixmap::setFormat(int w, int h, SDL_PixelFormat* format)
{
    setFormat(w, h, format->BitsPerPixel,
              format->Rmask, format->Gmask,
              format->Bmask, format->Amask);
}

/** Modify this surface's format. This will destroy the current contents.
    \param depth bits per pixel
    \param rmask,gmask,bmask,amask for surfaces with more than 8 bpp,
                 the RGB masks. Must be continuous bit strings. */
void
GfxPixmap::setFormat(int w, int h, int depth, uint32 rmask, uint32 gmask,
                     uint32 bmask, uint32 amask)
{
    SDL_Surface* sfc = SDL_CreateRGBSurface(SDL_SWSURFACE,
                                            w, h, depth,
                                            rmask, gmask, bmask, amask);
    if (sfc)
        setSurface(sfc, true);
    else
        throw GraphicsIOException(SDL_GetError());
}

/** Convert to specified format.
    \returns new surface, converted to specified format; 0 on error. */
Ptr<GfxPixmap>
GfxPixmap::convertTo(SDL_PixelFormat* fmt)
{
    SDL_Surface* sfc = SDL_ConvertSurface(this->sfc, fmt, SDL_SWSURFACE);
    if (sfc)
        return new GfxPixmap(sfc, true);
    else
        return 0;
}

/** Convert to screen format.
    \returns new surface, converted to pixel format of screen; 0 on error. */
Ptr<GfxPixmap>
GfxPixmap::convertToScreenFormat()
{
    SDL_Surface* sfc = SDL_DisplayFormat(this->sfc);
    if (sfc) {
        return new GfxPixmap(sfc, true);
    } else {
        return 0;
    }
}

/** Blit Pixmap on Canvas. Displays this pixmap on the specified
    canvas, at the specified position. */
void
GfxPixmap::blit(GfxCanvas* can, int x, int y)
{
    can->blitSurface(x, y, sfc, GfxRect(sfc));
}

/** Blit Pixmap on Canvas. Displays a sub-area of this pixmap on
    the specified canvas.
    \param can  Canvas
    \param x,y  Coordinates on canvas
    \param subr Area of this pixmap to use, in pixmap coordinates.

    \note The /x/ and /y/ parameters specify the coordinates of the
    upper-left point after clipping the pixmap at /subr/. For
    GfxCanvas::blitSurface(), they specify the coordinates before
    clipping. */
void
GfxPixmap::blit(GfxCanvas* can, int x, int y, GfxRect subr)
{
    x -= subr.x;
    y -= subr.y;
    subr.clipSurface(sfc);
    can->blitSurface(x, y, sfc, subr);
}

/** Blit Pixmap on canvas. Ensures that the area x+wi/y+he is covered,
    filling with color 0. */
void
GfxPixmap::blitSized(GfxCanvas* can, int x, int y, int wi, int he)
{
    GfxRect pix_rect(sfc);

    if (pix_rect.h > he) {
        /* pixmap is bigger than our width -- reduce */
        pix_rect.y += (pix_rect.h - he) / 2;
        pix_rect.h = he;
    } else if (pix_rect.h < he) {
        /* pixmap is too small. Draw outside */
        int overlap = he - pix_rect.h, otop = overlap/2, obot = overlap - otop;
        can->drawBar(GfxRect(x, y,             wi, otop), 0, 0, SOLID_FILL, SDL_ALPHA_OPAQUE);
        can->drawBar(GfxRect(x, y + he - obot, wi, obot), 0, 0, SOLID_FILL, SDL_ALPHA_OPAQUE);
        y += otop;
        he = pix_rect.h;
    }

    if (pix_rect.w > wi) {
        /* pixmap is bigger than our width -- reduce */
        pix_rect.x += (pix_rect.w - wi) / 2;
        pix_rect.w = wi;
    } else if (pix_rect.w < wi) {
        /* pixmap is too small. Draw outside */
        /* |   |XXXX|   |
           x            x+wi */
        int overlap = wi - pix_rect.w, oleft = overlap/2, oright = overlap - oleft;
        can->drawBar(GfxRect(x,               y, oleft,  he), 0, 0, SOLID_FILL, SDL_ALPHA_OPAQUE);
        can->drawBar(GfxRect(x + wi - oright, y, oright, he), 0, 0, SOLID_FILL, SDL_ALPHA_OPAQUE);
        x += oleft;
        wi = pix_rect.w;
    }
    ASSERT(pix_rect.w == wi && pix_rect.h == he);
    can->blitSurface(x - pix_rect.x, y - pix_rect.y, sfc, pix_rect);
}

/** Flip Pixmap Vertically. Flips the pixmap data so that the pixmap
    is now upside-down. */
void
GfxPixmap::flipVertical()
{
    if (!sfc->h)
        return;

    unsigned ptr1 = 0;
    unsigned ptr2 = sfc->h - 1;
    Array<char> tmp(sfc->pitch);

    while (ptr1 < ptr2) {
        char* chr1 = static_cast<char*>(sfc->pixels) + sfc->pitch * ptr1;
        char* chr2 = static_cast<char*>(sfc->pixels) + sfc->pitch * ptr2;

        std::memcpy (tmp, chr1, sfc->pitch);
        std::memcpy (chr1, chr2, sfc->pitch);
        std::memcpy (chr2, tmp, sfc->pitch);

        ++ptr1;
        --ptr2;
    }
}

template<class T>
void
flipHorizontalLine(uint8* data, unsigned width)
{
    unsigned l = 0;
    unsigned r = width-1;
    while (l < r) {
        uint8* pl = T::add (data, l);
        uint8* pr = T::add (data, r);
        typename T::pix_type tmp = T::peek (pl);
        T::poke (pl, T::peek (pr));
        T::poke (pr, tmp);
        ++l;
        --r;
    }
}

/** Flip Pixmap Horizontally. Flips the pixmap so that left becomes
    right and vice versa. */
void
GfxPixmap::flipHorizontal()
{
    uint8* pix = static_cast<uint8*>(sfc->pixels);

    if (!sfc->w)
        return;

    for (int i = 0; i < sfc->h; ++i) {
        GFX_MODE_SWITCH(flipHorizontalLine, (pix, sfc->w));
        pix += sfc->pitch;
    }
}

/** Clone this pixmap. Returns an exact copy (maybe modulo SDL blit
    settings). */
Ptr<GfxPixmap>
GfxPixmap::clone()
{
    return convertTo(sfc->format);
}

/** Tile area with pixmap.
    \param can    Target canvas
    \param area   Area to tile
    \param alt    Alteration of X coordinate
    \see tileAnchored() */
void
GfxPixmap::tile(GfxCanvas* can, const GfxRect& area, int alt)
{
    tileAnchored(can, area, area.x, area.y, alt);
}

/** Tile area with pixmap, anchored.
    \param can    Target canvas
    \param area   Area to tile
    \param ax,ay  Anchor point
    \param alt    Alteration of X coordinate

    To have multiple tiled regions fit to each other nicely, use the same
    anchor point. Using tile() instead will anchor the tiles at the top-left
    point of the area; this way, the join points will be visible.

    The alteration specifies how the X anchor point is alterated on each
    line. With alt=0, you'll get a "chequered" pattern; alt=16 starts the
    first line at ax, the second at ax-16, the third at ax again, etc. This
    makes large tiled regions look much nicer. */
void
GfxPixmap::tileAnchored(GfxCanvas* can, const GfxRect& area, int ax, int ay, int alt)
{
    int y2 = area.y + area.h, x2 = area.x + area.w;
    int y = ay, x0 = ax;
    int mult = -1;

    while (y < y2) {
        int x = x0;
        while (x < x2) {
            GfxRect blit_rect (sfc);

            GfxRect rclip = area;
            rclip.moveBy(-x, -y);
            blit_rect.clipRect(rclip);
            if (blit_rect.exists())
                can->blitSurface(x, y, sfc, blit_rect);

            x += sfc->w;
        }
        x0 += mult * alt;
        mult = -mult;
        y += sfc->h;
    }

//     int dx = 0;
//     int wi = sfc->w, he = sfc->h;
//     while (ax < area.x - wi)
//         ax += wi;
//     while (ay < area.y - he)
//         ay += he, dx ^= alt;

//     Ptr<GfxClipRectFilter> f = new GfxClipRectFilter(area, can);
//     while (ay < area.y - he) {
//         int x = ax - dx;
//         while (x < area.x - wi) {
//             blit(f, x, ay);
//             x += wi;
//         }
//         ay += he;
//         dx ^= alt;
//     }
}

/***************************** Image Loading *****************************/

/** Convert PCC v1 color to equivalent PCC v2 color. */
static inline unsigned char
convertColor(unsigned char c)
{
    if (c >= 0x30)
        return c - 0x20;
    else
        return c & 0x0F;
}

/** Load .cc image. Those are used by PCC v1 in 16 color mode.
    \pre magic number has been read.
    \assumes stream need not be seekable.

    File format:
    - 2 bytes magic ('CC')
    - 2 bytes width
    - 2 bytes height
    - 5x3 bytes palette (slots 10..14, entry ranges 0..63)
    - pixel data */
static Ptr<GfxPixmap>
loadCCImage(Ptr<Stream> s)
{
    // FIXME: should put "user" colors elsewhere to allow reordering
    // in 8-bit mode
    char head[4 + 5*3];
    if (s->read(head, sizeof head) != sizeof head)
        return 0;

    unsigned w = getUint16(head, 0);
    unsigned h = getUint16(head, 2);
    if (w == 0 || h == 0 || w > 1000 || h > 1000)
        return 0;

    unsigned bytes_per_line = (w+1) / 2;
    Array<char> line (bytes_per_line);

    /* Image seems to be valid. Load it. */
    Ptr<GfxPixmap> pix = new GfxPixmap(w, h, 8, 0, 0, 0, 0);
    SDL_Surface* sfc = pix->getSurface();

    ASSERT(!SDL_MUSTLOCK(sfc));

    for (unsigned i = 0; i < h; ++i) {
        char* pixel_data = static_cast<char*>(sfc->pixels) + sfc->pitch * i;
        char* pixel_src = line;

        if (s->read(line, bytes_per_line) != bytes_per_line)
            throw FileTooShortException(s->getName());
        for (unsigned x = 0; x < w; ++x)
            if (x & 1)
                *pixel_data++ = convertColor((*pixel_src++ >> 4) & 0x0F);
            else
                *pixel_data++ = convertColor(*pixel_src & 0x0F);
    }

    /* set up palette */
    SDL_SetPalette(sfc, SDL_LOGPAL | SDL_PHYSPAL, standard_palette, 0, 16);
    for (int i = 0; i < 5; ++i) {
        SDL_Color c = {
            head[4 + 3*i] * 255 / 63,
            head[5 + 3*i] * 255 / 63,
            head[6 + 3*i] * 255 / 63,
            0
        };
        SDL_SetPalette(sfc, SDL_LOGPAL | SDL_PHYSPAL, &c, 10+i, 1);
    }
    return pix;
}

/** Load .cc image. Those are used by PCC v1 in 256 color mode.
    \pre magic number has been read.
    \assumes stream need not be seekable.

    File format:
    - 2 bytes magic ('CD')
    - 2 bytes width
    - 2 bytes height
    - 64x3 bytes palette (slots 192..255, ranges 0..63)
    - pixel data

    Palette will be shuffled accordingly. */
static Ptr<GfxPixmap>
loadCDImage(Ptr<Stream> s)
{
    char head[4 + 64*3];
    if (s->read(head, sizeof head) != sizeof head)
        return 0;

    unsigned w = getUint16(head, 0);
    unsigned h = getUint16(head, 2);
    if (w == 0 || h == 0 || w > 1000 || h > 1000)
        return 0;

    /* Image seems to be valid. Load it. */
    Ptr<GfxPixmap> pix = new GfxPixmap(w, h, 8, 0, 0, 0, 0);
    SDL_Surface* sfc = pix->getSurface();

    ASSERT(!SDL_MUSTLOCK(sfc));

    for (unsigned i = 0; i < h; ++i) {
        char* pixel_data = static_cast<char*>(sfc->pixels) + sfc->pitch * i;

        if (s->read(pixel_data, w) != w)
            throw FileTooShortException(s->getName());
        for (unsigned x = 0; x < w; ++x)
            pixel_data[x] = convertColor(pixel_data[x]);
    }

    /* set up palette */
    SDL_SetPalette(sfc, SDL_LOGPAL | SDL_PHYSPAL, standard_palette, 0, 160);
    for (int i = 0; i < 64; ++i) {
        SDL_Color c = {
            head[4 + 3*i] * 255 / 63,
            head[5 + 3*i] * 255 / 63,
            head[6 + 3*i] * 255 / 63,
            0
        };
        SDL_SetPalette(sfc, SDL_LOGPAL | SDL_PHYSPAL, &c, 160+i, 1);
    }
    return pix;
}

/** Load .gfx image. These are used for various things in PCC v1, and are
    the native format of the graphics library PCC v1 is based on. Unlike
    the other formats, we do not translate those; we assume they have the
    right palette.

    File format:
    - 2 bytes magic (0, 8)
    - 2 bytes width
    - 2 bytes height
    - pixel data. FF means transparent */
static Ptr<GfxPixmap>
loadGFXImage(Ptr<Stream> s)
{
    char head[4];
    if (s->read(head, 4) != 4)
        return 0;

    unsigned w = getUint16(head, 0);
    unsigned h = getUint16(head, 2);
    if (w == 0 || h == 0 || w > 1000 || h > 1000)
        return 0;

    /* Image seems to be valid. Load it. */
    Ptr<GfxPixmap> pix = new GfxPixmap(w, h, 8, 0, 0, 0, 0);
    SDL_Surface* sfc = pix->getSurface();

    pix->lock();

    bool saw_ff = false;
    for (unsigned i = 0; i < h; ++i) {
        char* pixel_data = static_cast<char*>(sfc->pixels) + sfc->pitch * i;
        if (s->read(pixel_data, w) != w)
            throw FileTooShortException(s->getName());

        if (!saw_ff)
            for (unsigned j = 0; j < w; ++j)
                if (static_cast<unsigned char>(pixel_data[j]) == 0xFFU) {
                    saw_ff = true;
                    break;
                }
    }
    pix->unlock();
    pix->update(GfxRect(sfc));

    /* set up palette */
    SDL_SetPalette(sfc, SDL_LOGPAL | SDL_PHYSPAL, standard_palette, 0, NUM_STANDARD_COLORS);
    if (saw_ff)
        SDL_SetColorKey(sfc, SDL_SRCCOLORKEY, 0xFFU);

    return pix;
}

/** Load image. Attempts to automatically figure out the type.
    If was_compr is true, s is a decompressed RLE stream; these
    may contain only .cc/.cd/.gfx images as used in PCC v1.
    If was_compr is false, stream must be seekable and can contain
    anything SDL (SDL_image) can read. */
static Ptr<GfxPixmap>
loadImageInternal(Ptr<Stream> s, bool was_compr)
{
    char magic[2];

    s->seek(0);
    if (s->read(magic, 2) == 2) {
        if (magic[0] == 'C' && magic[1] == 'C')
            return loadCCImage(s);
        else if (magic[0] == 'C' && magic[1] == 'D')
            return loadCDImage(s);
        else if (magic[0] == 0 && magic[1] == 8)
            return loadGFXImage(s);
    }

    if (!was_compr) {
#ifdef CONF_FILEIO_SDL_INTERFACE
        SDL_Surface* sfc;
        s->seek(0);
#  ifdef HAVE_LIBSDL_IMAGE
        sfc = IMG_Load_RW(s.ptr(), 0);
#  else
        sfc = SDL_LoadBMP_RW(s.ptr(), 0);
#  endif
        if (sfc)
            return new GfxPixmap(sfc, true);
#endif
    }
    return 0;
}

/** Image Loading, main entry point. Load image contained in stream s.
    For convenience, s may be 0 (returns 0 then).
    \pre stream is at position 0 and seekable. */
Ptr<GfxPixmap>
loadImage(Ptr<Stream> s)
{
    if (!s)
        return 0;
    Ptr<GfxPixmap> pix = loadImageInternal(s, false);
    if (pix.ptr())
        return pix;

    s->seek(0);
    Ptr<RleReaderStream> r = new RleReaderStream(s);
    return loadImageInternal(r, true);
}
