/**
  *  \file gfx/sprite.cc
  *  \brief Sprites
  *
  *  Sprites have the following properties:
  *  - to make use of a sprite, it must be under control of a sprite
  *    manager;
  *  - sprites are layered, that is, there's a Z order. Lower-numbered
  *    layers are at the front (always visible), higher-numbered layers
  *    are at the bottom (obscured by lower-numbered sprites).
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "gfx/sprite.h"
#include "gfx/gfxutil.h"
#include "gfx/palette.h"
#include "gfx/multiclip.h"
#include "cpluslib/assert.h"

/** Sprite Manager Data. Each sprite is associated with one object of
    this type. Its contents is private to the implementation of
    GfxSprite/GfxSpriteManager. */
struct GfxSpriteManagerData {
    /** The sprite is dirty and wants to be redrawn in-place. */
    bool dirty : 1;

    /** The sprite requests to be deleted. */
    bool delete_requested : 1;

    /** The sprite requests to be moved to the specified place. */
    bool move_requested : 1;
    GfxRect old_pos;

    GfxSpriteManagerData();
};

GfxSpriteManagerData::GfxSpriteManagerData()
    : dirty(false), delete_requested(false), move_requested(false)
{ }

/******************************* GfxSprite *******************************/

/** Create a new Sprite.
    \param apos   position and size
    \param alayer layer. Small integer starting from 0.
                  (Implementation detail: GfxSpriteManager
                  makes a vector indexed by layer. */
GfxSprite::GfxSprite(const GfxRect& apos, int alayer)
    : mgr(new GfxSpriteManagerData), manager(0),
      pos(apos), layer(alayer)
{ }

GfxSprite::~GfxSprite()
{ }

/** Associate sprite with manager. Called by GfxSpriteManager when
    this sprite is added/removed. */
void
GfxSprite::setManager(GfxSpriteManager* amanager)
{
    ASSERT(!manager || !amanager);
    manager = amanager;
}

/** Move Sprite. Causes the sprite to be redrawn later.
    \param x,y  new top-left coordinates */
void
GfxSprite::moveTo(int x, int y)
{
    if (!mgr->move_requested) {
        mgr->move_requested = true;
        mgr->old_pos        = pos;
    }
    pos.moveTo(x, y);
}

/** Move Sprite Center. Like moveTo(), but uses center coordinates. */
void
GfxSprite::moveToC(int x, int y)
{
    moveTo(x - pos.w / 2, y - pos.h / 2);
}

/** Animation tick. */
void
GfxSprite::tick()
{ }

/** Causes sprite to be redrawn next time. */
void
GfxSprite::update()
{
    mgr->dirty = true;
}

/** Causes sprite to be deleted next time. */
void
GfxSprite::deleteThis()
{
    mgr->delete_requested = true;
    mgr->dirty = true;
}

/**************************** GfxSpriteManager ***************************/

/** \class GfxSpriteManager
    A container for sprites. Manages redraw and update. */   

/** Create empty sprite manager. */
GfxSpriteManager::GfxSpriteManager()
    : bbox(0, 0, ~0U, ~0U)
{ }

GfxSpriteManager::~GfxSpriteManager()
{
    /* d'ooh! */
    for (unsigned i = 0; i < sprite_list.size(); ++i)
        for (Ptr<GfxSprite> p = sprite_list[i]; p != 0; p = p->next)
            p->prev = 0;
}

/** Add sprite to manager. Each sprite can concurrently be in only one
    sprite manager. */
void
GfxSpriteManager::addSprite(Ptr<GfxSprite> asprite)
{
    ASSERT(asprite);
    ASSERT(asprite->getLayer() >= 0);

    unsigned int layer = asprite->getLayer();
    if (layer >= sprite_list.size())
        sprite_list.resize(layer+1);

    /* add the sprite to the front of the list. */
    asprite->next = sprite_list[layer];
    asprite->prev = 0;
    if (asprite->next)
        asprite->next->prev = asprite;
    sprite_list[layer] = asprite;

    /* mark sprite to be drawn. */
    GfxSpriteManagerData* mgr = asprite->getManagerData();
    mgr->dirty = true;
    mgr->move_requested = mgr->delete_requested = false;
    asprite->setManager(this);
}

/** Move a sprite. Same as GfxSprite::moveTo(). */
void
GfxSpriteManager::moveSprite(Ptr<GfxSprite> asprite, int ax, int ay)
{
    asprite->moveTo(ax, ay);
}

/** Delete a sprite. Same as GfxSprite::deleteThis(). */
void
GfxSpriteManager::deleteSprite(Ptr<GfxSprite> asprite)
{
    asprite->deleteThis();
}

/** Delete a sprite layer. Causes all sprites on the specified layer,
    if existing, to be deleted. */
void
GfxSpriteManager::deleteLayer(int layer)
{
    if (layer >= 0 && unsigned(layer) < sprite_list.size()) {
        for (Ptr<GfxSprite> p = sprite_list[layer]; p != 0; p = p->next) {
            p->deleteThis();
        }
    }
}

/** Redraw whole sprite set. This function must not be called on the
    same canvas as tick() when there are some dirty (moved or deleted)
    sprites; otherwise, tick() may get confused about what to redraw
    and how. Calling it on a different canvas, or during "clean"
    periods should be safe though.

    Canvas must be locked. */
void
GfxSpriteManager::draw(Ptr<GfxCanvas> can)
{
    /* draw back to front */
    drawVoid(can, bbox);
    for (unsigned i = sprite_list.size(); i > 0; --i) {
        for (Ptr<GfxSprite> s = sprite_list[i-1]; s.ptr(); s = s->next) {
            s->draw(can);
        }
    }
    can->update(bbox);
}

/** Do one animation tick. Calls GfxSprite::tick() on all sprites
    (in no particular guaranteed order), and updates the screen
    accordingly. */
void
GfxSpriteManager::tick(Ptr<GfxCanvas> can)
{
    /* "tick" all sprites */
    for (unsigned i = 0; i < sprite_list.size(); ++i)
        for (Ptr<GfxSprite> s = sprite_list[i]; s.ptr(); s = s->next)
            if (! s->getManagerData()->delete_requested)
                s->tick();

    /* screen update. Now we're getting tricky. */

    /* first version. easy way: draw front to back, only dirty parts. */

    // 1. find dirty regions, moved sprites
    Ptr<GfxMultiClip> mc = new GfxMultiClip(can.ptr());
    bool deleted = false;
    for (unsigned i = 0; i < sprite_list.size(); ++i) {
        for (Ptr<GfxSprite> s = sprite_list[i]; s.ptr(); s = s->next) {
            GfxSpriteManagerData* d = s->getManagerData();
            if (d->delete_requested) {
                deleted = true;
                if (d->move_requested)
                    mc->add(d->old_pos);
                else
                    mc->add(s->getPos());
            } else if (d->move_requested && d->old_pos != s->getPos()) {
                mc->add(d->old_pos);
                mc->add(s->getPos());
                d->move_requested = false;
                d->dirty = true;
            } else if (d->dirty)
                mc->add(s->getPos());
        }
    }

    // 2. delete
    if (deleted) {
        for (unsigned i = 0; i < sprite_list.size(); ++i) {
            for (Ptr<GfxSprite> s = sprite_list[i]; s.ptr(); s = s->next) {
                GfxSpriteManagerData* d = s->getManagerData();
                if (d->delete_requested) {
                    if (s->prev)
                        s->prev->next = s->next;
                    if (s->next)
                        s->next->prev = s->prev;
                    if (!s->prev)
                        sprite_list[i] = s->next;
                    s->setManager(0);
                }
            }
        }
    }

    // 3. draw background
    mc->clipRegionAtRectangle(bbox);
    const GfxRectSet& rs = mc->getRectSet();
    can->lock();
    for (GfxRectSet::iterator i = rs.begin(); i != rs.end(); ++i)
        drawVoid(can, *i);
    can->unlock();
    
    // 4. draw sprites
    can->lock();
    for (unsigned i = sprite_list.size(); i > 0; --i) {
        for (Ptr<GfxSprite> s = sprite_list[i-1]; s.ptr(); s = s->next) {
            GfxSpriteManagerData* d = s->getManagerData();
            if (d->dirty || mc->isVisible(s->getPos())) {
                s->draw(mc);
                d->dirty = false;
            }
        }
    }
    can->unlock();

    // 5. Update
    for (GfxRectSet::iterator i = rs.begin(); i != rs.end(); ++i)
        can->update(*i);
}

/** Do one animation tick. Calls GfxSprite::tick() on all sprites
    (in no particular guaranteed order), but does not update the
    screen. */
void
GfxSpriteManager::tickSkip(bool call_tick)
{
    /* "tick" all sprites */
    if (call_tick)
        for (unsigned i = 0; i < sprite_list.size(); ++i)
            for (Ptr<GfxSprite> s = sprite_list[i]; s.ptr(); s = s->next)
                if (! s->getManagerData()->delete_requested)
                    s->tick();

    for (unsigned i = 0; i < sprite_list.size(); ++i) {
        for (Ptr<GfxSprite> s = sprite_list[i]; s.ptr(); s = s->next) {
            GfxSpriteManagerData* d = s->getManagerData();
            if (d->delete_requested) {
                if (s->prev)
                    s->prev->next = s->next;
                if (s->next)
                    s->next->prev = s->prev;
                if (!s->prev)
                    sprite_list[i] = s->next;
                s->setManager(0);
            } else {
                d->dirty = d->move_requested = false;
            }
        }
    }
}

/** Draw Background. Draw background in rectangle r. Can be overridden
    to produce a nice background if needed. Can assume the canvas be
    locked. drawVoid() need not update the appropriate regions.
    \default fills the area with black. */
void
GfxSpriteManager::drawVoid(Ptr<GfxCanvas> can, GfxRect r)
{
    can->drawBar(r, standard_colors[COLOR_BLACK], GFX_TRANSPARENT,
                 SOLID_FILL, SDL_ALPHA_OPAQUE);
}

/** Return true iff specified layer is empty. */
bool
GfxSpriteManager::isLayerEmpty(int layer) const
{
    return layer < 0 || unsigned(layer) >= sprite_list.size() || !sprite_list[layer];
}

/**************************** GfxPixmapSprite ****************************/

/** \class GfxPixmapSprite
    \brief Static Pixmap Sprite

    This sprite displays a fixed pixmap; it does not move or change in
    any way. Derived classes may override tick() to add movement. */

/** Create GfxPixmapSprite. x,y specifies the top-left corner. */
GfxPixmapSprite::GfxPixmapSprite(int x, int y, int l, Ptr<GfxPixmap> apix)
    : GfxSprite(GfxRect(x, y, apix->getWidth(), apix->getHeight()), l),
      pix(apix)
{ }

/** Create GfxPixmapSprite. x,y specifies the sprite center. The bool
    parameter is unused and only exists to make the constructor signature
    unique. */
GfxPixmapSprite::GfxPixmapSprite(int x, int y, int l, Ptr<GfxPixmap> apix, bool)
    : GfxSprite(GfxRect(x - apix->getWidth()/2, y - apix->getHeight()/2,
                        apix->getWidth(), apix->getHeight()),
                l),
      pix(apix)
{ }

void
GfxPixmapSprite::draw(Ptr<GfxCanvas> can)
{
    pix->blit(can, pos.x, pos.y);
}
