/**
  *  \file int/bytecode.cc
  *  \brief Bytecode
  *
  *  The main reason behind byte-compilation is correctness, not speed.
  *  Of course, nobody will complain when we get faster :)
  *
  *  The main problem is that certain errors will not be found by the
  *  PCC 1.x interpreter. Compilation would solve those. The other
  *  problem is that complete byte-compilation is not possible due to
  *  the simple syntax of the language: a line containing just `foo+1'
  *  could mean `compute the sum of `foo' and `1' and discard the
  *  result', or `call the subroutine `foo' with argument `+1''. We do
  *  not know whether a subroutine `foo' will be defined by the time
  *  this is executed. Hence we only compile built-in commands with
  *  special syntax (among those, the control structures that cause
  *  trouble in PCC 1.x), and handle the rest with `Eval' instructions.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <iostream>
#include <cstring>
#include "int/bytecode.h"
#include "cpluslib/assert.h"
#include "cpluslib/misc.h"

IntByteCodeObject::IntByteCodeObject()
{ }

IntByteCodeObject::~IntByteCodeObject()
{ }

void
IntByteCodeObject::add(uint32 code)
{
    opcodes.push_back(code);
}

void
IntByteCodeObject::pushInt(int32 i)
{
    uint32 n = i;               // ISO C++ guarantees 2's complement expansion
    if (i > -0x7FFFFF && i < 0x7FFFFF) {
        add(mkCode(op_PushInt, n));
    } else {
        add(mkCode(op_PushInt,   n & 0xFFFFFF));
        add(mkCode(op_PushIntHi, (n >> 24) & 255));
    }
}

void
IntByteCodeObject::pushFloat(double d)
{
    add(mkCode(op_PushFloat, allocFloat(d)));
}

void
IntByteCodeObject::pushStr(const string_t& str)
{
    add(mkCode(op_PushString, allocStr(str)));
}

IntByteCodeObject::index_t
IntByteCodeObject::allocFloat(double d)
{
    float_table.push_back(d);
    return float_table.size() - 1;
}

IntByteCodeObject::index_t
IntByteCodeObject::allocStr(const string_t& str)
{
    string_table.push_back(str);
    return string_table.size() - 1;
}

IntByteCodeObject::label_t
IntByteCodeObject::allocLabel()
{
    label_table.push_back(-1);
    return label_table.size() - 1;
}

void
IntByteCodeObject::setLabel(label_t l)
{
    ASSERT(l < label_table.size());
    ASSERT(label_table[l] == uint32(-1));
    label_table[l] = opcodes.size();
}

static const char*const insn_def[] = {
    "call1 N",
    "call2 N",
    "call3 N",

    "pushn",
    "pushb N",
    "pushi N",
    "pushf F",
    "pushs S",
    "pushih N",
    "pop N",

    "qload S",
    "qstore S",
    "uload S",
    "ustore S",

    "apply N",
    "astore N",

    "jnever L",
    "jt L",
    "jf L",
    "jne L",
    "je L",
    "jte L",
    "jfe L",
    "j L",
    "jneverp L",
    "jtp L",
    "jfp L",
    "jnep L",
    "jep L",
    "jtep L",
    "jfep L",
    "jp L",

    "eval S",
    "setline N",
};
CTASSERT(op_SetLine+1 == countof(insn_def));

void
IntByteCodeObject::dump(std::ostream& out)
{
    for (std::size_t i = 0; i < opcodes.size(); ++i) {
        for (std::size_t n = 0; n < label_table.size(); ++n)
            if (label_table[n] == i)
                out << "L" << n << ":\n";
        const char* p = insn_def[getOpcodeFromCode(opcodes[i])];
        uint32 arg = getArgFromCode(opcodes[i]);
        const char* q = strchr(p, ' ');
        if (q) {
            out << "\t" << string_t(p, q-p) << "\t";
            switch (q[1]) {
             case 'N':
                out << arg << "\n";
                break;
             case 'S':
                out << "\"" << string_table[arg] << "\" @ " << arg << "\n";
                break;
             case 'F':
                out << float_table[arg] << " @ " << arg << "\n";
                break;
             case 'L':
                out << "L" << arg << "\n";
                break;
             default:
                FAIL();
            }
        } else
            out << "\t" << p << "\n";
    }
    for (std::size_t n = 0; n < label_table.size(); ++n)
        if (label_table[n] == opcodes.size())
            out << "L" << n << ":\n";
}
