/**
  *  \file int/bytecode.h
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_BYTECODE_H
#define PCC_V2_INT_BYTECODE_H

#include <vector>
#include <cstddef>
#include <iosfwd>
#include "cpluslib/string.h"
#include "cpluslib/types.h"

enum IntOpcode {
    /* r:ARG -> (op_CallUnary, index into unary_table) -> r:result */
    op_CallUnary,
    /* r:A1:A2 -> (op_CallBinary, index into binary_table) -> r:result */
    op_CallBinary,
    /* r:A1:A2:A3 -> (op_CallTernary, index into ternary_table) -> r:result */
    op_CallTernary,

    /* r -> (op_PushNull) -> r:NULL */
    op_PushNull,
    /* r -> (op_Push{Bool,Int}, V) -> r:V */
    op_PushBool,
    op_PushInt,
    /* r -> (op_Push{Float,String}, index) -> r:V */
    op_PushFloat,
    op_PushString,
    /* r:V -> (op_PushIntHi, HI) -> r:(HI<<24 + V) */
    op_PushIntHi,
    /* r:A1:...:An -> (op_Pop, n) -> r */
    op_Pop,

    /* r:CONTEXT -> (op_QualLookup, index into string table (name)) -> r:VALUE */
    op_QualLookup,
    /* r:VALUE:CONTEXT -> (op_QualAssign, -"-) -> r */
    op_QualAssign,
    /* r -> (op_UnqualLookup, -"-) -> r:VALUE */
    op_UnqualLookup,
    /* r:VALUE -> (op_UnqualAssign, -"-) -> r */
    op_UnqualAssign,

    /* r:A1:...:An:FUN -> (op_Apply, n) -> r:VALUE */
    op_Apply,
    /* r:VALUE:A1:...:An:FUN -> (op_ApplyAssign, n) -> r */
    op_ApplyAssign,

    /* keep this order. Basic idea: after subtracting op_DontJump,
       we get a 4-bit field:
         bit 0:    jump if value is true
         bit 1:    jump if value is false
         bit 2:    jump if value is empty
         bit 3:    pop TOS after successful jump */
    /* r:VALUE -> (op_JumpXX, address) -> r:VALUE */
    /* r:VALUE -> (op_JumpXXPop, address) -> r */
    op_DontJump,
    op_JumpT,                   /* if true */
    op_JumpF,                   /* if false */
    op_JumpTF,
    op_JumpE,                   /* if empty */
    op_JumpTE,
    op_JumpFE,
    op_Jump,
    op_DontJumpPop,
    op_JumpTPop,                /* if true */
    op_JumpFPop,                /* if false */
    op_JumpTFPop,
    op_JumpEPop,                /* if empty */
    op_JumpTEPop,
    op_JumpFEPop,
    op_JumpPop,

    /* r -> (op_EvalStr, index into string table) -> r */
    op_EvalStr,
    /* r -> (op_SetLine, line nr) -> r */
    op_SetLine,

    /* Aliases to existing insns go here */
    op_JumpNE = op_JumpTF
};

inline uint32 mkCode(IntOpcode opc, uint32 arg) { return (uint32(opc) << 24) | (arg & 0xFFFFFF); }
inline uint32 getArgFromCode(uint32 x) { return x & 0xFFFFFF; }
inline IntOpcode getOpcodeFromCode(uint32 x) { return IntOpcode(x >> 24); }

class IntByteCodeObject {
    std::vector<string_t> string_table;
    std::vector<double>   float_table;
    std::vector<uint32>   opcodes;
    std::vector<uint32>   label_table;
 public:
    typedef std::size_t index_t;
    typedef uint32      label_t;
    
    IntByteCodeObject();
    ~IntByteCodeObject();

    /* code generation */
    void add(uint32 code);
    void pushInt(int32);
    void pushFloat(double d);
    void pushStr(const string_t& str);

    index_t allocFloat(double d);
    index_t allocStr(const string_t& str);

    /* labels */
    label_t allocLabel();
    void    setLabel(label_t);

    /* accessors */
    const string_t& getStr(index_t i) const { return string_table[i]; }
    const double    getFloat(index_t i) const { return float_table[i]; }

    void dump(std::ostream&);
};

#endif
