/**
  *  \file int/data.cc
  *  \brief Interpreter Data Structures
  *
  *  This defines the data model used in the interpreter.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/data.h"
#include "int/lex.h"
#include "cpluslib/format.h"

IntContext::IntContext()
{ }

IntContext::~IntContext()
{ }

IntFunction::IntFunction()
{ }

IntFunction::~IntFunction()
{ }

IntArray::IntArray()
{ }

IntArray::~IntArray()
{ }

/**
  *  \class IntObject
  *  \brief Interpreter Object
  *
  *  An object (variable value) used in the interpreter.
  *
  *  Can be either a scalar (all the PCC 1.x data types: Empty, Bool,
  *  Int, Float, String), or a non-scalar (Function, Context).
  */

/*
  Valid conversions:
      bool,int         -> float  in arithmetic when needed
      bool             -> int    likewise
      int,float,string -> bool   in bool context
 */

IntObject::IntObject(const string_t& asv)
{
    initFromString(asv);
}

IntObject::IntObject(const IntObject& rhs)
    : tag(rhs.tag), u(rhs.u)
{
    inc();
}

/** Initialize from ternary-logic value. i=0 or 1 -> boolean, other -> empty */
IntObject::IntObject(int i, bool)
{
    switch(i) {
     case 0:
     case 1:
        tag = res_Bool;
        u.int_val = i;
        break;
     default:
        tag = res_Null;
    }
}

IntObject::~IntObject()
{
    dec();
}

/** Increment reference count. */
void
IntObject::inc()
{
    switch(tag) {
     case res_Null:
     case res_Bool:
     case res_Int:
     case res_Float:
        break;
     case res_String:
        ++u.string_ptr.rc->ptr_count;
        break;
     case res_Function:
        ++u.func_ptr.rc->ptr_count;
        break;
     case res_Context:
        ++u.context_ptr.rc->ptr_count;
        break;
    }
}

/** Decrement reference count. Frees object if needed. */
void
IntObject::dec()
{
    switch(tag) {
     case res_Null:
     case res_Bool:
     case res_Int:
     case res_Float:
        break;
     case res_String:
        if (!--u.string_ptr.rc->ptr_count)
            delete u.string_ptr.ptr;
        break;
     case res_Function:
        if (!--u.func_ptr.rc->ptr_count)
            delete u.func_ptr.ptr;
        break;
     case res_Context:
        if (!--u.context_ptr.rc->ptr_count)
            delete u.context_ptr.ptr;
        break;
    }
}

IntObject&
IntObject::operator=(int32 aiv)
{
    dec();
    tag       = res_Int;
    u.int_val = aiv;
    return *this;
}

IntObject&
IntObject::operator=(bool abv)
{
    dec();
    tag       = res_Bool;
    u.int_val = abv;
    return *this;
}

IntObject&
IntObject::operator=(double adv)
{
    dec();
    tag         = res_Float;
    u.float_val = adv;
    return *this;
}

IntObject&
IntObject::operator=(string_t asv)
{
    dec();
    initFromString(asv);
    return *this;
}

IntObject&
IntObject::operator=(const IntObject& rhs)
{
    /* This can probably be optimized */
    IntObject tmp_copy (rhs);
    dec();
    tag = tmp_copy.tag;
    u = tmp_copy.u;
    tmp_copy.tag = res_Null;    // inc(); not needed
    return *this;
}

/** Assign a string. Same as operator= with string argument. */
void
IntObject::setString(const char* asv)
{
    setString(string_t(asv));
}

/** Set this object to be empty. */
void
IntObject::setEmpty()
{
    dec();
    tag = res_Null;
}

/** Extract "Context" value. */
Ptr<IntContext>
IntObject::getContextVal() const
{
    ASSERT(tag == res_Context);
    Ptr<IntContext> con;
    con.ref = u.context_ptr;
    con.inc();
    return con;
}

Ptr<IntFunction>
IntObject::getFunctionVal() const
{
    ASSERT(tag == res_Function);
    Ptr<IntFunction> fun;
    fun.ref = u.func_ptr;
    fun.inc();
    return fun;
}

/** Return boolean/tristate value. Returns 0 if False, 1 if True, -1
    if empty or not convertible to bool. */
int
IntObject::getBoolVal() const
{
    switch (tag) {
     case res_Null:
        return -1;
     case res_Bool: case res_Int:
        return u.int_val != 0;
     case res_Float:
        return (u.float_val > -IntEPSILON && u.float_val < IntEPSILON);
     default:
        /* context, function -- should not be here? */
        return -1;
    }
}

/** Make sure object is a scalar, throw error message if not. */
void
IntObject::ensureScalar(const char* msg)
{
    if (!isScalar())
        throw IntException(msg);
}

/** Do numeric promotions. Converts this object and rhs to a common type,
    widening as needed.
    \param rhs   peer in operation
    \param typ   minimum type (res_Bool/res_Int/res_Float) */
void
IntObject::doPromotions(IntObject& rhs, IntType typ)
{
    ASSERT(isNumeric() && rhs.isNumeric());
    IntType t = tag > rhs.tag ? tag : rhs.tag;
    if (typ > t)
        t = typ;
    doPromote(t);
    rhs.doPromote(t);
}

/** Promote this value to a new numeric type.
    \pre this object is a number (int/float/bool)
    \param typ type to convert to, larger or equal than this type */
void
IntObject::doPromote(IntType typ)
{
    ASSERT(isNumeric());
    ASSERT(typ >= tag);
    if (typ == tag)
        return;
    if (typ == res_Int) {
        tag = res_Int;          // source type can only be bool
    } else {
        tag = res_Float;
        u.float_val = u.int_val;
    }
}

/** Convert to string. Same as `Str' function with one argument.
    \returns true if conversion succeeded, false if not possible (value
    is not scalar). */
bool
IntObject::convertToString()
{
    switch(getType()) {
     case res_Null:
        setString("");
        return true;
     case res_Float:
        /* FIXME: this generates exponential format */
        setString(format("%.2g", getFloatVal()));
        return true;
     case res_Int:
        setString(format("%d", getIntVal()));
        return true;
     case res_String:
        return true;
     case res_Bool:
        /* FIXME? PCC 1.x uses Yes/No here */
        setString(getIntVal() ? "True" : "False");
        return true;
     default:
        return false;
    }
}

/** Initialize object from a string. Call this instead of manually
    messing with u.string_ptr, so that we have a chance to merge
    identical strings here. */
void
IntObject::initFromString(const string_t& asv)
{
    tag = res_String;
    u.string_ptr.ptr = new string_t(asv);
    u.string_ptr.rc  = ReferenceCount::MapPointer(u.string_ptr.ptr);
    ++u.string_ptr.rc->ptr_count;
}
