/**
  *  \file int/data.h
  *  \brief Interpreter Data Structures
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_INT_DATA_H
#define PCC_V2_INT_DATA_H

#include <string>
#include "cpluslib/misc.h"
#include "cpluslib/string.h"
#include "cpluslib/smartptr.h"
#include "cpluslib/types.h"
#include "cpluslib/assert.h"

/** Type of a value. The actual names are, umm, historical.
    Preserve the order if at all possible (see IntObject::isScalar, doPromote). */
enum IntType {
    res_Null,                   ///< "empty".
    res_Bool,                   ///< boolean.
    res_Int,                    ///< integer.
    res_Float,                  ///< float.
    res_String,                 ///< string.

    res_Function,               ///< function. Something that can be called with args.
    res_Context                 ///< context. Something that can be dereferenced using `.'.
};

class IntObject;

/** Context. */
class IntContext : public Object {
 public:
    typedef int32 handle_t;
    Ptr<IntContext> parent;

    IntContext();
    ~IntContext();

    virtual bool next() = 0;

    virtual handle_t  lookUp(string_t name) = 0;
    virtual IntObject read(handle_t) = 0;
    virtual void      store(handle_t, IntObject) = 0;
};

class IntArray;

/** Function or Array. Functions can be applied to a number of
    arguments. Arrays are special functions which have the following
    additional properties:
    - arguments are integers; arrays know the required number of arguments
      and their range.
    - arrays can be referenced as whole (`ForEach Ship'). */
class IntFunction : public Object {
 public:
    IntFunction();
    ~IntFunction();

    virtual IntArray* getArray() const = 0;
    virtual IntObject apply(size_t nargs, IntObject args[]) = 0;
    virtual void      assign(size_t nargs, IntObject args[], IntObject value) = 0;
};

class IntArray : public IntFunction {
 public:
    IntArray();
    ~IntArray();
    virtual int getMinIndex() = 0;
    virtual int getMaxIndex() = 0;
    virtual Ptr<IntContext> getFirst() = 0;
};

const double IntEPSILON = 1.0e-6;

class IntObject {
    /** Type. */
    IntType tag;

    /** Value. All members are PODs, the reference counts are managed
        in ctors/dtors and operators. */
    union {
        int32                 int_val;      ///< res_Int, res_Bool.
        Ptr<string_t>::Ref    string_ptr;   ///< res_String.
        double                float_val;    ///< res_Float.
        Ptr<IntContext>::Ref  context_ptr;  ///< res_Context.
        Ptr<IntFunction>::Ref func_ptr;     ///< res_Function.
    } u;
    void dec();
    void inc();
    void initFromString(const string_t& asv);
 public:
    IntObject()
        : tag(res_Null) { }
    IntObject(int32 aiv)
        : tag(res_Int) { u.int_val = aiv; }
    IntObject(double afv)
        : tag(res_Float) { u.float_val = afv; }
    IntObject(bool abv)
        : tag(res_Bool) { u.int_val = abv; }
    IntObject(const string_t& asv);
    IntObject(const IntObject& arhs);
    IntObject(int i, bool);
    ~IntObject();

    bool isEmpty() const  { return tag == res_Null; }
    bool isBool() const   { return tag == res_Bool; }
    bool isInt() const    { return tag == res_Int; }
    bool isString() const { return tag == res_String; }
    bool isFloat() const  { return tag == res_Float; }
    bool isScalar() const { return tag >= res_Null && tag <= res_String; }
    bool isNumeric() const { return tag >= res_Bool && tag <= res_Float; }
    bool isContext() const { return tag == res_Context; }
    bool isFunction() const { return tag == res_Function; }

    IntType getType() const { return tag; }

    IntObject& operator=(const IntObject& rhs);
    IntObject& operator=(int32 aiv);
    IntObject& operator=(bool abv);
    IntObject& operator=(double adv);
    IntObject& operator=(string_t asv);
    void setEmpty();
    void setString(const string_t& asv) { *this = asv; }
    void setString(const char* asv);

    int32 getIntVal() const {
        ASSERT(tag == res_Bool || tag == res_Int);
        return u.int_val;
    }
    const string_t& getStringVal() const {
        ASSERT(tag == res_String);
        return *u.string_ptr.ptr;
    }
    double getFloatVal() const {
        ASSERT(tag == res_Float);
        return u.float_val;
    }
    Ptr<IntContext> getContextVal() const;
    IntContext&     getContextRef() const {
        ASSERT(tag == res_Context);
        return *u.context_ptr.ptr;
    }
    Ptr<IntFunction> getFunctionVal() const;
    IntFunction&     getFunctionRef() const{
        ASSERT(tag == res_Function);
        return *u.func_ptr.ptr;
    }

    int getBoolVal() const;
    void ensureScalar(const char* msg);
    void doPromotions(IntObject& rhs, IntType typ = res_Null);
    void doPromote(IntType typ);
    bool convertToString();
};

#endif
