/**
  *  \file int/expr.cc
  *  \brief Interpreter Core (Expressions)
  *
  *  This contains the interpreter core as far as expression
  *  evaluation is concerned. In particular, it implements expression
  *  parsing (converting a string into a parse tree) and the built-in
  *  operators.
  *
  *  Handling errors: in PCC I, it was easy because every error caused
  *  a signal on which we could throw an exception (via longjmp). For
  *  example, computing the logarithm just meant calling `log'. This
  *  is not, was never and will never be portable. Here, we check for
  *  possible errors beforehand and generate error messages manually.
  *  This is how perl does it, so it *must* work :)
  *
  *  \todo Integer arithmetic (in particular, DIV and MOD of negative
  *  numbers) works according to the native system's rules. This
  *  should be changed to use the x86 rules everywhere.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#define _BSD_SOURCE   /* finite, isinf, etc. if present */
#include <cmath>
#include "int/expr.h"
#include "util/nls.h"
#include "defs.h"
#include "cpluslib/buffer.h"
#include "int/function.h"
#include "int/bytecode.h"

static bool case_insensitive = true;

double
IntExprNode::checkFloatVal(double d)
{
    using namespace std;
#if defined(HAVE_ISFINITE)
    if (!isfinite(d))
#elif defined(HAVE_FINITE)
    if (!finite(d))
#else
    if (0
#  ifdef HAVE_ISNAN
        || isnan(d)
#  endif
#  ifdef HAVE_ISINF
        || isinf(d)
#  endif
        )
#endif
        throw IntException("Math error");
    return d;
}

static bool
checkArithmetic(IntObject& a, IntObject& b)
{
    if (a.isEmpty() || b.isEmpty()) {
        return false;
    } else if (a.isNumeric() && b.isNumeric()) {
        a.doPromotions(b, res_Int);
        return true;
    } else if (a.getType() == b.getType()) {
        return true;
    } else {
        throw IntException("Types don't match");
    }
}

/******************************* Functions *******************************/

/** "Sequence" operator (`a; b'). Caller guarantees left-to-right
    evaluation. */
IntObject
FSequence(IntObject /*a*/, IntObject b)
{
    return b;
}

/** Operator `Not': logical negation */
IntObject
FNot(IntObject a)
{
    a.ensureScalar("Operand to `Not' must be scalar");
    return IntObject(a.getBoolVal() ^ 1, true);
}

/** This implements `Not Not x' aka `cast to boolean' */
IntObject
FBool(IntObject a)
{
    a.ensureScalar("Operand to `Not' must be scalar");
    return IntObject(a.getBoolVal(), true);
}

/** Comparison operators.
    \pre checkArithmetic(a,b) has been called and returned true
    \returns comparison result, in the style of strcmp() */
int
doCompare(IntObject& a, IntObject& b)
{
    switch (a.getType()) {
     case res_Int:
        if (a.getIntVal() == b.getIntVal())
            return 0;
        else if (a.getIntVal() < b.getIntVal())
            return -1;
        else
            return 1;
     case res_Float:
        if (a.getFloatVal() == b.getFloatVal())
            return 0;
        else if (a.getFloatVal() < b.getFloatVal())
            return -1;
        else
            return 1;
     case res_String:
        if (case_insensitive)
            return strUCase(a.getStringVal()).compare(strUCase(b.getStringVal()));
        else
            return a.getStringVal().compare(b.getStringVal());
     default:
        /* because checkArithmetic returned true, we needn't bother
           with res_Null */
        throw IntException("Types can't be compared");
    }
}

/** "Less than" operator. */
IntObject
FLT(IntObject a, IntObject b)
{
    if (checkArithmetic(a, b))
        return doCompare(a, b) < 0;
    else
        return IntObject();
}

/** "Greater than" operator. */
IntObject
FGT(IntObject a, IntObject b)
{
    if (checkArithmetic(a, b))
        return doCompare(a, b) > 0;
    else
        return IntObject();
}

/** "Less or Equal" operator. */
IntObject
FLE(IntObject a, IntObject b)
{
    if (checkArithmetic(a, b))
        return doCompare(a, b) <= 0;
    else
        return IntObject();
}

/** "Greater or Equal" operator. */
IntObject
FGE(IntObject a, IntObject b)
{
    if (checkArithmetic(a, b))
        return doCompare(a, b) >= 0;
    else
        return IntObject();
}

/** "Equal" operator. */
IntObject
FEQ(IntObject a, IntObject b)
{
    if (checkArithmetic(a, b))
        return doCompare(a, b) == 0;
    else
        return IntObject();
}

/** "Not Equal" operator. */
IntObject
FNE(IntObject a, IntObject b)
{
    if (checkArithmetic(a, b))
        return doCompare(a, b) != 0;
    else
        return IntObject();
}

/** Arithmetic addition. Handles number+number and string+string. */
IntObject
FPlus(IntObject a, IntObject b)
{
    if (!checkArithmetic(a, b))
        return IntObject();

    switch(a.getType()) {
     case res_Int:
        return a.getIntVal() + b.getIntVal();
     case res_Float:
        RETURN_FLOAT_VAL(a.getFloatVal() + b.getFloatVal());
     case res_String:
        return IntObject(a.getStringVal() + b.getStringVal());
     default:
        throw IntException("Invalid types for operator `+'");
    }
}

/** Arithmetic minus. */
IntObject
FMinus(IntObject a, IntObject b)
{
    if (!checkArithmetic(a, b))
        return IntObject();

    switch(a.getType()) {
     case res_Int:
        return a.getIntVal() - b.getIntVal();
     case res_Float:
        RETURN_FLOAT_VAL(a.getFloatVal() - b.getFloatVal());
     default:
        throw IntException("Invalid types for operator `-'");
    }
}

/** Multiplication. */
IntObject
FMul(IntObject a, IntObject b)
{
    if (!checkArithmetic(a, b))
        return IntObject();

    switch(a.getType()) {
     case res_Int:
        return a.getIntVal() * b.getIntVal();
     case res_Float:
        RETURN_FLOAT_VAL(a.getFloatVal() * b.getFloatVal());
     default:
        throw IntException("Invalid types for operator `*'");
    }
}

/** Integer division (operator "\"). */
IntObject
FDiv(IntObject a, IntObject b)
{
    if (a.isEmpty() || b.isEmpty()) {
        return IntObject();
    } else if (a.isNumeric() && b.isNumeric() && (a.doPromotions(b), a.isInt())) {
        if (b.getIntVal() == 0)
            throw IntException("Divide by zero");
        else
            return a.getIntVal() / b.getIntVal();
    } else
        throw IntException("`\\' requires integer arguments");
}

/** Remainder. */
IntObject
FMod(IntObject a, IntObject b)
{
    if (a.isEmpty() || b.isEmpty()) {
        return IntObject();
    } else if (a.isNumeric() && b.isNumeric() && (a.doPromotions(b), a.isInt())) {
        if (b.getIntVal() == 0) {
            throw IntException("Divide by zero");
        } else {
            /* FIXME: implementation-defined behaviour. */
            return a.getIntVal() % b.getIntVal();
        }
    } else
        throw IntException("`Mod' requires integer arguments");
}

/** Real (fp) division. */
IntObject
FRealDiv(IntObject a, IntObject b)
{
    if (a.isEmpty() || b.isEmpty()) {
        return IntObject();
    } else if (a.isNumeric() && b.isNumeric()) {
        a.doPromotions(b, res_Float);
        if (b.getFloatVal() == 0.0)
            throw IntException("Divide by zero");
        else
            RETURN_FLOAT_VAL(a.getFloatVal() / b.getFloatVal());
    } else
        throw IntException("`/' requires numeric arguments");
}

/** Unary minus. */
IntObject
FNegate(IntObject a)
{
    switch (a.getType()) {
     case res_Null:
        return IntObject();
     case res_Int:
     case res_Bool:
        return -a.getIntVal();
     case res_Float:
        return -a.getFloatVal();
     default:
        throw IntException("Unary `-' needs a numeric argument");
    }
}

/** String concatenation. Empty values show up as empty strings. */
IntObject
FConcat(IntObject a, IntObject b)
{
    if (a.convertToString() && b.convertToString())
        return IntObject(a.getStringVal() + b.getStringVal());
    else
        throw IntException("Invalid types for operator `&'");
}

/** String concatenation. Empty values cause result to be empty. */
IntObject
FHash(IntObject a, IntObject b)
{
    if (a.isEmpty() || b.isEmpty())
        return IntObject();
    else if (a.convertToString() && b.convertToString())
        return IntObject(a.getStringVal() + b.getStringVal());
    else
        throw IntException("Invalid types for operator `#'");
}

/** Exponentiation. */
IntObject
FPow(IntObject a, IntObject b)
{
    if (a.isEmpty() || b.isEmpty()) {
        return IntObject();
    } else if (a.isNumeric() && b.isNumeric()) {
        a.doPromotions(b, res_Float);
        RETURN_FLOAT_VAL(std::pow(a.getFloatVal(), b.getFloatVal()));
    } else
        throw IntException("Invalid types for operator `^'");
}

/******************************** Classes ********************************/

IntExprNode::IntExprNode()
{ }

IntExprNode::~IntExprNode()
{ }

void
IntExprLValue::compileA(IntByteCodeObject&)
{
    throw IntException("Assignment to lvalue attempted");
}

IntExprFuncall1::IntExprFuncall1(IntUnaryFunction f, IntExprNodePtr aa)
    : func(f), a(aa)
{ }

IntExprFuncall1::~IntExprFuncall1()
{
//     delete a;
}

IntObject
IntExprFuncall1::eval(IntExecutionContext& con)
{
    return unary_table[func](a->eval(con));
}

void
IntExprFuncall1::compileV(IntByteCodeObject& bco)
{
    a->compileV(bco);
    bco.add(mkCode(op_CallUnary, func));
}

void
IntExprFuncall1::compileS(IntByteCodeObject& bco)
{
    a->compileS(bco);
}

IntExprFuncall2::IntExprFuncall2(IntBinaryFunction f,
                                 IntExprNodePtr aa, IntExprNodePtr bb)
    : func(f), a(aa), b(bb)
{ }

IntExprFuncall2::~IntExprFuncall2()
{
//     delete a;
//     delete b;
}

IntObject
IntExprFuncall2::eval(IntExecutionContext& con)
{
    IntObject va = a->eval(con);
    IntObject vb = b->eval(con);
    return binary_table[func](va, vb);
}

void
IntExprFuncall2::compileV(IntByteCodeObject& bco)
{
    a->compileV(bco);
    b->compileV(bco);
    bco.add(mkCode(op_CallBinary, func));
}

void
IntExprFuncall2::compileS(IntByteCodeObject& bco)
{
    a->compileS(bco);
    b->compileS(bco);
}

/***************************** IntExprLiteral ****************************/

IntExprLiteral::IntExprLiteral(const IntObject& v)
    : value(v)
{ }

IntExprLiteral::~IntExprLiteral()
{ }

IntObject
IntExprLiteral::eval(IntExecutionContext&)
{
    return value;
}

void
IntExprLiteral::compileV(IntByteCodeObject& bco)
{
    switch(value.getType()) {
     case res_Null:
        bco.add(mkCode(op_PushNull, 0));
        break;
     case res_Bool:
        bco.add(mkCode(op_PushBool, value.getIntVal()));
        break;
     case res_Int:
        bco.pushInt(value.getIntVal());
        break;
     case res_Float:
        bco.pushFloat(value.getFloatVal());
        break;
     case res_String:
        bco.pushStr(value.getStringVal());
        break;
     default:
        FAIL();
    }
}

void
IntExprLiteral::compileS(IntByteCodeObject&)
{ }

/*************************** Special Functions ***************************/

namespace {
    /** "And" operator. Short circuit evaluation. */
    class IntAndNode : public IntExprLValue {
        IntExprNodePtr a, b;
     public:
        IntAndNode(IntExprNodePtr aa, IntExprNodePtr bb)
            : a(aa), b(bb) { }
        IntObject eval(IntExecutionContext& con);
        void compileV(IntByteCodeObject&);
        void compileS(IntByteCodeObject&);
    };

    /** "Or" operator. Short circuit evaluation. */
    class IntOrNode : public IntExprLValue {
        IntExprNodePtr a, b;
     public:
        IntOrNode(IntExprNodePtr aa, IntExprNodePtr bb)
            : a(aa), b(bb) { }
        IntObject eval(IntExecutionContext& con);
        void compileV(IntByteCodeObject&);
        void compileS(IntByteCodeObject&);
    };

    /** "Xor" operator. "Short circuit" evaluation. */
    class IntXorNode : public IntExprLValue {
        IntExprNodePtr a, b;
     public:
        IntXorNode(IntExprNodePtr aa, IntExprNodePtr bb)
            : a(aa), b(bb) { }
        IntObject eval(IntExecutionContext& con);
        void compileV(IntByteCodeObject&);
        void compileS(IntByteCodeObject&);
    };

    /** Qualified name lookup (foo(bar).name). */
    class IntQualLookup : public IntExprNode {
        IntExprNodePtr a;
        string_t       name;
     public:
        IntQualLookup(IntExprNodePtr aa, const string_t& aname)
            : a(aa), name(aname) { }
        IntObject eval(IntExecutionContext& con);
        void compileV(IntByteCodeObject&);
        void compileS(IntByteCodeObject&);
        void compileA(IntByteCodeObject&);
    };

    /** Unqualified name lookup (lone identifier in code). */
    class IntUnqualLookup : public IntExprNode {
        string_t name;
     public:
        IntUnqualLookup(const string_t& aname)
            : name(aname) { }
        IntObject eval(IntExecutionContext& con);
        void compileV(IntByteCodeObject&);
        void compileS(IntByteCodeObject&);
        void compileA(IntByteCodeObject&);
    };

    /** Apply arguments to function (foo(x,y,z)). */
    class IntApply : public IntExprNode {
        IntExprNodePtr        a;
        std::size_t           nargs;
        Array<IntExprNodePtr> args;
     public:
        IntApply(IntExprNodePtr aa, std::size_t aargs, IntExprParser& apar);
        IntObject eval(IntExecutionContext& con);
        void compileV(IntByteCodeObject&);
        void compileS(IntByteCodeObject&);
        void compileA(IntByteCodeObject&);
    };
};

IntObject
IntAndNode::eval(IntExecutionContext& con)
{
    /* a -> _f_t_e_
          f| f f f
          t| f t e
          e| f e e */
    IntObject r = a->eval(con);
    r.ensureScalar("Operand to `And' must be scalar");
    int rl = r.getBoolVal();
    if (rl == 0) {
        return r;
    } else {
        r = b->eval(con);
        r.ensureScalar("Operand to `And' must be scalar");
        if (int rr = r.getBoolVal())
            return IntObject(rl | rr, true);
        else
            return IntObject(false);
    }
}

void
IntAndNode::compileV(IntByteCodeObject& /*bco*/)
{
    FAIL();
}

void
IntAndNode::compileS(IntByteCodeObject& bco)
{
    IntByteCodeObject::label_t label = bco.allocLabel();
    a->compileV(bco);
    bco.add(mkCode(op_JumpFPop, label));
    b->compileS(bco);
    bco.setLabel(label);
}

IntObject
IntOrNode::eval(IntExecutionContext& con)
{
    /* a -> _f_t_e_
          f| f t e
          t| t t t
          e| e t e */
    IntObject r = a->eval(con);
    r.ensureScalar("Operand to `Or' must be scalar");
    int rl = r.getBoolVal();
    if (rl == 1) {
        return IntObject(true);
    } else {
        r = b->eval(con);
        r.ensureScalar("Operand to `Or' must be scalar");
        int rr = r.getBoolVal();
        if (rr == 1)
            return IntObject(true);
        else
            return IntObject(rl | rr, true);
    }
}

void
IntOrNode::compileV(IntByteCodeObject& /*bco*/)
{
    FAIL();
}

void
IntOrNode::compileS(IntByteCodeObject& bco)
{
    IntByteCodeObject::label_t label = bco.allocLabel();
    a->compileV(bco);
    bco.add(mkCode(op_JumpTPop, label));
    b->compileS(bco);
    bco.setLabel(label);
}

IntObject
IntXorNode::eval(IntExecutionContext& con)
{
    /* a -> _f_t_e_
          f| f t e
          t| t f e
          e| e e e */
    IntObject r = a->eval(con);
    r.ensureScalar("Operand to `Xor' must be scalar");
    if (r.isEmpty()) {
        return IntObject();
    } else {
        int rl = r.getBoolVal();
        r = b->eval(con);
        r.ensureScalar("Operand to `Xor' must be scalar");
        if (r.isEmpty())
            return IntObject();
        else
            return IntObject(rl | r.getBoolVal(), true);
    }
}

void
IntXorNode::compileV(IntByteCodeObject& /*bco*/)
{
    FAIL();
}

void
IntXorNode::compileS(IntByteCodeObject& bco)
{
    IntByteCodeObject::label_t label = bco.allocLabel();
    a->compileV(bco);
    bco.add(mkCode(op_JumpEPop, label));
    b->compileS(bco);
    bco.setLabel(label);
}

/* `a.name' */
IntObject
IntQualLookup::eval(IntExecutionContext& con)
{
    IntObject c = a->eval(con);
    if (c.getType() != res_Context)
        throw IntException("Left operand of `.' is not a context");

    IntContext::handle_t handle = c.getContextRef().lookUp(name);
    if (handle < 0)
        throw IntException("Unknown field `" + name + "'");

    return c.getContextRef().read(handle);
}

void
IntQualLookup::compileV(IntByteCodeObject& bco)
{
    a->compileV(bco);
    bco.add(mkCode(op_QualLookup, bco.allocStr(name)));
}

void
IntQualLookup::compileS(IntByteCodeObject& bco)
{
    compileV(bco);
    bco.add(mkCode(op_Pop, 1));
}

void
IntQualLookup::compileA(IntByteCodeObject& bco)
{
    a->compileV(bco);
    bco.add(mkCode(op_QualAssign, bco.allocStr(name)));
}

IntObject
IntUnqualLookup::eval(IntExecutionContext& con)
{
    Ptr<IntContext> c = con.context;
    while (c) {
        IntContext::handle_t handle = c->lookUp(name);
        if (handle >= 0)
            return c->read(handle);
        c = c->parent;
    }
    throw IntException("Unknown identifier `" + name + "'");
}

void
IntUnqualLookup::compileV(IntByteCodeObject& bco)
{
    bco.add(mkCode(op_UnqualLookup, bco.allocStr(name)));
}

void
IntUnqualLookup::compileS(IntByteCodeObject& bco)
{
    compileV(bco);
    bco.add(mkCode(op_Pop, 1));
}

void
IntUnqualLookup::compileA(IntByteCodeObject& bco)
{
    bco.add(mkCode(op_UnqualAssign, bco.allocStr(name)));
}

IntApply::IntApply(IntExprNodePtr aa, std::size_t aargs, IntExprParser& apar)
    : a(aa), nargs(aargs), args(aargs)
{
    for (std::size_t i = aargs; i > 0; --i)
        args[i-1] = apar.pop();
}

IntObject
IntApply::eval(IntExecutionContext& con)
{
    IntObject fun = a->eval(con);
    if (!fun.isFunction())
        throw IntException("Object to be called is not a function");

    Array<IntObject> values(nargs);
    for (std::size_t i = 0; i < nargs; ++i)
        values[i] = args[i]->eval(con);

    return fun.getFunctionRef().apply(nargs, values.ptr());
}

void
IntApply::compileV(IntByteCodeObject& bco)
{
    for (std::size_t i = 0; i < nargs; ++i)
        args[i]->compileV(bco);
    a->compileV(bco);
    bco.add(mkCode(op_Apply, nargs));
}

void
IntApply::compileS(IntByteCodeObject& bco)
{
    compileV(bco);
    bco.add(mkCode(op_Pop, 1));
}

void
IntApply::compileA(IntByteCodeObject& bco)
{
    for (std::size_t i = 0; i < nargs; ++i)
        args[i]->compileV(bco);
    a->compileV(bco);
    bco.add(mkCode(op_ApplyAssign, nargs));
}

/** File handle node. When evaluated, these signal an error. Functions
    that take a file handle manually dissect it. */
IntExprFileHandle::IntExprFileHandle(IntExprNodePtr aa)
    : a(aa)
{ }

IntExprFileHandle::~IntExprFileHandle()
{ }

IntObject
IntExprFileHandle::eval(IntExecutionContext&)
{
    throw IntException("File handle not allowed here");
}

void
IntExprFileHandle::compileV(IntByteCodeObject&)
{
    throw IntException("File handle not allowed here");
}

void
IntExprFileHandle::compileS(IntByteCodeObject&)
{
    throw IntException("File handle not allowed here");
}

/***************************** IntExprParser *****************************/

/** \class IntExprParser
    \brief Expression Parser

    This parses an expression from a token stream (aka IntLexer).
    It builds up the parse tree step by step using an intermediate
    stack. The stack is in the IntExprParser object, not on the runtime
    stack. This way, only the IntExprParser has to deal with stack
    unwinding and stuff. */
IntExprParser::IntExprParser(IntLexer& l)
    : lex(l)
{ }

IntExprParser::~IntExprParser()
{
//     while (IntExprNode* n = pop())
//         delete n;
}

void
IntExprParser::push(IntExprNodePtr n)
{
    ASSERT(n);
    stack.push_front(n);
}

/** Generate call to unary function. Pops arg, pushes tree for f1(arg). */
void
IntExprParser::pushFuncall(IntUnaryFunction f1)
{
    ASSERT(stack.size() >= 1);
    IntExprNodePtr a = pop();
    push(new IntExprFuncall1(f1, a));
}

/** Generate call to binary function. Pops arg1,arg2 in correct order,
    pushes tree for f2(arg1,arg2). */
void
IntExprParser::pushFuncall(IntBinaryFunction f2)
{
    ASSERT(stack.size() >= 2);
    IntExprNodePtr b = pop();
    IntExprNodePtr a = pop();
    push(new IntExprFuncall2(f2, a, b));
}

/** Remove top of stack, and return it. */
IntExprNodePtr
IntExprParser::pop()
{
    ASSERT(stack.size() > 0);
    IntExprNodePtr n = stack.front();
    stack.pop_front();
    return n;
}

/* expr0 ::= expr0b
           | expr0 ';' expr0b */
void
IntExprParser::parse0()
{
    parse0b();
    while (lex.getType() == tt_Semi) {
        if (lex.lex() == tt_Null)
            throw IntException("Lone `;' at end of line is not allowed");
        parse0b();
        pushFuncall(FX_Sequence);
    }
}

/* expr0b ::= '#'? expr0a
            | '#'? expr0a ':=' expr0b */
void
IntExprParser::parse0b()
{
    bool hash = false;
    if (lex.lookingAt(tt_Hash))
        hash = true;

    parse0a();
    if (lex.lookingAt(tt_Asgn)) {
        parse0b();
        // FIXME FIXME FIXME: Assigment
    }
    if (hash)
        push(new IntExprFileHandle(pop()));
}

/* "additive logic operators"
   expr0a ::= expr1
            | expr0a 'XOR' expr1
            | expr0a 'OR' expr1 */
void
IntExprParser::parse0a()
{
    parse1();
    while (1) {
        if (lex.lookingAt(tt_XOR)) {
            parse1();
            IntExprNodePtr b = pop(), a = pop();
            push(new IntXorNode(a, b));
        } else if (lex.lookingAt(tt_OR)) {
            parse1();
            IntExprNodePtr b = pop(), a = pop();
            push(new IntOrNode(a, b));
        } else
            break;
    }
}

/* "multiplicative logic operators"
   expr1 ::= expr2
           | expr1 'AND' expr2 */
void
IntExprParser::parse1()
{
    parse2();
    while (lex.lookingAt(tt_AND)) {
        parse2();
        IntExprNodePtr b = pop(), a = pop();
        push(new IntAndNode(a, b));
    }
}

/* "unary logic operators"
   expr2 ::= expr3
           | 'NOT' expr2 */
void
IntExprParser::parse2()
{
    unsigned n = 0;
    while (lex.lookingAt(tt_NOT))
        ++n;
    parse3();
    if (n) {
        if (n & 1)
            pushFuncall(FX_Not);
        else
            pushFuncall(FX_Bool);
    }
}

/* "relational operators"
   expr3 ::= expr4
           | expr3 relop expr4
   relop ::= '<' | '>' | '<=' | '>=' | '<>' | '=' */
void
IntExprParser::parse3()
{
    parse4();
    while (1) {
        if (lex.lookingAt(tt_LT)) {
            parse4();
            pushFuncall(FX_LT);
        } else if (lex.lookingAt(tt_GT)) {
            parse4();
            pushFuncall(FX_GT);
        } else if (lex.lookingAt(tt_LE)) {
            parse4();
            pushFuncall(FX_LE);
        } else if (lex.lookingAt(tt_GE)) {
            parse4();
            pushFuncall(FX_GE);
        } else if (lex.lookingAt(tt_NE)) {
            parse4();
            pushFuncall(FX_NE);
        } else if (lex.lookingAt(tt_EQ)) {
            parse4();
            pushFuncall(FX_EQ);
        } else
            break;
    }
}

/* "concatenation operators"
   expr4 ::= expr5
           | expr4 '&' expr5
           | expr4 '#' expr5 */
void
IntExprParser::parse4()
{
    parse5();
    while (1) {
        if (lex.lookingAt(tt_Hash)) {
            parse5();
            pushFuncall(FX_Hash);
        } else if (lex.lookingAt(tt_Join)) {
            parse5();
            pushFuncall(FX_Concat);
        } else
            break;
    }
}

/* "additive operators"
   expr5 ::= expr6
           | expr5 '+' expr6
           | expr5 '-' expr6 */
void
IntExprParser::parse5()
{
    parse6();
    while (1) {
        if (lex.lookingAt(tt_Plus)) {
            parse6();
            pushFuncall(FX_Plus);
        } else if (lex.lookingAt(tt_Minus)) {
            parse6();
            pushFuncall(FX_Minus);
        } else
            break;
    }
}

/* "multiplicative operators"
   expr6 ::= expr7
           | expr6 mulop expr7
   mulop ::= '*' | '/' | '\' | 'mod' */
void
IntExprParser::parse6()
{
    parse7();
    while (1) {
        if (lex.lookingAt(tt_Mul)) {
            parse7();
            pushFuncall(FX_Mul);
        } else if (lex.lookingAt(tt_Div)) {
            parse7();
            pushFuncall(FX_Div);
        } else if (lex.lookingAt(tt_RealDiv)) {
            parse7();
            pushFuncall(FX_RealDiv);
        } else if (lex.lookingAt(tt_MOD)) {
            parse7();
            pushFuncall(FX_Mod);
        } else
            break;
    }
}

/* "unary operators"
   expr7 ::= expr8
           | '-' expr7
           | '+' expr7 */
void
IntExprParser::parse7()
{
    unsigned n = 0;
    while (1) {
        if (lex.lookingAt(tt_Plus))
            ;
        else if (lex.lookingAt(tt_Minus))
            ++n;
        else
            break;
    }
    parse8();
    if (n & 1)
        pushFuncall(FX_Negate);
}

/* "exponentiation operators"
   expr8 ::= expr9
           | expr9 '^' expr7 */
void
IntExprParser::parse8()
{
    parse9();
    if (lex.lookingAt(tt_Pot)) {
        parse7();
        pushFuncall(FX_Pow);
    }
}

/* "operand". Grammar is, sort of, this one:
   expr9 ::= string-lit | int-lit | float-lit | bool-lit
           | compound
   compound ::= '(' expr0 ')'
              | reference
              | compound '.' reference
   reference ::= id
              | id '(' ')'
              | id '(' arglist ')'
   arglist ::= expr0
             | expr0 ',' arglist */
void
IntExprParser::parse9()
{
    switch (lex.getType()) {
     case tt_Left:
        lex.lex();
        parse0();
        if (!lex.lookingAt(tt_Right))
            throw IntException("Expected `)'");
        parse10();
        break;
     case tt_String:
        push(new IntExprLiteral(IntObject(lex.getString())));
        lex.lex();
        break;
     case tt_Int:
        push(new IntExprLiteral(IntObject(lex.getInt())));
        lex.lex();
        break;
     case tt_Bool:
        push(new IntExprLiteral(IntObject(bool(lex.getInt()))));
        lex.lex();
        break;
     case tt_Float:
        push(new IntExprLiteral(IntObject(lex.getFloat())));
        lex.lex();
        break;
     case tt_Id:
        if (isBuiltinFunction(lex.getString())) {
            /* must be a function call */
            string_t name = lex.getString();
            lex.lex();
            if (!lex.lookingAt(tt_Left))
                throw IntException("`(' expected");
            std::size_t n = parseArglist();
            if (!lex.lookingAt(tt_Right))
                throw IntException("`)' expected");
            createFunctionNode(name, n, *this);
            parse10();
        } else {
            push(new IntUnqualLookup(lex.getString()));
            lex.lex();
            if (lex.lookingAt(tt_Left)) {
                IntExprNodePtr p = pop();
                std::size_t n = parseArglist();
                if (!lex.lookingAt(tt_Right))
                    throw IntException("`)' expected");
                push(new IntApply(p, n, *this));
                parse10();
            }
        }
        break;
     default:
        throw IntException("Invalid operand");
    }

}

void
IntExprParser::parse10()
{
    while (lex.lookingAt(tt_Dot)) {
        if (lex.getType() != tt_Id)
            throw IntException("Identifier expected after `.'");
        push(new IntQualLookup(pop(), lex.getString()));
        lex.lex();
        if (lex.lookingAt(tt_Left)) {
            IntExprNodePtr p = pop();
            std::size_t n = parseArglist();
            if (!lex.lookingAt(tt_Right))
                throw IntException("`)' expected");
            push(new IntApply(p, n, *this));
        } else
            break;
    }
}

/** Parse parameter list. Assumes opening paren already read. Returns
    number of arguments read (and pushed onto stack). Caller must read
    terminating token.

    FIXME: This one has possible terminating tokens hardwired to `)'
    and EOL, to be able to recognize nullary calls. */
std::size_t
IntExprParser::parseArglist()
{
    if (lex.getType() == tt_Right || lex.getType() == tt_Null)
        return 0;

    std::size_t n = 0;
    do {
        parseExpr();
        ++n;
    } while (lex.lookingAt(tt_Comma));
    return n;
}
