/**
  *  \file int/function.cc
  *  \brief Built-in Functions
  *
  *  This implements the wired-in functions. Some of these actually
  *  have special (call-by-name) semantics, others are just implemented
  *  here for reasons of convenience/speed.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include <cstdlib>
#include "int/function.h"
#include "util/bsearch.h"
#include "cpluslib/string.h"
#include "int/lex.h"
#include "int/bytecode.h"
#include "cpluslib/format.h"

/******************************* Functions *******************************/

IntObject
FAbs(IntObject a)
{
    switch(a.getType()) {
     case res_Null:
        return IntObject();
     case res_Int:
     case res_Bool:
        /* bcc doesn't have abs(long) */
        return std::labs(a.getIntVal());
     case res_Float:
        return std::fabs(a.getFloatVal());
     default:
        throw IntException("`Abs' needs a numeric argument");
    }
}

IntObject
FInt(IntObject a)
{
    switch(a.getType()) {
     case res_Null:
        return IntObject();
     case res_Int:
     case res_Bool:
        return a.getIntVal();
     case res_Float:
        /* FIXME? These limits are hardwired in PCC 1.x */
        if (a.getFloatVal() < -0x7FFFFFFF || a.getFloatVal() > 0x7FFFFFFF)
            throw IntException("Value out of range for `Int'");
        return int32(a.getFloatVal());
     default:
        throw IntException("`Int' needs a numeric argument");
    }
}

IntObject
FRound(IntObject a)
{
    switch(a.getType()) {
     case res_Null:
        return IntObject();
     case res_Int:
     case res_Bool:
        return a.getIntVal();
     case res_Float:
        /* FIXME? These limits are hardwired in PCC 1.x */
        if (a.getFloatVal() < 0) {
            if (a.getFloatVal() > -0x7FFFFFFF)
                return int32(a.getFloatVal() - 0.5);
        } else {
            if (a.getFloatVal() < 0x7FFFFFFF)
                return int32(a.getFloatVal() + 0.5);
        }
        throw IntException("Value out of range for `Round'");
     default:
        throw IntException("`Round' needs a numeric argument");
    }
}

IntObject
FSqrt(IntObject a)
{
    if (a.isEmpty())
        return IntObject();
    if (!a.isNumeric())
        throw IntException("`Sqrt' needs a numeric argument");
    a.doPromote(res_Float);
    if (a.getFloatVal() < 0)
        throw IntException("Argument to `Sqrt' is out of range");
    RETURN_FLOAT_VAL(std::sqrt(a.getFloatVal()));
}

IntObject
FZap(IntObject a)
{
    if (a.isEmpty() || a.getBoolVal() == 0)
        return IntObject();
    else
        return a;
}

IntObject
FChr(IntObject a)
{
    if (a.isEmpty())
        return IntObject();
    else if (a.isInt())
        return string_t(1, char(a.getIntVal()));
    else
        throw IntException("`Chr' needs an integer argument");
}

IntObject
FAsc(IntObject a)
{
    if (a.isEmpty())
        return IntObject();
    if (a.convertToString()) {
        if (a.getStringVal().length())
            return int32(static_cast<unsigned char>(a.getStringVal()[0]));
        else
            return IntObject();
    } else {
        throw IntException("`Asc' needs a string argument");
    }
}

/********************************** Trig *********************************/

static double
prepareTrig(IntObject& a)
{
    if (a.isNumeric()) {
        a.doPromote(res_Float);
        return a.getFloatVal() * (IntLexer::PI / 180.0);
    } else
        throw IntException("Numeric argument expected");
}

IntObject
FCos(IntObject a)
{
    if (a.isEmpty())
        return IntObject();
    else
        RETURN_FLOAT_VAL(std::cos(prepareTrig(a)));
}

IntObject
FSin(IntObject a)
{
    if (a.isEmpty())
        return IntObject();
    else
        RETURN_FLOAT_VAL(std::sin(prepareTrig(a)));
}

IntObject
FTan(IntObject a)
{
    if (a.isEmpty()) {
        return IntObject();
    } else {
        double d = prepareTrig(a);
        double c = std::cos(d);
        if (c == 0.0)
            throw IntException("`Tan' not defined for this angle");
        RETURN_FLOAT_VAL(std::sin(d) / c);
    }
}

namespace {
    class IntIfNode : public IntExprLValue {
        IntExprNodePtr a, b, c;
     public:
        IntIfNode(IntExprParser& par, int nargs);
        IntObject eval(IntExecutionContext& con);
        void compileV(IntByteCodeObject&);
        void compileS(IntByteCodeObject&);
    };

    class IntUnaryOrBinaryNode : public IntExprLValue {
     protected:
        IntExprNodePtr a, b;
     public:
        IntUnaryOrBinaryNode(IntExprParser& par, int nargs);
    };

    class IntATanNode : public IntUnaryOrBinaryNode {
     public:
        IntATanNode(IntExprParser& par, int nargs)
            : IntUnaryOrBinaryNode(par, nargs) { }
        IntObject eval(IntExecutionContext& con);
        void compileV(IntByteCodeObject&);
        void compileS(IntByteCodeObject&);
    };

    class IntStrNode : public IntUnaryOrBinaryNode {
     public:
        IntStrNode(IntExprParser& par, int nargs)
            : IntUnaryOrBinaryNode(par, nargs) { }
        IntObject eval(IntExecutionContext& con);
        void compileV(IntByteCodeObject&);
        void compileS(IntByteCodeObject&);
    };

    enum IntBuiltinCallType {
        bct_Unary,
        bct_Binary,
        bct_ATan,
        bct_If,
        bct_Str
    };

    struct IntBuiltinFunction {
        const char* name;
        short       min_args, max_args;
        IntBuiltinCallType call;
        //IntGenericFunctionPtr ptr;
        int         code;
    };
}

IntIfNode::IntIfNode(IntExprParser& par, int nargs)
{
    if (nargs > 2)
        c = par.pop();
    b = par.pop();
    a = par.pop();
}

IntObject
IntIfNode::eval(IntExecutionContext& con)
{
    IntObject av = a->eval(con);
    if (! av.isScalar())
        throw IntException("First argument to `If' must be scalar");
    if (av.getBoolVal() == 1)
        return b->eval(con);
    else
        if (c.ptr())
            return c->eval(con);
        else
            return IntObject();
}

void
IntIfNode::compileV(IntByteCodeObject& bco)
{
    IntByteCodeObject::label_t f = bco.allocLabel(), end = bco.allocLabel();
    a->compileV(bco);
    bco.add(mkCode(op_JumpFEPop, f));
    b->compileV(bco);
    bco.add(mkCode(op_Jump, end));
    bco.setLabel(f);
    if (c)
        c->compileV(bco);
    else
        bco.add(mkCode(op_PushNull, 0));
    bco.setLabel(end);
}

void
IntIfNode::compileS(IntByteCodeObject& bco)
{
    IntByteCodeObject::label_t f = bco.allocLabel(), end = bco.allocLabel();
    a->compileV(bco);
    bco.add(mkCode(op_JumpFEPop, f));
    b->compileS(bco);
    bco.add(mkCode(op_Jump, end));
    bco.setLabel(f);
    if (c)
        c->compileS(bco);
    bco.setLabel(end);
}

IntUnaryOrBinaryNode::IntUnaryOrBinaryNode(IntExprParser& par, int nargs)
{
    if (nargs > 1)
        b = par.pop();
    a = par.pop();
}

static double
arctangent(double y, double x)
{
    if (std::fabs(x) < IntEPSILON) {
        if (std::fabs(y) < IntEPSILON)
            throw IntException("`ATan' not defined for (0,0)");
        else if (y > 0)
            return 90;
        else
            return 270;
    } else {
        double f = std::atan(y / x) * (180 / IntLexer::PI);
        if (x < 0)
            f += 180;
        if (f < 0)
            f += 360;
        if (f > 360)
            f -= 360;
        return f;
    }
}

IntObject
IntATanNode::eval(IntExecutionContext& con)
{
    /* A bit more verbose than it needs to be, because we
       guarantee strict by-value evaluation. */
    /* FIXME: `ATan("string", EMPTY)' is accepted while
       `ATan(EMPTY, "string")' is not. */
    IntObject av = a->eval(con);
    double    divi = 1.0;
    if (b) {
        IntObject bv = b->eval(con);
        if (bv.isEmpty())
            return IntObject();
        if (! bv.isNumeric())
            throw IntException("`ATan' needs numeric arguments");
        bv.doPromote(res_Float);
        divi = bv.getFloatVal();
    }
    if (av.isEmpty())
        return IntObject();
    if (! av.isNumeric())
        throw IntException("`ATan' needs numeric arguments");
    av.doPromote(res_Float);
    return arctangent(av.getFloatVal(), divi);
}

void
IntATanNode::compileV(IntByteCodeObject&)
{
    FAIL();
}

void
IntATanNode::compileS(IntByteCodeObject&)
{
    FAIL();
}

IntObject
IntStrNode::eval(IntExecutionContext& con)
{
    IntObject av = a->eval(con);
    if (b.ptr()) {
        /* second argument is precision, and is only evaluated for
           for numeric output */
        IntObject bv = b->eval(con);
        if (bv.isEmpty())
            return IntObject();
        if (! (bv.isInt() || bv.isBool()) || bv.getIntVal() < 0)
            throw IntException("Second argument to `Str' must be positive integer");
        if (av.isInt() || av.isFloat()) {
            av.doPromote(res_Float);
            string_t fmtstr = format("%%.%df" /* hooray! */, bv.getIntVal());
            return string_t(format(fmtstr.c_str(), av.getFloatVal()));
        }
    }
    if (av.isEmpty())
        return IntObject();
    if (! av.isScalar())
        throw IntException("First argument to `Str' must be scalar");
    av.convertToString();
    return av;
}

void
IntStrNode::compileV(IntByteCodeObject&)
{
    FAIL();
}

void
IntStrNode::compileS(IntByteCodeObject&)
{
    FAIL();
}

/************************ Interface to the Outside ***********************/

#define FUNC_UNARY(f)   1, 1, bct_Unary,  f
#define FUNC_BINARY(f)  2, 2, bct_Binary, f

static IntBuiltinFunction builtin_functions[] = {
    { "ABS",     FUNC_UNARY(FX_Abs) },
    { "ASC",     FUNC_UNARY(FX_Asc) },
    { "ATAN",    1, 2, bct_ATan, 0 },
    { "CHR",     FUNC_UNARY(FX_Chr) },
    { "CHR$",    FUNC_UNARY(FX_Chr) },
    { "COS",     FUNC_UNARY(FX_Cos) },
    { "IF",      2, 3, bct_If, 0 },
    { "INT",     FUNC_UNARY(FX_Int) },
    { "ROUND",   FUNC_UNARY(FX_Round) },
    { "SIN",     FUNC_UNARY(FX_Sin) },
    { "SQR",     FUNC_UNARY(FX_Sqrt) },
    { "SQRT",    FUNC_UNARY(FX_Sqrt) },
    { "STR",     1, 2, bct_Str, 0 },
    { "TAN",     FUNC_UNARY(FX_Tan) },
    { "Z",       FUNC_UNARY(FX_Zap) },
    { "ZAP",     FUNC_UNARY(FX_Zap) },
};

//     +#3'ASC'#6'STRING'#7'STRING$'#3'LEN'#3'CFG'
//     +#5'RTRIM'#5'LTRIM'#4'TRIM'#4'LEFT'
//     +#5'RIGHT'#3'MID'#5'INSTR'#3'VAL'#8'SHIPNAME'
//     +#10'PLANETNAME'#7'STRCASE'#6'RANDOM'#3'MIN'#3'MAX'
//     +#4'CMUL'#4'CSUB'#4'CADD'#8'CCOMPARE'
//     +#8'CEXTRACT'#7'CREMOVE'#10'COUNTSHIPS'
//     +#8'FINDSHIP'#12'COUNTPLANETS'#10'FINDPLANET'
//     +#8'DISTANCE'
//     +#4'ATOM'#7'ATOMSTR'#8'TRUEHULL'#4'FPOS'
//     +#5'FSIZE'#7'GETBYTE'#7'GETWORD'#7'GETLONG'
//     +#6'GETSTR'#8'FREEFILE'#5'ISNUM'#8'ISSTRING'
//     +#5'FIRST'#4'REST'#4'SHIP'

namespace {
struct IntCaseblindCompare {
    const string_t& aname;
    IntCaseblindCompare(const string_t& an)
        : aname(an) { }
    bool isLess(const IntBuiltinFunction& c) const {
        return strCaseCmp(aname, c.name) < 0;
    }
    bool isEqual(const IntBuiltinFunction& c) const {
        return strCaseCmp(aname, c.name) == 0;
    }
};
}

/** Check whether name is a built-in function. */
bool
isBuiltinFunction(const string_t& name)
{
    int index = bin_search(builtin_functions, countof(builtin_functions),
                           IntCaseblindCompare(name));
    return index >= 0;
}

/** Create a function node for name.
    \param nargs   number of arguments encountered so far
    \param parser  parameter parser (containing arg stack) */
bool
createFunctionNode(const string_t& name, int nargs, IntExprParser& parser)
{
    int index = bin_search(builtin_functions, countof(builtin_functions),
                           IntCaseblindCompare(name));
    if (index < 0)
        return false;
    if (nargs < builtin_functions[index].min_args)
        throw IntException("Too few arguments for `" + name + "'");
    if (nargs > builtin_functions[index].max_args)
        throw IntException("Too many arguments for `" + name + "'");

    switch(builtin_functions[index].call) {
     case bct_If:
        parser.push(new IntIfNode(parser, nargs));
        break;
     case bct_ATan:
        parser.push(new IntATanNode(parser, nargs));
        break;
     case bct_Str:
        parser.push(new IntStrNode(parser, nargs));
        break;
     case bct_Unary:
        parser.pushFuncall(IntUnaryFunction(builtin_functions[index].code));
        break;
     case bct_Binary:
        parser.pushFuncall(IntBinaryFunction(builtin_functions[index].code));
        break;
    }
    return true;
}
