/**
  *  \file int/lex.cc
  *  \brief CCScript Lexer
  *
  *  Very similar to PCC 1.x lexer.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include "int/lex.h"
#include "util/nls.h"
#include "cpluslib/stringcv.h"

using namespace std;

const double IntLexer::PI = 3.14159265358979323844;

static bool
isIdentifierChar(char c)
{
    if ('A' == 65 && 'a' == 97 && 'Z' == 90) {
        /* we have reason to believe this is an ASCII system */
        return (c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z')
            || (c >= '0' && c <= '9');
    } else {
        /* not ASCII. Use portable but locale-dependant way */
        return isalnum(c);
    }
}

/** \class IntLexer
    CCScript Lexer. Very similar to PCC 1.x lexer, only differences
    known:
    - we use double, not real (bigger range)
    - no line/identifier/whatever length limits
    - we do no longer permit unterminated string constants (rationale:
      in "foo := 'bar\\r", does the \\r belong to the string?) */

IntLexer::IntLexer(string_t line)
    : the_line(line), index(0)
{ }

char
IntLexer::current() const
{
    if (index >= the_line.length())
        return 0;
    else
        return the_line[index];
}

#define RETURN(c) return (current_token = (c))

/** Main entry point. Read a token and return its code. Parameters
    (string for tt_String, tt_Id; value for tt_Float, tt_Bool, tt_Int)
    can be inquired later with getString(), getInt(), getFloat()). */
IntTokenType
IntLexer::lex()
{
    char c;
    do {
        if (index >= the_line.length())
            RETURN(tt_Null);
        c = the_line[index++];
    } while (isspace(c));
    switch (c) {
     case '%':
        index = string_t::npos;
        RETURN(tt_Null);
     case '&':
        RETURN(tt_Join);
     case '#':
        RETURN(tt_Hash);
     case '+':
        RETURN(tt_Plus);
     case '-':
        RETURN(tt_Minus);
     case ';':
        RETURN(tt_Semi);
     case '*':
        RETURN(tt_Mul);
     case '/':
        RETURN(tt_RealDiv);
     case '\\':
        RETURN(tt_Div);
     case '^':
        RETURN(tt_Pot);
     case '(':
        RETURN(tt_Left);
     case ')':
        RETURN(tt_Right);
     case ',':
        RETURN(tt_Comma);
     case '=':
        RETURN(tt_EQ);
     case ':':
        if (current() == '=') {
            ++index;
            RETURN(tt_Asgn);
        } else {
            RETURN(tt_Colon);
        }
     case '<':
        if (current() == '>') {
            ++index;
            RETURN(tt_NE);
        } else if (current() == '=') {
            ++index;
            RETURN(tt_LE);
        } else {
            RETURN(tt_LT);
        }
     case '>':
        if (current() == '=') {
            ++index;
            RETURN(tt_GE);
        } else {
            RETURN(tt_GT);
        }
     case '\'':
        {
            string_t::size_type n = the_line.find('\'', index);
            if (n == string_t::npos)
                throw IntException(_("Unterminated string constant"));
            current_str.assign(the_line, index, n - index);
            index = n+1;
            RETURN(tt_String);
        }
     case '"':
        {
            current_str.clear();
            while (1) {
                if (index >= the_line.length())
                    throw IntException(_("Unterminated string constant"));
                c = the_line[index++];
                if (c == '"')
                    break;
                if (c == '\\') {
                    if (index >= the_line.length())
                        throw IntException(_("Unterminated string constant"));
                    c = the_line[index++];
                }
                current_str.append(1, c);
            }
            RETURN(tt_String);
        }
     case '0': case '1': case '2': case '3': case '4':
     case '5': case '6': case '7': case '8': case '9':
     case '.':
        if (c == '.' && (current() < '0' || current() > '9')) {
            RETURN(tt_Dot);
        } else {
            uint32 ival = 0;
            double rval = 0;
            --index;            // re-read current character
            while ((c = current()) >= '0' && c <= '9') {
                unsigned v = c - '0';
                ival = 10*ival + v;     // unsigned -> no need to handle overflow
                rval = 10*rval + v;
                ++index;
            }
            if (c == '.') {
                /* it's going to be a real number */
                double imul = 1;
                ++index;
                while ((c = current()) >= '0' && c <= '9') {
                    imul /= 10;
                    rval += imul * (c - '0');
                    ++index;
                }
                current_dbl = rval;
                RETURN(tt_Float);
            } else {
                /* integer if in range, float otherwise */
                if (rval <= 0x7FFFFFFFL) {     // FIXME: numeric_limits...
                    current_int = ival;
                    RETURN(tt_Int);
                } else {
                    current_dbl = rval;
                    RETURN(tt_Float);
                }
            }
        }
     default:
        if (isIdentifierChar(c)) {
            current_str.assign(1, char(toupper(c)));
            while (index < the_line.length()
                   && (isIdentifierChar((c = the_line[index]))
                       || c == '$' || c == '_' || c == '.'))
            {
                current_str.append(1, char(toupper(c)));
                ++index;
            }
            if (current_str == "TRUE") {
                current_int = 1;
                RETURN(tt_Bool);
            } else if (current_str == "FALSE") {
                current_int = 0;
                RETURN(tt_Bool);
            } else if (current_str == "PI") {
                current_dbl = PI;
                RETURN(tt_Float);
            } else if (current_str == "AND") {
                RETURN(tt_AND);
            } else if (current_str == "OR") {
                RETURN(tt_OR);
            } else if (current_str == "XOR") {
                RETURN(tt_XOR);
            } else if (current_str == "MOD") {
                RETURN(tt_MOD);
            } else if (current_str == "NOT") {
                RETURN(tt_NOT);
            }
            RETURN(tt_Id);
        }
    }
    RETURN(tt_Invalid);
}

IntException::IntException(const string_t& awhat)
    : std::runtime_error(toStandardString(awhat))
{ }
