/**
  *  \file game/pvcr.cc
  *  \brief PVCR Implementation
  *
  *  (c) 2001 Stefan Reuther <Streu@gmx.de>
  */

#include "game/pvcrt.h"
#include "game/config.h"
#include "game/spec.h"

/** \class VcrPlayerPHost
    \brief PHost VCR Player

    This is the PHost VCR player, started as a port from the PCC
    v1.0.15 player. The initial port was 50% slower than the PDK
    implementation; this one is 33% faster than the PDK.

    Things remaining to do:
    - some things should be done in subroutines (killFighter...)
    - cache weapon specs */

/*
  This seems to be pretty optimisation-resistant.

  + original:        AC: 15.74,  Non-AC: 3.87
  + inlining:        AC: 15.27,  Non-AC: 3.72
  + PVCR_INTEGER:    AC: 15.25,  Non-AC: 3.74(!)
  + f_getDamage():   AC: 16.11(!) [not used]
  + caching config:  AC: 13.58,  Non-AC: 3.59
  + fighterMove:     AC: 11.91,  Non-AC: 3.41
  + beamFindFighter: AC:  6.70,  Non-AC: 2.04

  For comparison, PDK 4.2:  AC: 9.50,  Non-AC: 2.44
  PDK 4.3 (beamFindFighter) AC: 5.53,  Non-AC: 1.72

  Some things which do not get anything:
  + Status& st = status[side]
  + side^1 instead of !side (although it should give better code;
    in theory, that is)
*/

#ifdef PCC_STATIC_VCR
TVcr VcrPlayerPHost::vcr;
VcrPlayerPHost::Status VcrPlayerPHost::status[2];
uint16 VcrPlayerPHost::seed;
bool VcrPlayerPHost::done;
int VcrPlayerPHost::one_f, VcrPlayerPHost::right_probab;
#endif

/** Get config /KEY/ for object on /SIDE/. Use this during
    initialisation. */
#define CONFIG1(KEY, SIDE) (config.KEY[vcr.objects[SIDE].owner-1])

/** Get config /KEY/ for object on /SIDE/. Use this during
    the battle. This uses the cached values. */
#define CONFIG2(KEY, SIDE) (status[SIDE].f.KEY)

/** Distance. Shortcut for later optimisatitons. */
#define DISTANCE() (long(status[1].r.obj_x) - status[0].r.obj_x)

/********************************** RNG **********************************/

/** Random Number Generator. Basic linear congruence. */
uint16
VcrPlayerPHost::random64k()
{
    seed = 13489 - seed * 259;
    return seed;
}

/** Random Number Generator. Returns a random number uniformly
    in [0, max) */
int
VcrPlayerPHost::randomRange(int max)
{
    int divisor = 65535U / max;
    int i;
    do
        i = random64k() / divisor;
    while (i >= max);
    return i;
}

/** Check for planet. Returns true iff /side/ is the planet. */
PVCR_TEMPLATE
inline bool
VcrPlayerPHost::isPlanet()
{
    return side && vcr.battle_type;
}

/************************** Standard "Hit" Stuff *************************/

#ifdef PVCR_INTEGER
static inline int
rdiv(int32 a, int32 b)
{
    return (a + b/2) / b;
}

/** Compute shield damage. Documented formula. */
PVCR_TEMPLATE
int32
VcrPlayerPHost::computeShieldDamageS(int expl, int mass)
{
    if (config.AllowAlternativeCombat) {
        int32 damage = CONFIG2(ShieldDamageScaling, side) * int32(expl);
        return std::min(damage, status[side].f.max_scaled);
    } else {
        int32 damage = rdiv(CONFIG2(ShieldDamageScaling, side) * int32(expl),
                            mass + 1) + 1;
        if (damage > 10000)
            return 10000;
        else
            return damage;
    }
}

PVCR_TEMPLATE
int32
VcrPlayerPHost::computeHullDamageS(int expl, int mass)
{
    if (config.AllowAlternativeCombat) {
        int32 d = int32(expl) * CONFIG2(HullDamageScaling, side);
        return std::min(d, status[side].f.max_scaled);
    } else {
        int32 d = rdiv(computeShieldDamageS PVCR_TPLCALL (expl, mass) * CONFIG2(HullDamageScaling, side),
                       mass + 1) + 1;
        if (d > 10000)
            return 10000;
        else
            return d;
    }
}

PVCR_TEMPLATE
int32
VcrPlayerPHost::computeCrewKilledS(int kill, int mass)
{
    if (config.AllowAlternativeCombat) {
        return CONFIG2(CrewKillScaling, side) * int32(kill);
    } else {
        return rdiv(CONFIG2(CrewKillScaling, side) * int32(kill), mass + 1);
    }
}
#else
/** Compute shield damage. Documented formula. */
PVCR_TEMPLATE
VcrPlayerPHost::damage_t
VcrPlayerPHost::computeShieldDamage(int expl, int mass)
{
    double damage = CONFIG2(ShieldDamageScaling, side) * double(expl) / double(mass + 1);
    if (damage > 10000)
        damage = 10000;
    if (! config.AllowAlternativeCombat)
        return (int) (damage + 1.5);
    else
        return damage;
}

PVCR_TEMPLATE
VcrPlayerPHost::damage_t
VcrPlayerPHost::computeHullDamage(int expl, int mass)
{
    if (config.AllowAlternativeCombat) {
        double d = double(expl) * CONFIG2(HullDamageScaling, side) / double(mass + 1);
        if (d > 10000)
            return 10000;
        else
            return d;
    } else {
        double d = computeShieldDamage PVCR_TPLCALL (expl, mass) * CONFIG2(HullDamageScaling, side) / double(mass + 1);
        if (d > 10000)
            return 10000;
        else
            return (int) (d + 1.5);
    }
}

PVCR_TEMPLATE
VcrPlayerPHost::damage_t
VcrPlayerPHost::computeCrewKilled(int kill, int mass)
{
    double d = CONFIG2(CrewKillScaling, side) * double(kill) / double(mass + 1);
    if (config.AllowAlternativeCombat)
        return d;
    else
        return (int) (d + 0.5);
}
#endif

/** Hit object.
    \param side  object being hit
    \param kill  anti-life (x-ray) power of weapon
    \param expl  explosive power of weapon */
PVCR_TEMPLATE
bool
VcrPlayerPHost::hit(int kill, int expl)
{
#ifdef PVCR_INTEGER
    int damage_rate = 100;
    bool over = false;
    Status& st = status[side];
    
    if (kill <= 0)
        kill = 1;
    if (expl <= 0)
        expl = 1;

    /* Hit shields */
    if (st.r.shield_scaled > 0) {
        int32 damage_s = computeShieldDamageS PVCR_TPLCALL (expl, vcr.mass[side]);
        if (st.r.shield_scaled <= damage_s) {
            // Shields are completely worn down by this hit
            damage_rate = (damage_s - st.r.shield_scaled) * 100 / damage_s;
            st.r.shield_scaled = 0;
        } else {
            // Shields still hold
            damage_rate = 0;
            st.r.shield_scaled -= damage_s;
        }
    }

    /* Shields are down -- do damage */
    if (damage_rate > 0) {
        int32 damage_s = computeHullDamageS PVCR_TPLCALL (expl, vcr.mass[side]) * damage_rate;
        st.r.damage_scaled2 += damage_s;

        if (st.r.damage_scaled2 >= st.f.damage_limit_scaled) {
            over = true;
        } else if (!isPlanet PVCR_TPLCALL()) {
            int32 killed_s = computeCrewKilledS PVCR_TPLCALL (kill, vcr.mass[side]) * damage_rate;
            st.r.crew_scaled2 -= killed_s;
            if (st.r.crew_scaled2 < 50*st.f.scale) {
                st.r.crew_scaled2 = 0;
                over = true;
            }
        }            
    }
    return over;
#else
    int damage_rate = 100;
    bool over = false;
    
    if (kill <= 0)
        kill = 1;
    if (expl <= 0)
        expl = 1;

    /* Hit shields */
    if (status[side].r.shield > 0) {
        double damage = computeShieldDamage PVCR_TPLCALL (expl, vcr.mass[side]);
        if (status[side].r.shield <= damage) {
            // Shields are completely worn down by this hit
            damage_rate = int((damage - status[side].r.shield) * 100 / damage);
            status[side].r.shield = 0;
        } else {
            // Shields still hold
            damage_rate = 0;
            status[side].r.shield -= damage;
        }
    }

    /* Shields are down -- do damage */
    if (damage_rate > 0) {
        double damage = computeHullDamage PVCR_TPLCALL (expl, vcr.mass[side]) * damage_rate / 100.0;
        status[side].r.damage += damage;

        if (int(status[side].r.damage + 0.5) >= status[side].f.damage_limit) {
            over = true;
        } else if (!isPlanet PVCR_TPLCALL()) {
            double killed = computeCrewKilled PVCR_TPLCALL (kill, vcr.mass[side]) * damage_rate / 100.0;
            status[side].r.crew -= killed;
            if (status[side].r.crew < 0.5) {
                status[side].r.crew = 0;
                over = true;
            }
        }            
    }

    return over;
#endif
}

/******************************** Fighters *******************************/

int
VcrPlayerPHost::computeBayRechargeRate(int num, int side)
{
    int i = CONFIG1(BayRechargeBonus, side) * num
        + CONFIG1(BayRechargeRate, side);
    return i > 1 ? i : 1;
}

/** Recharge Fighter Bays. */
PVCR_TEMPLATE
void
VcrPlayerPHost::fighterRecharge()
{
    register const int mx = vcr.objects[side].bay_count;
    register Status& st = status[side];
    for (int i = 0; i < mx; ++i)
        if (st.r.bay_status[i] < 1000)
            st.r.bay_status[i] += randomRange(st.f.bay_recharge);
}

/** Launch Fighters. */
PVCR_TEMPLATE
void
VcrPlayerPHost::fighterLaunch()
{
    register Status& st = status[side];
    
    /* can we launch a fighter? */
    if (vcr.num_fighters[side] == 0
        || st.r.launched_fighters >= CONFIG2(MaxFightersLaunched, side)
        || st.r.launch_countdown > 0)
        return;

    /* yes, we can! */
    int bay_mx = vcr.objects[side].bay_count;
    for (int bay = 0; bay < bay_mx; ++bay)
        if (st.r.bay_status[bay] >= 1000)
            for (int track = 0; track < CONFIG2(MaxFightersLaunched, side); ++track)
                if (st.r.fighter_status[track] == 0) {
                    /* okay, we have a bay which is ready, and an empty track. */
                    st.r.fighter_status[track]     = 1;
                    st.r.fighter_x[track]          = st.r.obj_x;
                    st.r.fighter_shots_left[track] = CONFIG2(StrikesPerFighter, side);
                    st.r.bay_status[bay]           = 0;
                    st.r.launched_fighters++;
                    vcr.num_fighters[side]--;
                    st.r.launch_countdown = CONFIG2(BayLaunchInterval, side);
                    vis->startFighter(side, track);
                    return;
                }
}

/** Move Fighters. Takes back returned fighters. */
PVCR_TEMPLATE
void
VcrPlayerPHost::fighterMove()
{
    if (status[side].r.launched_fighters == 0)
        return;
    
    int dir = (side == 0) ? CONFIG2(FighterMovementSpeed, 0) : -CONFIG2(FighterMovementSpeed, 1);

    for (int track = 0; track < CONFIG2(MaxFightersLaunched, side); ++track)
        if (status[side].r.fighter_status[track] == 1) {
            status[side].r.fighter_x[track] += dir;
        } else if (status[side].r.fighter_status[track] == 2) {
            status[side].r.fighter_x[track] -= dir;
            if ((side == 0)
                ? status[side].r.fighter_x[track] < status[side].r.obj_x
                : status[side].r.fighter_x[track] > status[side].r.obj_x)
            {
                /* fighter comes back to baseship */
                status[side].r.fighter_status[track] = 0;
                status[side].r.launched_fighters--;
                vcr.num_fighters[side]++;
                vis->landFighter(side, track);
            }
        }
}

/** Do fighter intercepts. */
void
VcrPlayerPHost::fighterIntercept()
{
    static const int NEVER = -0x4000;
    
    /* fighter intercept only happens if both have fighters */
    if (status[0].r.launched_fighters == 0 || status[1].r.launched_fighters == 0)
        return;

    /* compute screen positions */
    int match[2][VCR_MAX_FTRS];
    for (int w = 0; w < 2; ++w)
        for (int i = 0; i < CONFIG2(MaxFightersLaunched, w); ++i)
            if (status[w].r.fighter_status[i] == 0)
                match[w][i] = NEVER;
            else
                match[w][i] = status[w].r.fighter_x[i] >> 7;

    for (int lf = 0; lf < CONFIG2(MaxFightersLaunched, 0); ++lf)
        if (match[0][lf] != NEVER)
            if (randomRange(100) < one_f)
                for (int rf = 0; rf < CONFIG2(MaxFightersLaunched, 1); ++rf)
                    if (match[1][rf] == match[0][lf]) {
                        /* two fighters at the same place, intercepting */
                        /* theoretically, here is a slight imbalance. If
                           the "right" fighter survives, it can fire again
                           while a "left" one can not. Whether this is
                           relevant in practice is unknown. */
                        if (randomRange(100) < right_probab) {
                            vis->fireBeam(1, rf, lf, 1, CONFIG2(FighterBeamExplosive, 1), CONFIG2(FighterBeamKill, 1));
                            status[0].r.fighter_status[lf] = 0;
                            status[0].r.launched_fighters--;
                            vis->killFighter(0, lf);
                        } else {
                            vis->fireBeam(0, lf, rf, 1, CONFIG2(FighterBeamExplosive, 0), CONFIG2(FighterBeamKill, 0));
                            status[1].r.fighter_status[rf] = 0;
                            status[1].r.launched_fighters--;
                            vis->killFighter(1, rf);
                            match[1][rf] = NEVER;
                        }
                        break;
                    }
}

/** Fighters attack enemy. */
PVCR_TEMPLATE
bool
VcrPlayerPHost::fighterAttack()
{
    /* only if we have fighters */
    if (status[side].r.launched_fighters == 0)
        return false;

    int enemy_x = status[!side].r.obj_x;
    for (int i = 0; i < CONFIG2(MaxFightersLaunched, side); ++i)
        if (status[side].r.fighter_status[i] == 1
            && std::abs(int32(status[side].r.fighter_x[i]) - enemy_x) <= CONFIG2(FighterFiringRange, side))
        {
            status[side].r.fighter_shots_left[i]--;
            if (!status[side].r.fighter_shots_left[i])
                status[side].r.fighter_status[i] = 2;

            bool hitres = hit<!side>(CONFIG2(FighterBeamKill, side), CONFIG2(FighterBeamExplosive, side));
            vis->fireBeam(side, i, -1, 1,
                          CONFIG2(FighterBeamExplosive, side),
                          CONFIG2(FighterBeamKill, side));
            if (hitres)
                return true;            
        }
    return false;
}

/********************************* Beams *********************************/

int
VcrPlayerPHost::computeBeamHitOdds(int beam, int side)
{
    GBeam& b = getBeam(beam);
    int i = CONFIG1(BeamHitBonus, side) * (b.getKill() + b.getDamage()) / 100
        + CONFIG1(BeamHitOdds, side);
    return i < 0 ? 0 : i;
}

int
VcrPlayerPHost::computeBeamRechargeRate(int beam, int side)
{
    GBeam& b = getBeam(beam);
    int i = ((b.getKill() + b.getDamage()) * CONFIG1(BeamRechargeBonus, side)) / 100
        + CONFIG1(BeamRechargeRate, side);
    return i < 1 ? 1 : i;
}

PVCR_TEMPLATE
void
VcrPlayerPHost::beamRecharge()
{
    register const int mx = vcr.objects[side].beam_count;
    for (int i = 0; i < mx; ++i) {
        if (status[side].r.beam_status[i] < 1000) {
            status[side].r.beam_status[i] += randomRange(status[side].f.beam_recharge);
            vis->updateBeam(side, i);
        }
    }
}

PVCR_TEMPLATE
int
VcrPlayerPHost::beamFindNearestFighter()
{
    const int opp = !side;
    long mindist = 100000;
    int fighter = -1;

    /* When enemy does not have a fighter, don't fire at one... */
    if (status[opp].r.launched_fighters == 0)
        return -1;

    long my_x = status[side].r.obj_x;
    
    /* Look for a matching fighters. When we're asked to fire on
       attacking fighters only, accept only those first. */
    for (int i = 0; i < CONFIG2(MaxFightersLaunched, opp); ++i)
        if (status[opp].r.fighter_status[i] == 1
            || (!config.FireOnAttackFighters && status[opp].r.fighter_status[i] != 0))
        {
            long d = std::abs(my_x - status[opp].r.fighter_x[i]);
            if (d < mindist) {
                mindist = d;
                fighter = i;
            }
        }

    /* No fighter found. Check retreating fighters. Any of them does. */
    if (fighter < 0 && config.FireOnAttackFighters) {
        for (int i = 0; i < CONFIG2(MaxFightersLaunched, opp); ++i)
            if (status[opp].r.fighter_status[i] != 0) {
                fighter = i;
                break;
            }
        return fighter;
    } else {
        /* Found a fighter on first pass. Must be in range. */
        if (mindist > CONFIG2(BeamHitFighterRange, side))
            return -1;
        else
            return fighter;
    }
}

/** Fire beams on /side/. \returns true iff battle ends. */
PVCR_TEMPLATE
bool
VcrPlayerPHost::beamFire()
{
    const int opp = !side;
    for (int beam = 0; beam < vcr.objects[side].beam_count; ++beam) {
        const GBeam& gb = getBeam(vcr.objects[side].beam_type);
        /* Can we fire at a fighter? */
        if (status[side].r.beam_status[beam] >= CONFIG2(BeamHitFighterCharge, side)) {
            /* PVCR tests `&& distance <= 100000', but that's not needed */
            bool missing = randomRange(100) >= status[side].f.beam_hit_odds;
            int fighter = beamFindNearestFighter<side>();
            if (fighter >= 0) {
                /* We fire at a fighter. */
                vis->fireBeam(side, -1-beam, fighter, missing ? -1 : +1,
                              gb.getDamage(), gb.getKill());
                status[side].r.beam_status[beam] = 0;
                vis->updateBeam(side, beam);
                if (!missing) {
                    vis->killFighter(opp, fighter);
                    status[opp].r.fighter_status[fighter] = 0;
                    status[opp].r.launched_fighters--;
                }
                return false;
            }
        }

        /* Can we fire at the enemy? */
        if (status[opp].r.launched_fighters <= 0
            && status[side].r.beam_status[beam] >= CONFIG2(BeamHitShipCharge, side)
            && DISTANCE() <= CONFIG2(BeamFiringRange, side))
        {
            bool missing = randomRange(100) >= status[side].f.beam_hit_odds;
            int kill = gb.getKill()   * (status[side].r.beam_status[beam]/10) / 100;
            int dest = gb.getDamage() * (status[side].r.beam_status[beam]/10) / 100;
            if (raceId(vcr.objects[side].owner) == 5)
                kill *= 3;
            
            status[side].r.beam_status[beam] = 0;
            vis->updateBeam(side, beam);
            
            if (!missing) {
                bool hitr = hit<opp>(kill, dest);
                vis->fireBeam(side, -1-beam, -1, 1, dest, kill);
                if (hitr)
                    return true;
            } else {
                vis->fireBeam(side, -1-beam, -1, -1, dest, kill);
            }
            return false;
        }
    }
    return false;
}

/******************************* Torpedoes *******************************/

int
VcrPlayerPHost::computeTorpHitOdds(int torp, int side)
{
    const GTorpLauncher& tl = getTorp(torp);
    int i = ((CONFIG1(TorpHitBonus, side) * (tl.getKill() + tl.getDamage())) / 100
             + CONFIG1(TorpHitOdds, side));
    return i < 0 ? 0 : i;
}

int
VcrPlayerPHost::computeTubeRechargeRate(int torp, int side)
{
    const GTorpLauncher& tl = getTorp(torp);
    int i = ((CONFIG1(TubeRechargeBonus, side) * (tl.getKill() + tl.getDamage())) / 100
             + CONFIG1(TubeRechargeRate, side));
    return i < 1 ? 1 : i;
}

PVCR_TEMPLATE
void
VcrPlayerPHost::torpsRecharge()
{
    register const int mx = vcr.torp_launcher_count[side];
    for (int i = 0; i < mx; ++i)
        if (status[side].r.torp_status[i] < 1000) {
            status[side].r.torp_status[i] += randomRange(status[side].f.torp_recharge);
            vis->updateTorp(side, i);
        }
}

PVCR_TEMPLATE
bool
VcrPlayerPHost::torpsFire()
{
    if (DISTANCE() > CONFIG2(TorpFiringRange, side) || vcr.num_torps[side] == 0)
        return false;

    /* Scaling factor for torpedo effect. Tim scales with 2 for some reason. */
    int factor = config.AllowAlternativeCombat ? 1 : 2;

    for (int launcher = 0; launcher < vcr.torp_launcher_count[side]; ++launcher)
        if (status[side].r.torp_status[launcher] >= 1000) {
            /* we're firing a torpedo */
            int rr = randomRange(100);
            
            vcr.num_torps[side]--;
            status[side].r.torp_status[launcher] = 0;
            vis->updateTorp(side, launcher);
            if (rr <= status[side].f.torp_hit_odds) {
                /* we hit the enemy */
                GTorpLauncher& tl = getTorp(vcr.objects[side].torp_type);
                bool hitr = hit<!side>(tl.getKill() * factor, tl.getDamage() * factor);
                vis->fireTorp(side, rr, launcher);
                return hitr;
            } else {
                /* miss */
                vis->fireTorp(side, -1-rr, launcher);
                return false;
            }
        }
    return false;
}

/******************************** Control ********************************/

void
VcrPlayerPHost::moveObjects()
{
    long remain = DISTANCE() - config.StandoffDistance;
    if (remain <= 0)
        return;
    
    /* move objects. Ensure StandoffDistance is not violated */
    long move_left = std::min (remain, long(CONFIG2(ShipMovementSpeed, 0)));
    status[0].r.obj_x += move_left;
    remain -= move_left;

    if (!isPlanet<1>()) {
        long move_right = std::min (remain, long(CONFIG2(ShipMovementSpeed, 1)));
        status[1].r.obj_x -= move_right;
    }
}

bool
VcrPlayerPHost::canStillFight(int side)
{
    return vcr.objects[side].beam_count > 0
        || (vcr.num_fighters[side] > 0 && vcr.objects[side].bay_count > 0)
        || (status[side].r.launched_fighters > 0)
        || (vcr.num_torps[side] > 0);
}

bool
VcrPlayerPHost::playCycle()
{
    if (!canStillFight(0) && !canStillFight(1))
        done = true;
    
    if (done)
        return false;

    ++time;
    if (status[0].r.launch_countdown > 0)
        --status[0].r.launch_countdown;
    if (status[1].r.launch_countdown > 0)
        --status[1].r.launch_countdown;

    fighterRecharge<0>();
    fighterRecharge<1>();
    beamRecharge<0>();
    beamRecharge<1>();
    torpsRecharge<0>();
    torpsRecharge<1>();
    fighterLaunch<0>();
    fighterLaunch<1>();

    if (fighterAttack<0>() || fighterAttack<1>()
        || torpsFire<0>() || torpsFire<1>()
        || beamFire<0>() || beamFire<1>()) {
        done = true;
    } else {
        fighterIntercept();
        fighterMove<0>();
        fighterMove<1>();
        moveObjects();
    }
    return true;
}

VcrPlayerPHost::VcrPlayerPHost(bool av3)
    : phost3(av3)
{ }

bool
VcrPlayerPHost::checkVcrSide(TVcr& vcr, int side)
{
    bool err = false;
    TVcrObject& o = vcr.objects[side];

    if (o.owner <= 0 || o.owner > 12)
        o.owner = 12;

    if (o.beam_type < 0 || o.beam_type > NUM_BEAMS) {
        o.beam_type = o.beam_count = 0;
        err = true;
    }

    if (o.torp_type < 0 || o.torp_type > NUM_TORPS) {
        o.torp_type = vcr.torp_launcher_count[side] = 0;
        err = true;
    }

    if (o.beam_count > VCR_MAX_BEAMS) {
        o.beam_count = VCR_MAX_BEAMS;
        err = true;
    }

    if (vcr.torp_launcher_count[side] > VCR_MAX_TORPS) {
        vcr.torp_launcher_count[side] = VCR_MAX_TORPS;
        err = true;
    }

    if (o.bay_count > VCR_MAX_BAYS) {
        o.bay_count = VCR_MAX_BAYS;
        err = true;
    }

    return err;
}

bool
VcrPlayerPHost::checkVcr(TVcr& vcr)
{
    bool l = checkVcrSide(vcr, 0);
    bool r = checkVcrSide(vcr, 1);
    return l || r;
}

void
VcrPlayerPHost::initVcr(TVcr& avcr)
{
    vcr = avcr;
    if (checkVcr(vcr)) {
        status_word = VCRS_INVALID;
        return;
    }

    /* Initialize Playback */
    status_word = 0;
    time = 0;
    seed = vcr.random_seed;
    done = false;

    status[0].r.obj_x = -29000;
    status[1].r.obj_x = +29000;
    for (int side = 0; side < 2; ++side) {
#ifdef PVCR_INTEGER
        if (config.AllowAlternativeCombat)
            status[side].f.scale = vcr.mass[side] + 1;
        else
            status[side].f.scale = 1;
        status[side].r.shield_scaled  = vcr.shield[side] * int32(status[side].f.scale);
        status[side].r.crew_scaled2   = vcr.objects[side].crew * int32(status[side].f.scale) * 100;
        status[side].r.damage_scaled2 = vcr.objects[side].damage * int32(status[side].f.scale) * 100;
        status[side].f.max_scaled = int32(10000) * status[side].f.scale;
#else
        status[side].r.shield  = vcr.shield[side];
        status[side].r.crew    = vcr.objects[side].crew;
        status[side].r.damage  = vcr.objects[side].damage;
#endif

        if (vcr.objects[side].beam_count > 0) {
            status[side].f.beam_hit_odds = computeBeamHitOdds     (vcr.objects[side].beam_type, side);
            status[side].f.beam_recharge = computeBeamRechargeRate(vcr.objects[side].beam_type, side);
        } else {
            status[side].f.beam_hit_odds = status[side].f.beam_recharge = 0;
        }
        if (vcr.torp_launcher_count[side] > 0) {
            status[side].f.torp_hit_odds = computeTorpHitOdds(vcr.objects[side].torp_type, side);
            status[side].f.torp_recharge = computeTubeRechargeRate(vcr.objects[side].torp_type, side);
        } else {
            status[side].f.torp_hit_odds = status[side].f.torp_recharge = 0;
        }
        if (vcr.objects[side].bay_count > 0) {
            status[side].f.bay_recharge = computeBayRechargeRate(vcr.objects[side].bay_count, side);
        } else {
            status[side].f.bay_recharge = 0;
        }

        // FIXME: this is still missing, but since the original
        // line is bogus (should be p.Shield, not VCR.Shield) and it works,
        // it's probably not needed.
        //   IF NOT IsPlanet(Who) AND IsFreighter(Who) THEN VCR.Shield[Who]:=0;
        // Change line below if needed.

        int charge = (vcr.shield[side] == 100) ? 1000 : 0;
        std::fill_n(status[side].r.beam_status, int(VCR_MAX_BEAMS), charge);
        std::fill_n(status[side].r.torp_status, int(VCR_MAX_TORPS), charge);
        std::fill_n(status[side].r.bay_status,  int(VCR_MAX_BAYS),  0); // !!!
        std::fill_n(status[side].r.fighter_status, int(VCR_MAX_FTRS), 0);
        std::fill_n(status[side].r.fighter_x, int(VCR_MAX_FTRS), 0);  // not needed?
        std::fill_n(status[side].r.fighter_shots_left, int(VCR_MAX_FTRS), 0);  // not needed?
        status[side].r.launch_countdown = status[side].r.launched_fighters = 0;

#ifdef PVCR_INTEGER
        /* Original comparison was:
            if (st.r.damage_scaled2 + 50 * st.f.scale >=
                st.f.damage_limit * 100 * st.f.scale)          */
        status[side].f.damage_limit_scaled =
            raceId(vcr.objects[side].owner) == 2
            ? (150 * 2 - 1) * 50 * status[side].f.scale
            : (100 * 2 - 1) * 50 * status[side].f.scale;
#else
        status[side].f.damage_limit = (raceId(vcr.objects[side].owner) == 2 ? 150 : 100);
#endif

        CONFIG2(ShieldDamageScaling, side)  = CONFIG1(ShieldDamageScaling, side);
        CONFIG2(HullDamageScaling, side)    = CONFIG1(HullDamageScaling, side);
        CONFIG2(CrewKillScaling, side)      = CONFIG1(CrewKillScaling, side);
        CONFIG2(MaxFightersLaunched, side)  = CONFIG1(MaxFightersLaunched, side);
        CONFIG2(StrikesPerFighter, side)    = CONFIG1(StrikesPerFighter, side);
        CONFIG2(BayLaunchInterval, side)    = CONFIG1(BayLaunchInterval, side);
        CONFIG2(FighterMovementSpeed, side) = CONFIG1(FighterMovementSpeed, side);
        CONFIG2(FighterBeamExplosive, side) = CONFIG1(FighterBeamExplosive, side);
        CONFIG2(FighterBeamKill, side)      = CONFIG1(FighterBeamKill, side);
        CONFIG2(FighterFiringRange, side)   = CONFIG1(FighterFiringRange, side);
        CONFIG2(BeamHitFighterRange, side)  = CONFIG1(BeamHitFighterRange, side);
        CONFIG2(BeamHitFighterCharge, side) = CONFIG1(BeamHitFighterCharge, side);
        CONFIG2(BeamFiringRange, side)      = CONFIG1(BeamFiringRange, side);
        CONFIG2(BeamHitShipCharge, side)    = CONFIG1(BeamHitShipCharge, side);
        CONFIG2(TorpFiringRange, side)      = CONFIG1(TorpFiringRange, side);
        CONFIG2(ShipMovementSpeed, side)    = CONFIG1(ShipMovementSpeed, side);
    }
    
    /* pre-compute fighter intercept probabilities */
    if (phost3) {
        int left_f   = (100 - CONFIG1(FighterKillOdds, 0)) * CONFIG1(FighterKillOdds, 1);
        int right_f  = (100 - CONFIG1(FighterKillOdds, 1)) * CONFIG1(FighterKillOdds, 0);
        one_f        = (left_f + right_f) / 100;
        right_probab = right_f / one_f;
        // FIXME: battle.c seems to do `right_probab = left_f / one_f'
    } else {
        /* In PHost 2, combat options were not arrayized. Hence, for
           a valid pconfig, all FighterKillOdds values are the same
           and we can pick any one */
        one_f        = config.FighterKillOdds[0];
        right_probab = 50;
    }

    vis->init();
}

void
VcrPlayerPHost::doneVcr(TVcr& avcr)
{
    if (status_word == VCRS_INVALID) {
        avcr = vcr;
        return;
    }

    ASSERT(done);

    for (int side = 0; side < 2; ++side) {
#ifdef PVCR_INTEGER
        if (config.AllowAlternativeCombat) {
            vcr.objects[side].damage = rdiv(status[side].r.damage_scaled2, status[side].f.scale * 100);
            vcr.objects[side].crew   = rdiv(status[side].r.crew_scaled2,   status[side].f.scale * 100);
            vcr.shield[side]         = rdiv(status[side].r.shield_scaled,  status[side].f.scale);
        } else {
            vcr.objects[side].damage = rdiv(status[side].r.damage_scaled2, 100);
            vcr.objects[side].crew   = rdiv(status[side].r.crew_scaled2, 100);
            vcr.shield[side]         = status[side].r.shield_scaled;
        }
#else
        vcr.objects[side].damage = int(status[side].r.damage + 0.5);
        vcr.objects[side].crew   = int(status[side].r.crew + 0.5);
        vcr.shield[side]         = int(status[side].r.shield + 0.5);
#endif
        if (vcr.objects[side].damage > 100)   // FIXME: take out?
            vcr.objects[side].damage = 100;

#ifdef PVCR_INTEGER
        if (canStillFight(side) && vcr.objects[side].damage < 100)
#else
        if (canStillFight(side) && status[side].r.damage < 99.5)
#endif
        {
            for (int i = 0; i < CONFIG2(MaxFightersLaunched, side); ++i) {
                if (status[side].r.fighter_status[i]) {
                    vis->landFighter(side, i);
                    status[side].r.fighter_status[i] = 0;
                    ++vcr.num_fighters[side];
                    --status[side].r.launched_fighters;
                }
            }
            ASSERT(status[side].r.launched_fighters == 0);
        }
        status[side].r.launched_fighters = 0;
    }

    status_word = 0;
    if (vcr.battle_type) {
        if (vcr.objects[0].damage >= 100 || vcr.objects[0].crew <= 0)
            status_word |= VCRS_LEFT_DESTROYED;
        if (vcr.objects[1].damage >= 100)
            status_word |= VCRS_RIGHT_DESTROYED;
    } else {
        //status[0].f.damage_limit = status[1].f.damage_limit = 100;
        if (vcr.objects[0].damage >= 100)
            status_word |= VCRS_LEFT_DESTROYED;
        else if (vcr.objects[0].crew <= 0) {
            if (vcr.objects[0].damage >= 100)
                status_word |= VCRS_LEFT_DESTROYED;
            else
                status_word |= VCRS_LEFT_CAPTURED;
        }

        if (vcr.objects[1].damage >= 100)
            status_word |= VCRS_RIGHT_DESTROYED;
        else if (vcr.objects[1].crew <= 0) {
            if (vcr.objects[1].damage >= 100)
                status_word |= VCRS_RIGHT_DESTROYED;
            else
                status_word |= VCRS_RIGHT_CAPTURED;
        }
    }
    if (status_word & VCRS_LEFT_DESTROYED)
        vis->killObject(0);
    if (status_word & VCRS_RIGHT_DESTROYED)
        vis->killObject(1);

    avcr = vcr;
}

void
VcrPlayerPHost::playFastForward()
{ }

int
VcrPlayerPHost::getBeamStatus(int side, int id)
{
    return status[side].r.beam_status[id] / 10;
}

int
VcrPlayerPHost::getTorpStatus(int side, int id)
{
    return status[side].r.torp_status[id] / 25;
}

int
VcrPlayerPHost::getTorpCount(int side)
{
    return vcr.num_torps[side];
}

int
VcrPlayerPHost::getFighterCount(int side)
{
    return vcr.num_fighters[side];
}

int
VcrPlayerPHost::getShields(int side)
{
#ifdef PVCR_INTEGER
    return status[side].r.shield_scaled / status[side].f.scale;
#else
    return status[side].r.shield;
#endif
}

int
VcrPlayerPHost::getDamage(int side)
{
#ifdef PVCR_INTEGER
    return status[side].r.damage_scaled2 / status[side].f.scale / 100;
#else
    return status[side].r.damage;
#endif
}

int
VcrPlayerPHost::getCrew(int side)
{
#ifdef PVCR_INTEGER
    return status[side].r.crew_scaled2 / status[side].f.scale / 100;
#else
    return status[side].r.crew;
#endif
}

int
VcrPlayerPHost::getFighterX(int side, int id)
{
    return (status[side].r.fighter_x[id]) / 128 + 320;
}

int
VcrPlayerPHost::getFighterStatus(int side, int id)
{
    return status[side].r.fighter_status[id];
}

int
VcrPlayerPHost::getObjectX(int side)
{
    return (status[side].r.obj_x) / 128 + 320;
}

string_t
VcrPlayerPHost::getPlayerName() const
{
    return string_t("PHost");
}

Ptr<VcrStatus>
VcrPlayerPHost::getStatusToken()
{
    // FIXME
}

void
VcrPlayerPHost::setStatus(Ptr<VcrStatus> status)
{
    // FIXME
}
