/**
  *  \file picedit/picedit.cc
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <SDL.h>
#include "gfx/pixmap.h"
#include "gfx/gfx.h"
#include "ui/ui.h"
#include "picedit/tools.h"
#include "arch/platform.h"
#include "util/nls.h"
#include "gfx/init.h"
#include "fileio/dirs.h"
#include "util/cmdline.h"
#include "ui/uiglobal.h"
#include "util/console.h"
#include "ui/layout.h"
#include "ui/widgets.h"
#include "gfx/palette.h"
#include "cpluslib/assert.h"
#include "util/debug.h"
#include "ui/msgbox.h"
#include "fileio/streams.h"
#include "cpluslib/holder.h"

enum {
    cm_ToolChanged = 1000
};

Ptr<PETool>    tool;
Ptr<GfxPixmap> clip;
string_t       filename;

class PEPainter : public UIWidget {
    Ptr<GfxPixmap> pixmap;
    Ptr<GfxPixmap> backup;
    Ptr<PETool> old_tool;
    GfxPoint old_pos;
    bool mouse_is_down;
    void handleMouse(GfxPoint pt, int b);
    int scale;
 public:
    PEPainter(Ptr<GfxPixmap> pixmap);
    void draw(Ptr<GfxCanvas> can, UpdateInfo& ui);
    bool handleEvent(const UIEvent& e, bool second_pass);
    void getLayoutInfo(LayoutInfo& info);
    void stateChanged(int flags, bool enable);
    void setScale(int scale);
    void setImage(Ptr<GfxPixmap> pix);
    GfxPoint screenToWorld(GfxPoint pt);
    GfxPoint worldToScreen(GfxPoint pt);
};

class PEGroup : public UIGroup {
 public:
    PEGroup(Ptr<UILayoutManager> p) : UIGroup(p) { }
    void drawInterior(Ptr<GfxCanvas> can, UpdateInfo& ui);
};

class PEToolName : public UIWidget {
 public:
    PEToolName();
    void draw(Ptr<GfxCanvas> can, UpdateInfo& ui);
    bool handleEvent(const UIEvent& e, bool second_pass);
    void getLayoutInfo(LayoutInfo& info);
};

class PEToolButton : public UIButton {
    Ptr<PETool> this_tool;
 public:
    PEToolButton(Ptr<PETool> tool);
    void setTool(int);
};

class PEScaleButton : public UIButton {
    Ptr<PEPainter> pt;
    void setScale(int);
 public:
    PEScaleButton(Ptr<PEPainter> pt, int scale, string_t name)
        : UIButton(scale, name, 0, aNothing), pt(pt)
        { sig_fired.add(this, &PEScaleButton::setScale); }
};

class PEColorSelector : public UIWidget {
 public:
    PEColorSelector();
    void draw(Ptr<GfxCanvas> can, UpdateInfo& ui);
    bool handleEvent(const UIEvent& e, bool second_pass);
    void getLayoutInfo(LayoutInfo& info);
};

PEPainter::PEPainter(Ptr<GfxPixmap> pixmap)
    : pixmap(pixmap), mouse_is_down(false), scale(0)
{ }

#include "picedit/blit.inc"

void
PEPainter::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    switch (scale) {
     case 0:
        pixmap->blit(can, extent.x, extent.y, GfxRect(0, 0, extent.w, extent.h));
        break;
     case 1:
        {
            SDL_Surface* sfc = can->getSurface();
            GFX_MODE_SWITCH(blit2, (sfc, pixmap->getSurface(), extent));
        }
        break;
     case 2:
        {
            SDL_Surface* sfc = can->getSurface();
            GFX_MODE_SWITCH(blit4, (sfc, pixmap->getSurface(), extent));
        }
        break;
    }
    int r = pixmap->getWidth() << scale;
    int b = pixmap->getHeight() << scale;
    if (r < extent.w) {
        drawSolidBar(can, GfxRect(extent.x + r, extent.y, extent.w - r, extent.h), standard_colors[COLOR_BLACK]);
        if (b < extent.h)
            drawSolidBar(can, GfxRect(extent.x, extent.y + b, r, extent.y - extent.h), standard_colors[COLOR_BLACK]);
    } else {
        drawSolidBar(can, GfxRect(extent.x, extent.y + b, extent.w, extent.y - extent.h), standard_colors[COLOR_BLACK]);
    }
    ui.update(extent);
}

void
PEPainter::setScale(int sc)
{
    if (sc != scale) {
        scale = sc;
        drawWidget(false);
    }
}

GfxPoint
PEPainter::screenToWorld(GfxPoint pt)
{
    return GfxPoint((pt.x - extent.x) >> scale, (pt.y - extent.y) >> scale);
}

GfxPoint
PEPainter::worldToScreen(GfxPoint pt)
{
    return GfxPoint((pt.x + extent.x) << scale, (pt.y + extent.y) << scale);
}

bool
PEPainter::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Mouse && extent.contains(e.mouse.x, e.mouse.y)) {
        select();
        handleMouse(screenToWorld(GfxPoint(e.mouse.x, e.mouse.y)), e.mouse.b);
        return true;
    } else if (e.type == ev_Command && e.cmd.code == cm_ToolChanged && mouse_is_down) {
        select();
        handleMouse(old_pos, 0);
        return false;
    } else if (e.type == ev_Key && e.key.code == SDLK_ESCAPE && mouse_is_down) {
        select();
        if (old_tool->needsPreview()) {
            backup->blit(pixmap, 0, 0);
            drawWidget(false);
        }
        mouse_is_down = false;
        return true;
    }
    return UIWidget::handleEvent(e, second_pass);
}

void
PEPainter::handleMouse(GfxPoint pt, int b)
{
    if (b) {
        if (!mouse_is_down) {
            /* mouse pressed. */
            old_tool = tool;
            if (old_tool->needsPreview()) {
                if (!backup)
                    backup = pixmap->clone();
                else
                    pixmap->blit(backup, 0, 0);
            }
            if (b & 1)
                PETool::setSwapColors(0);
            else
                PETool::setSwapColors(1);
            old_tool->click(pixmap, pt);
            drawWidget(false);
        } else {
            /* mouse still pressed. */
            if (old_tool->needsPreview())
                backup->blit(pixmap, 0, 0);
            old_tool->drag(pixmap, pt);
            drawWidget(false);
        }
    } else {
        if (mouse_is_down) {
            /* mouse released. */
            if (old_tool->needsPreview())
                backup->blit(pixmap, 0, 0);
            old_tool->release(pixmap, pt);
            drawWidget(false);
        }
    }
    mouse_is_down = b != 0;
    old_pos       = pt;
}

void
PEPainter::stateChanged(int flags, bool enable)
{
    if ((flags & st_Selected) && !enable && mouse_is_down)
        handleMouse(old_pos, 0);
    UIWidget::stateChanged(flags, enable);
}

void
PEPainter::setImage(Ptr<GfxPixmap> pix)
{
    if (pix) {
        if (mouse_is_down)
            handleMouse(old_pos, 0);
        pixmap = pix;
        backup = 0;
        drawWidget(false);
    }
}

void
PEPainter::getLayoutInfo(LayoutInfo& i)
{
    i.flags = li_Flexible;
    i.min_size = i.pref_size = GfxPoint(pixmap->getWidth(), pixmap->getHeight());
}

void
PEGroup::drawInterior(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    drawSolidBar(can, extent, standard_colors[COLOR_BLACK]);
    ui.update(extent);
}

PEToolName::PEToolName()
{ }

void
PEToolName::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    GfxParameters p;
    p.color = standard_colors[COLOR_WHITE];
    p.bgcolor = standard_colors[COLOR_BLACK];
    outTextF(can, p, fonts[FONT_NORMAL], extent.x, extent.y, extent.w, tool->getName());
    ui.update(extent);
}

bool
PEToolName::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Command && e.cmd.code == cm_ToolChanged)
        drawWidget(false);
    return UIWidget::handleEvent(e, second_pass);
}

void
PEToolName::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.min_size = info.pref_size = GfxPoint(12*font_ems[FONT_NORMAL],
                                              font_heights[FONT_NORMAL]);
}

PEToolButton::PEToolButton(Ptr<PETool> tool)
    : UIButton(0, tool->getName(), 0, aNothing), this_tool(tool)
{
    sig_fired.add(this, &PEToolButton::setTool);
}

void
PEToolButton::setTool(int)
{
    if (tool != this_tool) {
        tool = this_tool;
        putCommand(cm_ToolChanged, 0);
    }
}

void
PEScaleButton::setScale(int)
{
    pt->setScale(getId());
}

PEColorSelector::PEColorSelector()
{ }

void
PEColorSelector::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    for (int x = 0; x < 16; ++x) {
        for (int y = 0; y < 16; ++y) {
            int c = 16*y+x;
            GfxRect r;
            r.x = extent.x + x * extent.w / 16;
            r.y = extent.y + y * extent.h / 16;
            r.w = extent.x + (x+1) * extent.w / 16 - r.x;
            r.h = extent.y + (y+1) * extent.h / 16 - r.y;
            drawSolidBar(can, r, standard_colors[c]);
            if (PETool::getColorN(0) == c) {
                drawSolidBar(can, GfxRect(r.x, r.y, r.w, 1), standard_colors[COLOR_WHITE]);
                drawSolidBar(can, GfxRect(r.x, r.y, 1, r.h), standard_colors[COLOR_WHITE]);
            }
            if (PETool::getColorN(1) == c) {
                drawSolidBar(can, GfxRect(r.x, r.y+r.h-1, r.w, 1), standard_colors[COLOR_WHITE]);
                drawSolidBar(can, GfxRect(r.x+r.w-1, r.y, 1, r.h), standard_colors[COLOR_WHITE]);
            }
        }
    }
    ui.update(extent);
}

bool
PEColorSelector::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Mouse && e.mouse.b && extent.contains(e.mouse.x, e.mouse.y)) {
        int x = (e.mouse.x - extent.x) * 16 / extent.w;
        int y = (e.mouse.y - extent.y) * 16 / extent.h;
        int c = 16*y + x;
        int i = (e.mouse.b & 1) ? 0 : 1;
        if (PETool::getColorN(i) != c) {
            PETool::setColorN(i, c);
            drawWidget(false);
        }
        return true;
    }
    return UIWidget::handleEvent(e, second_pass);
}

void
PEColorSelector::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.min_size = info.pref_size = GfxPoint(160, 160);
}

/********************************** Main *********************************/

Ptr<PEPainter> the_painter;

static void
load(string_t fn)
{
    Ptr<Stream> s = Stream::openFileNT(fn, Stream::C_READ);
    if (s) {
        Ptr<GfxPixmap> pix = loadImage(s);
        if (pix) {
            the_painter->setImage(pix->convertToScreenFormat());
            filename = fn;
        }
    } else {
        messageBox("Unable to open file `" + fn + "'", "Error");
    }
}

static void
load(int=0)
{
    string_t fn = filename;
    if (UIInputLine::run(fn, string_t::npos, "Load file",
                         "File name:", 40, 0))
        load(fn);
}

static void
quit(int)
{
    ObjectHolder h;
    UIWindow w("Menu", UIWindow::BLUE_WINDOW, COLOR_GRAY);
    w.setLayout(new UIVBoxLayout(5));
    w.add(h.add(new UIButton(1, "Load", 'l', UIButton::aStop)));
    w.add(h.add(new UIButton(2, "Save", 's', UIButton::aStop)));
    w.add(h.add(new UIButton(3, "New",  'n', UIButton::aStop)));
    w.add(h.add(new UIButton(4, "Quit", 'q', UIButton::aStop)));
    w.add(h.add(new UIButton(0, "Cancel", SDLK_ESCAPE, UIButton::aStop)));
    w.pack();
    GfxRect r = w.getExtent();
    r.moveTo(ui_root->getExtent().w - r.w, ui_root->getExtent().h - r.h);
    w.setExtent(r);
    ui_root->add(w);
    int rv = ui_root->run();
    ui_root->close(w);
    switch (rv) {
     case 1:
        load(0);
        break;
     case 2:
     case 3:
        break;
     case 4:
        ui_root->stop(0);
        break;
     case 0:
        break;
    }
}

class PEArgs : public StandardCommandLineParser {
    bool had_name;
    string_t name;
 public:
    PEArgs() : had_name(0) { }
    void handleNonoption(const string_t& x);
    void handleError(const string_t& x);
    bool hadName() const { return had_name; }
    const string_t& getName() const { return name; }
};

void
PEArgs::handleNonoption(const string_t& x)
{
    if (had_name) {
        handleError("too many parameters");
    } else {
        had_name = true;
        name = x;
    }
}

void
PEArgs::handleError(const string_t& x)
{
    console.write(LOG_ERROR, x);
    Platform::showOutput();
    std::exit(1);
}

int main(int argc, char** argv)
{
    Platform::initOutput();
    initNLS();
    initDebug();
    PEArgs parser;
    Platform::initOptions(parser);
    initGraphicsOptions(parser);
    parser.parse(argv);

    SDL_Surface* sfc = initGraphics(0, "PCC II Picture Editor");
    if (!sfc) {
        Platform::showOutput();
        return 1;
    }

    try {
        initUIGlobals(new GfxSurface(sfc, false),
                      Directory::openDir(Platform::getResourceDirectory()));

        tool = new PEPaint();
        PETool::setColorN(0, COLOR_WHITE);
        PETool::setColorN(1, COLOR_BLACK);

        /* don't care about object lifetime here. They all live
           as long as this program lives. */
        the_painter = new PEPainter(new GfxPixmap(105, 93, ui_screen->getSurface()->format));
        Ptr<UIGroup> main = new PEGroup(new UIHBoxLayout(5));
        Ptr<UIGroup> bs = new UIGroup(new UIVBoxLayout(5));
        bs->add(*new PEToolButton(tool));
        bs->add(*new PEToolButton(new PELine()));
        bs->add(*new PEToolButton(new PERect()));
        bs->add(*new PEToolButton(new PEBar()));
        bs->add(*new PEToolName());
        bs->add(*new PEScaleButton(the_painter, 0, "Normal"));
        bs->add(*new PEScaleButton(the_painter, 1, "Double"));
        bs->add(*new PEScaleButton(the_painter, 2, "Large"));
        bs->add(*new PEColorSelector());
        bs->add(*new UISpacer());
        Ptr<UIButton> q = new UIButton(0, "Menu", SDLK_ESCAPE, UIButton::aNothing);
        q->sig_fired.add(quit);
        bs->add(*q);
        main->add(*the_painter);
        main->add(*bs);
        main->setExtent(GfxRect(ui_screen->getSurface()));
        main->doLayout();

        ui_root->add(*main);

        if (parser.hadName())
            load(parser.getName());
        
        ui_root->run();
    }
    catch(std::exception& e) {
        console.write(LOG_ERROR, string_t(_("Error during initialisation: ")) + e.what());
        Platform::showOutput();
    }
    return 0;
}
