/**
  *  \file resmgr/resmgr.cc
  *  \brief Resource Manager -- Interface Implementation
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "resmgr/resmgr.h"
#include "cpluslib/assert.h"
#include "fileio/iostr.h"
#include "util/nls.h"
#include "util/console.h"
#include "cpluslib/format.h"
#include "arch/platform.h"

ResProvider::ResProvider()
{ }

ResProvider::~ResProvider()
{ }

Ptr<Stream>
ResProvider::getEntry(ResID /*id*/, ResClass /*rc*/)
{
    return 0;
}

Ptr<Stream>
ResProvider::getNamedEntry(string_t /*name*/)
{
    return 0;
}

/****************************** ResDirectory *****************************/

ResDirectory::AliasMap ResDirectory::aliases;

ResDirectory::ResDirectory(Ptr<Directory> dir)
    : dir(dir)
{
    Ptr<Stream> s = dir->openFileNT("INDEX", Stream::C_READ);
    if (s) {
        StandardStream ss(s);
        loadAliases(ss);
    }
}

ResDirectory::~ResDirectory()
{ }

void
ResDirectory::loadAliases(std::istream& in)
{
    string_t line, l, r;
    int line_nr = 0;
    while (getline(in, line)) {
        ++line_nr;

        /* strip comment */
        string_t::size_type p = line.find_first_of(";#");
        if (p != line.npos)
            line.erase(p);
        line = strTrim(line);
        if (line.empty())
            continue;

        /* must be "normal" line */
        if (strSplit(line, l, r, "=")) {
            aliases[strTrim(l)] = strTrim(r);
        } else {
            console.write(LOG_ERROR, format(_("%s:%d: file format error -- line ignored")));
        }
    }
}

static const char* const gfx_suffix_list[] = { ".cd", ".gfx", ".bmp", ".jpg", ".png", 0 };
static const char* const sfx_suffix_list[] = { ".wav", 0 };
static const char* const fnt_suffix_list[] = { ".fnt", 0 };
static const char* const no_suffix_list[] = { "", 0 };

Ptr<Stream>
ResDirectory::getNamedEntry(string_t name)
{
    AliasMap::iterator i = aliases.find(name);
    if (i != aliases.end())
        return dir->openFileNT(i->second, Stream::C_READ);

    /* "legacy" way */
    const char*const* suffixes = no_suffix_list;
    if (name.compare(0, 2, "g.", 2) == 0) {
        name.erase(0, 2);
        suffixes = gfx_suffix_list;
    } else if (name.compare(0, 2, "s.", 2) == 0) {
        name.erase(0, 2);
        suffixes = sfx_suffix_list;
    } else if (name.compare(0, 2, "f.", 2) == 0) {
        name.erase(0, 2);
        suffixes = fnt_suffix_list;
    }

    string_t::size_type n = 0;
    while ((n = name.find('.', n)) != string_t::npos)
        name[n] = Platform::PATH_SEPARATOR;

    while (*suffixes) {
        Ptr<Stream> x = dir->openFileNT(name + *suffixes, Stream::C_READ);
        if (x)
            return x;
        ++suffixes;
    }
    return 0;
}

/******************************* ResManager ******************************/

ResManager ResManager::instance;

ResManager::ResManager()
{ }

ResManager::~ResManager()
{ }

void
ResManager::addProvider(Ptr<ResProvider> p)
{
    ASSERT(!p->next);
    p->next = first;
    first = p;
}

Ptr<Stream>
ResManager::getEntry(ResID id, ResClass rc)
{
    Ptr<ResProvider> f = first;
    while (f) {
        if (Ptr<Stream> s = f->getEntry(id, rc))
            return s;
        f = f->next;
    }
    return 0;
}

Ptr<Stream>
ResManager::getNamedEntry(string_t name)
{
    Ptr<ResProvider> f = first;
    while (f) {
        if (Ptr<Stream> s = f->getNamedEntry(name))
            return s;
        f = f->next;
    }
    return 0;
}
