/**
  *  \file resmgr/resmgrcf.cc
  *  \brief Resource Manager -- Configuration
  *
  *  This provides a function to read `cc-res.cfg' style files, and
  *  load them into a resource manager.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "resmgr/resmgr.h"
#include "resmgr/res-cc1.h"
#include "resmgr/res-wp.h"
#include "cpluslib/string.h"
#include "util/except.h"
#include "util/console.h"
#include "util/nls.h"
#include "resmgr/res-wpvcr.h"
#include "cpluslib/format.h"
#include "fileio/dirs.h"

namespace {
    typedef Ptr<ResProvider> (*ResProviderGen)(string_t s);

    struct ResProviderElem {
        const char* name;
        ResProviderGen gen;
    };

    template<class T>
    Ptr<ResProvider>
    genResProvider(string_t s) {
        return new T(s);
    }

    Ptr<ResProvider>
    genResDirectory(string_t s) {
        return new ResDirectory(Directory::openDir(s));
    }
};

static const ResProviderElem providers[] = {
    { "res",   &genResProvider<ResProviderResFile> },
    { "wp",    &genResProvider<ResProviderWinplanBMP> },
    { "wpvcr", &genResProvider<ResProviderWPVCR> },
    { "dir",   &genResDirectory }
};

/** Load `cc-res.cfg'-style file into a resource manager.
    \returns true if everything okay, false if there were errors.
    \param mgr resource manager
    \param in  file to read configuration from */
bool
loadResourceConfig(ResManager& mgr, std::istream& in)
{
    string_t s;
    bool rv = true;
    while (getline(in, s)) {
        const ResProviderElem* ele = &providers[0];
        string_t::size_type n = s.find(';');
        if (n != s.npos)
            s.erase(n);
        s = strTrim(s);
        if (s.empty())
            continue;

        n = s.find(':');
        if (n != s.npos) {
            string_t type = strLCase(strRTrim(s.substr(0, n)));
            for (unsigned i = 0; i < countof(providers); ++i) {
                if (type == providers[i].name) {
                    ele = &providers[i];
                    s = strLTrim(s.substr(n+1));
                    break;
                }
            }
        }

        try {
            mgr.addProvider(ele->gen(s));
            console.write(LOG_INFO, format(_("Added resource `%s'."), s));
        }
        catch(std::exception& e) {
            console.write(LOG_ERROR, s + ": " + e.what());
            rv = false;
        }
    }
    return rv;
}
