#!/usr/bin/perl -w
#
#  Simple "man" like interface to doxygen generated docs
#
#  "dman foo" -> show page for foo
#  "dman -k foo" -> show everything that matches foo
#
#  "foo" can be a qualified name of the form "class::member"
#  or "file.cc::member".
#
#  This needs a Tagfile; ad-hoc parsing is applied to create a
#  Whatis file which contains the relevant information (this
#  sort of "caching" gives a speedup of 3).
#

use strict;
use vars qw($opt_k $opt_keyword);

$opt_k = 0;
foreach (@ARGV) {
    if ($_ eq '-k') {
        $opt_k = 1;
    } elsif (! defined $opt_keyword) {
        $opt_keyword = $_;
    } else {
        print STDERR "usage: $0 [-k] keyword\n";
        exit 1;
    }
}
if (! defined $opt_keyword) {
    print STDERR "you must specify a keyword\n";
    exit 1;
}

my $dir = locate_docs();

my %possibilities;
open WHATIS, "< $dir/Whatis" or die "$dir/Whatis: $!\n";
while (<WHATIS>) {
    my ($mem_name, $url, @rest) = split / - /, $_;
    if ($mem_name =~ /\b\Q$opt_keyword\E\b/i) {
        $possibilities{$mem_name} = "$dir/html/$url";
    }
}
close WHATIS;

if ($opt_k) {
    foreach (sort keys %possibilities) {
        print "$_ - $possibilities{$_}\n";
    }
} else {
    if (defined $opt_keyword && exists $possibilities{$opt_keyword}) {
        show_page ($possibilities{$opt_keyword});
    } elsif (1 == keys %possibilities) {
        show_page (values %possibilities);
    } elsif (0 == keys %possibilities) {
        print STDERR "no match", (defined $opt_keyword ? " for $opt_keyword" : ""), "\n";
    } else {
        my %copy = %possibilities;
        if ($opt_keyword !~ /::/) {
            # DWIM mode. First remove all headers.
            foreach (keys %copy) {
                if ($_ =~ /\.h::/) {
                    delete $copy{$_};
                }
            }
            # Still not all? Remove class members (i.e. we want
            # `dman foo' access a global function `foo', not a
            # class member `Foo::foo')
            if (keys(%copy) > 1) {
                foreach (keys %copy) {
                    if ($_ !~ /\./) {
                        delete $copy{$_};
                    }
                }
            }
            if (1 == keys %copy) {
                show_page (values %copy);
                exit 0;
            }
        }
        print STDERR "ambiguous:\n";
        foreach (sort keys %possibilities) {
            print "$_\n";
        }
    }
}
exit 0;

sub show_page {
    exec "links", @_;
    die "unable to execute browser: $!\n";
}

sub locate_docs {
    foreach (qw(doxy_docs ../doxy_docs)) {
        if (open TAGFILE, "< $_/Tagfile") {
            $dir = "$_";
            last;
        }
    }
    die "unable to find documentation\n"
        if not defined $dir;

    # Compile "Whatis database"
    my $tagfile_mtime = (stat TAGFILE)[9];
    my $whatis_mtime = (stat "$dir/Whatis")[9];
    if (!defined $whatis_mtime || $whatis_mtime < $tagfile_mtime) {
        print STDERR "updating database...\n";
        open WHATIS, "> $dir/Whatis" or die "$dir/Whatis: $!\n";

        my ($t_kind, $t_name, $t_filename, $t_member_t, $t_mem_name, $t_mem_anchor, $t_mem_kind);
        $t_member_t = 0;
        while (<TAGFILE>) {
            if (m{<compound kind="(.*?)"}) {
                $t_kind = $1;
            } elsif (m{<name>(.*?)</name>}) {
                if ($t_member_t) {
                    $t_mem_name = $1;
                } else {
                    $t_name = $1;
                }
            } elsif (m{</compound>}) {
                $t_kind = $t_name = $t_filename = "";
            } elsif (m{<filename>(.*?)</filename>}) {
                $t_filename = $1;
                print WHATIS "$t_name - $t_filename - $t_kind\n";
            } elsif (m{<anchor>(.*?)</anchor>}) {
                $t_mem_anchor = $1;
            } elsif (m{<member kind="(.*?)"}) {
                $t_member_t = 1;
                $t_mem_kind = $1;
            } elsif (m{</member}) {
                $t_member_t = 0;
                print WHATIS "${t_name}::$t_mem_name - $t_filename#$t_mem_anchor - $t_mem_kind\n";
                $t_mem_name = $t_mem_anchor = $t_mem_kind = "";
            }
        }
        close WHATIS;
    }
    close TAGFILE;
    $dir;
}
