#!/usr/bin/perl -w
#
#  Compile message file into binary format
#
#  Why not use GNU gettext? First: GPL. Second: we need a solution
#  that also works on Windows etc., with data files portable to all
#  OSes. GNU gettext is endian-dependant, AFAICT.
#

use strict;
use vars qw{$line $lineno $token $token_arg %strings $input $output};

foreach (@ARGV) {
    if (!defined $input) {
        $input = $_;
    } elsif (!defined $output) {
        $output = $_;
    } else {
        &fail ("usage: $0 input output");
    }
}
if (defined $input && !defined $output && $input !~ /[^a-z]/i) {
    $output = "resource/$input.lang";
    $input = "po/$input.po";
}
if (!defined $output) {
    &fail ("usage: $0 input output\n   or: $0 languagecode");
}

open INPUT, "< $input" or &fail ("can't open $input: $!");

################################ Read Input ###############################

$line = "";
$lineno = 0;
%strings = ();

&lex ();
while (defined $token) {
    &expect ('msgid');
    my $key = &expect_str ();
    &expect ('msgstr');
    my $value = &expect_str ();
    &warn ("Duplicate message id `$key'") if exists $strings{$key};
    next if $value eq "" || $value eq $key;
    $strings{$key} = $value;
}

############################# Generate Output #############################

my $in     = "";
my $out    = "";
my $inptr  = "";
my $outptr = "";
foreach (sort keys %strings) {
    my ($instr, $outstr) = ("$_\0", "$strings{$_}\0");
    $inptr .= pack 'VV', length ($in), length ($instr);
    $outptr .= pack 'VV', length ($out), length ($outstr);
    $in .= $instr;
    $out .= $outstr;
}
open OUT, "> $output" or &fail ("can't create $output: $!");
binmode OUT;
print OUT "CClang0\032",
    pack ("VVVVVVV", 
          length ($inptr) / 8,  # number of entries
          36,                   # input ptr pos
          36 + length ($inptr), # output ptr pos
          36 + length ($inptr) + length ($outptr),    # input data pos
          length ($in),         # input data length
          36 + length ($inptr) + length ($outptr) + length ($in),
          length ($out));
print OUT $inptr, $outptr, $in, $out;
close OUT;
exit 0;

############################### Subroutines ###############################

sub expect_str {
    my $rv;
    while (defined $token && $token eq '"') {
        $rv .= $token_arg;
        &lex ();
    }
    $rv;
}

sub expect {
    my $rv;
    foreach (@_) {
        $rv = $token_arg;
        if ($token ne $_) {
            &fail ("expected `$_'");
        }
        &lex ();
    }
    $rv;
}

sub lex {
    while (1) {
        if ($line eq "") {
            $line = <INPUT>;
            return ($token = undef) if !defined $line;
            ++$lineno;
            chomp $line;
            next;
        }
        if ($line =~ s/^\s+//) {
            next;
        } elsif ($line =~ s/^\#.*//) {
            next;
        } elsif ($line =~ s/^(\w+)//) {
            ($token, $token_arg) = ($1, "");
            return $token;
        } elsif ($line =~ s/^\"(([^\"\\]|\\.)*)\"//) {
            my $text = $1;
            $text =~ s/\\(.)/&unescape ($1)/eg;
            ($token, $token_arg) = ('"', $text);
            return $token;
        } else {
            &fail ("syntax error (unable to parse)");
        }
    }
}

sub unescape {
    my $what = shift;
    $what eq "n" and return "\n";
    $what eq "\\" and return "\\";
    &warn ("unknown escape sequence `\\$what'");
    return "?";
}

sub warn {
    print STDERR "$input:$lineno: warning: @_\n";
}

sub fail {
    if (defined $lineno) {
        print STDERR "$input:$lineno: @_\n";
    } else {
        print STDERR "$0:@_\n";
    }
    exit 1;
}
