/**
  *  \file sound/sound-mix.cc
  *  \brief Mixing Code
  *
  *  This code does (most of) the actual sound mixing.
  *
  *  This is a pretty simple mixer, still it is efficient enough for
  *  my taste. Hey, this is not a 12 MHz AT :-)
  *
  *  CPU usage on Pentium 200 MHz:
  *  - idle, mixer not running (SDL_PauseAudio(1)): <1%
  *  - idle, mixer running (no active channel): <1%
  *  - 3 active sounds: <3%
  *  - 32 active sounds (maximum): 25%
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <SDL_audio.h>
#include "sound/sound.h"
#include "util/except.h"
#include "util/console.h"
#include "cpluslib/types.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "cpluslib/assert.h"

/** \class SfxSampleData
    \brief One Sample

    Users don't use SfxSampleData objects; they operate with SfxSample
    handles instead. This makes it clear that all SfxSampleData
    objects are owned by the sound.cc module. */

/** Create sample from file. */
SfxSampleData::SfxSampleData(Ptr<Stream> file)
    : sound_data(0), sound_size(0)
{
    loadWav(file);
}

SfxSampleData::~SfxSampleData()
{
    delete[] sound_data;
}

/** Load sample from wav file. */
void
SfxSampleData::loadWav(Stream* s)
{
    uint8*         wav_buffer;
    Uint32         wav_size;
    SDL_AudioSpec  wav_spec;
    if (! SDL_LoadWAV_RW(s, 0, &wav_spec, &wav_buffer, &wav_size))
        throw FileFormatException(format(_("Unable to load sample `%s': %s"),
                                         SDL_GetError(),
                                         s->getName()));

//     console.write(LOG_DEBUG, format(_("Loaded `%s': %d bytes, %d channels, %d Hz, %s"))
//                   << s->getName()
//                   << wav_size
//                   << wav_spec.channels
//                   << wav_spec.freq
//                   << getAudioFormatName(wav_spec.format));
    
    SDL_AudioCVT converter;
    if (SDL_BuildAudioCVT(&converter, 
                          wav_spec.format, wav_spec.channels, wav_spec.freq,
                          sound_hardware.format, 1, sound_hardware.freq) < 0)
        throw FileFormatException(format(_("Unable to convert sample `%s': %s"),
                                         SDL_GetError(),
                                         s->getName()));
    try {
        if (converter.len_mult == 1) {
            /* convert in-place in the wav buffer, then copy out */
            converter.buf = wav_buffer;
            converter.len = wav_size;
            SDL_ConvertAudio(&converter);   // can't fail
            
            sound_size = converter.len_cvt;
            sound_data = new uint8[sound_size];
            std::copy(wav_buffer, wav_buffer + sound_size, sound_data);
            SDL_FreeWAV(wav_buffer);
        } else {
            /* length increases. First make new buffer, then
               convert there. */
            //ASSERT(converter.len_ratio > 1.0);
            sound_size = uint32(wav_size * converter.len_mult);
            sound_data = new uint8[sound_size];
            std::copy(wav_buffer, wav_buffer + wav_size, sound_data);
            SDL_FreeWAV(wav_buffer);

            converter.buf = sound_data;
            converter.len = wav_size;
            SDL_ConvertAudio(&converter);

            /* It can happen that the buffer is larger than the final .wav.
               Shrink it again to save memory. */
            if (converter.len_cvt < sound_size) {
                sound_size = converter.len_cvt;
                try {
                    uint8* tmp = new uint8[sound_size];
                    std::copy(sound_data, sound_data + sound_size, tmp);
                    delete[] sound_data;
                    sound_data = tmp;
                }
                catch(...) { }
            }
        }
    }
    catch(...) {
        SDL_FreeWAV(wav_buffer);
        throw;
    }
}

#include "sound/add-tbl.inc"

/* add with saturation, unsigned */
static inline uint8 addsu(uint8 a, uint8 b)
{
    return uint8_add_table[a + b];
}

/* add with saturation, signed */
static inline int8 addss(int8 a, int8 b)
{
    return int8_add_table[a + b + 256];
}

static inline uint8 adjustVolume(uint8 volume, uint8 mult)
{
    return volume * mult / 255;
}

struct SfxU8Traits {
    typedef uint8 type;
    static inline void mix(type* dest, type sample, uint8 vol) {
        *dest = addsu(*dest, (sample * vol) >> 8);
    }
};

struct SfxS8Traits {
    typedef uint8 type;
    static inline void mix(type* dest, type sample, uint8 vol) {
        *dest = addss(*dest, (sample * vol) >> 8);
    }
};

struct SfxU16Traits {
    typedef uint16 type;
    static inline void mix(uint16* dest, uint16 sample, uint8 vol) {
        int32 sum = int32(*dest) + (int32(sample) - 0x8000) * vol / 256;
        if (sum < 0)
            *dest = 0;
        else if (sum > 0xFFFF)
            *dest = 0xFFFF;
        else
            *dest = sum;
    }
};

struct SfxS16Traits {
    typedef int16 type;
    static inline void mix(int16* dest, int16 sample, uint8 vol) {
        int32 sum = int32(*dest) + (int32(sample) * vol / 256);
        if (sum < -0x8000)
            *dest = -0x8000;
        else if (sum > 0x7FFF)
            *dest = 0x7FFF;
        else
            *dest = sum;
    }
};

/** Mix a mono sample into a stereo stream, with panning.
    \param dest_buf_u    output mixing buffer
    \param size          requested mixing buffer size
    \param lvol,rvol     left/right channel volume
    \param sound_buf     sample data
    \param sound_size    sample data size (bytes)
    \param sound_pos     opaque position indicator [sample number] */
template<class Traits>
void
mixOneTwo(void* dest_buf_u, uint8 lvol, uint8 rvol,
          const void* sound_buf, uint32 sound_pos, int nsamples)
{
    typedef typename Traits::type type;
    type* dest_buf = static_cast<type*>(dest_buf_u);
    const type* src_buf = static_cast<const type*>(sound_buf) + sound_pos;
    while (nsamples--) {
        type sample = *src_buf++;
        Traits::mix(dest_buf++, sample, lvol);
        Traits::mix(dest_buf++, sample, rvol);
    }
}

/** Mix a mono sample into another mono stream.
    \param dest_buf_u    output mixing buffer
    \param size          requested mixing buffer size
    \param volume        volume
    \param sound_buf     sample data
    \param sound_size    sample data size (bytes)
    \param sound_pos     opaque position indicator [sample number] */
template<class Traits>
void
mixOneOne(void* dest_buf_u,
          uint8 vol,
          const void* sound_buf, uint32 sound_pos, int nsamples)
{
    typedef typename Traits::type type;
    type* dest_buf = static_cast<type*>(dest_buf_u);
    const type* src_buf = static_cast<const type*>(sound_buf) + sound_pos;

    while (nsamples--) {
        Traits::mix(dest_buf++, *src_buf++, vol);
    }
}

/** Mix this sample into an audio stream.
    \param dest_buf   output audio buffer (sample will be added to it).
    \param size       size of audio buffer in bytes
    \param volume     volume (0..255)
    \param pan        panning (0=left, 255=right)
    \param pos        current position

    The current position is a value internal to this function; we
    currently interpret it as a number of samples already played
    (i.e. half the number of bytes already played for a 16-bit sound).

    \returns new position, 0 if sample finished playing. */
uint32
SfxSampleData::mix(uint8* dest_buf, int size, uint8 volume, uint8 pan, uint32 pos)
{
    // we have to mix:
    //   (U8/S8/U16/S16/U16S/S16S) x (one->two, one->one)
    //   12 mixing functions

    unsigned bytes_per_sample = (sound_hardware.format == AUDIO_U8 || sound_hardware.format == AUDIO_S8) ? 1 : 2;

    if (sound_hardware.channels == 1) {
        /* one->one */
        uint32 now = std::min(uint32(size) / bytes_per_sample, sound_size / bytes_per_sample - pos);
        switch (sound_hardware.format) {
         case AUDIO_U8:
            mixOneOne<SfxU8Traits>(dest_buf, volume, sound_data, pos, now);
            break;
         case AUDIO_S8:
            mixOneOne<SfxS8Traits>(dest_buf, volume, sound_data, pos, now);
            break;
         case AUDIO_U16SYS:
            mixOneOne<SfxU16Traits>(dest_buf, volume, sound_data, pos, now);
            break;
         case AUDIO_S16SYS:
            mixOneOne<SfxS16Traits>(dest_buf, volume, sound_data, pos, now);
            break;
        }
        pos += now;
        if (pos*bytes_per_sample >= sound_size)
            return 0;
        else
            return pos;
    } else {
        /* one->two */
        uint32 now = std::min(uint32(size) / (2*bytes_per_sample), sound_size/bytes_per_sample - pos);
        uint8 lvol = adjustVolume(volume, pan);
        uint8 rvol = adjustVolume(volume, 255-pan);
        switch (sound_hardware.format) {
         case AUDIO_U8:
            mixOneTwo<SfxU8Traits>(dest_buf, lvol, rvol, sound_data, pos, now);
            break;
         case AUDIO_S8:
            mixOneTwo<SfxS8Traits>(dest_buf, lvol, rvol, sound_data, pos, now);
            break;
         case AUDIO_U16SYS:
            mixOneTwo<SfxU16Traits>(dest_buf, lvol, rvol, sound_data, pos, now);
            break;
         case AUDIO_S16SYS:
            mixOneTwo<SfxS16Traits>(dest_buf, lvol, rvol, sound_data, pos, now);
            break;
        }
        pos += now;
        if (pos*bytes_per_sample >= sound_size)
            return 0;
        else
            return pos;
    }
    
    return 0;
}
