/**
  *  \file sound/sound.cc
  *  \brief Sound High-level API
  *
  *  This module provides sound initialisation and shutdown, and a means
  *  of handling samples.
  *
  *  To request a sample, use requestSound(). This will return an opaque
  *  SfxSample object, which see.
  *
  *  When the sound is reconfigured, all samples get outdated. They will
  *  simply refuse to respond to play(). You can check that and reload,
  *  as in
  *  \code
  *     if (s.isOutdated())
  *        s = requestSound("name");
  *     s.play(...);
  *  \endcode
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <SDL.h>
#include "sound/sound.h"
#include "util/except.h"
#include "util/console.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "cpluslib/types.h"
#include "cpluslib/assert.h"
#include "resmgr/resmgr.h"

#undef AUDIO_U16SWAB
#undef AUDIO_S16SWAB
#if AUDIO_U16SYS == AUDIO_U16LSB
#  define AUDIO_U16SWAB AUDIO_U16MSB
#  define AUDIO_S16SWAB AUDIO_S16MSB
#else
#  define AUDIO_U16SWAB AUDIO_U16LSB
#  define AUDIO_S16SWAB AUDIO_S16LSB
#endif

/* sound specification as configured by user */
static SfxConfig sound_config = {
    22500,                      // 22 kHz
    true,                       // stereo
    true,                       // 16 bits
    true,                       // enabled
    false,                      // headphone
    false                       // reverse stereo
};

/** Current sound hardware. This is the "obtained" from SDL_OpenAudio(). */
SDL_AudioSpec sound_hardware;

/** True iff sound is initialized and running. */
static bool sound_initialized;

/** Sample cache. */
typedef std::map<string_t, SfxSampleData*> SfxSampleCache;
SfxSampleCache sample_cache;

/** Current generation of samples. This is used to detect outdated
    samples. */
int SfxSample::current_generation;

/***************************** High-Level API ****************************/

/** Maximum number of currently-playing voices. */
static const int NUM_VOICES = 32;

namespace {
    struct Voice {
        /** Pointer to voice data. */
        SfxSampleData* sample;
        /** Playing position (handed to mix functions) */
        uint32         pos;
        /** Importance of this sample. When no play slots are free,
            we stop playing the least-important sample. Playing will
            *not* be resumed. */
        int            importance;
        /** Age of this sample. When no play slots are free and there
            is no less important sample, we remove the oldest. */
        int            age;
        /** Volume. */
        uint8          volume;
        /** Panning. 0 is left, 255 is right. */
        uint8          pan;

        inline void mix(Uint8* stream, int size) volatile {
            ++age;
            pos = sample->mix(stream, size, volume, pan, pos);
            if (!pos)
                sample = 0;
        }
    };
}

/** Currently-playing voices. volatile because it is changed from an
    interrupt. */
static volatile Voice voices[NUM_VOICES];
static bool audio_paused;
static int reserved_voices;

extern "C" {
    static void mixerFunction(void* /*arg*/, Uint8* stream, int len) {
        if (sound_hardware.format == AUDIO_U8 || sound_hardware.format == AUDIO_U16)
            std::fill_n(stream, len, sound_hardware.silence);
        else
            std::fill_n(static_cast<uint16*>(static_cast<void*>(stream)),
                        len/2, sound_hardware.silence);
        for (unsigned i = 0; i < countof(voices); ++i)
            if (voices[i].sample)
                voices[i].mix(stream, len);
    }
}

static bool
isAudioFormatKnown(uint16 fmt)
{
    return (fmt == AUDIO_U8 || fmt == AUDIO_S8
            || fmt == AUDIO_U16SYS || fmt == AUDIO_S16SYS
          /* || fmt == AUDIO_U16SWAB || fmt == AUDIO_S16SWAB */);
}

static const char*
getAudioFormatName(uint16 fmt)
{
    switch (fmt) {
     case AUDIO_U8:
        return _("8 bit, unsigned");
     case AUDIO_S8:
        return _("8 bit, signed");
     case AUDIO_U16SYS:
        return _("16 bit, unsigned");
     case AUDIO_S16SYS:
        return _("16 bit, signed");
     case AUDIO_U16SWAB:
        return _("16 bit, unsigned, swapped");
     case AUDIO_S16SWAB:
        return _("16 bit, signed, swapped");
     default:
        return _("unknown audio format");
    }
}

/** Find voice playing a sample. When we want to play a sound effect which
    is already running, we just change its volume and panning to save
    a channel. */
static volatile Voice*
findPlayingVoice(SfxSampleData* sample)
{
    /* we allow an imprecision of 3 mixer ticks (~0.3 seconds) to avoid
       overloading the mixer (and the user!) */
    for (unsigned i = reserved_voices; i < countof(voices); ++i)
        if (voices[i].sample == sample && voices[i].age < 3 /* arbitrary */)
            return &voices[i];
    return 0;
}

/** Find a free voice. \returns 0 if none found, or pointer to voice
    which is in an indeterminate state. */
static volatile Voice*
findVoice(int importance)
{
    for (unsigned i = reserved_voices; i < countof(voices); ++i)
        if (!voices[i].sample)
            return &voices[i];

    volatile Voice* v = 0;
    for (unsigned i = reserved_voices; i < countof(voices); ++i)
        if (voices[i].importance < importance
            && voices[i].age > 3
            && (!v
                || (voices[i].importance < v->importance)
                || (voices[i].importance == v->importance
                    && voices[i].age > v->age)))
            v = &voices[i];

    return v;
}

/** Play this sample.
    \param volume     volume
    \param pan        panning (0=left, 255=right)
    \param importance importance. Important samples kick out unimportant
                      ones if not enough voices left. */
void
SfxSample::play(uint8 volume, uint8 pan, int importance)
{
    if (isOutdated() || !data || !volume || !sound_initialized)
        return;

    /* Stereo tricks.
       - Reverse stereo is straightforward.
       - Headphone mode means we pan to [38, 217] instead of [0, 255]
         (30% smaller range), so that "left" isn't to the left of the
         listener but left in front of him. */
    if (sound_config.reverse_stereo)
        pan = 255 - pan;
    if (sound_config.headphone)
        pan = pan * 179 / 255 + 38;

    SDL_LockAudio();
    volatile Voice* v;
    if ((v = findPlayingVoice(data)) != 0) {
        /* voice already playing. Adjust volume and panning */
        v->pan = (v->volume * v->pan + volume * pan) / (v->volume + volume);
        if (volume < volume_Max - v->volume)
            v->volume += volume;
        else
            v->volume = volume_Max;
    } else if ((v = findVoice(volume)) != 0) {
        /* start anew */
        v->sample     = data;
        v->pos        = 0;
        v->importance = importance;
        v->age        = 0;
        v->volume     = volume;
        v->pan        = pan;
    } else {
        /* unable to find a free slot. */
    }
    SDL_UnlockAudio();
    if (audio_paused)
        updateSoundStatus();
}

/****************************** Sample Cache *****************************/

SfxSample
requestSound(string_t name)
{
    if (!sound_initialized)
        return 0;

    SfxSampleCache::iterator i = sample_cache.find(name);
    if (i != sample_cache.end())
        return i->second;

    /* we do intentionally insert null samples into the cache, so
       we know that a sample is missing later. */
    Ptr<Stream> s = ResManager::getInstance().getNamedEntry("s." + name);
    SfxSampleData* d = 0;
    try {
        if (s)
            d = new SfxSampleData(s);
    }
    catch(...) { }
    sample_cache.insert(SfxSampleCache::value_type(name, d));
    return d;
}

SfxSample
requestSound(string_t name, int index)
{
    if (!sound_initialized)
        return 0;

    SfxSample n = requestSound(name + itoa(index));
    if (!n.isValid())
        n = requestSound(name);
    return n;
}

SfxSample
requestSound(const char* name)
{
    return requestSound(string_t(name));
}

SfxSample
requestSound(const char* name, int index)
{
    return requestSound(string_t(name), index);
}

static void
clearSampleCache()
{
    for (SfxSampleCache::iterator i = sample_cache.begin(); i != sample_cache.end(); ++i)
        delete i->second;
    SfxSampleCache().swap(sample_cache);
}

/******************************* Player API ******************************/

/** Possibly put sound into "pause" mode. Call this to improve
    performance when sound is *not* playing. */
void
updateSoundStatus()
{
    int pause = 1;
    if (!sound_initialized)
        return;
    for (unsigned i = 0; i < countof(voices); ++i) {
        if (voices[i].sample) {
            pause = 0;
            break;
        }
    }
    SDL_PauseAudio(pause);
    audio_paused = pause;
}

/** Initialize sound. */
void
initSound()
{
    if (sound_initialized)
        return;

    if (!sound_config.enabled)
        return;

    if (!SDL_WasInit(SDL_INIT_AUDIO)) {
        if (SDL_InitSubSystem(SDL_INIT_AUDIO) != 0) {
            console.write(LOG_INFO, string_t(_("Sound disabled: ")) + SDL_GetError());
            return;
        }
    }

    SDL_AudioSpec desired;
    desired.freq     = sound_config.frequency;
    desired.format   = sound_config.sixteen_bits ? AUDIO_S16SYS : AUDIO_S8;

    /* we want a small sound latency; "* 30" means 1/15 to 1/30 sec
       latency (in theory...). SDL insists on getting a power of two,
       hence we get 1/21 sec with usual sound cards:
         22050 Hz / 1024 samples = 21/s */
    desired.samples = 512;
    while (desired.samples * 30 < sound_config.frequency)
        desired.samples *= 2;

    desired.callback = mixerFunction;
    desired.channels = sound_config.stereo ? 2 : 1;
    desired.userdata = 0;

    if (SDL_OpenAudio(&desired, &sound_hardware) < 0) {
        console.write(LOG_INFO, string_t(_("Sound disabled: ")) + SDL_GetError());
        return;
    }

    if (!isAudioFormatKnown(sound_hardware.format)) {
        console.write(LOG_INFO, _("Sound disabled: unsupported audio format."));
        return;
    }

    sound_initialized = true;
    clearSampleCache();
    SfxSample::newGeneration(); // this invalidates all current samples

    char driver_name[100];
    if (SDL_AudioDriverName(driver_name, sizeof driver_name)) {
        console.write(LOG_INFO,
                      format(_("Sound driver: %s (%d Hz, %d channels, %s; %d samples)"))
                      << driver_name
                      << sound_hardware.freq
                      << sound_hardware.channels
                      << getAudioFormatName(sound_hardware.format)
                      << sound_hardware.samples);
    }

    updateSoundStatus();
}

/** Shut down sound. \post !sound_initialized */
void
doneSound()
{
    if (!sound_initialized)
        return;

    sound_initialized = 0;
    SDL_CloseAudio();

    for (unsigned i = 0; i < countof(voices); ++i)
        voices[i].sample = 0;

    clearSampleCache();
}

/** Change sound configuration. */
void
configureSound(const SfxConfig& config)
{
    /* "headphone" or "reverse_stereo" don't need reinitialisation
       because we do it all ourself. */
    if (config.frequency != sound_config.frequency
        || config.stereo != sound_config.stereo
        || config.sixteen_bits != sound_config.sixteen_bits
        || config.enabled != sound_config.enabled
        || !sound_initialized)
    {
        doneSound();
        sound_config = config;
        initSound();
    } else {
        sound_config = config;
    }
}

const SfxConfig&
getSoundConfig()
{
    return sound_config;
}
