// A simple test program demonstrating 'istringstream' and
// 'ostringstream' plus the use of some floating point formatting.

// For 'string' and IOStream stuff and in particular for
// 'istringstream' and 'ostringstream':
#include <sstream>

// For 'EXIT_FAILURE' and 'EXIT_SUCCESS':
#include <cstdlib>

int main(int ac, char *av[])
{
  // The first thing to do is to check for the correct arguments. There
  // are just two arguments but especially the second argument can
  // consist of multiple floating point values:
  if (ac < 2)
  {
    cerr << "Usage: " << av[0] << " <value>" << endl;
    return EXIT_FAILURE;
  }

  // Next, the second argument, a C-style string is converted to a
  // 'string':
  string input(av[1]);

  // The 'string' is used to initialize an 'istringstream' to read from:
  istringstream in(input);

  // A 'double' is used to store the value for manipulation:
  double value = 0.0;

  // The following loop will iterate over all values found in the
  // 'string':

  while (!in.eof())
  {
    // Nothing new over here: Just reading in a value...
    in >> value;

    // ... and checking that there was no failure.
    if (in.fail())
    {
      if (in.eof())
        return EXIT_SUCCESS; // OK, we have reached eof...
      cerr << "ERROR: Reading of floating point value failed." << endl;
      return EXIT_FAILURE;
    }
    // Otherwise, the conversion succeeded (maybe 'eof' is reached but
    // don't bother with this yet).

    // ... an 'ostringstream' is created...
    ostringstream out;
    // ... and setup to print a floating point with a precision of
    // '2' in "fixed" format:
    out.precision(2);
    out.setf(ios::fixed, ios::floatfield);

    // Then, some "arithmetic" is performed on the value read...
    value += 1;

    // ... and the result is printed:
    out << value;

    // Finally, the printed result is converted back to a string...
    string output = out.str();

    // ... which in turn is converted to a 'char*' and then printed
    // on the standard output device (the conversion to 'char*' is,
    // of course, not strictly necessary as there is an insertor for
    // 'string').
    char const *cstr = output.c_str();
    cout << "The result is: " << cstr << endl;

    // Note: The character sequence obtained by the 'c_str()' member
    // of 'string' may not be modified. It may also become invalid by
    // the next call of a non-const member of for corresponding
    // 'string'. In short, the character sequence is owned by the
    // 'string' object.
  }

  return EXIT_SUCCESS;
}
