/*
   reading .vpm files

   This reads a .vpm file and outputs a ship summary to standard output.

   Not all values are 100% exact because the movie recorder doesn't
   record everything correctly and uniquely:
   - carriers may be shown to have torps, if the ship previously in that
     slot had some
   - the turn numbers when the ship was built might be guestimated wrong
 */

#include <iostream>
#include <cstdlib>
#include "fileio/streams.h"
#include "cpluslib/buffer.h"
#include "util/except.h"
#include "game/struct.h"
#include "cpluslib/types.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/global.h"
#include "game/spec.h"

static void
skipStream(Stream* s, unsigned int bytes)
{
    char buffer[1024];
    while (bytes > sizeof buffer)
        s->read(buffer, sizeof buffer), bytes -= sizeof buffer;
    if (bytes)
        s->read(buffer, bytes);
}

struct Ship {
    uint8   hull;
    uint8   owner_engine;
    uint8   beams;
    uint8   torps;
    uint8   damage;
    uint16  x, y;
    int     built;
};

struct Mine {
    uint16  x, y, radius;
    uint8   owner_type;
};

static int  planets[NUM_PLANETS];
static Ship ships[NUM_SHIPS];
static Mine mines[500];
static int  turn;

static int& p(int x) {
    static int dummy;
    if (x > 0 && x <= NUM_PLANETS)
        return planets[x-1];
    else
        return dummy;
}

static Ship& s(int x) {
    static Ship dummy;
    if (x > 0 && x <= NUM_SHIPS)
        return ships[x-1];
    else
        return dummy;
}

static Mine& m(int x) {
    static Mine dummy;
    if (x > 0 && x <= countof(mines))
        return mines[x-1];
    else
        return dummy;
}

static unsigned
readPlanets(Buffer<char>& buf, unsigned ip)
{
    while (1) {
        if (ip >= buf.getSize())
            throw FileFormatException("overrun in readPlanets");

        uint16 pid = getUint16(buf, ip);
        ip += 2;
        if (!pid)
            break;

        p(pid & 511) = pid >> 9;
    }
    return ip;
}

static unsigned
readShips(Buffer<char>& buf, unsigned ip)
{
    while (1) {
        if (ip >= buf.getSize())
            throw FileFormatException("overrun in readShips");

        uint16 sid = getUint16(buf, ip);
        ip += 2;
        if (!sid)
            break;

        Ship& ref = s(sid >> 7);
        if (sid & 1)
            ref.hull = buf[ip++], ref.built = turn;
        if (sid & 2) {
            ref.owner_engine = buf[ip++];
            if ((ref.owner_engine & 15) == 0) {
                ref.beams = ref.damage = ref.torps = 0;
                if (! (sid & 1))
                    ref.hull = 0;
            }
        }
        if (sid & 4)
            ref.beams = buf[ip++];
        if (sid & 8)
            ref.torps = buf[ip++];
        if (sid & 16)
            ref.damage = buf[ip++];
        if (sid & 32)
            ref.x = getInt16(buf, ip), ip += 2;
        if (sid & 64)
            ref.y = getInt16(buf, ip), ip += 2;
    }
    return ip;
}

static unsigned
readMines(Buffer<char>& buf, unsigned ip)
{
    while (1) {
        if (ip >= buf.getSize())
            throw FileFormatException("overrun in readMines");

        uint16 mid = getUint16(buf, ip);
        ip += 2;
        if (!mid)
            break;

        Mine& ref = m(mid & 511);
        if (mid & 0x8000) {
            ref.radius = getInt16(buf, ip), ip += 2;
        } else {
            ref.x = getInt16(buf, ip), ip += 2;
            ref.y = getInt16(buf, ip), ip += 2;
            ref.radius = getInt16(buf, ip), ip += 2;
            ref.owner_type = buf[ip++];
        }
    }
    return ip;
}

static unsigned
readTurn(Buffer<char>& buf, unsigned ip)
{
    ++turn;
    while (1) {
        if (ip >= buf.getSize())
            throw FileFormatException("overrun in readTurn");
        char opc = buf[ip++];
        if (opc == 0)
            break;
        else if (opc == 1)
            ip = readPlanets(buf, ip);
        else if (opc == 2)
            ip = readShips(buf, ip);
        else if (opc == 3)
            ip = readMines(buf, ip);
        else
            throw FileFormatException(format("invalid block type %d", opc));
    }
    return ip;
}

void
printSummary(int pid)
{
    std::cout << "Summary for player " << pid
              << "\n======================\n\n";
    std::cout << "Ship#  Type                            Beams  Torps     X,Y     Built\n"
              << "-----  ------------------------------  -----  -----  ---------  -----\n";
    for (int sid = 1; sid <= NUM_SHIPS; ++sid) {
        Ship& s = ::s(sid);
        if ((s.owner_engine & 15) == pid) {
            std::cout << format("%4d   %-30s  ", sid, getHull(s.hull).getName());
            if (s.beams > 15)
                std::cout << format("%2dx%-2d  ", s.beams >> 4, s.beams & 15);
            else
                std::cout << "       ";
            if (s.torps > 15)
                std::cout << format("%2dx%-2d  ", s.torps >> 4, s.torps & 15);
            else
                std::cout << "       ";
            std::cout << format("%4d,%4d  %4d\n", s.x, s.y, s.built);
        }
    }
    std::cout << "\n\n";
}

int main(int argc, char** argv)
{
    try {
        if (argc != 2 && argc != 3) {
            std::cerr << "usage: " << argv[0] << " filename.vpm [gamedir]\n";
            std::exit(1);
        }

        initGameGlobals(argc == 3 ? argv[2] : ".", ".");
        initSpecification();

        Ptr<Stream> s = Stream::openFile(argv[1], Stream::C_READ);
        skipStream(s, 10 + 2000 + 682);  // ignore for now

        Buffer<char> buf;
        s->readBuffer(buf);
        unsigned ip = 0;
        while (ip < buf.getSize()) {
            ip = readTurn(buf, ip);
        }
        for (int i = 1; i <= NUM_PLAYERS; ++i)
            printSummary(i);
        std::cout << "Notes: information is not 100% exact. The part of the movie recorder which\n"
                     "records ship equipment is seems not to be very well-tested (the player does\n"
                     "not display that information). The `Built' column is an estimate when this\n"
                     "ship was built. Weapon information is in format `count x type index'.\n";
    }
    catch(std::exception& e) {
        std::cerr << "exception: " << e.what() << "\n";
    }
}
