/**
  *  \file tests/sendfile.cc
  *  \brief Sending files within a .TRN
  *
  *  This program allows you to send files within your TRN.
  *  - PHost 4.0 supports a new TRN command that just sends you back
  *    a util.dat record. We use this to send a file transfer record,
  *    so we'll get the file back with the next RST.
  *
  *  - THost 3.2 accepts a changed TRN format ("TacCom-enhanced")
  *    which allows file upload to the host. This file is not sent
  *    back.
  */

#include <iostream>
#include <cstdlib>
#include "game/trn.h"
#include "cpluslib/buffer.h"
#include "fileio/streams.h"
#include "fileio/io.h"
#include "util/except.h"

enum Mode { use_Sendback, use_Taccom };

static char* prog_name;
static char* trn_name;
static char* file_name;
static Mode  how = use_Sendback;

static void
die(const char* data)
{
    std::cerr << prog_name << ": " << data << "\n";
    std::exit(1);
}

static void
doAttach()
{
    Ptr<Stream> s = Stream::openFileT(trn_name, Stream::C_READ);
    Ptr<Stream> f = Stream::openFileT(file_name, Stream::C_READ);

    if (f->getSize() > 1000*1000L)
        die("file exceeds reasonable size limits -- aborting");
    
    GTurnfile trn(s);
    Buffer<char> buffer;
    f->readBuffer(buffer);

    switch (how) {
     case use_Sendback: {
         char command_data[4+13];

         if (buffer.getSize() > 32000)
             die("file too large to be sent with SendBack -- aborting");
         trn.addCommand(cm_SendBack, 0);
         storeUint16(command_data + 0, 34 /* Type: FTP */);
         storeUint16(command_data + 2, buffer.getSize() + 13);
         storeBasicStringN(command_data + 4, 12, file_name);
         command_data[16] = 0;  // flag
         trn.addData(command_data, sizeof command_data);
         trn.addData(buffer.ptr(), buffer.getSize());
         break;
     }
     case use_Taccom:
        if (trn.addFile(buffer, file_name) < 0)
            die("Unable to attach file, too many files already there.");
        break;
    }
    trn.update();
    trn.write(Stream::openFileT(trn_name, Stream::C_CREATE));
}

static void
help()
{
    std::cout << "SendFile -- Send a File with a .TRN file\n\n"
        "sendfile [-s|-t] playerX.trn file.dat\n\n"
        "-s\tUse SendBack command. The file will be sent back in your UTIL.DAT.\n"
        "\tWorks with PHost 4.0 and later ONLY.\n\n"
        "-t\tUse TacCom TRN format (\"Leech\" method). File will be sent to\n"
        "\thost. Works with THost 3.2 and later ONLY.\n\n"
        "Default format is `-s'.\n";
    std::exit(0);
}

int
main(int argc, char** argv)
{
    prog_name = argv[0];
    while (*++argv) {
        if (**argv == '-') {
            char* p = *argv;
            while (*++p) {
                if (*p == 't')
                    how = use_Taccom;
                else if (*p == 's')
                    how = use_Sendback;
                else if (*p == 'h')
                    help();
                else
                    die("invalid option. Use `sendfile -h' for help.");
            }
        } else if (!trn_name) {
            trn_name = *argv;
        } else if (!file_name) {
            file_name = *argv;
        } else {
            die("too many parameters. Usage: `sendfile playerX.trn file.dat'");
        }
    }

    if (!file_name) {
        die("too few parameters. Usage: `sendfile playerX.trn file.dat'");
    }

    try {
        doAttach();
    }
    catch(std::exception& e) {
        die(e.what());
    }
}
