/*
 *  TEST: VcrDatabase, VCR players
 *
 *  This takes up to three parameters:
 *  - a VCR file
 *  - a spec file directory (defaults to /usr/games/planets)
 *  - a repeat count (defaults to 1)
 *  The VCR file is read and played; results are displayed on stdout.
 */
#include <iostream>
#include <cstdlib>
#include <vector>
#include <iomanip>
#include "game/struct.h"
#include "game/tvcr.h"
#include "fileio/streams.h"
#include "util/except.h"
#include "util/console.h"
#include "game/global.h"
#include "game/spec.h"
#include "game/config.h"
#include "game/vcrdb.h"
#include "game/player.h"

static const char*const results[] = {
    "left-destroyed",
    "right-destroyed",
    "left-captured",
    "right-captured",
    "timeout"
};

void
help()
{
    std::cout << "Usage: playvcr vcrfile [rootdir [repeat]]\n";
    std::exit(1);
}

void
loadVcrs(VcrDatabase& vcrs, const char* name)
{
    Ptr<Stream> s = Stream::openFile(name, Stream::C_READ);

//     int n;
//     char buffer[TVcr::size];

//     if (s->read(buffer, 2) != 2)
//         throw FileFormatException("invalid VCR file");
//     n = getUint16(buffer);
//     std::cout << "VCR Header: " << n << " records\n";

//     while(n-- > 0) {
//         if (s->read(buffer, TVcr::size) != TVcr::size)
//             throw FileFormatException("VCR file too short");
//         TVcr v;
//         getStructure(buffer, v);
//         vcrs.push_back(v);
//     }

    vcrs.loadFromFile(s);
    std::cout << "VCR file contains " << vcrs.getVcrCount() << " entries\n";
}

int
main(int, char** argv)
{
    char* file = 0;
    char* dir = 0;
    bool repeat_set = false;
    int repeat = 1;

    while (*++argv) {
        if (!file)
            file = *argv;
        else if (!dir)
            dir = *argv;
        else if (!repeat_set) {
            repeat_set = true;
            repeat = std::atoi(*argv);
            if (repeat <= 0)
                help();
        } else
            help();
    }

    if (!file)
        help();

    try {    
        enableStdioConsole();
        initDirectories(dir ? dir : ".", "/usr/games/planets");
        initSpecification();
        initRaceNames();
        initConfig();
        
        //std::vector<TVcr> vcrs;
        VcrDatabase vcrs;
        loadVcrs(vcrs, file);
        
        Ptr<VcrPlayer> ply = vcrs.getVcrPlayer();//new VcrPlayerTHost();
        ply->setVisualizer(new VcrNullVisualizer());
        
        std::cout << "--- Starting Playback ---" << std::endl;

        for (int rc = 0; rc < repeat; ++rc) {
            for(unsigned i = 0; i < vcrs.getVcrCount(); ++i) {
                Ptr<VcrEntry> en = vcrs.getEntry(i);
                std::cout << "Record #" << i << ":\n";
                
                TVcr the_vcr = en->getVcr();
                if (ply->checkVcr(the_vcr)) {
                    std::cout << "\tinvalid (checkVcr)\n";
                    continue;
                }
                ply->playVcr(the_vcr);
                
                TVcr result;
                ply->doneVcr(result);
                int status = ply->getStatusWord();
                int time = ply->getTime();
                std::cout << "\tEnding time " << time
                          << " (" << time/60 << ":" << setw(2) << setfill('0') << time%60
                          << ")\n";
                if (status == VcrPlayer::VCRS_INVALID) {
                    std::cout << "\tinvalid (play)\n";
                } else {
                    char t = '\t';
                    for (int i = 0; i < 5; ++i) {
                        if (status & (1 << i)) {
                            std::cout << t << results[i];
                            t = ' ';
                        }
                    }
                    if (t == '\t')
                        std::cout << "\tnone\n";
                    else
                        std::cout << "\n";
#define VALUE(v,x) "  " << v << ":" << setw(3) << x
#define AMMO(t,u) (t.objects[u].torp_type ? t.num_torps[u] : t.num_fighters[u])
                    std::cout << setfill(' ')
                              << VALUE("S", result.shield[0])
                              << VALUE("D", result.objects[0].damage)
                              << VALUE("C", result.objects[0].crew)
                              << VALUE("A", AMMO(result, 0))
                              << "   |   "
                              << VALUE("S", result.shield[1])
                              << VALUE("D", result.objects[1].damage)
                              << VALUE("C", result.objects[1].crew)
                              << VALUE("A", AMMO(result, 1))
                              << "\n";
#undef VALUE
                }
            }
        }
    }
    catch (std::exception& e) {
        std::cout << "Exception: " << e.what() << "\n";
    }
}
 
