/**
  *  \file ui/event.cc
  *  \brief Event Management
  *
  *  Basically, our event management is the same as SDL's. This module,
  *  however, adds some important properties:
  *  - we use an own event format for (sort-of) independance of SDL,
  *    and for better handling of own events;
  *  - there is a synchronous event queue. SDL's built-in queue is
  *    asynchronous; however, we need the ability to post sequences
  *    of events without the user getting in the way;
  *  - there is a timer, running at TIMER_HZ (50) Hertz. The timer
  *    posts ev_Tick events, as well as incrementing a timer variable
  *    (like good ol' MS-DOS `MemL[$40:$6C]').
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <list>
#include <csignal>
#include <cmath>
#include <SDL.h>
#include "ui/event.h"
#include "util/except.h"
#include "util/nls.h"
#include "cpluslib/assert.h"
#include "util/recode.h"

/** Internal Event Queue. We need a separate event queue for
    command events we posted, to guarantee that our command events
    are processed before user input posted by SDL. */
static std::list<UIEvent> event_queue;

/** Timer. This is incremented on every timer tick.
    \todo This is incremented from a timer which runs asynchronously
    (interrupt, signal). Maybe we need to lock it somehow for platforms
    where a `long' can not be read/written atomically. */
static volatile long global_timer;
static volatile long lost_ticks;

/** SDL Timer Id. The identificator for our SDL timer. */
static SDL_TimerID global_timer_id;

/** Mouse mode. If true, we want infinite mouse movement. Mouse events
    will report relative movement in x/y. If false, we want normal
    mouse movement (absolute coordinates in x/y). */
static bool mouse_is_infinite;

/** Time of last mouse click. Needed for double-click recognition. */
static long last_click;
/** Position of last mouse click. */
static int last_x, last_y, last_b;

int double_click_delay = TIMER_HZ / 3;

/******************************* Functions *******************************/

/** Merge mouse and key modifiers. We do not report key modifiers when
    the mouse was released, so mouse-release events see a zero button
    mask. */
static inline uint32
mergeMod(uint32 mouse, uint32 key)
{
    /* we assume 3-button mice at most. Wheel events generate
       "button 4" and "button 5" press events which are never
       released, leaving these bits permanently on. */
    mouse &= 7;

    return mouse ? mouse | UIEvent::simplMod(key) : 0;
}

/** Simplify an SDL keysym. We translate the keypad keys into their
    "gray" equivalent. We need not care for the status of NumLock,
    when that is on, SDL will translate the keys into their ASCII /
    Unicode value for us which is handled elsewhere. */
static inline uint32
simplSym(uint32 sym)
{
    CTASSERT(SDLK_KP9 - SDLK_KP0 == 9);
    static uint32 numpad_map[] = {
        SDLK_INSERT,            // 0
        SDLK_END,               // 1
        SDLK_DOWN,              // 2
        SDLK_PAGEDOWN,          // 3
        SDLK_LEFT,              // 4
        SDLK_KP5,               // 5
        SDLK_RIGHT,             // 6
        SDLK_HOME,              // 7
        SDLK_UP,                // 8
        SDLK_PAGEUP             // 9
    };

    if (sym == SDLK_KP_ENTER)
        return SDLK_RETURN;
    else if (sym >= SDLK_KP0 && sym <= SDLK_KP9)
        return numpad_map[sym - SDLK_KP0];
    else if (sym == SDLK_KP_PERIOD)
        return SDLK_DELETE;
    else
        return sym;
}

/** Make a key event. */
static void
makeKeyEvent(UIEvent& e, uint32 code)
{
    /* discount shift on alphanumeric input */
    if ((code & ~ss_Shift) < 256)
        code &= ~ss_Shift;

    e.type         = ev_Key;
    e.key.sdl_type = 0;
    e.key.arg      = 0;
    e.key.code     = code;
    e.key.unicode  = 0;
    if (code && code >= 32 && code != 127 && code < 256)
        e.key.unicode = code;
}

/** Convert SDL event into our format.
    \param infinite true iff we are in "infinite movement" mode and want
    relative mouse movement to be reported.
    \returns true iff successful, false if event doesn't map to our scheme. */
static bool
convertEvent(const SDL_Event& se, UIEvent& ue, bool infinite)
{
    uint16 u;
    uint32 m;
    switch(se.type) {
     case SDL_KEYDOWN:
        u = se.key.keysym.unicode;
        m = UIEvent::simplMod(uint32(se.key.keysym.mod) << 16);
        ue.type         = ev_Key;
        ue.key.sdl_type = se.type;
        ue.key.arg      = 0;    // FIXME: prefix arg
        ue.key.unicode  = u;
        if (u && u >= 32 && u != 127) {
            /* it is an "alphabetic" key according to Unicode */
            unsigned char xl = Recoder::getInternCS().getCharacter(u);
            if (! xl)           // unencodable character
                return false;
            ue.key.code = m | xl;
        } else {
            /* it is a special key */
            ue.key.code = m | simplSym(se.key.keysym.sym);
        }
        /* discount shift on alphanumeric input */
        if (ue.key.code > ss_Shift+32 && ue.key.code < ss_Shift+256)
            ue.key.code &= ~ss_Shift;
        return true;
     case SDL_MOUSEMOTION:
        ue.type           = ev_Mouse;
        ue.mouse.sdl_type = se.type;
        if (infinite) {
            ue.mouse.x    = se.motion.xrel;
            ue.mouse.y    = se.motion.yrel;
        } else {
            ue.mouse.x    = se.motion.x;
            ue.mouse.y    = se.motion.y;
        }
        ue.mouse.b        = mergeMod(se.motion.state,
                                     uint32(SDL_GetModState()) << 16);
        return true;
     case SDL_MOUSEBUTTONDOWN:
        /* SDL 1.2.1 has no mouse wheel API, but the Windows backend
           reports a mouse wheel as buttons 4 and 5 (the fbcon backend
           detects an imaginary Z axis, but that doesn't yet fit into the
           SDL_MouseMotionEvent. */
        if (se.button.button == 4) {
            makeKeyEvent(ue, vk_WheelUp);
            return true;
        } else if (se.button.button == 5) {
            makeKeyEvent(ue, vk_WheelDown);
            return true;
        }
        /* FALLTHROUGH */
     case SDL_MOUSEBUTTONUP:
        ue.type           = ev_Mouse;
        ue.mouse.sdl_type = se.type;
        if (infinite) {
            ue.mouse.x = ue.mouse.y = 0;
        } else {
            ue.mouse.x    = se.button.x;
            ue.mouse.y    = se.button.y;
        }
        ue.mouse.b        = mergeMod(SDL_GetMouseState(0, 0),
                                     uint32(SDL_GetModState()) << 16);
        if (!ue.mouse.b && last_b) {
            /* mouse was released */
            if (last_click > 0
                && getTimer() - last_click < double_click_delay
                && std::abs(last_x - ue.mouse.x) < 5 && std::abs(last_y - ue.mouse.y) < 5)
            {
                ue.mouse.b |= mb_Double;
                last_click = 0;
            } else {
                last_click = getTimer();
            }
            last_x = ue.mouse.x;
            last_y = ue.mouse.y;
        }
        last_b = ue.mouse.b & ~mb_Double;
        return true;
     case ev_Tick:
        ue.type = ev_Tick;
        return true;
     case SDL_NOEVENT:
        ue.type = ev_Idle;
        return true;
     case SDL_QUIT:
        makeKeyEvent(ue, vk_Quit);
        return true;
     default:
        return false;
    }
}

/** Get Event from Queue. Fetches one event and returns it in
    \c event. If \c idle is false (default), waits for an event.
    If \c idle is true, returns an event with type ev_Idle
    if there is no event in the queue. If \c infinite is false,
    returns normal mouse events; otherwise returns relative
    mouse movement. */
void
getEvent(UIEvent& event, bool idle, bool infinite)
{
    setMouseMode(infinite);
    if (event_queue.empty()) {
        SDL_Event ev;
        do {
            if (idle) {
                if (! SDL_PollEvent(&ev))
                    ev.type = SDL_NOEVENT;
            } else {
                SDL_WaitEvent(&ev);
            }
        } while (! convertEvent(ev, event, infinite));
    } else {
        event = event_queue.front();
        event_queue.pop_front();
    }
}

/** Get Event from Queue. \overload */
void
getEvent(UIEvent& event, bool idle)
{
    getEvent(event, idle, false);
}

/** Put event into Queue. Enqueues the specified event. It is
    guaranteed that events enqueued with this function are processed
    in the right order and before all user input. */
void
putEvent(const UIEvent& event)
{
    event_queue.push_back(event);
}

/** Put command event into Queue. \overload */
void
putCommand(int32 command, int32 info)
{
    putCommand(command, info, 0, 0);
}

/** Put command event into Queue. \overload */
void
putCommand(int32 command, int32 info, void* pinfo)
{
    putCommand(command, info, 0, pinfo);
}

/** Put command event into Queue.
    \param command  command code
    \param info     first parameter
    \param info2    second parameter
    \param pinfo    info pointer
    \see putEvent */
void
putCommand(int32 command, int32 info, int32 info2, void* pinfo)
{
    UIEvent e;
    e.type         = ev_Command;
    e.cmd.sdl_type = 0;
    e.cmd.arg      = 0;
    e.cmd.code     = command;
    e.cmd.info     = info;
    e.cmd.info2    = info2;
    e.cmd.infop    = pinfo;
    putEvent(e);
}

/** Enqueue key event (simulate it).
    Secret knowledge: "faked" keys can be told apart from normal keypresses
    because faked keys have 0 in sdl_type, real ones have SDL_KEYDOWN. */
void
putKey(uint32 code)
{
    UIEvent e;
    makeKeyEvent(e, code);
    putEvent(e);
}

/** Unget Event. Undoes the effect of a getEvent(). This causes the
    specified event be processed again (before enqueued and user events). */
void
ungetEvent(const UIEvent& event)
{
    event_queue.push_front(event);
}

/** Check whether we have enqueued events.
    \return true iff there is a putEvent() event available. User events
    do not cause this function to return true. */
bool
haveQueuedEvent()
{
    return !event_queue.empty();
}

/** Get current timer. The timer is increased TIMER_HZ times per
    second (default: 50 Hz). */
long
getTimer()
{
    return global_timer;
}

long
getLostTicks()
{
    return lost_ticks;
}

/** Post a Mouse Event. This must be called after certain ``structural
    changes'' on the GUI. */
void
postMouseEvent()
{
    UIEvent e;
    int x, y;

    /* SDL_PumpEvents();  -- do not do that. We want to post the state as
                             it was at the last regular mouse event. */
    e.type           = ev_Mouse;
    e.mouse.b        = SDL_GetMouseState(&x, &y);
    e.mouse.sdl_type = 0;
    e.mouse.x        = x;
    e.mouse.y        = y;
    putEvent(e);
}

/*************************** Back-End Functions **************************/

extern "C" {
    /** SDL Timer Function. This function is responsible for posting timer
        events and updating the \c global_timer. */
    static Uint32
    timerFunction(Uint32 /*interval*/, void* /*param*/)
    {
        SDL_Event evt;
        evt.type = ev_Tick;
        evt.user.code = 0;
        evt.user.data1 = evt.user.data2 = 0;
        if (SDL_PeepEvents(NULL, 1, SDL_PEEKEVENT, 1 << ev_Tick) == 0)
            SDL_PushEvent (&evt);
        else
            ++lost_ticks;

        // FIXME: atomic
        ++global_timer;
        return TIMER_MS;
    }

    /** QUIT handler. When SDL itself sees a QUIT event, it will think
        we handle it, and close our window. This does not fit our
        model. We want a QUIT event to generate an event but not
        disrupt the normal event stream. Hence, we post the event
        manually and lie to SDL. */
    static int
    quitHandler(const SDL_Event* event)
    {
        if (event->type == SDL_QUIT) {
            /* SDL_PushEvent is SDL_PeepEvents in disguise, but doesn't
               modify *event */
            SDL_PushEvent(const_cast<SDL_Event*>(event));
            return 0;
        }
        return 1;
    }
}

/** Initialize Event Subsytem. */
void
initEvents()
{
    ASSERT (!global_timer_id);

    if (!SDL_WasInit(SDL_INIT_TIMER))
        if (SDL_InitSubSystem(SDL_INIT_TIMER) != 0)
            throw InitException(string_t(_("Unable to initialize timer subsystem: ")) + SDL_GetError());

    global_timer_id = SDL_AddTimer(TIMER_MS, timerFunction, 0);
    if (!global_timer_id)
        throw InitException(string_t(_("Unable to start timer: ")) + SDL_GetError());

    SDL_SetEventFilter(quitHandler);
    SDL_EnableUNICODE(1);
    SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);

    SDL_EventState(SDL_ENABLE, ev_Tick);
    SDL_EventState(SDL_ENABLE, SDL_KEYDOWN);
    SDL_EventState(SDL_ENABLE, SDL_MOUSEBUTTONUP);
    SDL_EventState(SDL_ENABLE, SDL_MOUSEBUTTONDOWN);
    SDL_EventState(SDL_ENABLE, SDL_MOUSEMOTION);

    mouse_is_infinite = false;
}

/** Shutdown Event Subsystem. */
void
doneEvents()
{
    if (global_timer_id) {
        SDL_RemoveTimer(global_timer_id);
        global_timer_id = 0;
    }
    SDL_SetEventFilter(0);
}

/** Set mouse mode.
    \param infinite=false normal mode. Mouse events report normal x/y
           coordinates. Pointer is visible.
    \param infinite=true infinite movement mode. Mouse events report
           relative x/y coordinates. Pointer is invisible. */
void
setMouseMode(bool infinite)
{
    if (infinite == mouse_is_infinite)
        return;
    mouse_is_infinite = infinite;
    if (infinite) {
        SDL_ShowCursor(0);
        SDL_WM_GrabInput(SDL_GRAB_ON);
    } else {
        SDL_ShowCursor(1);
        SDL_WM_GrabInput(SDL_GRAB_OFF);
    }
}

/**************************** UIPollableTimer ****************************/

/*! \class UIPollableTimer
    This timer is intended to be used for visual effects, to provide
    a higher resolution than our "system timer" aka ev_Tick events.
    It is intended as a replacement for SDL_Delay. Instead of
    \code
      while(1) {
        doSomething();
        SDL_Delay(10);
      }
    \endcode
    use the following:
    \code
      while(1) {
        UIPollableTimer timer(10);    // a 10 ms (100 Hz) timer
        doSomething();
        timer.waitTime(1);            // wait until the first interval ends
      }
    \endcode
    This can be improved by creating only one timer and re-using it;
    use `waitTime(++tick)' or something like that. If you expect
    `doSomething()' to overrun its timeslice, you can also check
    `getTime()' to possibly skip some cycles. */

extern "C"
{
    static Uint32 pollTimerFunction(Uint32 interval, void* param)
    {
        // FIXME: atomic
        ++*(int*)param;
        return interval;
    }
}

/** Construct new timer.
    The timer increments every ainterval milliseconds. */
UIPollableTimer::UIPollableTimer(int ainterval)
    : interval(ainterval), status(0)
{
    timer = SDL_AddTimer(interval, pollTimerFunction, &status);
    // FIXME: failure
}

UIPollableTimer::~UIPollableTimer()
{
    if (timer)
        SDL_RemoveTimer(timer);
}

/** Get current time. */
int
UIPollableTimer::getTime() const
{
    // FIXME: atomic
    return status;
}

/** Set timer. This does not restart the current timeslice, though. */
void
UIPollableTimer::setTime(int time)
{
    // FIXME: atomic
    status = time;
}

/** Wait until time reached. Busy waiting. */
void
UIPollableTimer::waitTime(int time)
{
    while (status < time)
        SDL_PumpEvents();
}
