/**
  *  \file ui/event.h
  *  \brief Event Management Prototypes
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UI_EVENT_H
#define PCC_V2_UI_EVENT_H

#include <SDL_events.h>
#include <SDL_timer.h>
#include "cpluslib/types.h"

/** Our own event types. Note that code in event.cc assumes that these
    are valid SDL event codes. */
enum UIEventType {
    ev_Key        = SDL_KEYDOWN,
    ev_Mouse      = SDL_MOUSEMOTION,
    ev_Idle       = SDL_NOEVENT,
    ev_Command    = SDL_USEREVENT,
    //ev_Broadcast,
    ev_Tick
};

enum {
    mb_Left   = SDL_BUTTON(SDL_BUTTON_LEFT),
    mb_Right  = SDL_BUTTON(SDL_BUTTON_RIGHT),
    mb_Middle = SDL_BUTTON(SDL_BUTTON_MIDDLE),
    mb_Double = 0x8000U
};

enum {
    ss_Shift = KMOD_LSHIFT << 16,
    ss_Ctrl  = KMOD_LCTRL  << 16,
    ss_Alt   = KMOD_LALT   << 16,
    ss_Meta  = KMOD_LMETA  << 16,
    
    ss_AllMod = ss_Shift | ss_Ctrl | ss_Alt | ss_Meta
};

/** Virtual keys. These are posted as normal key events. */
enum {
    vk_WheelUp = SDLK_LAST,
    vk_WheelDown,
    vk_Quit
};

typedef uint32 UIKey;

/* these ought to be macros so that they can be used in `switch' */
/* #define MOD(mod) (((mod) & (KMOD_LCTRL|KMOD_LALT|KMOD_LMETA|KMOD_LSHIFT)) \
                   | (((mod)>>1) & (KMOD_LCTRL|KMOD_LALT|KMOD_LMETA|KMOD_LSHIFT))) */

/** Our own event structure. We support fewer events than SDL, but use
    a more regular format. */
struct UIEvent {
    struct Mouse {
        int   sdl_type;         // SDL event type
        int   x, y;
        int32 b;                // button state + double-click flag
    };
    struct Key {
        int   sdl_type;
        int   arg;
        uint32 code;
        uint16 unicode;
    };
    struct Command {
        int   sdl_type;
        int   arg;
        int32 code;
        int32 info, info2;
        void* infop;
    };
    UIEventType type;
    union {
        /* ev_Key */
        Key     key;
        /* ev_Mouse */
        Mouse   mouse;
        /* ev_Command, ev_Broadcast */
        Command cmd;
    };

    static inline uint32 simplMod(uint32 m) {
        return (m & ss_AllMod) | ((m >> 1) & ss_AllMod);
    }
};

enum {
    TIMER_HZ = 50,                 ///< Timer frequency (Hertz).
    TIMER_MS = 1000 / TIMER_HZ     ///< Timer delay (milliseconds).
};

void getEvent(UIEvent& event, bool idle = false);
void getEvent(UIEvent& event, bool idle, bool infinite);
void putEvent(const UIEvent& event);
void putCommand(int32 command, int32 info);
void putCommand(int32 command, int32 info, void* pinfo);
void putCommand(int32 command, int32 info, int32 info2, void* pinfo=0);
void putKey(uint32 code);
void ungetEvent(const UIEvent& event);
void postMouseEvent();
bool haveQueuedEvent();

long getTimer();
long getLostTicks();

void initEvents();
void doneEvents();

void setMouseMode(bool infinite);

//! Pollable timer for short intervals.
class UIPollableTimer {
    int interval;
    int status;
    SDL_TimerID timer;
 public:
    UIPollableTimer(int ainterval);
    ~UIPollableTimer();

    int getTime() const;
    void setTime(int time);
    void waitTime(int time);
};

#endif
