/**
  *  \file ui/layout.cc
  *  \brief Layout Management
  *
  *  Layout managers arrange widgets in a group, without needing to
  *  manually compute coordinates. Layouts automatically adapt to
  *  changing font sizes and so forth, provided the widgets report
  *  correct values via getLayoutInfo().
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <climits>
#include <algorithm>
#include <vector>
#include "ui/layout.h"
#include "cpluslib/format.h"
#include "cpluslib/assert.h"

namespace {
    /** Information for HBox/VBox layouts. */
    struct LayoutInfo {
        std::vector<int> pref_sizes, min_sizes, flags;
        int used_space;
        int used_outer;

        LayoutInfo();
        std::vector<int>& doLayout(int space, int outer, int have_size, int flex_flag);
    };
};

LayoutInfo::LayoutInfo()
{ }

/** Compute generic layout.
    \pre pref_sizes, min_sizes, flags filled in
    \param space      inter-widget space we want to use
    \param outer      outer space we want to leave
    \param have_size  available space
    \returns vector of widget sizes
    \post used_outer = actual outer space to leave
    \post used_space = actual inter-widget space to leave */
std::vector<int>&
LayoutInfo::doLayout(int space, int outer, int have_size, int flex_flag)
{
    typedef std::vector<int>::size_type index;
    ASSERT(pref_sizes.size() == min_sizes.size());
    ASSERT(pref_sizes.size() == flags.size());
        
    /* First, figure out widget counts and sizes */
    int num_layout = 0, num_flex = 0;
    int total_pref = 0, total_min = 0;
    for (index i = 0; i < flags.size(); ++i) {
        if (flags[i] & UIWidget::li_NoLayout) {
            ;
        } else {
            ++num_layout;
            total_pref += pref_sizes[i];
            total_min  += min_sizes[i];
            if (flags[i] & flex_flag)
                ++num_flex;
        }
    }

    /* Space that would be needed to layout this */
    used_space = space;
    if (num_layout > 1) {
        // too little space, reduce inter-widget gap
        while (used_space > 0 && (have_size < (num_layout-1) * used_space + total_min))
            --used_space;
        have_size -= (num_layout-1) * used_space;
    }

    used_outer = outer;
    while (used_outer > 0 && (have_size < 2*used_outer + total_min))
        // too little space, reduce outer space
        --used_outer;
    have_size -= 2*used_outer;

    if (have_size < total_pref) {
        /* Total size too small. Scale down relatively using minimum sizes. */
        /* FIXME: we can probably do better for have_size \in [min_total, pref_total] */
        int total = have_size;
        for (index i = 0; i < min_sizes.size(); ++i) {
            if (flags[i] & UIWidget::li_NoLayout) {
                min_sizes[i] = 0;
            } else {
                int new_size = total_min ? min_sizes[i] * total / total_min : 0;
                total     -= new_size;
                total_min -= min_sizes[i];
                min_sizes[i] = new_size;
            }
        }
        ASSERT(total_min == 0);
        ASSERT(total == 0);
        return min_sizes;
    } else if (have_size > total_pref) {
        /* Total size larger than preferred size. Scale up using
           preferred sizes. */
        int delta = have_size - total_pref;  // total amount to enlarge
        if (num_flex) {
            /* we have flexible components; enlarge only those */
            for (index i = 0; i < pref_sizes.size(); ++i) {
                if (flags[i] & flex_flag) {
                    int d = delta / num_flex--;
                    pref_sizes[i] += d;
                    delta -= d;
                }
            }
            ASSERT(num_flex == 0);
            ASSERT(delta == 0);
        } else {
            /* no flexible: enlarge all components */
            for (index i = 0; i < pref_sizes.size(); ++i) {
                if (!(flags[i] & UIWidget::li_NoLayout)) {
                    int d = delta / num_layout--;
                    pref_sizes[i] += d;
                    delta -= d;
                }
            }
            ASSERT(delta == 0);
            ASSERT(num_layout == 0);
        }
        return pref_sizes;
    } else
        return pref_sizes;
}

/****************************** UIHBoxLayout *****************************/

/** \class UIHBoxLayout
    \brief Horizontal Box Layout

    Widgets will be arranged horizontally, all the same height,
    from left to right. */

/** Create a Horizontal Box layout.
    \param aspace space to leave between widgets, in pixels.
    \param aouter space to leave at each side. */
UIHBoxLayout::UIHBoxLayout(int aspace, int aouter)
    : space(aspace), outer(aouter)
{ }

void
UIHBoxLayout::getLayoutInfo(UIWidgetContainer* container, UIWidget::LayoutInfo& info)
{
    info.min_size = info.pref_size = GfxPoint(2*outer, 0);

    int n = 0;
    for (UIWidget* w = container->getFirstWidget(); w; w = w->getNext()) {
        UIWidget::LayoutInfo winfo;
        w->getLayoutInfo(winfo);
        if (winfo.flags & UIWidget::li_NoLayout)
            continue;
        ++n;
        info.min_size.x  += winfo.min_size.x;
        info.min_size.y   = std::max(info.min_size.y, winfo.min_size.y);
        info.pref_size.x += winfo.pref_size.x;
        info.pref_size.y  = std::max(info.pref_size.y, winfo.pref_size.y);
    }
    if (n) {
        info.min_size.x  += (n-1) * space;
        info.pref_size.x += (n-1) * space;
    }
}

void
UIHBoxLayout::doLayout(UIWidgetContainer* container, GfxRect size)
{
    LayoutInfo lay;
    for (UIWidget* w = container->getFirstWidget(); w; w = w->getNext()) {
        UIWidget::LayoutInfo winfo;
        w->getLayoutInfo(winfo);
        lay.pref_sizes.push_back(winfo.pref_size.x);
        lay.min_sizes.push_back(winfo.min_size.x);
        lay.flags.push_back(winfo.flags);
    }
    if (size.w == 0)
        return;
    
    std::vector<int>& v = lay.doLayout(space, outer, size.w, UIWidget::li_FlexibleH);
    doLayoutWithWidths(container, size, v, lay.used_space, lay.used_outer);
}

void
UIHBoxLayout::doLayoutWithWidths(UIWidgetContainer* container, GfxRect size,
                                 std::vector<int>& sizes, int space, int outer)
{
    std::vector<int>::size_type i = 0;
    size.x += outer;
    for (UIWidget* w = container->getFirstWidget(); w; w = w->getNext(), ++i) {
        UIWidget::LayoutInfo li;
        w->getLayoutInfo(li);
        if (li.flags & UIWidget::li_NoLayout)
            continue;
        w->setExtent(GfxRect(size.x, size.y, sizes[i], size.h), true);
        w->doLayout();
        size.x += sizes[i];
        size.x += space;
    }
}

/****************************** UIVBoxLayout *****************************/

/** \class UIVBoxLayout
    \brief Vertical Box Layout

    Widgets will be arranged vertically, all the same width, below
    each other from top to bottom. */

/** Create a Vertical Box layout.
    \param aspace space to leave between widgets, in pixels.
    \param aouter space to leave at the sides */
UIVBoxLayout::UIVBoxLayout(int aspace, int aouter)
    : space(aspace), outer(aouter)
{ }

void
UIVBoxLayout::doLayout(UIWidgetContainer* container, GfxRect size)
{
    LayoutInfo lay;
    for (UIWidget* w = container->getFirstWidget(); w; w = w->getNext()) {
        UIWidget::LayoutInfo winfo;
        w->getLayoutInfo(winfo);
        lay.pref_sizes.push_back(winfo.pref_size.y);
        lay.min_sizes.push_back(winfo.min_size.y);
        lay.flags.push_back(winfo.flags);
    }
    if (size.h == 0)
        return;
    
    std::vector<int>& v = lay.doLayout(space, outer, size.h, UIWidget::li_FlexibleV);
    doLayoutWithHeights(container, size, v, lay.used_space, lay.used_outer);
}

void
UIVBoxLayout::getLayoutInfo(UIWidgetContainer* container, UIWidget::LayoutInfo& info)
{
    info.min_size = info.pref_size = GfxPoint(0, 2*outer);

    int n = 0;
    for (UIWidget* w = container->getFirstWidget(); w; w = w->getNext()) {
        UIWidget::LayoutInfo winfo;
        w->getLayoutInfo(winfo);
        if (winfo.flags & UIWidget::li_NoLayout)
            continue;
        ++n;
        info.min_size.y  += winfo.min_size.y;
        info.min_size.x   = std::max(info.min_size.x, winfo.min_size.x);
        info.pref_size.y += winfo.pref_size.y;
        info.pref_size.x  = std::max(info.pref_size.x, winfo.pref_size.x);
    }
    if (n) {
        info.min_size.y  += (n-1) * space;
        info.pref_size.y += (n-1) * space;
    }
}

void
UIVBoxLayout::doLayoutWithHeights(UIWidgetContainer* container, GfxRect size,
                                  std::vector<int>& sizes, int space, int outer)
{
    std::vector<int>::size_type i = 0;
    size.y += outer;
    for (UIWidget* w = container->getFirstWidget(); w; w = w->getNext(), ++i) {
        UIWidget::LayoutInfo li;
        w->getLayoutInfo(li);
        if (li.flags & UIWidget::li_NoLayout)
            continue;
        w->setExtent(GfxRect(size.x, size.y, size.w, sizes[i]), true);
        w->doLayout();
        size.y += sizes[i];
        size.y += space;
    }
}
