/**
  *  \file ui/listbox.cc
  *  \brief Listbox Widgets
  *
  *  Widgets displaying (vertical) lists, and related things.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "ui/listbox.h"
#include "ui/uiglobal.h"
#include "ui/commands.h"
#include "gfx/palette.h"
#include "cpluslib/assert.h"
#include "ui/layout.h"
#include "ui/widgets.h"
#include "util/nls.h"
#include "gfx/gfxcompl.h"
#include "ui/util.h"

/****************************** UIScrollable *****************************/

/** \class UIScrollable
    \brief Scrollable thing

    This class provides basic operations on a scrollable thing.
    A scrollable thing has the following properties:

    - a current position (cursor)
    - a size
    - a window on the whole data set (top position, page size)

    \invariant 0 <= pos < size */

UIScrollable::UIScrollable(int id)
    : UIWidget(id), pos(0), top(0), size(0), page(1)
{ }

UIScrollable::~UIScrollable()
{ }

/** \function UIScrollable::scrollTo(int pos, int corr)
    Set current position. Derived classes must override this to
    implement the actual scrolling policy.
    \param pos   desired new selected element
    \param corr  (-1 or +1) if object pos can't be selected,
                 correct by going into specified direction */

/** Adjust current position to account for changes in underlying
    data. */
void
UIScrollable::adjustPosition()
{
    if (pos >= size)
        pos = size - 1;
    if (pos < 0)
        pos = 0;
    if (pos < top) {
        top = pos;
        if (top > 0 && !isAccessible(top-1))
            --top;
    }
    if (pos >= top + page)
        top = pos - page + 1;
}

/** Check whether element is accessible. Derived classes can override
    this to implement inaccessible list items.
    \return true if object can be selected by user
    \pre 0 <= arg < getSize() */
bool
UIScrollable::isAccessible(int)
{
    return true;
}

/******************************* UIListbox *******************************/

/** \class UIListbox
    \brief Listbox Widget

    This is a standard list box widget. It provides the basic user interface
    for scrolling around. Derived classes should override drawPart() to
    actually draw the list box.

    The default height of a list cell is the height of FONT_NORMAL; you
    can change that (from the constructor) using setItemHeight(). */

/** Constructor.
    \param id     widget Id
    \param lines  the preferred line count. May be -1 to use the
                  maximum (see setDefaultHeight()).
    \param width  the preferred line width in pixels */
UIListbox::UIListbox(int id, int lines, int width, int items)
    : UIScrollable(id), lines(lines), width(width),
      item_height(font_heights[FONT_NORMAL]), mdown(false)
{
    setSize(items);
    if (lines < 0)
        setDefaultHeight();
}

/** Compute the default (=maximum) height. This computes a nice value
    for /lines/ so that the widget fits on one screen. This does not
    immediately affect display; it just sets the parameters for the
    next layout run. */
void
UIListbox::setDefaultHeight()
{
    int max_lines = ui_root->getExtent().h / getItemHeight() - 10;

    lines = getSize();
    if (lines < 3)
        lines = 3;
    else if (lines > max_lines)
        lines = max_lines;
}

/** Scroll to position. This accounts for disabled items.
    \param pos    desired position
    \param corr   correction, +1 or -1. When the desired position is
                  not accessible, pos is adjusted in that direction. */
void
UIListbox::scrollTo(int pos, int corr)
{
    ASSERT(corr);
    if (pos >= getSize())
        pos = getSize() - 1;
    if (pos < 0)
        pos = 0;
    if (getSize() == 0)
        return;

    /* figure out new position */
    int n = 0;
    while (!isAccessible(pos)) {
        if (pos <= 0)
            corr = 1, ++n;
        else if (pos >= getSize()-1)
            corr = -1, ++n;
        if (n > 2)
            return;
        pos += corr;
    }

    ASSERT(pos >= 0 && pos < getSize() && isAccessible(pos));

    int ot = getTop(), op = getCurrent();
    setCurrent(pos);
    adjustPosition();
    if (ot == getTop() && op == getCurrent())
        return;

    if (!owner || !hasState(st_Exposed))
        return;

    sig_moved.raise(getId());
    if (ot == getTop()) {
        int a, b;
        Ptr<GfxCanvas> can = owner->getCanvasFor(this);
        can->lock();
        a = op - ot;
        b = getCurrent() - ot;
        if (a > b)
            std::swap(a, b);
        ++b;
        drawPart(can, a, b);
        can->unlock();
        can->update(GfxRect(extent.x, extent.y + a*item_height,
                            extent.w, (b-a)*item_height));
    } else {
        drawWidget(false);
    }
}

void
UIListbox::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    drawPart(can, 0, getPageSize());
    ui.update(getExtent());
}

void
UIListbox::stateChanged(int state, bool enable)
{
    UIScrollable::stateChanged(state, enable);
    if (state & st_Focused)
        updateCurrent();
}

bool
UIListbox::handleEvent(const UIEvent& e, bool second_pass)
{
    int fuzz;
    switch (e.type) {
     case ev_Mouse:
        if (hasState(st_Disabled))
            break;

        /* recognize clicks which are doubtlessly inside, and clicks which
           are inside with fuzz factor, for scrolling */
        fuzz = (second_pass && !hasState(lb_Menu)) ? 8 : 0;
        if ((e.mouse.b || hasState(lb_Menu))
            && e.mouse.x >= extent.x && e.mouse.x < extent.x + extent.w
            && e.mouse.y >= extent.y - fuzz && e.mouse.y < extent.y + extent.h + fuzz)
        {
            select();
            setFocus();

            if (e.mouse.y < extent.y) {
                /* scroll up */
                scrollTo(getTop()-1, -1);
            } else if (e.mouse.y >= extent.y + extent.h) {
                /* scroll down */
                scrollTo(getTop()+getPageSize(), +1);
            } else {
                int y = getTop() + (e.mouse.y - extent.y) / getItemHeight();
                if (isAccessible(y))
                    scrollTo(y, 1);
            }

            if (hasState(lb_Menu)) {
                /* menu: cursor follows mouse, single click posts event */
                if (!e.mouse.b) {
                    if (mdown)
                        sig_dblclick.raise(getId());
                    mdown = false;
                } else {
                    mdown = true;
                }
            } else if (e.mouse.b & mb_Right) {
                sig_menu.raise(getId());
            } else if (e.mouse.b & mb_Double) {
                sig_dblclick.raise(getId());
            }
            return true;
        }
        break;
     case ev_Key:
        if (hasState(st_Disabled) || !hasState(st_Focused))
            break;
        switch (e.key.code) {
         case SDLK_UP:
         case vk_WheelUp:
            select();
            scrollTo(getCurrent()-1, -1);
            return true;
         case SDLK_DOWN:
         case vk_WheelDown:
            select();
            scrollTo(getCurrent()+1, +1);
            return true;
         case SDLK_HOME:
            select();
            scrollTo(0, 1);
            return true;
         case SDLK_END:
            select();
            scrollTo(getSize()-1, -1);
            return true;
         case SDLK_PAGEUP:
            if (hasState(lb_NoPageKeys))
                break;
            /* FALLTHROUGH */
         case ss_Shift + SDLK_PAGEUP:
            select();
            scrollTo(getCurrent() - getPageSize(), -1);
            return true;
         case SDLK_PAGEDOWN:
            if (hasState(lb_NoPageKeys))
                break;
            /* FALLTHROUGH */
         case ss_Shift + SDLK_PAGEDOWN:
            select();
            scrollTo(getCurrent() + getPageSize(), -1);
            return true;
         case '#':
         case '\\':
         case SDLK_MENU:        // why not
            if (hasState(lb_KeyboardMenu)) {
                select();
                sig_menu.raise(getId());
                return true;
            }
            break;
            // c-pgup, c-pgdn, c-home, c-end
        }
        break;
     case ev_Command:
        break;
     default:;
    }
    return UIScrollable::handleEvent(e, second_pass);
}

void
UIListbox::getLayoutInfo(LayoutInfo& info)
{
    info.flags     = li_Flexible;
    info.min_size  = GfxPoint(width, item_height * 3);
    info.pref_size = GfxPoint(width, item_height * lines);
}

void
UIListbox::setExtent(GfxRect newpos, bool by_layout)
{
    setPageSize(newpos.h / item_height);

    if (getPageSize() >= getSize()) {
        setTop(0);
    } else {
        /* try to put Top so that Pos is in center */
        int n = getCurrent() - getPageSize() / 2;
        if (n < 0)
            setTop(0);
        else
            setTop(n);
    }
    adjustPosition();
    UIScrollable::setExtent(newpos, by_layout);
    sig_moved.raise(getId());
}

/** Redraw the current entry of the list box. */
void
UIListbox::updateCurrent()
{
    if (!owner || !hasState(st_Exposed))
        return;

    int a = getCurrent() - getTop(), b = a+1;
    Ptr<GfxCanvas> can = owner->getCanvasFor(this);
    can->lock();
    drawPart(can, a, b);
    can->unlock();
    can->update(GfxRect(extent.x, extent.y + a*item_height,
                        extent.w, (b-a)*item_height));
}

/****************************** UIStringList *****************************/

/** \class UIStringList
    \brief Ordered key/string pair list

    This provides a container of pairs of keys (integers) and strings,
    to be used for list boxes and similar things. */

UIStringList::UIStringList() { }

UIStringList::~UIStringList()
{ }

void
UIStringList::add(int key, const string_t& s)
{
    data.push_back(std::make_pair(key, s));
}

void
UIStringList::add(int key, const char* s)
{
    add(key, string_t(s));
}

/** Get item at specified position.
    \param index position, [0, getSize()).
    \returns iterator, or end() if index too big. */
UIStringList::iterator
UIStringList::getItem(int index) const
{
    iterator i = data.begin();
    while (index > 0 && i != data.end())
        ++i, --index;
    return i;
}

/** Get item by key.
    \returns iterator, end() if not found.
    \post ret != end() <=> ret->first == key */
UIStringList::iterator
UIStringList::getItemByKey(int key) const
{
    iterator i = data.begin();
    while (i != data.end() && i->first != key)
        ++i;
    return i;
}

/** Get item by key. Like getItemByKey, but this one returns an index,
    not an iterator.
    \returns index, -1 if not found.
    \post ret != -1 <=> getItem(ret)->first == key */
int
UIStringList::getItemIndexByKey(int key) const
{
    iterator i = data.begin();
    int n = 0;
    while (i != data.end() && i->first != key)
        ++i, ++n;
    return (i == data.end() ? -1 : n);
}

/*************************** UIStandardListbox ***************************/

UIStandardListbox::UIStandardListbox(int id, int lines, int width, UIStringList& data)
    : UIListbox(id, lines, width, data.getCount()), data(data)
{ }

UIStandardListbox::~UIStandardListbox()
{ }

void
UIStandardListbox::drawPart(Ptr<GfxCanvas> can, int from, int to)
{
    GfxParameters parm;

    UIStringList::iterator it = data.getItem(from + getTop());
    int y = extent.y + from * getItemHeight();
    for (int i = from; i < to; ++i) {
        if (i + getTop() == getCurrent()) {
            if (hasState(st_Focused)) {
                parm.color   = standard_colors[COLOR_WHITE];
                parm.bgcolor = standard_colors[COLOR_BLACK];
            } else {
                parm.color   = standard_colors[COLOR_WHITE];
                parm.bgcolor = standard_colors[COLOR_GRAY];
            }
        } else {
            parm.color   = standard_colors[COLOR_BLACK];
            parm.bgcolor = standard_colors[COLOR_GRAY];
        }
        if (it != data.end()) {
            outTextF(can, parm, fonts[FONT_NORMAL],
                     extent.x, y, extent.w, " " + it->second);
            ++it;
        } else {
            outTextF(can, parm, fonts[FONT_NORMAL],
                     extent.x, y, extent.w, "");
        }
        y += getItemHeight();
    }
}

void
UIStandardListbox::scrollToKey(int key)
{
    int n = data.getItemIndexByKey(key);
    if (n >= 0)
        scrollTo(n, 1);
}

bool
UIListbox::doStandardListbox(string_t title, int help_id)
{
    UIWindow w(title, UIWindow::BLUE_WINDOW, COLOR_GRAY);
    w.setLayout(new UIVBoxLayout(5));
    UIGroup h(new UIHBoxLayout(5));
    UIButton b_ok(cm_Enter,   _("OK"), SDLK_RETURN, UIButton::aStop);
    UIButton b_esc(cm_Escape, _("Cancel"), SDLK_ESCAPE, UIButton::aStop);
    UIButton b_help(cm_Help, _("H"), 'H');
    UISpacer sp;
    
    w.add(*this);
    w.add(h);
    h.add(b_ok);
    h.add(b_esc);
    h.add(sp);
    if (help_id)                // FIXME: not handled yet
        h.add(b_help);
    w.pack();
    w.setState(UIWindow::st_Modal, true);
    ui_root->centerWidget(w);
    ui_root->add(w);
    int rv = ui_root->run();
    ui_root->close(w);
    return (rv == cm_Enter);
}

namespace {
    class MenuFrame : public UIFrameGroup {
        int target_height;
        long next_tick;
        enum { TIMER_DIVI = 0 };
     public:
        MenuFrame(int target_height);
        ~MenuFrame();
        void drawInterior(Ptr<GfxCanvas> can, UpdateInfo& ui);
        bool handleEvent(const UIEvent& e, bool second_pass);
        void stopValid(int) { stop(1); }
    };
}

MenuFrame::MenuFrame(int target_height)
    : UIFrameGroup(new UIHBoxLayout(0), 2), target_height(target_height),
      next_tick(getTimer() + TIMER_DIVI)
{ }

MenuFrame::~MenuFrame()
{ }

void
MenuFrame::drawInterior(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    GfxRect out = getExtent(), in = getClientRectFor(getExtent());
    in.grow(1, 1);
    
    drawFrameUp(can, out);
    drawFrameDown(can, in);
    ui.update(getExtent());
}

bool
MenuFrame::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Tick && getTimer() > next_tick) {
        next_tick = getTimer() + TIMER_DIVI;
        GfxRect ex = getExtent();
        int     scr_height = ui_root->getExtent().h;
        if (ex.h < target_height) {
            ex.h += font_heights[FONT_NORMAL];
            if (ex.h > target_height)
                ex.h = target_height;
            if (ex.y + ex.h > scr_height)
                ex.y = scr_height - ex.h;
            setExtent(ex);
            doLayout();
        }
    } else if (e.type == ev_Key) {
        if (e.key.code == SDLK_ESCAPE) {
            stop(0);
            return true;
        } else if (e.key.code == SDLK_RETURN) {
            stop(1);
            return true;
        }
    }
    return UIFrameGroup::handleEvent(e, second_pass);
}

bool
UIListbox::doStandardMenu(int x, int y)
{
    MenuFrame mf(font_heights[FONT_NORMAL]*getSize() + 4 /* FIXME */);
    void* keks = sig_dblclick.add(&mf, &MenuFrame::stopValid);
    setState(lb_Menu, true);
    mf.add(*this);
    mf.setExtent(GfxRect(x, y, width+4, font_heights[FONT_NORMAL] + 4));
    mf.doLayout();
    ui_root->add(mf);
    int i = ui_root->run();
    ui_root->close(mf);
    sig_dblclick.remove(keks);// FIXME: should be automated
    return i != 0;
}

bool
UIListbox::doStandardMenu(const UIEvent& ref_event)
{
    TReferencePointArg rpa(ref_event);
    UIEvent  ev;
    ev.type         = ev_Command;
    ev.cmd.code     = cm_INT_QueryReferencePoint;
    ev.cmd.arg      = 0;
    ev.cmd.sdl_type = 0;
    ev.cmd.info     = 0;
    ev.cmd.info2    = 0;
    ev.cmd.infop    = &rpa;
    ui_root->handleEvent(ev, false) || ui_root->handleEvent(ev, true);
    // FIXME: if it's a button, push it
    bool rv;
    if (rpa.widget) {
        const GfxRect& p = rpa.widget->getExtent();
        rv = doStandardMenu(p.x, p.y + p.h);
    } else {
        rv = doStandardMenu(rpa.x, rpa.y);
    }
    // FIXME: if it's a button, release it
    return rv;
}

/****************************** UIScrollbar ******************************/

UIScrollbar::UIScrollbar(UIScrollable& scrl)
    : UIWidget(), scrl(scrl), cookie(scrl.sig_moved.add(this, &UIScrollbar::scrolled))
{
    sync();
}

UIScrollbar::~UIScrollbar()
{
    scrl.sig_moved.remove(cookie);
}

void
UIScrollbar::scrolled(int)
{
    drawWidget(false);
}

void
UIScrollbar::sync()
{
    if (scrl.getPageSize() >= scrl.getSize()) {
        /* no scrollbar needed */
        bf[0] = bf[1] = st_Disabled;
    } else {
        /* scrollbar needed */
        bf[0] = bf[1] = 0;
    }
    drawWidget(false);
}

void
UIScrollbar::setBF(int id, int flags)
{
    if (flags != bf[id]) {
        bf[id] = flags;
        drawWidget(false);
    }
}

void
UIScrollbar::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    UIButton::drawButton(can, GfxRect(extent.x, extent.y,
                                      extent.w, 20), "\036", bf[0]);
    UIButton::drawButton(can, GfxRect(extent.x, extent.y + extent.h - 20,
                                      extent.w, 20), "\037", bf[1]);

    /* scrollbox */
    GfxRect interior = extent;
    interior.grow(0, -20);
    drawFrameUp(can, interior);
    interior.grow(-1, -1);

    if (scrl.getSize() > scrl.getPageSize()) {
        /* draw scroll box */
        int top = scrl.getTop();
        if (top + scrl.getPageSize() > scrl.getSize())
            top = scrl.getSize() - scrl.getPageSize();
        int divi = scrl.getSize();

        int yy1 = interior.h * int32(top) / divi;
        int yy2 = interior.h * int32(top + scrl.getPageSize()) / divi;

        if (yy1 > 0)
            drawSolidBar(can, GfxRect(interior.x, interior.y, interior.w, yy1),
                         standard_colors[COLOR_SHIELD + 3]);
        drawFrameUp(can, GfxRect(interior.x, interior.y + yy1, interior.w, yy2 - yy1));
        if (yy1 < yy2-2)
            drawSolidBar(can, GfxRect(interior.x+1, interior.y + yy1+1,
                                      interior.w-2, yy2-yy1-2),
                         standard_colors[COLOR_SHIELD + 1]);
        if (yy2 < interior.h)
            drawSolidBar(can, GfxRect(interior.x, interior.y+yy2,
                                      interior.w, interior.h-yy2),
                         standard_colors[COLOR_SHIELD + 3]);
    } else {
        drawSolidBar(can, interior, standard_colors[COLOR_SHIELD + 5]);
    }

    ui.update(extent);
}

void
UIScrollbar::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_FlexibleV;
    info.min_size = GfxPoint(20, 60);
    info.pref_size = GfxPoint(20, 80);
}

bool
UIScrollbar::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Mouse) {
        int old_bf1 = bf[0], old_bf2 = bf[1];
        bf[0] &= ~(st_Selected | UIButton::bf_Pressed);
        bf[1] &= ~(st_Selected | UIButton::bf_Pressed);
        if (isInside(e.mouse.x, e.mouse.y)) {
            int where = -1;
            if (e.mouse.y < extent.y + 20)
                where = 0;
            else if (e.mouse.y >= extent.y + extent.h - 20)
                where = 1;

            if (where >= 0) {
                select();
                bf[where] |= st_Selected;
                if (e.mouse.b)
                    bf[where] |= UIButton::bf_Pressed;
            }
        }
        if (bf[0] != old_bf1 || bf[1] != old_bf2) {
            drawWidget(false);
            return true;
        }
        /* FALLTHROUGH */
    } else if (e.type == ev_Tick && hasState(st_Selected)) {
        if (bf[0] & UIButton::bf_Pressed) {
            scrl.setFocus();
            scrl.scrollTo(scrl.getTop() - 1, -1);
        } else if (bf[1] & UIButton::bf_Pressed) {
            scrl.setFocus();
            scrl.scrollTo(scrl.getTop() + scrl.getPageSize(), +1);
        }
        /* FALLTHROUGH */
    }
    return UIWidget::handleEvent(e, second_pass);
}

void
UIScrollbar::stateChanged(int state, bool enable)
{
    if ((state & st_Selected) && !enable) {
        setBF(0, bf[0] & ~(st_Selected | UIButton::bf_Pressed));
        setBF(1, bf[1] & ~(st_Selected | UIButton::bf_Pressed));
    }
    UIWidget::stateChanged(state, enable);
}
