/**
  *  \file ui/ui.cc
  *  \brief Simple User Interface Kernel
  *
  *  This module contains the user interface kernel: basic widget and
  *  group class, plus ultra-simple widgets.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <sstream>
#include <climits>
#include <typeinfo>             // Debugging only
#include "ui/ui.h"
#include "gfx/multiclip.h"
#include "ui/layout.h"
#include "cpluslib/stringcv.h"
#include "cpluslib/assert.h"

/** \class UIWidget
    \brief Widget

    This is a basic widget of the user interface system.

    - can handle events
    - can redraw itself
    - resizable, movable */

UIWidget::UIWidget()
    : extent(), owner(0), next(0), prev(0), status(0), options(0), id(0)
{ }

UIWidget::UIWidget(int aid)
    : extent(), owner(0), next(0), prev(0), status(0), options(0), id(aid)
{ }

UIWidget::~UIWidget()
{
    /* say goodbye */
    if (getOwner())
        getOwner()->removeWidget(*this);
}

/** Set status. Call this to change the status flags. Derived widgets
    can override stateChanged() to react on status changes. */
void
UIWidget::setState(int astat, bool enable)
{
    int old = status;

    if (enable)
        status |= astat;
    else
        status &= ~astat;

    old ^= status;
    if (old)
        stateChanged(old, enable);
}

/** State changed. Called by setState() after the widget status changed.
    \param bits     the status flags that changed
    \param enabled  true iff the flags were set, false if cleared
    \default propagate idle-notify events, propagate redraw requests. */
void
UIWidget::stateChanged(int bits, bool enabled)
{
    if (bits & st_IdleNotify)
        if (owner)
            owner->setIdleNotify(enabled);
    if (enabled && (bits & (st_WantRedraw | st_ChildRedraw)))
        if (owner)
            owner->setState(st_ChildRedraw, true);
}

/** \fn UIWidget::draw(Ptr<GfxCanvas> c, UpdateInfo& ui)
    \brief Draw widget

    This function redraws the widget. Derived classes must override
    it to provide actual visual layout.

    \param c  locked canvas; widget can immediately be drawn
    \param ui used to manage screen update. Widgets should use
              ui.update() to update dirty screen regions after
              drawing.

    \todo Maybe it's better to do update scheduling in a specialized
    canvas? (i.e. collect all GfxCanvas::update requests and perform
    them in batch; not in UpdateInfo but in a canvas) */

/** Draw widget. This function calls draw() with the right parameters.
    When you need to update a widget, call this function.

    \param immed=true  call draw() immediately
    \param immed=false arrange for draw() to be called in the future
                       (next run through user event loop actually)

    Most applications use immed=false because that optimizes redraw and
    combines successive requests. */
void
UIWidget::drawWidget(bool immed)
{
    if (!hasState(st_Exposed) || !owner)
        return;

    if (immed) {
        Ptr<GfxCanvas> can = owner->getCanvasFor(this);
        if (!can.ptr())
            return;

        UpdateInfo ui (can->getSurface());

        can->lock();
        draw(can, ui);
        can->unlock();
        ui.doUpdates();
    } else {
        setState(st_WantRedraw, true);
    }
}

/** Preform deferred draw. If the widget requested to be redrawn
    (drawWidget(false)), do it now. */
void
UIWidget::drawDeferred(Ptr<GfxCanvas> canvas, UpdateInfo& ui)
{
    if (hasState(st_WantRedraw)) {
        setState(st_WantRedraw, false);
        draw(canvas, ui);
    }
}

/** Event handler. Process the event (if wanted).
    \param event the event structure
    \param second_pass event handlers are called twice. The first run
    is intended to deliver user input to the focused widget.
    The second one allows other widgets also receive events.

    \returns true iff the event has been handled. For broadcast
    events (ev_Broadcast), return false.
    \default returns false */
bool
UIWidget::handleEvent(const UIEvent& event, bool second_pass)
{
    (void) &event;
    (void) &second_pass;

    return false;
}

/** Check whether coordinates are inside. Widgets that are not strictly
    rectangular can override this. */
bool
UIWidget::isInside(int x, int y)
{
    return extent.contains(x, y);
}

void
UIWidget::UpdateInfo::update(const GfxRect& r)
{
    // FIXME: unlock etc.
    GfxRect r1 = r;
    r1.clipSurface(sfc);
    if (r1.w && r1.h)
        SDL_UpdateRect(sfc, r1.x, r1.y, r1.w, r1.h);
}

UIWidget::LayoutInfo::LayoutInfo()
    : flags(0), min_size(), pref_size()
{ }

void
UIWidget::setLinks(UIWidget* prev, UIWidget* next)
{
    this->prev = prev;
    this->next = next;
}

/** Get widget by Id.
    \default returns this widget if id matches, 0 otherwise.

    Group widgets must override this to recurse through the widget
    hierarchy and return the appropriate widget.

    Note: can't be const because it might return /this/. */
UIWidget*
UIWidget::getWidgetById(int id)
{
    if (id == this->id)
        return this;
    else
        return 0;
}

/** Get focused widget. When called on a focused group, returns the
    innermost focused child widget.

    Note: can't be const because it might return /this/. */
UIWidget*
UIWidget::getFocusedWidget()
{
    if (hasState(st_Focused))
        return this;
    else
        return 0;
}

/** Run. Starts an event loop. The root group must overwrite this
    method.
    \default hands the request up to the owner.
    \returns the value that was passed to stop() */
int
UIWidget::run()
{
    ASSERT(owner);
    return owner->run();
}

/** Stop. Stops the enclosing event loop and returns the specified
    argument. The root group must override this method.
    \default hands the request up to the owner */
void
UIWidget::stop(int arg)
{
    ASSERT(owner);
    owner->stop(arg);
}

/** Obtain logical focus. You must call this function prior
    reacting to user actions. A well-behaved widget
    - calls select(), when the user talks to it. This sets this
      widget's st_Selected status flag;
    - possibly overrides stateChanged() to react when it loses st_Selected.

    This is to prevent things like two buttons being depressed
    simultaneously because one of them did not notice that the user
    moved the mouse out of it. */
void
UIWidget::select()
{
    if (owner) {
        owner->selectWidget(this);
    } else {
        /* if there's noone else we have to take care of... */
        setState(st_Selected, true);
    }
    ASSERT(hasState(st_Selected));
}

/** Obtain physical (keyboard) focus. */
void
UIWidget::setFocus()
{
    owner->setFocus(this);
}

/** Convert widget to string representation. For debugging. */
string_t
UIWidget::toString()
{
    std::ostringstream s;
    s << "{ extent = " << extent << ", id = " << id << "}";
    return Object::toString() + toStringType(s.str());
}

/** Compute Optimum Layout (Pack Widgets).
    Groups that provide a layout override this.
    \default does nothing. */
void
UIWidget::pack()   // FIXME: do we actually need this or does doLayout suffice?
{ }

void
UIWidget::doLayout()
{ }

/** Change Widget Size. This is a shorthand for setExtent(). */
void
UIWidget::setSize(int w, int h)
{
    setExtent(GfxRect(extent.x, extent.y, w, h));
}

void
UIWidget::moveTo(int x, int y)
{
    setExtent(GfxRect(x, y, extent.w, extent.h));
}

/** Change Widget Position. Moves and/or resizes the widget
    to the specified rectangle.

    \param by_layout true iff the caller is the parent's layout manager.
                     You usually pass false here. */
void
UIWidget::setExtent(GfxRect newpos, bool by_layout)
{
    GfxRect oldsize = extent;
    extent = newpos;
    if (!by_layout && owner)
        owner->handleChildResize(this, oldsize);
}

/***************************** UIFixedWidget *****************************/

/** \class UIFixedWidget
    \brief Fixed-size widget

    Widgets which have a fixed size and position (i.e. which want to be
    ignored by layout managers) can derive from this class. It is identical
    to UIWidget, but provides a pre-implemented getLayoutInfo() function. */

UIFixedWidget::~UIFixedWidget()
{ }

void
UIFixedWidget::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_NoLayout;
}

/*************************** UIWidgetContainer ***************************/

/** \class UIWidgetContainer
    \brief Widget Container

    This class provides a container for grouping widgets: the
    mechanism to build widget trees. Derived classes must provide the
    actual look and feel. */

UIWidgetContainer::UIWidgetContainer(int aid)
    : UIWidget(aid), first(0), last(0), selected(0), focused(0)
{ }

UIWidgetContainer::~UIWidgetContainer()
{
    /* remove exposed-ness from everyone, so noone gets the weird idea
       of drawing something on the screen. */
    setState(st_Exposed, false);
    UIWidget* p = first;
    while (p) {
        UIWidget* pp = p;
        p = p->getNext();
        pp->setOwner(0);
        pp->setLinks(0, 0);
        pp->setState(st_IdleNotify | st_Focused | st_Selected, false);
    }
}

/** Give logical focus to widget.
    If w points to a widget, that widget wants to get the input focus.
    Take it away from the one which currently has it, and give it to
    the specified widget. Also grab the logical focus for this
    UIWidgetContainer, to make this process happen recursively.

    If w is null, take the input focus away from whoever currently
    has it.

    \pre w != 0 => w->getParent() == this */
void
UIWidgetContainer::selectWidget(UIWidget* w)
{
    if (w == selected)
        return;

    /* we're going to change something. */
    if (!w) {
        /* noone wants to be selected. */
        selected->setState(st_Selected, false);
        selected = 0;
    } else {
        select();
        if (selected)
            selected->setState(st_Selected, false);
        selected = w;
        selected->setState(st_Selected, true);
    }
}

/** Give physical input focus to widget.
    If w points to a widget, that widget wants to get the physical
    input focus. Also grab the focus for this UIWidgetContainer, to
    make this process happen recursively.

    If w is null, take the input focus away from whoever currently
    has it, but keep this group focused.

    \pre w != 0 => w->getParent() == this */
void
UIWidgetContainer::setFocus(UIWidget* w)
{
    if (!w) {
        /* noone wants to be selected */
        if (focused)
            focused->setState(st_Focused, false);
        focused = 0;
    } else {
        /* get focus for us */
        if (owner)
            owner->setFocus(this);

        if (focused != w) {
            if (focused)
                focused->setState(st_Focused, false);
            focused = w;
            focused->setState(st_Focused, true);
        }
    }
}

/** Get widget by id. This function is the same for all derived classes.
    \returns pointer to the requested widget, or null. */
UIWidget*
UIWidgetContainer::getWidgetById(int id)
{
    if (id == this->id)
        return this;
    for (UIWidget* p = first; p != 0; p = p->getNext())
        if (UIWidget* q = p->getWidgetById(id))
            return q;
    return 0;
}

/** Get focused widget. */
UIWidget*
UIWidgetContainer::getFocusedWidget()
{
    if (hasState(st_Focused))
        if (focused)
            return focused->getFocusedWidget();
        else
            return this;
    else
        return 0;
}

/** Get client rectangle for extent.
    \returns coordinates usable by client widgets if size were /ext/.
             This will be the size of the UIWidgetContainer, sans some
             decorative borders.
    \default returns ext, i.e. the whole area is usable.
    \inv when r' = getClientRectFor(getExtentFor(r)), r' should equal r
    (possibly slightly larger due to rounding effects) */
GfxRect
UIWidgetContainer::getClientRectFor(GfxRect ext)
{
    return ext;
}

/** Get extent for client rectangle.
    \returns extent value needed to have a client area sized /cli/.
    \default returns cli, i.e. client area is same as extent
    \inv when r' = getExtentFor(getClientRectFor(r)), r' should equal r
    (possibly slightly larger due to rounding effects) */
GfxRect
UIWidgetContainer::getExtentFor(GfxRect cli)
{
    return cli;
}

/** Add widget. Derived classes must override this method to do the
    actual add operation (i.e. inserting into widget chain).

    \default does the "red tape" operations only: set the widget's
    owner and status, and draw it. This does not put the widget on
    the widget chain. */
void
UIWidgetContainer::add(UIWidget& widget)
{
    ASSERT(&widget);
    widget.setOwner(this);
    if (!focused && !widget.hasState(st_Disabled)) {
        focused = &widget;
        widget.setState(st_Focused, hasState(st_Focused));
    } else {
        widget.setState(st_Focused, false);
    }
    if (widget.hasState(st_IdleNotify))
        setIdleNotify(true);

    if (hasState(st_Exposed)) {
        widget.setState(st_Exposed, true);
        widget.drawWidget(true);
    }
}

/** Remove widget. Derived classes must override this method to handle
    possible redraws.

    \default does the "red tape" operations only: unchain the widget and
    change status. */
void
UIWidgetContainer::removeWidget(UIWidget& w)
{
    ASSERT(&w);
    ASSERT(w.getOwner() == this);
    if (UIWidget* p = w.getPrev())
        p->setLinks(p->getPrev(), w.getNext());
    if (UIWidget* n = w.getNext())
        n->setLinks(w.getPrev(), n->getNext());
    if (&w == first)
        first = w.getNext();
    if (&w == last)
        last = w.getPrev();

    if (selected == &w)
        selected = 0;
    if (focused == &w) {
        focused = 0;
        for (UIWidget* p = first; p != 0; p = p->getNext())
            if (!p->hasState(st_Disabled))
                setFocus(p);
    }

    /* take away exposedness first so it doesn't try to paint */
    w.setState(st_Exposed, false);

    w.setLinks(0, 0);
    w.setOwner(0);
    if (w.hasState(st_IdleNotify))
        setIdleNotify(false);

    /* take away other flags in case widget survives a while. Do that
       after removing the owner so that it doesn't waste time trying
       to update its status */
    w.setState(st_Selected | st_Focused | st_IdleNotify, false);
}

/** Status change. When we lose the logical focus, take it away
    from our child which has it. */
void
UIWidgetContainer::stateChanged(int bits, bool enable)
{
    UIWidget::stateChanged(bits, enable);
    if ((bits & st_Selected) && !enable) {
        /* we lost the selection. Take it away from our child which has it. */
        selectWidget(0);
    }
    if ((bits & st_Focused)) {
        /* we lost or got the focus. We do remember who has the focus,
           so we can give it back to him when regaining focus. */
        if (focused)
            focused->setState(st_Focused, enable);
    }

    if (bits & st_Exposed) {
        /* propagate visibility to children */
        for (UIWidget* p = first; p; p = p->getNext())
            p->setState(st_Exposed, enable);
    }

    /* we're clearing the redraw status. This means we must be inside
       drawDeferred() and our children are going to be redrawn soon.
       So they need not go through drawDeferred() themselves. */
    if (!enable && (bits & (st_WantRedraw | st_ChildRedraw)))
        for (UIWidget* p = first; p; p = p->getNext())
            p->setState(st_WantRedraw | st_ChildRedraw, false);
}

/** Child st_IdleNotify change. UIWidget::setState calls this function
    when the st_IdleNotify status changes.
    \param enable=true flag was set. This means that this container's
                       st_IdleNotify must also be set.
    \param enable=false flag was cleared. We must check whether we need
                       it, too (if there are some widgets with
                       st_IdleNotify remaining). */
void
UIWidgetContainer::setIdleNotify(bool enable)
{
    if (enable) {
        /* easy way. */
        setState(st_IdleNotify, true);
    } else {
        /* someone turned off Idle Notify. */
        for (UIWidget* p = first; p; p = p->getNext()) {
            if (p->hasState(st_IdleNotify)) {
                setState(st_IdleNotify, true);
                return;
            }
        }
        setState(st_IdleNotify, false);
    }
}

/** Deferred redraw. The main loop calls this to process st_WantRedraw
    and st_ChildRedraw requests. */
void
UIWidgetContainer::drawDeferred(Ptr<GfxCanvas> canvas, UpdateInfo& ui)
{
    if (hasState(st_WantRedraw)) {
        draw(canvas, ui);
        setState(st_WantRedraw | st_ChildRedraw, false);
    } else if (hasState(st_ChildRedraw)) {
        for (UIWidget* p = first; p; p = p->getNext())
            // FIXME: this can be improved a lot: we ignore the canvas
            // we got and re-fetch a new one
            if (p->hasState(st_WantRedraw | st_ChildRedraw))
                p->drawDeferred(getCanvasFor(p), ui);
        setState(st_WantRedraw | st_ChildRedraw, false);
    }
}

/** Get string representation. Contains a nicely-indented list of
    child widgets. */
string_t
UIWidgetContainer::toString()
{
    string_t children;
    for (UIWidget* p = first; p; p = p->getNext()) {
        children += '\n';
        children += p->toString();
    }

    return UIWidget::toString() + "{" + strIndent("  ", children) + "}";
}

/** Change status of a widget. Locates the widget with identifier id
    and calls its setState(). Does not fail if there is no widget /id/. */
void
UIWidgetContainer::setStateOf(int id, int astat, bool enable)
{
    if (UIWidget* w = getWidgetById(id))
        w->setState(astat, enable);
}

//////////////////////////////////////////////////////////////////////

/** \class UIGroup
    \brief Widget Group

    A group of non-overlapping widgets.

    By default, groups have no visual appearance of their own, and can
    be nested without negative effects. */

UIGroup::UIGroup(int aid)
    : UIWidgetContainer(aid)
{ }

UIGroup::UIGroup(Ptr<UILayoutManager> alayout, int aid)
    : UIWidgetContainer(aid),
      layout(alayout)
{ }

UIGroup::~UIGroup()
{ }

/** Add widget with specified coordinates. */
void
UIGroup::add(const GfxRect& pos, UIWidget& widget)
{
    widget.setExtent(pos, true);
    add(widget);
}

/** Add widget to group. */
void
UIGroup::add(UIWidget& widget)
{
//     if (!widget)
//         return;

    /* Insert widget into list */
    if (!first) {
        first = last = &widget;
        widget.setLinks(0, 0);
    } else {
        last->setLinks(last->getPrev(), &widget);
        widget.setLinks(last, 0);
        last = &widget;
    }

    /* red tape */
    UIWidgetContainer::add(widget);
}

void
UIGroup::removeWidget(UIWidget& w)
{
    UIWidgetContainer::removeWidget(w);
    drawWidget(false);            // simple & effective
}

/** Draw interior. This function should draw the interior of the
    group. The interior is logically below the widgets contained in
    the group. If you wish the group to have some background,
    overload this method.
    \default does nothing, i.e. group has no visual appearance of its own. */
void
UIGroup::drawInterior(Ptr<GfxCanvas> /*canvas*/, UpdateInfo& /*ui*/)
{ }

void
UIGroup::draw(Ptr<GfxCanvas> canvas, UpdateInfo& ui)
{
    drawInterior(canvas, ui);
    for (UIWidget* p = first; p; p = p->getNext())
        p->draw(canvas, ui);
}

void
UIGroup::drawBelow(UIWidget* /*child*/, Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    /* Only the interior is topologically below all children. */
    drawInterior(can, ui);
}

/** Get layout manager. */
Ptr<UILayoutManager>
UIGroup::getLayout() const
{
    return layout;
}

/** Set layout manager. This will affect future layout operations. */
void
UIGroup::setLayout(Ptr<UILayoutManager> alay)
{
    layout = alay;
}

/** Process events. Sends the events to all child widgets, first to
    the logically selected one. Returns true if one of the children
    handled the event. */
bool
UIGroup::handleEvent(const UIEvent& event, bool second_pass)
{
    /* save current selection. It might change during event processing
       (broadcast) */
    UIWidget* sel = selected;

    if (sel)
        if (sel->handleEvent(event, second_pass))
            return true;

    for (UIWidget* p = first; p; p = p->getNext())
        if (p != sel)
            if (p->handleEvent(event, second_pass))
                return true;

    return UIWidget::handleEvent(event, second_pass);
}

Ptr<GfxCanvas>
UIGroup::getCanvasFor(UIWidget* /*child*/)
{
    return owner->getCanvasFor(this);
}

/** Handle Child Resize. */
void
UIGroup::handleChildResize(UIWidget* child, GfxRect oldsize)
{
    if (!hasState(st_Exposed))
        return;

    if (child->getExtent().contains(oldsize)) {
        /* child completely covers old position. */
        child->drawWidget(true);
    } else {
        /* child has moved. Complete redraw in case something previously
           obscured (i.e. background) now becomes visible. */
        Ptr<GfxMultiClip> mc = new GfxMultiClip(owner->getCanvasFor(this));
        mc->add(child->getExtent());
        mc->add(oldsize);

        UpdateInfo ui(mc->getSurface());
        mc->lock();
        draw(mc, ui);
        mc->unlock();
        ui.doUpdates();
    }
}

void
UIGroup::setExtent(GfxRect r, bool by_layout)
{
    /* Move children */
    int dx = r.x - extent.x;
    int dy = r.y - extent.y;

    for (UIWidget* p = first; p; p = p->getNext()) {
        GfxRect child_pos = p->getExtent();
        child_pos.moveBy(dx, dy);
        p->setExtent(child_pos, true);
    }

    /* Move the group. This will redraw the whole widget. */
    UIWidgetContainer::setExtent(r, by_layout);
}

static GfxPoint
expand(UIGroup* g, GfxPoint pt)
{
    GfxRect r(0, 0, pt.x, pt.y);
    r = g->getExtentFor(r);
    return GfxPoint(r.w, r.h);
}

void
UIGroup::getLayoutInfo(LayoutInfo& i)
{
    if (layout) {
        layout->getLayoutInfo(this, i);
        i.pref_size = expand(this, i.pref_size);
        i.min_size  = expand(this, i.min_size);
    } else {
        i.flags = li_NoLayout;
    }
}

void
UIGroup::pack()
{
    if (layout) {
        LayoutInfo info;
        layout->getLayoutInfo(this, info);
        GfxRect new_int (0, 0, info.pref_size.x, info.pref_size.y);
        GfxRect new_ext = getExtentFor(new_int);
        new_int.moveBy(new_ext.moveTo(extent.x, extent.y));
        layout->doLayout(this, new_int);
        setExtent(new_ext, true);
        drawWidget(false);
    }
}

void
UIGroup::doLayout()
{
    if (layout) {
        layout->doLayout(this, getClientRectFor(getExtent()));
        drawWidget(false);
    }
}

/******************************** UISpacer *******************************/

/** \class UISpacer
    \brief Spacer

    An invisible spacer. When put into a group with other widgets,
    spacers will expand to fill available space. Spacers themselves
    are invisible. */

UISpacer::UISpacer()
    : UIWidget()
{ }

void
UISpacer::draw(Ptr<GfxCanvas> /*sfc*/, UpdateInfo& /*ui*/)
{ }

void
UISpacer::getLayoutInfo(LayoutInfo& i)
{
    i.flags = li_Flexible;
    i.min_size = i.pref_size = GfxPoint(0,0);
}
