/**
  *  \file ui/ui.h
  *  \brief Simple User Interface Kernel
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UI_UI_H
#define PCC_V2_UI_UI_H

#include <SDL_events.h>
#include <vector>
#include "gfx/gfx.h"
#include "ui/event.h"
#include "cpluslib/misc.h"
#include "cpluslib/string.h"
#include "cpluslib/smartptr.h"

class UIWidgetContainer;
class UILayoutManager;

class UIWidget : public Object {
 protected:
    GfxRect            extent;  /**< Position and size of widget. Absolute coordinates. */
    UIWidgetContainer* owner;   /**< Owner of this widget. */
    UIWidget*          next;
    UIWidget*          prev;
    
    int       status;           /**< Status flags. These change during lifetime. */
    int       options;          /**< Options. */
    int       id;               /**< Identifier. */

    struct UpdateInfo;
 public:
    struct LayoutInfo {
        int      flags;
        /** Minimum size. Layout manager should make sure widget does
            not get smaller than this. */
        GfxPoint min_size;
        /** Preferred size. When asked to compute the optimum size (pack()),
            all widgets come out with this size if possible. */
        GfxPoint pref_size;
        LayoutInfo();
    };
    enum {
        /** Widget does not want to be layouted. Layout managers will
            ignore it. */
        li_NoLayout = 1,
        /** Widget is flexible. It might be enlarged when there is
            more space than needed. */
        li_FlexibleH = 2,
        li_FlexibleV = 4,
        li_Flexible  = 6
    };
    enum {
        st_Exposed     = 1,     ///< Component is visible.
        st_Focused     = 2,     ///< Component has physical input focus.
        st_Selected    = 4,     ///< Component has logical input focus.
        st_Disabled    = 8,     ///< Component cannot be selected.
        st_IdleNotify  = 16,    ///< Component wants to receive `Idle' events. This flag only works for leaf widgets, not for UIWidgetContainer descendants.
        st_Modal       = 32,    ///< Component is modal.
        st_WantRedraw  = 64,    ///< Component wishes to be redrawn.
        st_ChildRedraw = 128,   ///< Component has a child that wishes to be redrawn.

        st_Last        = 128
    };
    Ptr<Object> layout_arg;     /** Layout argument. */

    UIWidget();
    UIWidget(int id);
    virtual ~UIWidget();

    /* Look & Feel */
    void         drawWidget(bool immediate);
    virtual void draw(Ptr<GfxCanvas> canvas, UpdateInfo& ui) = 0;
    virtual void drawDeferred(Ptr<GfxCanvas> canvas, UpdateInfo& ui);
    virtual bool handleEvent(const UIEvent& event, bool second_pass);
    virtual bool isInside(int x, int y);

    /* Status Management */
    int          getState() const { return status; }
    bool         hasState(int what) const { return (status & what) != 0; }
    void         setState(int astat, bool enable);
    virtual void stateChanged(int bits, bool enabled);

    void         setFocus();
    void         select();

    /* Administrivia */
    void                  setOwner(UIWidgetContainer* g) { owner = g; }
    UIWidgetContainer*    getOwner() const { return owner; }
    UIWidget*             getNext() const { return next; }
    UIWidget*             getPrev() const { return prev; }
    void                  setLinks(UIWidget* prev, UIWidget* next);

    int                   getId() const { return id; }
    virtual UIWidget*     getWidgetById(int id);
    virtual UIWidget*     getFocusedWidget();

    const GfxRect&        getExtent() const { return extent; }

    /* Size Changes */
    virtual void          getLayoutInfo(LayoutInfo&) = 0;
    virtual void          pack();
    virtual void          doLayout();

    void                  setSize(int w, int h);
    void                  moveTo(int x, int y);

    virtual void          setExtent(GfxRect newpos, bool by_layout = false);

    /* Main loops */
    virtual int           run();
    virtual void          stop(int arg);

    string_t              toString();
};

struct UIWidget::UpdateInfo {
    SDL_Surface* sfc;

    UpdateInfo(SDL_Surface* asfc)
        : sfc(asfc) { }
    void update(const GfxRect& r);
    void doUpdates() { }
};

class UIFixedWidget : public UIWidget {
 public:
    UIFixedWidget(int aid) : UIWidget(aid) { }
    UIFixedWidget() : UIWidget() { }
    ~UIFixedWidget();

    void getLayoutInfo(LayoutInfo& info);
};

class UIWidgetContainer : public UIWidget {
 protected:
    UIWidget* first;
    UIWidget* last;
    UIWidget* selected;
    UIWidget* focused;

    virtual void add(UIWidget& widget) = 0;
 public:
    UIWidgetContainer(int aid = 0);
    ~UIWidgetContainer();

    /* Look & Feel */
    virtual Ptr<GfxCanvas> getCanvasFor(UIWidget* child) = 0;
    virtual void           drawBelow(UIWidget* child, Ptr<GfxCanvas> can, UpdateInfo& ui) = 0;
    virtual void           drawDeferred(Ptr<GfxCanvas> can, UpdateInfo& ui);

    virtual GfxRect        getClientRectFor(GfxRect ext);
    virtual GfxRect        getExtentFor(GfxRect cli);

    virtual void           removeWidget(UIWidget& w) = 0;

    UIWidget*              getWidgetById(int id);
    UIWidget*              getFocusedWidget();
    void                   selectWidget(UIWidget* w);
    void                   setFocus(UIWidget* w);
    void setFocus() { UIWidget::setFocus(); }
    virtual void handleChildResize(UIWidget* child, GfxRect oldsize) = 0;
    void stateChanged(int astat, bool enable);
    void setIdleNotify(bool enable);
    string_t toString();

    UIWidget* getFirstWidget() { return first; }

    void setStateOf(int id, int astat, bool enable);
};

class UIGroup : public UIWidgetContainer {
    Ptr<UILayoutManager> layout;
 public:
    UIGroup(int aid = 0);
    UIGroup(Ptr<UILayoutManager> alayout, int aid = 0);
    ~UIGroup();

    void setLayout(Ptr<UILayoutManager> alayout);
    Ptr<UILayoutManager> getLayout() const;

    virtual void drawInterior(Ptr<GfxCanvas> canvas, UpdateInfo& ui);
    void draw(Ptr<GfxCanvas> canvas, UpdateInfo& u);
    void drawBelow(UIWidget* child, Ptr<GfxCanvas> can, UpdateInfo& ui);

    bool handleEvent(const UIEvent& event, bool second_pass);
    Ptr<GfxCanvas> getCanvasFor(UIWidget* child);
    void handleChildResize(UIWidget* child, GfxRect oldsize);
    void add(const GfxRect& apos, UIWidget& widget);
    void add(UIWidget& widget);
    void removeWidget(UIWidget& w);

    void setExtent(GfxRect r, bool by_layout = false);

    void getLayoutInfo(LayoutInfo&);
    void pack();
    void doLayout();
};

/// Spacer.
class UISpacer : public UIWidget {
 public:
    UISpacer();
    void draw(Ptr<GfxCanvas> sfc, UpdateInfo& ui);
    void getLayoutInfo(LayoutInfo&);
};

#endif
