/**
  *  \file ui/uiglobal.cc
  *  \brief User Interface Global Variables
  *
  *  Global Variables and initialisation routines thereof.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdio>
#include "ui/uiglobal.h"
#include "util/console.h"
#include "gfx/pixmap.h"
#include "ui/event.h"
#include "fileio/streams.h"
#include "util/nls.h"
#include "gfx/palette.h"
#include "resmgr/resmgr.h"
#include "fileio/iostr.h"
#include "util/except.h"
#include "cpluslib/assert.h"
#include "cpluslib/format.h"

/** All the fonts. Fonts are loaded at startup. */
Ptr<GfxFont> fonts[MAX_FONTS];

/** Font heights. Cached here for simplicity. */
int font_heights[MAX_FONTS];

/** "em" widths for all fonts. Cached here for simplicity. */
int font_ems[MAX_FONTS];

/** Root widget. */
Ptr<UIRoot> ui_root;

/** Screen surface. This is the surface on which all user interface
    activities happen. */
Ptr<GfxCanvas> ui_screen;

/** Bit depth of screen surface. Cached here for simplicity. */
int ui_screen_bits;

/** Resource directory. */
Ptr<Directory> resource_directory;

static bool
loadFont(int i)
{
    try {
        Ptr<Stream> s = getNamedResourceEntry(format("f.font%d", i));
        if (!s)
            return false;

        fonts[i-1] = new GfxStdFont(s);
        font_heights[i-1] = fonts[i-1]->getTextHeight("Tp");
        font_ems[i-1] = fonts[i-1]->getTextWidth("m");
        return true;
    }
    catch(...) {
        return false;
    }
}

/* FIXME: I don't really like this one. It should probably go into
   resmgr/, along with proper unified caching for pixmaps. */
Ptr<GfxPixmap>
requestPixmap(string_t name)
{
    return ResManager::getInstance().getNamedPixmap("g." + name);
}

Ptr<GfxPixmap>
requestPixmap(string_t name, int index)
{
    if (Ptr<GfxPixmap> pix = requestPixmap(name + itoa(index)))
        return pix;
    else
        return requestPixmap(name);
}

Ptr<GfxPixmap>
requestPixmap(const char* name)
{
    return requestPixmap(string_t(name));
}

/** Initialize. This sets up all GUI-related variables.
    \param screen  Screen canvas to use.
    \param resdir  Resource directory. */
void
initUIGlobals(Ptr<GfxCanvas> screen, Ptr<Directory> resdir)
{
    SDL_Surface* sfc = screen->getSurface();
    char driver_name[100];
    ASSERT(sfc);

    console.write(LOG_INFO, _("=== Initializing GUI... ==="));

    if (SDL_VideoDriverName(driver_name, sizeof driver_name)) {
        const SDL_VideoInfo* vi = SDL_GetVideoInfo();
        string_t flags;
        if (vi->wm_available)   // do not translate the flag strings
            flags += ", wm";
        if (vi->hw_available)
            flags += ", hw";
        if (vi->blit_hw)
            flags += ", hw->hw";
        if (vi->blit_sw)
            flags += ", sw->hw";
        if (vi->blit_fill)
            flags += ", fill";
        console.write(LOG_INFO, format(_("Video driver: %s (%dk%s)"),
                                       driver_name, vi->video_mem, flags));
    }

    /* initialize UI */
    initPalette(sfc);

    resource_directory = resdir;
    ui_screen = screen;
    ui_screen_bits = screen->getSurface()->format->BitsPerPixel;
    ui_root = new UIRoot(screen);

    ResManager::getInstance().addProvider(new ResDirectory(resource_directory));

    /* initialize resources */
    try {
        StandardStream s(resdir->openFile("cc-res.cfg", Stream::C_READ | Stream::C_TEXT));
        loadResourceConfig(ResManager::getInstance(), s);
    }
    catch(...) { }

    /* load fonts */
    int i = 1;
    while (loadFont(i))
        ++i;
    console.write(LOG_INFO, format(_("%d font%!1{s%} loaded"), i-1));
    if (i <= MIN_FONTS)
        throw GraphicsIOException(format(_("Too few fonts available, need at least %d"), int(MIN_FONTS)));

    initEvents();
}

void
postTimerEvent()
{
    if (ui_root)
        ui_root->postTimerEvent();
}

/** Take a screenshot. Screenshot will be saved in current directory. */
void
doScreenshot()
{
    // FIXME: this is pretty C-ish.
    using namespace std;
    FILE* fp;
    int index = 0;
    char fname[100];
    while (1) {
        sprintf(fname, "file%04d.bmp", ++index);
        if ((fp = fopen(fname, "rb")) == 0)
            break;
        fclose(fp);
        if (index >= 9999) {
            console.write(LOG_ERROR,
                          _("Unable to find a free file name for screenshot"));
            return;
        }
    }
    SDL_SaveBMP(ui_screen->getSurface(), fname);
    console.write(LOG_INFO, format(_("Screenshot saved as `%s'."), fname));
}
