/**
  *  \file ui/util.cc
  *  \brief UI Utilities
  *
  *  Graphics primitives, rather specific to our UI.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <vector>
#include "ui/util.h"
#include "gfx/palette.h"

/** Tile area with pixmap.
    \param can    canvas to use
    \param r      area to tile with pixmap
    \param pix    pixmap to use, may be null
    \param color  when /pix/ is null, the image is filled with this color
    \param alter  X coordinate alteration. With alteration 0, the area is
                  tiled with a regular grid, like on a checkered paper.
                  With nonzero alteration, the second row is shifted
                  that many pixels to the left, the next one is shifted
                  to the right again, etc, to make the pattern look more
                  interesting. */
void
drawTiledArea(GfxCanvas* can, GfxRect r, Ptr<GfxPixmap> pix,
              uint32 color, int alter)
{
    /* fallback method. */
    if (!pix.ptr()) {
        can->drawBar(r, color, GFX_TRANSPARENT, SOLID_FILL, SDL_ALPHA_OPAQUE);
        return;
    }
    pix->tile(can, r, alter);
}

/** Draw "up" frame. Like drawRectangle(), but the rectangle is colored
    to give the effect of a 3D raised panel. */
void
drawFrameUp(GfxCanvas* can, GfxRect r)
{
    int x2 = r.x + r.w - 1;
    int y2 = r.y + r.h - 1;

    can->drawHLine(r.x, r.y,   x2, standard_colors[COLOR_WHITE], ~0U, SDL_ALPHA_OPAQUE);
    can->drawVLine(r.x, r.y+1, y2, standard_colors[COLOR_WHITE], ~0U, SDL_ALPHA_OPAQUE);
    can->drawHLine(r.x, y2,    x2, standard_colors[COLOR_BLACK], ~0U, SDL_ALPHA_OPAQUE);
    can->drawVLine(x2,  r.y, y2+1, standard_colors[COLOR_BLACK], ~0U, SDL_ALPHA_OPAQUE);
}

/** Draw "down" frame. Like drawRectangle(), but the rectangle is colored
    to give the effect of a 3D lowered panel. */
void
drawFrameDown(GfxCanvas* can, GfxRect r)
{
    int x2 = r.x + r.w - 1;
    int y2 = r.y + r.h - 1;

    can->drawHLine(r.x, r.y,   x2, standard_colors[COLOR_BLACK], ~0U, SDL_ALPHA_OPAQUE);
    can->drawVLine(r.x, r.y+1, y2, standard_colors[COLOR_BLACK], ~0U, SDL_ALPHA_OPAQUE);
    can->drawHLine(r.x, y2,    x2, standard_colors[COLOR_WHITE], ~0U, SDL_ALPHA_OPAQUE);
    can->drawVLine(x2,  r.y, y2+1, standard_colors[COLOR_WHITE], ~0U, SDL_ALPHA_OPAQUE);
}

void
drawColorFrame(GfxCanvas* can, GfxRect r, TFrameColor fc, bool update)
{
    static int colors[][2] = {
        { COLOR_DARK,         COLOR_DARK },
        { COLOR_GREENSCALE+8, COLOR_GREENSCALE+10 },
        { COLOR_FIRE+6,       COLOR_FIRE+8 },
        { COLOR_DARKYELLOW,   COLOR_BRIGHTYELLOW },
        { COLOR_GRAY,         COLOR_GRAY }
    };

    GfxParameters parm;
    r.x--; r.y--; r.w += 2; r.h += 2;
    parm.color = standard_colors[colors[fc][1]];
    drawRectangle(can, parm, r);
    r.x--; r.y--; r.w += 2; r.h += 2;
    parm.color = standard_colors[colors[fc][0]];
    drawRectangle(can, parm, r);
    if (update)
        can->update(r);
}

/** Split text into lines.
    \param font      Font to use
    \param width     Available width in pixels
    \param s         Text. Can contain '\n' for forced linefeeds.
    \param lines     (out) Text split into lines
    \post lines.size() != 0 */
static void
splitText(GfxFont* font, int width, const string_t s,
          std::vector<string_t>& lines)
{
    string_t line;
    int last_split_point = 0, last_width = 0;
    for(string_t::size_type i = 0; i < s.length(); ++i) {
        char c = s[i];
        if (c == '\n') {
            /* new line */
            lines.push_back (line);
            last_width = last_split_point = 0;
            line.erase();
        } else if (c == ' ') {
            /* add a word */
            line += c;
            last_width = font->getTextWidth(line);
            last_split_point = line.length();
            if (last_width >= width) {
                /* need to split here */
                lines.push_back(line.substr(0, line.length()-1));
                last_width = last_split_point = 0;
                line.erase();
            }
        } else {
            /* add a character */
            line += c;
            if (font->getTextWidth(line) > width) {
                /* when there is no split point (extra-ordinarily
                   long word), split in the middle of a word. When
                   the width is too small, exceed it to avoid endless
                   loops. */
                if (!last_split_point)
                    last_split_point = line.length() - 1;
                if (!last_split_point)
                    last_split_point = 1;
                lines.push_back(line.substr(0, last_split_point));
                line.erase(0, last_split_point);
                last_width = last_split_point = 0;
            }
        }
    }
    /* add terminating partial line. */
    if (!line.empty() || lines.empty())
        lines.push_back(line);
}

/** Count number of lines occupied by text.
    \param font     Font to use
    \param width    Available width
    \param s        Text. Can contain '\n' */
int
countLines(GfxFont* font, int width, string_t s)
{
    std::vector<string_t> lines;
    splitText(font, width, s, lines);
    return lines.size();
}

/** Output text, with word wrap.
    \param can     Canvas
    \param parm    Parameter set. Currently only uses whatever the
                   font supports (not alignment settings etc.)
    \param font    Font to use
    \param x,y     Upper-left coordinate
    \param s       Text. Can contain '\n' */
void
outTextFormatted(GfxCanvas* can, GfxParameters& parm, GfxFont* font,
                 int x, int y, int width, string_t s)
{
    std::vector<string_t> lines;
    splitText(font, width, s, lines);
    for (unsigned i = 0; i < lines.size(); ++i) {
        font->outText(can, parm, x, y, lines[i]);
        y += font->getTextHeight(lines[i]);
    }
}
