/**
  *  \file ui/widget-button.cc
  *  \brief Push-button-like Widgets
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/widgets.h"
#include "ui/event.h"
#include "gfx/gfxcompl.h"
#include "gfx/gfx.h"
#include "gfx/font.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "cpluslib/assert.h"

/**************************** UIAbstractButton ***************************/

/** \class UIAbstractButton
    \brief Standard Button

    This class provides the event handling mechanisms of a standard
    clickable button (checkbox, radiobutton, ...). The button can be
    have an associated hot-key if so desired.

    Buttons are triggered by a keystroke, mouse click and possibly a
    command. For example, a aStop button will call stop(id) whenever
    it is clicked, the associated key is pressed, or when it sees the
    command (id) (possibly from a UIKeyTranslator).

    This class provides four standard actions:
    - aCommand: post a command (id, prefix) when triggered.
    - aStop: end a modal dialog (stop(id)) when triggered.
    - aKey: post a key event when triggered. These buttons are not
      triggered by a keystroke.
    - aNothing: do nothing.

    aCommand buttons should have the bf_NoCommand style to avoid that
    they trigger themselves again. aKey buttons \em must have the bf_Key
    style.

    Derived classes must override draw().

    You can attach own handlers to sig_fired. */

/** Construct new command button. The button will post a command /aid/ when clicked.
    \param aid    widget id, used by derived classes for emitting
                  commands etc.
    \param aname  button text, will usually be drawn on button
    \param akey   key that triggers this button. */
UIAbstractButton::UIAbstractButton(int aid, string_t aname, UIKey akey)
    : UIWidget(aid),
      name(aname),
      key(akey)
{
    sig_fired.add(this, &UIButton::postCommand);
    setState(bf_NoCommand, true);
}

/** Construct new key button. The button will post a UIKey event.
    \param aname  button text, will usually be drawn on button
    \param akey   key that triggers this button. */
UIAbstractButton::UIAbstractButton(string_t name, UIKey akey)
    : UIWidget(),
      name(name),
      key(akey)
{
    setState(bf_Key, true);
}

/** Construct new button.
    \param aid    widget id, used by derived classes for emitting
                  commands etc.
    \param aname  button text, will usually be drawn on button
    \param akey   key that triggers this button.
    \param action action to do when clicked. */
UIAbstractButton::UIAbstractButton(int aid, string_t aname, UIKey akey, Action action)
    : UIWidget(aid),
      name(aname),
      key(akey)
{
    if (action == aCommand) {
        sig_fired.add(this, &UIButton::postCommand);
        setState(bf_NoCommand, true);
    } else if (action == aStop) {
        sig_fired.add(this, &UIButton::callStop);
    } else if (action == aKey) {
        setState(bf_Key, true);
    }
}

UIAbstractButton::~UIAbstractButton()
{ }

/** Change button state. When the button loses logical focus, it must
    be un-pressed. Certain events also need redraw. */
void
UIAbstractButton::stateChanged(int astate, bool enable)
{
    UIWidget::stateChanged(astate, enable);

    if ((status & (bf_Pressed | st_Selected)) == bf_Pressed)
        status &= ~bf_Pressed;

    if (astate & (bf_Pressed | st_Selected | bf_Highlight | st_Disabled))
        drawWidget(false);
}

/** Event handler. Processes mouse and keyboard events, and calls
    fire() when needed. */
bool
UIAbstractButton::handleEvent(const UIEvent& e, bool second_pass)
{
    switch(e.type) {
     case ev_Key:
        if (!hasState(bf_Key) && e.key.code == key) {
            select();
            fire();
            return true;
        }
        break;
     case ev_Mouse:
        if (extent.contains(e.mouse.x, e.mouse.y)) {
            /* mouse is in widget. */
            select();
            if (e.mouse.b) {
                /* mouse pressed in button */
                setState(bf_Pressed, true);
            } else if (hasState(bf_Pressed)) {
                /* mouse released in button */
                setState(bf_Pressed, false);
                if (hasState(bf_Key)) {
                    UIEvent ne;
                    ne.type         = ev_Key;
                    ne.key.sdl_type = SDL_KEYDOWN;
                    ne.key.arg      = 0;// FIXME
                    ne.key.code = key | UIEvent::simplMod(uint32(SDL_GetModState()) << 16);
                    ne.key.unicode  = 0;// FIXME?
                    putEvent(ne);
                }
                fire();
            }
            return true;
        } else {
            /* automatically lose selection when mouse is out. */
            if (hasState(st_Selected))
                owner->selectWidget(0);

            /* let others handle the event, too. */
            return false;
        }
        break;
     case ev_Command:
        if (e.cmd.code == getId() && !hasState(bf_NoCommand)) {
            if (hasState(bf_Key))  // FIXME: unify with ev_Mouse branch?
                putKey(key);
            fire();
            return true;
        } else if (e.cmd.code == cm_INT_QueryReferencePoint) {
            ASSERT(e.cmd.infop);
            TReferencePointArg& ra = *static_cast<TReferencePointArg*>(e.cmd.infop);
            if ((ra.event.type == ev_Key && ra.event.key.code == key)
                || (ra.event.type == ev_Command && ra.event.cmd.code == getId()))
            {
                ra.x      = getExtent().x;
                ra.y      = getExtent().y;
                ra.widget = this;
                return true;
            }
        }
        break;
     default:;
    }
    return UIWidget::handleEvent(e, second_pass);
}

void
UIAbstractButton::postCommand(int prefix)
{
    ::putCommand(id, prefix);
}

void
UIAbstractButton::callStop(int)
{
    stop(id);
}

/** Reaction on button click. */
void
UIAbstractButton::fire()
{
    /* FIXME: arg should be prefix arg */
    sig_fired.raise(0);
}

/******************************** UIButton *******************************/

/** \class UIButton
    \brief Push button

    This class provides the look of a standard push-button. */

UIButton::~UIButton()
{ }

void
UIButton::drawButton(Ptr<GfxCanvas> can, GfxRect extent,
                     const string_t& name, int flags)
{
    // FIXME: use actual PCC I look. Maybe skinnable?
    const int fnr = (flags & bf_SmallSize) ? FONT_SMALL : FONT_TITLE;
    int delta = 0;
    if (flags & bf_Pressed) {
        drawFrameDown(can, extent);
        ++delta;
    } else {
        drawFrameUp(can, extent);
    }

    int color = (flags & bf_Highlight
                 ? COLOR_GRAYSCALE + 12
                 : COLOR_GRAYSCALE + 9);
    if (flags & st_Selected)
        color++;

    can->drawBar(GfxRect(extent.x + 1, extent.y + 1,
                         extent.w - 2, extent.h - 2),
                 standard_colors[color],
                 GFX_TRANSPARENT,
                 SOLID_FILL,
                 SDL_ALPHA_OPAQUE);

    GfxParameters parm;
    parm.color = standard_colors[COLOR_SHIELD + 5];

    int x = extent.x + extent.w/2 - fonts[fnr]->getTextWidth(name)/2 + delta;
    int y = extent.y + extent.h/2 - fonts[fnr]->getTextHeight(name)/2 + delta;
    fonts[fnr]->outText(can.ptr(), parm, x, y, name);
}    

/** Draw the button. */
void
UIButton::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    drawButton(can, extent, getName(), getState());
    ui.update(extent);
}

void
UIButton::getLayoutInfo(LayoutInfo& info)
{
    if (hasState(bf_SmallSize)) {
        info.flags = 0;
        info.min_size = info.pref_size
            = GfxPoint(getName().length() == 1 ? 15 : 8 + fonts[FONT_SMALL]->getTextWidth(getName()), 15);
    } else {
        info.flags = 0;
        info.min_size = info.pref_size
            = GfxPoint(getName().length() == 1 ? 25 : 15 + fonts[FONT_TITLE]->getTextWidth(getName()), 25);
    }
}
