/**
  *  \file ui/widget-cb.cc
  *  \brief Checkbox-like Widgets
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/widgets.h"
#include "resmgr/cc1resid.h"
#include "resmgr/resmgr.h"
#include "gfx/palette.h"
#include "cpluslib/format.h"

/*************************** UIAbstractCheckbox **************************/

/** \class UIAbstractCheckbox
    \brief Abstract Checkbox

    This module provides drawing and event handling for an abstract
    checkbox. Descendants have to override getPixmap() and click() to
    provide the actual policy. */

UIAbstractCheckbox::UIAbstractCheckbox(int id, string_t name, UIKey key)
    : UIAbstractButton(id, name, key, aNothing)
{
    sig_fired.add(this, &UIAbstractCheckbox::click);
}

UIAbstractCheckbox::~UIAbstractCheckbox()
{ }

void
UIAbstractCheckbox::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    if (Ptr<GfxPixmap> pix = getPixmap())
        pix->blit(can, extent.x + 12 - pix->getWidth()/2, extent.y + 12 - pix->getHeight()/2);

    GfxParameters parm;

    if (getName().length()) {
        parm.color = standard_colors[COLOR_BLACK];
        outTextF(can, parm, fonts[FONT_TITLE], extent.x + 24, extent.y + 2, extent.w - 24, getName());
    }
    
    if (hasState(st_Focused))
        parm.color = standard_colors[COLOR_BLACK];
    else
        parm.color = standard_colors[COLOR_GRAY];
    drawRectangle(can, parm, GfxRect(extent.x, extent.y, 24, 24));

    ui.update(getExtent());
}

void
UIAbstractCheckbox::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_FlexibleH;
    info.min_size = GfxPoint(24, 24);
    if (getName().length())
        info.min_size.x += fonts[FONT_TITLE]->getTextWidth(getName()) + 2;
    info.pref_size = info.min_size;
}

void
UIAbstractCheckbox::stateChanged(int state, bool enabled)
{
    UIAbstractButton::stateChanged(state, enabled);
    if (state & st_Focused)
        drawWidget(false);
}

bool
UIAbstractCheckbox::handleEvent(const UIEvent& e, bool second_pass)
{
    if (hasState(st_Focused) && e.type == ev_Key && e.key.code == ' ') {
        select();
        click(0);
        return true;
    }
    return UIAbstractButton::handleEvent(e, second_pass);
}

/******************************* UICheckbox ******************************/

/** \class UICheckbox
    \brief Classic "yes/no" checkbox

    This provides a yes/no checkbox which users can toggle by clicking.
    Descendants can override getPixmap() to provide different images. */

/** Create checkbox.
    \param aid    Widget id
    \param aname  Text
    \param akey   Hotkey
    \param max    Number of states (2 if "yes/no"). Clicking brings
                  the checkbox through the states max-1, max-2, ... 0 */
UICheckbox::UICheckbox(int aid, string_t aname, UIKey akey, int max)
    : UIAbstractCheckbox(aid, aname, akey), value(0), max(max)
{ }

UICheckbox::~UICheckbox()
{ }

void
UICheckbox::click(int)
{
    int i = getValue();
    if (i == 0)
        i = max;
    setValue(i-1);
    setFocus();
}

void
UICheckbox::setValue(int avalue)
{
    if (value != avalue) {
        value = avalue;
        drawWidget(false);
        sig_changed.raise(getId());
    }
}

Ptr<GfxPixmap>
UICheckbox::getPixmap()
{
    return requestPixmap(format("cb%d", getValue()+1));
}

/***************************** UIRadioButton *****************************/

/** \class UIRadioButton
    \brief Radio button

    From all radio buttons in a group, only one may be checked at a time.
    The actual value of a radio button group is stored in an object of
    type UIRadioButton::Value. */

/** Create radio button.
    \param id      widget id
    \param name    text
    \param key     hotkey
    \param valuep  pointer to value object for this radio group
    \param myvalue the value of this radio button */
UIRadioButton::UIRadioButton(int id, string_t name, UIKey key, Ptr<Value> valuep, int myvalue)
    : UIAbstractCheckbox(id, name, key),
      value(valuep), myvalue(myvalue)
{
    cookie = getSignal().add(this, &UIRadioButton::changed);
}

UIRadioButton::~UIRadioButton()
{
    getSignal().remove(cookie);
}

void
UIRadioButton::click(int)
{
    setValue(myvalue);
    setFocus();
}

void
UIRadioButton::changed(int)
{
    drawWidget(false);
}

Ptr<GfxPixmap>
UIRadioButton::getPixmap()
{
    if (getValue() == myvalue)
        return requestPixmap("radio2");
    else
        return requestPixmap("radio1");
}
