/**
  *  \file ui/widget-il.cc
  *  \brief Input-line Widget
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "gfx/font.h"
#include "gfx/gfx.h"
#include "gfx/gfxcompl.h"
#include "ui/layout.h"
#include "ui/widgets.h"
#include "ui/window.h"
#include "ui/util.h"
#include "gfx/palette.h"

/****************************** UIInputLine ******************************/

/** \class UIInputLine
    \brief Input Line

    Standard text entry field */

/** Input line.
    \param amaxl   maximum length of input (a hard limit)
    \param ahot    hot-key that brings focus to this widget
    \param aflags  options
    \param aid     widget Id */
UIInputLine::UIInputLine(string_t::size_type amaxl, UIKey ahot,
                         int aflags, int aid)
    : UIWidget(aid), hotkey(ahot), maxlen(amaxl), curx(0), firstx(0),
      text(), font(FONT_TITLE), pref_len(amaxl > 40 ? amaxl : 40)
{
    setState(aflags | ilf_TypeErase, true);
}

/** Input line.
    \param amaxl   maximum length of input (a hard limit)
    \param aprefl  preferred width of input, in "em" widths
    \param ahot    hot-key that brings focus to this widget
    \param aflags  options
    \param aid     widget Id */
UIInputLine::UIInputLine(string_t::size_type amaxl, int apref_len,
                         UIKey ahot,
                         int aflags, int aid)
    : UIWidget(aid), hotkey(ahot), maxlen(amaxl), curx(0), firstx(0),
      text(), font(FONT_TITLE), pref_len(apref_len)
{
    setState(aflags | ilf_TypeErase, true);
}

/** Set contents of input line. */
void
UIInputLine::setText(string_t s)
{
    text = s;
    curx = text.length();
    firstx = 0;
    scroll();
    drawWidget(false);
    sig_changed.raise(getId());
}

void
UIInputLine::stateChanged(int astate, bool enable)
{
    UIWidget::stateChanged(astate, enable);
    if ((astate & st_Focused) && enable)
        status |= ilf_TypeErase;
    if (astate & (st_Focused | st_Disabled | ilf_TypeErase))
        drawWidget(false);
}

/** Get width of space available for text. */
int
UIInputLine::getTextWidth()
{
    int rv = extent.w;
    if (hasState(ilf_Framed))
        rv -= 4;
    return rv;
}

/** Get height of space needed for text. */
int
UIInputLine::getTextHeight()
{
    int rv = fonts[font]->getTextHeight("Tp");
    if (hasState(ilf_Framed))
        return rv + 4;
    else
        return rv;
}

/** Adjust display so that cursor is visible. */
void
UIInputLine::scroll()
{
    if (curx > text.length())
        curx = text.length();
    if (curx < 3)
        firstx = 0;
    else if (firstx > curx - 3)
        firstx = curx - 3;

    string_t t = hasState(ilf_Hidden) ? string_t(text.length(), '*') : text;
    int ln = fonts[font]->getTextWidth(t.substr(firstx, curx - firstx));
    int av = getTextWidth() - CursorWidth;
    if (curx) {
        while (firstx < curx-1 && ln > av) {
            ++firstx;
            ln = fonts[font]->getTextWidth(t.substr(firstx, curx - firstx));
        }
    }
}

/** Insert text at current cursor position. Respects flags
    ilf_NonEditable (=request is ignored) and ilf_TypeErase (=replace
    current text). */
void
UIInputLine::insert(string_t s)
{
    if (hasState(ilf_NonEditable))
        return;
    if (hasState(ilf_TypeErase)) {
        text = s;
        curx = s.length();
        setState(ilf_TypeErase, false);
    } else {
        if (text.length() >= maxlen)
            return;
        string_t::size_type remain = maxlen - text.length();
        if (remain < s.length())
            s.erase(remain);
        text.insert(curx, s);
        curx += s.length();
    }
    scroll();
    drawWidget(false);
    sig_changed.raise(getId());
}

/** Move cursor to position pos.
    \param pos cursor position, [0, text.length()] */
void
UIInputLine::setCursor(pos_type pos)
{
    if (pos > text.length())
        pos = text.length();
    if (pos != curx) {
        setState(ilf_TypeErase, false);
        curx = pos;
        scroll();
        drawWidget(false);
    }
}

bool
UIInputLine::handleEvent(const UIEvent& e, bool second_pass)
{
    switch(e.type) {
     case ev_Mouse:
        if (e.mouse.b && extent.contains(e.mouse.x, e.mouse.y)) {
            select();
            if (!hasState(st_Focused)) {
                owner->setFocus(this);
                putCommand(id, 0);
            }

            pos_type new_pos = firstx;
            int      x0      = hasState(ilf_Framed) ? extent.x + 2 : extent.x;
            if (e.mouse.x < x0) {
                if (new_pos)
                    --new_pos;
            } else {
                int delta = e.mouse.x - x0;
                if (hasState(ilf_Hidden)) {
                    while (new_pos < text.length() &&
                           fonts[font]->getTextWidth(string_t(new_pos - firstx, '*')) < delta)
                        ++new_pos;
                } else {
                    while (new_pos < text.length() &&
                           fonts[font]->getTextWidth(text.substr(firstx, new_pos - firstx)) < delta)
                        ++new_pos;
                }
            }
            // FIXME: consumeprefix()
            curx = new_pos;
            drawWidget(false);
            return true;
        }
        break;
     case ev_Key:
        if (!hasState(st_Focused) && !hasState(st_Disabled)) {
            if (e.key.code == hotkey) {
                select();
                setFocus();
                return true;
            }
        } else {
            switch(e.key.code) {
             case SDLK_LEFT:
                select();
                if (curx > 0)
                    setCursor(curx - 1);
                return true;
             case SDLK_RIGHT:
                select();
                if (curx < text.length())
                    setCursor(curx + 1);
                return true;
             case SDLK_BACKSPACE:
                select();
                if (curx > 0) {
                    --curx;
                    text.erase(curx, 1);
                    setState(ilf_TypeErase, false);
                    drawWidget(false);
                    sig_changed.raise(getId());
                }
                return true;
             case SDLK_DELETE:
             case ss_Ctrl + 'd':
                select();
                if (curx < text.length()) {
                    text.erase(curx, 1);
                    setState(ilf_TypeErase, false);
                    drawWidget(false);
                    sig_changed.raise(getId());
                }
                return true;
             default:
                if ((e.key.code >= '0' && e.key.code <= '9')
                    || (e.key.code >= ' ' && e.key.code <= 126 && !hasState(ilf_NumbersOnly))
                    || (e.key.code >= 128 && e.key.code <= 255 && !hasState(ilf_NumbersOnly | ilf_NoHi)))
                {
                    select();
                    insert(string_t(1, char(e.key.code)));
                    return true;
                }
            }
        }
        break;
     case ev_Command:
//         if (e.cmd.code == id) {
//             select();
//             owner->setFocus(this);
//             return true;
//         }
        break;
     default:;
    }
    return UIWidget::handleEvent(e, second_pass);
}

void
UIInputLine::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    GfxParameters parm;
    GfxRect r = getExtent();
    if (hasState(ilf_Framed)) {
        drawFrameDown(can, r);
        r.x += 2;
        r.y += 2;
        r.w -= 4;
        r.h -= 4;
    }
    if (hasState(st_Focused)) {
        scroll();
        if (hasState(ilf_TypeErase)) {
            parm.color   = standard_colors[COLOR_GRAY];
            parm.bgcolor = standard_colors[COLOR_BLUE];
        } else {
            parm.color   = standard_colors[COLOR_BLUE];
            parm.bgcolor = standard_colors[COLOR_GRAY];
        }
    } else {
        parm.color   = standard_colors[COLOR_BLACK];
        parm.bgcolor = standard_colors[COLOR_GRAY];
        firstx = 0;
    }

    string_t t = (hasState(ilf_Hidden)
                  ? string_t(text.length() - firstx, '*')
                  : string_t(text.substr(firstx)));
    outTextF(can, parm, fonts[font], r.x, r.y, r.w, t);
    /* outTextF sets parm.x to the end of the output */
    drawSolidBar(can, GfxRect(parm.x, r.y, r.x + r.w - parm.x, r.h), standard_colors[COLOR_GRAY]);
    if (hasState(st_Focused) && curx >= firstx) {
        unsigned cx = fonts[font]->getTextWidth(t.substr(0, curx-firstx));
        if (cx + 10 < r.w)
            /* FIXME: magic numbers */
            drawSolidBar(can, GfxRect(r.x + cx, r.y + 19, 10, 2), standard_colors[COLOR_BLACK]);
    }
    ui.update(getExtent());
}

void
UIInputLine::getLayoutInfo(LayoutInfo& info)
{
    info.flags     = li_FlexibleH;
    info.min_size  = GfxPoint(4 * font_ems[font], getTextHeight());
    info.pref_size = GfxPoint(pref_len * font_ems[font], getTextHeight());
}

namespace {
    class KeyHandler : public UIWidget {
     public:
        KeyHandler() { }
        bool handleEvent(const UIEvent& e, bool second_pass) {
            if (e.type == ev_Key) {
                if (e.key.code == SDLK_RETURN) {
                    stop(1);
                    return true;
                } else if (e.key.code == SDLK_ESCAPE) {
                    stop(0);
                    return true;
                }
            }
            return UIWidget::handleEvent(e, second_pass);
        }
        void draw(Ptr<GfxCanvas>, UpdateInfo&) { }
        void getLayoutInfo(LayoutInfo& info) {
            info.flags = li_NoLayout;
        }
    };
}

/** Do a standard input line dialog.
    \param str     string to enter/modify
    \param maxlen  maximum length for string
    \param title,req  title and prompt of dialog window
    \param width   width of input field in "em" widths
    \param flags   flags for input field */
bool
UIInputLine::run(string_t& str, int maxlen,
                 const string_t& title, const string_t& req,
                 int width, int flags)
{
    if (width * font_ems[FONT_TITLE] > ui_root->getExtent().w - 40)
        width = (ui_root->getExtent().w - 40) / font_ems[FONT_TITLE];

    UIWindow win(title, UIWindow::BLUE_WINDOW, COLOR_GRAY, 10);
    win.setLayout(new UIVBoxLayout(10));
    win.setState(UIWidget::st_Modal, true);

    UIStaticText w_text(req, UIStaticText::BlackTitle);
    win.add(w_text);

    UIInputLine il(maxlen, width, 0, flags, 0);
    win.add(il);

    KeyHandler kh;
    win.add(kh);
    win.pack();
    win.setFocus(&il);
    il.setText(str);
    ui_root->centerWidget(win);
    ui_root->add(win);
    int rc = ui_root->run();
    ui_root->close(win);
    if (rc)
        str = il.getText();
    return rc;
}
