/**
  *  \file ui/widgets.cc
  *  \brief Standard Widgets
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/widgets.h"
#include "ui/util.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "ui/layout.h"

/**************************** UIKeyTranslator ****************************/

/**
  *  \class UIKeyTranslator
  *  \brief Key to Command Translation
  *
  *  This class provides a means of translating keystrokes into commands.
  *  This widget is invisible.
  *
  *  All action happens in event processing phase 2, so we don't steal
  *  alphanumeric input from widgets that need it more than we.
  */
UIKeyTranslator::UIKeyTranslator()
    : UIWidget()
{ }

/** Initialize from static table.
    \param akeys points to \c n Init structures;
    \param n     number of initial mappings. */
UIKeyTranslator::UIKeyTranslator(const Init* akeys, unsigned n)
    : UIWidget()
{
    while (n--) {
        translation[akeys->key] = akeys->command;
        ++akeys;
    }
}

bool
UIKeyTranslator::handleEvent(const UIEvent& e, bool second_pass)
{
    if (second_pass && e.type == ev_Key) {
        std::map<UIKey, int>::iterator i = translation.find(e.key.code);
        if (i != translation.end()) {
            select();
            putCommand(i->second, 0);
            return true;
        }
    }
    return UIWidget::handleEvent(e, second_pass);
}

void
UIKeyTranslator::draw(Ptr<GfxCanvas>, UpdateInfo&)
{ }

/** Add a key translation. */
void
UIKeyTranslator::add(UIKey key, int command)
{
    translation[key] = command;
}

void
UIKeyTranslator::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_NoLayout;
}

/***************************** UIImageWidget *****************************/

/** \class UIImageWidget
    \brief Static widget displaying a pixmap */
UIImageWidget::UIImageWidget(Ptr<GfxPixmap> apix)
    : UIWidget(0), pix(apix)
{ }

UIImageWidget::UIImageWidget(const string_t& name)
    : UIWidget(0), pix(requestPixmap(name))
{ }

void
UIImageWidget::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    can->drawBar(extent, standard_colors[0], GFX_TRANSPARENT,
                 SOLID_FILL, SDL_ALPHA_OPAQUE);
    drawFrameDown(can, extent);
    if (pix) {
        GfxRect clip(0, 0,
                     std::min(extent.w - 2, pix->getWidth()),
                     std::min(extent.h - 2, pix->getHeight()));
        can->blitSurface(extent.x+1, extent.y+1, pix->getSurface(), clip);
    }
    ui.update(extent);
}

void
UIImageWidget::getLayoutInfo(LayoutInfo& info)
{
    // FIXME: should be made smarter
    info.flags = li_NoLayout;
}

/****************************** UIStaticText *****************************/

/** \class UIStaticText
    \brief Static line of text

    This widget can display a particular piece of text in one of several
    styles. */

static int text_types[] = {
    FONT_TITLE,
    FONT_NORMAL,
    FONT_TITLE,
    FONT_TITLE,
};
static int text_colors[] = {
    COLOR_BLACK,
    COLOR_BLACK,
    COLOR_GREENSCALE + 7,
    COLOR_WHITE,
};

UIStaticText::UIStaticText(const string_t& text, Scheme scheme)
    : UIWidget(0), text(text), scheme(scheme)
{
    setState(st_Disabled, true);
}

UIStaticText::UIStaticText(const char* text, Scheme scheme)
    : UIWidget(0), text(text), scheme(scheme)
{
    setState(st_Disabled, true);
}

void
UIStaticText::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.min_size = info.pref_size =
        GfxPoint(fonts[text_types[scheme]]->getTextWidth(text),
                 font_heights[text_types[scheme]]);
}

void
UIStaticText::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    GfxParameters parm;
    parm.color = standard_colors[text_colors[scheme]];
    outTextF(can, parm, fonts[text_types[scheme]],
             extent.x, extent.y, extent.w, text);
    ui.update(extent);
}

/**************************** UIFocusIterator ****************************/

/** \class UIFocusIterator
    \brief Cursor Focus Control

    This invisible widget provides a means of manipulating focus using
    the keyboard. Given a range of widget Ids, it provides keys to
    move within this range.
    - fi_Horiz: left and right arrows
    - fi_Vert: up and down arrows
    - fi_Tab: Tab and Shift-Tab keys */

UIFocusIterator::UIFocusIterator(int from, int to, int flags)
    : UIWidget(), from(from), to(to), flags(flags)
{
    setState(st_Disabled, true);  // so it doesn't try to get the focus
}

UIFocusIterator::~UIFocusIterator()
{ }

void
UIFocusIterator::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_NoLayout;
}

void
UIFocusIterator::draw(Ptr<GfxCanvas> /*can*/, UpdateInfo& /*ui*/)
{ }

bool
UIFocusIterator::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        int type = 0;
        switch(e.key.code) {
         case SDLK_TAB:            type = fi_Tab; break;
         case SDLK_TAB + ss_Shift: type = fi_Tab + fi_INTERNAL_NEGATE; break;
         case SDLK_UP:             type = fi_Vert + fi_INTERNAL_NEGATE; break;
         case SDLK_DOWN:           type = fi_Vert; break;
         case SDLK_LEFT:           type = fi_Horiz + fi_INTERNAL_NEGATE; break;
         case SDLK_RIGHT:          type = fi_Horiz; break;
        }
        if (type & flags) {
            UIWidget* focused_w = owner->getFocusedWidget();
            /* Iterate from the focused widget up until we find one which
               is in our range. */
            while (focused_w) {
                int focus_id = focused_w->getId();
                if (focus_id >= from && focus_id <= to) {
                    int new_id = focus_id;
                    do {
                        if (type & fi_INTERNAL_NEGATE)
                            --new_id;
                        else
                            ++new_id;
                        if (new_id < from) {
                            if (type & fi_Tab)
                                new_id = to;
                            else
                                return true;
                        } else if (new_id > to) {
                            if (type & fi_Tab)
                                new_id = from;
                            else
                                return true;
                        }
                    } while (new_id != focus_id && owner->getWidgetById(new_id)->hasState(st_Disabled));
                    if (new_id != focus_id) {
                        select();
                        owner->getWidgetById(new_id)->setFocus();
                    }
                    return true;
                } /* not ours */
                if (owner == focused_w)
                    break;
                focused_w = focused_w->getOwner();
            }
        } /* not a focus iterator key */
    } /* not a key */
    return UIWidget::handleEvent(e, second_pass);
}

/****************************** UIFrameGroup *****************************/

/** \class UIFrameGroup
    \brief Framed Group

    This widget is like UIGroup, but it will draw a "down" frame
    around its child widgets. */

UIFrameGroup::UIFrameGroup(Ptr<UILayoutManager> lay, int frame)
    : UIGroup(lay), frame(frame)
{ }

void
UIFrameGroup::drawInterior(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    drawFrameDown(can, extent);
    ui.update(extent);
}

GfxRect
UIFrameGroup::getExtentFor(GfxRect client)
{
    client.grow(frame, frame);
    return client;
}

GfxRect
UIFrameGroup::getClientRectFor(GfxRect ex)
{
    ex.grow(-frame, -frame);
    return ex;
}

/****************************** UIFocusGroup *****************************/

/** \class UIFocusGroup
    \brief Group that can have Focus, Visibly

    This is a group which can indicate visibly that it has focus.
    It also ensures that keyboard focus is distributed correctly.
    Otherwise it is exactly like UIGroup. */

UIFocusGroup::UIFocusGroup(int id)
    : UIGroup(id)
{ }

UIFocusGroup::UIFocusGroup(Ptr<UILayoutManager> lay, int id)
    : UIGroup(lay, id)
{ }

UIFocusGroup::~UIFocusGroup()
{ }

void
UIFocusGroup::drawInterior(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    GfxParameters parm;
    parm.color = hasState(st_Focused) ? standard_colors[COLOR_BLACK] : standard_colors[COLOR_GRAY];
    drawRectangle(can, parm, extent);
    ui.update(extent);
}

void
UIFocusGroup::stateChanged(int state, bool enable)
{
    if (state & st_Focused)
        drawWidget(false);
    UIGroup::stateChanged(state, enable);
}

bool
UIFocusGroup::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key && !hasState(st_Focused))
        return false;
    if (e.type == ev_Mouse && e.mouse.b
        && !hasState(st_Focused) && extent.contains(e.mouse.x, e.mouse.y))
    {
        select();
        setFocus();
        /* FALLTHROUGH */
    }
    return UIGroup::handleEvent(e, second_pass);
}

GfxRect
UIFocusGroup::getExtentFor(GfxRect client)
{
    client.grow(1, 1);
    return client;
}

GfxRect
UIFocusGroup::getClientRectFor(GfxRect ex)
{
    ex.grow(-1, -1);
    return ex;
}
