/**
  *  \file ui/widgets.h
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UI_WIDGETS_H
#define PCC_V2_UI_WIDGETS_H

#include <string>
#include <map>
#include "ui/ui.h"
#include "ui/uiglobal.h"
#include "cpluslib/sigslot.h"

/* Button (checkbox, radiobutton, ...) */
class UIAbstractButton : public UIWidget {
    void postCommand(int);
    void callStop(int);

    string_t name;
    UIKey    key;
 public:
    enum {
        bf_Pressed   = st_Last*2,     //!< Button is currently pressed.
        bf_Highlight = st_Last*4,     //!< Button is highlighted ("on").
        bf_Key       = st_Last*8,     //!< This is a "key" button.
        bf_NoCommand = st_Last*16,    //!< Do not react on command events. Normal buttons fire their event when getting a ev_Command with their Id.
        bf_SmallSize = st_Last*32     //!< Use "small" font.
    };
    enum Action {
        aNothing,
        aCommand,
        aKey,
        aStop
    };
    UIAbstractButton(int aid, string_t aname, UIKey akey);
    UIAbstractButton(string_t aname, UIKey akey);
    UIAbstractButton(int aid, string_t aname, UIKey akey, Action a);
    ~UIAbstractButton();

    bool handleEvent(const UIEvent& e, bool second_pass);
    void stateChanged(int astate, bool enable);
    void fire();

    const string_t& getName() const { return name; }

    Signal1<int> sig_fired;
};

/* Push-button */
class UIButton : public UIAbstractButton {
 public:
    UIButton(int aid, string_t aname, UIKey akey)
        : UIAbstractButton(aid, aname, akey) { }
    UIButton(string_t aname, UIKey akey)
        : UIAbstractButton(aname, akey) { }
    UIButton(int aid, string_t aname, UIKey akey, Action a)
        : UIAbstractButton(aid, aname, akey, a) { }
    ~UIButton();

    void draw(Ptr<GfxCanvas> can, UpdateInfo& ui);

    void getLayoutInfo(LayoutInfo&);

    static void drawButton(Ptr<GfxCanvas> can, GfxRect extent,
                           const string_t& name, int flags);
};

class UIAbstractCheckbox : public UIAbstractButton {
 protected:
    UIAbstractCheckbox(int aid, string_t aname, UIKey akey);
    ~UIAbstractCheckbox();

    virtual Ptr<GfxPixmap> getPixmap() = 0;
    virtual void click(int) = 0;
 public:
    void draw(Ptr<GfxCanvas> can, UpdateInfo& ui);
    void getLayoutInfo(LayoutInfo&);
    void stateChanged(int state, bool enabled);
    bool handleEvent(const UIEvent& e, bool second_pass);
};

class UICheckbox : public UIAbstractCheckbox {
 protected:
    void click(int);

    int value, max;
 public:
    UICheckbox(int aid, string_t aname, UIKey akey, int max);
    ~UICheckbox();

    void setValue(int value);
    int  getValue() const { return value; }
    Ptr<GfxPixmap> getPixmap();

    Signal1<int> sig_changed;
};

class UIRadioButton : public UIAbstractCheckbox {
 public:
    typedef Observable<int> Value;
 private:
    Ptr<Value> value;
    int myvalue;
    void* cookie;

    void click(int);
    void changed(int);
 public:
    UIRadioButton(int id, string_t name, UIKey key, Ptr<Value> valuep, int myvalue);
    ~UIRadioButton();

    void          setValue(int newvalue) { value->set(newvalue); }
    int           getValue() const { return value->get(); }
    Signal1<int>& getSignal() const { return value->sig_changed; }

    Ptr<GfxPixmap> getPixmap();
};

/* Key Translator */
class UIKeyTranslator : public UIWidget {
    std::map<UIKey, int> translation;
 public:
    struct Init;

    UIKeyTranslator();
    UIKeyTranslator(const Init* akeys, unsigned n);

    bool handleEvent(const UIEvent& event, bool second_pass);
    void draw(Ptr<GfxCanvas> can, UpdateInfo& ui);

    void add(UIKey key, int command);
    void getLayoutInfo(LayoutInfo&);
};

/** Key Translator Initialisation. This POD type allows initialisation
    of a UIKeyTranslator from a static table, without the need for
    run-time initialisation for the data structure. */
struct UIKeyTranslator::Init {
    UIKey key;
    int command;
};

class UIImageWidget : public UIWidget {
    Ptr<GfxPixmap> pix;
 public:
    UIImageWidget(Ptr<GfxPixmap> apix);
    UIImageWidget(const string_t& name);

    void draw(Ptr<GfxCanvas> can, UpdateInfo& ui);
    void getLayoutInfo(LayoutInfo& info);
};

class UIStaticText : public UIWidget {
 public:
    enum Scheme {
        BlackTitle,
        BlackNormal,
        MedGreenTitle,
        WhiteTitle
    };
 private:
    string_t text;
    Scheme scheme;
 public:
    UIStaticText(const string_t& text, Scheme scheme);
    UIStaticText(const char* text, Scheme scheme);
    void getLayoutInfo(LayoutInfo&);
    void draw(Ptr<GfxCanvas> can, UpdateInfo& ui);
};

class UIInputLine : public UIWidget {
 public:
    typedef string_t::size_type pos_type;
 private:
    UIKey     hotkey;
    pos_type  maxlen;
    pos_type  curx, firstx;
    string_t  text;
    int       font;
    int       pref_len;         // preferred length in em
    void scroll();
    int getTextWidth();
    int getTextHeight();
    enum {
        CursorWidth = 2
    };
 public:
    enum {
        ilf_NumbersOnly = 2*st_Last,    // accept only digits
        ilf_Framed      = 4*st_Last,    // show frame
        ilf_Hidden      = 8*st_Last,    // "star out" text
        ilf_NoHi        = 16*st_Last,   // don't accept high-ascii chars
        ilf_NonEditable = 32*st_Last,   // non-editable
        ilf_Notify      = 64*st_Last,   // generate cm_TypeNotify events
        ilf_TypeErase   = 128*st_Last
    };
    UIInputLine(string_t::size_type amaxl, UIKey ahot, int aflags, int aid);
    UIInputLine(string_t::size_type amaxl, int apref_len, UIKey ahot, int aflags, int aid);
    void setText(string_t s);
    string_t getText() const { return text; }

    bool handleEvent(const UIEvent& e, bool second_pass);
    void draw(Ptr<GfxCanvas> can, UpdateInfo& ui);
    void insert(string_t s);
    void setCursor(pos_type);
    void stateChanged(int astate, bool enable);

    void getLayoutInfo(LayoutInfo& info);

    static bool run(string_t& str, int maxlen,
                    const string_t& title, const string_t& req,
                    int width, int flags);

    Signal1<int> sig_changed;
};

class UIFocusIterator : public UIWidget {
    int from, to, flags;
 public:
    UIFocusIterator(int from, int to, int flags);
    ~UIFocusIterator();

    void getLayoutInfo(LayoutInfo& info);
    void draw(Ptr<GfxCanvas> can, UpdateInfo& ui);
    bool handleEvent(const UIEvent& e, bool second_pass);

    enum {
        fi_Horiz = 1,
        fi_Vert  = 2,
        fi_Tab   = 4,
        fi_INTERNAL_NEGATE = 64
    };
};

class UIFrameGroup : public UIGroup {
    int frame;
 public:
    UIFrameGroup(Ptr<UILayoutManager> lay, int frame = 1);
    void drawInterior(Ptr<GfxCanvas> can, UpdateInfo& ui);
    GfxRect getExtentFor(GfxRect client);
    GfxRect getClientRectFor(GfxRect ex);
};

class UIFocusGroup : public UIGroup {
 public:
    UIFocusGroup(int id = 0);
    UIFocusGroup(Ptr<UILayoutManager> lay, int id = 0);
    ~UIFocusGroup();
    void drawInterior(Ptr<GfxCanvas> can, UpdateInfo& ui);
    void stateChanged(int state, bool enable);
    bool handleEvent(const UIEvent& e, bool second_pass);
    GfxRect getExtentFor(GfxRect client);
    GfxRect getClientRectFor(GfxRect ex);
};

class UINumberSelector : public UIWidget {
    long current;
    long low, high;
    int  step;
 public:
    UINumberSelector(int id, long low, long high, int step);
    ~UINumberSelector();
    void setValue(long val);
    long getValue() const { return current; }
    bool handleEvent(const UIEvent& e, bool second_pass);

    long getMin() const { return low; }
    long getMax() const { return high; }
    int  getStep() const { return step; }

    Signal1<int> sig_changed;
};

// class UIDecimalSelector : public UINumberSelector {
//  public:
// };

#endif
