/**
  *  \file ui/window.cc
  *  \brief Top-Level Windows and Root Window
  *
  *  Implementation of the Top-Level Window widget and a root group.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <SDL_timer.h>
#include "ui/window.h"
#include "gfx/multiclip.h"
#include "ui/uiglobal.h"
#include "gfx/palette.h"
#include "ui/util.h"
#include "ui/event.h"
#include "gfx/clip.h"
#include "cpluslib/assert.h"

/**
  *  \class UIRoot
  *  \brief Root Group
  *
  *  This implements a ``root group'': a group which does not need
  *  a owner group.
  *
  *  This particular incarantion supports:
  *  - overlapping windows
  *  - painting on a GfxCanvas (which it expects to own completely)
  *  - nested invocations of the event loop
  */

/** Create Root Group on Canvas. The canvas should be a GfxSurface
    or something similar. */
UIRoot::UIRoot(Ptr<GfxCanvas> ascreen)
    : UIWidgetContainer(),
      multi_clip(new GfxMultiClip(ascreen)),
      screen(ascreen)
{
    setExtent(GfxRect(ascreen->getSurface()), true);
    setState(st_Exposed | st_Focused | st_Selected, true);
}

Ptr<GfxCanvas>
UIRoot::getCanvasFor(UIWidget* child)
{
    sync();
    
    if (child == first || !child) {
        return screen;
    }

    Ptr<GfxMultiClip> mc = new GfxMultiClip(screen);
    mc->add(GfxRect(screen->getSurface()));
    for (UIWidget* p = first; p && p != child; p = p->getNext())
        mc->remove(p->getExtent());

    return mc;
}

/** Draw Background. Child classes can override this to do something
    more interesting. \default fills the background with COLOR_BLACK. */
void
UIRoot::drawBackground(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    can->drawBar(getExtent(), standard_colors[COLOR_BLACK], GFX_TRANSPARENT,
                 SOLID_FILL, SDL_ALPHA_OPAQUE);
    ui.update(getExtent());
}

void
UIRoot::draw(Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    Ptr<GfxMultiClip> mc = new GfxMultiClip(can.ptr());
    mc->add(GfxRect(screen->getSurface()));

    for (UIWidget* p = first; p; p = p->getNext()) {
        p->draw (mc, ui);
        mc->remove (p->getExtent());
    }
    if (mc->isVisible(getExtent()))
        drawBackground(mc, ui);
}

void
UIRoot::drawBelow(UIWidget* child, Ptr<GfxCanvas> can, UpdateInfo& ui)
{
    sync();
    
    Ptr<GfxMultiClip> mc = new GfxMultiClip(can.ptr());
    mc->add(GfxRect(screen->getSurface()));
    for (UIWidget* p = child->getNext(); p; p = p->getNext()) {
        p->draw (mc, ui);
        mc->remove(p->getExtent());
    }
    if (mc->isVisible(getExtent()))
        drawBackground(mc, ui);
}

/** Event Handling. This does the focus policy:
    - keyboard and command events go to topmost window only;
    - mouse events go to all windows that can see it;
    - rest (timer, broadcast) goes to everyone. */
bool
UIRoot::handleEvent(const UIEvent& e, bool second_pass)
{
    /* events we do ourselves */
    if (e.type == ev_Key) {
        if (e.key.code == ss_Ctrl + ss_Alt + 's') {
            doScreenshot();
            return true;
        }
        if (e.key.code == ss_Ctrl + ss_Alt + 'q') {
            putKey(vk_Quit);
            return true;
        }
    }

    /* propagate stuff */
    if (e.type == ev_Key) {
        /* top widget only if modal, propagate otherwise. */
        for (UIWidget* p = first; p; p = p->getNext()) {
            if (p->handleEvent(e, second_pass))
                return true;
            if (p->hasState(st_Modal))
                return false;
        }
        return false;
    }        
    if (e.type == ev_Command) {
        /* top widget only */
        if (first)
            return first->handleEvent(e, second_pass);
        else
            return false;
    }
    if (e.type == ev_Mouse) {
        /* widgets which see it only */
        int mx = e.mouse.x;
        int my = e.mouse.y;
        for (UIWidget* p = first; p; p = p->getNext()) {
            if (p->handleEvent(e, second_pass))
                return true;
            if (p->isInside(mx, my) || p->hasState(st_Modal))
                return false;
        }
        return false;
    }
    /* other widgets: broadcast */
    for (UIWidget* p = first; p; p = p->getNext())
        if (p->handleEvent(e, second_pass))
            return true;

    /* UIWidget::handleEvent, actually */
    return UIWidgetContainer::handleEvent(e, second_pass);
}

int
UIRoot::run()
{
    stop_requested = false;
    while (!stop_requested) {
        UIEvent e;
        if (!haveQueuedEvent())
            doDeferredRedraws();
        getEvent(e, hasState(st_IdleNotify));
        if (! handleEvent(e, false))
            handleEvent(e, true);
    }
    stop_requested = false;  /* in case handleEvent() calls run again */
    return stop_arg;
}

void
UIRoot::stop(int arg)
{
    stop_arg       = arg;
    stop_requested = true;
}

/** Process Timer Event. Post timer event to topmost window, if needed.
    Call this occasionally while doing expensive computations. This
    permits display of `idle' indicators.

    \note that this is different from standard timer events which are
    posted to all windows. This one goes to the topmost window only
    for performance reasons. Wo don't want to degrade the performance
    of the above-mentioned heavy processing by stupid event handling. */
void
UIRoot::postTimerEvent()
{
    static long old_timer = 0;
    long this_timer = getTimer();

    if (this_timer == old_timer)
        return;

    old_timer = this_timer;

    UIEvent e;
    e.type = ev_Tick;
    if (first)
        first->handleEvent(e, false) || first->handleEvent(e, true);
}

/** Center Widget on Screen. */
void
UIRoot::centerWidget(UIWidget& w)
{
    SDL_Surface* sfc = screen->getSurface();
    int x0 = (sfc->w - w.getExtent().w) / 2;
    int y0 = (sfc->h - w.getExtent().h) / 2;

    w.setExtent(GfxRect(x0, y0, w.getExtent().w, w.getExtent().h));
}

/** Handle Child Resize.
    \todo Right now, this is almost a copy of UIGroup::handleChildResize(). */
void
UIRoot::handleChildResize(UIWidget* child, GfxRect oldsize)
{
    sync();
    if (child->getExtent().contains(oldsize)) {
        /* child completely covers old position. */
        child->drawWidget(true);
    } else {
        /* child has moved. Complete redraw in case something previously
           obscured (i.e. background) now becomes visible. */
        Ptr<GfxMultiClip> mc = new GfxMultiClip(screen);
        mc->add(child->getExtent());
        mc->add(oldsize);

        UpdateInfo ui(mc->getSurface());
        draw(mc, ui);
        ui.doUpdates();
    }
}

/** Add Window. Adds the window as the top-most (current) window.
    It need not be a UIWindow. */
void
UIRoot::add(UIWidget& w)
{
    multi_clip->remove(w.getExtent());
    sync();
    if (hasState (st_Exposed) && screen->getSurface()->format->BytesPerPixel > 1) {
#if 0
#if 1
        GfxRect size = w.getExtent();
        Ptr<GfxPixmap> pix = new GfxPixmap(size.w, size.h, screen->getSurface()->format);
        Ptr<GfxCanvas> sfc = new GfxOffset(-size.x, -size.y, pix);
        UpdateInfo ui(pix->getSurface());
        w.draw(sfc, ui);
        ui.doUpdates();

        UIPollableTimer timer(30);
        static const int offsets[] = {
            0, 8, 4, 12, 2, 10, 6, 14, 3, 11, 7, 15, 5, 13, 9, 1
        };
        for (int i = 0; i < 16; ++i) {
            SDL_Surface* p = pix->getSurface();
            for (int y = offsets[i]; y < size.h; y += 16)
                screen->blitSurface(size.x, size.y, p,
                                    GfxRect(0, y, size.w, 1));
            screen->update(size);
            timer.waitTime(i+1);
        }
#else
        /* funky effect stolen from "Master of Antares" by EAD */
        UIPollableTimer timer(30);
        GfxRect size = w.getExtent();
#if 0
        for (int i = 1; i < 10; i += 2) {
            screen->drawBar(size, standard_colors[COLOR_WHITE], GFX_TRANSPARENT,
                            SOLID_FILL, i*i*2);
            screen->update(size);
            timer.waitTime(i/2+1);
        }
        screen->drawBar(size, standard_colors[COLOR_WHITE], GFX_TRANSPARENT,
                        SOLID_FILL, SDL_ALPHA_OPAQUE);
#endif
        Ptr<GfxPixmap> pix = new GfxPixmap(size.w, size.h, screen->getSurface()->format);
        Ptr<GfxCanvas> sfc = new GfxOffset(-size.x, -size.y, pix);
        UpdateInfo ui(pix->getSurface());
        w.draw(sfc, ui);
        ui.doUpdates();

        timer.setTime(0);
        for (int i = 1; i < 10; i += 2) {
            SDL_Surface* p = pix->getSurface();
            SDL_SetAlpha(p, SDL_SRCALPHA, i*i*2);
            screen->blitSurface(size.x, size.y, p, GfxRect(p));
            screen->update(size);
            timer.waitTime(i/2+1);
        }
#endif
#endif
    }

    /* build window stack */
    if (first)
        first->setLinks(&w, first->getNext());
    w.setLinks(0, first);
    w.setOwner(this);
    first = &w;
    if (!last)
        last = &w;
    setFocus(0);                // remove focus, so add will give it to our window
    UIWidgetContainer::add(w);
    selectWidget(&w);
}

/** Remove the specified widget. */
void
UIRoot::removeWidget(UIWidget& w)
{
    if (&w == first) {
        /* closing topmost window */
        multi_clip->add(w.getExtent());
        UIWidgetContainer::removeWidget(w);
    } else {
        /* closing "middle" window */
        sync();
        UIWidget* p = first;
        while (p && p->getNext() != &w)
            p = p->getNext();
        ASSERT(p);              // if this fails, w is not a child of us

        UIWidgetContainer::removeWidget(w);

        /* FIXME: this can be optimized to not redraw windows before w */
        Ptr<GfxClipRectFilter> cr =
            new GfxClipRectFilter(w.getExtent(), screen);
        UpdateInfo ui (screen->getSurface());
        draw (cr, ui);
        ui.doUpdates();
    }
}

/** Redraw everything. Redraws the whole screen. */
void
UIRoot::redraw()
{
    sync();
    UpdateInfo ui (screen->getSurface());
    draw (screen, ui);
    ui.doUpdates();
}

void
UIRoot::doDeferredRedraws()
{
    sync();
    UpdateInfo ui (screen->getSurface());
    drawDeferred (screen, ui);
    ui.doUpdates();
}

void
UIRoot::sync()
{
    if (multi_clip->empty())
        return;
    UpdateInfo ui (screen->getSurface());
    draw (multi_clip, ui);
    ui.doUpdates();
    multi_clip->clear();
}

void
UIRoot::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_NoLayout;
}

/******************************** UIWindow *******************************/

/**
  *  \class UIWindow
  *  \brief Standard Window
  *
  *  This is a standard framed window.
  */

Ptr<GfxPixmap> UIWindow::frame_blue;

UIWindow::UIWindow(const GfxRect& size, string_t aname, int afs, int abs,
                   int aborder)
    : UIGroup(0),
      name(aname),
      frame_style(afs),
      bg_style(abs),
      border_space(aborder)
{
    setExtent(size, true);
    if (!frame_blue.ptr())
        frame_blue = requestPixmap("winblue");
}

UIWindow::UIWindow(string_t aname, int afs, int abs, int aborder)
    : UIGroup(0),
      name(aname),
      frame_style(afs),
      bg_style(abs),
      border_space(aborder)
{
    if (!frame_blue.ptr())
        frame_blue = requestPixmap("winblue");
}

/** Draw a window. Available as static function to be callable from
    outside (widgets that look like windows but aren't, like the VCR
    screen) */
void
UIWindow::draw(Ptr<GfxCanvas> canvas, const GfxRect& extent,
               int bg_style, string_t name, bool bg)
{
    GfxParameters parm;
    parm.color   = standard_colors[COLOR_WHITE];
    parm.just_x  = 1;

    drawFrameUp(canvas, extent);
    drawTiledArea(canvas, GfxRect(extent.x + 1, extent.y + 1,
                                  extent.w - 2, 22),
                  frame_blue, standard_colors[COLOR_BLUEBLACK], 16);
    drawTiledArea(canvas, GfxRect(extent.x + 1, extent.y + 23,
                                  2, extent.h - 26),
                  frame_blue, standard_colors[COLOR_BLUEBLACK], 16);
    drawTiledArea(canvas, GfxRect(extent.x + extent.w - 3, extent.y + 23,
                                  2, extent.h - 26),
                  frame_blue, standard_colors[COLOR_BLUEBLACK], 16);
    drawTiledArea(canvas, GfxRect(extent.x + 1, extent.y + extent.h - 3,
                                  extent.w - 2, 2),
                  frame_blue, standard_colors[COLOR_BLUEBLACK], 16);
    
    drawFrameDown(canvas, GfxRect(extent.x + 3, extent.y + 23,
                                  extent.w - 6, extent.h - 26));
    if (bg)
        // synchronize this with drawInterior
        canvas->drawBar(GfxRect(extent.x + 4, extent.y + 24,
                                extent.w - 8, extent.h - 28),
                        standard_colors[bg_style],
                        GFX_TRANSPARENT,
                        SOLID_FILL,
                        SDL_ALPHA_OPAQUE);

    outTextF(canvas, parm, fonts[0], extent.x + extent.w/2, extent.y + 2, extent.w, name);
}

void
UIWindow::draw(Ptr<GfxCanvas> canvas, UpdateInfo& ui)
{
    UIWindow::draw(canvas, getExtent(), bg_style, name, false);
    ui.update(getExtent());
    UIGroup::draw(canvas, ui);
}

void
UIWindow::drawInterior(Ptr<GfxCanvas> canvas, UpdateInfo& ui)
{
    // synchronize this with static draw()
    GfxRect r(extent.x + 4, extent.y + 24,
              extent.w - 8, extent.h - 28);
    canvas->drawBar(r, standard_colors[bg_style], GFX_TRANSPARENT,
                    SOLID_FILL, SDL_ALPHA_OPAQUE);
    ui.update(r);
}

/** Event Handling. Standard event handling, owner does focus policy. */
bool
UIWindow::handleEvent(const UIEvent& e, bool second_pass)
{
    return UIGroup::handleEvent(e, second_pass);
}

GfxRect
UIWindow::getClientRectFor(GfxRect ext)
{
    ext.w -= 8 + 2*border_space;
    ext.h -= 28 + 2*border_space;
    ext.x += 4 + border_space;
    ext.y += 24 + border_space;

    return ext;
}

GfxRect
UIWindow::getExtentFor(GfxRect cli)
{
    cli.w += 8 + 2*border_space;
    cli.h += 28 + 2*border_space;
    cli.x -= 4 + border_space;
    cli.y -= 24 + border_space;

    return cli;
}
