/**
  *  \file bsearch.h
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_BSEARCH_H
#define PCC_V2_BSEARCH_H

#include <iterator>

/** Binary Search.
    \param beg     random access iterator, beginning of sequence.
    \param count   length of sequence
    \param comp    search predicate. Must be a class with public
                   methods `bool isLess(T)' and `bool isEqual(T)',
                   where T is the type of *Iter. These should return
                   true iff the value to search is less than/equal
                   to the parameter.
    \return index of found object, or -1. */
template<class Iter, class Comp>
typename std::iterator_traits<Iter>::difference_type
bin_search(Iter beg, typename std::iterator_traits<Iter>::difference_type count,
           const Comp& comp)
{
    typename std::iterator_traits<Iter>::difference_type offset = 0;
    while (count > 4) {         /* arbitrary */
        typename std::iterator_traits<Iter>::difference_type mid = (count-1)/2;
        if (comp.isLess(beg[mid])) {
            count = mid;
        } else {
            offset += mid;
            beg += mid;
            count -= mid;
        }
    }
    while (count > 0) {
        --count;
        if(comp.isEqual(beg[count]))
            return count + offset;
    }
    return -1;
}

/** Binary Search on Sequence.
    This function is like bin_search, but handles the standard STL
    conventions for sequences.
    \param beg     random access iterator, beginning of sequence.
    \param end     end of sequence
    \param comp    search predicate. Must be a class with public
                   methods `bool isLess(T)' and `bool isEqual(T)',
                   where T is the type of *Iter. These should return
                   true iff the value to search is less than/equal
                   to the parameter.
    \return iterator pointing to found object, or end. */
template<class Iter, class Comp>
Iter
bin_search_seq(Iter beg, Iter end, const Comp& comp)
{
    typename std::iterator_traits<Iter>::difference_type match =
        bin_search(beg, std::distance(beg, end), comp);

    if (match >= 0)
        return std::advance(beg, match);
    else
        return end;
}

#endif
