#!/usr/bin/perl -w
#
#  Convert charsets.in into charsets.out (which can then be edited into
#  recode.cc)
#
#  usage: chargen.pl <charsets.in >charsets.out
#
#  Input file format:
#  `rfc FILENAME'          read RFC 1345 for short mnemonic character names
#  `cs NAME'               begin definition of a character set
#  `name TEXT'             set dexcription of current character set
#  `note TEXT'             add comment (source-code) to current cs
#  `file NAME'             include file verbatim
#  `XX ZZZZ'               (XX and ZZZZ are hex numbers) character XX is U+ZZZZ
#  `XX'                    (XX is hex number) character XX is undefined
#  `alias XXXX=YYYY'       treat U+XXXX and U+YYYY identical
#  `cname XXXX NAME'       set short mnemonic for U+XXXX
#

use strict;
use IO::Handle;
use vars qw{ %in_csnames %in_cs_chars %in_csyms %in_used_csyms 
             $cur_cs %in_aliases %in_notes };

sub output_format { 2 }
sub singleton_warnings { 0 }

read_file ("<stdin>", \*STDIN);

sub read_file {
    my $fn = shift;
    my $fd = shift;
    while (<$fd>) {
        chomp;
        s/\s*\#.*//;
        s/^\s+//;
        s/\s+$//;
        s/\032//g;
        next if /^$/;
        if (/^rfc\s*(.*)/) {
            read_rfc ($1);
        } elsif (/^cs\s*(.*)/) {
            $cur_cs = $1;
        } elsif (/^name\s*(.*)/) {
            die "$fn:$.: no `cs' directive seen yet\n" if !defined $cur_cs;
            die "$fn:$. we already have a `name' for `$cur_cs'\n" if exists $in_csnames{$cur_cs};
            $in_csnames{$cur_cs} = $1;
        } elsif (/^note\s*(.*)/) {
            die "$fn:$.: no `cs' directive seen yet\n" if !defined $cur_cs;
            if (exists $in_notes{$cur_cs}) {
                $in_notes{$cur_cs} .= "\n       $1";
            } else {
                $in_notes{$cur_cs} = $1;
            }
        } elsif (/^file\s*(.*)/) {
            my $fd = new IO::Handle;
            open $fd, $1 or die "$1: $!\n";
            read_file ($1, $fd);
            close $fd;
        } elsif (/^0?x?([0-9A-F]{2})\s+0?x?([0-9A-F]{4})/i) {
            die "$fn:$.: no `cs' directive seen yet\n" if !defined $cur_cs;
            if (hex($1) >= 128) {
                $in_cs_chars{$cur_cs}[hex($1) - 128] = hex $2;
            } else {
                if (hex($1) != hex($2)) {
                    print STDERR "$fn:$.: character $1 maps to unicode $2 in `$cur_cs'\n";
                }
            }
        } elsif (/^0?x?([0-9A-F]{2})\s*$/i) {
            # undefined character slot. Ignore
        } elsif (/^alias\s*([0-9a-f]{4})\s*=\s*([0-9a-f]{4})/i) {
            $in_aliases{hex $1} = unalias(hex $2);
        } elsif (/^cname\s*([0-9a-f]{4})\s+(.*)/i) {
            $in_csyms{unalias(hex $1)} = $2;
        } else {
            die "$fn:$.: unknown command `$_'\n";
        }
    }
}

foreach (sort keys %in_csnames) {
    do_output ($_);
}
if (output_format() == 2) {
    my @output;
    foreach (keys %in_csyms) {
        if (exists $in_used_csyms{$in_csyms{$_}}) {
            push @output, sprintf "    { 0x%04x, \"%s\" },\n", $_, c_escape($in_csyms{$_});
        }
    }
    print sort @output;
}

if (singleton_warnings()) {
    foreach (sort keys %in_used_csyms) {
        if (@{$in_used_csyms{$_}} == 1) {
            print STDERR "note: `$_' only appears in `$in_used_csyms{$_}[0]'\n";
        }
    }
}

exit 0;

sub read_rfc {
    my $name = shift;
    open FILE, $name or die "$name: $!\n";
    while (<FILE>) {
        if (/^ (\S+)\s+([0-9a-f]{4})\s+[^a-z]{3}/) {
            $in_csyms{hex $2} = $1;
        }
    }
    close FILE;
}

sub do_output {
    my $csname = shift;
    if (!exists $in_cs_chars{$csname}) {
        print STDERR "no characters defined for character set `$csname'\n";
        return;
    }
    if (exists $in_notes{$csname}) {
        print "    /* ", $in_notes{$csname}, " */\n";
    }
    print "    { \"", c_escape ($csname), "\",\n";
    print "      \"", c_escape ($in_csnames{$csname}), "\",\n";
    print "      {\n";
    if (output_format() == 1) {
        # Output format 1 (RFC 1345-based)
        foreach (0..127) {
            print "         " if ($_ % 8) == 0;
            my $char = unalias($in_cs_chars{$csname}[$_]);
            if (!defined $char) {
                print ' "??",';
            } elsif (! defined $in_csyms{$char}) {
                printf STDERR "character U+%04x has no RFC1345 name\n", $char;
                print ' "??",';
            } else {
                print ' "', c_escape($in_csyms{$char}), '",';
                push @{$in_used_csyms{$in_csyms{$char}}}, $csname;
            }
            print "\n" if ($_ % 8) == 7;
        } 
    } else {
        # Output format 2 (Unicode-based)
        foreach (0..127) {
            print "         " if ($_ % 8) == 0;
            my $char = unalias($in_cs_chars{$csname}[$_]);
            if (!defined $char) {
                print ' 0x0000,';
            } elsif (! defined $in_csyms{$char}) {
                printf STDERR "character U+%04x has no RFC1345 name\n", $char;
                print ' 0x0000,';
            } else {
                printf ' 0x%04x,', $char;
                push @{$in_used_csyms{$in_csyms{$char}}}, $csname;
            }
            print "\n" if ($_ % 8) == 7;
        }
    }
    print "      }\n";
    print "    },\n\n";
}

sub c_escape {
    my $text = shift;
    $text =~ s/([\\\"])/\\$1/g;
    $text;
}

sub unalias {
    my $c = shift;
    defined($c) && exists($in_aliases{$c}) ? $in_aliases{$c} : $c;
}
