/**
  *  \file util/cmdline.cc
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <iostream>
#include "util/cmdline.h"
#include "util/nls.h"
#include "cpluslib/format.h"

/*! \class CommandLineParser
    Encapsulates the logic for parsing a command line into
    options and non-options. Derived classes implement the
    actual parameter interpretation.

    Our syntax:
    - options start with `-' or `--', e.g. `-display', `--help'.
      Single-dash and double-dash is identical.
    - an argument consisting of just two dashes stops interpretation
      of options and causes the rest to be interpreted as non-options.
    - everything else is a non-option. Note that `-' is a non-option,
      too; this is useful because it is usually used in place of a
      file name to mean stdin/stdout.

    With `traditional', it uses the traditional syntax, where
    a single dash introduces a block of combined options
    (`-blah' == `-b -l -a -h'). */

const char* CommandLineParser::prog;

/** Initialize. */
CommandLineParser::CommandLineParser(bool atrad)
    : argv(0), traditional(atrad)
{ }

CommandLineParser::~CommandLineParser()
{ }

/** Internal: get next element from parameter vector. With
    "saturation" at end. */
const char*
CommandLineParser::getNext()
{
    if (*argv)
        return *argv++;
    else
        return 0;
}

/** Get argument. handleOption() can call this to request the current
    argument. */
const char*
CommandLineParser::getArg()
{
    if (!arg)
        arg = getNext();
    return arg;
}

/** Parse. This one does the actual work. */
void
CommandLineParser::parse(char** aargv)
{
    argv = aargv;
    prog = getNext();
    parseStarted();
    while (const char* p = getNext()) {
        arg = 0;
        if (p[0] != '-' || p[1] == 0) {
            /* non-option: doesn't start with `-' or only one character */
            handleNonoption(p);
        } else if(p[1] == '-' && p[2] == 0) {
            /* option-stop `--': rest is non-options */
            while (const char* q = getNext())
                handleNonoption(q);
            break;
        } else {
            /* option. Skip at most two dashes */
            ++p;
            if (traditional && *p != '-') {
                /* traditional handling. `-blah' == `-b -l -a -h' */
                while (*p)
                    handleOption(string_t(1, *p++), false);
            } else {
                /* X11/GNU style parsing. `-display' == `--display' */
                if (*p == '-')
                    ++p;
                const char* eq = std::strchr(p, '=');
                if (!eq) {
                    handleOption(p, false);
                } else {
                    arg = eq+1;
                    handleOption(string_t(p, eq-p), true);
                }
            }
        }
    }
    parseFinished();
}

/** Hook: called before parsing the command line.
    \default does nothing */
void
CommandLineParser::parseStarted()
{ }

/** Hook: called after successfully parsing the command line.
    \default does nothing */
void
CommandLineParser::parseFinished()
{ }

/** \fn CommandLineParser::handleOption(const string_t&, bool)
    Handle an option. The \c have_arg parameter is true when the option
    was specified with a parameter (`--param=foo'); this is so that
    the parser can reject options that ought not to take parameters
    but the user gave some. No matter how the argument was specified,
    handleOption() can request it using getArg(). */

/** \fn CommandLineParser::handleNonoption(const string_t&)
    Handle a non-option. */

/*********************** StandardCommandLineParser ***********************/

StandardCommandLineParser::StandardCommandLineParser(bool atrad)
    : CommandLineParser(atrad)
{ }

StandardCommandLineParser::~StandardCommandLineParser()
{ }

void
StandardCommandLineParser::addOptionSet(const Option* opt,
                                        std::size_t count)
{
    if (count)
        options.insert(options.end(), opt, opt + count);
}

void
StandardCommandLineParser::handleOption(const string_t& opt, bool have_arg)
{
    /* Find option */
    int index = -1;
    string_t error;
    for (std::vector<Option>::size_type i = 0; i < options.size(); ++i) {
        std::size_t len = std::strlen(options[i].name);
        if (opt.length() >= len && opt.compare(0, len, options[i].name, len) == 0) {
            /* Option matches. */
            if (index < 0) {
                index = i;
            } else {
                /* ambiguous */
                if (! error.length())
                    /* translator: this should yield a partial
                       sentence, we'll append a number of tokens of
                       the form ", `name'" to this text */
                    error = format(_("Option `%s' is ambiguous, could be `%s'"),
                                     opt, options[index].name);
                else
                    error += format(", `%s'", options[i].name);
            }
        }
    }
    if (error.length()) {
        handleError(error);
    } else if (index < 0) {
        handleError(format(_("Unknown option `%s'"), opt));
    } else if (have_arg && !options[index].takes_arg) {
        handleError(format(_("Option `%s' does not take an argument"), opt));
    } else if (options[index].takes_arg && !getArg()) {
        handleError(format(_("Option `%s' requires an argument"), opt));
    } else {
        options[index].handler(*this, options[index]);
    }
}

void
StandardCommandLineParser::printHelp(std::ostream& os)
{
    for (std::vector<Option>::iterator i = options.begin(); i != options.end(); ++i)
        if (i->help)
            os << _(i->help) << std::endl;
}

void
StandardCommandLineParser::handleError(const string_t& msg)
{
    throw CommandLineException(msg);
}
