/**
  *  \file util/conf-file.cc
  *  \brief High-Level Configuration File Access
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <map>
#include <fstream>
#include <cerrno>
#include "util/confpars.h"
#include "util/misc.h"
#include "util/console.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "cpluslib/string.h"

struct ConfigFile::Data {
    struct Entry {
        string_t value;
        bool modified;
    };
    typedef std::map<string_t, Entry> MapType;

    MapType values;
};

namespace {
    class CfParser : public ConfigParser {
     protected:
        ConfigFile::Data& data;
        string_t name;
     public:
        CfParser(ConfigFile::Data& data, const string_t& name)
            : data(data), name(name) { }
        void error(int line_id, string_t msg);
    };
    class CfReader : public CfParser {
     public:
        CfReader(ConfigFile::Data& data, const string_t& name)
            : CfParser(data, name) { }
        void assign(int line_id, string_t name, string_t value, string_t line);
    };
    class CfWriter : public CfParser {
        string_t output_pre;
        string_t output_post;
     public:
        CfWriter(ConfigFile::Data& data, const string_t& name)
            : CfParser(data, name) { }
        void assign(int line_id, string_t name, string_t value, string_t line);
        void unparsed(int line_id, string_t line);
        void write(std::ostream& out, const string_t& section, bool in_section);
    };
};

void
CfParser::error(int line_id, string_t msg)
{
    /* this message is the same as in game/config.cc */
    console.write(LOG_WARN, format(_("%s:%d: %s -- line has been ignored"), name, line_id, msg));
}

void
CfReader::assign(int /*line_id*/, string_t name, string_t value, string_t /*line*/)
{
    ConfigFile::Data::Entry& i = data.values[strUCase(name)];
    i.value = value;
    i.modified = false;
}

void
CfWriter::assign(int /*line_id*/, string_t name, string_t /*value*/, string_t line)
{
    output_pre += output_post;
    output_post.clear();

    ConfigFile::Data::MapType::iterator i = data.values.find(name);
    if (i != data.values.end() && i->second.modified) {
        /* a variable we know and we want to modify */
        output_pre += name + " = " + i->second.value + "\n";
        output_post = string_t();
        i->second.modified = false;
    } else {
        /* a variable in our section which we don't want to modify */
        output_pre += line + "\n";
    }
}

void
CfWriter::unparsed(int /*line_id*/, string_t line)
{
    output_post += line + "\n";
}

void
CfWriter::write(std::ostream& out, const string_t& section, bool in_section)
{
    if (!output_pre.length() && !in_section) {
        out << "% " << section << "\n";
    } else {
        out << output_pre;
    }

    for (ConfigFile::Data::MapType::iterator i = data.values.begin(); i != data.values.end(); ++i) {
        if (i->second.modified) {
            out << i->first << " = " << i->second.value << "\n";
            i->second.modified = false;
        }
    }

    out << output_post;
}

/******************************* ConfigFile ******************************/

/** \class ConfigFile
    \brief High-Level Configuration File Access

    This class provides a high-level configuration file access
    mechanism. It reads configuration files, and can also write
    modified versions thereof out again, trying to preserve the
    user's structure. */

/** Create new, empty data set. */
ConfigFile::ConfigFile()
    : data(*new Data)
{ }

ConfigFile::~ConfigFile()
{
    delete &data;
}

/** Load configuration file. All information from the file will become
    available through this object.
    \param file_name  file name
    \param section    section to read
    \param in_section true iff we start reading even without a section
                      delimiter.
    \returns true iff file could be opened. */
bool
ConfigFile::load(string_t file_name, string_t section, bool in_section)
{
    CfReader reader(data, file_name);
    std::ifstream input(file_name.c_str());
    if (!input)
        return false;
    reader.parse(input, section, in_section);
    return true;
}

bool
ConfigFile::load(string_t file_name)
{
    return load(file_name, string_t(), true);
}

/** Write back changes to disk. */
void
ConfigFile::save(string_t file_name, string_t section, bool in_section)
{
    CfWriter writer(data, file_name);
    {
        std::ifstream input(file_name.c_str());
        if (input)
            writer.parse(input, section, in_section);
    }
    {
        std::ofstream output(file_name.c_str());
        if (!output)
            console.write(LOG_ERROR, format(_("Unable to create `%s'."), file_name));
        else
            writer.write(output, section, in_section);
    }
}

void
ConfigFile::save(string_t file_name)
{
    save(file_name, string_t(), true);
}

/** Fetch value, nondestructively. This reads a value from the container
    if it exists, and does not add it when it does not exist.
    \param key  key to read
    \param ret  result is returned here
    \returns true iff found */
bool
ConfigFile::getValueNC(string_t key, string_t& ret) const
{
    Data::MapType::iterator i = data.values.find(strUCase(key));
    if (i != data.values.end()) {
        ret = i->second.value;
        return true;
    } else {
        return false;
    }
}

/** Modify value. */
void
ConfigFile::setValue(string_t key, string_t value)
{
    key = strUCase(key);    
    Data::MapType::iterator i = data.values.find(key);
    if (i != data.values.end() && i->second.value == value)
        return;

    Data::Entry& v = data.values[key];
    v.value = value;
    v.modified = true;
}

void
ConfigFile::setValue(string_t key, const char* value)
{
    setValue(key, string_t(value));
}

void
ConfigFile::setValue(string_t key, int value)
{
    setValue(key, itoa(value));
}

void
ConfigFile::setValue(string_t key, bool value)
{
    setValue(key, value ? "yes" : "no");
}

/** Fetch string value. When the value is not stored in the file, the
    default value will be stored.
    \param key   key
    \param value [in] default value [out] value from file
    \return true iff read from file */
bool
ConfigFile::getValue(string_t key, string_t& value)
{
    if (!getValueNC(key, value)) {
        setValue(key, value);
        return false;
    } else {
        return true;
    }
}

/** Fetch integer value. Like getValue(string_t, string_t&), but
    returns an integer. Invalid integers are treated as non-existant
    and are overwritten. */
bool
ConfigFile::getValue(string_t key, int& value)
{
    string_t tmp;
    if (getValueNC(key, tmp) && stringValue(tmp, value)) {
        return true;
    } else {
        setValue(key, itoa(value));
        return false;
    }
}

/** Fetch boolean value. Like getValue(string_t, string_t&), but
    returns a bool. Invalid bools are treated as non-existant and are
    overwritten. */
bool
ConfigFile::getValue(string_t key, bool& value)
{
    string_t tmp;
    if (getValueNC(key, tmp)) {
        int tmpi;
        if (stringValue(tmp, tmpi) && (tmpi==0 || tmpi==1)) {
            value = (tmpi != 0);
            return true;
        }
        if (stringMatch("Yes", tmp) || stringMatch("True", tmp)) {
            value = true;
            return true;
        }
        if (stringMatch("No", tmp) || stringMatch("False", tmp)) {
            value = false;
            return true;
        }
    }
    setValue(key, value ? "yes" : "no");
    return false;
}
