/**
  *  \file confpars.cc
  *  \brief Configuration Parser
  *
  *  Logic for parsing pconfig.src-like config files.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <iostream>
#include "util/confpars.h"
#include "util/except.h"
#include "util/nls.h"
#include "cpluslib/string.h"

/** Preprocess config line. Removes comments and leading/trailing
    whitespace. */
static string_t
preprocess(string_t line)
{
    string_t::size_type n = line.find('#');
    if (n != string_t::npos)
        line.erase(n);
    
    return strTrim(line);
}

/** \class ConfigParser
    \brief Configuration File Parser (pconfig.src alike)

    This class encapsulates the logic for parsing PCONFIG.SRC-alike
    configuration files. In a nutshell:
    - sections separated by `% sectionname'
    - comments starting with `\#'
    - assignments `key = value' in each section

    To parse a configuration file, derive a class from this and implement
    at least the assign() method. */
    
/** Line not parsed. This function is called when a line does not
    contain a `key = value' pair, but is nonetheless valid. This is
    the case for
    - section delimiters
    - blank lines/comment lines
    - lines in different sections

    \param line_id line number
    \param line    verbatim text of the line

    \default does nothing. You can override it to record the the
    un-parsed lines, too. */
void
ConfigParser::unparsed(int, string_t)
{ }

/** Syntax error. This function is called when the parser encounters
    a syntax error.

    \param line_id line number
    \param msg     error message
    
    \default Throws a FileFormatException. */
void
ConfigParser::error(int line_id, string_t s)
{
    throw FileFormatException(itoa(line_id) + ": " + s);
}

/** Parser Main Entry Point.
    \param in         Stream to read from
    \param section    The section we want to read (case-insensitive)
    \param in_section If true, assume we are in our target section
                      even if the file does not start with a section
                      delimiter. For example, if pconfig.src does not
                      start with `% phost', PHost will read config options
                      anyway.
    \param line_id    How many lines you already read from the stream.
                      This is only to get line numbers right for error
                      messages. Defaults to 0 if omitted. */
void
ConfigParser::parse(std::istream& in, string_t section, bool in_section,
                    int line_id)
{
    string_t line;
    while(getline(in, line)) {
        string_t ppline = preprocess(line);
        ++line_id;
        
        if (ppline.empty()) {
            /* blank line, or comment line */
            unparsed(line_id, line);
            continue;
        }

        if (ppline[0] == '%') {
            if (strCaseCmp(section, strLTrim(ppline.substr(1))) == 0)
                in_section = true;
            else
                in_section = false;
            unparsed(line_id, line);
            continue;
        }

        if (!in_section) {
            unparsed(line_id, line);
            continue;
        }

        string_t::size_type eqpos = ppline.find('=');
        if (eqpos == string_t::npos) {
            error(line_id, _("Syntax error"));
            continue;
        }

        string_t key = strRTrim(ppline.substr(0, eqpos));
        if (key.empty()) {
            error(line_id, _("Syntax error"));
            continue;
        }
        
        assign(line_id, key, strLTrim(ppline.substr(eqpos+1)), line);
    }
}
