/**
  *  \file console.cc
  *  \brief Console Implementation
  *
  *  Implementation of the Console class (management, scrollback) and
  *  output to stdout.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <iostream>
#include "util/console.h"
#include "util/compat.h"
#include "cpluslib/assert.h"

/** \class Console
    \brief Game Console

    This provides a simple console for output, which can be used by
    the various subsystems. Console output is line-based, and has
    importance tags. The Console class maintains a variable-sized
    scrollback buffer. To actually display the console, something must
    be attached to Console::console_changed.

    You can also use the standard stream operations on a Console
    object (because Console is derived from std::ostream). */

/** Standard Console. */
Console console;

/** Console Streambuf. ``Unbuffered'' output, as far as iostreams
    are concerned. This collects characters in a string, and
    outputs it to its console, line by line. */
class Console::Streambuf : public std::streambuf {
    Console* the_console;
    string_t the_line;
 public:
    Streambuf(Console* aconsole);
    int overflow(int c);
    int sync();
};

Console::Streambuf::Streambuf(Console* aconsole)
    : the_console(aconsole)
{ }

int
Console::Streambuf::overflow(int c)
{
    if (c == STREAM_CHAR_TRAITS::eof())
        return 0;
    if (c == '\n') {
        the_console->writeLine(LOG_INFO, the_line);
        the_line.erase();
    } else {
        the_line += STREAM_CHAR_TRAITS::narrow(c);
    }
    return 0;
}

int
Console::Streambuf::sync()
{
    if (!the_line.empty()) {
        the_console->writeLine(LOG_INFO, the_line);
        the_line.erase();
    }
    return 0;
}

/******************************** Console ********************************/

/** Create a new Console. */
Console::Console()
    : std::ostream(0),
      buf(new Streambuf(this)),
      scrollback_length(DEFAULT_SCROLLBACK_SIZE),
      scrollback_buffer(new Message[unsigned(DEFAULT_SCROLLBACK_SIZE)]),
      buf_head(0), buf_fill(0)
{
    init(buf);
}

/** Delete this Console. */
Console::~Console()
{
    delete[] scrollback_buffer;
    delete buf;
}

inline Console::Message&
Console::getLineInternal(int n) const
{
    return scrollback_buffer[(buf_head - n + scrollback_length) % scrollback_length];
}

/** Get console line n. Line 0 is the most recent one, 1 is the one
    before, etc. */
const Console::Message&
Console::getLine(int n) const
{
    return getLineInternal(n);
}

/** Write string. Roughly identical to `some_stream << text <<
    std::endl', but allows specification of message class and format. */
void
Console::write(MessageClass mc, string_t text, int format)
{
    string_t::size_type n;
    while ((n = text.find('\n')) != string_t::npos) {
        writeLine(mc, text.substr(0, n), format);
        text.erase(0, n+1);
    }
    writeLine(mc, text, format);
}

void
Console::writeLine(MessageClass mc, string_t text, int format)
{
    ++buf_head;
    if (buf_head >= scrollback_length)
        buf_head = 0;
    if (buf_fill < scrollback_length)
        ++buf_fill;

    getLineInternal(0) = Message(mc, text, format);
    console_changed.raise(*this, 1);
}

/** Change size of scrollback buffer. This will clear the console. */
void
Console::setScrollbackSize(int n)
{
    ASSERT(n>0);

    delete[] scrollback_buffer;

    scrollback_buffer = new Message[n];
    scrollback_length = n;
    buf_fill          = 0;
    buf_head          = 0;

    console_changed.raise(*this, 0);
}

/** Clear Console. After this command, the console is (appears to be)
    empty. */
void
Console::clear()
{
    buf_fill = 0;
    console_changed.raise(*this, 0);
}

/************************** Standard I/O Console *************************/

static void* stdiocon_cookie;

static void
stdioConsoleHandler(const Console& con, int n)
{
    while (n--) {
        const Console::Message& m = con.getLine(n);
        if (m.mc >= LOG_WARN)
            std::cerr << m.text << std::endl;
        else
            std::cout << m.text << std::endl;
    }
}

/** Turn on Console on standard output. */
void
enableStdioConsole()
{
    if (stdiocon_cookie)
        return;

    stdiocon_cookie = console.console_changed.add(stdioConsoleHandler);
}

/** Turn off Console on standard output. */
void
disableStdioConsole()
{
    if (!stdiocon_cookie)
        return;

    console.console_changed.remove(stdiocon_cookie);
    stdiocon_cookie = 0;
}
