/**
  *  \file except.cc
  *
  *  We implement the "const char*", copy constructors and destructors
  *  manually to defeat compiler inlining. Without that, gcc generates
  *  about 500 bytes to 1k of code and data per throw (basically, the
  *  inlined copy constructors contain a string copy which itself
  *  might throw again, this needs a lot of code to handle the case
  *  where the exception object itself cannot be constructed(!)).
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <cerrno>
#include "util/except.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "cpluslib/stringcv.h"

FileOpenException::FileOpenException(const string_t& awhat)
    : std::runtime_error(toStandardString(awhat))
{ }

FileOpenException::FileOpenException(const char* awhat)
    : std::runtime_error(awhat)
{ }

FileOpenException::FileOpenException(const FileOpenException& e)
    : std::runtime_error(e)
{ }

FileOpenException::~FileOpenException() throw()
{ }

FileFormatException::FileFormatException(const string_t& awhat)
    : std::runtime_error(toStandardString(awhat))
{ }

FileFormatException::FileFormatException(const char* awhat)
    : std::runtime_error(awhat)
{ }

FileFormatException::FileFormatException(const FileFormatException& e)
    : std::runtime_error(e)
{ }

FileFormatException::~FileFormatException() throw()
{ }

FileTooShortException::FileTooShortException(const string_t& fname)
    : FileFormatException(format(_("File `%s' is too short"), fname))
{ }

FileTooShortException::FileTooShortException(const char* fname)
    : FileFormatException(format(_("File `%s' is too short"), fname))
{ }

FileTooShortException::FileTooShortException(const FileTooShortException& e)
    : FileFormatException(e)
{ }

FileTooShortException::~FileTooShortException() throw()
{ }

FileIOException::FileIOException(const string_t& awhat)
    : std::runtime_error(toStandardString(awhat))
{ }

FileIOException::FileIOException(const char* awhat)
    : std::runtime_error(awhat)
{ }

FileIOException::FileIOException(const FileIOException& e)
    : std::runtime_error(e)
{ }

FileIOException::~FileIOException() throw()
{ }

GraphicsIOException::GraphicsIOException(const string_t& awhat)
    : std::runtime_error(toStandardString(awhat))
{ }

GraphicsIOException::GraphicsIOException(const char* awhat)
    : std::runtime_error(awhat)
{ }

GraphicsIOException::GraphicsIOException(const GraphicsIOException& e)
    : std::runtime_error(e)
{ }

GraphicsIOException::~GraphicsIOException() throw()
{ }

UserException::UserException(const string_t& awhat)
    : std::runtime_error(toStandardString(awhat))
{ }

UserException::UserException(const char* awhat)
    : std::runtime_error(awhat)
{ }

UserException::UserException(const UserException& e)
    : std::runtime_error(e)
{ }

UserException::~UserException() throw()
{ }

InitException::InitException(const string_t& awhat)
    : std::runtime_error(toStandardString(awhat))
{ }

InitException::InitException(const char* awhat)
    : std::runtime_error(awhat)
{ }

InitException::InitException(const InitException& e)
    : std::runtime_error(e)
{ }

InitException::~InitException() throw()
{ }

string_t
perrorStr(const string_t& name)
{
    char* s = std::strerror(errno);
    return name + ": " + s;
}
