/**
  *  \file misc.cc
  *  \brief Miscellaneous utilities
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include <cstdlib>
#include <cerrno>
#include <climits>
#include <cstring>
#include "util/misc.h"

using namespace std;  /* to access unqualified isXXX/toXXX, some libc++'s
                         still have these as macros */

/** String Match, PHost way.
    \param pattern  Pattern we want to match against. Consists of
                    an initial sequence of capital letters, followed
                    by a sequence of other letters. Lower-case letters
                    are optional, the rest is mandatory.
    \param tester   The string to test against the pattern.

    Matching is case-insensitive.

    Example: stringMatch("ENglish", p) matches if p is
    "english", or any abbreviation of "english" up to "en". */
bool stringMatch(const char* pattern, const char* tester)
{
    while (*pattern) {
        unsigned char p = *pattern++;
        if (!*tester) {
            /* tester ends before pattern. Permitted? */
            return islower(p);
        } else {
            /* character must match exactly otherwise */
            if (tolower(p) != tolower((unsigned char) *tester))
                return false;
        }
        ++tester;
    }
    /* pattern ends -- tester must end, too */
    return *tester == 0;
}

/** Tokenize a string. Returns initial subsequence of line consisting
    entirely of (a) alphanumeric characters, (b) numeric characters,
    (c) one non-whitespace character. */
string_t
stringTokenize(string_t& line)
{
    string_t::size_type ptr = 0, endptr, remptr;
    while (ptr < line.length() && isspace(line[ptr]))
        ++ptr;
    if (ptr >= line.length()) {
        /* Line completely read */
        line.erase();
        return string_t();
    }

    endptr = ptr;
    if (isalpha(line[ptr])) {
        /* sequence of alphanumerics */
        while (endptr < line.length() && isalnum(line[endptr]))
            ++endptr;
    } else if (isdigit(line[ptr])) {
        /* sequence of digits */
        while (endptr < line.length() && isdigit(line[endptr]))
            ++endptr;
    } else {
        /* single special character */
        ++endptr;
    }
    remptr = endptr;
    while (remptr < line.length() && isspace(line[remptr]))
        ++remptr;

    string_t rv (line, ptr, endptr - ptr);
    line.erase(0, remptr);
    return rv;
}

bool
stringValue(const string_t& s, long& val)
{
    string_t::size_type pos;
    return stringValue(s, val, pos);
}

/** Parse string into integer.
    \param s     (in) Input
    \param val   (out) Value, only modified if s is completely valid
    \param pos   (out) Number of valid characters, if string is partially
                 valid
    \returns true if string is completely valid; false if not. If the
    string is not completely valid, pos is set to the number of valid
    characters */
bool
stringValue(const string_t& s, long& val, string_t::size_type& pos)
{
    const char* c = s.c_str();
    char* end;
    
    errno = 0;
    long v = std::strtol(c, &end, 10);
    if (errno == ERANGE) {
        pos = 0;
        return false;
    }
    if (!*c || *end) {
        pos = end - c;
        return false;
    }
    pos = 0;
    val = v;
    return true;
}

bool
stringValue(const string_t& s, int& val)
{
    string_t::size_type pos;
    return stringValue(s, val, pos);
}

bool
stringValue(const string_t& s, int& val, string_t::size_type& pos)
{
    long l;
    if (!stringValue(s, l, pos))
        return false;
    if (l < INT_MIN || l > INT_MAX) {
        pos = 0;
        return false;
    }
    val = l;
    return true;
}

/** Parse a flag list.
    \param s       string we got from user
    \param data    allowed options, in caps (e.g., for fcodes, "SPBRAU")
    \param value   pointer to integer variable with result (first option
                   has value 1, second has value 2, third has 4, etc.)
    \param races   pointer to integer variable with race bitmap
                   (1 << race, hence bit 0 is unused). Can be NULL, then
                   we'll not accept races. *races can be preinitialized
                   to a default value (*value can not).
    \param accept_bad_flags if set, accept unknown flags.
    \returns true on success, false on error */
bool
parseFlags(const string_t& s, const char* data, int* value, int* races, bool accept_bad_flags)
{
    static const char eleven[] = "123456789AB";
    string_t::size_type t = 0;
    
    *value = 0;
    while (t < s.length()) {
        unsigned char c = s[t];
        if (races && (c == '+' || c == '-')) {
            bool negate = (c == '-');
            *races = 0;
            while (++t < s.length()) {
                c = s[t];
                const char* p = strchr(eleven, toupper(c));
                if (!p)
                    return false;
                *races |= 2 << (p - eleven);
            }
            if (negate)
                *races = ~*races;
            break;
        } else {
            const char* p = strchr(data, toupper(c));
            if (p)
                *value |= 1 << (p - data);
            else if (!accept_bad_flags && !isspace(c))
                return false;
            ++t;
        }
    }
    return true;
}

string_t
numToString(int32 number)
{
    // FIXME: handle thousand-separator as configured by user
    return itoa(number);
}

string_t
clansToString(int32 number)
{
    // FIXME: handle thousand-separator as configured by user
    return itoa(100*number);
}
