/**
  *  \file util/nls.cc
  *  \brief National Language Support
  *
  *  This module provides message translation in a gettext-like fashion.
  *  We don't use GNU gettext because it's GPL and because its data files
  *  are not portable between architectures. Otherwise, this could be
  *  easily be changed to use gettext when available.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <algorithm>
#include "arch/platform.h"
#include "cpluslib/buffer.h"
#include "cpluslib/types.h"
#include "fileio/dirs.h"
#include "fileio/io.h"
#include "fileio/streams.h"
#include "util/bsearch.h"
#include "util/console.h"
#include "util/except.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "util/recode.h"

/* File format is:
   +0  8 BYTEs   magic: "CClang0",26
   +8    DWORD   number of entries
  +12    DWORD   pointer to "input pointer" array
  +16    DWORD   pointer to "output pointer" array
  +20    DWORD   pointer to input data
  +24    DWORD   length of input data
  +28    DWORD   pointer to output data
  +32    DWORD   length of output data */

namespace {
    struct LanguageItem {
        uint32 in;
        uint32 out;
    };
    Buffer<char>             language_db;
    Buffer<LanguageItem>     language_idx;
    CharacterSet             character_set;
    uint32                   recode_start, recode_len;
    
    struct LangItemCompare {
        bool operator()(const LanguageItem& a, const LanguageItem& b) {
            using namespace std;
            return strcmp(&language_db[a.in], &language_db[b.in]) < 0;
        }
    };
    struct LangItemFind {
        const char* what;
        LangItemFind(const char* p) : what(p) { }
        bool isLess(const LanguageItem& b) const {
            using namespace std;
            return strcmp(&language_db[b.in], what) > 0;
        }
        bool isEqual(const LanguageItem& b) const {
            using namespace std;
            return strcmp(&language_db[b.in], what) == 0;
        }
    };
};

const char*
translate(const char* in)
{
    if (language_idx.getSize()) {
        ptrdiff_t i = bin_search(language_idx.ptr(), language_idx.getSize(),
                                 LangItemFind(in));
        if (i >= 0) {
            /* Character set has changed. Recode whole text database. 
               FIXME: When we change the internal character set too
               often, this might lose information. */
            if (character_set != Recoder::getInternCS()) {
                Recoder re(character_set, Recoder::getInternCS());
                re.recodeArray(&language_db[recode_start], recode_len);
                character_set = Recoder::getInternCS();
            }
            return &language_db[language_idx[i].out];
        }
    }
    return in;
}

void
loadLanguageDatabase(Stream* s)
{
    using namespace std;

    enum {
        i_Count,
        i_InPtr,
        i_OutPtr,
        i_InText,
        i_InSize,
        i_OutText,
        i_OutSize
    };    

    language_db.clear();
    language_idx.clear();
    s->readBuffer(language_db);
    if (language_db.getSize() < 36 || memcmp(language_db.ptr(), "CClang0\032", 8) != 0) {
        language_db.clear();
        throw FileFormatException(format("File `%s' is not a language file", s->getName()));
    }

    uint32 data[7]; // 0-count 1-inptr 2-outptr 3-intxt 4-insize 5-outtxt 6-outsize
    getUint32(&language_db[8], data, 7);
    if (   data[i_InPtr]   >= language_db.getSize()
        || data[i_OutPtr]  >= language_db.getSize()
        || data[i_InText]  >= language_db.getSize()
        || data[i_OutText] >= language_db.getSize()
        || 8*data[i_Count] > language_db.getSize() - data[i_InPtr]
        || 8*data[i_Count] > language_db.getSize() - data[i_OutPtr]
        || data[i_InSize]  > language_db.getSize() - data[i_InText]
        || data[i_OutSize] > language_db.getSize() - data[i_OutText])
    {
        language_db.clear();
        throw FileFormatException(format("File `%s' has an invalid format", s->getName()));
    }

    /* File is valid */
    language_idx.ensureCapacity(data[i_Count]);
    for (uint32 i = 0; i < data[i_Count]; ++i) {
        LanguageItem n;
        n.in           = getUint32(language_db, data[i_InPtr] + 8*i);
        uint32 in_len  = getUint32(language_db, data[i_InPtr] + 8*i + 4);
        n.out          = getUint32(language_db, data[i_OutPtr] + 8*i);
        uint32 out_len = getUint32(language_db, data[i_OutPtr] + 8*i + 4);
        if (n.in > data[i_InSize] || n.out > data[i_OutSize]
            || in_len > data[i_InSize] - n.in || out_len > data[i_OutSize] - n.in)
        {
            /* FIXME: check zero-terminatedness */
            language_db.clear();
            language_idx.clear();
            throw FileFormatException(format("File `%s' has an invalid format", s->getName()));
        }
        n.in += data[i_InText];
        n.out += data[i_OutText];
        language_idx.append(n);
    }

    /* loaded ok, sort it */
    std::sort(language_idx.begin(), language_idx.end(), LangItemCompare());

    /* character set hacking */
    string_t n = translate("");
    character_set = 0;
    do {
        string_t line = strFirst(n, "\n");
        if (strCaseCmp(line.substr(0, 14), "X-PCC-Charset:") == 0) {
            character_set = Recoder::getCharsetByName(strTrim(line.substr(14)));
            break;
        }
    } while (strRemove(n, "\n"));
    
    recode_start = data[i_OutText];
    recode_len   = data[i_OutSize];

    /* Actual recoding is done in translate. */
}

void
loadLanguageDatabase(const char* fn)
{
    string_t res     = Platform::getResourceDirectory();
    string_t full_fn = Directory::makeFileName(res, fn);
    Ptr<Stream> s = Stream::openFile(full_fn, Stream::C_READ);
    loadLanguageDatabase(s);   
}

void
initNLS()
{
    try {
        string_t code = Platform::getLanguageCode();
        string_t::size_type n = code.find_first_of("_:-/.\\");
        if (n != code.npos)
            code.erase(n);
        if (code.length()) {
            code += ".lang";
            loadLanguageDatabase(code.c_str());
        }
    }
    catch(FileOpenException& e) {
        /* ignore */
    }
    catch(std::exception& e) {
        console.write(LOG_ERROR, string_t("Unable to initialize language support: ") + e.what());
    }
}
