/**
  *  \file misc/random.cc
  *  \brief Random Numbers
  *
  *  Instead of relying on the vague semantics of std::rand(), we use
  *  our own, simple random number generator. This one is expected to
  *  have a period of 4G (in other words, enough for us).
  *
  *  Note that this is called `myRandom()' because there is already
  *  a `random()' function in BSD.
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <ctime>
#include "util/random.h"

static uint32 seed;

namespace {
    struct Init {
        Init();
    };
    Init init;
};

Init::Init()
{
    seed = std::time(0);
}

/** Advance seed. Simple linear congruential sequence. */
static inline void
next()
{
    seed = 134775813UL * seed + 1;
}

/** Get random number in range [0, 2^16). */
uint16
myRandom()
{
    next();
    return seed >> 16;
}

/** Get random number in range [0, max). */
uint16
myRandom(uint16 max)
{
    next();

    /* seed * max >> 32        
        = (loword(seed) * max + hiword(seed) * max << 16) >> 32
          hence loword(result) = loword(seed)*max which is discarded anyways
        = (loword(seed) * max >> 16 + hiword(seed) * max) >> 16 */
    return (((seed & 0xFFFFU) * max >> 16) + (seed >> 16) * max) >> 16;
}
