/**
  *  \file util/recode.cc
  *  \brief Character-set Recoding
  *
  *  Because Planets is being played on a number of platforms, we must
  *  occasionally recode some strings.
  *
  *  The basic model is that we have the following character sets:
  *  - a game character set. This one is used for communication with
  *    the host and the other players. All players must agree on this
  *    character set.
  *  - an internal character set. This one is used within PCC II.
  *    It is the character set our fonts are coded in, usually Latin-1.
  *  - a character set for the text string databases. This is handled
  *    in util/nls.cc.
  *
  *  This module provides the mechanism of recoding.
  *
  *  We assume that all character sets have the standard ASCII set at
  *  0..127. The source code uses only ASCII as well. We use some
  *  additional glyphs in the low control character range 16..31, so
  *  the control characters we use must be below (\r \n \t). I guess
  *  all C++ implementations this code runs on fulfill this restriction
  *  (otherwise they'd have lotsa trouble reading Planets files :-)
  *
  *  (c) 2001,2002,2003 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "util/recode.h"
#include "cpluslib/misc.h"
#include "cpluslib/string.h"
#include "cpluslib/assert.h"
#include "cpluslib/types.h"

Recoder Recoder::intern_to_extern;
Recoder Recoder::extern_to_intern;
CharacterSet Recoder::intern_cs(0);
CharacterSet Recoder::extern_cs(0);

/*********************** Character Set Definitions ***********************/

struct CharacterSet::Data {
    const char*  name;
    const char*  full_name;
    uint16       chars[128];
};

/** Character set definitions.
    Character set definitions were generated using a small Perl
    script from the mapping tables on unicode.org.

    We only store the "high" half of each character set here, the
    lower half is assumed to be plain 7-bit ASCII. For each of the
    high 128 characters, we store the unicode value.

    An additional mapping from Unicode to RFC 1345's short mnemonics
    makes it simple to generate replacements for nonexact matches, and
    probably other interesting things in the future. */
static const CharacterSet::Data charsets[] = {
    { "cp1250",
      "Windows Codepage 1250 (Eastern Europe)",
      {
          0x20ac, 0x0000, 0x201a, 0x0000, 0x201e, 0x2026, 0x2020, 0x2021,
          0x0000, 0x2030, 0x0160, 0x2039, 0x015a, 0x0164, 0x017d, 0x0179,
          0x0000, 0x2018, 0x2019, 0x201c, 0x201d, 0x2219, 0x2013, 0x2014,
          0x0000, 0x2122, 0x0161, 0x203a, 0x015b, 0x0165, 0x017e, 0x017a,
          0x00a0, 0x02c7, 0x02d8, 0x0141, 0x00a4, 0x0104, 0x00a6, 0x00a7,
          0x00a8, 0x00a9, 0x015e, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x017b,
          0x00b0, 0x00b1, 0x02db, 0x0142, 0x00b4, 0x03bc, 0x00b6, 0x00b7,
          0x00b8, 0x0105, 0x015f, 0x00bb, 0x013d, 0x02dd, 0x013e, 0x017c,
          0x0154, 0x00c1, 0x00c2, 0x0102, 0x00c4, 0x0139, 0x0106, 0x00c7,
          0x010c, 0x00c9, 0x0118, 0x00cb, 0x011a, 0x00cd, 0x00ce, 0x010e,
          0x0110, 0x0143, 0x0147, 0x00d3, 0x00d4, 0x0150, 0x00d6, 0x00d7,
          0x0158, 0x016e, 0x00da, 0x0170, 0x00dc, 0x00dd, 0x0162, 0x00df,
          0x0155, 0x00e1, 0x00e2, 0x0103, 0x00e4, 0x013a, 0x0107, 0x00e7,
          0x010d, 0x00e9, 0x0119, 0x00eb, 0x011b, 0x00ed, 0x00ee, 0x010f,
          0x0111, 0x0144, 0x0148, 0x00f3, 0x00f4, 0x0151, 0x00f6, 0x00f7,
          0x0159, 0x016f, 0x00fa, 0x0171, 0x00fc, 0x00fd, 0x0163, 0x02d9,
      }
    },

    { "cp1251",
      "Windows Codepage 1251 (Cyrillic)",
      {
          0x0402, 0x0403, 0x201a, 0x0453, 0x201e, 0x2026, 0x2020, 0x2021,
          0x20ac, 0x2030, 0x0409, 0x2039, 0x040a, 0x040c, 0x040b, 0x040f,
          0x0452, 0x2018, 0x2019, 0x201c, 0x201d, 0x2219, 0x2013, 0x2014,
          0x0000, 0x2122, 0x0459, 0x203a, 0x045a, 0x045c, 0x045b, 0x045f,
          0x00a0, 0x040e, 0x045e, 0x0408, 0x00a4, 0x0490, 0x00a6, 0x00a7,
          0x0401, 0x00a9, 0x0404, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x0407,
          0x00b0, 0x00b1, 0x0406, 0x0456, 0x0491, 0x03bc, 0x00b6, 0x00b7,
          0x0451, 0x2116, 0x0454, 0x00bb, 0x0458, 0x0405, 0x0455, 0x0457,
          0x0410, 0x0411, 0x0412, 0x0413, 0x0414, 0x0415, 0x0416, 0x0417,
          0x0418, 0x0419, 0x041a, 0x041b, 0x041c, 0x041d, 0x041e, 0x041f,
          0x0420, 0x0421, 0x0422, 0x0423, 0x0424, 0x0425, 0x0426, 0x0427,
          0x0428, 0x0429, 0x042a, 0x042b, 0x042c, 0x042d, 0x042e, 0x042f,
          0x0430, 0x0431, 0x0432, 0x0433, 0x0434, 0x0435, 0x0436, 0x0437,
          0x0438, 0x0439, 0x043a, 0x043b, 0x043c, 0x043d, 0x043e, 0x043f,
          0x0440, 0x0441, 0x0442, 0x0443, 0x0444, 0x0445, 0x0446, 0x0447,
          0x0448, 0x0449, 0x044a, 0x044b, 0x044c, 0x044d, 0x044e, 0x044f,
      }
    },

    { "cp1252",
      "Windows Codepage 1252 (Western Europe, Latin-1)",
      {
          0x20ac, 0x0000, 0x201a, 0x0192, 0x201e, 0x2026, 0x2020, 0x2021,
          0x005e, 0x2030, 0x0160, 0x2039, 0x0152, 0x0000, 0x017d, 0x0000,
          0x0000, 0x2018, 0x2019, 0x201c, 0x201d, 0x2219, 0x2013, 0x2014,
          0x007e, 0x2122, 0x0161, 0x203a, 0x0153, 0x0000, 0x017e, 0x0178,
          0x00a0, 0x00a1, 0x00a2, 0x00a3, 0x00a4, 0x00a5, 0x00a6, 0x00a7,
          0x00a8, 0x00a9, 0x00aa, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x00af,
          0x00b0, 0x00b1, 0x00b2, 0x00b3, 0x00b4, 0x03bc, 0x00b6, 0x00b7,
          0x00b8, 0x00b9, 0x00ba, 0x00bb, 0x00bc, 0x00bd, 0x00be, 0x00bf,
          0x00c0, 0x00c1, 0x00c2, 0x00c3, 0x00c4, 0x00c5, 0x00c6, 0x00c7,
          0x00c8, 0x00c9, 0x00ca, 0x00cb, 0x00cc, 0x00cd, 0x00ce, 0x00cf,
          0x00d0, 0x00d1, 0x00d2, 0x00d3, 0x00d4, 0x00d5, 0x00d6, 0x00d7,
          0x00d8, 0x00d9, 0x00da, 0x00db, 0x00dc, 0x00dd, 0x00de, 0x00df,
          0x00e0, 0x00e1, 0x00e2, 0x00e3, 0x00e4, 0x00e5, 0x00e6, 0x00e7,
          0x00e8, 0x00e9, 0x00ea, 0x00eb, 0x00ec, 0x00ed, 0x00ee, 0x00ef,
          0x00f0, 0x00f1, 0x00f2, 0x00f3, 0x00f4, 0x00f5, 0x00f6, 0x00f7,
          0x00f8, 0x00f9, 0x00fa, 0x00fb, 0x00fc, 0x00fd, 0x00fe, 0x00ff,
      }
    },

    /* The infamous DOS character set. Most DOS planeteers use that one. */
    { "cp437",
      "MS-DOS Codepage 437",
      {
          0x00c7, 0x00fc, 0x00e9, 0x00e2, 0x00e4, 0x00e0, 0x00e5, 0x00e7,
          0x00ea, 0x00eb, 0x00e8, 0x00ef, 0x00ee, 0x00ec, 0x00c4, 0x00c5,
          0x00c9, 0x00e6, 0x00c6, 0x00f4, 0x00f6, 0x00f2, 0x00fb, 0x00f9,
          0x00ff, 0x00d6, 0x00dc, 0x00a2, 0x00a3, 0x00a5, 0x20a7, 0x0192,
          0x00e1, 0x00ed, 0x00f3, 0x00fa, 0x00f1, 0x00d1, 0x00aa, 0x00ba,
          0x00bf, 0x2310, 0x00ac, 0x00bd, 0x00bc, 0x00a1, 0x00ab, 0x00bb,
          0x2591, 0x2592, 0x2593, 0x2502, 0x2524, 0x2525, 0x2528, 0x2512,
          0x2511, 0x252b, 0x2503, 0x2513, 0x251b, 0x251a, 0x2519, 0x2510,
          0x2514, 0x2534, 0x252c, 0x251c, 0x2500, 0x253c, 0x251d, 0x2520,
          0x2517, 0x250f, 0x253b, 0x2533, 0x2523, 0x2501, 0x254b, 0x2537,
          0x2538, 0x252f, 0x2530, 0x2516, 0x2515, 0x250d, 0x250e, 0x2542,
          0x253f, 0x2518, 0x250c, 0x2588, 0x2584, 0x258c, 0x2590, 0x2580,
          0x03b1, 0x00df, 0x0393, 0x03c0, 0x03a3, 0x03c3, 0x03bc, 0x03c4,
          0x03a6, 0x0398, 0x03a9, 0x03b4, 0x221e, 0x03c6, 0x03b5, 0x2229,
          0x2261, 0x00b1, 0x2265, 0x2264, 0x2320, 0x2321, 0x00f7, 0x2248,
          0x00b0, 0x2219, 0x00b7, 0x221a, 0x207f, 0x00b2, 0x25a0, 0x00a0,
      }
    },

    /* Another DOS codepage. For some obscure reasons, this one is
       the recommended character set for most of Europe. Probably
       because it has more accented characters. But it lacks some of
       the interesting pseudographics ones. */
    { "cp850",
      "MS-DOS Codepage 850",
      {
          0x00c7, 0x00fc, 0x00e9, 0x00e2, 0x00e4, 0x00e0, 0x00e5, 0x00e7,
          0x00ea, 0x00eb, 0x00e8, 0x00ef, 0x00ee, 0x00ec, 0x00c4, 0x00c5,
          0x00c9, 0x00e6, 0x00c6, 0x00f4, 0x00f6, 0x00f2, 0x00fb, 0x00f9,
          0x00ff, 0x00d6, 0x00dc, 0x00f8, 0x00a3, 0x00d8, 0x00d7, 0x0192,
          0x00e1, 0x00ed, 0x00f3, 0x00fa, 0x00f1, 0x00d1, 0x00aa, 0x00ba,
          0x00bf, 0x00ae, 0x00ac, 0x00bd, 0x00bc, 0x00a1, 0x00ab, 0x00bb,
          0x2591, 0x2592, 0x2593, 0x2502, 0x2524, 0x00c1, 0x00c2, 0x00c0,
          0x00a9, 0x252b, 0x2503, 0x2513, 0x251b, 0x00a2, 0x00a5, 0x2510,
          0x2514, 0x2534, 0x252c, 0x251c, 0x2500, 0x253c, 0x00e3, 0x00c3,
          0x2517, 0x250f, 0x253b, 0x2533, 0x2523, 0x2501, 0x254b, 0x00a4,
          0x00f0, 0x00d0, 0x00ca, 0x00cb, 0x00c8, 0x0131, 0x00cd, 0x00ce,
          0x00cf, 0x2518, 0x250c, 0x2588, 0x2584, 0x00a6, 0x00cc, 0x2580,
          0x00d3, 0x00df, 0x00d4, 0x00d2, 0x00f5, 0x00d5, 0x03bc, 0x00fe,
          0x00de, 0x00da, 0x00db, 0x00d9, 0x00fd, 0x00dd, 0x00af, 0x00b4,
          0x00ad, 0x00b1, 0x2017, 0x00be, 0x00b6, 0x00a7, 0x00f7, 0x00b8,
          0x00b0, 0x00a8, 0x00b7, 0x00b9, 0x00b3, 0x00b2, 0x25a0, 0x00a0,
      }
    },

    /* This one is used in eastern Europe. */
    { "cp852",
      "MS-DOS Codepage 852",
      {
          0x00c7, 0x00fc, 0x00e9, 0x00e2, 0x00e4, 0x016f, 0x0107, 0x00e7,
          0x0142, 0x00eb, 0x0150, 0x0151, 0x00ee, 0x0179, 0x00c4, 0x0106,
          0x00c9, 0x0139, 0x013a, 0x00f4, 0x00f6, 0x013d, 0x013e, 0x015a,
          0x015b, 0x00d6, 0x00dc, 0x0164, 0x0165, 0x0141, 0x00d7, 0x010d,
          0x00e1, 0x00ed, 0x00f3, 0x00fa, 0x0104, 0x0105, 0x017d, 0x017e,
          0x0118, 0x0119, 0x00ac, 0x017a, 0x010c, 0x015f, 0x00ab, 0x00bb,
          0x2591, 0x2592, 0x2593, 0x2502, 0x2524, 0x00c1, 0x00c2, 0x011a,
          0x015e, 0x252b, 0x2503, 0x2513, 0x251b, 0x017b, 0x017c, 0x2510,
          0x2514, 0x2534, 0x252c, 0x251c, 0x2500, 0x253c, 0x0102, 0x0103,
          0x2517, 0x250f, 0x253b, 0x2533, 0x2523, 0x2501, 0x254b, 0x00a4,
          0x0111, 0x0110, 0x010e, 0x00cb, 0x010f, 0x0147, 0x00cd, 0x00ce,
          0x011b, 0x2518, 0x250c, 0x2588, 0x2584, 0x0162, 0x016e, 0x2580,
          0x00d3, 0x00df, 0x00d4, 0x0143, 0x0144, 0x0148, 0x0160, 0x0161,
          0x0154, 0x00da, 0x0155, 0x0170, 0x00fd, 0x00dd, 0x0163, 0x00b4,
          0x00ad, 0x02dd, 0x02db, 0x02c7, 0x02d8, 0x00a7, 0x00f7, 0x00b8,
          0x00b0, 0x00a8, 0x02d9, 0x0171, 0x0158, 0x0159, 0x25a0, 0x00a0,
      }
    },

    /* It seems most DOS planeteers in Russia & Co use this one. For
       example, VPA ships with such a character set; the PHost
       messages are in this character set, too. */
    { "cp866",
      "MS-DOS Codepage 866 (\"Alternative\" cyrillic)",
      {
          0x0410, 0x0411, 0x0412, 0x0413, 0x0414, 0x0415, 0x0416, 0x0417,
          0x0418, 0x0419, 0x041a, 0x041b, 0x041c, 0x041d, 0x041e, 0x041f,
          0x0420, 0x0421, 0x0422, 0x0423, 0x0424, 0x0425, 0x0426, 0x0427,
          0x0428, 0x0429, 0x042a, 0x042b, 0x042c, 0x042d, 0x042e, 0x042f,
          0x0430, 0x0431, 0x0432, 0x0433, 0x0434, 0x0435, 0x0436, 0x0437,
          0x0438, 0x0439, 0x043a, 0x043b, 0x043c, 0x043d, 0x043e, 0x043f,
          0x2591, 0x2592, 0x2593, 0x2502, 0x2524, 0x2525, 0x2528, 0x2512,
          0x2511, 0x252b, 0x2503, 0x2513, 0x251b, 0x251a, 0x2519, 0x2510,
          0x2514, 0x2534, 0x252c, 0x251c, 0x2500, 0x253c, 0x251d, 0x2520,
          0x2517, 0x250f, 0x253b, 0x2533, 0x2523, 0x2501, 0x254b, 0x2537,
          0x2538, 0x252f, 0x2530, 0x2516, 0x2515, 0x250d, 0x250e, 0x2542,
          0x253f, 0x2518, 0x250c, 0x2588, 0x2584, 0x258c, 0x2590, 0x2580,
          0x0440, 0x0441, 0x0442, 0x0443, 0x0444, 0x0445, 0x0446, 0x0447,
          0x0448, 0x0449, 0x044a, 0x044b, 0x044c, 0x044d, 0x044e, 0x044f,
          0x0401, 0x0451, 0x0404, 0x0454, 0x0407, 0x0457, 0x040e, 0x045e,
          0x00b0, 0x2219, 0x00b7, 0x221a, 0x2116, 0x00a4, 0x25a0, 0x00a0,
      }
    },

    /* This one's used in Cyrillic Unixen and on the net. */
    { "koi8-r",
      "Cyrillic (KOI8-R)",
      {
          0x2500, 0x2502, 0x250c, 0x2510, 0x2514, 0x2518, 0x251c, 0x2524,
          0x252c, 0x2534, 0x253c, 0x2580, 0x2584, 0x2588, 0x258c, 0x2590,
          0x2591, 0x2592, 0x2593, 0x2320, 0x25a0, 0x2219, 0x221a, 0x2248,
          0x2264, 0x2265, 0x00a0, 0x2321, 0x00b0, 0x00b2, 0x00b7, 0x00f7,
          0x2501, 0x2503, 0x250d, 0x0451, 0x250e, 0x250f, 0x2511, 0x2512,
          0x2513, 0x2515, 0x2516, 0x2517, 0x2519, 0x251a, 0x251b, 0x251d,
          0x2520, 0x2523, 0x2525, 0x0401, 0x2528, 0x252b, 0x252f, 0x2530,
          0x2533, 0x2537, 0x2538, 0x253b, 0x253f, 0x2542, 0x254b, 0x00a9,
          0x044e, 0x0430, 0x0431, 0x0446, 0x0434, 0x0435, 0x0444, 0x0433,
          0x0445, 0x0438, 0x0439, 0x043a, 0x043b, 0x043c, 0x043d, 0x043e,
          0x043f, 0x044f, 0x0440, 0x0441, 0x0442, 0x0443, 0x0436, 0x0432,
          0x044c, 0x044b, 0x0437, 0x0448, 0x044d, 0x0449, 0x0447, 0x044a,
          0x042e, 0x0410, 0x0411, 0x0426, 0x0414, 0x0415, 0x0424, 0x0413,
          0x0425, 0x0418, 0x0419, 0x041a, 0x041b, 0x041c, 0x041d, 0x041e,
          0x041f, 0x042f, 0x0420, 0x0421, 0x0422, 0x0423, 0x0416, 0x0412,
          0x042c, 0x042b, 0x0417, 0x0428, 0x042d, 0x0429, 0x0427, 0x042a,
      }
    },

    /* Most Windows/Unix planeteers use (an approximation of) this one. */
    { "latin1",
      "ISO 8859-1 (Latin-1, Windows/Unix)",
      {
          0x0080, 0x0081, 0x0082, 0x0083, 0x0084, 0x0085, 0x0086, 0x0087,
          0x0088, 0x0089, 0x008a, 0x008b, 0x008c, 0x008d, 0x008e, 0x008f,
          0x0090, 0x0091, 0x0092, 0x0093, 0x0094, 0x0095, 0x0096, 0x0097,
          0x0098, 0x0099, 0x009a, 0x009b, 0x009c, 0x009d, 0x009e, 0x009f,
          0x00a0, 0x00a1, 0x00a2, 0x00a3, 0x00a4, 0x00a5, 0x00a6, 0x00a7,
          0x00a8, 0x00a9, 0x00aa, 0x00ab, 0x00ac, 0x00ad, 0x00ae, 0x00af,
          0x00b0, 0x00b1, 0x00b2, 0x00b3, 0x00b4, 0x03bc, 0x00b6, 0x00b7,
          0x00b8, 0x00b9, 0x00ba, 0x00bb, 0x00bc, 0x00bd, 0x00be, 0x00bf,
          0x00c0, 0x00c1, 0x00c2, 0x00c3, 0x00c4, 0x00c5, 0x00c6, 0x00c7,
          0x00c8, 0x00c9, 0x00ca, 0x00cb, 0x00cc, 0x00cd, 0x00ce, 0x00cf,
          0x00d0, 0x00d1, 0x00d2, 0x00d3, 0x00d4, 0x00d5, 0x00d6, 0x00d7,
          0x00d8, 0x00d9, 0x00da, 0x00db, 0x00dc, 0x00dd, 0x00de, 0x00df,
          0x00e0, 0x00e1, 0x00e2, 0x00e3, 0x00e4, 0x00e5, 0x00e6, 0x00e7,
          0x00e8, 0x00e9, 0x00ea, 0x00eb, 0x00ec, 0x00ed, 0x00ee, 0x00ef,
          0x00f0, 0x00f1, 0x00f2, 0x00f3, 0x00f4, 0x00f5, 0x00f6, 0x00f7,
          0x00f8, 0x00f9, 0x00fa, 0x00fb, 0x00fc, 0x00fd, 0x00fe, 0x00ff,
      }
    },

    /* This one is used in eastern Europe, too. */
    { "latin2",
      "ISO 8859-2 (Latin-2)",
      {
          0x0080, 0x0081, 0x0082, 0x0083, 0x0084, 0x0085, 0x0086, 0x0087,
          0x0088, 0x0089, 0x008a, 0x008b, 0x008c, 0x008d, 0x008e, 0x008f,
          0x0090, 0x0091, 0x0092, 0x0093, 0x0094, 0x0095, 0x0096, 0x0097,
          0x0098, 0x0099, 0x009a, 0x009b, 0x009c, 0x009d, 0x009e, 0x009f,
          0x00a0, 0x0104, 0x02d8, 0x0141, 0x00a4, 0x013d, 0x015a, 0x00a7,
          0x00a8, 0x0160, 0x015e, 0x0164, 0x0179, 0x00ad, 0x017d, 0x017b,
          0x00b0, 0x0105, 0x02db, 0x0142, 0x00b4, 0x013e, 0x015b, 0x02c7,
          0x00b8, 0x0161, 0x015f, 0x0165, 0x017a, 0x02dd, 0x017e, 0x017c,
          0x0154, 0x00c1, 0x00c2, 0x0102, 0x00c4, 0x0139, 0x0106, 0x00c7,
          0x010c, 0x00c9, 0x0118, 0x00cb, 0x011a, 0x00cd, 0x00ce, 0x010e,
          0x0110, 0x0143, 0x0147, 0x00d3, 0x00d4, 0x0150, 0x00d6, 0x00d7,
          0x0158, 0x016e, 0x00da, 0x0170, 0x00dc, 0x00dd, 0x0162, 0x00df,
          0x0155, 0x00e1, 0x00e2, 0x0103, 0x00e4, 0x013a, 0x0107, 0x00e7,
          0x010d, 0x00e9, 0x0119, 0x00eb, 0x011b, 0x00ed, 0x00ee, 0x010f,
          0x0111, 0x0144, 0x0148, 0x00f3, 0x00f4, 0x0151, 0x00f6, 0x00f7,
          0x0159, 0x016f, 0x00fa, 0x0171, 0x00fc, 0x00fd, 0x0163, 0x02d9,
      }
    },

    /* The character set used by PCC 1.x. Almost the same as
       cp437, but adds some funny symbols. In addition, some
       character sets replace "|" with a small "/".
       We do not translate 155 = logo, 156 = (-\-) ornament,
       157 = (-/-) ornament */
    { "pcc1",
      "PCC 1.x (almost Codepage 437)",
      {
          0x00c7, 0x00fc, 0x00e9, 0x00e2, 0x00e4, 0x00e0, 0x00e5, 0x00e7,
          0x00ea, 0x00eb, 0x00e8, 0x00ef, 0x00ee, 0x00ec, 0x00c4, 0x00c5,
          0x00c9, 0x00e6, 0x00c6, 0x00f4, 0x00f6, 0x00f2, 0x00fb, 0x00f9,
          0x00ff, 0x00d6, 0x00dc, 0x00a2, 0x00a3, 0x00a5, 0x2258, 0x00d7,
          0x00e1, 0x00ed, 0x00f3, 0x00fa, 0x00f1, 0x00d1, 0x00aa, 0x00ba,
          0x00bf, 0x00ae, 0x2122, 0x00bd, 0x00bc, 0x00a1, 0x00ab, 0x00bb,
          0x2591, 0x2592, 0x2593, 0x2502, 0x2524, 0x2525, 0x2528, 0x2512,
          0x2511, 0x252b, 0x2503, 0x2513, 0x251b, 0x251a, 0x2519, 0x2510,
          0x2514, 0x2534, 0x252c, 0x251c, 0x2500, 0x253c, 0x251d, 0x2520,
          0x2517, 0x250f, 0x253b, 0x2533, 0x2523, 0x2501, 0x254b, 0x2537,
          0x2538, 0x252f, 0x2530, 0x2516, 0x2515, 0x250d, 0x250e, 0x2542,
          0x253f, 0x2518, 0x250c, 0x2588, 0x2584, 0x25c0, 0x25b6, 0x2580,
          0x03b1, 0x00df, 0x0393, 0x03c0, 0x03a3, 0x03c3, 0x03bc, 0x03c4,
          0x03a6, 0x0398, 0x03a9, 0x03b4, 0x221e, 0x03c6, 0x03b5, 0x2229,
          0x2261, 0x00b1, 0x2265, 0x2264, 0x2320, 0x2321, 0x00f7, 0x2248,
          0x00b0, 0x2219, 0x00b7, 0x221a, 0x207f, 0x00b2, 0x25a0, 0x00a0,
      }
    },
};

namespace {
    struct Uni2RFC {
        uint16 code;
        char   rfc[4];
    };
    const Uni2RFC uni_to_rfc[] = {
        { 0x005e, "'>" },
        { 0x007e, "'?" },
        { 0x0080, "PA" },
        { 0x0081, "HO" },
        { 0x0082, "BH" },
        { 0x0083, "NH" },
        { 0x0084, "IN" },
        { 0x0085, "NL" },
        { 0x0086, "SA" },
        { 0x0087, "ES" },
        { 0x0088, "HS" },
        { 0x0089, "HJ" },
        { 0x008a, "VS" },
        { 0x008b, "PD" },
        { 0x008c, "PU" },
        { 0x008d, "RI" },
        { 0x008e, "S2" },
        { 0x008f, "S3" },
        { 0x0090, "DC" },
        { 0x0091, "P1" },
        { 0x0092, "P2" },
        { 0x0093, "TS" },
        { 0x0094, "CC" },
        { 0x0095, "MW" },
        { 0x0096, "SG" },
        { 0x0097, "EG" },
        { 0x0098, "SS" },
        { 0x0099, "GC" },
        { 0x009a, "SC" },
        { 0x009b, "CI" },
        { 0x009c, "ST" },
        { 0x009d, "OC" },
        { 0x009e, "PM" },
        { 0x009f, "AC" },
        { 0x00a0, "NS" },
        { 0x00a1, "!I" },
        { 0x00a2, "Ct" },
        { 0x00a3, "Pd" },
        { 0x00a4, "Cu" },
        { 0x00a5, "Ye" },
        { 0x00a6, "BB" },
        { 0x00a7, "SE" },
        { 0x00a8, "':" },
        { 0x00a9, "Co" },
        { 0x00aa, "-a" },
        { 0x00ab, "<<" },
        { 0x00ac, "NO" },
        { 0x00ad, "--" },
        { 0x00ae, "Rg" },
        { 0x00af, "'m" },
        { 0x00b0, "DG" },
        { 0x00b1, "+-" },
        { 0x00b2, "2S" },
        { 0x00b3, "3S" },
        { 0x00b4, "''" },
        { 0x00b6, "PI" },
        { 0x00b7, ".M" },
        { 0x00b8, "'," },
        { 0x00b9, "1S" },
        { 0x00ba, "-o" },
        { 0x00bb, ">>" },
        { 0x00bc, "14" },
        { 0x00bd, "12" },
        { 0x00be, "34" },
        { 0x00bf, "?I" },
        { 0x00c0, "A!" },
        { 0x00c1, "A'" },
        { 0x00c2, "A>" },
        { 0x00c3, "A?" },
        { 0x00c4, "A:" },
        { 0x00c5, "AA" },
        { 0x00c6, "AE" },
        { 0x00c7, "C," },
        { 0x00c8, "E!" },
        { 0x00c9, "E'" },
        { 0x00ca, "E>" },
        { 0x00cb, "E:" },
        { 0x00cc, "I!" },
        { 0x00cd, "I'" },
        { 0x00ce, "I>" },
        { 0x00cf, "I:" },
        { 0x00d0, "D-" },
        { 0x00d1, "N?" },
        { 0x00d2, "O!" },
        { 0x00d3, "O'" },
        { 0x00d4, "O>" },
        { 0x00d5, "O?" },
        { 0x00d6, "O:" },
        { 0x00d7, "*X" },
        { 0x00d8, "O/" },
        { 0x00d9, "U!" },
        { 0x00da, "U'" },
        { 0x00db, "U>" },
        { 0x00dc, "U:" },
        { 0x00dd, "Y'" },
        { 0x00de, "TH" },
        { 0x00df, "ss" },
        { 0x00e0, "a!" },
        { 0x00e1, "a'" },
        { 0x00e2, "a>" },
        { 0x00e3, "a?" },
        { 0x00e4, "a:" },
        { 0x00e5, "aa" },
        { 0x00e6, "ae" },
        { 0x00e7, "c," },
        { 0x00e8, "e!" },
        { 0x00e9, "e'" },
        { 0x00ea, "e>" },
        { 0x00eb, "e:" },
        { 0x00ec, "i!" },
        { 0x00ed, "i'" },
        { 0x00ee, "i>" },
        { 0x00ef, "i:" },
        { 0x00f0, "d-" },
        { 0x00f1, "n?" },
        { 0x00f2, "o!" },
        { 0x00f3, "o'" },
        { 0x00f4, "o>" },
        { 0x00f5, "o?" },
        { 0x00f6, "o:" },
        { 0x00f7, "-:" },
        { 0x00f8, "o/" },
        { 0x00f9, "u!" },
        { 0x00fa, "u'" },
        { 0x00fb, "u>" },
        { 0x00fc, "u:" },
        { 0x00fd, "y'" },
        { 0x00fe, "th" },
        { 0x00ff, "y:" },
        { 0x0102, "A(" },
        { 0x0103, "a(" },
        { 0x0104, "A;" },
        { 0x0105, "a;" },
        { 0x0106, "C'" },
        { 0x0107, "c'" },
        { 0x010c, "C<" },
        { 0x010d, "c<" },
        { 0x010e, "D<" },
        { 0x010f, "d<" },
        { 0x0110, "D/" },
        { 0x0111, "d/" },
        { 0x0118, "E;" },
        { 0x0119, "e;" },
        { 0x011a, "E<" },
        { 0x011b, "e<" },
        { 0x0131, "i." },
        { 0x0139, "L'" },
        { 0x013a, "l'" },
        { 0x013d, "L<" },
        { 0x013e, "l<" },
        { 0x0141, "L/" },
        { 0x0142, "l/" },
        { 0x0143, "N'" },
        { 0x0144, "n'" },
        { 0x0147, "N<" },
        { 0x0148, "n<" },
        { 0x0150, "O\"" },
        { 0x0151, "o\"" },
        { 0x0152, "OE" },
        { 0x0153, "oe" },
        { 0x0154, "R'" },
        { 0x0155, "r'" },
        { 0x0158, "R<" },
        { 0x0159, "r<" },
        { 0x015a, "S'" },
        { 0x015b, "s'" },
        { 0x015e, "S," },
        { 0x015f, "s," },
        { 0x0160, "S<" },
        { 0x0161, "s<" },
        { 0x0162, "T," },
        { 0x0163, "t," },
        { 0x0164, "T<" },
        { 0x0165, "t<" },
        { 0x016e, "U0" },
        { 0x016f, "u0" },
        { 0x0170, "U\"" },
        { 0x0171, "u\"" },
        { 0x0178, "Y:" },
        { 0x0179, "Z'" },
        { 0x017a, "z'" },
        { 0x017b, "Z." },
        { 0x017c, "z." },
        { 0x017d, "Z<" },
        { 0x017e, "z<" },
        { 0x0192, "Fl" },
        { 0x02c7, "'<" },
        { 0x02d8, "'(" },
        { 0x02d9, "'." },
        { 0x02db, "';" },
        { 0x02dd, "'\"" },
        { 0x0393, "G*" },
        { 0x0398, "H*" },
        { 0x03a3, "S*" },
        { 0x03a6, "F*" },
        { 0x03a9, "W*" },
        { 0x03b1, "a*" },
        { 0x03b4, "d*" },
        { 0x03b5, "e*" },
        { 0x03bc, "m*" },
        { 0x03c0, "p*" },
        { 0x03c3, "s*" },
        { 0x03c4, "t*" },
        { 0x03c6, "f*" },
        { 0x0401, "IO" },
        { 0x0402, "D%" },
        { 0x0403, "G%" },
        { 0x0404, "IE" },
        { 0x0405, "DS" },
        { 0x0406, "II" },
        { 0x0407, "YI" },
        { 0x0408, "J%" },
        { 0x0409, "LJ" },
        { 0x040a, "NJ" },
        { 0x040b, "Ts" },
        { 0x040c, "KJ" },
        { 0x040e, "V%" },
        { 0x040f, "DZ" },
        { 0x0410, "A=" },
        { 0x0411, "B=" },
        { 0x0412, "V=" },
        { 0x0413, "G=" },
        { 0x0414, "D=" },
        { 0x0415, "E=" },
        { 0x0416, "Z%" },
        { 0x0417, "Z=" },
        { 0x0418, "I=" },
        { 0x0419, "J=" },
        { 0x041a, "K=" },
        { 0x041b, "L=" },
        { 0x041c, "M=" },
        { 0x041d, "N=" },
        { 0x041e, "O=" },
        { 0x041f, "P=" },
        { 0x0420, "R=" },
        { 0x0421, "S=" },
        { 0x0422, "T=" },
        { 0x0423, "U=" },
        { 0x0424, "F=" },
        { 0x0425, "H=" },
        { 0x0426, "C=" },
        { 0x0427, "C%" },
        { 0x0428, "S%" },
        { 0x0429, "Sc" },
        { 0x042a, "=\"" },
        { 0x042b, "Y=" },
        { 0x042c, "%\"" },
        { 0x042d, "JE" },
        { 0x042e, "JU" },
        { 0x042f, "JA" },
        { 0x0430, "a=" },
        { 0x0431, "b=" },
        { 0x0432, "v=" },
        { 0x0433, "g=" },
        { 0x0434, "d=" },
        { 0x0435, "e=" },
        { 0x0436, "z%" },
        { 0x0437, "z=" },
        { 0x0438, "i=" },
        { 0x0439, "j=" },
        { 0x043a, "k=" },
        { 0x043b, "l=" },
        { 0x043c, "m=" },
        { 0x043d, "n=" },
        { 0x043e, "o=" },
        { 0x043f, "p=" },
        { 0x0440, "r=" },
        { 0x0441, "s=" },
        { 0x0442, "t=" },
        { 0x0443, "u=" },
        { 0x0444, "f=" },
        { 0x0445, "h=" },
        { 0x0446, "c=" },
        { 0x0447, "c%" },
        { 0x0448, "s%" },
        { 0x0449, "sc" },
        { 0x044a, "='" },
        { 0x044b, "y=" },
        { 0x044c, "%'" },
        { 0x044d, "je" },
        { 0x044e, "ju" },
        { 0x044f, "ja" },
        { 0x0451, "io" },
        { 0x0452, "d%" },
        { 0x0453, "g%" },
        { 0x0454, "ie" },
        { 0x0455, "ds" },
        { 0x0456, "ii" },
        { 0x0457, "yi" },
        { 0x0458, "j%" },
        { 0x0459, "lj" },
        { 0x045a, "nj" },
        { 0x045b, "ts" },
        { 0x045c, "kj" },
        { 0x045e, "v%" },
        { 0x045f, "dz" },
        { 0x0490, "G3" },
        { 0x0491, "g3" },
        { 0x2013, "-N" },
        { 0x2014, "-M" },
        { 0x2017, "=2" },
        { 0x2018, "'6" },
        { 0x2019, "'9" },
        { 0x201a, ".9" },
        { 0x201c, "\"6" },
        { 0x201d, "\"9" },
        { 0x201e, ":9" },
        { 0x2020, "/-" },
        { 0x2021, "/=" },
        { 0x2026, "..." },
        { 0x2030, "%0" },
        { 0x2039, "<1" },
        { 0x203a, ">1" },
        { 0x207f, "nS" },
        { 0x20a7, "Pt" },
        { 0x20ac, "Eu" },
        { 0x2116, "N0" },
        { 0x2122, "TM" },
        { 0x2219, "Sb" },
        { 0x221a, "RT" },
        { 0x221e, "00" },
        { 0x2229, "(U" },
        { 0x2248, "?2" },
        { 0x2258, "=^" },
        { 0x2261, "=3" },
        { 0x2264, "=<" },
        { 0x2265, ">=" },
        { 0x2310, "NI" },
        { 0x2320, "Iu" },
        { 0x2321, "Il" },
        { 0x2500, "hh" },
        { 0x2501, "HH" },
        { 0x2502, "vv" },
        { 0x2503, "VV" },
        { 0x250c, "dr" },
        { 0x250d, "dR" },
        { 0x250e, "Dr" },
        { 0x250f, "DR" },
        { 0x2510, "dl" },
        { 0x2511, "dL" },
        { 0x2512, "Dl" },
        { 0x2513, "LD" },
        { 0x2514, "ur" },
        { 0x2515, "uR" },
        { 0x2516, "Ur" },
        { 0x2517, "UR" },
        { 0x2518, "ul" },
        { 0x2519, "uL" },
        { 0x251a, "Ul" },
        { 0x251b, "UL" },
        { 0x251c, "vr" },
        { 0x251d, "vR" },
        { 0x2520, "Vr" },
        { 0x2523, "VR" },
        { 0x2524, "vl" },
        { 0x2525, "vL" },
        { 0x2528, "Vl" },
        { 0x252b, "VL" },
        { 0x252c, "dh" },
        { 0x252f, "dH" },
        { 0x2530, "Dh" },
        { 0x2533, "DH" },
        { 0x2534, "uh" },
        { 0x2537, "uH" },
        { 0x2538, "Uh" },
        { 0x253b, "UH" },
        { 0x253c, "vh" },
        { 0x253f, "vH" },
        { 0x2542, "Vh" },
        { 0x254b, "VH" },
        { 0x2580, "TB" },
        { 0x2584, "LB" },
        { 0x2588, "FB" },
        { 0x258c, "lB" },
        { 0x2590, "RB" },
        { 0x2591, ".S" },
        { 0x2592, ":S" },
        { 0x2593, "?S" },
        { 0x25a0, "fS" },
        { 0x25b6, "PR" },
        { 0x25c0, "PL" },
        { 0xe023, "Fl" },
    };
}

/******************************** Recoder ********************************/

/** Initialize with transformation. */
Recoder::Recoder(CharacterSet from, CharacterSet to)
{
    setTransformation(from, to);
}

/** Initialize with identity transformation. */
Recoder::Recoder()
{
    setIdentityTransformation();
}

static bool
isLetter(char c)
{
    /* FIXME: this assumes ASCII, but so does the rest of this module :) */
    return (c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z');
}

/** Set transformation. When either character set is null, assume identity
    transformation.
    \param from  source character set
    \param to    target character set */
void
Recoder::setTransformation(CharacterSet from, CharacterSet to)
{
    if (!from.isValid() || !to.isValid() || from==to) {
        setIdentityTransformation();
        return;
    }
    for (unsigned i = 0; i < 128; ++i)
        table[i] = i;
    for (unsigned i = 0; i < 128; ++i) {
        uint16 result = 0;
        uint16 fchar = from.data->chars[i];
        /* Basic idea: if characters match exactly, be happy. If
           character is accented and we don't have a match, remove the
           accent. Punt otherwise. Unlike GNU recode, this does not
           try to make a reversible transformation at all costs. */
        if (fchar < 128) {
            /* mapping to a low-ASCII character (happens with cp1252) */
            result = fchar;
        } else {
            /* try to find exact match. */
            for (unsigned j = 0; j < 128; ++j) {
                if (fchar == to.data->chars[j]) {
                    result = 128+j;
                    break;
                }
            }
            if (!result) {
                /* no exact match. Look up RFC mnemonic. Could
                   use binary search, but this isn't speed-critical. */
                const char* rfc = 0;
                for (unsigned j = 0; j < countof(uni_to_rfc); ++j) {
                    if (uni_to_rfc[j].code == fchar) {
                        rfc = uni_to_rfc[j].rfc;
                        if (isLetter(rfc[0]) && !isLetter(rfc[1]))
                            result = rfc[0];
                        break;
                    }
                }
            }
            if (!result)
                result = '?';
        }
        table[128+i] = result;
    }
}

/** Set identity transformation. Future recodeFoo() calls will not
    modify their input. */
void
Recoder::setIdentityTransformation()
{
    for (unsigned i = 0; i < TABLE_SIZE; ++i)
        table[i] = i;
}

/** Recode a string. \returns translated string */
string_t
Recoder::recodeString(string_t str) const
{
    for (string_t::size_type i = 0; i < str.length(); ++i)
        str[i] = table[(unsigned char) str[i]];
    return str;
}

/** Recode a char array. */
void
Recoder::recodeArray(char* data, std::size_t size) const
{
    for (std::size_t i = 0; i < size; ++i)
        data[i] = table[(unsigned char) data[i]];
}

/** Recode a C string. */
void
Recoder::recodeCString(char* data) const
{
    while (*data) {
        *data = table[(unsigned char) *data];
        ++data;
    }
}

/************************* Character Set Enquiry *************************/

/* I don't want to publish the CharacterSet::Data structure, so these
   trivial accessor functions must be here. This won't be a
   performance bottleneck, promised :) */

/** Get a character set by name.
    \returns character set, null if no match. */
CharacterSet
Recoder::getCharsetByName(const string_t& name)
{
    for (std::size_t n = 0; n < countof(::charsets); ++n)
        if (strCaseCmp(name, charsets[n].name) == 0)
            return CharacterSet(&charsets[n]);
    return CharacterSet(0);
}

/** Get number of supported character sets. */
std::size_t
Recoder::getNumCharsets()
{
    return countof(::charsets);
}

/** Get a character set by index.
    \pre n \in [0, getNumCharsets()) */
CharacterSet
Recoder::getCharset(std::size_t n)
{
    ASSERT(n < countof(::charsets));
    return CharacterSet(&charsets[n]);
}

/****************************** CharacterSet *****************************/

const char*
CharacterSet::getName() const
{
    return data ? data->name : "default";
}

const char*
CharacterSet::getDescription() const
{
    return data ? data->full_name : "default";
}

/** Return Unicode value of a character. */
uint16
CharacterSet::getUnicode(char c) const
{
    unsigned char uc = c;
    return (data && uc >= 128) ? data->chars[uc-128] : uc;
}

/** Return character corresponding to a Unicode character.
    \returns 0 if mapping fails. */
char
CharacterSet::getCharacter(uint16 uni) const
{
    if (uni < 128)
        return uni;
    if (!data && uni < 256)
        return uni;
    for (unsigned i = 0; i < 128; ++i)
        if (data->chars[i] == uni)
            return i+128;
    return 0;
}

/***************************** Global Recoder ****************************/

/** Set character sets.
    \param intrn internal character set (used within program)
    \param extrn external character set (used in files) */
void
Recoder::setCharsets(CharacterSet intrn, CharacterSet extrn)
{
    intern_cs = intrn;
    extern_cs = extrn;
    intern_to_extern.setTransformation(intrn, extrn);
    extern_to_intern.setTransformation(extrn, intrn);
}
